/*  These are functions and Methods used by GE2KAP_GDAL.DLL*/

/******************************************************************************
 *
 * Project:  GDAL Utilities
 * Purpose:  Command line application to list info about a file.
 * Author:   Frank Warmerdam, warmerdam@pobox.com
 *
 * ****************************************************************************
 * Copyright (c) 1998, Frank Warmerdam
 * Copyright (c) 2007-2015, Even Rouault <even.rouault at spatialys.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *****************************************************************************/


#include <windows.h>   
#include <CommCtrl.h>
#include <stdio.h>
#include <vector>
#include "oorexxapi.h"

/* added for OgrInfo 
// #include <cstring>
// #include <algorithm>
#include "ogr_api.h"
#include "ogr_core.h"
#include "ogr_feature.h"
#include "ogr_geometry.h"
#include "ogr_p.h"
// #include "ogr_spatialref.h"
#include "ogrsf_frmts.h"
#include <set>
*/

// #include "ogr_api.h"
// #include "ogr_core.h"
// #include "ogr_p.h"
// #include "ogrsf_frmts.h"


// #include <stdint.h>
// #include <cmath.h>

/* KEEP for ogrinfo
bool bVerbose = true;
bool bSuperQuiet = false;
bool bSummaryOnly = false;
GIntBig nFetchFID = OGRNullFID;
char** papszOptions = nullptr;
*/


// for Rgb2Pct
#include "common.h" 
// #if defined(_WIN32)
#define _CDECL __cdecl
// #else
// #define _CDECL    
// #endif


#define MAX_DIGITS     9  

static bool StringCISortFunction( const CPLString& a, const CPLString& b )
{
    return STRCASECMP(a.c_str(), b.c_str()) < 0;
} 

/************************************************************************/
/*                       GDALInfoPrintMetadata()                        */
/************************************************************************/
/* KEEP
static void GDALInfoPrintMetadata( GDALMajorObjectH hObject,
                                   const char *pszDomain,
                                   const char *pszDisplayedname,
                                   const char *pszIndent )
{
    bool bIsxml = false;

    if( pszDomain != nullptr && STARTS_WITH_CI(pszDomain, "xml:") )
        bIsxml = true;

    char **papszMetadata = GDALGetMetadata(hObject, pszDomain);
    if( CSLCount(papszMetadata) > 0 )
    {
        printf("%s%s:\n", pszIndent, pszDisplayedname);
        for( int i = 0; papszMetadata[i] != nullptr; i++ )
        {
            if( bIsxml )
                printf("%s%s\n", pszIndent, papszMetadata[i]);
            else
                printf("%s  %s\n", pszIndent, papszMetadata[i]);
        }
    }
}

/************************************************************************/
/*                       GDALInfoReportMetadata()                       */
/************************************************************************/
/* KEEP
static void GDALInfoReportMetadata( GDALMajorObjectH hObject,
                                    bool bListMDD,
                                    bool bShowMetadata,
                                    char **papszExtraMDDomains )
{
    const char* pszIndent = "";

   
    //      Report list of Metadata domains                                 
    
    if( bListMDD )
    {
        char** papszMDDList = GDALGetMetadataDomainList(hObject);
        char** papszIter = papszMDDList;

        if( papszMDDList != nullptr )
            printf("%sMetadata domains:\n", pszIndent);
        while( papszIter != nullptr && *papszIter != nullptr )
        {
            if( EQUAL(*papszIter, "") )
                printf("%s  (default)\n", pszIndent);
            else
                printf("%s  %s\n", pszIndent, *papszIter);
            papszIter ++;
        }
        CSLDestroy(papszMDDList);
    }

    if( !bShowMetadata )
        return;

    
    //      Report default Metadata domain.                                 
   
    GDALInfoPrintMetadata(hObject, nullptr, "Metadata", pszIndent);

    
    //      Report extra Metadata domains                                   
    
    if( papszExtraMDDomains != nullptr )
    {
        char **papszExtraMDDomainsExpanded = nullptr;

        if( EQUAL(papszExtraMDDomains[0], "all") &&
            papszExtraMDDomains[1] == nullptr )
        {
            char** papszMDDList = GDALGetMetadataDomainList(hObject);
            char** papszIter = papszMDDList;

            while( papszIter != nullptr && *papszIter != nullptr )
            {
                if( !EQUAL(*papszIter, "") )
                {
                    papszExtraMDDomainsExpanded = CSLAddString(papszExtraMDDomainsExpanded, *papszIter);
                }
                papszIter ++;
            }
            CSLDestroy(papszMDDList);
        }
        else
        {
            papszExtraMDDomainsExpanded = CSLDuplicate(papszExtraMDDomains);
        }

        for( int iMDD = 0;
             papszExtraMDDomainsExpanded != nullptr &&
               papszExtraMDDomainsExpanded[iMDD] != nullptr;
             iMDD++ )
        {
            char pszDisplayedname[256];
            snprintf(pszDisplayedname, 256, "Metadata (%s)",
                     papszExtraMDDomainsExpanded[iMDD]);
            GDALInfoPrintMetadata(hObject, papszExtraMDDomainsExpanded[iMDD],
                                  pszDisplayedname, pszIndent);
        }

        CSLDestroy(papszExtraMDDomainsExpanded);
    }
}
*/
/************************************************************************/
/*                           ReportOnLayer()                            */
/************************************************************************/
/* KEEP
static void ReportOnLayer( OGRLayer * poLayer, const char *pszWHERE,
                           const char* pszGeomField,
                           OGRGeometry *poSpatialFilter,
                           bool bListMDD,
                           bool bShowMetadata,
                           char** papszExtraMDDomains,
                           bool bFeatureCount,
                           bool bExtent )
{
    OGRFeatureDefn      *poDefn = poLayer->GetLayerDefn();


   //     Set filters if provided.                                        

    if( pszWHERE != nullptr )
    {
        if( poLayer->SetAttributeFilter(pszWHERE) != OGRERR_NONE )
        {
            printf("FAILURE: SetAttributeFilter(%s) failed.\n", pszWHERE);
            exit(1);
        }
    }

    if( poSpatialFilter != nullptr )
    {
        if( pszGeomField != nullptr )
        {
            const int iGeomField = poDefn->GetGeomFieldIndex(pszGeomField);
            if( iGeomField >= 0 )
                poLayer->SetSpatialFilter(iGeomField, poSpatialFilter);
            else
                printf("WARNING: Cannot find geometry field %s.\n",
                       pszGeomField);
        }
        else
        {
            poLayer->SetSpatialFilter(poSpatialFilter);
        }
    }


   //      Report various overall information.                             

    if( !bSuperQuiet )
    {
        printf("\n");
        printf("Layer name: %s\n", poLayer->GetName());
    }

    GDALInfoReportMetadata(static_cast<GDALMajorObjectH>(poLayer),
                           bListMDD,
                           bShowMetadata,
                           papszExtraMDDomains);

    if( bVerbose )
    {
        const int nGeomFieldCount =
            poLayer->GetLayerDefn()->GetGeomFieldCount();
        if( nGeomFieldCount > 1 )
        {
            for(int iGeom = 0;iGeom < nGeomFieldCount; iGeom ++ )
            {
                OGRGeomFieldDefn* poGFldDefn =
                    poLayer->GetLayerDefn()->GetGeomFieldDefn(iGeom);
                printf("Geometry (%s): %s\n", poGFldDefn->GetNameRef(),
                       OGRGeometryTypeToName(poGFldDefn->GetType()));
            }
        }
        else
        {
            printf("Geometry: %s\n",
                   OGRGeometryTypeToName(poLayer->GetGeomType()));
        }

        if( bFeatureCount )
            printf("Feature Count: " CPL_FRMT_GIB "\n",
                   poLayer->GetFeatureCount());

        OGREnvelope oExt;
        if( bExtent && nGeomFieldCount > 1 )
        {
            for( int iGeom = 0;iGeom < nGeomFieldCount; iGeom ++ )
            {
                if( poLayer->GetExtent(iGeom, &oExt, TRUE) == OGRERR_NONE )
                {
                    OGRGeomFieldDefn* poGFldDefn =
                        poLayer->GetLayerDefn()->GetGeomFieldDefn(iGeom);
                    CPLprintf("Extent (%s): (%f, %f) - (%f, %f)\n",
                              poGFldDefn->GetNameRef(),
                              oExt.MinX, oExt.MinY, oExt.MaxX, oExt.MaxY);
                }
            }
        }
        else if( bExtent && poLayer->GetExtent(&oExt, TRUE) == OGRERR_NONE )
        {
            CPLprintf("Extent: (%f, %f) - (%f, %f)\n",
                      oExt.MinX, oExt.MinY, oExt.MaxX, oExt.MaxY);
        }

        if( nGeomFieldCount > 1 )
        {
            for(int iGeom = 0;iGeom < nGeomFieldCount; iGeom ++ )
            {
                OGRGeomFieldDefn* poGFldDefn =
                    poLayer->GetLayerDefn()->GetGeomFieldDefn(iGeom);
                OGRSpatialReference* poSRS = poGFldDefn->GetSpatialRef();
                char *pszWKT = nullptr;
                if( poSRS == nullptr )
                {
                    pszWKT = CPLStrdup("(unknown)");
                }
                else
                {
                    poSRS->exportToPrettyWkt(&pszWKT);
                }

                printf("SRS WKT (%s):\n%s\n",
                       poGFldDefn->GetNameRef(), pszWKT);
                CPLFree(pszWKT);
            }
        }
        else
        {
            char *pszWKT = nullptr;
            if( poLayer->GetSpatialRef() == nullptr )
            {
                pszWKT = CPLStrdup("(unknown)");
            }
            else
            {
                poLayer->GetSpatialRef()->exportToPrettyWkt(&pszWKT);
            }

            printf("Layer SRS WKT:\n%s\n", pszWKT);
            CPLFree(pszWKT);
        }

        if( strlen(poLayer->GetFIDColumn()) > 0 )
            printf("FID Column = %s\n",
                   poLayer->GetFIDColumn());

        for(int iGeom = 0;iGeom < nGeomFieldCount; iGeom ++ )
        {
            OGRGeomFieldDefn* poGFldDefn =
                poLayer->GetLayerDefn()->GetGeomFieldDefn(iGeom);
            if( nGeomFieldCount == 1 &&
                EQUAL(poGFldDefn->GetNameRef(), "") &&
                poGFldDefn->IsNullable() )
                break;
            printf("Geometry Column ");
            if( nGeomFieldCount > 1 )
                printf("%d ", iGeom + 1);
            if( !poGFldDefn->IsNullable() )
                printf("NOT NULL ");
            printf("= %s\n", poGFldDefn->GetNameRef());
        }

        for( int iAttr = 0; iAttr < poDefn->GetFieldCount(); iAttr++ )
        {
            OGRFieldDefn *poField = poDefn->GetFieldDefn(iAttr);
            const char* pszType = (poField->GetSubType() != OFSTNone)
                ? CPLSPrintf(
                      "%s(%s)",
                      poField->GetFieldTypeName(poField->GetType()),
                      poField->GetFieldSubTypeName(poField->GetSubType()))
                : poField->GetFieldTypeName(poField->GetType());
            printf("%s: %s (%d.%d)",
                   poField->GetNameRef(),
                   pszType,
                   poField->GetWidth(),
                   poField->GetPrecision());
            if( !poField->IsNullable() )
                printf(" NOT NULL");
            if( poField->GetDefault() != nullptr )
                printf(" DEFAULT %s", poField->GetDefault());
            printf("\n");
        }
    }


   //      Read, and dump features.                                        

    if( nFetchFID == OGRNullFID && !bSummaryOnly )
    {
        OGRFeature *poFeature = nullptr;
        while( (poFeature = poLayer->GetNextFeature()) != nullptr )
        {
            if( !bSuperQuiet )
                poFeature->DumpReadable(nullptr, papszOptions);
            OGRFeature::DestroyFeature(poFeature);
        }
    }
    else if( nFetchFID != OGRNullFID )
    {
        OGRFeature *poFeature = poLayer->GetFeature(nFetchFID);
        if( poFeature == nullptr )
        {
            printf("Unable to locate feature id " CPL_FRMT_GIB
                   " on this layer.\n",
                   nFetchFID);
        }
        else
        {
            poFeature->DumpReadable(nullptr, papszOptions);
            OGRFeature::DestroyFeature(poFeature);
        }
    }
}
*/
/************************************************************************/
/*                             RemoveBOM()                              */
/************************************************************************/
/* KEEP
// Remove potential UTF-8 BOM from data (must be NUL terminated) 
static void RemoveBOM(GByte* pabyData)
{
    if( pabyData[0] == 0xEF && pabyData[1] == 0xBB && pabyData[2] == 0xBF )
    {
        memmove(pabyData, pabyData + 3,
                strlen(reinterpret_cast<char *>(pabyData) + 3) + 1);
    }
}

static void RemoveSQLComments(char*& pszSQL)
{
    char** papszLines = CSLTokenizeStringComplex(pszSQL, "\r\n", FALSE, FALSE);
    CPLString osSQL;
    for( char** papszIter = papszLines; papszIter && *papszIter; ++papszIter )
    {
        if( !STARTS_WITH(*papszIter, "--") )
        {
            osSQL += *papszIter;
            osSQL += " ";
        }
    }
    CSLDestroy(papszLines);
    CPLFree(pszSQL);
    pszSQL = CPLStrdup(osSQL);
} 
*/
/************************************************************************/
/*                       GDALTranslateOptionsForBinaryNew()             */
/************************************************************************/

static GDALTranslateOptionsForBinary *GDALTranslateOptionsForBinaryNew(void)
{
    return static_cast<GDALTranslateOptionsForBinary *>(
        CPLCalloc(1, sizeof(GDALTranslateOptionsForBinary)));
}  



/************************************************************************/
/*                       GDALTranslateOptionsForBinaryFree()            */
/************************************************************************/

static void GDALTranslateOptionsForBinaryFree( GDALTranslateOptionsForBinary* psOptionsForBinary )
{
    if( psOptionsForBinary == nullptr )
        return;

    CPLFree(psOptionsForBinary->pszSource);
    CPLFree(psOptionsForBinary->pszDest);
    CSLDestroy(psOptionsForBinary->papszOpenOptions);
    CPLFree(psOptionsForBinary->pszFormat);
    CPLFree(psOptionsForBinary);
}

/************************************************************************/
/*                 GDALVectorTranslateOptionsForBinaryNew()             */
/************************************************************************/

static GDALVectorTranslateOptionsForBinary *
GDALVectorTranslateOptionsForBinaryNew()
{
    return static_cast<GDALVectorTranslateOptionsForBinary *>(
        CPLCalloc(1, sizeof(GDALVectorTranslateOptionsForBinary)));
}

/************************************************************************/
/*                  GDALVectorTranslateOptionsForBinaryFree()           */
/************************************************************************/

static void GDALVectorTranslateOptionsForBinaryFree(
    GDALVectorTranslateOptionsForBinary* psOptionsForBinary )
{
    if( psOptionsForBinary )
    {
        CPLFree(psOptionsForBinary->pszDataSource);
        CPLFree(psOptionsForBinary->pszDestDataSource);
        CSLDestroy(psOptionsForBinary->papszOpenOptions);
        CPLFree(psOptionsForBinary->pszFormat);
        CPLFree(psOptionsForBinary);
    }
}

/************************************************************************/
/*                         GDALInfoOptionsForBinary()                   */
/************************************************************************/
static GDALInfoOptionsForBinary *GDALInfoOptionsForBinaryNew(void)
{
    return static_cast<GDALInfoOptionsForBinary *>(
        CPLCalloc(1, sizeof(GDALInfoOptionsForBinary)));
}
/************************************************************************/
/*                       GDALInfoOptionsForBinaryFree()                 */
/************************************************************************/
static void GDALInfoOptionsForBinaryFree( GDALInfoOptionsForBinary* psOptionsForBinary )
{
    if( psOptionsForBinary )
    {
        CPLFree(psOptionsForBinary->pszFilename);
        CSLDestroy(psOptionsForBinary->papszOpenOptions);
        CPLFree(psOptionsForBinary);
    }
}

struct ScaledBand   // for pansharpen
{
	ScaledBand() :
		oversample(0), lo_w(0), lo_h(0), hi_w(0), hi_h(0),
		delta_x(0), delta_y(0), band(NULL), line_buf_idx(0)
	{ }

	int oversample;
	size_t lo_w, lo_h;
	size_t hi_w, hi_h;
	int delta_x, delta_y;
	std::vector<std::vector<double> > kernel_x;
	std::vector<std::vector<double> > kernel_y;
	GDALRasterBandH band;
	std::vector<std::vector<double> > lines_buf;
	int line_buf_idx;
};

/****************************** GdalInfo *********************************************
 calls the GdalInfo function
*******************************************************************************************/ 

RexxRoutine5(int, GdalInfo, CSTRING, sourceFile, CSTRING, options, int, diag, RexxArrayObject, diagmsgs, RexxArrayObject, output)
{
   char buffer[2000]; 
   int rtnrc = 0;
   int rc;
   // DEBUG EarlySetConfigOptions(argc, argv);
   
   // Create a buffer for the output
   unsigned int size_buf = 20000;
   char * out_buf = (char*)malloc(size_buf);
   if (out_buf==NULL)
   {
      sprintf(buffer,"ERROR: Unable to allocate out_buf of 20,000 bytes");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
     return 1;
   }   
   
   GDALAllRegister();
   
   // if there is a source file
   if (strlen(sourceFile) > 0)
   {
      sprintf(buffer,"DIAG GdalInfo sourceFile=%s",sourceFile);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   }
   else
   {
       sprintf(buffer,"DIAG GdalInfo no sourceFile");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   }
   
   sprintf(buffer,"DIAG GdalInfo options=%s",options);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   // Create a dummy argv
   std::vector<std::string> args; 
   std::vector<char*> argv; 
   // Scan the options and add to argv
   int buf_offset=0;  // buffer offset 
   int i=0;
   int len = lstrlen(options);
   sprintf(buffer,"DIAG GdalInfo options len=%d",len);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   // If there are any options
   if (len > 0)
   {
      // Scan the options and add each word to the argv 
      byte thischar;
      for(int m=0;m <= len ;m++)
      {  
         thischar = options[m] ;
         if (thischar ==  0x20 || m == len) // if we found a blank  or end of options
         {
            buffer[buf_offset] =  '\0'; // put in the string delimiter
            args.push_back(buffer);
            buf_offset = 0;
            // m++;
         }
         else 
         {
            // move the char into the buffer
            
            // if(thischar == 63) thischar = 32;  // change any ? to blanks
            buffer[buf_offset] = thischar;
            buf_offset++;  
         }    
      }
      // Add all the args to argv
      for (const auto& arg : args)
          argv.push_back((char*)arg.data());
   }      
   // add a NULL entry at end    
   argv.push_back(nullptr);
   
   int argc = argv.size();
   sprintf(buffer,"DIAG GdalInfo argc=%d",argc);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 

   GDALAllRegister();

   // Scan the options looking for special ones
   for (int i = 0; i < argc; i++)
   {
      if(argv[i] == NULL);
      else
      { 
         if (EQUAL(argv[i], "--version"))
         {
            sprintf(out_buf,"%s",GDALVersionInfo("RELEASE_NAME"));
            rc = context->ArrayAppendString(output, out_buf, strlen(out_buf));  
           
         return 0;
         }
      }         
   }    

   GDALInfoOptionsForBinary* psOptionsForBinary = GDALInfoOptionsForBinaryNew();

   GDALInfoOptions *psOptions
     = GDALInfoOptionsNew(argv.data(), psOptionsForBinary);

   
   //      Open dataset.                                                   
   

   char** papszOpenOptions=nullptr;
   GDALDatasetH hDataset;
         
   hDataset = GDALOpenEx(sourceFile, GDAL_OF_RASTER | GDAL_OF_VERBOSE_ERROR, NULL,
     (const char* const*)papszOpenOptions, NULL);            

    if( hDataset == nullptr )
    {

      sprintf(buffer,"ERROR: GdalInfo failed - unable to open %s",sourceFile);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;
   }

   
   //     Read specified subdataset if requested.                        
   
   if ( psOptionsForBinary->nSubdataset > 0 )
   {
     char **papszSubdatasets = GDALGetMetadata( hDataset, "SUBDATASETS" );
     int nSubdatasets = CSLCount( papszSubdatasets );
   
     if ( nSubdatasets > 0 && psOptionsForBinary->nSubdataset <= nSubdatasets )
     {
         char szKeyName[1024];
         char *pszSubdatasetName;
   
         snprintf( szKeyName, sizeof(szKeyName),
                   "SUBDATASET_%d_NAME", psOptionsForBinary->nSubdataset );
         szKeyName[sizeof(szKeyName) - 1] = '\0';
         pszSubdatasetName =
             CPLStrdup( CSLFetchNameValue( papszSubdatasets, szKeyName ) );
         GDALClose( hDataset );
         hDataset = GDALOpen( pszSubdatasetName, GA_ReadOnly );
         CPLFree( pszSubdatasetName );
     }
     else
     {
         sprintf(buffer,"DIAG GdalInfo warning: subdataset %d of %d requested. Reading the main dataset.",
                  psOptionsForBinary->nSubdataset, nSubdatasets);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
                 
     }
   }

   char* pszGDALInfoOutput = GDALInfo( hDataset, psOptions );
   
   // Check the size of  pszGDALInfoOutput
   if (strlen(pszGDALInfoOutput) > size_buf)
   {
      sprintf(buffer,"ERROR: GdalInfo pszGDALInfoOutput size:%d is too large",strlen(pszGDALInfoOutput));
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;
   }
    
   if( pszGDALInfoOutput )
   {
      // printf("DEBUG length of pszGDALInfoOutput=%d\n",strlen(pszGDALInfoOutput));
      
      sprintf(out_buf,"%s",pszGDALInfoOutput);
      rc = context->ArrayAppendString(output, out_buf, strlen(out_buf));  
   }   
   
   CPLFree( pszGDALInfoOutput );

   GDALClose( hDataset );

   GDALInfoOptionsForBinaryFree(psOptionsForBinary);
   
   GDALInfoOptionsFree( psOptions );
   
   GDALDumpOpenDatasets( stderr );
   
   //  DEBUG GDALDestroyDriverManager();
   
   CPLDumpSharedList( nullptr );
   CPLCleanupTLS();
   
   if (out_buf) free(out_buf); 
   
   return 0;
}

/****************************** Gdal_Translate *****************************************************************************
 calls the Gdal_translate function  from: C:\release-1911-gdal-2-4-1-mapserver-7-2-2-src\gdal\apps\gdal_translate_bin.cpp
 
 called:  rc = Gdal_Translate(sourceFile,dstFile,options,0,diagmsgs)
 
***************************************************************************************************************************/ 
RexxRoutine5(int, Gdal_Translate, CSTRING, sourceFile, CSTRING, dstFile, CSTRING, options, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[2048]; 
   int rtnrc = 0;
   int rc;
   // logical_t flag;
   // RexxObjectPtr dataobj;
   
   GDALAllRegister();
   
   if (sourceFile == NULL)
   {
      sprintf(buffer,"ERROR: Gdal_Translate no source file specified");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;
   }

   if (dstFile == NULL)
   {
      sprintf(buffer,"ERROR: Gdal_Translate no destination file specified");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;
   }
   
   sprintf(buffer,"DIAG Gdal_Translate sourceFile=%s",sourceFile);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   sprintf(buffer,"DIAG Gdal_Translate dstFile=%s ",dstFile);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   sprintf(buffer,"DIAG Gdal_Translate options=%s",options);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   // Create a dummy argv
   std::vector<std::string> args; 
   std::vector<char*> argv; 
   // Scan the options and add to argv
   int buf_offset=0;  // buffer offset 
   int i=0;
   int len = lstrlen(options);
   sprintf(buffer,"DIAG Gdal_Translate options len=%d",len);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   // If there are any options
   if (len > 0)
   {
      // Scan the options and add each word to the argv 
      byte thischar;
      for(int m=0;m <= len ;m++)
      {  
         thischar = options[m] ;
         if (thischar ==  0x20 || m == len) // if we found a blank  or end of options
         {
            buffer[buf_offset] =  '\0'; // put in the string delimiter
            args.push_back(buffer);
            buf_offset = 0;
            // m++;
         }
         else 
         {
            // move the char into the buffer
            
            // if(thischar == 63) thischar = 32;  // change any ? to blanks
            buffer[buf_offset] = thischar;
            buf_offset++;  
         }    
      }
      // Add all the args to argv
      for (const auto& arg : args)
          argv.push_back((char*)arg.data());
   }      
   // add a NULL entry at end    
   argv.push_back(nullptr);
   
   int argc = argv.size();
   sprintf(buffer,"DIAG Gdal_Translate argc=%d",argc);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   if (CPLGetConfigOption("GDAL_MAX_DATASET_POOL_SIZE", NULL) == NULL)
   {
     CPLSetConfigOption("GDAL_MAX_DATASET_POOL_SIZE", "450");
   }
   
   char** papszOpenOptions=nullptr;
   
   // Scan the options looking for any open options and if found NULL them
   int j,k;
   for (int i = 0; i < argc; i++)
   {
      if(argv[i] == NULL);
      else
      { 
         if (EQUAL(argv[i], "--config") && i < argc - 2)
         {
            j = i +1;
            k = i +2;
            papszOpenOptions = CSLAddNameValue(papszOpenOptions, argv[j],argv[k]);
            argv[i] = NULL;
            argv[j] = NULL;
            argv[k] = NULL;
         }
      }
   } 
   
   GDALTranslateOptionsForBinary* psOptionsForBinary = GDALTranslateOptionsForBinaryNew();
   GDALTranslateOptions *psOptions = GDALTranslateOptionsNew(argv.data(), psOptionsForBinary);
   // CSLDestroy(argv.data());
   
   
   if (!(psOptionsForBinary->bQuiet))
   {
     GDALTranslateOptionsSetProgress(psOptions, GDALTermProgress, NULL);
   }
  
   GDALDatasetH hDataset;
   int bUsageError;

   // Attempt to open the source using anyopen options
   /*      
   hDataset = GDALOpenEx(sourceFile, GDAL_OF_RASTER | GDAL_OF_VERBOSE_ERROR, NULL,
     (const char* const*)papszOpenOptions, NULL);
   */  
   hDataset = GDALOpenEx(sourceFile, GDAL_OF_RASTER | GDAL_OF_VERBOSE_ERROR, NULL,
     papszOpenOptions, NULL);  
   if (hDataset == NULL)
   {
      const char* errmsg = CPLGetLastErrorMsg();
      sprintf(buffer,"ERROR:%s\n",errmsg);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;
   }
   
   CSLDestroy( papszOpenOptions ); 
   
   bool bQuiet = false, bCopySubDatasets = false, bFormatExplicitlySet = false;
   char* pszFormat = NULL;

   // Scan the options looking for special ones
   for (int i = 0; i < argc; i++)
   {
      if(argv[i] == NULL);
      else
      { 
         if (EQUAL(argv[i], "-of") && i < argc - 1)
         {
            ++i;
            pszFormat = CPLStrdup(argv[i]);
            bFormatExplicitlySet = TRUE;
         }
         else if (EQUAL(argv[i], "-q") || EQUAL(argv[i], "-quiet"))
         {
            bQuiet = TRUE;
         }
         else if (EQUAL(argv[i], "-sds"))
         {
            bCopySubDatasets = TRUE;
         }       
         else if (EQUAL(argv[i], "-oo") && i + 1 < argc)
         {
            i++;
            papszOpenOptions = CSLAddString(papszOpenOptions, argv[i]);
         }
      }
   } 
   
   GDALDatasetH hOutDS;   
   
   // GDALTranslateOptions *psOptions = GDALTranslateOptionsNew(argv.data(), NULL);
   // CSLDestroy(argv.data());
     
   if (!bQuiet)
   {
     GDALTranslateOptionsSetProgress(psOptions, GDALTermProgress, NULL);
   }   

   /* -------------------------------------------------------------------- */
   /*      Handle subdatasets.                                             */
   /* -------------------------------------------------------------------- */
   if (!bCopySubDatasets
     && CSLCount(GDALGetMetadata(hDataset, "SUBDATASETS")) > 0
     && GDALGetRasterCount(hDataset) == 0)
   {
      fprintf(stderr,
         "Input file contains subdatasets. Please, select one of them for reading.\n");
      const char* errmsg = CPLGetLastErrorMsg();
      sprintf(buffer,"ERROR:%s\n",errmsg);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 2;
   }
   
   if (CSLCount(GDALGetMetadata(hDataset, "SUBDATASETS")) > 0
     && bCopySubDatasets)
   {
     char **papszSubdatasets = GDALGetMetadata(hDataset, "SUBDATASETS");
     char *pszSubDest = (char *)CPLMalloc(strlen(dstFile) + 32);
     int i;
   
     CPLString osPath = CPLGetPath(dstFile);
     CPLString osBasename = CPLGetBasename(dstFile);
     CPLString osExtension = CPLGetExtension(dstFile);
     CPLString osTemp;
   
     const char* pszFormat = NULL;
     if (CSLCount(papszSubdatasets) / 2 < 10)
     {
         pszFormat = "%s_%d";
     }
     else if (CSLCount(papszSubdatasets) / 2 < 100)
     {
         pszFormat = "%s_%002d";
     }
     else
     {
         pszFormat = "%s_%003d";
     }
   
     const char* pszDest = pszSubDest;
   
     for (i = 0; papszSubdatasets[i] != NULL; i += 2)
     {
         char* pszSource = CPLStrdup(strstr(papszSubdatasets[i], "=") + 1);
         osTemp = CPLSPrintf(pszFormat, osBasename.c_str(), i / 2 + 1);
         osTemp = CPLFormFilename(osPath, osTemp, osExtension);
         strcpy(pszSubDest, osTemp.c_str());
         hDataset = GDALOpenEx(pszSource, GDAL_OF_RASTER, NULL,
             (const char* const*)papszOpenOptions, NULL);
         CPLFree(pszSource);
         if (!bQuiet) 
          {
            sprintf(buffer,"DIAG Gdal_Translate 2 source file size is %d, %d",GDALGetRasterXSize(hDataset), GDALGetRasterYSize(hDataset));
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         }
         hOutDS = GDALTranslate(pszDest, hDataset, psOptions, &bUsageError);
         if (bUsageError == TRUE)
         {
            const char* errmsg = CPLGetLastErrorMsg();
            sprintf(buffer,"ERROR:%s\n",errmsg);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            return 3;
         }    
         if (hOutDS == NULL)
             break;
         GDALClose(hOutDS);
     }
   
     GDALTranslateOptionsFree(psOptions);
   
     CPLFree(pszSubDest);
   
      const char* errmsg = CPLGetLastErrorMsg();
      sprintf(buffer,"ERROR:%s\n",errmsg);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 4;
   }

   if (!bQuiet)
   {
      sprintf(buffer,"DIAG Gdal_Translate source file size is %d, %d",GDALGetRasterXSize(hDataset), GDALGetRasterYSize(hDataset));
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   } 
                                 
   // Issue the translate
   //  hOutDS = GDALTranslate(psOptionsForBinary->pszDest, hDataset, psOptions, &bUsageError);
   hOutDS = GDALTranslate(dstFile, hDataset, psOptions, &bUsageError);
   if (bUsageError == TRUE)
   {
   GDALClose(hOutDS);
   const char* errmsg = CPLGetLastErrorMsg();
   sprintf(buffer,"ERROR:%s\n",errmsg);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   return 5;
   }
   
   GDALFlushCache( hOutDS );     
   GDALClose(hOutDS);
   hOutDS = nullptr;
   
   /* Close hOutDS before hDataset for the -f VRT case */
   GDALClose(hDataset);
   GDALTranslateOptionsFree(psOptions);
   GDALTranslateOptionsForBinaryFree(psOptionsForBinary);
    
   return 0;
}

/****************************** ogr2ogr *****************************************************************************
 calls the ogr2ogr function  from: C:\release-1911-gdal-2-4-1-mapserver-7-2-2-src\gdal\apps\ogr2ogr_bin.cpp
 called:  rc = ogr2ogr(sourceFile,destFile,options,0,diagmsgs)
***************************************************************************************************************************/ 
RexxRoutine5(int, ogr2ogr, CSTRING, sourceFile, CSTRING, destFile, CSTRING, options, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[2048]; 
   int rtnrc = 0;
   int rc;
   GDALDatasetH hDS = nullptr;
   GDALDatasetH hODS = nullptr;
   bool bCloseODS = true;
   GDALDatasetH hDstDS = nullptr;
   
   /* -------------------------------------------------------------------- */
   /*      Register format(s).                                             */
   /* -------------------------------------------------------------------- */
   OGRRegisterAll();
   
   /* KEEP if uncommented need the 
   #include <cstring>
   #include <algorithm>
   #include "ogr_api.h"
   #include "ogr_core.h"
   #include "ogr_p.h"
   #include "ogrsf_frmts.h"
   
   std::vector<CPLString> aoSetDrivers;
   OGRSFDriverRegistrar *poR = OGRSFDriverRegistrar::GetRegistrar();
   for( int iDriver = 0; iDriver < poR->GetDriverCount(); iDriver++ )
    {
        GDALDriver *poDriver = poR->GetDriver(iDriver);

        if( CPLTestBool( CSLFetchNameValueDef(poDriver->GetMetadata(),
                                              GDAL_DCAP_CREATE, "FALSE")) )
            aoSetDrivers.push_back(poDriver->GetDescription());      
    }
    
    std::sort(aoSetDrivers.begin(), aoSetDrivers.end(), StringCISortFunction);
    for( const auto &oDriver : aoSetDrivers )
    {
        printf("     -f \"%s\"\n", oDriver.c_str());
    }
    */
   
   GDALVectorTranslateOptionsForBinary* psOptionsForBinary = nullptr;
   GDALVectorTranslateOptions *psOptions = nullptr;
      
   if (sourceFile == NULL)
   {
      sprintf(buffer,"ERROR: ogr2ogr no source file specified");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;
   }

   if (destFile == NULL)
   {
      sprintf(buffer,"ERROR: ogr2ogr no destination file specified");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;
   }
   
   sprintf(buffer,"DIAG ogr2ogr sourceFile=%s",sourceFile);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   sprintf(buffer,"DIAG ogr2ogr destFile=%s ",destFile);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   sprintf(buffer,"DIAG ogr2ogr options=%s",options);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   printf("sourceFile=%s destFile=%s options=%s\n",sourceFile,destFile,options);
   
   // Create a dummy argv
   std::vector<std::string> args; 
   std::vector<char*> argv; 
   // Scan the options and add to argv
   int buf_offset=0;  // buffer offset 
   int i=0;
   int len = lstrlen(options);
   sprintf(buffer,"DIAG ogr2ogr options len=%d",len);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   // If there are any options
   if (len > 0)
   {
      // Scan the options and add each parameter to the argv 
      byte thischar,next_char;
      bool skip = false;  // skip to next double quote added in V10.0.9.2 
      for(int m=0;m <= len ;m++)
      {  
         thischar = options[m];
         if (m < len) next_char =  options[m+1];   // get the next char
         else  next_char = NULL;
         if(thischar == 0x22 && next_char == 0x22)  ;// if thischar is a double quote and the next_char is too do nothing
         else 
         {
            if ((thischar == 0x20 || m == len) && !skip) // if we found a blank  or end of options and not skip
            {
               buffer[buf_offset] =  '\0'; // put in the string delimiter
               if(strlen(buffer) == 0) ;  // if the length of the buffer is zero do nothing
               else  // else add the parameter
               {
                  args.push_back(buffer);  // add to the args
                  buf_offset = 0;
               }
            }
            else 
            {
               if(thischar == 0x22) // if this is a double quote its a parameter enclosed in double quotes
               {
                  if(!skip) skip = true; // if skip is not set its the first double quote wait for next one
                  else  // else this is the second double quote make this the end of the parameter
                  {
                     buffer[buf_offset] =  '\0'; // put in the string delimiter
                     args.push_back(buffer);   // add to the args
                     buf_offset = 0;
                     skip = false;   // turn off skipping
                  }
               }
               else
               { 
                  // move the char into the buffer
                  buffer[buf_offset] = thischar;
                  buf_offset++;  
               }
            
            }    
         }
      }
      // Add all the args to argv
      for (const auto& arg : args)
          argv.push_back((char*)arg.data());
   }      
   // add a NULL entry at end    
   argv.push_back(nullptr);
   
   int argc = argv.size();
    
   sprintf(buffer,"DIAG ogr2ogr argc=%d",argc);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   char** papszOpenOptions=nullptr;
   
   // Print all the options 
   // int j,k;
   for (int i = 0; i < argc; i++)
   {
      sprintf(buffer,"DIAG ogr2ogr argv[%d]=%s\n",i,argv[i]);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      // printf("DEBUG ARG%d=%s\n",i,argv[i]);
   } 
   
   psOptionsForBinary = GDALVectorTranslateOptionsForBinaryNew();
   psOptions = GDALVectorTranslateOptionsNew(argv.data(), psOptionsForBinary);
   
   // Avoid opening twice the same datasource if it is both the input and
    // output Known to cause problems with at least FGdb, SQlite and GPKG
    // drivers. See #4270
   if (psOptionsForBinary->eAccessMode != ACCESS_CREATION &&
         strcmp(sourceFile,
               destFile) == 0)
   {
      /* Open the source file*/
      // printf("DEBUG 1 opening sourceFile=%s\n",sourceFile);
      hODS = GDALOpenEx(
               // psOptionsForBinary->pszDataSource,       
               sourceFile,
               GDAL_OF_UPDATE | GDAL_OF_VECTOR, nullptr,
               psOptionsForBinary->papszOpenOptions, nullptr);
               
      GDALDriverH hDriver =
               hODS != nullptr ? GDALGetDatasetDriver(hODS) : nullptr;
               
      // Restrict to those 3 drivers. For example it is known to break with
           // the PG driver due to the way it manages transactions.
           if( hDriver && !(EQUAL(GDALGetDescription(hDriver), "FileGDB") ||
                            EQUAL(GDALGetDescription(hDriver), "SQLite") ||
                            EQUAL(GDALGetDescription(hDriver), "GPKG")) )
           {
               // printf("DEBUG 2  opening sourceFile=%s\n",sourceFile);
               hDS = GDALOpenEx(sourceFile,
                                GDAL_OF_VECTOR, nullptr,
                                psOptionsForBinary->papszOpenOptions, nullptr);
           }
           else
           {
               hDS = hODS;
               bCloseODS = false;
           }
   }
   else
   {
      // printf("DEBUG 3 opening sourceFile=%s\n",sourceFile);
      hDS = GDALOpenEx(sourceFile,
                         GDAL_OF_VECTOR, nullptr,
                         psOptionsForBinary->papszOpenOptions, nullptr);
   }
   
   /* -------------------------------------------------------------------- */
   /*      Report failure                                                  */
   /* -------------------------------------------------------------------- */
   if( hDS == nullptr )
   {
      sprintf(buffer,"ERROR:ogr2ogr Failed to open Source file=%s\n",sourceFile);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;
   } 
   
   printf("DEBUGR calling GDALVectorTranslate\n");
   int bUsageError = FALSE;
   hDstDS = GDALVectorTranslate(destFile, hODS, 1, &hDS, psOptions, &bUsageError);
   if( bUsageError )
   {
      sprintf(buffer,"ERROR:ogr2ogr GDALVectorTranslate Failed Source file=%s\n",sourceFile);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      printf("DEBUGR GDALVectorTranslate Failed Source file=%s\n",sourceFile);
      return 1;
   }
   
   printf("DEBUGR calling GDALVectorTranslateOptionsFree\n");
   
   GDALVectorTranslateOptionsFree(psOptions);
   GDALVectorTranslateOptionsForBinaryFree(psOptionsForBinary);
   
   if( hDS )
     GDALClose(hDS);
   if( bCloseODS )
     GDALClose(hDstDS);
     
   GDALDestroyDriverManager();  
    
   OGRCleanupAll();
   
   return 0;
}

/****************************** ogrinfo *****************************************************************************
 calls the ogr2ogr function  from: C:\release-1911-gdal-2-4-1-mapserver-7-2-2-src\gdal\apps\ogrinfo.cpp
 called:  rc = ogr2ogr(sourceFile,options,0,diagmsgs)
***************************************************************************************************************************/ 
/* KEEP
RexxRoutine4(int, ogrinfo, CSTRING, sourceFile, CSTRING, options, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[2048]; 
   int rtnrc = 0;
   int rc;
   
   if (sourceFile == NULL)
   {
      sprintf(buffer,"ERROR: ogrinfo no source file specified");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;
   }
   
   sprintf(buffer,"DIAG ogrinfo sourceFile=%s",sourceFile);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   sprintf(buffer,"DIAG ogrinfo options=%s",options);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   // printf("DEBUG sourceFile=%s\n",sourceFile);
   
   OGRRegisterAll();
   
   // Create a dummy argv
   std::vector<std::string> args; 
   std::vector<char*> argv; 
   // Scan the options and add to argv
   int buf_offset=0;  // buffer offset 
   int i=0;
   int len = lstrlen(options);
   sprintf(buffer,"DIAG ogrinfo options len=%d",len);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   // If there are any options
   if (len > 0)
   {
      // Scan the options and add each parameter to the argv 
      byte thischar,next_char;
      bool skip = false;  // skip to next double quote added in V10.0.9.2 
      for(int m=0;m <= len ;m++)
      {  
         thischar = options[m];
         if (m < len) next_char =  options[m+1];   // get the next char
         else  next_char = NULL;
         if(thischar == 0x22 && next_char == 0x22)  ;// if thischar is a double quote and the next_char is too do nothing
         else 
         {
            if ((thischar == 0x20 || m == len) && !skip) // if we found a blank  or end of options and not skip
            {
               buffer[buf_offset] =  '\0'; // put in the string delimiter
               if(strlen(buffer) == 0) ;  // if the length of the buffer is zero do nothing
               else  // else add the parameter
               {
                  args.push_back(buffer);  // add to the args
                  buf_offset = 0;
               }
            }
            else 
            {
               if(thischar == 0x22) // if this is a double quote its a parameter enclosed in double quotes
               {
                  if(!skip) skip = true; // if skip is not set its the first double quote wait for next one
                  else  // else this is the second double quote make this the end of the parameter
                  {
                     buffer[buf_offset] =  '\0'; // put in the string delimiter
                     args.push_back(buffer);   // add to the args
                     buf_offset = 0;
                     skip = false;   // turn off skipping
                  }
               }
               else
               { 
                  // move the char into the buffer
                  buffer[buf_offset] = thischar;
                  buf_offset++;  
               }
            
            }    
         }
      }
      // Add all the args to argv
      for (const auto& arg : args)
          argv.push_back((char*)arg.data());
   }      
   
   int argc = argv.size();
   sprintf(buffer,"DIAG ogrinfo argc=%d",argc);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   char** papszOpenOptions=nullptr;
   
   // Print all the options 
   for (int i = 0; i < argc; i++)
   {
      sprintf(buffer,"DIAG ogrinfo argv[%d]=%s\n",i,argv[i]);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      // printf("DEBUG ARG%d=%s\n",i,argv[i]);
   } 
   
   bool bVerbose = true;
   bool bSuperQuiet = false;
   bool bSummaryOnly = false;
   GIntBig nFetchFID = OGRNullFID;
   char** papszOptions = nullptr;
   
   char *pszWHERE = nullptr;
   const char *pszDataSource = nullptr;
   char **papszLayers = nullptr;
   OGRGeometry *poSpatialFilter = nullptr;
   int nRepeatCount = 1;
   bool bAllLayers = false;
   char *pszSQLStatement = nullptr;
   const char *pszDialect = nullptr;
   int nRet = 0;
   const char* pszGeomField = nullptr;
   char **papszExtraMDDomains = nullptr;
   bool bListMDD = false;
   bool bShowMetadata = true;
   bool bFeatureCount = true;
   bool bExtent = true;
   bool bDatasetGetNextFeature = false;
   bool bReadOnly = false;
   bool bUpdate = false;
   
   #define CHECK_HAS_ENOUGH_ADDITIONAL_ARGS(nExtraArg) \
    do { if (iArg + nExtraArg >= argc) \
         sprintf(buffer,"Error: ogrinfo %s option requires %d argument(s)",argv[iArg], nExtraArg);  \
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  } while( false )
                            
   // scan all the arguments and set the flags
   for (int iArg= 0; iArg < argc; iArg++)
   {
      if (EQUAL(argv[iArg], "-ro") ) bReadOnly = true;
      else if(EQUAL(argv[iArg], "-update") )  bUpdate = true;
      
      else if( EQUAL(argv[iArg], "-q") || EQUAL(argv[iArg], "-quiet")) bVerbose = false;
      else if( EQUAL(argv[iArg], "-qq") )
      {
         
         bVerbose = false;
         bSuperQuiet = true;
      }
      else if( EQUAL(argv[iArg], "-fid") )
      {
         CHECK_HAS_ENOUGH_ADDITIONAL_ARGS(1);
         nFetchFID = CPLAtoGIntBig(argv[++iArg]);
      }
      else if( EQUAL(argv[iArg], "-spat") )
      {
         CHECK_HAS_ENOUGH_ADDITIONAL_ARGS(4);
      
         OGRLinearRing oRing;
         oRing.addPoint(CPLAtof(argv[iArg+1]),
                        CPLAtof(argv[iArg+2]));
         oRing.addPoint(CPLAtof(argv[iArg+1]),
                        CPLAtof(argv[iArg+4]));
         oRing.addPoint(CPLAtof(argv[iArg+3]),
                        CPLAtof(argv[iArg+4]));
         oRing.addPoint(CPLAtof(argv[iArg+3]),
                        CPLAtof(argv[iArg+2]));
         oRing.addPoint(CPLAtof(argv[iArg+1]),
                        CPLAtof(argv[iArg+2]));
      
         poSpatialFilter = new OGRPolygon();
         static_cast<OGRPolygon *>(poSpatialFilter)->addRing(&oRing);
         iArg += 4;
      }
      else if( EQUAL(argv[iArg], "-geomfield") )
      {
         CHECK_HAS_ENOUGH_ADDITIONAL_ARGS(1);
         pszGeomField = argv[++iArg];
      }
      else if( EQUAL(argv[iArg], "-where") )
      {
         CHECK_HAS_ENOUGH_ADDITIONAL_ARGS(1);
         iArg++;
         CPLFree(pszWHERE);
         GByte* pabyRet = nullptr;
         if( argv[iArg][0] == '@' &&
             VSIIngestFile(nullptr, argv[iArg] + 1, &pabyRet,
                           nullptr, 1024*1024) )
         {
             RemoveBOM(pabyRet);
             pszWHERE = reinterpret_cast<char *>(pabyRet);
         }
         else
         {
             pszWHERE = CPLStrdup(argv[iArg]);
         }
      }
      else if( EQUAL(argv[iArg], "-sql") )
      {
         CHECK_HAS_ENOUGH_ADDITIONAL_ARGS(1);
         iArg++;
         CPLFree(pszSQLStatement);
         GByte* pabyRet = nullptr;
         if( argv[iArg][0] == '@' &&
             VSIIngestFile(nullptr, argv[iArg] + 1, &pabyRet,
                           nullptr, 1024*1024) )
         {
             RemoveBOM(pabyRet);
             pszSQLStatement = reinterpret_cast<char *>(pabyRet);
             RemoveSQLComments(pszSQLStatement);
         }
         else
         {
             pszSQLStatement = CPLStrdup(argv[iArg]);
         }
      }
      else if( EQUAL(argv[iArg], "-dialect") )
      {
         CHECK_HAS_ENOUGH_ADDITIONAL_ARGS(1);
         pszDialect = argv[++iArg];
      }
      else if( EQUAL(argv[iArg], "-rc") )
      {
         CHECK_HAS_ENOUGH_ADDITIONAL_ARGS(1);
         nRepeatCount = atoi(argv[++iArg]);
      }
      else if( EQUAL(argv[iArg], "-al") )
      {
         bAllLayers = true;
         // printf("DEBUG bAllLayers=%d\n",bAllLayers);
      } 
      else if( EQUAL(argv[iArg], "-so") ||
              EQUAL(argv[iArg], "-summary")  )
      {
         bSummaryOnly = true;
      }
      else if( STARTS_WITH_CI(argv[iArg], "-fields=") )
      {
         char* pszTemp =
             static_cast<char *>(CPLMalloc(32 + strlen(argv[iArg])));
         snprintf(pszTemp,
                 32 + strlen(argv[iArg]),
                 "DISPLAY_FIELDS=%s", argv[iArg] + strlen("-fields="));
         papszOptions = CSLAddString(papszOptions, pszTemp);
         CPLFree(pszTemp);
      }
      else if( STARTS_WITH_CI(argv[iArg], "-geom=") )
      {
         char* pszTemp =
             static_cast<char *>(CPLMalloc(32 + strlen(argv[iArg])));
         snprintf(pszTemp,
                 32 + strlen(argv[iArg]),
                 "DISPLAY_GEOMETRY=%s", argv[iArg] + strlen("-geom="));
         papszOptions = CSLAddString(papszOptions, pszTemp);
         CPLFree(pszTemp);
      }
      else if( EQUAL(argv[iArg], "-oo") )
      {
         CHECK_HAS_ENOUGH_ADDITIONAL_ARGS(1);
         papszOpenOptions = CSLAddString(papszOpenOptions,
                                         argv[++iArg]);
      }
      else if( EQUAL(argv[iArg], "-nomd") )
         {
            bShowMetadata = false;
         }
         else if( EQUAL(argv[iArg], "-listmdd") )
         {
            bListMDD = true;
         }
         else if( EQUAL(argv[iArg], "-mdd") )
         {
            CHECK_HAS_ENOUGH_ADDITIONAL_ARGS(1);
            papszExtraMDDomains = CSLAddString(papszExtraMDDomains,
                                               argv[++iArg]);
         }
         else if( EQUAL(argv[iArg], "-nocount") )
         {
            bFeatureCount = false;
         }
         else if( EQUAL(argv[iArg], "-noextent") )
         {
            bExtent = false;
         }
         else if( EQUAL(argv[iArg], "-rl"))
         {
            bDatasetGetNextFeature = true;
         }
         else if( argv[iArg][0] == '-' )
         {
            sprintf(buffer,"Error: ogrinfo Unknown option name '%s",argv[iArg]);  
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         }
         else
         {
            papszLayers = CSLAddString(papszLayers, argv[iArg]);
            bAllLayers = false;
         } 
      }
      
      // printf("DEBUG HERE AFTER SElection options\n");
         
      if( pszDialect != nullptr && pszWHERE != nullptr &&
        pszSQLStatement == nullptr )
        printf("Warning: -dialect is ignored with -where. Use -sql instead");
      
      if( bDatasetGetNextFeature && pszSQLStatement )
      {
         sprintf(buffer,"Error: ogrinfo -rl is incompatible with -sql");  
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      }  
      
      // printf("DEBUG Opening the source=%s\n",sourceFile);
      
     
      //      Open data source.                                               
      
      GDALDataset *poDS = static_cast<GDALDataset *>(GDALOpenEx(
        // pszDataSource,
        sourceFile,
        ((bReadOnly || pszSQLStatement == nullptr) &&
         !bUpdate ? GDAL_OF_READONLY : GDAL_OF_UPDATE) | GDAL_OF_VECTOR,
        nullptr, papszOpenOptions, nullptr));
      if( poDS == nullptr && !bReadOnly && !bUpdate &&
        pszSQLStatement == nullptr )
      {
        // In some cases (empty geopackage for example), opening in read-only
        // mode fails, so retry in update mode
        // if( GDALIdentifyDriverEx(pszDataSource, GDAL_OF_VECTOR,
        if( GDALIdentifyDriverEx(sourceFile, GDAL_OF_VECTOR,
                                 nullptr, nullptr) )
        {
            poDS = static_cast<GDALDataset *>(GDALOpenEx(
                // pszDataSource,
                sourceFile,
                GDAL_OF_UPDATE | GDAL_OF_VECTOR, nullptr,
                papszOpenOptions, nullptr));
        }
      }
      if( poDS == nullptr && !bReadOnly && !bUpdate &&
        pszSQLStatement != nullptr )
      {
        poDS = static_cast<GDALDataset *>(GDALOpenEx(
            // pszDataSource,
            sourceFile,
            GDAL_OF_READONLY | GDAL_OF_VECTOR, nullptr,
            papszOpenOptions, nullptr));
        if( poDS != nullptr && bVerbose )
        {
            printf("Had to open data source read-only.\n");
        }
      }
      
      GDALDriver *poDriver = nullptr;
      if( poDS != nullptr )
        poDriver = poDS->GetDriver();
        
      // printf("DEBUG After Opening the source=%s\n",sourceFile);  
      
      //      Report failure                                                  
      
      if( poDS == nullptr )
      {
        printf("FAILURE:\n"
               "Unable to open datasource `%s' with the following drivers.\n",
              sourceFile);
      
        OGRSFDriverRegistrar *poR = OGRSFDriverRegistrar::GetRegistrar();
        for( int iDriver = 0; iDriver < poR->GetDriverCount(); iDriver++ )
        {
            printf("  -> %s\n", poR->GetDriver(iDriver)->GetDescription());
        }
      
        nRet = 1;
        goto end;
      }
      
      CPLAssert(poDriver != nullptr);   
      
      
      //      Some information messages.                                      
      
      if( bVerbose )
        printf("INFO: Open of `%s'\n"
               "      using driver `%s' successful.\n",
               sourceFile, poDriver->GetDescription());
      
      if( bVerbose && !EQUAL(sourceFile,poDS->GetDescription()) )
      {
        printf("INFO: Internal data source name `%s'\n"
               "      different from user name `%s'.\n",
               poDS->GetDescription(), sourceFile);
      }
      
      GDALInfoReportMetadata(static_cast<GDALMajorObjectH>(poDS),
                           bListMDD,
                           bShowMetadata,
                           papszExtraMDDomains);
      
      if( bDatasetGetNextFeature )
      {
        nRepeatCount = 0;  // skip layer reporting.
      
      
      //      Set filters if provided.                                       
      
      if( pszWHERE != nullptr || poSpatialFilter != nullptr )
        {
            for( int iLayer = 0; iLayer < poDS->GetLayerCount(); iLayer++ )
            {
                OGRLayer *poLayer = poDS->GetLayer(iLayer);
      
                if( poLayer == nullptr )
                {
                    printf("FAILURE: Couldn't fetch advertised layer %d!\n",
                           iLayer);
                    exit(1);
                }
      
                if( pszWHERE != nullptr )
                {
                    if( poLayer->SetAttributeFilter(pszWHERE) != OGRERR_NONE )
                    {
                        printf("WARNING: SetAttributeFilter(%s) "
                               "failed on layer %s.\n",
                               pszWHERE, poLayer->GetName());
                    }
                }
      
                if( poSpatialFilter != nullptr )
                {
                    if( pszGeomField != nullptr )
                    {
                        OGRFeatureDefn *poDefn = poLayer->GetLayerDefn();
                        const int iGeomField =
                            poDefn->GetGeomFieldIndex(pszGeomField);
                        if( iGeomField >= 0 )
                            poLayer->SetSpatialFilter(iGeomField,
                                                      poSpatialFilter);
                        else
                            printf("WARNING: Cannot find geometry field %s.\n",
                                   pszGeomField);
                    }
                    else
                    {
                        poLayer->SetSpatialFilter(poSpatialFilter);
                    }
                }
            }
        }
      
        std::set<OGRLayer*> oSetLayers;
        while( true )
        {
            OGRLayer* poLayer = nullptr;
            OGRFeature* poFeature = poDS->GetNextFeature(&poLayer, nullptr,
                                                         nullptr, nullptr);
            if( poFeature == nullptr )
                break;
            if( papszLayers == nullptr || poLayer == nullptr ||
                CSLFindString(papszLayers, poLayer->GetName()) >= 0 )
            {
                if( bVerbose && poLayer != nullptr &&
                    oSetLayers.find(poLayer) == oSetLayers.end() )
                {
                    oSetLayers.insert(poLayer);
                    const bool bSummaryOnlyBackup = bSummaryOnly;
                    bSummaryOnly = true;
                    ReportOnLayer(poLayer, nullptr, nullptr, nullptr,
                                  bListMDD, bShowMetadata,
                                  papszExtraMDDomains,
                                  bFeatureCount,
                                  bExtent);
                    bSummaryOnly = bSummaryOnlyBackup;
                }
                if( !bSuperQuiet && !bSummaryOnly )
                    poFeature->DumpReadable(nullptr, papszOptions);
            }
            OGRFeature::DestroyFeature(poFeature);
        }
      }
      
      //     Special case for -sql clause.  No source layers required.       
      
      else if( pszSQLStatement != nullptr )
      {
        nRepeatCount = 0;  // skip layer reporting.
      
        if( CSLCount(papszLayers) > 0 )
            printf("layer names ignored in combination with -sql.\n");
      
        OGRLayer *poResultSet =
            poDS->ExecuteSQL(
                pszSQLStatement,
                pszGeomField == nullptr ? poSpatialFilter : nullptr,
                pszDialect);
      
        if( poResultSet != nullptr )
        {
            if( pszWHERE != nullptr )
            {
                if( poResultSet->SetAttributeFilter(pszWHERE) != OGRERR_NONE )
                {
                    printf("FAILURE: SetAttributeFilter(%s) failed.\n",
                           pszWHERE);
                    exit(1);
                }
            }
      
            if( pszGeomField != nullptr )
                ReportOnLayer(poResultSet, nullptr,
                              pszGeomField, poSpatialFilter,
                              bListMDD, bShowMetadata, papszExtraMDDomains,
                              bFeatureCount, bExtent);
            else
                ReportOnLayer(poResultSet, nullptr, nullptr, nullptr,
                              bListMDD, bShowMetadata, papszExtraMDDomains,
                              bFeatureCount, bExtent);
            poDS->ReleaseResultSet(poResultSet);
        }
      }
      
      // coverity[tainted_data]
      for( int iRepeat = 0; iRepeat < nRepeatCount; iRepeat++ )
      {
        if( papszLayers == nullptr || *papszLayers == nullptr )
        {
            if( iRepeat == 0 )
                CPLDebug("OGR", "GetLayerCount() = %d\n",
                         poDS->GetLayerCount());
      
            
            //      Process each data source layer.                                 
            
            for( int iLayer = 0; iLayer < poDS->GetLayerCount(); iLayer++ )
            {
                OGRLayer *poLayer = poDS->GetLayer(iLayer);
      
                if( poLayer == nullptr )
                {
                    printf("FAILURE: Couldn't fetch advertised layer %d!\n",
                           iLayer);
                    exit(1);
                }
      
                if( !bAllLayers )
                {
                    printf("%d: %s", iLayer + 1, poLayer->GetName());
      
                    const int nGeomFieldCount =
                        poLayer->GetLayerDefn()->GetGeomFieldCount();
                    if( nGeomFieldCount > 1 )
                    {
                        printf(" (");
                        for( int iGeom = 0; iGeom < nGeomFieldCount; iGeom++ )
                        {
                            if( iGeom > 0 )
                                printf(", ");
                            OGRGeomFieldDefn* poGFldDefn =
                                poLayer->GetLayerDefn()->
                                    GetGeomFieldDefn(iGeom);
                            printf(
                                "%s",
                                OGRGeometryTypeToName(
                                    poGFldDefn->GetType()));
                        }
                        printf(")");
                    }
                    else if( poLayer->GetGeomType() != wkbUnknown )
                        printf(" (%s)",
                               OGRGeometryTypeToName(
                                   poLayer->GetGeomType()));
      
                    printf("\n");
                }
                else
                {
                    if( iRepeat != 0 )
                        poLayer->ResetReading();
      
                    ReportOnLayer(poLayer, pszWHERE,
                                  pszGeomField, poSpatialFilter,
                                  bListMDD, bShowMetadata, papszExtraMDDomains,
                                  bFeatureCount, bExtent);
                }
            }
        }
        else
        {
            
            //      Process specified data source layers.                           
      
            for( char** papszIter = papszLayers;
                 *papszIter != nullptr;
                 ++papszIter )
            {
                OGRLayer *poLayer = poDS->GetLayerByName(*papszIter);
      
                if( poLayer == nullptr )
                {
                    printf("FAILURE: Couldn't fetch requested layer %s!\n",
                           *papszIter);
                    exit(1);
                }
      
                if( iRepeat != 0 )
                    poLayer->ResetReading();
      
                ReportOnLayer(poLayer, pszWHERE, pszGeomField, poSpatialFilter,
                              bListMDD, bShowMetadata, papszExtraMDDomains,
                              bFeatureCount, bExtent);
            }
        }
      }
      
      
      //      Close down.                                                     
      
      GDALClose(poDS);
      
      CSLDestroy(papszLayers);
      CSLDestroy(papszOptions);
      CSLDestroy(papszOpenOptions);
      CSLDestroy(papszExtraMDDomains);
      if( poSpatialFilter )
        OGRGeometryFactory::destroyGeometry(poSpatialFilter);
      CPLFree(pszSQLStatement);
      CPLFree(pszWHERE);
      
      OGRCleanupAll();
    
end:
   return 0;
}
*/
/**************************** Rgb2Pct ***********************************************
  Written by Paul Higgins June, 2016
 from the RGB2PCT.PY written by  Frank Warmerdam
#******************************************************************************
#  Copyright (c) 2001, Frank Warmerdam
# 
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
# 
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
# 
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#******************************************************************************
 * Takes an RGB tif file and creates a one band tif file with the colors reduced
 * This can be used as input to TIF2BSB
 * Syntax:	rc = Rgb2Pct(filein, fileout, numcolors(optional), diag, diag_array)
 * This invokes GDAL.  It is similiar to Rgb2Pct.PY  
 *
 * Params:		
 *			   src_fn - filename of the tif input file
 *			   dst_fn  - filename of the tif outputfile file
 *			   numcolors  - number of colors in the output (optional) default is 127
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages   
 *
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine5(int, Rgb2Pct, CSTRING, src_fn, CSTRING, dst_fn, OPTIONAL_int, numcolors, int, diag, RexxArrayObject, diagmsgs)
{
   using namespace dangdal;
   char buffer[512]; 
   int rtnrc = 0;
   int rc, color_count;
  
   if ( diag>0 )
   {
      sprintf(buffer,"DIAG Rgb2Pct src_fn=%s dst_fn=%s",src_fn, dst_fn);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   } 
   
   // If the numcolors is provided then use it
	if ( argumentExists(3) ) color_count = numcolors;
   else color_count  = 127 ;
	if ( diag>0 ) 
   {
      sprintf(buffer,"DIAG Rgb2Pct color_count=%d",color_count);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }  

   std::string tif_fn;
   
	// std::vector<GDALDatasetH> src_ds;     
	std::string output_format = "GTiff";
   
	GDALAllRegister();

	//////// open source //////// 
   GDALDatasetH src_ds = GDALOpen(src_fn, GA_ReadOnly);      
	if(src_ds == NULL) 
   {
      sprintf(buffer,"ERROR: Rgb2Pct open failed for src_fn");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }
   
   // it must have 3 rasters
   if(GDALGetRasterCount(src_ds) < 3)
   { 
      sprintf(buffer,"ERROR: Rgb2Pct src_ds need 3 rasters for input red, green and blue.");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }
   
   // Ensure we recognise the driver.
   GDALDriverH dst_driver = GDALGetDriverByName(output_format.c_str());
	if(!dst_driver) 
   {
      sprintf(buffer,"ERROR: unrecognized output format (%s)", output_format.c_str());
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }
   
   // Generate palette
   GDALColorTableH  ct;
   ct = GDALCreateColorTable((GDALPaletteInterp)1);
   if( ct == NULL )
        fatal_error("Color Table is Null.");
   
   int err=GDALComputeMedianCutPCT(GDALGetRasterBand(src_ds, 1),
                                    GDALGetRasterBand(src_ds, 2),
                                    GDALGetRasterBand(src_ds, 3),
                                    NULL,
                                    color_count, 
                                    ct,
                                    NULL, 
                                    NULL);
   if ( err!=0 ) 
   {                                 
      sprintf(buffer,"ERROR: GDALComputeMedianCutPCT failed err=%d", err);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   }
                                      
   tif_fn = dst_fn;
   
   GDALDatasetH tif_ds = GDALCreate(dst_driver, tif_fn.c_str(), GDALGetRasterXSize(src_ds), GDALGetRasterYSize(src_ds), 1 , (GDALDataType)1, NULL);
   if ( tif_ds == NULL ) 
   {                                 
      sprintf(buffer,"ERROR: Rgb2Pct tif_ds = NULL!");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   }
   
   err= GDALSetRasterColorTable(GDALGetRasterBand(tif_ds,1), ct );
   
   // ----------------------------------------------------------------------------
   //  We should copy projection information and so forth at this point.
   
   // PY tif_ds.SetProjection( src_ds.GetProjection() )
   GDALSetProjection(tif_ds,  GDALGetProjectionRef(src_ds));
   
   /* PY
   tif_ds.SetGeoTransform( src_ds.GetGeoTransform() )
   if src_ds.GetGCPCount() > 0:
    tif_ds.SetGCPs( src_ds.GetGCPs(), src_ds.GetGCPProjection() )
   */
   err = GDALSetGeoTransform(tif_ds, (double *)GDALGetGCPProjection(src_ds));
   if ( err!=0 ) 
   {                                 
      sprintf(buffer,"ERROR: GDALSetGeoTransform failed err=%d", err);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   }
   
   if (GDALGetGCPCount(src_ds) > 0) 
   {
      err= GDALSetGCPs(tif_ds, GDALGetGCPCount(src_ds), GDALGetGCPs(src_ds), GDALGetGCPProjection(src_ds));
      if ( err!=0 ) 
      {                                 
         sprintf(buffer,"ERROR: GDALGetGCPCount failed err=%d", err);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      }
   }
   
   // ----------------------------------------------------------------------------
   // Actually transfer and dither the data.
         
   err = GDALDitherRGB2PCT( GDALGetRasterBand(src_ds, 1),
                             GDALGetRasterBand(src_ds, 2),
                             GDALGetRasterBand(src_ds, 3),
                             GDALGetRasterBand(tif_ds, 1),
                             ct,
                             NULL,
                             NULL );
   if ( err!=0 ) 
   {                            
      sprintf(buffer,"ERROR:  GDALDitherRgb2Pct err=%d",err);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }
   
   GDALClose(tif_ds);                             
   tif_ds = NULL;
   
   GDALClose(src_ds);
	// GDALClose(dst_ds);

	GDALTermProgress(1, NULL, NULL);

	return 0;
}

/**************************** Merge_Simple ***********************************************
 * Merges several images into one image with many bands.
 * This can be used as input to TIF2BSB
 * Syntax:	rc = Merge_Simple(filein, fileout, numcolors(optional), diag, diag_array)
 * This invokes GDAL.  It is similiar to Rgb2Pct.PY  
 *
 * Params:		
 *			   dst_fn  - filename of the tif outputfile file
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages  
 *			   in1_fn  - filename of first input file
 *			   in2_fn  - optional filename of 
 *			   in3_fn  - optional filename of third input file

 *
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine6(int, Merge_Simple, CSTRING,dst_fn, int,diag, RexxArrayObject,diagmsgs, CSTRING,in1_fn, OPTIONAL_CSTRING,in2_fn, OPTIONAL_CSTRING,in3_fn)
{
   using namespace dangdal;
   char buffer[512]; 
   int rtnrc = 0;
   int rc;
  
   if ( diag>0 )
   {
      sprintf(buffer,"DIAG Merge_Simple dst_fn=%s",dst_fn);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   } 
   
   // Function declaration
   void copyGeoCode(GDALDatasetH dst_ds, GDALDatasetH src_ds);

	std::string output_format = "GTiff";
   std::vector<GDALDatasetH> src_ds;

	GDALAllRegister();

	//////// open source ////////   
   GDALDatasetH ds;
   ds = GDALOpen(in1_fn, GA_ReadOnly);   
   if (ds == NULL) 
   {
      sprintf(buffer,"ERROR: Merge_Simple GDALOpen failed for in1:%s",in1_fn);   
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }
	// put it in the list
	src_ds.push_back(ds);
   if ( diag>0 )
   {
      sprintf(buffer,"DIAG Merge_Simple added:%s", in1_fn);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }
   
   // if there is a an IN2 exists open it
   if ( argumentExists(5))
   {

      ds = GDALOpen(in2_fn, GA_ReadOnly);
      if (ds == NULL) 
      {
         sprintf(buffer,"ERROR: Merge_Simple GDALOpen failed for in2_fn:%s",in2_fn);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
      }
   	// put it in the list
   	src_ds.push_back(ds); 
      if ( diag>0 )
      {
         sprintf(buffer,"DIAG Merge_Simple added:%s", in2_fn);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      } 
   }
   
   // if there is a an in3 exists open it
   if ( argumentExists(6))
   {
      ds = GDALOpen(in3_fn, GA_ReadOnly);
      if (ds == NULL) 
      {
         sprintf(buffer,"ERROR: Merge_Simple GDALOpen failed for in3_fn:%s",in3_fn);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
      }
   	// put it in the list
   	src_ds.push_back(ds); 
      if ( diag>0 )
      {
         sprintf(buffer,"DIAG Merge_Simple added:%s", in3_fn);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }
   }
   

	std::vector<GDALRasterBandH> src_bands;

	size_t w=0, h=0;

	for(size_t ds_idx=0; ds_idx<src_ds.size(); ds_idx++) {
		size_t ds_w = GDALGetRasterXSize(src_ds[ds_idx]);
		size_t ds_h = GDALGetRasterYSize(src_ds[ds_idx]);
		if(!ds_w || !ds_h) fatal_error("missing width/height");
		if(w) {
			if(ds_w != w || ds_h != h) 
         {
            sprintf(buffer,"ERROR: Merge_Simple size mismatch for inputs");
            printf(buffer);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            return 1;
         }
		} else 
      {
			w = ds_w;
			h = ds_h;
		}

		int nb = GDALGetRasterCount(src_ds[ds_idx]);
		for(int i=0; i<nb; i++) {
			src_bands.push_back(GDALGetRasterBand(src_ds[ds_idx], i+1));
		}
	}

	size_t band_count = src_bands.size();

	//////// open output ////////
	// printf("Output size is %zd x %zd x %zd\n", w, h, band_count);

	GDALDriverH dst_driver = GDALGetDriverByName(output_format.c_str());
	if(!dst_driver) fatal_error("unrecognized output format (%s)", output_format.c_str());
	GDALDatasetH dst_ds = GDALCreate(dst_driver, dst_fn, w, h, band_count, GDT_Byte, NULL);
	if(!dst_ds) fatal_error("could not create output");
	copyGeoCode(dst_ds, src_ds[0]);

	std::vector<GDALRasterBandH> dst_bands;
	for(size_t i=0; i<band_count; i++) {
		dst_bands.push_back(GDALGetRasterBand(dst_ds, i+1));
	}
   

	//////// process data ////////

	int chunk_size = 200;

	// FIXME - handle other datatypes
	std::vector<uint8_t> buf(w * chunk_size);

	for(size_t row=0; row<h; row+=chunk_size) {
		GDALTermProgress((double)row/(double)h, NULL, NULL);

		int num_lines = chunk_size;
		if(row+num_lines > h) num_lines = h - row;
		for(size_t band_idx=0; band_idx<band_count; band_idx++) {
			if(GDALRasterIO(
				src_bands[band_idx], GF_Read,
				0, row, w, num_lines,
				&buf[0], w, num_lines,
				GDT_Byte, 0, 0
			) != CE_None) fatal_error("read error");
			if(GDALRasterIO(
				dst_bands[band_idx], GF_Write,
				0, row, w, num_lines,
				&buf[0], w, num_lines,
				GDT_Byte, 0, 0
			) != CE_None) 
         {
            sprintf(buffer,"ERROR: Merge_Simple write error");
            printf(buffer);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            return 1;      
         }   
		}
	}

	//////// shutdown ////////

	for(size_t ds_idx=0; ds_idx<src_ds.size(); ds_idx++) {
		GDALClose(src_ds[ds_idx]);
	}
     
	GDALClose(dst_ds);

	GDALTermProgress(1, NULL, NULL);

	return 0;
}

/**************************** Pansharpen ***********************************************
 * Pansharpens low_res RGB images using a Hi-res panchromatic image
 * Syntax:	rc = Pansharpen(pan_tif, dst_fn, diag, diag_array, rgb1_fn [',rgb12_fn] [',rgb13_fn])
 * This invokes GDAL.  It is similiar to Rgb2Pct.PY  
 *
 * Params:		
 *			   pan_fn - filename of the panchromatic image
 *			   dst_fn  - filename of the tif outputfile file
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages  
 *			   rgb1_fn  - filename of first low_res image
 *			   rgb2_fn  - optional filename of second low_res image
 *			   rgb3_fn  - optional filename of third low_res image

 *
 * Return: 0 if OK else 1
 *       
   Copyright (c) 2013, Regents of the University of Alaska
   
   All rights reserved.
   
   Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
   
       * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
       * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
       * Neither the name of the Geographic Information Network of Alaska nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
   
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS     
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
   
   This code was developed by Dan Stahlke for the Geographic Information Network of Alaska.
 *********************************************************************/
RexxRoutine7(int, Pansharpen, CSTRING, pan_fn, CSTRING, dst_fn, int,diag, RexxArrayObject, diagmsgs, CSTRING, rgb1_fn, OPTIONAL_CSTRING, rgb2_fn, OPTIONAL_CSTRING, rgb3_fn)
{
   using namespace dangdal;
   char buffer[512]; 
   int rtnrc = 0, rc;
   
   // Functions Declararations for PanSharpen
   void copyGeoCode(GDALDatasetH dst_ds, GDALDatasetH src_ds);
   ScaledBand getScaledBand(GDALDatasetH lores_ds, int band_id, GDALDatasetH hires_ds);
   void readLineScaled(ScaledBand &sb, int row, double *hires_buf);
  
   if ( diag>0 )
   {
      sprintf(buffer,"DIAG Pansharpen pan_fn=%s dst_fn=%s",pan_fn, dst_fn);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   } 

	std::vector<GDALDatasetH> rgb_ds;  // list of the lowres RGBs

   std::vector<GDALDatasetH> lum_ds;
   std::vector<double> lum_weights;
   std::string output_format = "GTiff";
	double ndv = 0;
	bool use_ndv = 0;

	GDALAllRegister();
   
   GDALDatasetH ds;
   // Open first low_res RGB
   ds = GDALOpen(rgb1_fn, GA_ReadOnly);   
   if (ds == NULL) 
   {
      sprintf(buffer,"ERROR: Pansharpen GDALOpen failed for rgb1:%s",rgb1_fn);   
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }
	// put it in the list
	rgb_ds.push_back(ds); 
   if ( diag>0 )
   {
      sprintf(buffer,"DIAG Pansharpen added:%s", rgb1_fn);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }
   
   // if there is a an RGB2 exists open it
   if ( argumentExists(6))
   {
      // Open first low_res RGB
      ds = GDALOpen(rgb2_fn, GA_ReadOnly);
      if (ds == NULL) 
      {
         sprintf(buffer,"ERROR: Pansharpen GDALOpen failed for rgb2_fn:%s",rgb2_fn);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
      }
   	// put it in the list
   	rgb_ds.push_back(ds); 
      if ( diag>0 )
      {
         sprintf(buffer,"DIAG Pansharpen added:%s", rgb2_fn);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      } 
   }
   
   // if there is a an RGB3 exists open it
   if ( argumentExists(7))
   {
      // Open first low_res RGB
      ds = GDALOpen(rgb2_fn, GA_ReadOnly);
      if (ds == NULL) 
      {
         sprintf(buffer,"ERROR: Pansharpen GDALOpen failed for rgb2_fn:%s",rgb2_fn);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
      }
   	// put it in the list
   	rgb_ds.push_back(ds); 
      if ( diag>0 )
      {
         sprintf(buffer,"DIAG Pansharpen added:%s", rgb3_fn);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }
   }

	//////// open source ////////

	GDALDatasetH pan_ds = GDALOpen(pan_fn, GA_ReadOnly);
   if (pan_ds == NULL) 
   {
      {
         sprintf(buffer,"ERROR: Pansharpen GDALOpen failed for rgb2_fn:%s",rgb2_fn);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
      }
   }
	
	size_t w = GDALGetRasterXSize(pan_ds);
	size_t h = GDALGetRasterYSize(pan_ds);

	if(GDALGetRasterCount(pan_ds) != 1) 
   {
      sprintf(buffer,"ERROR: Pansharpen pan input must be only one band");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }

	GDALRasterBandH pan_band = GDALGetRasterBand(pan_ds, 1);

	// this will be updated via GDALDataTypeUnion as RGB bands are opened
	GDALDataType out_dt = GDALGetRasterDataType(pan_band);

	/////

	std::vector<ScaledBand> rgb_bands;
   printf("Processing rgb bands\n");
	for(size_t ds_idx=0; ds_idx<rgb_ds.size(); ds_idx++) {
		int nb = GDALGetRasterCount(rgb_ds[ds_idx]);
		for(int i=0; i<nb; i++) {
			rgb_bands.push_back(getScaledBand(rgb_ds[ds_idx], i+1, pan_ds));
			out_dt = GDALDataTypeUnion(out_dt, GDALGetRasterDataType(
				GDALGetRasterBand(rgb_ds[ds_idx], i+1)));
		}
	}
	size_t rgb_band_count = rgb_bands.size();

	std::vector<ScaledBand> lum_bands;
 
	lum_bands = rgb_bands;
	lum_weights.assign(rgb_band_count, 1);

	size_t lum_band_count = lum_bands.size();

	double lum_weight_total = 0;
	for(size_t i=0; i<lum_band_count; i++) 
   {
		lum_weight_total += lum_weights[i];
	}
   
	// printf("lum weights:");
	for(size_t i=0; i<lum_band_count; i++) 
   {
		lum_weights[i] /= lum_weight_total;
		printf(" %lf", lum_weights[i]);
	}

	GDALDriverH dst_driver = GDALGetDriverByName(output_format.c_str());
	if(!dst_driver)
   {
      sprintf(buffer,"ERROR: Pansharpen unrecognized output format");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   } 

	GDALDatasetH dst_ds = GDALCreate(dst_driver, dst_fn, w, h, rgb_band_count, out_dt, NULL);
   if (dst_ds == NULL) 
   {
      sprintf(buffer,"ERROR: Pansharpen GDALCreate failed for dst_fn");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   } 
   
	copyGeoCode(dst_ds, pan_ds);

	std::vector<GDALRasterBandH> dst_bands;
	for(size_t i=0; i<rgb_band_count; i++) {
		dst_bands.push_back(GDALGetRasterBand(dst_ds, i+1));
		if(use_ndv) {
			GDALSetRasterNoDataValue(dst_bands[i], ndv);
		}
	}

	//////// process data ////////

	std::vector<std::vector<double> > lum_buf(lum_band_count);
	for(size_t band_idx=0; band_idx<lum_band_count; band_idx++) {
		lum_buf[band_idx].resize(w);
	}
	std::vector<double> pan_buf(w);
	std::vector<double> rgb_buf(w);
	std::vector<double> out_buf(w);
	std::vector<double> scale_buf(w);

	for(size_t row=0; row<h; row++) {
		GDALTermProgress((double)row/h, NULL, NULL);

		GDALRasterIO(pan_band, GF_Read, 0, row, w, 1, &pan_buf[0], w, 1, GDT_Float64, 0, 0);
		for(size_t band_idx=0; band_idx<lum_band_count; band_idx++) {
			readLineScaled(lum_bands[band_idx], row, &lum_buf[band_idx][0]);
		}

		for(size_t col=0; col<w; col++) {
			bool skip = 0;

			if(use_ndv) {
				if(pan_buf[col] == ndv) skip = 1;
				for(size_t band_idx=0; band_idx<lum_band_count; band_idx++) {
					if(lum_buf[band_idx][col] == ndv) {
						skip = 1;
					}
				}
			}

			if(skip) {
				scale_buf[col] = 1;
			} else {
				double lum_out = (double)pan_buf[col];
				double lum_in = 0;
				for(size_t i=0; i<lum_band_count; i++) {
					lum_in += lum_buf[i][col] * lum_weights[i];
				}

				scale_buf[col] = lum_in>0 ? lum_out/lum_in : 0;
			} // skip
		} // col

		for(size_t band_idx=0; band_idx<rgb_band_count; band_idx++) {
			readLineScaled(rgb_bands[band_idx], row, &rgb_buf[0]);

			for(size_t col=0; col<w; col++) {
				if(use_ndv && rgb_buf[col] == ndv) {
					out_buf[col] = ndv;
				} else {
					double dbl_val = rgb_buf[col] * scale_buf[col];
					out_buf[col] = dbl_val;
				}
			}

			GDALRasterIO(dst_bands[band_idx], GF_Write, 0, row, w, 1,
				&out_buf[0], w, 1, GDT_Float64, 0, 0);
		}
	} // row

	for(size_t i=0; i<rgb_ds.size(); i++) {
		GDALClose(rgb_ds[i]);
	}
	for(size_t i=0; i<lum_ds.size(); i++) {
		GDALClose(lum_ds[i]);
	}
	GDALClose(pan_ds);
	GDALClose(dst_ds);

	GDALTermProgress(1, NULL, NULL);

	return 0;
}


/* ADDED  */
double round (double r) 
{
      return  (r> 0.0) ? floor(r + 0.5): ceil(r - 0.5);
}
      
void copyGeoCode(GDALDatasetH dst_ds, GDALDatasetH src_ds) 
{
	double affine[6];
	if(GDALGetGeoTransform(src_ds, affine) == CE_None) {
		GDALSetGeoTransform(dst_ds, affine);
	}
	GDALSetProjection(dst_ds, GDALGetProjectionRef(src_ds));
}


ScaledBand getScaledBand(GDALDatasetH lores_ds, int band_id, GDALDatasetH hires_ds) 
{
	double lores_affine[6];
	double hires_affine[6];
	GDALGetGeoTransform(lores_ds, lores_affine);
   
   GDALGetGeoTransform(hires_ds, hires_affine); 

	ScaledBand sb;

	double band_res = lores_affine[1];
	double base_res = hires_affine[1];
	double offset_x =  (hires_affine[0] - lores_affine[0]) / band_res;
	double offset_y = -(hires_affine[3] - lores_affine[3]) / band_res;
	double scale = band_res / base_res;
	sb.oversample = (int)round(scale);

	sb.lo_w = GDALGetRasterXSize(lores_ds);
	sb.lo_h = GDALGetRasterYSize(lores_ds);

	sb.hi_w = GDALGetRasterXSize(hires_ds);
	sb.hi_h = GDALGetRasterYSize(hires_ds);

	sb.band = GDALGetRasterBand(lores_ds, band_id);

	sb.delta_x = (int)floor(offset_x);
	offset_x -= (double)sb.delta_x;
	sb.delta_y = (int)floor(offset_y);
	offset_y -= (double)sb.delta_y;

	// Cubic convolution resampling.  For more info see:
	// http://www.imgfsr.com/ResamplingCVPR.pdf
	sb.kernel_x.resize(sb.oversample);
	sb.kernel_y.resize(sb.oversample);
	for(int mod=0; mod<sb.oversample; mod++) {
		sb.kernel_x[mod].resize(4);
		sb.kernel_y[mod].resize(4);
		double t = offset_x + (double)mod / (double)sb.oversample;
		sb.kernel_x[mod][0] = -0.5*t*t*t + 1.0*t*t - 0.5*t;
		sb.kernel_x[mod][1] =  1.5*t*t*t - 2.5*t*t + 1;
		sb.kernel_x[mod][2] = -1.5*t*t*t + 2.0*t*t + 0.5*t;
		sb.kernel_x[mod][3] =  0.5*t*t*t - 0.5*t*t;
		t = offset_y + (double)mod / (double)sb.oversample;
		sb.kernel_y[mod][0] = -0.5*t*t*t + 1.0*t*t - 0.5*t;
		sb.kernel_y[mod][1] =  1.5*t*t*t - 2.5*t*t + 1;
		sb.kernel_y[mod][2] = -1.5*t*t*t + 2.0*t*t + 0.5*t;
		sb.kernel_y[mod][3] =  0.5*t*t*t - 0.5*t*t;
	}

	sb.lines_buf.resize(4);
	for(int j=0; j<4; j++) {
		sb.lines_buf[j].resize(sb.hi_w);
	}
	sb.line_buf_idx = -1000000;

	return sb;
}

void readLineScaled1D(ScaledBand &sb, int row, double *hires_buf) 
{
	std::vector<double> lores_buf(sb.lo_w);

	if(row < 0 || size_t(row) >= sb.lo_h) {
		for(size_t col=0; col<sb.hi_w; col++) {
			hires_buf[col] = 0;
		}
	} else {
		GDALRasterIO(sb.band, GF_Read, 0, row, sb.lo_w, 1, &lores_buf[0], sb.lo_w, 1, GDT_Float64, 0, 0);
		for(int mx=0; mx<sb.oversample; mx++) {
			double *kernel = &sb.kernel_x[mx][0];
			for(int x0=0; ; x0++) {
				size_t col = x0 * sb.oversample + mx;
				if(col >= sb.hi_w) break;
				double accum = 0;
				for(int i=0; i<4; i++) {
					int x = x0 - 1 + i + sb.delta_x;
					double v = (x<0 || size_t(x) >= sb.lo_w) ? 0 : lores_buf[x];
					accum += v * kernel[i];
				}
				hires_buf[col] = accum;
			}
		}
	}
}

void readLineScaled(ScaledBand &sb, int row, double *hires_buf) 
{
	int y0 = row / sb.oversample;
	int my = row % sb.oversample;
	double *kernel = &sb.kernel_y[my][0];

	int top_y = y0 - 1 + sb.delta_y;
	if(top_y == sb.line_buf_idx) {
		// no action
	} else if(top_y == sb.line_buf_idx+1) {
		std::vector<double> tmp;
		std::swap(tmp, sb.lines_buf[0]);
		for(int j=0; j<3; j++) {
			std::swap(sb.lines_buf[j], sb.lines_buf[j+1]);
		}
		std::swap(sb.lines_buf[3], tmp);
		readLineScaled1D(sb, top_y+3, &sb.lines_buf[3][0]);
		sb.line_buf_idx = top_y;
	} else {
		for(int j=0; j<4; j++) {
			readLineScaled1D(sb, top_y+j, &sb.lines_buf[j][0]);
		}
		sb.line_buf_idx = top_y;
	}

	for(size_t col=0; col<sb.hi_w; col++) {
		double accum = 0;
		for(int j=0; j<4; j++) {
			accum += sb.lines_buf[j][col] * kernel[j];
		}
		hires_buf[col] = accum;
	}
}

// build the actual entry list
RexxRoutineEntry ge2kap_gdal_routines[] = {         
    REXX_TYPED_ROUTINE(GdalInfo, GdalInfo),
    REXX_TYPED_ROUTINE(Gdal_Translate, Gdal_Translate),
    REXX_TYPED_ROUTINE(ogr2ogr, ogr2ogr),    
    // REXX_TYPED_ROUTINE(ogrinfo, ogrinfo),
    REXX_TYPED_ROUTINE(Rgb2Pct, Rgb2Pct),
    REXX_TYPED_ROUTINE(Merge_Simple, Merge_Simple),
    REXX_TYPED_ROUTINE(Pansharpen, Pansharpen),
    REXX_LAST_ROUTINE()
};  


RexxPackageEntry window_package_entry =
{
    STANDARD_PACKAGE_HEADER
    REXX_INTERPRETER_4_0_0,              // needs at least the 4.0.0 interpreter
    "GE2KAP_GDAL_Funcs",                 // name of the package
    "2.0.0",                             // package information
    NULL,                                // no load/unload functions
    NULL,
    ge2kap_gdal_routines,                    // the exported functions
};

// package loading stub.
OOREXX_GET_PACKAGE(window);