/*  These are functions and Methods used by Sat2Chart.REX              */

#include <windows.h>    
#include <CommCtrl.h>
#include <stdio.h>
#include "sqlite3.h"    

// for Rgb2Pct
//#include "common.h" 
// #if defined(_WIN32)
// #define _CDECL __cdecl
// #else
// #define _CDECL    
// #endif
// #include "oorexxapi.h"        

// #include "C:\Program Files (x86)\Visual Leak Detector\include\vld.h"    //      added for leak detection in V31     
#include <vector>
#include <iostream>             
#include <fstream>
#include<Psapi.h>  // added in V31 for MemUsage
#pragma comment(lib, "psapi.lib")

#include "ximage.h"
#include <atltypes.h>    
#include <atlstr.h>    
// #include <shlwapi.h>//need to add library shlwapi!   

#include <GdiPlus.h>
#pragma comment( lib, "gdiplus" )

#define MULTI_TO_WIDE( x,y )  MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, y, -1, x, _MAX_PATH );

#include "Quantize.h"
#include "bsb.h"
#include <tiffio.h>		/* libtiff - TIFF file I/O */

#include "oorexxapi.h"

#include "georef.h"
#include "cutil.h"

#include <tchar.h>
#include "zlib.h"

#define HAVE_SSIZE_T

#define DTOL                 1e-12

#define YANDEX_Rn (6378137.0)
// #define YANDEX_E (0.0818191908426)  0.0818191909028683
#define YANDEX_E 0.0818191909028683
#define YANDEX_A (20037508.342789)
#define YANDEX_F (53.5865938)
#define YANDEX_AB (0.00335655146887969400)
#define YANDEX_BB (0.00000657187271079536)
#define YANDEX_CB (0.00000001764564338702)
#define YANDEX_DB (0.00000000005328478445)


#define RadiusA 6378137 
#define RadiusB  6356752

#define HALFPI  1.5707963267948966
#define SPI     3.14159265358979323846
#define TWOPI   6.2831853071795864769
#define ONEPI   3.14159265358979323846
#define DEG_TO_RAD ONEPI/180.0
#define MERI_TOL 1e-9

#define D_R (M_PI / 180.0)
#define R_D (180.0 / M_PI)
#define R_MAJOR 6378137.0
#define R_MINOR 6356752.0
#define RATIO (R_MINOR/R_MAJOR)
#define ECCENT (sqrt(1.0 - (RATIO * RATIO)))
#define COM (0.5 * ECCENT)

#define _USE_MATH_DEFINES
// Additions for MS Windows compilation:
#ifndef M_PI
	#define M_PI acos(-1.0)
#endif
#ifndef M_PI_2
	#define M_PI_2 1.57079632679489661922
#endif

#ifndef M_PI_4
#define M_PI_4 0.785398163397448309616
#endif

typedef struct
{
    std::uint32_t biSize;
    std::int32_t  biWidth;
    std::int32_t  biHeight;
    std::uint16_t  biPlanes;
    std::uint16_t  biBitCount;
    std::uint32_t biCompression;
    std::uint32_t biSizeImage;
    std::int32_t  biXPelsPerMeter;
    std::int32_t  biYPelsPerMeter;
    std::uint32_t biClrUsed;
    std::uint32_t biClrImportant;
} DIB;

typedef struct
{
    std::uint16_t type;
    std::uint32_t bfSize;
    std::uint32_t reserved;
    std::uint32_t offset;
} HEADER;

typedef struct
{
    HEADER header;
    DIB dib;
} BMP;

inline double fmin(double x, double y) { return(x < y ? x : y); }
inline double fmax(double x, double y) { return(x > y ? x : y); }

static int getMaxPixels(int zoom_in) 
{     
   int worldSize [26];
	for (int zoom = 0; zoom < 26; zoom++)
		worldSize[zoom] = 256 * (1 << zoom);
	return worldSize[zoom_in];
}

/*
public static double getPowZoom(float zoom){
		if(zoom >= 0 && zoom - Math.floor(zoom) < 0.001f){
			return 1 << ((int)zoom); 
		} else {
			return Math.pow(2, zoom);
		}
	}
*/  

static double rho_math_atanh(double n)
{
    return 0.5 * log((1 + n)/(1 - n));
} 


static uint64_t rho_math_pow2(int n)
{
    uint64_t ret;
    for (ret = 1; n > 0; --n)
        ret *= 2;
    return ret;
}

static double Clip(double n, double minValue, double maxValue)
{
   return min(max(n, minValue), maxValue);
}

static double getPowZoom(double zoom)
{
	if(zoom >= 0 && zoom - floor(zoom) < 0.001)
   {
		return 1 << ((int)zoom); 
	} 
   else 
   {
		return pow(2, zoom);
	}
}

static double toRadians(double angdeg) 
{
   return angdeg / 180.0 * SPI;
}

static double checkLongitude(double longitude) 
{
	while (longitude < -180 || longitude > 180) 
   {
		if (longitude < 0) 
      {
			longitude += 360;
		} 
      else 
      {
				longitude -= 360;
		}
	}
		
   return longitude;
}

/**************************** getTileNumberX **************************
 * * Local version converts Spheroid Longitude to tilex
 * * @see osmand.util
 *********************************************************************/
// RexxRoutine2(double,getTileNumberX, double, zoom, double, longitude )
static double L_getTileNumberX(double zoom, double longitude) 
{
   if(longitude == 180.0) 
   {
   	return getPowZoom(zoom) - 1;
   }
   longitude = checkLongitude(longitude);
   return (longitude + 180.0)/360.0 * getPowZoom(zoom);
}
/**************************** getLongitudeFromTile **************************
 *  Local version get Longitude from tilex
 * * @see osmand.util
 *********************************************************************/
// RexxRoutine2(double,getLongitudeFromTile, double, zoom, double, x )
static double L_getLongitudeFromTile(double zoom,  double x) 
{  /*  
   i	public static double getLongitudeFromTile(float zoom, double x) {
		return x / getPowZoom(zoom) * 360.0 - 180.0;
   */
	return x / getPowZoom(zoom) * 360.0 - 180.0;
}
/**************************** getTileEllipsoidNumberY **************************
 *  Local version converts Latitude to Ellipsoid tiley
 * * @see osmand.util
 *********************************************************************/
// RexxRoutine2(double,getTileEllipsoidNumberY, double, zoom, double, latitude )
static double L_getTileEllipsoidNumberY(double zoom,  double latitude)
{

   double E2 = (double) latitude * SPI / 180;
   double sradiusa = 6378137;
   double sradiusb = 6356752;
   double J2 = (double) sqrt((sradiusa * sradiusa) - (sradiusb * sradiusb))	/ sradiusa;
   double M2 = (double) log((1 + sin(E2))
   		/ (1 - sin(E2)))/ 2- J2	* log((1 + J2 * sin(E2))/ (1 - J2 * sin(E2))) / 2;
   double B2 = getPowZoom(zoom);
   return B2 / 2 - M2 * B2 / 2 / SPI;
}
/**************************** getLatitudeFromEllipsoidTileY **************************
 *  Local Version get Latitude from  Ellipsoid tiley
 * * @see osmand.util       DEBUG SOME TIMES GIVES nan use
 *                         pixely = ul_tiley * 256
                           lat = cYToLat(pixely, zm) 
 *********************************************************************/
// RexxRoutine2(double,getLatitudeFromEllipsoidTileY, double, zoom, double, tileNumberY )
static double L_getLatitudeFromEllipsoidTileY(double zoom,  double tileNumberY)
{

   double MerkElipsK = 0.0000001;
   double sradiusa = 6378137;
   double sradiusb = 6356752;
   double FExct = (double) sqrt((sradiusa * sradiusa)  - (sradiusb * sradiusb)) / sradiusa;
   double TilesAtZoom = getPowZoom(zoom);
   double result = (tileNumberY - TilesAtZoom / 2) / -(TilesAtZoom / (2 * SPI));
   result = (2 * atan(exp(result)) - SPI / 2) * 180 / SPI;
   double Zu = result / (180 / SPI);
   double yy = (tileNumberY - TilesAtZoom / 2);
   double Zum1 = Zu;
   Zu = asin(1 - ((1 + sin(Zum1)) * pow(1 - FExct * sin(Zum1), FExct))
   		/ (exp((2 * yy) / -(TilesAtZoom / (2 * SPI))) * pow(1 + FExct * sin(Zum1), FExct)));
   while (abs(Zum1 - Zu) >= MerkElipsK) 
   {
   	Zum1 = Zu;
   	Zu = asin(1 - ((1 + sin(Zum1)) * pow(1 - FExct * sin(Zum1), FExct))
   			/ (exp((2 * yy) / -(TilesAtZoom / (2 * SPI))) * pow(1 + FExct * sin(Zum1), FExct)));
   }
   
   return Zu * 180 / SPI;
}
/**************************** getTileNumberY **************************
 *  Local version converts Spheroid Latitude to tiley
 * * @see osmand.util
 *********************************************************************/
// RexxRoutine2(double,getTileNumberY, double, zoom, double, latitude )
static double L_getTileNumberY(double zoom,  double latitude)
{
   
   double eval = log(tan(toRadians(latitude)) + 1/cos(toRadians(latitude)) );
   return (1 - eval / SPI) / 2 * getPowZoom(zoom);
}
/**************************** getLatitudeFromTile **************************
 *  Local version get latitude from tiley
 * * @see osmand.util
 *********************************************************************/
// RexxRoutine2(double,getLatitudeFromTile, double, zoom, double, y )
static double L_getLatitudeFromTile(double zoom,  double y)
{
	int sign = y < 0 ? -1 : 1;
	return atan(sign*sinh(SPI * (1 - 2 * y / getPowZoom(zoom)))) * 180.0 / SPI;
}
static  int falseNorthing(int aZoomlevel) 
{
	return (-1 * getMaxPixels(aZoomlevel) / 2);
}
static  double radius(int zoom) 
{
		return getMaxPixels(zoom) / (2.0 * SPI);
}


// Iteratively solve equation (7-9) from Snyder. 
static double cphi2(double ts)
{
   double iteration_tolerance = 1E-12;
   double maximum_iterations = 200;
   double excentricity = 0.081819790992;
	double eccnth = 0.5 * excentricity;
	double phi = (SPI / 2) - 2.0 * atan(ts);
	for (int i = 0; i < maximum_iterations; i++) 
   {
		double con = excentricity * sin(phi);
		double dphi = (SPI / 2) - 2.0 * atan(ts * pow((1 - con) / (1 + con), eccnth)) - phi;
		phi += dphi;
		if (abs(dphi) <= iteration_tolerance) 
      {
			return phi;
		}
	}
   return NULL;
}
static double rad_deg (double ang) 
{
        return ang * R_D;
}
/**************************** cYToLat **************************
 * * Transforms pixel coordinate ellipsoid Y to latitude
 * * @see mobac.mapsources.mapspace.MercatorPower2MapSpace#cYToLat(int, int)
 *********************************************************************/
// RexxRoutine2(double,cYToLat, int, y, int, zoom)
static double L_cYToLat(int y,  int zoom)
{ 
	int y2 = y + falseNorthing(zoom);
	double latitude = exp(-y2 / radius(zoom));
	latitude = cphi2(latitude);
	return -1 * rad_deg(latitude);
}
/*******************************************************************/  
// Local version R_Get_Tile_XY public 
// for a given lat/lon return the tile's x/y and tile lat/lon
/*******************************************************************/ 
static void L_Get_Tile_XY(double lat,  double lon, int zm, int projection,
                 int &tilex, int &tiley, double &tile_lat, double &tile_lon)
{
// use arg comm.,self,lat,lon,zm,projection

   // int therc;
   int pixely;
   
   // get the tilex/y and tilex/y lat/lon of this point
   if (projection == 2)  // Ellipsodial maps
   { 
      tilex = (int)(L_getTileNumberX(zm, lon)); 
      tile_lon = L_getLongitudeFromTile(zm, tilex);
      tiley = (int)(L_getTileEllipsoidNumberY(zm, lat));
      pixely = tiley * 256;    
      tile_lat = L_cYToLat(pixely, zm);
   } 
   else   // else do for Spheroid type projection
   { 
      tilex = (int)(L_getTileNumberX(zm, lon));
      tiley = (int)(L_getTileNumberY(zm, lat)); 
      tile_lat = L_getLatitudeFromTile(zm, tiley); 
      tile_lon = L_getLongitudeFromTile(zm, tilex);
   } 
}
/******************************************************************* 
Trival
-- creates a value for the 3 points either the above/below or the left/right
-- if we are going up or down the value is to the left and right of xp,yp
-- if we are going left or right the value is for above and below of xp,yp
-- parms: - tiles array 
--        - xp, yp, 
--        - x_incr, y_incr
-- returns: trival as 3 top bits of one hex value: '80'x = tri_1
--                                                 '40'x = curr
--                                                 '20'x = tri_2
*******************************************************************/ 
byte Trival(int xp, int yp, int x_incr, int y_incr, int bmp_width, int bmp_height, char * tiles_all)
{

   byte trival = 0;
   
   int tri_1 = 0;
   int tri_2 = 0;
   
   int direction = 0;
   
   if (x_incr != 0) direction = 1;
   if (y_incr != 0) direction = 2;
   
   // Calculate the tival
   
   switch(direction)
   {
      case 1 :  // Moving Left or Right
         if (yp == 0);  // at the top  
         else // Else set tri_1 to the value of the tile above  
         {   
            tri_1 = tiles_all[xp + ((yp-1) * bmp_width)];      
         }   
         if (yp == bmp_height - 1); // At the bottom
         else  // Else set tri_2 to the value of the tile below
         {
            tri_2 = tiles_all[xp + ((yp+1) * bmp_width)]; 
         }  
   
         break;
         
      case 2 :  // Moving Up or Down
         if (xp == 0); // At the Left side
         else  // Else set tri_1 to the value of the tile to the left
         {
            tri_1 = tiles_all[xp-1 + (yp * bmp_width)]; 
            
         }
         
         if (xp == bmp_width - 1); // At the right side
         else  // Else set tri_2 to the value of the tile to the right
         {
            tri_2 = tiles_all[xp+1 + (yp * bmp_width)]; 
         }
   
         break;
      
      default:  // Both 0  set tri_1 and Tri_2 to value of the current tile
         tri_1 = tiles_all[xp + (yp * bmp_width)];  
         tri_2 = tiles_all[xp + (yp * bmp_width)]; 
         
         break;
   }
   
   // Set the curr to the value of the current tile
   int curr = tiles_all[xp + (yp * bmp_width)]; 
   
   // Set the trival bits
   byte bit80 = 0x80;  
   byte bit40 = 0x40;  
   byte bit20 = 0x20;   
   if (tri_1 == 1) trival = trival | bit80;                
   if (curr == 1) trival = trival | bit40;  
   if (tri_2 == 1) trival = trival | bit20;  
   
   return trival;
   
}

struct MonitorRects
{
    std::vector<RECT>   rcMonitors;

    static BOOL CALLBACK MonitorEnum(HMONITOR hMon,HDC hdc,LPRECT lprcMonitor,LPARAM pData)
    {
        MonitorRects* pThis = reinterpret_cast<MonitorRects*>(pData);
        pThis->rcMonitors.push_back(*lprcMonitor);
        return TRUE;
    }

    MonitorRects()
    {
        EnumDisplayMonitors(0, 0, MonitorEnum, (LPARAM)this);
    }
};
	
static double checkLatitude(double latitude) 
{
	while (latitude < -90 || latitude > 90) 
   {
		if (latitude < 0) 
      {
			latitude += 180;
		} 
      else 
      {
			latitude -= 180;
		}
	}
	if(latitude < -85.0511) 
   {
		return -85.0511;
	} 
   else if(latitude > 85.0511)
   {
		return 85.0511;
	}
	return latitude;
}

static double CalcTeta(double teta0, double y1, double e)
{
   double math_e = 2.71828182845905;
   return asin(1.0 - (1.0 + sin(teta0)) * pow(1.0 - e * sin(teta0), e) / (pow(math_e, 2.0 * y1 / 1.0) * pow(1 + e * sin(teta0), e)));
}

static double deg_rad (double ang) {
        return ang * D_R;
}

/* Convert x pixels to longitude  degrees */
static double pixel2lon(int zoom,double x) 
{  
   double mapsize = 256 * pow(2.0, zoom);
   // return fmod(fmod(x / mapsize * 360.0, 360.0) + 360.0, 360.0) - 180;
  double pixels = 256 * pow(2.0, zoom);      
  return x / pixels * 360 - 180;
} 

/* Convert y pixels to latitude degrees */
//Latitude = Degree(ATan(SinH(#PI * (1.0 - 2.0 * *Coords\y / n))))
static double pixel2lat(int zoom,double y) 
{
  double mapsize = 256 * pow(2.0, zoom);
  // return rad_deg(atan(sinh(SPI * (1.0 - 2.0 * y / mapsize))));
  double pixels = 256 * pow(2.0, zoom);
  double n = SPI - 2 * SPI * y / pixels;
  return 180 / PI * atan(0.5 * (exp(n) - exp(-n)));
}

double adjlon (double lon) 
{

   if (fabs(lon) <= SPI) return( lon );
   lon += ONEPI;  /* adjust to 0..2pi rad */
   lon -= TWOPI * floor(lon / TWOPI); /* remove integral # of 'revolutions'*/
   lon -= ONEPI;  /* adjust back to -pi..pi rad */
   return( lon );
}

/*********************************************************************/
/* adjustAngleToPI ---This method adjusts the input angle to         */
/*              be within the range of 0.0 to PI. (international     */
/*              date line adjustment, go short way around earth,...) */
/*        Input: longitude in degrees                                    */
/*********************************************************************/
double adjustAngleToPI(double lon)
{  
   // convert input to radians 
   lon = ONEPI/180 * lon;  
   
   if (lon > ONEPI) lon = lon - 2.0 * ONEPI ;
   else if (lon < (-ONEPI)) lon = lon + (2.0*ONEPI);
   
   // convert output to degrees 
   lon = (180 / ONEPI) * lon;

   return (lon);
}


/*********************************************************************/
/* ckinbounds -- checks if a lat/lon are with a box         
the first two parms are the north latitude and west longitude of the box
the next two are the south latitude and east longitude  of the box
the last two parms parm are the lat/lon to check
           returns true if it is within the box else returns false   */
/*********************************************************************/
bool ckinbounds(double nl, double wl, double sl, double el, double cklat, double cklon)
{
   bool lon0,rtcd;
   rtcd = false;
   
   // check if crossing lon0
   if (wl < 0 && el > 0) lon0 = true;
   else lon0 = false;
   // convert all the longitudes to 0-360
   if (wl  < 0) wl += 360;
   if (el  < 0) el += 360;
   if (cklon < 0) cklon += 360;
   // convert all the latitudes to 0-180
   nl += 90;
   sl += 90;
   cklat += 90;
   
   // check depending if chart crosses lon0
   if (lon0) 
   { 
      if ((cklon >= wl && cklon >= el && cklat <= nl && cklat >= sl) || (cklon <= wl && cklon <= el && cklat <= nl && cklat >= sl))   // added = in V30.0.3.5
      {
         rtcd = true;  
      }
   }
   else  // not crossing
   {
      if (cklon >= wl && cklon <= el && cklat <= nl && cklat >= sl) rtcd = true;    // added = in V30.0.3.5
   }
   
   return rtcd;

}

/*********************************************************************/
/* Meridian180Crossed   
    Returns true if the 180 meridian is crossed else false
                       -- parms lon1 and lon2                    
*********************************************************************/ 
bool Meridian180Crossed(double lon1, double lon2)
{

   int dir = 0;
   bool idl = false;
   double diff;
   if (lon1 < 0 && lon2 > 0) dir = 1;
   if (lon2 < 0 && lon1 > 0) dir = 2;
   if (lon1 < 0) lon1 = 360.0 - fabs(lon1);
   if (lon2 < 0) lon2 = 360.0 - fabs(lon2);
   diff = fabs(lon1 - lon2);
   if (dir > 0 && diff < 180) idl = true;
   
   // if idl then self~logmsg('Meridian180Crossed yes')
   
   return idl;
}


#define TOL 1e-10
#define CONV      1e-10
#define N_ITER    10
#define I_ITER 20
#define ITOL 1.e-12

#define Sat2Chart 0
#define ENTLEN 7
#define ENTSIZE sizeof(LONG) * ENTLEN

// Added in V28.0.1.2
#define HISTLEN 3
#define HISTSIZE sizeof(LONG) * ENTLEN

#ifdef __cplusplus
extern "C" {
#endif

// TCHAR* FindExtension(const TCHAR * name);

#ifdef _WIN32
#define DIR_SEPARATOR '\\'
#else
#define DIR_SEPARATOR '/'
#endif

/* MSVC doesn't supply a strcasecmp(), so use the MSVC workalike */
#ifdef _MSC_VER
#define strcasecmp(s1, s2) stricmp(s1, s2)
#endif

#define MAX_DIGITS     9

/*********************************************************************/
/*  Various definitions used by the math functions                   */
/*********************************************************************/
#define SINE        0                  /* trig function defines...   */
#define COSINE      3                  /* the ordering is important, */
#define TANGENT     1                  /* as these get transformed   */
#define COTANGENT   2                  /* depending on the angle     */
#define MAXTRIG     3                  /* value                      */
#define ARCSINE     0                  /* defines for arc trig       */
#define ARCCOSINE   1                  /* functions.  Ordering is    */
#define ARCTAN2     2                  /* not as important here      */

#define pi  3.14159265358979323846l    /* pi value                   */

#define MAX_PRECISION     16           /* maximum available precision*/
#define MIN_PRECISION     1            /* minimum available precision*/

 // simple class for handling numeric values
class NumericFormatter
{
public:
    NumericFormatter(RexxCallContext *c, bool explicitPrecision, uint32_t p)
    {
        optionError = false;
        precision = p;
        context = c;
        if (explicitPrecision)
        {
            if (p == 0)
            {
                // raise an error on return
                context->InvalidRoutine();
                // remember we rejected this
                optionError = true;
            }
        }
        else
        {
            precision = (int)context->GetContextDigits();
        }
        // cap this value
        if (precision > MAX_PRECISION)
        {
            precision = MAX_PRECISION;
        }
    }

    RexxObjectPtr format(double x)
    {
        // we've already raised an error here, so don't bother formatting a value
        if (optionError)
        {
            return NULLOBJECT;
        }

        return context->DoubleToObjectWithPrecision(x, precision);
    }

protected:
    uint32_t precision;
    RexxCallContext *context;

    bool optionError;     // had invalid options and we're going to raise an error
};

class TrigFormatter : public NumericFormatter
{
public:

    TrigFormatter(RexxCallContext *c, bool explicitPrecision, uint32_t p, const char *u) : NumericFormatter(c, explicitPrecision, p)
    {
        units = DEGREES;
        // process any units option
        if (u != NULL)
        {
            switch (*u)
            {
                case 'D':
                case 'd':
                    units = DEGREES;
                    break;

                case 'R':
                case 'r':
                    units = RADIANS;
                    break;

                case 'G':
                case 'g':
                    units = GRADES;
                    break;

                default:
                    context->InvalidRoutine();
                    optionError = true;
            }
        }
    }

    RexxObjectPtr evaluate(double angle, int function)
    {
        // if we've had an option error, we can't contine
        if (optionError)
        {
            return NULLOBJECT;
        }

        double    nsi;                       /* convertion factor          */
        double    nco;                       /* convertion factor          */
        double    result;                    /* result                     */

        nsi = 1.;                            /* set default conversion     */
        nco = 1.;                            /* set default conversion     */

        switch (units)
        {
            case DEGREES:         {            /* need to convert degrees    */
                    nsi = (angle < 0.) ? -1. : 1.;   /* get the direction          */
                    angle = fmod(fabs(angle), 360.); /* make modulo 360            */
                    if (angle <= 45.)                /* less than 45?              */
                    {
                        angle = angle * pi / 180.;
                    }
                    else if (angle < 135.)
                    {         /* over on the other side?    */
                        angle = (90. - angle) * pi / 180.;
                        function = MAXTRIG - function; /* change the function        */
                        nco = nsi;                     /* swap around the conversions*/
                        nsi = 1.;
                    }
                    else if (angle <= 225.)
                    {        /* around the other way?      */
                        angle = (180. - angle) * pi / 180.;
                        nco = -1.;
                    }
                    else if (angle < 315.)
                    {         /* close to the origin?       */
                        angle = (angle - 270.) * pi / 180.;
                        function = MAXTRIG - function; /* change the function        */
                        nco = -nsi;
                        nsi = 1.;
                    }
                    else
                    {
                        angle = (angle - 360.) * pi / 180.;
                    }
                    break;
                }

            case GRADES:              {        /* need to convert degrees    */
                    nsi = (angle < 0.) ? -1. : 1.;   /* get the direction          */
                    angle = fmod(fabs(angle), 400.); /* make modulo 400            */
                    if (angle <= 50.)
                    {
                        angle = angle * pi / 200.;
                    }
                    else if (angle < 150.)
                    {
                        angle = (100. - angle) * pi / 200.;
                        function = MAXTRIG - function; /* change the function        */
                        nco = nsi;                     /* swap the conversions       */
                        nsi = 1.;
                    }
                    else if (angle <= 250.)
                    {
                        angle = (200. - angle) * pi / 200.;
                        nco = -1.;
                    }
                    else if (angle < 350.)
                    {
                        angle = (angle - 300.) * pi / 200.;
                        function = MAXTRIG - function; /* change the function        */
                        nco = -nsi;
                        nsi = 1.;
                    }
                    else
                    {
                        angle = (angle - 400.) * pi / 200.;
                    }
                    break;
                }

                // radians are already ok
            case RADIANS:
                break;
        }
        switch (function)                /* process the function       */
        {
            case SINE:                       /* Sine function              */
                result = nsi * sin(angle);
                break;
            case COSINE:                     /* Cosine function            */
                result = nco * cos(angle);
                break;
            case TANGENT:                    /* Tangent function           */
                result = nsi * nco * tan(angle);
                break;
            case COTANGENT:                  /* cotangent function         */
                /* overflow?                  */
                if ((result = tan(angle)) == 0.0)
                {
                    context->InvalidRoutine();
                    return NULLOBJECT;
                }
                result = nsi * nco / result; /* real result                */
                break;
        }

        // now format based on precision setting
        return format(result);
    }

    RexxObjectPtr evaluateArc(double x, double y, int function)
    {
        // if we've had an option error, we can't contine
        if (optionError)
        {
            return NULLOBJECT;
        }

        double    angle;                     /* working angle              */
        double    nsi;                       /* convertion factor          */
        double    nco;                       /* convertion factor          */

        nsi = 1.;                            /* set default conversion     */
        nco = 1.;                            /* set default conversion     */

        switch (function)                /* process the function       */
        {
            case ARCSINE:                    /* ArcSine function           */
                angle = asin(x);
                break;
            case ARCCOSINE:                  /* ArcCosine function         */
                angle = acos(x);
                break;
            case ARCTAN2:                 /* ArcTangent function  2      */
                angle = atan2(x,y);
                break;
        }
        if (units == DEGREES)              /* have to convert the result?*/
        {
            angle = angle * 180. / pi;       /* make into degrees          */
        }
        else if (units == GRADES)          /* need it in grades?         */
        {
            angle = angle * 200. / pi;       /* convert to base 400        */
        }
        // now format based on precision setting
        return format(angle);
    }

protected:
    typedef enum
    {
        DEGREES,
        RADIANS,
        GRADES
    } Units;

    Units units;              // the type of units to process
};

// LITEUNZIP   This code was adapted from code wriiten by By Jeff Glatt      
#include "LiteUnzip.h"
#define DIRSLASH_CHAR	'\\'
#define IDS_OK        20
#define IDS_UNKNOWN   21
#ifndef CP_UTF8
#define CP_UTF8		65001
#endif

// =========================== Defines ======================
// Basic data types
typedef unsigned int	uInt;		// 16 bits or more
typedef unsigned char	UCH;
typedef unsigned long	ULG;
typedef unsigned long	lutime_t;	// define it ourselves since we don't include time.h

#define ZIP_MEMORY		0x01
#define ZIP_FILENAME	0x02
#define ZIP_HANDLE		0x04
#define ZIP_RAW			0x08
#define ZIP_ALREADYINIT	0x40000000
#define ZIP_UNICODE		0x80000000   

//global structures for task enumeration
typedef struct _TASK_LIST {
    DWORD       dwProcessId;
    DWORD       dwInheritedFromProcessId;
    BOOL        flags;
    HWND      hwnd;
    CHAR        status[256];
    TCHAR        WindowTitle[MAX_PATH];
} TASK_LIST, *PTASK_LIST;

typedef struct _TASK_LIST_ENUM {
    PTASK_LIST  tlist;
    DWORD       numtasks;
} TASK_LIST_ENUM, *PTASK_LIST_ENUM;
typedef struct tagCxImageInfo {
	DWORD	dwEffWidth;			///< DWORD aligned scan line width
	BYTE*	pImage;				///< THE IMAGE BITS
	CxImage* pGhost;			///< if this is a ghost, pGhost points to the body
	CxImage* pParent;			///< if this is a layer, pParent points to the body
	DWORD	dwType;				///< original image format
	char	szLastError[256];	///< debugging
	long	nProgress;			///< monitor
	long	nEscape;			///< escape
	long	nBkgndIndex;		///< used for GIF, PNG, MNG
	RGBQUAD nBkgndColor;		///< used for RGB transparency
	float	fQuality;			///< used for JPEG, JPEG2000 (0.0f ... 100.0f)
	BYTE	nJpegScale;			///< used for JPEG [ignacio]
	long	nFrame;				///< used for TIF, GIF, MNG : actual frame
	long	nNumFrames;			///< used for TIF, GIF, MNG : total number of frames
	DWORD	dwFrameDelay;		///< used for GIF, MNG
	long	xDPI;				///< horizontal resolution
	long	yDPI;				///< vertical resolution
	RECT	rSelectionBox;		///< bounding rectangle
	BYTE	nAlphaMax;			///< max opacity (fade)
	bool	bAlphaPaletteEnabled; ///< true if alpha values in the palette are enabled.
	bool	bEnabled;			///< enables the painting functions
	long	xOffset;
	long	yOffset;
	DWORD	dwCodecOpt[CMAX_IMAGE_FORMATS];	///< for GIF, TIF : 0=def.1=unc,2=fax3,3=fax4,4=pack,5=jpg
	RGBQUAD last_c;				///< for GetNearestIndex optimization
	BYTE	last_c_index;
	bool	last_c_isvalid;
	long	nNumLayers;
	DWORD	dwFlags;			///< 0x??00000 = reserved, 0x00??0000 = blend mode, 0x0000???? = layer id - user flags
	BYTE	dispmeth;
	bool	bGetAllFrames;
	bool	bLittleEndianHost;

} CXIMAGEINFO;

typedef struct tagCxTextInfo
{
   TCHAR    text[4096]; ///< text (char -> TCHAR for UNICODE [Cesar M])
	LOGFONT  lfont;      ///< font and codepage data
   COLORREF fcolor;     ///< foreground color
   long     align;      ///< DT_CENTER, DT_RIGHT, DT_LEFT aligment for multiline text
   BYTE     smooth;     ///< text smoothing option. Default is false.
   BYTE     opaque;     ///< text has background or hasn't. Default is true.
   				 ///< data for background (ignored if .opaque==FALSE) 
   COLORREF bcolor;     ///< background color
   float    b_opacity;  ///< opacity value for background between 0.0-1.0 Default is 0. (opaque)
   BYTE     b_outline;  ///< outline width for background (zero: no outline)
   BYTE     b_round;    ///< rounding radius for background rectangle. % of the height, between 0-50. Default is 10.
                      ///< (backgr. always has a frame: width = 3 pixel + 10% of height by default.)
} CXTEXTINFO;

// The deflate compression method (the only one supported in this version)
#define Z_DEFLATED		8


// preset dictionary flag in zlib header 
#define PRESET_DICT		0x20

// Maximum size of dynamic tree. The maximum found in a long but non-
// exhaustive search was 1004 huft structures (850 for length/literals
// and 154 for distances, the latter actually the result of an
// exhaustive search). The actual maximum is not known, but the
// value below is more than safe.
#define MANY		1440


// =========================== Structs ======================

typedef struct
{
	unsigned int tm_sec;	// seconds after the minute - [0,59]
	unsigned int tm_min;	// minutes after the hour - [0,59]
	unsigned int tm_hour;	// hours since midnight - [0,23]
	unsigned int tm_mday;	// day of the month - [1,31]
	unsigned int tm_mon;	// months since January - [0,11]
	unsigned int tm_year;	// years - [1980..2044]
} TM_UNZ;

// ZIPENTRYINFO holds information about one entry in the zip archive.
// This header appears in the ZIP archive.

#define UNZ_BUFSIZE (16384)
#define UNZ_MAXFILENAMEINZIP (256)
#define SIZECENTRALDIRITEM (0x2e)
#define SIZEZIPLOCALHEADER (0x1e)

#define ZIP_FIELDS_REFORMAT 0x000060F0
#define NUM_FIELDS_REFORMAT 15

#pragma pack(1)
typedef struct
{
	unsigned short	version;			// version made by					2 bytes
	unsigned short	version_needed;		// version needed to extract		2 bytes
	unsigned short	flag;				// general purpose bit flag			2 bytes
	unsigned short	compression_method;	// compression method				2 bytes
	unsigned long	dosDate;			// last mod file date in Dos fmt	4 bytes
	unsigned long	crc;				// crc-32							4 bytes
	unsigned long	compressed_size;	// compressed size					4 bytes
	unsigned long	uncompressed_size;	// uncompressed size				4 bytes
	unsigned short	size_filename;		// filename length					2 bytes
	unsigned short	size_file_extra;	// extra field length				2 bytes
	unsigned short	size_file_comment;	// file comment length				2 bytes
	unsigned short	disk_num_start;		// disk number start				2 bytes
	unsigned short	internal_fa;		// internal file attributes			2 bytes
	unsigned long	external_fa;		// external file attributes			4 bytes
	unsigned long	offset;				// Byte offset of local header		4 bytes
} ZIPENTRYINFO;
#pragma pack()

// readEntry() updates next_in and avail_in when avail_in has
// dropped to zero. It updates next_out and avail_out when avail_out
// has dropped to zero. All other fields are set by low level
// DEFLATE routines and must not be updated by the higher level.
//
// The fields total_in and total_out can be used for statistics or
// progress reports. After compression, total_in holds the total size of
// the uncompressed data and may be saved for use in the decompressor
// (particularly if the decompressor wants to decompress everything in
// a single step)

/* DEfined same as ZLIB.h  below
typedef struct {
	UCH	*		next_in;	// next input byte
	DWORD		avail_in;	// # of bytes available at next_in
	DWORD		total_in;	// total # of input bytes read so far
	UCH	*		next_out;	// next output byte should be put there
	DWORD		avail_out;	// remaining free space at next_out
	DWORD		total_out;	// total # of bytes output so far
#ifndef NDEBUG
	char *		msg;		// last error message, NULL if no error
#endif
	INTERNAL_STATE *state;
//	int			data_type;	// best guess about the data type: ascii or binary
//	ULG			adler;		// adler32 value of the uncompressed data
} Z_STREAM;
*/

typedef struct  {
    Bytef    *next_in;  /* next input byte */
    uInt     avail_in;  /* number of bytes available at next_in */
    uLong    total_in;  /* total nb of input bytes read so far */

    Bytef    *next_out; /* next output byte should be put there */
    uInt     avail_out; /* remaining free space at next_out */
    uLong    total_out; /* total nb of bytes output so far */

    char     *msg;      /* last error message, NULL if no error */
    struct internal_state FAR *state; /* not visible by applications */

    alloc_func zalloc;  /* used to allocate the internal state */
    free_func  zfree;   /* used to free the internal state */
    voidpf     opaque;  /* private data object passed to zalloc and zfree */

    int     data_type;  /* best guess about the data type: binary or text */
    uLong   adler;      /* adler32 value of the uncompressed data */
    uLong   reserved;   /* reserved for future use */
} Z_STREAM;


// ENTRYREADVARS holds tables/variables used when reading and decompressing an entry
typedef struct
{
	UCH			*InputBuffer;				// Buffer for reading in compressed data of the current entry
	Z_STREAM	stream;						// structure for inflate()
//	DWORD		PosInArchive;				// Current "file position" within the archive
	ULG			RunningCrc;					// crc32 of all data uncompressed
	DWORD		RemainingCompressed;		// Remaining number of bytes to be decompressed
	DWORD		RemainingUncompressed;		// Remaining number of bytes to be obtained after decomp
	unsigned long Keys[3];					// Decryption keys, initialized by initEntry()
	DWORD		RemainingEncrypt;			// The first call(s) to readEntry will read this many encryption-header bytes first
	char		CrcEncTest;					// If encrypted, we'll check the encryption buffer against this
} ENTRYREADVARS;

// For TUNZIP->flags
#define TZIP_ARCMEMORY			0x0000001	// Set if TZIP->archive is memory, instead of a file, handle.
#define TZIP_ARCCLOSEFH			0x0000002	// Set if we open the file handle in archiveOpen() and must close it later.
#define TZIP_GZIP				0x0000004	// Set if a GZIP archive.
#define TZIP_RAW				0x0000008	// Set if "raw" mode


// TUNZIP holds information about the ZIP archive itself.
typedef struct
{
	DWORD			Flags;
	HANDLE			ArchivePtr;					// Points to a handle, or a buffer if TZIP_ARCMEMORY
	DWORD			LastErr;					// Holds the last TUNZIP error code
	DWORD			InitialArchiveOffset;		// Initial offset within a file where the ZIP archive begins. This allows reading a ZIP archive contained within another file
	DWORD			ArchiveBufLen;				// Size of memory buffer
	DWORD			ArchiveBufPos;				// Current position within "ArchivePtr" if TZIP_ARCMEMORY
	DWORD			TotalEntries;				// Total number of entries in the current disk of this archive
	DWORD			CommentSize;				// Size of the global comment of the archive
	DWORD			ByteBeforeZipArchive;		// Byte before the archive, (>0 for sfx)
	DWORD			CurrentEntryNum;			// Number of the entry (in the archive) that is currently selected for
												// unzipping. -1 if none.
	DWORD			CurrEntryPosInCentralDir;	// Position of the current entry's header within the central dir
//	DWORD			CentralDirPos;				// Byte offset to the beginning of the central dir
	DWORD			CentralDirOffset;			// Offset of start of central directory with respect to the starting disk number
	unsigned char	*Password;					// Password, or 0 if none.
	unsigned char	*OutBuffer;					// Output buffer (where we decompress the current entry when unzipping it).
	ZIPENTRYINFO	CurrentEntryInfo;			// Info about the currently selected entry (gotten from the Central Dir)
	ENTRYREADVARS	EntryReadVars;				// Variables/buffers for decompressing the current entry
	unsigned char	Rootdir[MAX_PATH];			// Root dir for unzipping entries. Includes a trailing slash. Must be the last field!!!
} TUNZIP;


// ======================== Function Declarations ======================

// Diagnostic functions
#ifndef NDEBUG
#define LuAssert(cond,msg)
#define LuTrace(x)
#define LuTracev(x)
#define LuTracevv(x)
#define LuTracec(c,x)
#define LuTracecv(c,x)
#endif

static DWORD setCurrentEntry(TUNZIP *, ZIPENTRY *, DWORD);


// simplify the use of the INFLATE_HUFT type with some defines
// defines for inflate input/output
//   update pointers and return 
#define UPDBITS {s->bitb = b; s->bitk = k;}
#define UPDIN {z->avail_in = n; z->total_in += (ULG)(p-z->next_in); z->next_in = p;}
#define UPDOUT {s->write = q;}
#define UPDATE {UPDBITS UPDIN UPDOUT}
#define LEAVE {UPDATE return(inflate_flush(s,z,r));}
//   get bytes and bits 
#define LOADIN {p = z->next_in; n = z->avail_in; b = s->bitb; k = s->bitk;}
#define NEEDBYTE {if (n) r = Z_OK; else LEAVE}
#define NEXTBYTE (n--, *p++)
#define NEEDBITS(j) {while(k<(j)){NEEDBYTE;b|=((ULG)NEXTBYTE)<<k;k+=8;}}
#define DUMPBITS(j) {b >>= (j); k -= (j);}
//   output bytes 
#define WAVAIL (uInt)(q < s->read ? s->read - q - 1: s->end - q)
#define LOADOUT {q=s->write;m=(uInt)WAVAIL;}
#define WRAP {if(q==s->end&&s->read!=s->window){q=s->window;m=(uInt)WAVAIL;}}
#define FLUSH {UPDOUT r=inflate_flush(s,z,r); LOADOUT}
#define NEEDOUT {if(m==0){WRAP if(m==0){FLUSH WRAP if(m==0) LEAVE}}r=Z_OK;}
#define OUTBYTE(a) {*q++=(UCH)(a);m--;}
//   load local pointers 
#define LOAD {LOADIN LOADOUT}


// NOTE: I specify this data section to be Shared (ie, each running rexx
// script shares these variables, rather than getting its own copies of these
// variables). This is because, since I have only globals that are read-only
// or whose value is the same for all processes, I don't need a separate copy
// of these for each process that uses this DLL. In Visual C++'s Linker
// settings, I add "/section:Shared,rws"

#ifdef WIN32
#pragma data_seg("ZIPSHARE")

static HINSTANCE	ThisInstance;
#endif

// Table of valid spinwheel values      
static const double valid_spinwheel_colors[16] = 
{6525877, 6657464, 6525878, 6657463, 6525619, 6591670, 6591671, 6657207, 6591413, 
 6657206, 6657208, 6722743, 6722999, 6460084, 6591157, 6656948};
  
                    

// Table of CRC-32's of all single-byte values (made by make_Crc_table)
static const ULG Crc_table[256] = {
  0x00000000L, 0x77073096L, 0xee0e612cL, 0x990951baL, 0x076dc419L,
  0x706af48fL, 0xe963a535L, 0x9e6495a3L, 0x0edb8832L, 0x79dcb8a4L,
  0xe0d5e91eL, 0x97d2d988L, 0x09b64c2bL, 0x7eb17cbdL, 0xe7b82d07L,
  0x90bf1d91L, 0x1db71064L, 0x6ab020f2L, 0xf3b97148L, 0x84be41deL,
  0x1adad47dL, 0x6ddde4ebL, 0xf4d4b551L, 0x83d385c7L, 0x136c9856L,
  0x646ba8c0L, 0xfd62f97aL, 0x8a65c9ecL, 0x14015c4fL, 0x63066cd9L,
  0xfa0f3d63L, 0x8d080df5L, 0x3b6e20c8L, 0x4c69105eL, 0xd56041e4L,
  0xa2677172L, 0x3c03e4d1L, 0x4b04d447L, 0xd20d85fdL, 0xa50ab56bL,
  0x35b5a8faL, 0x42b2986cL, 0xdbbbc9d6L, 0xacbcf940L, 0x32d86ce3L,
  0x45df5c75L, 0xdcd60dcfL, 0xabd13d59L, 0x26d930acL, 0x51de003aL,
  0xc8d75180L, 0xbfd06116L, 0x21b4f4b5L, 0x56b3c423L, 0xcfba9599L,
  0xb8bda50fL, 0x2802b89eL, 0x5f058808L, 0xc60cd9b2L, 0xb10be924L,
  0x2f6f7c87L, 0x58684c11L, 0xc1611dabL, 0xb6662d3dL, 0x76dc4190L,
  0x01db7106L, 0x98d220bcL, 0xefd5102aL, 0x71b18589L, 0x06b6b51fL,
  0x9fbfe4a5L, 0xe8b8d433L, 0x7807c9a2L, 0x0f00f934L, 0x9609a88eL,
  0xe10e9818L, 0x7f6a0dbbL, 0x086d3d2dL, 0x91646c97L, 0xe6635c01L,
  0x6b6b51f4L, 0x1c6c6162L, 0x856530d8L, 0xf262004eL, 0x6c0695edL,
  0x1b01a57bL, 0x8208f4c1L, 0xf50fc457L, 0x65b0d9c6L, 0x12b7e950L,
  0x8bbeb8eaL, 0xfcb9887cL, 0x62dd1ddfL, 0x15da2d49L, 0x8cd37cf3L,
  0xfbd44c65L, 0x4db26158L, 0x3ab551ceL, 0xa3bc0074L, 0xd4bb30e2L,
  0x4adfa541L, 0x3dd895d7L, 0xa4d1c46dL, 0xd3d6f4fbL, 0x4369e96aL,
  0x346ed9fcL, 0xad678846L, 0xda60b8d0L, 0x44042d73L, 0x33031de5L,
  0xaa0a4c5fL, 0xdd0d7cc9L, 0x5005713cL, 0x270241aaL, 0xbe0b1010L,
  0xc90c2086L, 0x5768b525L, 0x206f85b3L, 0xb966d409L, 0xce61e49fL,
  0x5edef90eL, 0x29d9c998L, 0xb0d09822L, 0xc7d7a8b4L, 0x59b33d17L,
  0x2eb40d81L, 0xb7bd5c3bL, 0xc0ba6cadL, 0xedb88320L, 0x9abfb3b6L,
  0x03b6e20cL, 0x74b1d29aL, 0xead54739L, 0x9dd277afL, 0x04db2615L,
  0x73dc1683L, 0xe3630b12L, 0x94643b84L, 0x0d6d6a3eL, 0x7a6a5aa8L,
  0xe40ecf0bL, 0x9309ff9dL, 0x0a00ae27L, 0x7d079eb1L, 0xf00f9344L,
  0x8708a3d2L, 0x1e01f268L, 0x6906c2feL, 0xf762575dL, 0x806567cbL,
  0x196c3671L, 0x6e6b06e7L, 0xfed41b76L, 0x89d32be0L, 0x10da7a5aL,
  0x67dd4accL, 0xf9b9df6fL, 0x8ebeeff9L, 0x17b7be43L, 0x60b08ed5L,
  0xd6d6a3e8L, 0xa1d1937eL, 0x38d8c2c4L, 0x4fdff252L, 0xd1bb67f1L,
  0xa6bc5767L, 0x3fb506ddL, 0x48b2364bL, 0xd80d2bdaL, 0xaf0a1b4cL,
  0x36034af6L, 0x41047a60L, 0xdf60efc3L, 0xa867df55L, 0x316e8eefL,
  0x4669be79L, 0xcb61b38cL, 0xbc66831aL, 0x256fd2a0L, 0x5268e236L,
  0xcc0c7795L, 0xbb0b4703L, 0x220216b9L, 0x5505262fL, 0xc5ba3bbeL,
  0xb2bd0b28L, 0x2bb45a92L, 0x5cb36a04L, 0xc2d7ffa7L, 0xb5d0cf31L,
  0x2cd99e8bL, 0x5bdeae1dL, 0x9b64c2b0L, 0xec63f226L, 0x756aa39cL,
  0x026d930aL, 0x9c0906a9L, 0xeb0e363fL, 0x72076785L, 0x05005713L,
  0x95bf4a82L, 0xe2b87a14L, 0x7bb12baeL, 0x0cb61b38L, 0x92d28e9bL,
  0xe5d5be0dL, 0x7cdcefb7L, 0x0bdbdf21L, 0x86d3d2d4L, 0xf1d4e242L,
  0x68ddb3f8L, 0x1fda836eL, 0x81be16cdL, 0xf6b9265bL, 0x6fb077e1L,
  0x18b74777L, 0x88085ae6L, 0xff0f6a70L, 0x66063bcaL, 0x11010b5cL,
  0x8f659effL, 0xf862ae69L, 0x616bffd3L, 0x166ccf45L, 0xa00ae278L,
  0xd70dd2eeL, 0x4e048354L, 0x3903b3c2L, 0xa7672661L, 0xd06016f7L,
  0x4969474dL, 0x3e6e77dbL, 0xaed16a4aL, 0xd9d65adcL, 0x40df0b66L,
  0x37d83bf0L, 0xa9bcae53L, 0xdebb9ec5L, 0x47b2cf7fL, 0x30b5ffe9L,
  0xbdbdf21cL, 0xcabac28aL, 0x53b39330L, 0x24b4a3a6L, 0xbad03605L,
  0xcdd70693L, 0x54de5729L, 0x23d967bfL, 0xb3667a2eL, 0xc4614ab8L,
  0x5d681b02L, 0x2a6f2b94L, 0xb40bbe37L, 0xc30c8ea1L, 0x5a05df1bL,
  0x2d02ef8dL
};


// Masks for lower bits. And'ing with mask[n] masks the lower n bits
static const uInt inflate_mask[17] = {0x0000,
    0x0001, 0x0003, 0x0007, 0x000f, 0x001f, 0x003f, 0x007f, 0x00ff,
    0x01ff, 0x03ff, 0x07ff, 0x0fff, 0x1fff, 0x3fff, 0x7fff, 0xffff
};

static const uInt Fixed_bl = 9;
static const uInt Fixed_bd = 5;

// Order of the bit length code lengths
static const uInt Border[] = {16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15};

// Tables for deflate from PKZIP's appnote.txt. 
static const uInt CpLens[31] = { // Copy lengths for literal codes 257..285
			3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 15, 17, 19, 23, 27, 31,
			35, 43, 51, 59, 67, 83, 99, 115, 131, 163, 195, 227, 258, 0, 0};
        // see note #13 above about 258
static const uInt CpLExt[31] = { // Extra bits for literal codes 257..285
			0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2,
			3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 0, 112, 112}; // 112==invalid
static const uInt CpDist[30] = { // Copy offsets for distance codes 0..29
			1, 2, 3, 4, 5, 7, 9, 13, 17, 25, 33, 49, 65, 97, 129, 193,
			257, 385, 513, 769, 1025, 1537, 2049, 3073, 4097, 6145,
			8193, 12289, 16385, 24577};
static const uInt CpDExt[30] = { // Extra bits for distance codes 
			0, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6,
			7, 7, 8, 8, 9, 9, 10, 10, 11, 11,
			12, 12, 13, 13};

// Error messages
#ifndef WIN32
static const char UnknownErr[] = "Unknown zip result code";
static const char ErrorMsgs[] = "Success\0\
Can't create/open file\0\
Failed to allocate memory\0\
Error writing to file\0\
Entry not found in the zip archive\0\
Still more data to unzip\0\
Zip archive is corrupt or not a zip archive\0\
Error reading file\0\
The entry is in a format that can't be decompressed by this Unzip add-on\0\
Faulty arguments\0\
Can get memory only of a memory-mapped zip\0\
Not enough space allocated for memory zip\0\
There was a previous error\0\
Additions to the zip have already been ended\0\
The anticipated size turned out wrong\0\
Mixing creation and opening of zip\0\
Trying to seek the unseekable\0\
Tried to change mind, but not allowed\0\
An internal error during flation\0\
Password is incorrect\0";
#endif

#ifdef WIN32
#pragma data_seg()
#endif

// ==================== String functions ====================

#ifndef WIN32

int lstrcmpiA(const char *s1, const char *s2)
{
	  int	c1, c2;

	for (;;)
	{
		c1 = tolower((unsigned char)*s1++);
		c2 = tolower((unsigned char)*s2++);
		if (!c1 || c1 != c2) return(c1 - c2);
	}
}

#endif
void toSM_ECC(double lat, double lon, double lat0, double lon0, double *x, double *y)
{
      const double f = 1.0 / WGSinvf;       // WGS84 ellipsoid flattening parameter
      const double e2 = 2 * f - f * f;      // eccentricity^2  .006700
      const double e = sqrt(e2);

      const double z = WGS84_semimajor_axis_meters * mercator_k0;

      *x = (lon - lon0) * DEGREE * z;

// y =.5 ln( (1 + sin t) / (1 - sin t) )
      const double s = sin(lat * DEGREE);
      const double y3 = (.5 * log((1 + s) / (1 - s))) * z;

      const double s0 = sin(lat0 * DEGREE);
      const double y30 = (.5 * log((1 + s0) / (1 - s0))) * z;
      const double y4 = y3 - y30;

    //Add eccentricity terms

      const double falsen =  z *log(tan(PI/4 + lat0 * DEGREE / 2)*pow((1. - e * s0)/(1. + e * s0), e/2.));
      const double test =    z *log(tan(PI/4 + lat  * DEGREE / 2)*pow((1. - e * s )/(1. + e * s ), e/2.));
      *y = test - falsen;
}

/* --------------------------------------------------------------------------------- */
/*
// Given the lat/long of starting point, and traveling a specified distance,
// at an initial bearing, calculates the lat/long of the resulting location.
// using elliptic earth model.
*/
/* --------------------------------------------------------------------------------- */
void ll_gc_ll(double lat, double lon, double brg, double dist, double *dlat, double *dlon)
{
    
    double th1,costh1,sinth1,sina12,cosa12,M,N,c1,c2,D,P,s1;
    int merid, signS;
    
    /*   Input/Output from geodesic functions   */
    double al12;           /* Forward azimuth */
    double al21;           /* Back azimuth    */
    double geod_S;         /* Distance        */
    double phi1, lam1, phi2, lam2;
    
    int ellipse;
    double geod_f;
    double geod_a;
    double es, onef, f, f64, f2, f4;
    
    /*      Setup the static parameters  */
    phi1 = lat * DEGREE;            /* Initial Position  */
    lam1 = lon * DEGREE;
    al12 = brg * DEGREE;            /* Forward azimuth */
    geod_S = dist * 1852.0;         /* Distance        */
    
    
    //void geod_pre(struct georef_state *state) 
    {
        
        /*   Stuff the WGS84 projection parameters as necessary
         *      To avoid having to include <geodesic,h>
         */
        ellipse = 1;
        f = 1.0 / WGSinvf;       /* WGS84 ellipsoid flattening parameter */
        geod_a = WGS84_semimajor_axis_meters;
        
        es = 2 * f - f * f;
        onef = sqrt(1. - es);
        geod_f = 1 - onef;
        f2 = geod_f/2;
        f4 = geod_f/4;
        f64 = geod_f*geod_f/64;
        
        al12 = adjlon(al12); /* reduce to  +- 0-PI */
        signS = fabs(al12) > HALFPI ? 1 : 0;
        th1 = ellipse ? atan(onef * tan(phi1)) : phi1;
        costh1 = cos(th1);
        sinth1 = sin(th1);
        if ((merid = fabs(sina12 = sin(al12)) < MERI_TOL)) {
            sina12 = 0.;
            cosa12 = fabs(al12) < HALFPI ? 1. : -1.;
            M = 0.;
        } else {
            cosa12 = cos(al12);
            M = costh1 * sina12;
        }
        N = costh1 * cosa12;
        if (ellipse) {
            if (merid) {
                c1 = 0.;
                c2 = f4;
                D = 1. - c2;
                D *= D;
                P = c2 / D;
            } else {
                c1 = geod_f * M;
                c2 = f4 * (1. - M * M);
                D = (1. - c2)*(1. - c2 - c1 * M);
                P = (1. + .5 * c1 * M) * c2 / D;
            }
        }
        if (merid) s1 = HALFPI - th1;
        else {
            s1 = (fabs(M) >= 1.) ? 0. : acos(M);
            s1 =  sinth1 / sin(s1);
            s1 = (fabs(s1) >= 1.) ? 0. : acos(s1);
        }
    }
    
    
    //void  geod_for(struct georef_state *state)
    {
        double d,sind,u,V,X,ds,cosds,sinds,ss,de;
        
        ss = 0.;
        
        if (ellipse) {
            d = geod_S / (D * geod_a);
            if (signS) d = -d;
            u = 2. * (s1 - d);
            V = cos(u + d);
            X = c2 * c2 * (sind = sin(d)) * cos(d) * (2. * V * V - 1.);
            ds = d + X - 2. * P * V * (1. - 2. * P * cos(u)) * sind;
            ss = s1 + s1 - ds;
        } else {
            ds = geod_S / geod_a;
            if (signS) ds = - ds;
        }
        cosds = cos(ds);
        sinds = sin(ds);
        if (signS) sinds = - sinds;
        al21 = N * cosds - sinth1 * sinds;
        if (merid) {
            phi2 = atan( tan(HALFPI + s1 - ds) / onef);
            if (al21 > 0.) {
                al21 = PI;
                if (signS)
                    de = PI;
                else {
                    phi2 = - phi2;
                    de = 0.;
                }
            } else {
                al21 = 0.;
                if (signS) {
                    phi2 = - phi2;
                    de = 0;
                } else
                    de = PI;
            }
        } else {
            al21 = atan(M / al21);
            if (al21 > 0)
                al21 += PI;
            if (al12 < 0.)
                al21 -= PI;
            al21 = adjlon(al21);
            phi2 = atan(-(sinth1 * cosds + N * sinds) * sin(al21) /
            (ellipse ? onef * M : M));
            de = atan2(sinds * sina12 ,
                       (costh1 * cosds - sinth1 * sinds * cosa12));
            if (ellipse){
                if (signS)
                    de += c1 * ((1. - c2) * ds +
                    c2 * sinds * cos(ss));
                else
                    de -= c1 * ((1. - c2) * ds -
                    c2 * sinds * cos(ss));
            }
        }
        lam2 = adjlon( lam1 + de );
    }
    
    
    *dlat = phi2 / DEGREE;
    *dlon = lam2 / DEGREE;
}


/* From two lat/lons calculate the range, in meters and bearing*/
void DistanceBearingMercator(double lat0, double lon0, double lat1, double lon1, double *brg, double *dist)
{
      //    Calculate bearing by conversion to SM (Mercator) coordinates, then simple trigonometry

      double lon0x = lon0;
      double lon1x = lon1;

      //    Make lon points the same phase
      if((lon0x * lon1x) < 0.)
      {
          lon0x < 0.0 ? lon0x += 360.0 : lon1x += 360.0;
            //    Choose the shortest distance
            if(fabs(lon0x - lon1x) > 180.)
            {
                lon0x > lon1x ? lon0x -= 360.0 : lon1x -= 360.0;
            }

            //    Make always positive
            lon1x += 360.;
            lon0x += 360.;
      }

      //    Classic formula, which fails for due east/west courses....
      if(dist)
      {
          //    In the case of exactly east or west courses
          //    we must make an adjustment if we want true Mercator distances

          //    This idea comes from Thomas(Cagney)
          //    We simply require the dlat to be (slightly) non-zero, and carry on.
          //    MAS022210 for HamishB from 1e-4 && .001 to 1e-9 for better precision
          //    on small latitude diffs
          const double mlat0 = fabs(lat1 - lat0) < 1e-9 ? lat0 + 1e-9 : lat0;

          double east, north;
          toSM_ECC(lat1, lon1x, mlat0, lon0x, &east, &north);
          const double C = atan2(east, north);
          if(cos(C))
          {
              const double dlat = (lat1 - mlat0) * 60.;              // in minutes
              *dist = (dlat /cos(C));
          }
          else
          {
              *dist = DistGreatCircle(lat0, lon0, lat1, lon1);
          }
      }

      //    Calculate the bearing using the un-adjusted original latitudes and Mercator Sailing
      if(brg)
      {
          double east, north;
          toSM_ECC(lat1, lon1x, lat0, lon0x, &east, &north);

          const double C = atan2(east, north);
          const double brgt = 180. + (C * 180. / PI);
          if (brgt < 0)
              *brg = brgt + 360.;
          else if (brgt >= 360.)
              *brg = brgt - 360.;
          else
              *brg = brgt;
      }
}


/* --------------------------------------------------------------------------------- */
/*
// Given the lat/long of starting point and ending point,
// calculates the distance along a geodesic curve, using elliptic earth model.
returns the distance in meters
*/
/* --------------------------------------------------------------------------------- */
/* Call example:
double d5 = DistGreatCircle( gLat, gLon, pActivePoint->m_lat, pActivePoint->m_lon );
        CurrentRngToActivePoint = d5;
*/

double DistGreatCircle(double slat, double slon, double dlat, double dlon)
{
//    double th1,costh1,sinth1,sina12,cosa12,M,N,c1,c2,D,P,s1;
//    int merid, signS;
    
    /*   Input/Output from geodesic functions   */
    double al12;           /* Forward azimuth */
    double al21;           /* Back azimuth    */
    double geod_S;         /* Distance        */
    double phi1, lam1, phi2, lam2;
    
    int ellipse;
    double geod_f;
    double geod_a;
    double es, onef, f, f64, f2, f4;
    
    double d5;
    phi1 = slat * DEGREE;
    lam1 = slon * DEGREE;
    phi2 = dlat * DEGREE;
    lam2 = dlon * DEGREE;
    
    //void geod_inv(struct georef_state *state)
    {
        double      th1,th2,thm,dthm,dlamm,dlam,sindlamm,costhm,sinthm,cosdthm,
        sindthm,L,E,cosd,d,X,Y,T,sind,tandlammp,u,v,D,A,B;
        
        
        /*   Stuff the WGS84 projection parameters as necessary
         *      To avoid having to include <geodesic,h>
         */
        
        ellipse = 1;
        f = 1.0 / WGSinvf;       /* WGS84 ellipsoid flattening parameter */
        geod_a = WGS84_semimajor_axis_meters;
        
        es = 2 * f - f * f;
        onef = sqrt(1. - es);
        geod_f = 1 - onef;
        f2 = geod_f/2;
        f4 = geod_f/4;
        f64 = geod_f*geod_f/64;
        
        
        if (ellipse) {
            th1 = atan(onef * tan(phi1));
            th2 = atan(onef * tan(phi2));
        } else {
            th1 = phi1;
            th2 = phi2;
        }
        thm = .5 * (th1 + th2);
        dthm = .5 * (th2 - th1);
        dlamm = .5 * ( dlam = adjlon(lam2 - lam1) );
        if (fabs(dlam) < DTOL && fabs(dthm) < DTOL) {
            al12 =  al21 = geod_S = 0.;
            return 0.0;
        }
        sindlamm = sin(dlamm);
        costhm = cos(thm);      sinthm = sin(thm);
        cosdthm = cos(dthm);    sindthm = sin(dthm);
        L = sindthm * sindthm + (cosdthm * cosdthm - sinthm * sinthm)
            * sindlamm * sindlamm;
        d = acos(cosd = 1 - L - L);
        if (ellipse) {
              E = cosd + cosd;
              sind = sin( d );
              Y = sinthm * cosdthm;
              Y *= (Y + Y) / (1. - L);
              T = sindthm * costhm;
              T *= (T + T) / L;
              X = Y + T;
              Y -= T;
              T = d / sind;
              D = 4. * T * T;
              A = D * E;
              B = D + D;
              geod_S = geod_a * sind * (T - f4 * (T * X - Y) +
                          f64 * (X * (A + (T - .5 * (A - E)) * X) -
                          Y * (B + E * Y) + D * X * Y));
              tandlammp = tan(.5 * (dlam - .25 * (Y + Y - E * (4. - X)) *
                          (f2 * T + f64 * (32. * T - (20. * T - A)
                          * X - (B + 4.) * Y)) * tan(dlam)));
        } else {
            geod_S = geod_a * d;
            tandlammp = tan(dlamm);
        }
        u = atan2(sindthm , (tandlammp * costhm));
        v = atan2(cosdthm , (tandlammp * sinthm));
        al12 = adjlon(TWOPI + v - u);
        al21 = adjlon(TWOPI - v - u);
    }
  
    // d5 = geod_S / 1852.0;
    d5 = geod_S;  // pass back the meters
    
    return d5;
}

/*============================================================================*/
// DXT Functions
// #include <stdint.h>
#include <stddef.h>
#include <assert.h>

/*
DXT1/DXT3/DXT5 texture decompression

The original code is from Benjamin Dobell, see below for details. Compared to
the original the code is now valid C89, has support for 64-bit architectures
and has been refactored. It also has support for additional formats and uses
a different PackRGBA order.

---

Copyright (c) 2012 - 2015 Matthus G. "Anteru" Chajdas (http://anteru.net)

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
of the Software, and to permit persons to whom the Software is furnished to do 
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.

---

Copyright (C) 2009 Benjamin Dobell, Glass Echidna

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
of the Software, and to permit persons to whom the Software is furnished to do 
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.

---
*/
static uint32_t PackRGBA (uint8_t r, uint8_t g, uint8_t b, uint8_t a)
{
	return r | (g << 8) | (b << 16) | (a << 24);
}

static float Int8ToFloat_SNORM (const uint8_t input)
{
	return (float)((int8_t)input) / 127.0f;
}

static float Int8ToFloat_UNORM (const uint8_t input)
{
	return (float)input / 255.0f;
}

/**
Decompress a BC 16x3 index block stored as
h g f e
d c b a
p o n m
l k j i

Bits packed as

| h | g | f | e | d | c | b | a | // Entry
|765 432 107 654 321 076 543 210| // Bit
|0000000000111111111112222222222| // Byte

into 16 8-bit indices.
*/
static void Decompress16x3bitIndices (const uint8_t* packed, uint8_t* unpacked)
{
	uint32_t tmp, block, i;

	for (block = 0; block < 2; ++block) {
		tmp = 0;

		// Read three bytes
		for (i = 0; i < 3; ++i) {
			tmp |= ((uint32_t)packed [i]) << (i * 8);
		}

		// Unpack 8x3 bit from last 3 byte block
		for (i = 0; i < 8; ++i) {
			unpacked [i] = (tmp >> (i*3)) & 0x7;
		}

		packed += 3;
		unpacked += 8;
	}
}

static void DecompressBlockBC1Internal (const uint8_t* block,
	unsigned char* output, uint32_t outputStride, const uint8_t* alphaValues)
{
	uint32_t temp, code;

	uint16_t color0, color1;
	uint8_t r0, g0, b0, r1, g1, b1;

	int i, j;

	color0 = *(const uint16_t*)(block);
	color1 = *(const uint16_t*)(block + 2);

	temp = (color0 >> 11) * 255 + 16;
	r0 = (uint8_t)((temp/32 + temp)/32);
	temp = ((color0 & 0x07E0) >> 5) * 255 + 32;
	g0 = (uint8_t)((temp/64 + temp)/64);
	temp = (color0 & 0x001F) * 255 + 16;
	b0 = (uint8_t)((temp/32 + temp)/32);

	temp = (color1 >> 11) * 255 + 16;
	r1 = (uint8_t)((temp/32 + temp)/32);
	temp = ((color1 & 0x07E0) >> 5) * 255 + 32;
	g1 = (uint8_t)((temp/64 + temp)/64);
	temp = (color1 & 0x001F) * 255 + 16;
	b1 = (uint8_t)((temp/32 + temp)/32);

	code = *(const uint32_t*)(block + 4);

	if (color0 > color1) {
		for (j = 0; j < 4; ++j) {
			for (i = 0; i < 4; ++i) {
				uint32_t finalColor, positionCode;
				uint8_t alpha;

				alpha = alphaValues [j*4+i];

				finalColor = 0;
				positionCode = (code >>  2*(4*j+i)) & 0x03;

				switch (positionCode) {
				case 0:
					finalColor = PackRGBA(r0, g0, b0, alpha);
					break;
				case 1:
					finalColor = PackRGBA(r1, g1, b1, alpha);
					break;
				case 2:
					finalColor = PackRGBA((2*r0+r1)/3, (2*g0+g1)/3, (2*b0+b1)/3, alpha);
					break;
				case 3:
					finalColor = PackRGBA((r0+2*r1)/3, (g0+2*g1)/3, (b0+2*b1)/3, alpha);
					break;
				}

				*(uint32_t*)(output + j*outputStride + i * sizeof (uint32_t)) = finalColor;
			}
		}
	} else {
		for (j = 0; j < 4; ++j) {
			for (i = 0; i < 4; ++i) {
				uint32_t finalColor, positionCode;
				uint8_t alpha;

				alpha = alphaValues [j*4+i];

				finalColor = 0;
				positionCode = (code >>  2*(4*j+i)) & 0x03;

				switch (positionCode) {
				case 0:
					finalColor = PackRGBA(r0, g0, b0, alpha);
					break;
				case 1:
					finalColor = PackRGBA(r1, g1, b1, alpha);
					break;
				case 2:
					finalColor = PackRGBA((r0+r1)/2, (g0+g1)/2, (b0+b1)/2, alpha);
					break;
				case 3:
					finalColor = PackRGBA(0, 0, 0, alpha);
					break;
				}

				*(uint32_t*)(output + j*outputStride + i * sizeof (uint32_t)) = finalColor;
			}
		}
	}
}

/*
Decompresses one block of a BC1 (DXT1) texture and stores the resulting pixels at the appropriate offset in 'image'.

uint32_t x:						x-coordinate of the first pixel in the block.
uint32_t y:						y-coordinate of the first pixel in the block.
uint32_t stride:				stride of a scanline in bytes.
const uint8_t* blockStorage:	pointer to the block to decompress.
uint32_t* image:				pointer to image where the decompressed pixel data should be stored.
*/
void DecompressBlockBC1 (uint32_t x, uint32_t y, uint32_t stride,
	const uint8_t* blockStorage, unsigned char* image)
{
	static const uint8_t const_alpha [] = {
		255, 255, 255, 255,
		255, 255, 255, 255,
		255, 255, 255, 255,
		255, 255, 255, 255
	};

	DecompressBlockBC1Internal (blockStorage,
		image + x * sizeof (uint32_t) + (y * stride), stride, const_alpha);
}

/*
Decompresses one block of a BC3 (DXT5) texture and stores the resulting pixels at the appropriate offset in 'image'.

uint32_t x:						x-coordinate of the first pixel in the block.
uint32_t y:						y-coordinate of the first pixel in the block.
uint32_t stride:				stride of a scanline in bytes.
const uint8_t *blockStorage:	pointer to the block to decompress.
uint32_t *image:				pointer to image where the decompressed pixel data should be stored.
*/
void DecompressBlockBC3 (uint32_t x, uint32_t y, uint32_t stride,
	const uint8_t* blockStorage, unsigned char* image)
{
	uint8_t alpha0, alpha1;
	uint8_t alphaIndices [16];

	uint16_t color0, color1;
	uint8_t r0, g0, b0, r1, g1, b1;

	int i, j;

	uint32_t temp, code;

	alpha0 = *(blockStorage);
	alpha1 = *(blockStorage + 1);

	Decompress16x3bitIndices (blockStorage + 2, alphaIndices);

	color0 = *(const uint16_t*)(blockStorage + 8);
	color1 = *(const uint16_t*)(blockStorage + 10);

	temp = (color0 >> 11) * 255 + 16;
	r0 = (uint8_t)((temp / 32 + temp) / 32);
	temp = ((color0 & 0x07E0) >> 5) * 255 + 32;
	g0 = (uint8_t)((temp / 64 + temp) / 64);
	temp = (color0 & 0x001F) * 255 + 16;
	b0 = (uint8_t)((temp / 32 + temp) / 32);

	temp = (color1 >> 11) * 255 + 16;
	r1 = (uint8_t)((temp / 32 + temp) / 32);
	temp = ((color1 & 0x07E0) >> 5) * 255 + 32;
	g1 = (uint8_t)((temp / 64 + temp) / 64);
	temp = (color1 & 0x001F) * 255 + 16;
	b1 = (uint8_t)((temp / 32 + temp) / 32);

	code = *(const uint32_t*)(blockStorage + 12);

	for (j = 0; j < 4; j++) {
		for (i = 0; i < 4; i++) {
			uint8_t finalAlpha;
			int alphaCode;
			uint8_t colorCode;
			uint32_t finalColor;

			alphaCode = alphaIndices [4 * j + i];

			if (alphaCode == 0) {
				finalAlpha = alpha0;
			} else if (alphaCode == 1) {
				finalAlpha = alpha1;
			} else {
				if (alpha0 > alpha1) {
					finalAlpha = (uint8_t)(((8 - alphaCode)*alpha0 + (alphaCode - 1)*alpha1) / 7);
				} else {
					if (alphaCode == 6) {
						finalAlpha = 0;
					} else if (alphaCode == 7) {
						finalAlpha = 255;
					} else {
						finalAlpha = (uint8_t)(((6 - alphaCode)*alpha0 + (alphaCode - 1)*alpha1) / 5);
					}
				}
			}

			colorCode = (code >> 2 * (4 * j + i)) & 0x03;
			finalColor = 0;

			switch (colorCode) {
			case 0:
				finalColor = PackRGBA (r0, g0, b0, finalAlpha);
				break;
			case 1:
				finalColor = PackRGBA (r1, g1, b1, finalAlpha);
				break;
			case 2:
				finalColor = PackRGBA ((2 * r0 + r1) / 3, (2 * g0 + g1) / 3, (2 * b0 + b1) / 3, finalAlpha);
				break;
			case 3:
				finalColor = PackRGBA ((r0 + 2 * r1) / 3, (g0 + 2 * g1) / 3, (b0 + 2 * b1) / 3, finalAlpha);
				break;
			}


			*(uint32_t*)(image + sizeof (uint32_t) * (i + x) + (stride * (y + j))) = finalColor;
		}
	}
}

/*
Decompresses one block of a BC2 (DXT3) texture and stores the resulting pixels at the appropriate offset in 'image'.

uint32_t x:						x-coordinate of the first pixel in the block.
uint32_t y:						y-coordinate of the first pixel in the block.
uint32_t stride:				stride of a scanline in bytes.
const uint8_t *blockStorage:	pointer to the block to decompress.
uint32_t *image:				pointer to image where the decompressed pixel data should be stored.
*/
void DecompressBlockBC2 (uint32_t x, uint32_t y, uint32_t stride,
	const uint8_t* blockStorage, unsigned char* image)
{
	int i;

	uint8_t alphaValues [16] = { 0 };

	for (i = 0; i < 4; ++i) {
		const uint16_t* alphaData = (const uint16_t*)(blockStorage);

		alphaValues [i * 4 + 0] = (((*alphaData) >> 0) & 0xF) * 17;
		alphaValues [i * 4 + 1] = (((*alphaData) >> 4) & 0xF) * 17;
		alphaValues [i * 4 + 2] = (((*alphaData) >> 8) & 0xF) * 17;
		alphaValues [i * 4 + 3] = (((*alphaData) >> 12) & 0xF) * 17;

		blockStorage += 2;
	}

	DecompressBlockBC1Internal (blockStorage,
		image + x * sizeof (uint32_t) + (y * stride), stride, alphaValues);
}

enum BC4Mode
{
	BC4_UNORM = 0,
	BC4_SNORM = 1
};

enum BC5Mode
{
	BC5_UNORM = 0,
	BC5_SNORM = 1
};

static void DecompressBlockBC4Internal (
	const uint8_t* block, unsigned char* output,
	uint32_t outputStride, const float* colorTable)
{
	uint8_t indices [16];
	int x, y;

	Decompress16x3bitIndices (block + 2, indices);

	for (y = 0; y < 4; ++y) {
		for (x = 0; x < 4; ++x) {
			*(float*)(output + x * sizeof (float)) = colorTable [indices [y*4 + x]];
		}

		output += outputStride;
	}
}

/*
Decompresses one block of a BC4 texture and stores the resulting pixels at the appropriate offset in 'image'.

uint32_t x:						x-coordinate of the first pixel in the block.
uint32_t y:						y-coordinate of the first pixel in the block.
uint32_t stride:				stride of a scanline in bytes.
const uint8_t* blockStorage:	pointer to the block to decompress.
float* image:					pointer to image where the decompressed pixel data should be stored.
*/
void DecompressBlockBC4 (uint32_t x, uint32_t y, uint32_t stride, enum BC4Mode mode,
	const uint8_t* blockStorage, unsigned char* image)
{
	float colorTable [8];
	float r0, r1;

	if (mode == BC4_UNORM) {
		r0 = Int8ToFloat_UNORM (blockStorage [0]);
		r1 = Int8ToFloat_UNORM (blockStorage [1]);

		colorTable [0] = r0;
		colorTable [1] = r1;

		if (r0 > r1) {
			// 6 interpolated color values
			colorTable [2] = (6*r0 + 1*r1)/7.0f; // bit code 010
			colorTable [3] = (5*r0 + 2*r1)/7.0f; // bit code 011
			colorTable [4] = (4*r0 + 3*r1)/7.0f; // bit code 100
			colorTable [5] = (3*r0 + 4*r1)/7.0f; // bit code 101
			colorTable [6] = (2*r0 + 5*r1)/7.0f; // bit code 110
			colorTable [7] = (1*r0 + 6*r1)/7.0f; // bit code 111
		} else {
			// 4 interpolated color values
			colorTable [2] = (4*r0 + 1*r1)/5.0f; // bit code 010
			colorTable [3] = (3*r0 + 2*r1)/5.0f; // bit code 011
			colorTable [4] = (2*r0 + 3*r1)/5.0f; // bit code 100
			colorTable [5] = (1*r0 + 4*r1)/5.0f; // bit code 101
			colorTable [6] = 0.0f;               // bit code 110
			colorTable [7] = 1.0f;               // bit code 111
		}
	} else if (mode == BC4_SNORM) {
		r0 = Int8ToFloat_SNORM (blockStorage [0]);
		r1 = Int8ToFloat_SNORM (blockStorage [1]);

		colorTable [0] = r0;
		colorTable [1] = r1;

		if (r0 > r1) {
		  // 6 interpolated color values
		  colorTable [2] = (6*r0 + 1*r1)/7.0f; // bit code 010
		  colorTable [3] = (5*r0 + 2*r1)/7.0f; // bit code 011
		  colorTable [4] = (4*r0 + 3*r1)/7.0f; // bit code 100
		  colorTable [5] = (3*r0 + 4*r1)/7.0f; // bit code 101
		  colorTable [6] = (2*r0 + 5*r1)/7.0f; // bit code 110
		  colorTable [7] = (1*r0 + 6*r1)/7.0f; // bit code 111
		} else {
		  // 4 interpolated color values
		  colorTable [2] = (4*r0 + 1*r1)/5.0f; // bit code 010
		  colorTable [3] = (3*r0 + 2*r1)/5.0f; // bit code 011
		  colorTable [4] = (2*r0 + 3*r1)/5.0f; // bit code 100
		  colorTable [5] = (1*r0 + 4*r1)/5.0f; // bit code 101
		  colorTable [6] = -1.0f;              // bit code 110
		  colorTable [7] =  1.0f;              // bit code 111
		}
	}

	DecompressBlockBC4Internal (blockStorage,
		image + x * sizeof (float) + (y * stride), stride, colorTable);
}


/*
Decompresses one block of a BC5 texture and stores the resulting pixels at the appropriate offset in 'image'.

uint32_t x:						x-coordinate of the first pixel in the block.
uint32_t y:						y-coordinate of the first pixel in the block.
uint32_t stride:				stride of a scanline in bytes.
const uint8_t* blockStorage:	pointer to the block to decompress.
float* image:					pointer to image where the decompressed pixel data should be stored.
*/
void DecompressBlockBC5 (uint32_t x, uint32_t y, uint32_t stride, enum BC5Mode mode,
	const uint8_t* blockStorage, unsigned char* image)
{
	// We decompress the two channels separately and interleave them when
	// writing to the output
	float c0 [16];
	float c1 [16];

	int dx, dy;

	DecompressBlockBC4 (0, 0, 4 * sizeof (float), (enum BC4Mode)mode, 
		blockStorage, (unsigned char*)c0);
	DecompressBlockBC4 (0, 0, 4 * sizeof (float), (enum BC4Mode)mode, 
		blockStorage + 8, (unsigned char*)c1);

	for (dy = 0; dy < 4; ++dy) {
		for (dx = 0; dx < 4; ++dx) {
			*(float*)(image + stride * y + (x + dx + 0) * sizeof (float) * 2) = c0 [dy * 4 + dx];
			*(float*)(image + stride * y + (x + dx + 1) * sizeof (float) * 2) = c1 [dy * 4 + dx];
		}
	}
}

// END DXT

/* ===========================================================================*/

// ======================= Unix/DOS time-conversion stuff =====================

/***************** timet2filetime() ******************
 * Converts a Unix time_t to a Windows FILETIME.
 */

static void timet2filetime(FILETIME *ft, const lutime_t t)
{
	LONGLONG i;

	i = Int32x32To64(t, 10000000) + 116444736000000000;
	ft->dwLowDateTime = (DWORD)i;
	ft->dwHighDateTime = (DWORD)(i >>32);
}

/*************** dosdatetime2filetime() ***************
 * Converts a DOS timestamp (date in high word, time in
 * low word) to a Windows FILETIME.
 */

static void dosdatetime2filetime(FILETIME *ft, DWORD dosdate, DWORD dostime)
{
	SYSTEMTIME		st;

	// date: bits 0-4 are day of month 1-31. Bits 5-8 are month 1..12. Bits 9-15 are year-1980
	// time: bits 0-4 are seconds/2, bits 5-10 are minute 0..59. Bits 11-15 are hour 0..23
	st.wYear = (WORD)(((dosdate>>9) & 0x7f) + 1980);
	st.wMonth = (WORD)((dosdate>>5) & 0xf);
	st.wDay = (WORD)(dosdate & 0x1f);
	st.wHour = (WORD)((dostime >> 11) & 0x1f);
	st.wMinute = (WORD)((dostime >> 5) & 0x3f);
	st.wSecond = (WORD)((dostime & 0x1f) * 2);
	st.wMilliseconds = 0;

	SystemTimeToFileTime(&st, ft);
}

/************************ ucrc32() ********************
 * Computes the CRC-32 of the bytes in the specified
 * buffer.
 *
 * crc =	Initial CRC-32 value.
 * buf =	Pointer to buffer containing bytes.
 * len =	Length of "buf" in bytes.
 *
 * RETURNS: The updated CRC-32.
 *
 * Copyright (C) 1995-1998 Mark Adler. For conditions of
 * distribution and use, see copyright notice in zlib.h
 */


#define CRC_DO1(buf) crc = Crc_table[((int)crc ^ (*buf++)) & 0xff] ^ (crc >> 8);
#define CRC_DO2(buf)  CRC_DO1(buf); CRC_DO1(buf);
#define CRC_DO4(buf)  CRC_DO2(buf); CRC_DO2(buf);
#define CRC_DO8(buf)  CRC_DO4(buf); CRC_DO4(buf);

static ULG ucrc32(ULG crc, const UCH *buf, DWORD len)
{
	crc = crc ^ 0xffffffffL;
	while (len >= 8)
	{
		CRC_DO8(buf);
		len -= 8;
	}
	if (len)
	{
		do
		{
			CRC_DO1(buf);
		} while (--len);
	}
	return(crc ^ 0xffffffffL);
}

// Does decryption of encrypted data
#define CRC32(c, b) (Crc_table[((int)(c)^(b))&0xff]^((c)>>8)) 
static void Uupdate_keys(unsigned long *keys, char c)
{
	keys[0] = CRC32(keys[0], c);
	keys[1] += keys[0] & 0xFF;
	keys[1] = keys[1]*134775813L +1;
	keys[2] = CRC32(keys[2], keys[1] >> 24);
}

#define IM_NEEDBYTE {if (z->avail_in == 0) return r; r=f;}
#define IM_NEXTBYTE (--z->avail_in, ++z->total_in, *(z->next_in)++)

/********************** seekInZip() *********************
 * Sets the current "file position" within the ZIP archive.
 */

static int seekInZip(TUNZIP *tunzip, long offset, DWORD reference)
{
	if (!(tunzip->Flags & TZIP_ARCMEMORY))
	{
#ifdef WIN32
		if (reference == FILE_BEGIN) offset = SetFilePointer(tunzip->ArchivePtr, tunzip->InitialArchiveOffset + offset, 0, FILE_BEGIN);
		else if (reference == FILE_CURRENT) offset = SetFilePointer(tunzip->ArchivePtr, offset, 0, FILE_CURRENT);
//		else if (reference == FILE_END) offset = SetFilePointer(tunzip->ArchivePtr, offset, 0, FILE_END);
#else
		if (reference == FILE_BEGIN) offset = lseek((int)tunzip->ArchivePtr, tunzip->InitialArchiveOffset + offset, SEEK_SET);
		else if (reference == FILE_CURRENT) offset = lseek((int)tunzip->ArchivePtr, offset, SEEK_CUR);
//		else if (reference == FILE_END) offset = lseek((int)tunzip->ArchivePtr, offset, SEEK_END);
#endif
		if (offset == -1) return(ZR_SEEK);
	}
	else
	{
		if (reference == FILE_BEGIN) tunzip->ArchiveBufPos = offset;
		else if (reference == FILE_CURRENT) tunzip->ArchiveBufPos += offset;
//		else if (reference == FILE_END) tunzip->ArchiveBufPos = tunzip->ArchiveBufLen + offset;
//		if (tunzip->ArchiveBufPos > tunzip->ArchiveBufLen) tunzip->ArchiveBufPos = tunzip->ArchiveBufLen;
	}
	return(0);
}

/********************** readFromZip() *********************
 * Reads the specified number of bytes from the ZIP archive
 * (starting at the current position) and copies them to
 * the specified buffer.
 *
 * RETURNS: The number of bytes actually read (could be less
 * than the requested number if the end of file is reached)
 * or 0 if no more bytes to read (or an error).
 *
 * NOTE: If an error, than TUNZIP->LastErr is non-zero.
 */

static DWORD readFromZip(TUNZIP *tunzip, void *ptr, DWORD toread)
{
	// Reading from a handle?
	if (!(tunzip->Flags & TZIP_ARCMEMORY))
	{
		DWORD	readb;

#ifdef WIN32
		if (!ReadFile(tunzip->ArchivePtr, ptr, toread, &readb, 0))
#else
		if ((readb = read((int)tunzip->ArchivePtr, ptr, toread)) == (DWORD)-1)
#endif
		{
			tunzip->LastErr = ZR_READ;
			readb = 0;
		}
		return(readb);
	}

	// Reading from memory
	if (tunzip->ArchiveBufPos + toread > tunzip->ArchiveBufLen) toread = tunzip->ArchiveBufLen - tunzip->ArchiveBufPos;
	CopyMemory(ptr, (unsigned char *)tunzip->ArchivePtr + tunzip->ArchiveBufPos, toread);
	tunzip->ArchiveBufPos += toread;
	return(toread);
}


static unsigned char * reformat_short(unsigned char *ptr)
{
 
   // Do nuthin its correct already	
	return(ptr);
}                                     

// Reads a short in LSB order from the given file.
static unsigned short getArchiveShort(TUNZIP *tunzip)
{
	unsigned short		x;

	if (!tunzip->LastErr && readFromZip(tunzip, &x, sizeof(unsigned short)))
		reformat_short((unsigned char *)&x);
	return(x);
}

static unsigned char * reformat_long(unsigned char *ptr)
{

   // Do nothing its OK as it is
	return(ptr);
}

// Reads a long in LSB order from the given file
static ULG getArchiveLong(TUNZIP *tunzip)
{
	ULG		x;

	x = 0;
	if (!tunzip->LastErr && readFromZip(tunzip, &x, sizeof(ULG)))
		reformat_long((unsigned char *)&x);
	return(x);
}

/********************* getEntryFN() *******************
 * Gets the currently selected entry's filename in the
 * zip archive.
 *
 * TUNZIP must be initialized to contain information about
 * that entry via a call to goToFirstEntry() or goToNextEntry(),
 * and getEntryInfo() must be called prior to this.
 *
 * NOTE: If an error, than TUNZIP->LastErr is set non-zero,
 * so caller must clear it first if needed.
 */

static void getEntryFN(TUNZIP *tunzip, char *szFileName)
{
	  DWORD		uSizeRead;
	  DWORD		lSeek;

	lSeek = tunzip->CurrentEntryInfo.size_filename;

	// Make sure filename will fit in the supplied buffer (MAX_PATH)
	if (lSeek < MAX_PATH)
		uSizeRead = lSeek;
	else
		uSizeRead = MAX_PATH - 1;

	// Read the filename
	if (uSizeRead && readFromZip(tunzip, szFileName, uSizeRead) != uSizeRead)
bad:	tunzip->LastErr = ZR_CORRUPT;
	else
	{
		// If there are any bytes left that we didn't read, skip them
		lSeek -= uSizeRead;
		if (lSeek && seekInZip(tunzip, lSeek, FILE_CURRENT)) goto bad;
	}

	szFileName[uSizeRead] = 0;
}


/********************* getEntryInfo() *******************
 * Gets info about the currently selected entry in the
 * zip archive.
 *
 * TUNZIP must be initialized to contain information about
 * that entry via a call to goToFirstEntry() or goToNextEntry().
 *
 * NOTE: If an error, than TUNZIP->LastErr is set non-zero,
 * so caller must clear it first.
 */

static void getEntryInfo(  TUNZIP *tunzip)
{
	DWORD			uSizeRead;
	DWORD			lSeek;

	// Seek to the start of this entry's info in the Central Directory
	if (!seekInZip(tunzip, tunzip->CurrEntryPosInCentralDir + tunzip->ByteBeforeZipArchive, FILE_BEGIN))
	{
		if (tunzip->Flags & TZIP_GZIP)
		{
			BYTE	flag, chr;

			ZeroMemory(&tunzip->CurrentEntryInfo, sizeof(ZIPENTRYINFO));

			if (tunzip->Flags & TZIP_RAW)
			{
				tunzip->CurrentEntryInfo.compression_method = Z_DEFLATED;
				tunzip->CurrentEntryInfo.offset = tunzip->CurrEntryPosInCentralDir + tunzip->ByteBeforeZipArchive;
				goto out;
			}

			if (readFromZip(tunzip, &flag, 1) != 1 ||
				readFromZip(tunzip, &tunzip->CurrentEntryInfo.dosDate, 4) != 4 ||
				seekInZip(tunzip, 2, FILE_CURRENT))
			{
				goto bad;
			}

			if ((flag & 0x04) && (readFromZip(tunzip, &tunzip->CurrentEntryInfo.disk_num_start, 2) != 2 || seekInZip(tunzip, tunzip->CurrentEntryInfo.disk_num_start, FILE_CURRENT))) goto bad;

			// Save offset to filename
			if (!(tunzip->Flags & TZIP_ARCMEMORY))
#ifdef WIN32
				uSizeRead = SetFilePointer(tunzip->ArchivePtr, 0, 0, FILE_CURRENT);
#else
				uSizeRead = lseek((int)tunzip->ArchivePtr, 0, SEEK_CUR);
#endif
			else
				uSizeRead = tunzip->ArchiveBufPos;

			// Skip over filename, counting length
			if (flag & 0x08)
			{
				for (;;)
				{
					if (readFromZip(tunzip, &chr, 1) != 1) goto bad;
					if (!chr) break;
					++tunzip->CurrentEntryInfo.size_filename;
				}
			}

			// Skip over comment
			if (flag & 0x10)
			{
				do
				{
					if (readFromZip(tunzip, &chr, 1) != 1) goto bad;
				} while (chr);
			}

			// Skip compress info and OS bytes
			if ((flag & 0x02) && seekInZip(tunzip, 2, FILE_CURRENT)) goto bad;

			tunzip->CurrentEntryInfo.compression_method = Z_DEFLATED;

			if (!(tunzip->Flags & TZIP_ARCMEMORY))
			{
				// Save offset to data
#ifdef WIN32
				tunzip->CurrentEntryInfo.offset = SetFilePointer(tunzip->ArchivePtr, 0, 0, FILE_CURRENT);
				tunzip->CurrentEntryInfo.compressed_size = SetFilePointer(tunzip->ArchivePtr, -8, 0, FILE_END) - tunzip->CurrentEntryInfo.offset;
#else
				tunzip->CurrentEntryInfo.offset = lseek((int)tunzip->ArchivePtr, 0, SEEK_CUR);
				tunzip->CurrentEntryInfo.compressed_size = lseek((int)tunzip->ArchivePtr, -8, SEEK_END) - tunzip->CurrentEntryInfo.offset;
#endif
				tunzip->CurrentEntryInfo.crc = getArchiveLong(tunzip);
				tunzip->CurrentEntryInfo.uncompressed_size = getArchiveLong(tunzip);
			}
			else
			{
				tunzip->CurrentEntryInfo.offset = tunzip->ArchiveBufPos;
				tunzip->CurrentEntryInfo.compressed_size = tunzip->ArchiveBufLen - tunzip->ArchiveBufPos - 8;
				CopyMemory((void *)tunzip->CurrentEntryInfo.crc, (unsigned char *)tunzip->ArchivePtr + tunzip->ArchiveBufLen - 8, 4);
				reformat_long((unsigned char *)&tunzip->CurrentEntryInfo.crc);
				CopyMemory((void *)tunzip->CurrentEntryInfo.uncompressed_size, (unsigned char *)tunzip->ArchivePtr + tunzip->ArchiveBufLen - 4, 4);
				reformat_long((unsigned char *)&tunzip->CurrentEntryInfo.uncompressed_size);
			}

			// Seek to name offset upon return
			seekInZip(tunzip, uSizeRead, FILE_BEGIN);
			goto out;
		}

		// Get the ZIP signature and check it
		uSizeRead = getArchiveLong(tunzip);
		if (uSizeRead == 0x02014b50 &&

			// Read in the ZIPENTRYINFO
			readFromZip(tunzip, &tunzip->CurrentEntryInfo, sizeof(ZIPENTRYINFO)) == sizeof(ZIPENTRYINFO))
		{
			unsigned char	*ptr;
			
			
         
			// Adjust the various fields to this CPU's byte order
			uSizeRead = ZIP_FIELDS_REFORMAT;
			ptr = (unsigned char *)&tunzip->CurrentEntryInfo;
			for (lSeek = 0; lSeek < NUM_FIELDS_REFORMAT; lSeek++)
			{
				if (0x00000001 & uSizeRead) ptr = reformat_long(ptr);
				else ptr = reformat_short(ptr);
				uSizeRead >>= 1;
			}
			 

out:		return;
		}
	}
bad:
	tunzip->LastErr = ZR_CORRUPT;
}

/******************** goToFirstEntry() *******************
 * Set the current entry to the first entry in the ZIP
 * archive.
 */

static void goToFirstEntry(  TUNZIP *tunzip)
{
	if (tunzip->TotalEntries)
	{
		tunzip->CurrEntryPosInCentralDir = tunzip->CentralDirOffset;
		tunzip->CurrentEntryNum = 0;
		getEntryInfo(tunzip);
	}
	else
		tunzip->LastErr = ZR_NOTFOUND;
}

/******************** goToNextEntry() *******************
 * Set the current entry to the next entry in the ZIP
 * archive.
 */

static void goToNextEntry(  TUNZIP *tunzip)
{
	if (tunzip->CurrentEntryNum + 1 < tunzip->TotalEntries)
	{
		tunzip->CurrEntryPosInCentralDir += SIZECENTRALDIRITEM + tunzip->CurrentEntryInfo.size_filename + tunzip->CurrentEntryInfo.size_file_extra + tunzip->CurrentEntryInfo.size_file_comment;
		++tunzip->CurrentEntryNum;
		getEntryInfo(tunzip);
	}
	else
		tunzip->LastErr = ZR_NOTFOUND;
}

/*********************************************************************
 Returns true if the IDL is crossed                                  
/*********************************************************************/ 
bool IsIDLCrossed( double lon1, double lon2) 
{
   int dir = 0;
   bool idl = false;
   
   if (lon1 < 0 && lon2 > 0) dir = 1 ;
   if (lon2 < 0 && lon1 > 0) dir = 2;
   if (lon1 < 0) lon1 = 360 - fabs(lon1);
   if (lon2 < 0) lon2 = 360 - fabs(lon2);
   double diff = fabs(lon1 - lon2);
   if (dir > 0 && diff < 180.0)  idl = true;
 
   return idl;
}


/*********************************************************************
 Returns the difference between latlon1 and latlon2                                 
 the TYPE is true if converting longitude or false for latitude   
/*********************************************************************/ 
double DiffLatLon( BOOL type, double latlon1, double latlon2) 
{
   double diff;
   if (type) // LON 
   {
      bool idl;
      idl = IsIDLCrossed(latlon1,latlon2);
      if (latlon1 < 0) latlon1 = 360 - fabs(latlon1);
      if (latlon2 < 0) latlon2 = 360 - fabs(latlon2);
      diff = latlon1 - latlon2;
      if (diff < 0.0 && (fabs(diff) > 180 ) )
         diff += 360;
      else if (diff > 0.0 && diff > 180.0)
      {   
         diff -= 360;
      }   
      //  If International Date Line is crossed reverse sign
      if (idl)
         diff = diff * -1;  //  reverse sign if idl crossed
   }
   else // LAT
   { 
      latlon1 += 90;  //  make south pole 0 to north pole=180
      latlon2 +=90;   
      diff = latlon2 - latlon1;
   }
   
   return diff;
}

/****************** Get_Pixval ********************************
 returns the pixval of a color 
 Get_Pixval(color)
 **********************************************************/
ULONG Get_Pixval(RGBQUAD color) 
{
	ULONG pixval;
	BYTE red,green,blue;
   
   red = color.rgbRed;
   green = color.rgbGreen;
   blue = color.rgbBlue;
   pixval = red * 65536;
   pixval += green * 256;
   pixval += blue;    
   return pixval;
}
/*******************************************************************
 * Returns the best palette index that matches a specified color.
 ******************************************************************/
uint8_t GetNearestIndex(HANDLE pDib, RGBQUAD c, uint8_t loadix, long x, long y, int chartnum, RexxArrayObject diagmsgs, RexxCallContext * context)
{
   char buffer[512]; 
   int rc;
   // Get the bitmap header and the CXimage info
   BITMAPINFOHEADER head;
	memcpy(&head,pDib, sizeof(head));
   CXIMAGEINFO info;
   memcpy(&info,(uint8_t*)(pDib) + sizeof(BITMAPINFOHEADER), sizeof(info));
	if ((pDib==NULL)||(head.biClrUsed==0)) return 0;

	// <RJ> check matching with the previous result
	if (info.last_c_isvalid && (*(int32_t*)&info.last_c == *(int32_t*)&c)) return info.last_c_index;
	info.last_c = c;
	info.last_c_isvalid = true;

	uint8_t* iDst = (uint8_t*)(pDib) + sizeof(BITMAPINFOHEADER);
	int32_t distance=200000;
	int32_t i,j = 0;
	int32_t k,l,pixval;
	int32_t m = (int32_t)(head.biClrImportant==0 ? head.biClrUsed : head.biClrImportant);
	for(i=0,l=0;i<m;i++,l+=sizeof(RGBQUAD))
   {
      pixval = iDst[l+2] * 65536 + iDst[l+1] * 256 + iDst[l];
      
      k = (iDst[l]-c.rgbBlue)*(iDst[l]-c.rgbBlue)+
			(iDst[l+1]-c.rgbGreen)*(iDst[l+1]-c.rgbGreen)+
			(iDst[l+2]-c.rgbRed)*(iDst[l+2]-c.rgbRed);
      
         // testing
         if (x==679 && y==508 && loadix > 80)
         {
            sprintf(buffer,"GetNearestIndex color=%d i=%d k=%d red=%d green=%d blue=%d pivxal=%d",*(int32_t*)&c,i,k,iDst[l],iDst[l+1], iDst[l+2], pixval);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }
//		k = abs(iDst[l]-c.rgbBlue)+abs(iDst[l+1]-c.rgbGreen)+abs(iDst[l+2]-c.rgbRed);
		if (k==0)
      {
			j=i;
			break;
		}
		if (k<distance)
      {
			distance=k;
			j=i;
		}
	}
   
   // if the loadix = 95 print some info 
   if (x==679 && y==508 && loadix > 80)
   {
      sprintf(buffer,"GetNearestIndex chartnum=%d loadix=%d x=%d y=%d color=%d distance=%d i=%d j=%d k=%d l=%d m=%d",chartnum, loadix, x,y, *(int32_t*)&c, distance, i, j, k, l, m);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      sprintf(buffer,"GetNearestIndex biClrUsed=%d biClrImportant=%d last_c_index=%d",head.biClrUsed, head.biClrImportant, info.last_c_index);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }
   
   info.last_c_index = (uint8_t)j;  // remember last palette index
   
	return (uint8_t)j;
}
/****************** listENT ********************************
 List all the entries in an array 
 of ENTLEN long words where each entry is ENTSIZE, 
 offset is the word offset of the word to sort on
 title is the title for the messages
 palsize is the number of entries in the array
 diagmsg is a pointer to the RexxArray Object 
 returns the numcolors
 eg:
 listENT("DIAG ConvertImage histogram image before sort",pArray, palsize,diagmsgs)
 **********************************************************/
INT listENT( ULONG * pArray, INT palsize, RexxArrayObject diagmsgs, RexxCallContext * context) 
{
	char buffer[512]; 
	int numcolors = 0;
	int rc;
	BYTE red,green,blue;
	ULONG pixval;
   uint32_t idx,newidx;
	for(int count,i=0; i<palsize; i++)
	{
      red = (BYTE)pArray[i*ENTLEN]; 
      green = (BYTE)pArray[i*ENTLEN+1]; 
      blue = (BYTE)pArray[i*ENTLEN+2]; 
      pixval = (LONG)pArray[i*ENTLEN+6]; 
      newidx = (LONG)pArray[i*ENTLEN+5]; 
      idx =   pArray[i*ENTLEN+3]; 
      count = pArray[i*ENTLEN+4]; 
      if (count > 0 ) numcolors++;
      sprintf(buffer,"%d: rgb(%d,%d,%d)(%d) idx=%d  count=%d newidx=%d\n",i,red,green,blue,pixval,idx,count,newidx);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   } 
   return numcolors;
} 
/****************** sortENT ********************************
 Bubble Sorts an array 
 of ENTLEN long words where each entry is ENTSIZE, 
 offset is the word offset of the word to sort on
 palsize is the number of entries in the array
 ascdesc if 0 sort in ascending order, 
 else sort in descending order
 **********************************************************/
static void sortENT(  ULONG * pArray, INT offset, INT palsize, INT ascdesc) 
{
   int exchanges,pass;
   LONG val1,val2;    
   LONG temp[ENTLEN]; 
   
   exchanges = -1;
   pass = 1;
   while (exchanges != 0)
   {
      exchanges = 0;
      for (int i=0; i < palsize-1; i++ )
      {
        val1 =  pArray[i*ENTLEN+offset];  // Get count of first entry
        val2 =  pArray[i*ENTLEN+ENTLEN+offset];  // Get count of 2nd entry
        if (ascdesc == 0)
        {
           if ( val1 > val2 ) // sort in ascending order
           {    //Swap the entries
               memcpy(temp,pArray + i*ENTLEN,ENTSIZE);  // save the current entry
               memcpy(pArray + i*ENTLEN,pArray + (i*ENTLEN+ENTLEN),ENTSIZE);  // copy the next entry into this one
               memcpy(pArray + (i*ENTLEN+ENTLEN),temp,ENTSIZE);  // copy the saved entry into the next
               exchanges++;
           }
        }
        else   // sort in descending order
        {
           if ( val1 < val2 )
           {    //Swap the entries
               memcpy(temp,pArray + i*ENTLEN,ENTSIZE);  // save the current entry
               memcpy(pArray + i*ENTLEN,pArray + (i*ENTLEN+ENTLEN),ENTSIZE);  // copy the next entry into this one
               memcpy(pArray + (i*ENTLEN+ENTLEN),temp,ENTSIZE);  // copy the saved entry into the next
               exchanges++;
           }            
        }
      }
      pass++;
   } 
}
/****************** cleanupEntry() ******************
 * Frees resources allocated by initEntry().
 *
 * NOTE: Must not alter TUNZIP->LastErr!
 */

static void cleanupEntry(  TUNZIP * tunzip)
{
	// Free the input buffer
	if (tunzip->EntryReadVars.InputBuffer) GlobalFree(tunzip->EntryReadVars.InputBuffer);
	tunzip->EntryReadVars.InputBuffer = 0;

	// Free stuff allocated for decompressing in DEFLATE mode
	if (tunzip->EntryReadVars.stream.state) inflateEnd((z_streamp)&tunzip->EntryReadVars);
	tunzip->EntryReadVars.stream.state = 0;

	// No currently selected entry
	tunzip->CurrentEntryNum = (DWORD)-1;
}

/********************** initEntry() *********************
 * Initializes structures in preparation of unzipping
 * the current entry.
 *
 * NOTE: Sets TUNZIP->LastErr non-zero if an error, so this
 * must be cleared before calling.
 */


static void initEntry(  TUNZIP *tunzip, ZIPENTRY *ze)
{
	DWORD	offset;
	int err;  
   
	// Clear out the ENTRYREADVARS struct
	ZeroMemory(&tunzip->EntryReadVars, sizeof(ENTRYREADVARS));

	// Get a read buffer to input, and decrypt, bytes from the ZIP archive
	if (!(tunzip->EntryReadVars.InputBuffer = (unsigned char *)GlobalAlloc(GMEM_FIXED, UNZ_BUFSIZE)))
	{
		tunzip->LastErr = ZR_NOALLOC;
badout:	cleanupEntry(tunzip);
		return;
	}

	// If DEFLATE compression method, we need to get some resources for the deflate routines
	if (tunzip->CurrentEntryInfo.compression_method)
	{  
      tunzip->EntryReadVars.stream.zalloc = (alloc_func)0;
      tunzip->EntryReadVars.stream.zfree = (free_func)0;
      tunzip->EntryReadVars.stream.opaque = (voidpf)0;
      tunzip->EntryReadVars.stream.next_in  = tunzip->EntryReadVars.InputBuffer;
      tunzip->EntryReadVars.stream.next_out = Z_NULL;
      tunzip->EntryReadVars.stream.avail_in = 0; 
      err = inflateInit2((z_streamp)&tunzip->EntryReadVars.stream, -MAX_WBITS);

#ifndef NDEBUG
		LuTracev((stderr, "inflate: allocated\n"));
#endif
	}

	// If raw mode, app must supply the compressed and uncompressed sizes
	if (tunzip->Flags & TZIP_RAW)
	{
		tunzip->CurrentEntryInfo.compressed_size = ze->CompressedSize;
		tunzip->CurrentEntryInfo.uncompressed_size = ze->UncompressedSize;
	}

	// Initialize running counts
	tunzip->EntryReadVars.RemainingCompressed = tunzip->CurrentEntryInfo.compressed_size;
	tunzip->EntryReadVars.RemainingUncompressed = tunzip->CurrentEntryInfo.uncompressed_size;

	// Initialize for CRC checksum
	if (tunzip->CurrentEntryInfo.flag & 8) tunzip->EntryReadVars.CrcEncTest = (char)((tunzip->CurrentEntryInfo.dosDate >> 8) & 0xff);
	else tunzip->EntryReadVars.CrcEncTest = (char)(tunzip->CurrentEntryInfo.crc >> 24);

	if (tunzip->Flags & TZIP_GZIP)
		offset = tunzip->CurrentEntryInfo.offset;
	else
	{
		{
		// Initialize encryption stuff
		  const unsigned char	*cp;

		// Entry is encrypted?
		if (tunzip->CurrentEntryInfo.flag & 1)
		{
			tunzip->EntryReadVars.RemainingEncrypt = 12;	// There is an additional 12 bytes at the beginning of each local header
			tunzip->EntryReadVars.Keys[0] = 305419896L;
			tunzip->EntryReadVars.Keys[1] = 591751049L;
			tunzip->EntryReadVars.Keys[2] = 878082192L;
			if ((cp = tunzip->Password))
			{
				while(*cp) Uupdate_keys(&tunzip->EntryReadVars.Keys[0], *(cp)++);
			}
		}
		}

		{
		unsigned short	extra_offset;

		// Seek to the entry's LOCALHEADER->extra_field_size
		if (seekInZip(tunzip, tunzip->CurrentEntryInfo.offset + tunzip->ByteBeforeZipArchive + 28, FILE_BEGIN) ||
			!readFromZip(tunzip, &extra_offset, sizeof(unsigned short)))
		{
badseek:	tunzip->LastErr = ZR_READ;
			goto badout;
		}

		// Get the offset to where the entry's compressed data starts within the archive
//		tunzip->EntryReadVars.PosInArchive = (DWORD)tunzip->CurrentEntryInfo.offset + SIZEZIPLOCALHEADER + (DWORD)tunzip->CurrentEntryInfo.size_filename + (DWORD)extra_offset;
		offset = (DWORD)tunzip->CurrentEntryInfo.offset + SIZEZIPLOCALHEADER + (DWORD)tunzip->CurrentEntryInfo.size_filename + (DWORD)extra_offset;
		}
	}

	// Seek so that a subsequent call to readEntry() will be at the correct position. (ie, We assume that
	// the DEFLATE routines will not be seeking around within the archive while the entry is decompressed)
//	if (seekInZip(tunzip, tunzip->EntryReadVars.PosInArchive, FILE_BEGIN)) goto badseek;
	if (seekInZip(tunzip, offset, FILE_BEGIN)) goto badseek;
}

/************************* readEntry() **********************
 * Reads bytes from the current position of the currently
 * selected entry (within the archive), unencrypts them,
 * and decompresses them into the passed buffer.
 *
 * buf =	Pointer to buffer where data must be copied.
 * len =	The size of buf.
 *
 * RETURNS: The number of byte copied. Could be 0 if the
 * end of file was reached. Sets TUNZIP->LastErr if an
 * error.
 */

// DWORD readEntry(  TUNZIP *tunzip, void *buf, DWORD len) PH99
DWORD readEntry(TUNZIP *tunzip, void *buf, DWORD len)
{
	int							err;
	DWORD						iRead;

	iRead = 0;
	tunzip->EntryReadVars.stream.next_out = (UCH *)buf;

	// Caller is asking for more bytes than remaining? If so, just give him the
	// remaining bytes
	if (len > tunzip->EntryReadVars.RemainingUncompressed) len = tunzip->EntryReadVars.RemainingUncompressed;
	tunzip->EntryReadVars.stream.avail_out = len;

	// More room in the output buffer?
	while (tunzip->EntryReadVars.stream.avail_out)
	{
		// Input buffer is empty? Fill it
		if (!tunzip->EntryReadVars.stream.avail_in && tunzip->EntryReadVars.RemainingCompressed)
		{
			DWORD	uReadThis;

			// Fill up the input buffer, or read as much as is available
			uReadThis = UNZ_BUFSIZE;
			if (uReadThis > tunzip->EntryReadVars.RemainingCompressed) uReadThis = tunzip->EntryReadVars.RemainingCompressed;

			// No more input (ie, done decompressing the entry)?
			if (!uReadThis)
			{
				// Check that checksum works out to what we expected
				if (tunzip->EntryReadVars.RunningCrc != tunzip->CurrentEntryInfo.crc)
					tunzip->LastErr = ZR_FLATE;
none:			return(0);
			}

			// Seek to where we last left off in the ZIP archive and fill the input buffer
			if (!readFromZip(tunzip, tunzip->EntryReadVars.InputBuffer, uReadThis))
//			if (seekInZip(tunzip, tunzip->EntryReadVars.PosInArchive + tunzip->ByteBeforeZipArchive, FILE_BEGIN) ||
//				!readFromZip(tunzip, tunzip->EntryReadVars.InputBuffer, uReadThis))
			{
				tunzip->LastErr = ZR_READ;
				goto none;
			}

			// Increment current position within archive
//			tunzip->EntryReadVars.PosInArchive += uReadThis;

			// Decrement remaining bytes to be decompressed
			tunzip->EntryReadVars.RemainingCompressed -= uReadThis;

			tunzip->EntryReadVars.stream.next_in = tunzip->EntryReadVars.InputBuffer;
			tunzip->EntryReadVars.stream.avail_in = uReadThis;

			// If encryption was applied, then decrypt the bytes
			if (tunzip->CurrentEntryInfo.flag & 1)
			{
				DWORD	i;

				for (i=0; i < uReadThis; i++)
				{
					DWORD	temp;

					temp = ((DWORD)tunzip->EntryReadVars.Keys[2] & 0xffff) | 2;
					tunzip->EntryReadVars.InputBuffer[i] ^= (char)(((temp * (temp ^ 1)) >> 8) & 0xff);
					Uupdate_keys(&tunzip->EntryReadVars.Keys[0], tunzip->EntryReadVars.InputBuffer[i]);
				}
			}
		}

		// Read the encrpytion header that is at the start of the entry, if we haven't already done so
		{
   		  DWORD	uDoEncHead;
   
   		uDoEncHead = tunzip->EntryReadVars.RemainingEncrypt;
   		if (uDoEncHead > tunzip->EntryReadVars.stream.avail_in) uDoEncHead = tunzip->EntryReadVars.stream.avail_in;
   		if (uDoEncHead)
   		{
   			char	bufcrc;
   
   			bufcrc = tunzip->EntryReadVars.stream.next_in[uDoEncHead - 1];
   			tunzip->EntryReadVars.RemainingUncompressed -= uDoEncHead;
   			tunzip->EntryReadVars.stream.avail_in -= uDoEncHead;
   			tunzip->EntryReadVars.stream.next_in += uDoEncHead;
   			tunzip->EntryReadVars.RemainingEncrypt -= uDoEncHead;
   
   			// If we've finished the encryption header, then do the CRC check with the password
   			if (!tunzip->EntryReadVars.RemainingEncrypt && bufcrc != tunzip->EntryReadVars.CrcEncTest)
   			{
   				tunzip->LastErr = ZR_PASSWORD;
   				goto none;
   			}
   		}
		}

		// STORE?
		if (!tunzip->CurrentEntryInfo.compression_method)
		{
			DWORD	uDoCopy;

			if (tunzip->EntryReadVars.stream.avail_out < tunzip->EntryReadVars.stream.avail_in)
				uDoCopy = tunzip->EntryReadVars.stream.avail_out;
			else
				uDoCopy = tunzip->EntryReadVars.stream.avail_in;
			CopyMemory(tunzip->EntryReadVars.stream.next_out, tunzip->EntryReadVars.stream.next_in, uDoCopy);

			tunzip->EntryReadVars.RunningCrc = ucrc32(tunzip->EntryReadVars.RunningCrc, tunzip->EntryReadVars.stream.next_out, uDoCopy);
			tunzip->EntryReadVars.RemainingUncompressed -= uDoCopy;
			tunzip->EntryReadVars.stream.avail_in -= uDoCopy;
			tunzip->EntryReadVars.stream.avail_out -= uDoCopy;
			tunzip->EntryReadVars.stream.next_out += uDoCopy;
			tunzip->EntryReadVars.stream.next_in += uDoCopy;
			tunzip->EntryReadVars.stream.total_out += uDoCopy;
			iRead += uDoCopy;

			if (!tunzip->EntryReadVars.RemainingUncompressed) break;
		}

		// DEFLATE
		else
		{
			DWORD		uTotalOutBefore,uTotalOutAfter;
			const UCH	*bufBefore;
			DWORD		uOutThis;

			uTotalOutBefore = tunzip->EntryReadVars.stream.total_out;
			bufBefore = tunzip->EntryReadVars.stream.next_out;
			if ((err = inflate((z_streamp)&tunzip->EntryReadVars.stream, Z_SYNC_FLUSH)) && err != Z_STREAM_END)
			{
			   printf("DIAG readEntry INFLATE ERR=%d MSG=%s\n",err,tunzip->EntryReadVars.stream.msg);
			   //  printf("DIAG readEntry AFTER INFLATE 2A ERR=%d\n",err);
				tunzip->LastErr = ZR_FLATE;
				goto none;
			}

			uTotalOutAfter = tunzip->EntryReadVars.stream.total_out;
			uOutThis = uTotalOutAfter - uTotalOutBefore;

			tunzip->EntryReadVars.RunningCrc = ucrc32(tunzip->EntryReadVars.RunningCrc, bufBefore, uOutThis);

			tunzip->EntryReadVars.RemainingUncompressed -= uOutThis;
			iRead += (uTotalOutAfter - uTotalOutBefore);

			if (err == Z_STREAM_END || !tunzip->EntryReadVars.RemainingUncompressed) break;
		}
	}

	return(iRead);
}

/************************ findEntry() ***********************
 * Locates an entry (within the archive) by name.
 *
 * flags =	1 (case-insensitive) perhaps OR'ed with ZIP_UNICODE.
 * index =	Where to return the index number of the located entry.
 * ze =		Where to return a filled in ZIPENTRY.
 *
 * NOTE: ZIPENTRY->Name[] must contain the name to match. This
 * will be overwritten.
 *
 * RETURNS: ZR_OK if success, ZR_NOTFOUND if not located, or some
 * other error.
 */

static DWORD findEntry(TUNZIP *tunzip, ZIPENTRY *ze, DWORD flags)
{
	char			name[MAX_PATH];

	if (flags & ZIP_UNICODE)
#ifdef WIN32
		WideCharToMultiByte(CP_UTF8, 0, (const WCHAR *)&ze->Name[0], -1, &name[0], MAX_PATH, 0, 0);
#else
	{
		  unsigned int offset;

		offset = 0;
		while ((name[offset] = (char)(ze->Name[offset]))) ++offset;
	}
#endif
	else
		lstrcpy(&name[0], (const char *)&ze->Name[0]);
	if (lstrlenA(&name[0]) >= UNZ_MAXFILENAMEINZIP) return(ZR_ARGS);

	{
	  char		*d;

	// Next we need to replace '\' with '/' chars
	d = name;
	while (*d)
	{
		if (*d == '\\') *d = '/';
		++d;
	}
	}

	// If there's a currently selected entry, free it
	cleanupEntry(tunzip);

	// No error yet
	tunzip->LastErr = 0;

	// Start with first entry and read its table from the Central Directory
	goToFirstEntry(tunzip);
	while (!tunzip->LastErr)
	{
		// Get this entry's filename
		getEntryFN(tunzip, (char *)&ze->Name[0]);
		if (!tunzip->LastErr)
		{
			// Do names match?
			if (!(flags & 0x01 ? lstrcmpiA(&name[0], (const char *)&ze->Name[0]) : lstrcmpA(&name[0], (const char *)&ze->Name[0])))
			{
				// Fill in caller's ZIPENTRY
				ze->Index = tunzip->CurrentEntryNum;
				return(setCurrentEntry(tunzip, ze, (flags & ZIP_UNICODE) | ZIP_ALREADYINIT));
			}

			goToNextEntry(tunzip);
/*
			// Another entry?
			if (tunzip->CurrentEntryNum + 1 < tunzip->TotalEntries)
			{
				// Skip the remainder of the current entry's table
				skipToEntryEnd(tunzip, 0);
				if (!tunzip->LastErr)
				{
					// Read the info for the next entry
					tunzip->CurrEntryPosInCentralDir += SIZECENTRALDIRITEM + tunzip->CurrentEntryInfo.size_filename + tunzip->CurrentEntryInfo.size_file_extra + tunzip->CurrentEntryInfo.size_file_comment;
					++tunzip->CurrentEntryNum;
					getEntryInfo(tunzip);
				}
			}
			else
				tunzip->LastErr = ZR_NOTFOUND;
*/
		}
	}

	cleanupEntry(tunzip);
	return(tunzip->LastErr);
}

/********************** setCurrentEntry() *********************
 * Sets the specified entry to the currently selected
 * entry within the ZIP archive, and fills in the
 * passed ZIPENTRY with information about that entry.
 *
 * tunzip =		Handle returned by UnzipOpen*() functions.
 * ze =			Pointer to a ZIPENTRY to fill in.
 * flags =		One of the following:
 *				ZIP_ALREADYINIT - ZIPENTRY already filled
 *				in by a call to findEntry().
 *				ZIP_UNICODE - Caller's ZIPENTRY uses UNICODE name.
 *
 * RETURNS: Z_OK if success, otherwise an error code.
 *
 * NOTE: ZIPENTRY->Index must be set to the desired
 * entry number before calling setCurrentEntry(). The value -1
 * means to return how many entries are in the ZIP archive.
 */

static DWORD setCurrentEntry(  TUNZIP *tunzip, ZIPENTRY *ze, DWORD flags)
{
	unsigned char	*extra;

	// No error yet
	tunzip->LastErr = 0;

	// Did findEntry() already init the ZIPENTRY?
	if (!(flags & ZIP_ALREADYINIT))
	{
		// Does caller want general information about the ZIP archive?
		if (ze->Index == (DWORD)-1)
		{
			ze->Index = tunzip->TotalEntries;
			goto good;
		}

		// If there is currently a selected entry, free its resources
		cleanupEntry(tunzip);

		// Seek to the point in the ZIP archive where this entry is found
		// and fill in the TUNZIP->CurrentEntryNum
		if (ze->Index < tunzip->CurrentEntryNum) goToFirstEntry(tunzip);
		while (!tunzip->LastErr && tunzip->CurrentEntryNum < ze->Index) goToNextEntry(tunzip);

		if (tunzip->LastErr)
		{
reterr:		cleanupEntry(tunzip);
			return(tunzip->LastErr);
		}

		// Get the entry's filename
		getEntryFN(tunzip, (char *)&ze->Name[0]);
	}

	// We support only STORE and DEFLATE compression methods
	if (tunzip->CurrentEntryInfo.compression_method && tunzip->CurrentEntryInfo.compression_method != Z_DEFLATED)
		tunzip->LastErr = ZR_NOTSUPPORTED;
	if (tunzip->LastErr) goto reterr;

	// If raw mode, app must supply the compressed and uncompressed sizes
	if (tunzip->Flags & TZIP_RAW) goto good2;

	// Get the extra header (which may contain some extra timestamp stuff)
	{
	DWORD	size;

	extra = 0;
	if ((size = tunzip->CurrentEntryInfo.size_file_extra))
	{
		if (!(extra = (unsigned char *)GlobalAlloc(GMEM_FIXED, size)))
		{
			tunzip->LastErr = ZR_NOALLOC;
			goto reterr;
		}

		if (readFromZip(tunzip, extra, size) != size)
		{
			GlobalFree(extra);
			tunzip->LastErr = ZR_CORRUPT;
			goto reterr;
		}
	}
	}

	// Copy the entry's name to ZIPENTRY->name[] (UNICODE or ANSI)
	{
	  char	*sfn;
	  char	*dfn;
	char			fn[MAX_PATH];
	unsigned char	previous;

	// As a safety feature: if the zip filename had sneaky stuff like "c:\windows\file.txt" or
	// "\windows\file.txt" or "fred\..\..\..\windows\file.txt" then we get rid of them all. That
	// way, when the application does UnzipItem(), it won't be a problem. (If the
	// programmer really does want to get the full evil information, then he can edit out this
	// security feature below). In particular, we chop off any prefixes that are "c:\" or
	// "\" or "/" or "[stuff]\.." or "[stuff]/.."

	// Copy the root dir name
	lstrcpyA(&fn[0], (const char *)&tunzip->Rootdir[0]);

	// Prepare to append entry's name
	sfn = &fn[0] + lstrlenA(&fn[0]);
	dfn = (char *)&ze->Name[0];

	// Skip the drive
	if (dfn[0] && dfn[1] == ':') dfn += 2;

	previous = DIRSLASH_CHAR;			// Skip leading slashes
	while (*dfn)
	{
		// Skip any "\..\" or "/../"
		if (dfn[0] == '\\' || dfn[0] == '/')
		{
			dfn[0] = DIRSLASH_CHAR;		// Change all '/' to '\' for Windows
			if (dfn[1] == '.' && dfn[2] == '.' && (dfn[3] == '\\' || dfn[3] == '/'))
			{
				dfn += 4;
				lstrcpyA(&fn[0], (const char *)&tunzip->Rootdir[0]);
				sfn = &fn[0] + lstrlenA(&fn[0]);
				previous = DIRSLASH_CHAR;
				continue;
			}

			if (previous == DIRSLASH_CHAR)
			{
				previous = 0;
				continue;
			}
		}

		*(sfn)++ = previous = *(dfn)++;
	}
	*sfn = 0;

	if (flags & ZIP_UNICODE)
	{
#ifdef WIN32
		MultiByteToWideChar(CP_UTF8, 0, &fn[0], -1, (WCHAR *)&ze->Name[0], MAX_PATH);
#else
		  unsigned int offset;

		offset = 0;
		while ((ze->Name[offset] = fn[offset])) ++offset;
#endif
	}
	else
		lstrcpy((char *)&ze->Name[0], &fn[0]);

	// Copy the attributes. zip has an 'attribute' 32bit value. Its lower half
	// is windows stuff. Its upper half is standard unix stat.st_mode. We use the
	// UNIX half, but in normal hostmodes these are overridden by the lower half
	{
	unsigned long host;

	host = tunzip->CurrentEntryInfo.version >> 8;
#ifdef WIN32
	if (!host || host==6 || host==10 || host==14)
		ze->Attributes = tunzip->CurrentEntryInfo.external_fa & (FILE_ATTRIBUTE_READONLY|FILE_ATTRIBUTE_HIDDEN|FILE_ATTRIBUTE_SYSTEM|FILE_ATTRIBUTE_DIRECTORY|FILE_ATTRIBUTE_ARCHIVE);
	else
	{
		ze->Attributes = FILE_ATTRIBUTE_ARCHIVE;
		if (tunzip->CurrentEntryInfo.external_fa & 0x40000000) ze->Attributes = FILE_ATTRIBUTE_ARCHIVE|FILE_ATTRIBUTE_DIRECTORY;
		if (!(tunzip->CurrentEntryInfo.external_fa & 0x00800000)) ze->Attributes |= FILE_ATTRIBUTE_READONLY;
	}
#else
	// MS-DOS/Windows/OS-2?
	if (!host || host==6 || host==10 || host==14)
	{
		if (tunzip->CurrentEntryInfo.external_fa & 0x10) ze->Attributes = S_IFDIR;
		else ze->Attributes = S_IFREG;
		if (tunzip->CurrentEntryInfo.external_fa & 0x01) ze->Attributes |= (S_IRUSR|S_IXUSR);
		else ze->Attributes |= (S_IRUSR|S_IXUSR|S_IWUSR);
	}
	else
		ze->Attributes = tunzip->CurrentEntryInfo.external_fa >> 16;
#endif
	}

	// Copy sizes
	ze->CompressedSize = tunzip->CurrentEntryInfo.compressed_size;
	ze->UncompressedSize = tunzip->CurrentEntryInfo.uncompressed_size;

	// Copy timestamp
#ifdef WIN32
	{
	FILETIME ftd;
	FILETIME ft;

	dosdatetime2filetime(&ftd, (tunzip->CurrentEntryInfo.dosDate >> 16) & 0xFFFF, tunzip->CurrentEntryInfo.dosDate & 0xFFFF);
	LocalFileTimeToFileTime(&ftd, &ft);
	ze->AccessTime = ft;
	ze->CreateTime = ft;
	ze->ModifyTime = ft;
	}
#else
	ze->ModifyTime = ze->AccessTime = ze->CreateTime = dos_to_unix_time(tunzip->CurrentEntryInfo.dosDate);
#endif

	{
	// the zip will always have at least that dostime. But if it also has
	// an extra header, then we'll instead get the info from that
	DWORD	epos;

	epos = 0;
	while (epos + 4 < tunzip->CurrentEntryInfo.size_file_extra)
	{
		char	etype[3];
		DWORD	flags;
		int		size;

		etype[0] = extra[epos+0];
		etype[1] = extra[epos+1];
		etype[2] = 0;
		size = extra[epos+2];
		if (!lstrcmpA(etype, "UT"))
		{
			flags = extra[epos + 4];
			epos += 5;
			if ((flags & 1) && epos < tunzip->CurrentEntryInfo.size_file_extra)
			{
#ifdef WIN32
				lutime_t modifyTime = ((extra[epos+0])<<0) | ((extra[epos+1])<<8) |((extra[epos+2])<<16) | ((extra[epos+3])<<24);
				timet2filetime(&ze->ModifyTime, modifyTime);
#else
				ze->ModifyTime = ((extra[epos+0])<<0) | ((extra[epos+1])<<8) |((extra[epos+2])<<16) | ((extra[epos+3])<<24);
#endif
				epos += 4;
			}

			if ((flags & 2) && epos < tunzip->CurrentEntryInfo.size_file_extra)
			{
#ifdef WIN32
				lutime_t accessTime = ((extra[epos+0])<<0) | ((extra[epos+1])<<8) |((extra[epos+2])<<16) | ((extra[epos+3])<<24);
				timet2filetime(&ze->AccessTime, accessTime);
#else
				ze->AccessTime = ((extra[epos+0])<<0) | ((extra[epos+1])<<8) |((extra[epos+2])<<16) | ((extra[epos+3])<<24);
#endif
				epos += 4;
			}

			if ((flags & 4) && epos < tunzip->CurrentEntryInfo.size_file_extra)
			{
#ifdef WIN32
				lutime_t createTime = ((extra[epos+0])<<0) | ((extra[epos+1])<<8) |((extra[epos+2])<<16) | ((extra[epos+3])<<24);
				timet2filetime(&ze->CreateTime, createTime);
#else
				ze->CreateTime = ((extra[epos+0])<<0) | ((extra[epos+1])<<8) |((extra[epos+2])<<16) | ((extra[epos+3])<<24);
#endif
				epos += 4;
			}

			break;
		}

		epos += 4 + size;
	}
	}

	if (extra) GlobalFree(extra);
	}
good2:
	// We clear the internal_fa field as a signal to unzipEntry() that a memory-buffer unzip is
	// starting for the first time
	tunzip->CurrentEntryInfo.internal_fa = 0;
good:
	return(ZR_OK);
}

/************************ str_chrA **********************
 * Searches for the first occurence of 'chr' in
 * nul-terminated 'str'. Same as C library's strchr().
 *
 * str =	pointer to string to search
 * chr =	character to search for
 *
 * RETURNS:	pointer to where 'chr' is found in 'str', or 0
 *			if not found.
 */

static char * str_chrA(char *str, char chr)
{
	  char	tempch;

	if ((tempch = *str))
	{
		do
		{
			if (tempch == chr) return(str);
		} while ((tempch = *(++str)));
	}

	// Not found
	return(0);
}

/********************* createMultDirsA() *******************
 * Creates as many dirs as are specified in the
 * nul-terminated string pointed to by dirname.
 *
 * dirname =	The names of directories to create, each
 *				separated by a backslash (but no backslash
 *				at the head or tail of the string).
 *
 * RETURNS: 1 if success, 0 if error.
 */

unsigned long createMultDirsA(char *dirname, BOOL isDir)
{
	  char *		ptr;
	  char *		pathbuf;
#ifdef WIN32
	SECURITY_ATTRIBUTES	sc;

	sc.nLength = sizeof(SECURITY_ATTRIBUTES);
	sc.lpSecurityDescriptor = 0;
	sc.bInheritHandle = TRUE;
#endif

	pathbuf = dirname;

	// Skip drive
#ifdef WIN32
	if (dirname[0] && dirname[1] == ':') dirname += 2;
#endif
	if (dirname[0] == DIRSLASH_CHAR) ++dirname;

	// Another sub-dir to create?
	while (*dirname)
	{
		// Isolate next sub-dir name
		if (!(ptr = str_chrA(dirname, DIRSLASH_CHAR)))
		{
			if (!isDir) break;
		}
		else
		{
			// Nul-term path
			*ptr = 0;
		}

#ifdef WIN32
		if (!CreateDirectoryA(pathbuf, &sc) && GetLastError() != ERROR_ALREADY_EXISTS)
#else
		if (mkdir(pathbuf, 0700) == -1 && errno != EEXIST)
#endif
		{
			if (ptr) *ptr = DIRSLASH_CHAR;
			return(0);
		}

		if (!ptr) break;

		// Restore overwritten char
		*ptr = DIRSLASH_CHAR;
		dirname = ++ptr;
	}

	return(1);
}

#ifdef WIN32

/************************ str_chrW **********************
 * Wide character version of str_chrA().
 */

static WCHAR * str_chrW(WCHAR *str, WCHAR chr)
{
	  WCHAR	ch;

	if ((ch = *str))
	{
		do
		{
			if (ch == chr) return(str);
		} while ((ch = *(++str)));
	}

	// Not found
	return(0);
}

/********************* createMultDirsW *******************
 * Wide character version of createMultDirsA().
 */

unsigned long createMultDirsW(WCHAR *dirname, BOOL isDir)
{
	  WCHAR *		ptr;
	  WCHAR *		pathbuf;
	SECURITY_ATTRIBUTES	sc;

	sc.nLength = sizeof(SECURITY_ATTRIBUTES);
	sc.lpSecurityDescriptor = 0;
	sc.bInheritHandle = TRUE;

	pathbuf = dirname;

	// Skip drive
	if (dirname[0] && dirname[1] == ':') dirname += 2;
	if (dirname[0] == DIRSLASH_CHAR) ++dirname;

	// Another sub-dir to create?
	while (*dirname)
	{
		// Isolate next sub-dir name
		if (!(ptr = str_chrW(dirname, DIRSLASH_CHAR)))
		{
			if (!isDir) break;
		}
		else
		{
			// Nul-term path
			*ptr = 0;
		}

		if (!CreateDirectoryW(pathbuf, &sc) && GetLastError() != ERROR_ALREADY_EXISTS)
		{
			if (ptr) *ptr = DIRSLASH_CHAR;
			return(0);
		}

		if (!ptr) break;

		// Restore overwritten char
		*ptr = DIRSLASH_CHAR;
		dirname = ++ptr;
	}

	return(1);
}

#endif

/********************* unzipEntry() *******************
 * Unzips the specified entry in the specified ZIP
 * archive. Can be unzipped to a pipe/handle, disk file,
 * or memory buffer.
 *
 * dst =	Handle to file where the entry is decompressed,
 *			or filename, or memory buffer pointer.
 * ze =		Filled in ZIPENTRY struct.
 * flags =	ZIP_MEMORY, ZIP_FILENAME, or ZIP_HANDLE. Also
 *			may be ZIP_UNICODE.
 */

static DWORD unzipEntry(  TUNZIP *tunzip, void *dst, ZIPENTRY *ze, DWORD flags)
{
	HANDLE		h;

	// Make sure TUNZIP if valid
	if (IsBadReadPtr(tunzip, 1) ||

		// Make sure we have a currently selected entry
		tunzip->CurrentEntryNum == (DWORD)-1)
	{
		return(ZR_ARGS);
	}
   
	// No error
	tunzip->LastErr = 0;
	// ============ Unzipping to memory ============
	if (flags & ZIP_MEMORY)
	{
		// Don't reinitialize if this is called as a result of ZR_MORE
		if (!tunzip->CurrentEntryInfo.internal_fa)
		{
			// initEntry(tunzip, ze); PH99
			if (tunzip->LastErr) goto out;
			tunzip->CurrentEntryInfo.internal_fa = 1;
		}

		ze->CompressedSize = readEntry(tunzip, dst, ze->CompressedSize);
		if (tunzip->LastErr || !tunzip->EntryReadVars.RemainingUncompressed)
		{
			tunzip->CurrentEntryInfo.internal_fa = 0;
			goto out;
		}

		// Filled the output buffer. Return ZR_MORE so the caller can flush it somewhere,
		// but don't close the entry, and leave internal_fa set to 1
		return(ZR_MORE);
	}

	// ============ Unzipping to disk/pipe ============

	// Is this entry a directory name?
	if (ze->Attributes & FILE_ATTRIBUTE_DIRECTORY)
	{
		// NOTE: We can't create a directory when spooling to a pipe
		if (flags & ZIP_FILENAME)
		{
			if (flags & ZIP_UNICODE) flags = createMultDirsW((WCHAR *)dst, 1);
			else  flags = createMultDirsA((char *)dst, 1);
			if (!flags) goto badf;
		}
		return(ZR_OK);
	}

	// Write the entry to a file/handle
	if (flags == ZIP_HANDLE)
		h = (HANDLE)dst;
	else
	{
		DWORD		res;
		// Create any needed directories
		if (flags & ZIP_UNICODE) res = createMultDirsW((WCHAR *)dst, 0);
		else res = createMultDirsA((char *)dst, 0);

		if (!res) goto badf;  
		// Create the file to which we'll write the uncompressed entry
		if (flags & ZIP_UNICODE)
		{
			h = CreateFileW((WCHAR *)dst, GENERIC_WRITE, 0, 0, CREATE_ALWAYS, ze->Attributes, 0);
      }
		else
		{
			h = CreateFileA((char *)dst, GENERIC_WRITE, 0, 0, CREATE_ALWAYS, ze->Attributes, 0);
      }
	}
	if (h == INVALID_HANDLE_VALUE)

badf:	tunzip->LastErr = ZR_NOFILE;
	else
	{
		// Allocate resources for decompressing
		initEntry(tunzip, ze);
		if (!tunzip->LastErr)
		{
			// Get an output buffer (where we decompress bytes)
			if (!tunzip->OutBuffer && !(tunzip->OutBuffer = (unsigned char *)GlobalAlloc(GMEM_FIXED, 16384))) tunzip->LastErr = ZR_NOALLOC;
			while (!tunzip->LastErr)
			{
				DWORD				writ;
				  DWORD read;
				// Decompress the bytes into the input buffer. If EOF, then get out of this loop
				if (!(read = readEntry(tunzip, tunzip->OutBuffer, 16384))) break;
				// Write the bytes
				if (!WriteFile(h, tunzip->OutBuffer, read, &writ, 0) || writ != read)
					tunzip->LastErr = ZR_WRITE;
			}
		}

		/* Set the file's timestamp  PH99
		if (!tunzip->LastErr)
			SetFileTime(h, &ze->CreateTime, &ze->AccessTime, &ze->ModifyTime); // may fail if it was a pipe
		*/	
		// Close the file if we opened it above	
		if (flags != ZIP_HANDLE)
		{
			CloseHandle(h);
		}	
	}
out:
	cleanupEntry(tunzip);
	return(tunzip->LastErr);
}

DWORD WINAPI UnzipItemToHandle(HUNZIP tunzip, HANDLE h, ZIPENTRY *ze)
{
	return(unzipEntry((TUNZIP *)tunzip, (void *)h, ze, ZIP_HANDLE));
}

DWORD WINAPI UnzipItemToFileA(HUNZIP tunzip, const char *fn, ZIPENTRY *ze)
{
	return(unzipEntry((TUNZIP *)tunzip, (void *)fn, ze, ZIP_FILENAME));
}

DWORD WINAPI UnzipItemToFileW(HUNZIP tunzip, const WCHAR *fn, ZIPENTRY *ze)
{
	return(unzipEntry((TUNZIP *)tunzip, (void *)fn, ze, ZIP_FILENAME|ZIP_UNICODE));
}

DWORD WINAPI UnzipItemToBuffer(HUNZIP tunzip, void *z, DWORD len, ZIPENTRY *ze)
{
	ze->CompressedSize = len;
	return(unzipEntry((TUNZIP *)tunzip, z, ze, ZIP_MEMORY));
}


DWORD WINAPI UnzipFormatMessageA(DWORD code, char *buf, DWORD len)
{
#ifdef WIN32
	if (code == ZR_OK) code = IDS_OK;
	if (!(code = LoadStringA(ThisInstance, code, buf, len)))
		code = LoadStringA(ThisInstance, IDS_UNKNOWN, buf, len);
#else
	  const char	*str;
	  char 			*dest;

	str = &ErrorMsgs[0];
	while (code-- && *str) str += (strlen(str) + 1);
	if (!(*str)) str = &UnknownErr[0];
	code = 0;
	if (len)
	{
		dest = buf;
		do
		{
			if (!(dest[code] = str[code])) goto out;
		} while (++code < len);
		dest[--code] = 0;
	}
out:
#endif
	return(code);
}

DWORD WINAPI UnzipFormatMessageW(DWORD code, WCHAR *buf, DWORD len)
{
#ifdef WIN32
	if (code == ZR_OK) code = IDS_OK;
	if (!(code = LoadStringW(ThisInstance, code, buf, len)))
		code = LoadStringW(ThisInstance, IDS_UNKNOWN, buf, len);
#else
	  const char	*str;
	  WCHAR 		*dest;

	str = &ErrorMsgs[0];
	while (code-- && *str) str += (strlen(str) + 1);
	if (!(*str)) str = &UnknownErr[0];
	code = 0;
	if (len)
	{
		dest = buf;
		do
		{
			if (!(dest[code] = (WCHAR)((unsigned char)str[code]))) goto out;
		} while (++code < len);
		dest[--code] = 0;
	}
out:
#endif
	return(code);
}

/********************* closeArchive() *****************
 * Closes the ZIP archive opened with openArchive().
 */

static void closeArchive(  TUNZIP *tunzip)
{
	cleanupEntry(tunzip);
	if (tunzip->Flags & TZIP_ARCCLOSEFH)
#ifdef WIN32
		CloseHandle(tunzip->ArchivePtr);
#else
		close((int)tunzip->ArchivePtr);
#endif
	if (tunzip->Password) GlobalFree(tunzip->Password);
	if (tunzip->OutBuffer) GlobalFree(tunzip->OutBuffer);
	GlobalFree(tunzip);
}

/********************* UnzipClose() **********************
 * Closes the ZIP file that was created/opened with one
 * of the UnzipOpen* functions.
 */

DWORD WINAPI UnzipClose(HUNZIP tunzip)
{
	  DWORD	result;

	// Make sure TUNZIP if valid
#ifdef WIN32
	if (IsBadReadPtr(tunzip, 1))
#else
	if (!tunzip)
#endif
		result = ZR_ARGS;
	else
	{
		result = ZR_OK;
		closeArchive((TUNZIP *)tunzip);
	}
	return(result);
}

DWORD WINAPI UnzipGetItemA(HUNZIP tunzip, ZIPENTRY *ze)
{
#ifdef WIN32
	if (IsBadReadPtr(tunzip, 1))
#else
	if (!tunzip)
#endif
		return(ZR_ARGS);
	return(setCurrentEntry((TUNZIP *)tunzip, ze, 0));
}

DWORD WINAPI UnzipGetItemW(HUNZIP tunzip, ZIPENTRY *ze)
{
#ifdef WIN32
	if (IsBadReadPtr(tunzip, 1))
#else
	if (!tunzip)
#endif
		return(ZR_ARGS);
	return(setCurrentEntry((TUNZIP *)tunzip, ze, ZIP_UNICODE));
}

DWORD WINAPI UnzipFindItemA(HUNZIP tunzip, ZIPENTRY *ze, BOOL ic)
{
#ifdef WIN32
	if (IsBadReadPtr(tunzip, 1))
#else
	if (!tunzip)
#endif
		return(ZR_ARGS);
	return(findEntry((TUNZIP *)tunzip, ze, (DWORD)ic));
}

DWORD WINAPI UnzipFindItemW(HUNZIP tunzip, ZIPENTRY *ze, BOOL ic)
{
#ifdef WIN32
	if (IsBadReadPtr(tunzip, 1))
#else
	if (!tunzip)
#endif
		return(ZR_ARGS);
	return(findEntry((TUNZIP *)tunzip, ze, (DWORD)ic | ZIP_UNICODE));
}

/******************* openArchive() ********************
 * Opens a ZIP archive, in preparation of unzipping
 * its entries.
 *
 * z =	A handle, pointer to filename, or pointer to a
 *		memory buffer where the ZIP archive resides.
 * len = If a memory buffer, its size.
 * flags = ZIP_HANDLE, ZIP_FILENAME, or ZIP_MEMORY. Also,
 *		can be OR'd with ZIP_UNICODE.
 *
 * RETURNS: Z_OK if success, otherwise an error number.
 */

#define BUFREADCOMMENT (0x400)

static DWORD openArchive(HANDLE *ptr, void *z, DWORD len, DWORD flags, const char *pwd)
{
	  TUNZIP		*tunzip;
	DWORD				centralDirPos; 

	// Get a TUNZIP
	if (!(tunzip = (TUNZIP *)GlobalAlloc(GMEM_FIXED, sizeof(TUNZIP))))
	{
badalloc:
		flags = ZR_NOALLOC;
bad:	return(flags);
	}
	ZeroMemory(tunzip, sizeof(TUNZIP) - MAX_PATH);
   
	// Store password if supplied
	if (pwd)
	{
		if (!(tunzip->Password = (unsigned char *)GlobalAlloc(GMEM_FIXED, lstrlenA(pwd) + 1)))
		{
			GlobalFree(tunzip);
			goto badalloc;
		}
		lstrcpyA((char *)tunzip->Password, pwd);
	}

	// No currently selected entry
	tunzip->CurrentEntryNum = (DWORD)-1;

	switch (flags & ~(ZIP_UNICODE|ZIP_ALREADYINIT|ZIP_RAW))
	{
		// ZIP archive is from a pipe or already open handle
		case ZIP_HANDLE:
		{
#ifdef WIN32
			if (!DuplicateHandle(GetCurrentProcess(), (HANDLE)z, GetCurrentProcess(), &tunzip->ArchivePtr, 0, 0, DUPLICATE_SAME_ACCESS)) goto mustclose;
#endif
			tunzip->ArchivePtr = (HANDLE)z;
			goto chk_handle;
		}
		
		// ZIP archive is a disk file
		case ZIP_FILENAME:
		{
			if (flags & ZIP_UNICODE)
				tunzip->ArchivePtr = CreateFileW((const WCHAR *)z, GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
			else
				tunzip->ArchivePtr = CreateFileA((const char *)z, GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
			if (tunzip->ArchivePtr == INVALID_HANDLE_VALUE)
			{
				flags = ZR_NOFILE;
bad2:			closeArchive(tunzip);
				goto bad;
			}

#ifdef WIN32
mustclose:
#endif
			tunzip->Flags = TZIP_ARCCLOSEFH;
chk_handle:	// Test if we can seek on it
#ifdef WIN32
			if ((tunzip->InitialArchiveOffset = SetFilePointer(tunzip->ArchivePtr, 0, 0, FILE_CURRENT)) == (DWORD)-1)
#else
			if ((tunzip->InitialArchiveOffset = (DWORD)lseek((int)tunzip->ArchivePtr, 0, SEEK_CUR)) == (DWORD)-1)
#endif
			{
				// Our unzip code requires that we be able to seek on the ZIP archive
				flags = ZR_SEEK;
				goto bad2;
			}

			break;
		}

		// A memory buffer
		case ZIP_MEMORY:
		{
			// If buffer pointer is supplied, then he passed the ZIP archive
			if ((tunzip->ArchivePtr = z))
				tunzip->ArchiveBufLen = len;

			// Otherwise, he wants the ZIP archive extracted from the EXE's resource, and "len" is the ID number
			else
			{
#ifdef WIN32
				HRSRC		hrsrc;
				HANDLE		hglob;

				z = GetModuleHandle(0);

				// Load the ZIP archive, which is in our EXE's resources. It is an RT_RCDATA type
				// of resource with an ID number of 'len'
				if (!(hrsrc = FindResource((HMODULE)z, MAKEINTRESOURCE(len), RT_RCDATA)) || !(hglob = LoadResource((HMODULE)z, hrsrc)))
				{
					flags = ZR_NOTFOUND;
					goto bad2;
				}

				// Get the size and ptr to the ZIP archive in memory
				if (!(tunzip->ArchivePtr = LockResource(hglob))) goto memerr;
				tunzip->ArchiveBufLen = SizeofResource((HMODULE)z, hrsrc);
#else
				flags = ZR_NOTFOUND;
				goto bad2;
#endif
			}

			tunzip->Flags = TZIP_ARCMEMORY;
			break;
		}

//		default:
//		{
//			flags = ZR_ARGS;
//			goto bad2;
//		}
	}

	// If RAW flag, then it's assumed there is only one file in the archive, no central directory,
	// and not even any GZIP indentification ID3 tag compression seems to be this way. Sigh... Wish
	// some programmers put a little more thought into their dodgy file formats _before_ releasing
	// code
	if (flags & ZIP_RAW)
	{
		ZeroMemory(&tunzip->CurrentEntryInfo, sizeof(ZIPENTRYINFO));
		tunzip->CurrentEntryInfo.compression_method = Z_DEFLATED;
		tunzip->CurrentEntryInfo.offset = tunzip->InitialArchiveOffset;
		tunzip->Flags |= TZIP_RAW;
		tunzip->CurrentEntryNum = 0;
		goto raw;
	}

	{
	// Find the central directory's offset within the archive
	DWORD			uMaxBack;
	DWORD			uSizeFile;
	DWORD			uBackRead;
	unsigned char	*buf;

	if (!(tunzip->Flags & TZIP_ARCMEMORY))
		uSizeFile = SetFilePointer(tunzip->ArchivePtr, 0, 0, FILE_END) - tunzip->InitialArchiveOffset;
	else
		uSizeFile = tunzip->ArchiveBufPos = tunzip->ArchiveBufLen;

	uMaxBack = 0xffff; // maximum size of global comment
	if (uMaxBack > uSizeFile) uMaxBack = uSizeFile;

	if ((buf = (unsigned char *)GlobalAlloc(GMEM_FIXED, BUFREADCOMMENT + 4)))
	{
		uBackRead = 4;
		while (uBackRead < uMaxBack)
		{
			DWORD		uReadSize;
			int			i;

			if (uBackRead + BUFREADCOMMENT > uMaxBack) uBackRead = uMaxBack;
			else uBackRead += BUFREADCOMMENT;
			centralDirPos = uSizeFile - uBackRead;
			uReadSize = ((BUFREADCOMMENT + 4) < (uSizeFile - centralDirPos)) ? (BUFREADCOMMENT+4) : (uSizeFile - centralDirPos);
			if (seekInZip(tunzip, centralDirPos, FILE_BEGIN) || !readFromZip(tunzip, buf, uReadSize)) break;
			for (i = (int)uReadSize - 3; (i--) >= 0;)
			{
				if (*(buf+i) == 0x50 && *(buf+i+1) == 0x4b && *(buf+i+2) == 0x05 && *(buf+i+3) == 0x06)
				{
					GlobalFree(buf);
					centralDirPos += i;
					seekInZip(tunzip, centralDirPos, FILE_BEGIN);
					goto gotdir;
				}
			}
		}

		GlobalFree(buf);
	}

	// Memory error?
	if (!buf)
	{
#ifdef WIN32
memerr:
#endif
		flags = ZR_NOALLOC;
		goto bad2;
	}
	}

	{
	unsigned short	number_disk;			// # of the current dist, used for spanning ZIP, unsupported here
	unsigned short	number_disk_with_CD;	// # of the disk with central dir, used for spanning ZIP, unsupported here
	unsigned short	totalEntries_CD;		// total number of entries in the central dir (same as TotalEntries on nospan)
	DWORD			centralDirSize;
	// Assume GZIP format (ie, no central dir because there is only 1 file)
	seekInZip(tunzip, 0, FILE_BEGIN);
	flags = getArchiveShort(tunzip);
	if (flags != 0x00008b1f)
	{
badzip:	flags = ZR_CORRUPT;
		goto bad2;
	}
	tunzip->CentralDirOffset = 3;
raw:
	tunzip->ByteBeforeZipArchive = tunzip->InitialArchiveOffset;
	tunzip->LastErr = 0;
	tunzip->Flags |= TZIP_GZIP;
	tunzip->TotalEntries = 1;
	goto gotgzip;

gotdir:
	// Get/skip the signature, already checked above
	getArchiveLong(tunzip);

	// Get the number of this disk. Used for spanning ZIP, unsupported here
	number_disk = getArchiveShort(tunzip);

	// Get number of the disk with the start of the central directory
	number_disk_with_CD = getArchiveShort(tunzip);

	// Get total number of entries in the central dir on this disk
	tunzip->TotalEntries = getArchiveShort(tunzip);

	// Get total number of entries in the central dir
	totalEntries_CD = getArchiveShort(tunzip);
	if (tunzip->LastErr) goto badzip;

	// We don't support disk-spanning here
	if (totalEntries_CD != tunzip->TotalEntries || number_disk_with_CD || number_disk)
	{
		flags = ZR_NOTSUPPORTED;
		goto bad2;
	}


	// Size of the central directory
	centralDirSize = getArchiveLong(tunzip);

	// Offset of start of central directory with respect to the starting disk number
	tunzip->CentralDirOffset = getArchiveLong(tunzip);

	// zipfile comment length
	tunzip->CommentSize = getArchiveShort(tunzip);
	if (tunzip->LastErr || centralDirPos + tunzip->InitialArchiveOffset < tunzip->CentralDirOffset + centralDirSize) goto badzip;
	tunzip->ByteBeforeZipArchive = centralDirPos + tunzip->InitialArchiveOffset - (tunzip->CentralDirOffset + centralDirSize);
//	tunzip->CentralDirPos = centralDirPos;
gotgzip:
	// Set Rootdir to current directory. (Assume we unzip there)
#ifdef WIN32
	if (!(centralDirPos = GetCurrentDirectoryA(MAX_PATH - 1, (LPSTR)&tunzip->Rootdir[0])) ||
		tunzip->Rootdir[centralDirPos - 1] != '\\')
#else
	if (getcwd((char *)&tunzip->Rootdir[0], PATH_MAX) && (centralDirPos = strlen((const char *)&tunzip->Rootdir[0])) &&
		tunzip->Rootdir[centralDirPos - 1] != '/')
#endif
	{
		tunzip->Rootdir[centralDirPos++] = DIRSLASH_CHAR;
		tunzip->Rootdir[centralDirPos] = 0;
	}

	// Return the TUNZIP
	*ptr = tunzip;
	
	return(ZR_OK);
	}
}

/******************** UnzipOpen*() ********************
 * Opens a ZIP archive, in preparation of decompressing
 * entries from it. The archive can be on disk, or in
 * memory, or from a pipe or an already open file.
 */

DWORD WINAPI UnzipOpenHandle(HUNZIP *tunzip, HANDLE h, const char *password)
{
	return(openArchive(tunzip, h, 0, ZIP_HANDLE, password));
}

DWORD WINAPI UnzipOpenHandleRaw(HUNZIP *tunzip, HANDLE h, const char *password)
{
	return(openArchive(tunzip, h, 0, ZIP_HANDLE|ZIP_RAW, password));
}

DWORD WINAPI UnzipOpenFileA(HUNZIP *tunzip, const char *fn, const char *password)
{
	return(openArchive(tunzip, (void *)fn, 0, ZIP_FILENAME, password));
}

DWORD WINAPI UnzipOpenFileW(HUNZIP *tunzip, const WCHAR *fn, const char *password)
{
	return(openArchive(tunzip, (void *)fn, 0, ZIP_FILENAME|ZIP_UNICODE, password));
}

DWORD WINAPI UnzipOpenFileRawA(HUNZIP *tunzip, const char *fn, const char *password)
{
	return(openArchive(tunzip, (void *)fn, 0, ZIP_FILENAME|ZIP_RAW, password));
}

DWORD WINAPI UnzipOpenFileRawW(HUNZIP *tunzip, const WCHAR *fn, const char *password)
{
	return(openArchive(tunzip, (void *)fn, 0, ZIP_FILENAME|ZIP_UNICODE|ZIP_RAW, password));
}

DWORD WINAPI UnzipOpenBuffer(HUNZIP *tunzip, void *z, DWORD len, const char *password)
{
	return(openArchive(tunzip, z, len, ZIP_MEMORY, password));
}

DWORD WINAPI UnzipOpenBufferRaw(HUNZIP *tunzip, void *z, DWORD len, const char *password)
{
	return(openArchive(tunzip, z, len, ZIP_MEMORY|ZIP_RAW, password));
}

/******************** unzipSetBaseDir() ******************
 * Sets the root directory.
 */

static DWORD unzipSetBaseDir(TUNZIP *tunzip, const void *dir, DWORD isUnicode)
{
	// Make sure TUNZIP if valid
#ifdef WIN32
	if (IsBadReadPtr(tunzip, 1))
#else
	if (!tunzip)
#endif
		isUnicode = ZR_ARGS;
	else
	{
		unsigned char	*lastchar;

		if (isUnicode)
#ifdef WIN32
			WideCharToMultiByte(CP_UTF8, 0, (const WCHAR *)dir, -1, (LPSTR)&tunzip->Rootdir[0], MAX_PATH, 0, 0);
#else
		{
			  const WCHAR *tempptr;

			isUnicode = 0;
			tempptr = (const WCHAR *)dir;
			while ((tunzip->Rootdir[isUnicode] = (char)(tempptr[isUnicode]))) ++isUnicode;
		}
#endif
		else
			lstrcpyA((char *)&tunzip->Rootdir[0], (const char *)dir);

		// Make sure it ends with '\\'
		lastchar = &tunzip->Rootdir[lstrlenA((char *)&tunzip->Rootdir[0]) - 1];
		if (*lastchar != DIRSLASH_CHAR)
		{
			lastchar[1] = DIRSLASH_CHAR;
			lastchar[2] = 0;
		}

		isUnicode = ZR_OK;
	}

	return(isUnicode);
}

DWORD WINAPI UnzipSetBaseDirA(HUNZIP tunzip, const char *dir)
{
	return(unzipSetBaseDir((TUNZIP *)tunzip, dir, 0));
}

DWORD WINAPI UnzipSetBaseDirW(HUNZIP tunzip, const WCHAR *dir)
{
	return(unzipSetBaseDir((TUNZIP *)tunzip, dir, 1));
}

// END   LITEUNZIP CODE

// Start MINIZIP CODE
/*
   minizip.c
   Version 1.01h, December 28th, 2009

   Copyright (C) 1998-2009 Gilles Vollant

Changes:
   Aug 3, 2006.     jg.    support storing files with out paths. (-j)
   Aug 3, 2006.     jg.    files with paths should not have leading slashes.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <fcntl.h>

#ifdef unix
# include <unistd.h>
# include <utime.h>
# include <sys/types.h>
# include <sys/stat.h>
#else
# include <direct.h>
# include <io.h>
#endif

#include "zip.h"

#ifdef WIN32
#define USEWIN32IOAPI
#include "iowin32.h"
#endif



#define WRITEBUFFERSIZE (16384)
#define MAXFILENAME (256)

#ifdef WIN32
uLong filetime(char *f, tm_zip *tmzip, uLong *dt)
    // char *f;                /* name of file to get info on */
    // tm_zip *tmzip;             /* return value: access, modific. and creation times */
    // uLong *dt;             /* dostime */   
{
  int ret = 0;
  {
      FILETIME ftLocal;
      HANDLE hFind;
      WIN32_FIND_DATAA ff32;

      hFind = FindFirstFileA(f,&ff32);
      if (hFind != INVALID_HANDLE_VALUE)
      {
        FileTimeToLocalFileTime(&(ff32.ftLastWriteTime),&ftLocal);
        FileTimeToDosDateTime(&ftLocal,((LPWORD)dt)+1,((LPWORD)dt)+0);
        FindClose(hFind);
        ret = 1;
      }
  }
  return ret;
}
#else
#ifdef unix
uLong filetime(f, tmzip, dt)
    char *f;               /* name of file to get info on */
    tm_zip *tmzip;         /* return value: access, modific. and creation times */
    uLong *dt;             /* dostime */
{
  int ret=0;
  struct stat s;        /* results of stat() */
  struct tm* filedate;
  time_t tm_t=0;

  if (strcmp(f,"-")!=0)
  {
    char name[MAXFILENAME+1];
    int len = strlen(f);
    if (len > MAXFILENAME)
      len = MAXFILENAME;

    strncpy(name, f,MAXFILENAME-1);
    /* strncpy doesnt append the trailing NULL, if the string is too long. */
    name[ MAXFILENAME ] = '\0';

    if (name[len - 1] == '/')
      name[len - 1] = '\0';
    /* not all systems allow stat'ing a file with / appended */
    if (stat(name,&s)==0)
    {
      tm_t = s.st_mtime;
      ret = 1;
    }
  }
  filedate = localtime(&tm_t);

  tmzip->tm_sec  = filedate->tm_sec;
  tmzip->tm_min  = filedate->tm_min;
  tmzip->tm_hour = filedate->tm_hour;
  tmzip->tm_mday = filedate->tm_mday;
  tmzip->tm_mon  = filedate->tm_mon ;
  tmzip->tm_year = filedate->tm_year;

  return ret;
}
#else
uLong filetime(f, tmzip, dt)
    char *f;                /* name of file to get info on */
    tm_zip *tmzip;             /* return value: access, modific. and creation times */
    uLong *dt;             /* dostime */
{
    return 0;
}
#endif
#endif

int check_exist_file(const char* filename)
    // const char* filename;
{
    FILE* ftestexist;
    int ret = 1;
    ftestexist = fopen(filename,"rb");
    if (ftestexist==NULL)
        ret = 0;
    else
        fclose(ftestexist);
    return ret;
}

/* iowin32.c -- IO base function header for compress/uncompress .zip
   files using zlib + zip or unzip API
   This IO API version uses the Win32 API (for Microsoft Windows)

   Version 1.01h, December 28th, 2009

   Copyright (C) 1998-2009 Gilles Vollant
*/

#include <stdlib.h>

#include "zlib.h"
#include "ioapi.h"
#include "iowin32.h"

#ifndef INVALID_HANDLE_VALUE
#define INVALID_HANDLE_VALUE (0xFFFFFFFF)
#endif

#ifndef INVALID_SET_FILE_POINTER
#define INVALID_SET_FILE_POINTER ((DWORD)-1)
#endif

voidpf ZCALLBACK win32_open_file_func OF((
   voidpf opaque,
   const char* filename,
   int mode));

uLong ZCALLBACK win32_read_file_func OF((
   voidpf opaque,
   voidpf stream,
   void* buf,
   uLong size));

uLong ZCALLBACK win32_write_file_func OF((
   voidpf opaque,
   voidpf stream,
   const void* buf,
   uLong size));

long ZCALLBACK win32_tell_file_func OF((
   voidpf opaque,
   voidpf stream));

long ZCALLBACK win32_seek_file_func OF((
   voidpf opaque,
   voidpf stream,
   uLong offset,
   int origin));

int ZCALLBACK win32_close_file_func OF((
   voidpf opaque,
   voidpf stream));

int ZCALLBACK win32_error_file_func OF((
   voidpf opaque,
   voidpf stream));

typedef struct
{
    HANDLE hf;
    int error;
} WIN32FILE_IOWIN;

voidpf ZCALLBACK win32_open_file_func (voidpf opaque, const char* filename, int mode)
   // voidpf opaque;
   // const char* filename;
   // int mode;
{
    const char* mode_fopen = NULL;
    DWORD dwDesiredAccess,dwCreationDisposition,dwShareMode,dwFlagsAndAttributes ;
    HANDLE hFile = 0;
    voidpf ret=NULL;

    dwDesiredAccess = dwShareMode = dwFlagsAndAttributes = 0;

    if ((mode & ZLIB_FILEFUNC_MODE_READWRITEFILTER)==ZLIB_FILEFUNC_MODE_READ)
    {
        dwDesiredAccess = GENERIC_READ;
        dwCreationDisposition = OPEN_EXISTING;
        dwShareMode = FILE_SHARE_READ;
    }
    else
    if (mode & ZLIB_FILEFUNC_MODE_EXISTING)
    {
        dwDesiredAccess = GENERIC_WRITE | GENERIC_READ;
        dwCreationDisposition = OPEN_EXISTING;
    }
    else
    if (mode & ZLIB_FILEFUNC_MODE_CREATE)
    {
        dwDesiredAccess = GENERIC_WRITE | GENERIC_READ;
        dwCreationDisposition = CREATE_ALWAYS;
    }

    if ((filename!=NULL) && (dwDesiredAccess != 0))
        hFile = CreateFile((LPCTSTR)filename, dwDesiredAccess, dwShareMode, NULL,
                      dwCreationDisposition, dwFlagsAndAttributes, NULL);

    if (hFile == INVALID_HANDLE_VALUE)
        hFile = NULL;

    if (hFile != NULL)
    {
        WIN32FILE_IOWIN w32fiow;
        w32fiow.hf = hFile;
        w32fiow.error = 0;
        ret = malloc(sizeof(WIN32FILE_IOWIN));
        if (ret==NULL)
            CloseHandle(hFile);
        else *((WIN32FILE_IOWIN*)ret) = w32fiow;
    }
    return ret;
}


uLong ZCALLBACK win32_read_file_func (voidpf opaque, voidpf stream, void* buf, uLong size)
   // voidpf opaque;
   // voidpf stream;
   // void* buf;
   // uLong size;
{
    uLong ret=0;
    HANDLE hFile = NULL;
    if (stream!=NULL)
        hFile = ((WIN32FILE_IOWIN*)stream) -> hf;
    if (hFile != NULL)
        if (!ReadFile(hFile, buf, size, &ret, NULL))
        {
            DWORD dwErr = GetLastError();
            if (dwErr == ERROR_HANDLE_EOF)
                dwErr = 0;
            ((WIN32FILE_IOWIN*)stream) -> error=(int)dwErr;
        }

    return ret;
}


uLong ZCALLBACK win32_write_file_func (voidpf opaque, voidpf stream, const void* buf, uLong size)
   // voidpf opaque;
   // voidpf stream;
   // const void* buf;
   // uLong size;
{
    uLong ret=0;
    HANDLE hFile = NULL;
    if (stream!=NULL)
        hFile = ((WIN32FILE_IOWIN*)stream) -> hf;

    if (hFile !=NULL)
        if (!WriteFile(hFile, buf, size, &ret, NULL))
        {
            DWORD dwErr = GetLastError();
            if (dwErr == ERROR_HANDLE_EOF)
                dwErr = 0;
            ((WIN32FILE_IOWIN*)stream) -> error=(int)dwErr;
        }

    return ret;
}

long ZCALLBACK win32_tell_file_func (voidpf opaque, voidpf stream)
   // voidpf opaque;
   // voidpf stream;
{
    long ret=-1;
    HANDLE hFile = NULL;
    if (stream!=NULL)
        hFile = ((WIN32FILE_IOWIN*)stream) -> hf;
    if (hFile != NULL)
    {
        DWORD dwSet = SetFilePointer(hFile, 0, NULL, FILE_CURRENT);
        if (dwSet == INVALID_SET_FILE_POINTER)
        {
            DWORD dwErr = GetLastError();
            ((WIN32FILE_IOWIN*)stream) -> error=(int)dwErr;
            ret = -1;
        }
        else
            ret=(long)dwSet;
    }
    return ret;
}

long ZCALLBACK win32_seek_file_func (voidpf opaque, voidpf stream, uLong offset, int origin)
   // voidpf opaque;
   // voidpf stream;
   // uLong offset;
   // int origin;
{
    DWORD dwMoveMethod=0xFFFFFFFF;
    HANDLE hFile = NULL;

    long ret=-1;
    if (stream!=NULL)
        hFile = ((WIN32FILE_IOWIN*)stream) -> hf;
    switch (origin)
    {
    case ZLIB_FILEFUNC_SEEK_CUR :
        dwMoveMethod = FILE_CURRENT;
        break;
    case ZLIB_FILEFUNC_SEEK_END :
        dwMoveMethod = FILE_END;
        break;
    case ZLIB_FILEFUNC_SEEK_SET :
        dwMoveMethod = FILE_BEGIN;
        break;
    default: return -1;
    }

    if (hFile != NULL)
    {
        DWORD dwSet = SetFilePointer(hFile, offset, NULL, dwMoveMethod);
        if (dwSet == INVALID_SET_FILE_POINTER)
        {
            DWORD dwErr = GetLastError();
            ((WIN32FILE_IOWIN*)stream) -> error=(int)dwErr;
            ret = -1;
        }
        else
            ret=0;
    }
    return ret;
}

int ZCALLBACK win32_close_file_func (voidpf opaque, voidpf stream)
   // voidpf opaque;
   // voidpf stream;
{
    int ret=-1;

    if (stream!=NULL)
    {
        HANDLE hFile;
        hFile = ((WIN32FILE_IOWIN*)stream) -> hf;
        if (hFile != NULL)
        {
            CloseHandle(hFile);
            ret=0;
        }
        free(stream);
    }
    return ret;
}

int ZCALLBACK win32_error_file_func (voidpf opaque, voidpf stream)
   // voidpf opaque;
   // voidpf stream;
{
    int ret=-1;
    if (stream!=NULL)
    {
        ret = ((WIN32FILE_IOWIN*)stream) -> error;
    }
    return ret;
}

void fill_win32_filefunc (zlib_filefunc_def* pzlib_filefunc_def)
  // zlib_filefunc_def* pzlib_filefunc_def;
{
    pzlib_filefunc_def->zopen_file = win32_open_file_func;
    pzlib_filefunc_def->zread_file = win32_read_file_func;
    pzlib_filefunc_def->zwrite_file = win32_write_file_func;
    pzlib_filefunc_def->ztell_file = win32_tell_file_func;
    pzlib_filefunc_def->zseek_file = win32_seek_file_func;
    pzlib_filefunc_def->zclose_file = win32_close_file_func;
    pzlib_filefunc_def->zerror_file = win32_error_file_func;
    pzlib_filefunc_def->opaque=NULL;
}


/* zip.c -- IO on .zip files using zlib
   Version 1.01h, December 28th, 2009

   27 Dec 2004 Rolf Kalbermatter
   Modification to zipOpen2 to support globalComment retrieval.

   Copyright (C) 1998-2009 Gilles Vollant

   Read zip.h for more info
*/


#ifndef local
#  define local static
#endif
/* compile with -Dlocal if your debugger can't find static symbols */

#ifndef VERSIONMADEBY
# define VERSIONMADEBY   (0x0) /* platform depedent */
#endif

#ifndef Z_BUFSIZE
#define Z_BUFSIZE (16384)
#endif

#ifndef Z_MAXFILENAMEINZIP
#define Z_MAXFILENAMEINZIP (256)
#endif

#ifndef ALLOC
# define ALLOC(size) (malloc(size))
#endif
#ifndef TRYFREE
# define TRYFREE(p) {if (p) free(p);}
#endif

/*
#define SIZECENTRALDIRITEM (0x2e)
#define SIZEZIPLOCALHEADER (0x1e)
*/

/* I've found an old Unix (a SunOS 4.1.3_U1) without all SEEK_* defined.... */

#ifndef SEEK_CUR
#define SEEK_CUR    1
#endif

#ifndef SEEK_END
#define SEEK_END    2
#endif

#ifndef SEEK_SET
#define SEEK_SET    0
#endif

#ifndef DEF_MEM_LEVEL
#if MAX_MEM_LEVEL >= 8
#  define DEF_MEM_LEVEL 8
#else
#  define DEF_MEM_LEVEL  MAX_MEM_LEVEL
#endif
#endif
const char zip_copyright[] =
   " zip 1.01 Copyright 1998-2004 Gilles Vollant - http://www.winimage.com/zLibDll";


#define SIZEDATA_INDATABLOCK (4096-(4*4))

#define LOCALHEADERMAGIC    (0x04034b50)
#define CENTRALHEADERMAGIC  (0x02014b50)
#define ENDHEADERMAGIC      (0x06054b50)

#define FLAG_LOCALHEADER_OFFSET (0x06)
#define CRC_LOCALHEADER_OFFSET  (0x0e)

#define SIZECENTRALHEADER (0x2e) /* 46 */

typedef struct linkedlist_datablock_internal_s
{
  struct linkedlist_datablock_internal_s* next_datablock;
  uLong  avail_in_this_block;
  uLong  filled_in_this_block;
  uLong  unused; /* for future use and alignement */
  unsigned char data[SIZEDATA_INDATABLOCK];
} linkedlist_datablock_internal;

typedef struct linkedlist_data_s
{
    linkedlist_datablock_internal* first_block;
    linkedlist_datablock_internal* last_block;
} linkedlist_data;


typedef struct
{
    z_stream stream;            /* zLib stream structure for inflate */
    int  stream_initialised;    /* 1 is stream is initialised */
    uInt pos_in_buffered_data;  /* last written byte in buffered_data */

    uLong pos_local_header;     /* offset of the local header of the file
                                     currenty writing */
    char* central_header;       /* central header data for the current file */
    uLong size_centralheader;   /* size of the central header for cur file */
    uLong flag;                 /* flag of the file currently writing */

    int  method;                /* compression method of file currenty wr.*/
    int  raw;                   /* 1 for directly writing raw data */
    Byte buffered_data[Z_BUFSIZE];/* buffer contain compressed data to be writ*/
    uLong dosDate;
    uLong crc32;
    int  encrypt;
#ifndef NOCRYPT
    unsigned long keys[3];     /* keys defining the pseudo-random sequence */
    const unsigned long* pcrc_32_tab;
    int crypt_header_size;
#endif
} curfile_info;

typedef struct
{
    zlib_filefunc_def z_filefunc;
    voidpf filestream;        /* io structore of the zipfile */
    linkedlist_data central_dir;/* datablock with central dir in construction*/
    int  in_opened_file_inzip;  /* 1 if a file in the zip is currently writ.*/
    curfile_info ci;            /* info on the file curretly writing */

    uLong begin_pos;            /* position of the beginning of the zipfile */
    uLong add_position_when_writting_offset;
    uLong number_entry;
#ifndef NO_ADDFILEINEXISTINGZIP
    char *globalcomment;
#endif
} zip_internal;



#ifndef NOCRYPT
#define INCLUDECRYPTINGCODE_IFCRYPTALLOWED
#include "crypt.h"
#endif

local linkedlist_datablock_internal* allocate_new_datablock()
{
    linkedlist_datablock_internal* ldi;
    ldi = (linkedlist_datablock_internal*)
                 ALLOC(sizeof(linkedlist_datablock_internal));
    if (ldi!=NULL)
    {
        ldi->next_datablock = NULL ;
        ldi->filled_in_this_block = 0 ;
        ldi->avail_in_this_block = SIZEDATA_INDATABLOCK ;
    }
    return ldi;
}

local void free_datablock(linkedlist_datablock_internal* ldi)
    // linkedlist_datablock_internal* ldi;
{
    while (ldi!=NULL)
    {
        linkedlist_datablock_internal* ldinext = ldi->next_datablock;
        TRYFREE(ldi);
        ldi = ldinext;
    }
}

local void init_linkedlist(linkedlist_data* ll)
    // linkedlist_data* ll;
{
    ll->first_block = ll->last_block = NULL;
}

local void free_linkedlist(linkedlist_data* ll)
    // linkedlist_data* ll;
{
    free_datablock(ll->first_block);
    ll->first_block = ll->last_block = NULL;
}


local int add_data_in_datablock(linkedlist_data* ll,const void* buf,uLong len)
    // linkedlist_data* ll;
    // const void* buf;
    // uLong len;
{
    linkedlist_datablock_internal* ldi;
    const unsigned char* from_copy;

    if (ll==NULL)
        return ZIP_INTERNALERROR;

    if (ll->last_block == NULL)
    {
        ll->first_block = ll->last_block = allocate_new_datablock();
        if (ll->first_block == NULL)
            return ZIP_INTERNALERROR;
    }

    ldi = ll->last_block;
    from_copy = (unsigned char*)buf;

    while (len>0)
    {
        uInt copy_this;
        uInt i;
        unsigned char* to_copy;

        if (ldi->avail_in_this_block==0)
        {
            ldi->next_datablock = allocate_new_datablock();
            if (ldi->next_datablock == NULL)
                return ZIP_INTERNALERROR;
            ldi = ldi->next_datablock ;
            ll->last_block = ldi;
        }

        if (ldi->avail_in_this_block < len)
            copy_this = (uInt)ldi->avail_in_this_block;
        else
            copy_this = (uInt)len;

        to_copy = &(ldi->data[ldi->filled_in_this_block]);

        for (i=0;i<copy_this;i++)
            *(to_copy+i)=*(from_copy+i);

        ldi->filled_in_this_block += copy_this;
        ldi->avail_in_this_block -= copy_this;
        from_copy += copy_this ;
        len -= copy_this;
    }
    return ZIP_OK;
}



/****************************************************************************/

#ifndef NO_ADDFILEINEXISTINGZIP
/* ===========================================================================
   Inputs a long in LSB order to the given file
   nbByte == 1, 2 or 4 (byte, short or long)
*/

local int ziplocal_putValue OF((const zlib_filefunc_def* pzlib_filefunc_def,
                                voidpf filestream, uLong x, int nbByte));
local int ziplocal_putValue (const zlib_filefunc_def* pzlib_filefunc_def, voidpf filestream, uLong x, int nbByte)
   /*
    const zlib_filefunc_def* pzlib_filefunc_def;
    voidpf filestream;
    uLong x;
    int nbByte;
      */
{
    unsigned char buf[4];
    int n;
    for (n = 0; n < nbByte; n++)
    {
        buf[n] = (unsigned char)(x & 0xff);
        x >>= 8;
    }
    if (x != 0)
      {     /* data overflow - hack for ZIP64 (X Roche) */
      for (n = 0; n < nbByte; n++)
        {
          buf[n] = 0xff;
        }
      }

    if (ZWRITE(*pzlib_filefunc_def,filestream,buf,nbByte)!=(uLong)nbByte)
        return ZIP_ERRNO;
    else
        return ZIP_OK;
}

local void ziplocal_putValue_inmemory OF((void* dest, uLong x, int nbByte));
local void ziplocal_putValue_inmemory (void* dest, uLong x, int nbByte)
   /*
    void* dest;
    uLong x;
    int nbByte;
    */
{
    unsigned char* buf=(unsigned char*)dest;
    int n;
    for (n = 0; n < nbByte; n++) {
        buf[n] = (unsigned char)(x & 0xff);
        x >>= 8;
    }

    if (x != 0)
    {     /* data overflow - hack for ZIP64 */
       for (n = 0; n < nbByte; n++)
       {
          buf[n] = 0xff;
       }
    }
}

/****************************************************************************/


local uLong ziplocal_TmzDateToDosDate(const tm_zip* ptm,uLong dosDate)
    // const tm_zip* ptm;
    // uLong dosDate;
{
    uLong year = (uLong)ptm->tm_year;
    if (year>=1980)
        year-=1980;
    else if (year>=80)
        year-=80;
    return
      (uLong) (((ptm->tm_mday) + (32 * (ptm->tm_mon+1)) + (512 * year)) << 16) |
        ((ptm->tm_sec/2) + (32* ptm->tm_min) + (2048 * (uLong)ptm->tm_hour));
}


/****************************************************************************/
local int ziplocal_getByte OF((
    const zlib_filefunc_def* pzlib_filefunc_def,
    voidpf filestream,
    int *pix));
    
    local int ziplocal_getByte(const zlib_filefunc_def* pzlib_filefunc_def,voidpf filestream,int *pix)
    // const zlib_filefunc_def* pzlib_filefunc_def;
    // voidpf filestream;
    // int *pix;
{
    unsigned char c;
    int err = (int)ZREAD(*pzlib_filefunc_def,filestream,&c,1);
    if (err==1)
    {
        *pix = (int)c;
        return ZIP_OK;
    }
    else
    {
        if (ZERROR(*pzlib_filefunc_def,filestream))
            return ZIP_ERRNO;
        else
            return ZIP_EOF;
    }
}


/* ===========================================================================
   Reads a long in LSB order from the given gz_stream. Sets
*/
local int ziplocal_getShort OF((
    const zlib_filefunc_def* pzlib_filefunc_def,
    voidpf filestream,
    uLong *pX));

local int ziplocal_getShort (const zlib_filefunc_def* pzlib_filefunc_def,voidpf filestream,uLong *pX)
   /*
    const zlib_filefunc_def* pzlib_filefunc_def;
    voidpf filestream;
    uLong *pX;
    */
{
    uLong x ;
    int i = 0;
    int err;

    err = ziplocal_getByte(pzlib_filefunc_def,filestream,&i);
    x = (uLong)i;

    if (err==ZIP_OK)
        err = ziplocal_getByte(pzlib_filefunc_def,filestream,&i);
    x += ((uLong)i)<<8;

    if (err==ZIP_OK)
        *pX = x;
    else
        *pX = 0;
    return err;
}

local int ziplocal_getLong OF((
    const zlib_filefunc_def* pzlib_filefunc_def,
    voidpf filestream,
    uLong *pX));

local int ziplocal_getLong (const zlib_filefunc_def* pzlib_filefunc_def,voidpf filestream,uLong *pX)
   /*
    const zlib_filefunc_def* pzlib_filefunc_def;
    voidpf filestream;
    uLong *pX;
    */
{
    uLong x ;
    int i = 0;
    int err;

    err = ziplocal_getByte(pzlib_filefunc_def,filestream,&i);
    x = (uLong)i;

    if (err==ZIP_OK)
        err = ziplocal_getByte(pzlib_filefunc_def,filestream,&i);
    x += ((uLong)i)<<8;

    if (err==ZIP_OK)
        err = ziplocal_getByte(pzlib_filefunc_def,filestream,&i);
    x += ((uLong)i)<<16;

    if (err==ZIP_OK)
        err = ziplocal_getByte(pzlib_filefunc_def,filestream,&i);
    x += ((uLong)i)<<24;

    if (err==ZIP_OK)
        *pX = x;
    else
        *pX = 0;
    return err;
}

#ifndef BUFREADCOMMENT
#define BUFREADCOMMENT (0x400)
#endif
/*
  Locate the Central directory of a zipfile (at the end, just before
    the global comment)
   Fix from Riccardo Cohen
*/
local uLong ziplocal_SearchCentralDir OF((
    const zlib_filefunc_def* pzlib_filefunc_def,
    voidpf filestream));

local uLong ziplocal_SearchCentralDir(const zlib_filefunc_def* pzlib_filefunc_def,voidpf filestream)
     // const zlib_filefunc_def* pzlib_filefunc_def;
     // voidpf filestream;
{
     unsigned char* buf;
     uLong uSizeFile;
     uLong uBackRead;
     uLong uMaxBack=0xffff; /* maximum size of global comment */
     uLong uPosFound=0;

     if (ZSEEK(*pzlib_filefunc_def,filestream,0,ZLIB_FILEFUNC_SEEK_END) != 0)
         return 0;


     uSizeFile = ZTELL(*pzlib_filefunc_def,filestream);

     if (uMaxBack>uSizeFile)
         uMaxBack = uSizeFile;

     buf = (unsigned char*)ALLOC(BUFREADCOMMENT+4);
     if (buf==NULL)
         return 0;

     uBackRead = 4;
     while (uBackRead<uMaxBack)
     {
         uLong uReadSize,uReadPos ;
         int i;
         if (uBackRead+BUFREADCOMMENT>uMaxBack)
             uBackRead = uMaxBack;
         else
             uBackRead+=BUFREADCOMMENT;
         uReadPos = uSizeFile-uBackRead ;

         uReadSize = ((BUFREADCOMMENT+4) < (uSizeFile-uReadPos)) ?
                      (BUFREADCOMMENT+4) : (uSizeFile-uReadPos);
         if (ZSEEK(*pzlib_filefunc_def,filestream,uReadPos,ZLIB_FILEFUNC_SEEK_SET)!=0)
             break;

         if (ZREAD(*pzlib_filefunc_def,filestream,buf,uReadSize)!=uReadSize)
             break;

         for (i=(int)uReadSize-3; (i--)>0;)
             if (((*(buf+i))==0x50) && ((*(buf+i+1))==0x4b) &&
                 ((*(buf+i+2))==0x05) && ((*(buf+i+3))==0x06))
             {
                 uPosFound = uReadPos+i;
                 break;
             }

         if (uPosFound!=0)
             break;
     }
     TRYFREE(buf);
     return uPosFound;
}

#endif /* !NO_ADDFILEINEXISTINGZIP*/

/************************************************************/
extern zipFile ZEXPORT zipOpen2 (const char *pathname, int append, zipcharpc* globalcomment, zlib_filefunc_def* pzlib_filefunc_def)
   /*
    const char *pathname;
    int append;
    zipcharpc* globalcomment;
    zlib_filefunc_def* pzlib_filefunc_def;
    */
{
    zip_internal ziinit;
    zip_internal* zi;
    int err=ZIP_OK;


    /* PH99 COMMENTED
    if (pzlib_filefunc_def==NULL)
        fill_fopen_filefunc(&ziinit.z_filefunc);
    else
    */
        ziinit.z_filefunc = *pzlib_filefunc_def;

    ziinit.filestream = (*(ziinit.z_filefunc.zopen_file))
                 (ziinit.z_filefunc.opaque,
                  pathname,
                  (append == APPEND_STATUS_CREATE) ?
                  (ZLIB_FILEFUNC_MODE_READ | ZLIB_FILEFUNC_MODE_WRITE | ZLIB_FILEFUNC_MODE_CREATE) :
                    (ZLIB_FILEFUNC_MODE_READ | ZLIB_FILEFUNC_MODE_WRITE | ZLIB_FILEFUNC_MODE_EXISTING));

    if (ziinit.filestream == NULL)
        return NULL;
    if (append == APPEND_STATUS_CREATEAFTER)
        ZSEEK(ziinit.z_filefunc,ziinit.filestream,0,SEEK_END);
    ziinit.begin_pos = ZTELL(ziinit.z_filefunc,ziinit.filestream);
    ziinit.in_opened_file_inzip = 0;
    ziinit.ci.stream_initialised = 0;
    ziinit.number_entry = 0;
    ziinit.add_position_when_writting_offset = 0;
    init_linkedlist(&(ziinit.central_dir));


    zi = (zip_internal*)ALLOC(sizeof(zip_internal));
    if (zi==NULL)
    {
        ZCLOSE(ziinit.z_filefunc,ziinit.filestream);
        return NULL;
    }

    /* now we add file in a zipfile */
#    ifndef NO_ADDFILEINEXISTINGZIP
    ziinit.globalcomment = NULL;
    if (append == APPEND_STATUS_ADDINZIP)
    {
        uLong byte_before_the_zipfile;/* byte before the zipfile, (>0 for sfx)*/

        uLong size_central_dir;     /* size of the central directory  */
        uLong offset_central_dir;   /* offset of start of central directory */
        uLong central_pos,uL;

        uLong number_disk;          /* number of the current dist, used for
                                    spaning ZIP, unsupported, always 0*/
        uLong number_disk_with_CD;  /* number the the disk with central dir, used
                                    for spaning ZIP, unsupported, always 0*/
        uLong number_entry;
        uLong number_entry_CD;      /* total number of entries in
                                    the central dir
                                    (same than number_entry on nospan) */
        uLong size_comment;

        central_pos = ziplocal_SearchCentralDir(&ziinit.z_filefunc,ziinit.filestream);
/* disable to allow appending to empty ZIP archive
        if (central_pos==0)
            err=ZIP_ERRNO;
*/
        if (ZSEEK(ziinit.z_filefunc, ziinit.filestream,
                                        central_pos,ZLIB_FILEFUNC_SEEK_SET)!=0)
            err=ZIP_ERRNO;

        /* the signature, already checked */
        if (ziplocal_getLong(&ziinit.z_filefunc, ziinit.filestream,&uL)!=ZIP_OK)
            err=ZIP_ERRNO;

        /* number of this disk */
        if (ziplocal_getShort(&ziinit.z_filefunc, ziinit.filestream,&number_disk)!=ZIP_OK)
            err=ZIP_ERRNO;

        /* number of the disk with the start of the central directory */
        if (ziplocal_getShort(&ziinit.z_filefunc, ziinit.filestream,&number_disk_with_CD)!=ZIP_OK)
            err=ZIP_ERRNO;

        /* total number of entries in the central dir on this disk */
        if (ziplocal_getShort(&ziinit.z_filefunc, ziinit.filestream,&number_entry)!=ZIP_OK)
            err=ZIP_ERRNO;

        /* total number of entries in the central dir */
        if (ziplocal_getShort(&ziinit.z_filefunc, ziinit.filestream,&number_entry_CD)!=ZIP_OK)
            err=ZIP_ERRNO;

        if ((number_entry_CD!=number_entry) ||
            (number_disk_with_CD!=0) ||
            (number_disk!=0))
            err=ZIP_BADZIPFILE;

        /* size of the central directory */
        if (ziplocal_getLong(&ziinit.z_filefunc, ziinit.filestream,&size_central_dir)!=ZIP_OK)
            err=ZIP_ERRNO;

        /* offset of start of central directory with respect to the
            starting disk number */
        if (ziplocal_getLong(&ziinit.z_filefunc, ziinit.filestream,&offset_central_dir)!=ZIP_OK)
            err=ZIP_ERRNO;

        /* zipfile global comment length */
        if (ziplocal_getShort(&ziinit.z_filefunc, ziinit.filestream,&size_comment)!=ZIP_OK)
            err=ZIP_ERRNO;

        if ((central_pos<offset_central_dir+size_central_dir) &&
            (err==ZIP_OK))
            err=ZIP_BADZIPFILE;

        if (err!=ZIP_OK)
        {
            ZCLOSE(ziinit.z_filefunc, ziinit.filestream);
            return NULL;
        }

        if (size_comment>0)
        {
            ziinit.globalcomment = (char*)ALLOC(size_comment+1);
            if (ziinit.globalcomment)
            {
               size_comment = ZREAD(ziinit.z_filefunc, ziinit.filestream,ziinit.globalcomment,size_comment);
               ziinit.globalcomment[size_comment]=0;
            }
        }

        byte_before_the_zipfile = central_pos -
                                (offset_central_dir+size_central_dir);
        ziinit.add_position_when_writting_offset = byte_before_the_zipfile;

        {
            uLong size_central_dir_to_read = size_central_dir;
            size_t buf_size = SIZEDATA_INDATABLOCK;
            void* buf_read = (void*)ALLOC(buf_size);
            if (ZSEEK(ziinit.z_filefunc, ziinit.filestream,
                  offset_central_dir + byte_before_the_zipfile,
                  ZLIB_FILEFUNC_SEEK_SET) != 0)
                  err=ZIP_ERRNO;

            while ((size_central_dir_to_read>0) && (err==ZIP_OK))
            {
                uLong read_this = SIZEDATA_INDATABLOCK;
                if (read_this > size_central_dir_to_read)
                    read_this = size_central_dir_to_read;
                if (ZREAD(ziinit.z_filefunc, ziinit.filestream,buf_read,read_this) != read_this)
                    err=ZIP_ERRNO;

                if (err==ZIP_OK)
                    err = add_data_in_datablock(&ziinit.central_dir,buf_read,
                                                (uLong)read_this);
                size_central_dir_to_read-=read_this;
            }
            TRYFREE(buf_read);
        }
        ziinit.begin_pos = byte_before_the_zipfile;
        ziinit.number_entry = number_entry_CD;

        if (ZSEEK(ziinit.z_filefunc, ziinit.filestream,
                  offset_central_dir+byte_before_the_zipfile,ZLIB_FILEFUNC_SEEK_SET)!=0)
            err=ZIP_ERRNO;
    }

    if (globalcomment)
    {
      *globalcomment = ziinit.globalcomment;
    }
#    endif /* !NO_ADDFILEINEXISTINGZIP*/

    if (err != ZIP_OK)
    {
#    ifndef NO_ADDFILEINEXISTINGZIP
        TRYFREE(ziinit.globalcomment);
#    endif /* !NO_ADDFILEINEXISTINGZIP*/
        TRYFREE(zi);
        return NULL;
    }
    else
    {
        *zi = ziinit;
        return (zipFile)zi;
    }
}

extern zipFile ZEXPORT zipOpen (const char *pathname, int append)
    // const char *pathname;
    // int append;
{
    return zipOpen2(pathname,append,NULL,NULL);
}

extern int ZEXPORT zipOpenNewFileInZip4 ( zipFile file,  const char* filename, const zip_fileinfo* zipfi, const void* extrafield_local,
                                          uInt size_extrafield_local,  const void* extrafield_global, uInt size_extrafield_global,
                                          const char* comment, int method, int level, int raw, int windowBits, int memLevel,
                                          int strategy, const char* password, uLong crcForCrypting, uLong versionMadeBy,uLong flagBase)
    /*                                          
    zipFile file;
    const char* filename;
    const zip_fileinfo* zipfi;
    const void* extrafield_local;
    uInt size_extrafield_local;
    const void* extrafield_global;
    uInt size_extrafield_global;
    const char* comment;
    int method;
    int level;
    int raw;
    int windowBits;
    int memLevel;
    int strategy;
    const char* password;
    uLong crcForCrypting;
    uLong versionMadeBy;
    uLong flagBase;
    */
{
    zip_internal* zi;
    uInt size_filename;
    uInt size_comment;
    uInt i;
    int err = ZIP_OK;

#    ifdef NOCRYPT
    if (password != NULL)
        return ZIP_PARAMERROR;
#    endif

    if (file == NULL)
        return ZIP_PARAMERROR;
    if ((method!=0) && (method!=Z_DEFLATED))
        return ZIP_PARAMERROR;

    zi = (zip_internal*)file;

    if (zi->in_opened_file_inzip == 1)
    {
        err = zipCloseFileInZip (file);
        if (err != ZIP_OK)
            return err;
    }


    if (filename==NULL)
        filename="-";

    if (comment==NULL)
        size_comment = 0;
    else
        size_comment = (uInt)strlen(comment);

    size_filename = (uInt)strlen(filename);

    if (zipfi == NULL)
        zi->ci.dosDate = 0;
    else
    {
        if (zipfi->dosDate != 0)
            zi->ci.dosDate = zipfi->dosDate;
        else zi->ci.dosDate = ziplocal_TmzDateToDosDate(&zipfi->tmz_date,zipfi->dosDate);
    }

    zi->ci.flag = flagBase;
    if ((level==8) || (level==9))
      zi->ci.flag |= 2;
    if ((level==2))
      zi->ci.flag |= 4;
    if ((level==1))
      zi->ci.flag |= 6;
    if (password != NULL)
      zi->ci.flag |= 1;

    zi->ci.crc32 = 0;
    zi->ci.method = method;
    zi->ci.encrypt = 0;
    zi->ci.stream_initialised = 0;
    zi->ci.pos_in_buffered_data = 0;
    zi->ci.raw = raw;
    zi->ci.pos_local_header = ZTELL(zi->z_filefunc,zi->filestream) ;
    zi->ci.size_centralheader = SIZECENTRALHEADER + size_filename +
                                      size_extrafield_global + size_comment;
    zi->ci.central_header = (char*)ALLOC((uInt)zi->ci.size_centralheader);

    ziplocal_putValue_inmemory(zi->ci.central_header,(uLong)CENTRALHEADERMAGIC,4);
    /* version info */
    ziplocal_putValue_inmemory(zi->ci.central_header+4,(uLong)versionMadeBy,2);
    ziplocal_putValue_inmemory(zi->ci.central_header+6,(uLong)20,2);
    ziplocal_putValue_inmemory(zi->ci.central_header+8,(uLong)zi->ci.flag,2);
    ziplocal_putValue_inmemory(zi->ci.central_header+10,(uLong)zi->ci.method,2);
    ziplocal_putValue_inmemory(zi->ci.central_header+12,(uLong)zi->ci.dosDate,4);
    ziplocal_putValue_inmemory(zi->ci.central_header+16,(uLong)0,4); /*crc*/
    ziplocal_putValue_inmemory(zi->ci.central_header+20,(uLong)0,4); /*compr size*/
    ziplocal_putValue_inmemory(zi->ci.central_header+24,(uLong)0,4); /*uncompr size*/
    ziplocal_putValue_inmemory(zi->ci.central_header+28,(uLong)size_filename,2);
    ziplocal_putValue_inmemory(zi->ci.central_header+30,(uLong)size_extrafield_global,2);
    ziplocal_putValue_inmemory(zi->ci.central_header+32,(uLong)size_comment,2);
    ziplocal_putValue_inmemory(zi->ci.central_header+34,(uLong)0,2); /*disk nm start*/

    if (zipfi==NULL)
        ziplocal_putValue_inmemory(zi->ci.central_header+36,(uLong)0,2);
    else
        ziplocal_putValue_inmemory(zi->ci.central_header+36,(uLong)zipfi->internal_fa,2);

    if (zipfi==NULL)
        ziplocal_putValue_inmemory(zi->ci.central_header+38,(uLong)0,4);
    else
        ziplocal_putValue_inmemory(zi->ci.central_header+38,(uLong)zipfi->external_fa,4);

    ziplocal_putValue_inmemory(zi->ci.central_header+42,(uLong)zi->ci.pos_local_header- zi->add_position_when_writting_offset,4);

    for (i=0;i<size_filename;i++)
        *(zi->ci.central_header+SIZECENTRALHEADER+i) = *(filename+i);

    for (i=0;i<size_extrafield_global;i++)
        *(zi->ci.central_header+SIZECENTRALHEADER+size_filename+i) =
              *(((const char*)extrafield_global)+i);

    for (i=0;i<size_comment;i++)
        *(zi->ci.central_header+SIZECENTRALHEADER+size_filename+
              size_extrafield_global+i) = *(comment+i);
    if (zi->ci.central_header == NULL)
        return ZIP_INTERNALERROR;

    /* write the local header */
    err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,(uLong)LOCALHEADERMAGIC,4);

    if (err==ZIP_OK)
        err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,(uLong)20,2);/* version needed to extract */
    if (err==ZIP_OK)
        err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,(uLong)zi->ci.flag,2);

    if (err==ZIP_OK)
        err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,(uLong)zi->ci.method,2);

    if (err==ZIP_OK)
        err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,(uLong)zi->ci.dosDate,4);

    if (err==ZIP_OK)
        err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,(uLong)0,4); /* crc 32, unknown */
    if (err==ZIP_OK)
        err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,(uLong)0,4); /* compressed size, unknown */
    if (err==ZIP_OK)
        err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,(uLong)0,4); /* uncompressed size, unknown */

    if (err==ZIP_OK)
        err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,(uLong)size_filename,2);

    if (err==ZIP_OK)
        err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,(uLong)size_extrafield_local,2);

    if ((err==ZIP_OK) && (size_filename>0))
        if (ZWRITE(zi->z_filefunc,zi->filestream,filename,size_filename)!=size_filename)
                err = ZIP_ERRNO;

    if ((err==ZIP_OK) && (size_extrafield_local>0))
        if (ZWRITE(zi->z_filefunc,zi->filestream,extrafield_local,size_extrafield_local)
                                                                           !=size_extrafield_local)
                err = ZIP_ERRNO;

    zi->ci.stream.avail_in = (uInt)0;
    zi->ci.stream.avail_out = (uInt)Z_BUFSIZE;
    zi->ci.stream.next_out = zi->ci.buffered_data;
    zi->ci.stream.total_in = 0;
    zi->ci.stream.total_out = 0;
    zi->ci.stream.data_type = Z_BINARY;

    if ((err==ZIP_OK) && (zi->ci.method == Z_DEFLATED) && (!zi->ci.raw))
    {
        zi->ci.stream.zalloc = (alloc_func)0;
        zi->ci.stream.zfree = (free_func)0;
        zi->ci.stream.opaque = (voidpf)0;

        if (windowBits>0)
            windowBits = -windowBits;

        err = deflateInit2(&zi->ci.stream, level,
               Z_DEFLATED, windowBits, memLevel, strategy);

        if (err==Z_OK)
            zi->ci.stream_initialised = 1;
    }
#    ifndef NOCRYPT
    zi->ci.crypt_header_size = 0;
    if ((err==Z_OK) && (password != NULL))
    {
        unsigned char bufHead[RAND_HEAD_LEN];
        unsigned int sizeHead;
        zi->ci.encrypt = 1;
        zi->ci.pcrc_32_tab = get_crc_table();
        /*init_keys(password,zi->ci.keys,zi->ci.pcrc_32_tab);*/

        sizeHead=crypthead(password,bufHead,RAND_HEAD_LEN,zi->ci.keys,zi->ci.pcrc_32_tab,crcForCrypting);
        zi->ci.crypt_header_size = sizeHead;

        if (ZWRITE(zi->z_filefunc,zi->filestream,bufHead,sizeHead) != sizeHead)
                err = ZIP_ERRNO;
    }
#    endif

    if (err==Z_OK)
        zi->in_opened_file_inzip = 1;
    return err;
}

extern int ZEXPORT zipOpenNewFileInZip2( zipFile file,    const char* filename,    const zip_fileinfo* zipfi,    const void* extrafield_local,
                                         uInt size_extrafield_local,    const void* extrafield_global,    uInt size_extrafield_global,
                                         const char* comment, int method,  int level, int raw)
    /*                                           
    zipFile file;
    const char* filename;
    const zip_fileinfo* zipfi;
    const void* extrafield_local;
    uInt size_extrafield_local;
    const void* extrafield_global;
    uInt size_extrafield_global;
    const char* comment;
    int method;
    int level;
    int raw;
      */
{
    return zipOpenNewFileInZip4 (file, filename, zipfi,
                                 extrafield_local, size_extrafield_local,
                                 extrafield_global, size_extrafield_global,
                                 comment, method, level, raw,
                                 -MAX_WBITS, DEF_MEM_LEVEL, Z_DEFAULT_STRATEGY,
                                 NULL, 0, VERSIONMADEBY, 0);
}

extern int ZEXPORT zipOpenNewFileInZip3 (zipFile file,
    const char* filename,
    const zip_fileinfo* zipfi,
    const void* extrafield_local,
    uInt size_extrafield_local,
    const void* extrafield_global,
    uInt size_extrafield_global,
    const char* comment,
    int method,
    int level,
    int raw,
    int windowBits,
    int memLevel,
    int strategy,
    const char* password,
    uLong crcForCrypting)
    /*
    zipFile file;
    const char* filename;
    const zip_fileinfo* zipfi;
    const void* extrafield_local;
    uInt size_extrafield_local;
    const void* extrafield_global;
    uInt size_extrafield_global;
    const char* comment;
    int method;
    int level;
    int raw;
    int windowBits;
    int memLevel;
    int strategy;
    const char* password;
    uLong crcForCrypting;
    */
{
    return zipOpenNewFileInZip4 (file, filename, zipfi,
                                 extrafield_local, size_extrafield_local,
                                 extrafield_global, size_extrafield_global,
                                 comment, method, level, raw,
                                 windowBits, memLevel, strategy,
                                 password, crcForCrypting, VERSIONMADEBY, 0);
}


extern int ZEXPORT zipOpenNewFileInZip (zipFile file,
    const char* filename,
    const zip_fileinfo* zipfi,
    const void* extrafield_local,
    uInt size_extrafield_local,
    const void* extrafield_global,
    uInt size_extrafield_global,
    const char* comment,
    int method,
    int level)
    /*
    zipFile file;
    const char* filename;
    const zip_fileinfo* zipfi;
    const void* extrafield_local;
    uInt size_extrafield_local;
    const void* extrafield_global;
    uInt size_extrafield_global;
    const char* comment;
    int method;
    int level;
    */
{
    return zipOpenNewFileInZip4 (file, filename, zipfi,
                                 extrafield_local, size_extrafield_local,
                                 extrafield_global, size_extrafield_global,
                                 comment, method, level, 0,
                                 -MAX_WBITS, DEF_MEM_LEVEL, Z_DEFAULT_STRATEGY,
                                 NULL, 0, VERSIONMADEBY, 0);
}

local int zipFlushWriteBuffer(zip_internal* zi)
  // zip_internal* zi;
{
    int err=ZIP_OK;

    if (zi->ci.encrypt != 0)
    {
#ifndef NOCRYPT
        uInt i;
        int t;
        for (i=0;i<zi->ci.pos_in_buffered_data;i++)
            zi->ci.buffered_data[i] = zencode(zi->ci.keys, zi->ci.pcrc_32_tab,
                                       zi->ci.buffered_data[i],t);
#endif
    }
    if (ZWRITE(zi->z_filefunc,zi->filestream,zi->ci.buffered_data,zi->ci.pos_in_buffered_data)
                                                                    !=zi->ci.pos_in_buffered_data)
      err = ZIP_ERRNO;
    zi->ci.pos_in_buffered_data = 0;
    return err;
}

extern int ZEXPORT zipWriteInFileInZip (zipFile file, const void* buf, unsigned len)
    /*
    zipFile file;
    const void* buf;
    unsigned len;
    */
{
    zip_internal* zi;
    int err=ZIP_OK;

    if (file == NULL)
        return ZIP_PARAMERROR;
    zi = (zip_internal*)file;

    if (zi->in_opened_file_inzip == 0)
        return ZIP_PARAMERROR;

    zi->ci.stream.next_in = (Bytef*)buf;
    zi->ci.stream.avail_in = len;
    zi->ci.crc32 = crc32(zi->ci.crc32,(const Bytef *)buf,(uInt)len);

    while ((err==ZIP_OK) && (zi->ci.stream.avail_in>0))
    {
        if (zi->ci.stream.avail_out == 0)
        {
            if (zipFlushWriteBuffer(zi) == ZIP_ERRNO)
                err = ZIP_ERRNO;
            zi->ci.stream.avail_out = (uInt)Z_BUFSIZE;
            zi->ci.stream.next_out = zi->ci.buffered_data;
        }


        if(err != ZIP_OK)
            break;

        if ((zi->ci.method == Z_DEFLATED) && (!zi->ci.raw))
        {
            uLong uTotalOutBefore = zi->ci.stream.total_out;
            err=deflate(&zi->ci.stream,  Z_NO_FLUSH);
            zi->ci.pos_in_buffered_data += (uInt)(zi->ci.stream.total_out - uTotalOutBefore) ;

        }
        else
        {
            uInt copy_this,i;
            if (zi->ci.stream.avail_in < zi->ci.stream.avail_out)
                copy_this = zi->ci.stream.avail_in;
            else
                copy_this = zi->ci.stream.avail_out;
            for (i=0;i<copy_this;i++)
                *(((char*)zi->ci.stream.next_out)+i) =
                    *(((const char*)zi->ci.stream.next_in)+i);
            {
                zi->ci.stream.avail_in -= copy_this;
                zi->ci.stream.avail_out-= copy_this;
                zi->ci.stream.next_in+= copy_this;
                zi->ci.stream.next_out+= copy_this;
                zi->ci.stream.total_in+= copy_this;
                zi->ci.stream.total_out+= copy_this;
                zi->ci.pos_in_buffered_data += copy_this;
            }
        }
    }

    return err;
}

extern int ZEXPORT zipCloseFileInZipRaw (zipFile file, uLong uncompressed_size, uLong crc32)
   /*
    zipFile file;
    uLong uncompressed_size;
    uLong crc32;
    */
{
    zip_internal* zi;
    uLong compressed_size;
    int err=ZIP_OK;

    if (file == NULL)
        return ZIP_PARAMERROR;
    zi = (zip_internal*)file;

    if (zi->in_opened_file_inzip == 0)
        return ZIP_PARAMERROR;
    zi->ci.stream.avail_in = 0;

    if ((zi->ci.method == Z_DEFLATED) && (!zi->ci.raw))
        while (err==ZIP_OK)
    {
        uLong uTotalOutBefore;
        if (zi->ci.stream.avail_out == 0)
        {
            if (zipFlushWriteBuffer(zi) == ZIP_ERRNO)
                err = ZIP_ERRNO;
            zi->ci.stream.avail_out = (uInt)Z_BUFSIZE;
            zi->ci.stream.next_out = zi->ci.buffered_data;
        }
        uTotalOutBefore = zi->ci.stream.total_out;
        err=deflate(&zi->ci.stream,  Z_FINISH);
        zi->ci.pos_in_buffered_data += (uInt)(zi->ci.stream.total_out - uTotalOutBefore) ;
    }

    if (err==Z_STREAM_END)
        err=ZIP_OK; /* this is normal */

    if ((zi->ci.pos_in_buffered_data>0) && (err==ZIP_OK))
        if (zipFlushWriteBuffer(zi)==ZIP_ERRNO)
            err = ZIP_ERRNO;

    if ((zi->ci.method == Z_DEFLATED) && (!zi->ci.raw))
    {
        int tmp_err=deflateEnd(&zi->ci.stream);
        if (err == ZIP_OK)
            err = tmp_err;
        zi->ci.stream_initialised = 0;
    }

    if (!zi->ci.raw)
    {
        crc32 = (uLong)zi->ci.crc32;
        uncompressed_size = (uLong)zi->ci.stream.total_in;
    }
    compressed_size = (uLong)zi->ci.stream.total_out;
#    ifndef NOCRYPT
    compressed_size += zi->ci.crypt_header_size;
#    endif

    ziplocal_putValue_inmemory(zi->ci.central_header+16,crc32,4); /*crc*/
    ziplocal_putValue_inmemory(zi->ci.central_header+20,
                                compressed_size,4); /*compr size*/
    if (zi->ci.stream.data_type == Z_ASCII)
        ziplocal_putValue_inmemory(zi->ci.central_header+36,(uLong)Z_ASCII,2);
    ziplocal_putValue_inmemory(zi->ci.central_header+24,
                                uncompressed_size,4); /*uncompr size*/

    if (err==ZIP_OK)
        err = add_data_in_datablock(&zi->central_dir,zi->ci.central_header,
                                       (uLong)zi->ci.size_centralheader);
    free(zi->ci.central_header);

    if (err==ZIP_OK)
    {
        long cur_pos_inzip = ZTELL(zi->z_filefunc,zi->filestream);
        if (ZSEEK(zi->z_filefunc,zi->filestream,
                  zi->ci.pos_local_header + 14,ZLIB_FILEFUNC_SEEK_SET)!=0)
            err = ZIP_ERRNO;

        if (err==ZIP_OK)
            err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,crc32,4); /* crc 32, unknown */

        if (err==ZIP_OK) /* compressed size, unknown */
            err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,compressed_size,4);

        if (err==ZIP_OK) /* uncompressed size, unknown */
            err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,uncompressed_size,4);

        if (ZSEEK(zi->z_filefunc,zi->filestream,
                  cur_pos_inzip,ZLIB_FILEFUNC_SEEK_SET)!=0)
            err = ZIP_ERRNO;
    }

    zi->number_entry ++;
    zi->in_opened_file_inzip = 0;

    return err;
}

extern int ZEXPORT zipCloseFileInZip (zipFile file)
    // zipFile file;
{
    return zipCloseFileInZipRaw (file,0,0);
}

extern int ZEXPORT zipClose (zipFile file, const char* global_comment)
    // zipFile file;
    // const char* global_comment;
{
    zip_internal* zi;
    int err = 0;
    uLong size_centraldir = 0;
    uLong centraldir_pos_inzip;
    uInt size_global_comment;
    if (file == NULL)
        return ZIP_PARAMERROR;
    zi = (zip_internal*)file;

    if (zi->in_opened_file_inzip == 1)
    {
        err = zipCloseFileInZip (file);
    }

#ifndef NO_ADDFILEINEXISTINGZIP
    if (global_comment==NULL)
        global_comment = zi->globalcomment;
#endif
    if (global_comment==NULL)
        size_global_comment = 0;
    else
        size_global_comment = (uInt)strlen(global_comment);

    centraldir_pos_inzip = ZTELL(zi->z_filefunc,zi->filestream);
    if (err==ZIP_OK)
    {
        linkedlist_datablock_internal* ldi = zi->central_dir.first_block ;
        while (ldi!=NULL)
        {
            if ((err==ZIP_OK) && (ldi->filled_in_this_block>0))
                if (ZWRITE(zi->z_filefunc,zi->filestream,
                           ldi->data,ldi->filled_in_this_block)
                              !=ldi->filled_in_this_block )
                    err = ZIP_ERRNO;

            size_centraldir += ldi->filled_in_this_block;
            ldi = ldi->next_datablock;
        }
    }
    free_linkedlist(&(zi->central_dir));

    if (err==ZIP_OK) /* Magic End */
        err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,(uLong)ENDHEADERMAGIC,4);

    if (err==ZIP_OK) /* number of this disk */
        err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,(uLong)0,2);

    if (err==ZIP_OK) /* number of the disk with the start of the central directory */
        err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,(uLong)0,2);

    if (err==ZIP_OK) /* total number of entries in the central dir on this disk */
        err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,(uLong)zi->number_entry,2);

    if (err==ZIP_OK) /* total number of entries in the central dir */
        err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,(uLong)zi->number_entry,2);

    if (err==ZIP_OK) /* size of the central directory */
        err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,(uLong)size_centraldir,4);

    if (err==ZIP_OK) /* offset of start of central directory with respect to the
                            starting disk number */
        err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,
                                (uLong)(centraldir_pos_inzip - zi->add_position_when_writting_offset),4);

    if (err==ZIP_OK) /* zipfile comment length */
        err = ziplocal_putValue(&zi->z_filefunc,zi->filestream,(uLong)size_global_comment,2);

    if ((err==ZIP_OK) && (size_global_comment>0))
        if (ZWRITE(zi->z_filefunc,zi->filestream,
                   global_comment,size_global_comment) != size_global_comment)
                err = ZIP_ERRNO;

    if (ZCLOSE(zi->z_filefunc,zi->filestream) != 0)
        if (err == ZIP_OK)
            err = ZIP_ERRNO;

#ifndef NO_ADDFILEINEXISTINGZIP
    TRYFREE(zi->globalcomment);
#endif
    TRYFREE(zi);

    return err;
}


// END MINIZIP CODE



   //  ellipsoid: index into the gEllipsoid[] array, in which
   //             a is the ellipsoid semimajor axis
   //             invf is the inverse of the ellipsoid flattening f
   //  dx, dy, dz: ellipsoid center with respect to WGS84 ellipsoid center
   //    x axis is the prime meridian
   //    y axis is 90 degrees east longitude
   //    z axis is the axis of rotation of the ellipsoid
   
   // The following values for dx, dy and dz were extracted from the output of
   // the GARMIN PCX5 program. The output also includes values for da and df, the
   // difference between the reference ellipsoid and the WGS84 ellipsoid semi-
   // major axis and flattening, respectively. These are replaced by the
   // data contained in the structure array gEllipsoid[], which was obtained from
   // the Defence Mapping Agency document number TR8350.2, "Department of Defense
   // World Geodetic System 1984."
   
   struct DATUM const gDatum[] = {
   //     name               ellipsoid   dx      dy       dz
         { "Adindan",                5,   -162,    -12,    206 },    // 0
         { "Afgooye",               15,    -43,   -163,     45 },    // 1
         { "Ain el Abd 1970",       14,   -150,   -251,     -2 },    // 2
         { "Anna 1 Astro 1965",      2,   -491,    -22,    435 },    // 3
         { "Arc 1950",               5,   -143,    -90,   -294 },    // 4
         { "Arc 1960",               5,   -160,     -8,   -300 },    // 5
         { "Ascension Island  58",  14,   -207,    107,     52 },    // 6
         { "Astro B4 Sorol Atoll",  14,    114,   -116,   -333 },    // 7
         { "Astro Beacon  E ",      14,    145,     75,   -272 },    // 8
         { "Astro DOS 71/4",        14,   -320,    550,   -494 },    // 9
         { "Astronomic Stn  52",    14,    124,   -234,    -25 },    // 10
         { "Australian Geod  66",    2,   -133,    -48,    148 },    // 11
         { "Australian Geod  84",    2,   -134,    -48,    149 },    // 12
         { "Bellevue (IGN)",        14,   -127,   -769,    472 },    // 13
         { "Bermuda 1957",           4,    -73,    213,    296 },    // 14
         { "Bogota Observatory",    14,    307,    304,   -318 },    // 15
         { "Campo Inchauspe",       14,   -148,    136,     90 },    // 16
         { "Canton Astro 1966",     14,    298,   -304,   -375 },    // 17
         { "Cape",                   5,   -136,   -108,   -292 },    // 18
         { "Cape Canaveral",         4,     -2,    150,    181 },    // 19
         { "Carthage",               5,   -263,      6,    431 },    // 20
         { "CH-1903",                3,    674,     15,    405 },    // 21
         { "Chatham 1971",          14,    175,    -38,    113 },    // 22
         { "Chua Astro",            14,   -134,    229,    -29 },    // 23
         { "Corrego Alegre",        14,   -206,    172,     -6 },    // 24
         { "Djakarta (Batavia)",     3,   -377,    681,    -50 },    // 25
         { "DOS 1968",              14,    230,   -199,   -752 },    // 26
         { "Easter Island 1967",    14,    211,    147,    111 },    // 27
         { "European 1950",         14,    -87,    -98,   -121 },    // 28
         { "European 1979",         14,    -86,    -98,   -119 },    // 29
         { "Finland Hayford",       14,    -78,   -231,    -97 },    // 30
         { "Gandajika Base",        14,   -133,   -321,     50 },    // 31
         { "Geodetic Datum  49",    14,     84,    -22,    209 },    // 32
         { "Guam 1963",              4,   -100,   -248,    259 },    // 33
         { "GUX 1 Astro",           14,    252,   -209,   -751 },    // 34
         { "Hjorsey 1955",          14,    -73,     46,    -86 },    // 35
         { "Hong Kong 1963",        14,   -156,   -271,   -189 },    // 36
         { "Indian Bangladesh",      6,    289,    734,    257 },    // 37
         { "Indian Thailand",        6,    214,    836,    303 },    // 38
         { "Ireland 1965",           1,    506,   -122,    611 },    // 39
         { "ISTS 073 Astro  69",    14,    208,   -435,   -229 },    // 40
         { "Johnston Island",       14,    191,    -77,   -204 },    // 41
         { "Kandawala",              6,    -97,    787,     86 },    // 42
         { "Kerguelen Island",      14,    145,   -187,    103 },    // 43
         { "Kertau 1948",            7,    -11,    851,      5 },    // 44
         { "L.C. 5 Astro",           4,     42,    124,    147 },    // 45
         { "Liberia 1964",           5,    -90,     40,     88 },    // 46
         { "Luzon Mindanao",         4,   -133,    -79,    -72 },    // 47
         { "Luzon Philippines",      4,   -133,    -77,    -51 },    // 48
         { "Mahe 1971",              5,     41,   -220,   -134 },    // 49
         { "Marco Astro",           14,   -289,   -124,     60 },    // 50
         { "Massawa",                3,    639,    405,     60 },    // 51
         { "Merchich",               5,     31,    146,     47 },    // 52
         { "Midway Astro 1961",     14,    912,    -58,   1227 },    // 53
         { "Minna",                  5,    -92,    -93,    122 },    // 54
         { "NAD27 Alaska",           4,     -5,    135,    172 },    // 55
         { "NAD27 Bahamas",          4,     -4,    154,    178 },    // 56
         { "NAD27 Canada",           4,    -10,    158,    187 },    // 57
         { "NAD27 Canal Zone",       4,      0,    125,    201 },    // 58
         { "NAD27 Caribbean",        4,     -7,    152,    178 },    // 59
         { "NAD27 Central",          4,      0,    125,    194 },    // 60
         { "NAD27 CONUS",            4,     -8,    160,    176 },    // 61
         { "NAD27 Cuba",             4,     -9,    152,    178 },    // 62
         { "NAD27 Greenland",        4,     11,    114,    195 },    // 63
         { "NAD27 Mexico",           4,    -12,    130,    190 },    // 64
         { "NAD27 San Salvador",     4,      1,    140,    165 },    // 65
         { "NAD83",                 11,      0,      0,      0 },    // 66
         { "Nahrwn Masirah Ilnd",    5,   -247,   -148,    369 },    // 67
         { "Nahrwn Saudi Arbia",     5,   -231,   -196,    482 },    // 68
         { "Nahrwn United Arab",     5,   -249,   -156,    381 },    // 69
         { "Naparima BWI",          14,     -2,    374,    172 },    // 70
         { "Observatorio 1966",     14,   -425,   -169,     81 },    // 71
         { "Old Egyptian",          12,   -130,    110,    -13 },    // 72
         { "Old Hawaiian",           4,     61,   -285,   -181 },    // 73
         { "Oman",                   5,   -346,     -1,    224 },    // 74
         { "Ord Srvy Grt Britn",     0,    375,   -111,    431 },    // 75
         { "Pico De Las Nieves",    14,   -307,    -92,    127 },    // 76
         { "Pitcairn Astro 1967",   14,    185,    165,     42 },    // 77
         { "Prov So Amrican  56",   14,   -288,    175,   -376 },    // 78
         { "Prov So Chilean  63",   14,     16,    196,     93 },    // 79
         { "Puerto Rico",            4,     11,     72,   -101 },    // 80
         { "Qatar National",        14,   -128,   -283,     22 },    // 81
         { "Qornoq",                14,    164,    138,   -189 },    // 82
         { "Reunion",               14,     94,   -948,  -1262 },    // 83
         { "Rome 1940",             14,   -225,    -65,      9 },    // 84
         { "RT 90",                  3,    498,    -36,    568 },    // 85
         { "Santo (DOS)",           14,    170,     42,     84 },    // 86
         { "Sao Braz",              14,   -203,    141,     53 },    // 87
         { "Sapper Hill 1943",      14,   -355,     16,     74 },    // 88
         { "Schwarzeck",            21,    616,     97,   -251 },    // 89
         { "South American  69",    16,    -57,      1,    -41 },    // 90
         { "South Asia",             8,      7,    -10,    -26 },    // 91
         { "Southeast Base",        14,   -499,   -249,    314 },    // 92
         { "Southwest Base",        14,   -104,    167,    -38 },    // 93
         { "Timbalai 1948",          6,   -689,    691,    -46 },    // 94
         { "Tokyo",                  3,   -128,    481,    664 },    // 95
         { "Tristan Astro 1968",    14,   -632,    438,   -609 },    // 96
         { "Viti Levu 1916",         5,     51,    391,    -36 },    // 97
         { "Wake-Eniwetok  60",     13,    101,     52,    -39 },    // 98
         { "WGS72",                 19,      0,      0,      5 },    // 99
         { "WGS84",                 20,      0,      0,      0 },    // 100
         { "Zanderij",              14,   -265,    120,   -358 }     // 101
   };
   
   struct ELLIPSOID const gEllipsoid[] = {
   //      name                               a        1/f
         {  "Airy 1830",                  6377563.396, 299.3249646   },    // 0
         {  "Modified Airy",              6377340.189, 299.3249646   },    // 1
         {  "Australian National",        6378160.0,   298.25        },    // 2
         {  "Bessel 1841",                6377397.155, 299.1528128   },    // 3
         {  "Clarke 1866",                6378206.4,   294.9786982   },    // 4
         {  "Clarke 1880",                6378249.145, 293.465       },    // 5
         {  "Everest (India 1830)",       6377276.345, 300.8017      },    // 6
         {  "Everest (1948)",             6377304.063, 300.8017      },    // 7
         {  "Modified Fischer 1960",      6378155.0,   298.3         },    // 8
         {  "Everest (Pakistan)",         6377309.613, 300.8017      },    // 9
         {  "Indonesian 1974",            6378160.0,   298.247       },    // 10
         {  "GRS 80",                     6378137.0,   298.257222101 },    // 11
         {  "Helmert 1906",               6378200.0,   298.3         },    // 12
         {  "Hough 1960",                 6378270.0,   297.0         },    // 13
         {  "International 1924",         6378388.0,   297.0         },    // 14
         {  "Krassovsky 1940",            6378245.0,   298.3         },    // 15
         {  "South American 1969",        6378160.0,   298.25        },    // 16
         {  "Everest (Malaysia 1969)",    6377295.664, 300.8017      },    // 17
         {  "Everest (Sabah Sarawak)",    6377298.556, 300.8017      },    // 18
         {  "WGS 72",                     6378135.0,   298.26        },    // 19
         {  "WGS 84",                     6378137.0,   298.257223563 },    // 20
         {  "Bessel 1841 (Namibia)",      6377483.865, 299.1528128   },    // 21
         {  "Everest (India 1956)",       6377301.243, 300.8017      }     // 22
   };
   

/* bsb_ntohl - portable ntohl */
static uint32_t bsb_ntohl(uint32_t netlong)
{
	static const uint32_t testvalue = 0x12345678;
	uint8_t *p = (uint8_t *)&testvalue;

	if (p[0] == 0x12 && p[3] == 0x78)		// big endian?
		return netlong;

	// for little endian swap bytes
	return	( (netlong & 0xff000000) >> 24
			| (netlong & 0x00ff0000) >> 8
			| (netlong & 0x0000ff00) << 8
			| (netlong & 0x000000ff) << 24 );
}

/* Copies the next newline-delimited line from *pp into line as a NUL		*/
/* terminated string and increments	the pp pointer to point to the start	*/
/* of the next line.														*/
/* Removes \r characters from the line (if any).							*/
static int next_line(char **pp, int len, char *line)
{
	char *p = *pp;
	char *q = line;

	while (*p != '\0')
	{
		/* Don't overflow destination buffer */
		if (q - line > len)
			return 0;		/* line didn't fit in buffer */

		if (*p == '\r')
		{
			p++;
			continue;
		}
		if (*p == '\n')
		{
			*q = '\0';
			p++;
			*pp = p;
			return 1;
		}
		*q++ = *p++;
	}
	return 0;				/* didn't find \n line terminator */
}

int bsb_get_header_size(FILE *fp)
{
	int		text_size = 0, c;

	/* scan for end-of-text marker and record size of text section */
	while ( (c = fgetc(fp)) != -1 )
	{
		if (c == 0x1a)		/* Control-Z */
			break;
		text_size++;
	}
	return text_size;
}

/* return 0 on failure */
int bsb_open_header(char *filename, BSBImage *p)
{
	int			text_size = 0, c, depth;
	char		*p_ext, *pt, *text_buf, line[1024];

	p->pFile = NULL;

	/* Look for an extension (if any) to test for '.NO1' files */
	/* which must be treated specially since they are obfuscated */
	if (	(p_ext = strrchr(filename, '.')) != NULL &&
			(p_ext > strrchr(filename, DIR_SEPARATOR)) &&
			(strcasecmp(p_ext, ".NO1") == 0) )
	{
		FILE *inputFile;

		if (! (inputFile = fopen(filename, "rb")))
		{
			perror(filename);
			return 0;
		}

		/* Open temporary file to store unobfuscated file */
		if (! (p->pFile = tmpfile()))
		{
			perror("tmpfile()");
			return 0;
		}

		/* .NO1 files are obfuscated using ROT-9 */
		while ((c = fgetc(inputFile)) != EOF)
		{
			int r = (c - 9) & 0xFF;
			fputc(r, p->pFile);
		}
		fflush(p->pFile);
		fseek(p->pFile, 0, SEEK_SET);
	}
	else
	{
		/* Normal unobfuscated BSB/NOS files can be opened straight away */
		if (! (p->pFile = fopen(filename, "rb")))
		{
			perror(filename);
			return 0;
		}
	}

	if ((text_size = bsb_get_header_size(p->pFile)) == 0)
		return 0;

	/* allocate space & read in the entire text header */
	text_buf = (char *)malloc(text_size + 1);
	if (text_buf == NULL)
	{
		fprintf(stderr,
				"malloc(%d) failed for text header - BSB file possibly corrupt",
				text_size + 1);
		return 0;
	}

	fseek(p->pFile, 0, SEEK_SET);
	if (fread(text_buf, text_size, 1, p->pFile) != 1)
		return 0;
	text_buf[text_size] = '\0';

	pt = text_buf;
	p->num_colors = 0;
	p->version = -1.0;
	p->width = -1;
	p->height = -1;
	while ( next_line(&pt, sizeof(line), line) )
	{
		char		*s;
		int			index, r, g, b, ifm_depth;
		float		ver;

		if (sscanf(line, "RGB/%d,%d,%d,%d", &index, &r, &g, &b) == 4)
		{
			if ((unsigned)index < sizeof(p->red))
			{
				if (index > 0)
				{
					p->red[index-1] = r;
					p->green[index-1] = g;
					p->blue[index-1] = b;
					p->num_colors++;
				}
			}
		}
		if ( (s = strstr(line, "RA=")) )
		{
			int x0, y0;

			/* Attempt to read old-style NOS (4 parameter) version of RA= */
			/* then fall back to newer 2-argument version */
			if ((sscanf(s,"RA=%d,%d,%d,%d",&x0,&y0,&p->width,&p->height)!=4) &&
				(sscanf(s,"RA=%d,%d", &p->width, &p->height) != 2))
			{
				fprintf(stderr, "failed to read width,height from RA=\n");
				return 0;
			}
		}
		if ( (s = strstr(line, "DX=")) )
		{
			if ( sscanf(s, "DX=%lf", &p->xresolution) != 1 )
			{
				fprintf(stderr, "failed to read xresolution\n");
				return 0;
			}
		}
		if ( (s = strstr(line, "DY=")) )
		{
			if ( sscanf(s, "DY=%lf", &p->yresolution) != 1 )
			{
				fprintf(stderr, "failed to read xresolution\n");
				return 0;
			}
		}
		if (sscanf(line, "IFM/%d", &ifm_depth) == 1)
		{
			p->depth = ifm_depth;
		}
		if (sscanf(line, "VER/%f", &ver) == 1)
		{
			p->version = ver;
		}
	}
	if (p->width == -1 || p->height == -1)
	{
		fprintf(stderr, "Error: Could not read RA=<width>,<height>\n");
		return 0;
	}

	/* Attempt to read depth from binary section */
	c = fgetc(p->pFile);		/* discard <Control-Z> */
	c = fgetc(p->pFile);		/* discard NUL */

	/* Test depth from bitstream */
	depth = fgetc(p->pFile);
	if (depth != p->depth)
		fprintf(stderr, "Warning: depth from IFM tag (%d) != depth from bitstream (%d)\n", p->depth, depth);

	free(text_buf);
	return 1;
}

/* bsb_seek_to_row(BSBImage *p, int row)
 *
 * p	- pointer to a BSBImage
 * row	- row to seek to starting from row 0 (BSB row 1)
 *
 * Returns 1 on success and 0 on error
 *
 * Uses the index table at the end of the BSB file to quickly jump to a row.
 */
int bsb_seek_to_row(BSBImage *p, int row)
{
	int st, start_of_index, start_of_rows;

	/* Read start-of-index offset */
	if (fseek(p->pFile, -4, SEEK_END) == -1)
		return 0;
	if (fread(&st, 4, 1, p->pFile) != 1)
		return 0;
	start_of_index = bsb_ntohl(st);

	/* Read start-of-rows offset */
	if (fseek(p->pFile, row*4 + start_of_index, SEEK_SET) == -1)
		return 0;
	if (fread(&st, 4, 1, p->pFile) != 1)
		return 0;
	start_of_rows = bsb_ntohl(st);

	/* seek to row offset */
	if (fseek(p->pFile, start_of_rows, SEEK_SET) == -1)
		return 0;
	return 1;
}

/* Table used for computing multiplier in bsb_read_row() */
static char mul_mask[8] = { 0, 63, 31, 15, 7, 3, 1, 0 };

/* bsb_read_row(BSBImage *p, int row, uint8_t *buf)
 *
 * p	- pointer to a BSBImage containing file pointer at the start of a row
 *			this occurs after bsb_open_header() or bsb_seek_to_row()
 * buf	- output buffer for uncompressed pixel data
 *
 * Returns 1 on success and 0 on error
 */
int bsb_read_row(BSBImage *p, uint8_t *buf)
{
	int		c, i, multiplier, row_num = 0;
	int		pixel = 1, written = 0;

	/* The row number is stored in the low 7 bits of each byte.				*/
	/* The 8th bit indicates if row number is continued in the next byte.	*/
	do {
		c = fgetc(p->pFile);
		row_num = ((row_num & 0x7f) << 7) + c;
	} while (c >= 0x80);

	/* Rows are terminated by '\0'.  Note that rows can contain a '\0'	*/
	/* as part of the run-length data, so '\0' does not delimit rows.	*/
	/* (This occurs when multiplier is a multiple of 128 - 1)			*/
	while ((c = fgetc(p->pFile)) != '\0')
	{
		if (c == EOF)
		{
			fprintf(stderr, "Warning: EOF reading row %d\n", row_num);
			return 0;
		}

		pixel = (c & 0x7f) >> (7 - p->depth);
		multiplier = c & mul_mask[(int)p->depth];

		while (c >= 0x80)
		{
			c = fgetc(p->pFile);
			multiplier = (multiplier << 7) + (c & 0x7f);
		}
		multiplier++;

		if (multiplier > p->width)		/* limit impact of corrupt BSB data */
			multiplier = p->width;

		for (i = 0; i < multiplier; i++)
		{
			/* For the lower depths, the "grain" of the multiplyer is		*/
			/* course, so don't write past the width of the buffer.			*/
			if (written < p->width)
				buf[written++] = pixel - 1;		/* BSB color idx starts at 1 */
		}
	}

	if (written < p->width)
	{
		int short_fall = p->width - written;

		/* It seems valid BSB rows sometimes don't include pixel data for	*/
		/* the very last pixel or two.  Perhaps the decoder is supposed		*/
		/* to merely repeat the last pixel until the width is reached.		*/
		/* A value of 8 was chosen as a guess since the intended behaviour	*/
		/* of a BSB reader in this situation is not known.					*/
		if (short_fall < 8)
		{
			/* Repeat the last pixel value for small short falls */
			while (written < p->width)
				buf[written++] = pixel - 1;
		}
		else
		{
			fprintf(stderr, "Warning: Short row for row %d written=%d width=%d\n", row_num, written, p->width);
			return 0;
		}
	}
	return 1;
}

int bsb_write_index(FILE *fp, int height, int index[])
{
	int j;

	/* Write index table */
	for (j = 0; j < height + 1; j++)
	{
		/* Indices must be written as big-endian */
		if ((fputc(index[j] >> 24, fp)) == EOF) return 0;
		if ((fputc((index[j] & 0x00ff0000) >> 16, fp)) == EOF) return 0;
		if ((fputc((index[j] & 0x0000ff00) >> 8, fp)) == EOF) return 0;
		if ((fputc(index[j] & 0x000000ff, fp)) == EOF) return 0;
	}
	return 1;
}

/* Convert an integer into the variable length encoded form used for row
 * numbers in the BSB bitstream.
 */
static int bsb_store_integer(int row, uint8_t *p)
{
	int		req_bytes, i, c;

	/* Using 32 bit integers, row numbers are stored in one to four bytes.	*/
	/* This limits row numbers to 2^28 - 1 (= 268,435,455).					*/
	/* Calculate required bytes to store integer using a table instead of	*/
	/* expensive calls to get log base 2.									*/

	req_bytes = 1;
	if (row > 2097151)						/* 2^(7*3) - 1 */
		req_bytes = 4;
		else if (row > 16383)				/* 2^(7*2) - 1 */
			req_bytes = 3;
			else if (row > 127)				/* 2^(7*1) - 1 */
				req_bytes = 2;

	for (i = req_bytes - 1; i >= 0; i--)
	{
		c = (row >> i * 7) & 0x7f;

		if (i > 0)
			c |= 0x80;						/* set sentinel high bit */

		*p++ = c;
	}
	return req_bytes;
}

/* bsb_compress_row(BSBImage *p, int row, uint8_t *aPixel, uint8_t *buf)
 *
 * p		- pointer to a BSBImage for the width & depth values
 * row		- row number stating at 0 (row 0 will be stored as BSB row 1)
 * aPixel	- array of uncompressed pixels
 * buf		- output buffer for compressed bitstream
 *
 * Returns length of encoded bitstream
 *
 * Use a simple run length encoding, storing the run length in the
 * sentinel bits where possible.  Lower depth images have more room
 * to store this information.
 */
int bsb_compress_row(BSBImage *p, int row, const uint8_t *aPixel, uint8_t *buf)
{
	int		width = p->width, depth = p->depth;
	int		run_length, sentinel_bits, max_sentinel_rep, sentinel_mask;
	int		ibuf, ipixel;
	uint8_t	last_pix, shifted_pix;

	/* Write the row number (add 1 since BSB rows number from 1 not 0) */
	ibuf = 0;
	ibuf = bsb_store_integer(row+1, buf);


	sentinel_bits = 7 - depth;		/* number of sentinel bits available */
	sentinel_mask = ((1 << depth) - 1) << (7 - depth);
	max_sentinel_rep = 1 << sentinel_bits;	/* max run length that fits	*/
											/* in sentinel bits			*/

	ipixel = 0;
	while ( ipixel < width )
	{
		last_pix = aPixel[ipixel];
		ipixel++;

		/* Count length of pixel 'run' - run length cannot be greater than width */
		run_length = 0;
		while ( (ipixel < width) && (aPixel[ipixel] == last_pix) )
		{
			ipixel++;
			run_length++;
		}
		ipixel--;

		/* BSB colormap never uses index 0, so add 1 to pixel index before use */
		shifted_pix = (last_pix + 1) << sentinel_bits;

		if ( run_length < max_sentinel_rep )
		{
			if ( run_length == 0 )
			{
				if ( (shifted_pix & 0xff) == 0 )
					shifted_pix = sentinel_mask;
			}
			buf[ibuf++] = shifted_pix | run_length;
		}
		else
		{
			int		i, rc_bytes, rc_bits;
			int		rc = run_length / 2;

			rc_bits = 1;
			while (rc > 0)
			{
				rc = rc / 2;
				rc_bits++;
			}

			rc_bytes = (int)(rc_bits / 7);
			if ( rc_bits - rc_bytes * 7 > sentinel_bits )
				rc_bytes++;

			for ( i = rc_bytes; i > 0; i-- )
			{
				buf[ibuf + i] = run_length & 0x7f;
				run_length = run_length >> 7;
			}

			buf[ibuf] = shifted_pix | run_length;

			for ( i = 0; i < rc_bytes; i++ )
				buf[ibuf + i] |= 0x80;

			ibuf++;
			ibuf = ibuf + rc_bytes;
		}
		ipixel++;
	}
	buf[ibuf] = 0;						/* terminate row with zero */
	return ibuf + 1;
}

int bsb_close(BSBImage *p)
{
	fclose(p->pFile);

	return 1;
}

int read_8bpp_tiff_line(TIFF* tif, uint16_t bits_per_sample, int width, uint8_t *tiff_row, uint32_t row)
{
	int i, result;

	result = TIFFReadScanline(tif, tiff_row, row, 0);

	/* Copy the TIFF row expanding to 8 bits per pixel if necessary */
	switch (bits_per_sample)
	{
        case 8:		/* no modification needed */
        {
			break;
		}
        case 4:
		{
			/* Expand 4 bits-per-sample into 8 bits-per-sample		*/

			unsigned char	*p = tiff_row;
			uint8_t			a, b;

			if ((width % 2) != 0)			// Handle boundary condition
			{
				i = (width - 1) / 2;
				a = (p[i] >> 4) & 0xf;		// left nibble
				p[ width - 1 ] = a;
			}

			/* Start at the end of the compressed row and expand each pair */
			for (i = (width - 2) / 2; i >= 0; i--)
			{
				a = (p[i] >> 4) & 0xf;		// left nibble
				b = p[i] & 0xf;				// right nibble
				p[i*2] = a;
				p[i*2+1] = b;
			}
			break;
		}
	}
	return result;
}

/* Count unique colors in the TIFF color map.							*/
int count_tiff_colors(TIFF* tif, BSBImage *pImage, uint16_t bits_per_sample)
{
	int			i, j, max_colors = 0;
	uint64_t	histogram[256] = {0};
	uint8_t		*tiff_row;

	tiff_row = (uint8_t *)malloc(pImage->width);

	/* Scan through entire image computing a histogram of colormap indices	*/
	for (j = 0; j < pImage->height; j++)
	{
		uint8_t *p = tiff_row;

		read_8bpp_tiff_line(tif, bits_per_sample, pImage->width, tiff_row, j);

		for (i = 0; i < pImage->width; i++)
			histogram[*p++]++;
	}
	/* Count down from highest colormap index for the first non-zero value.	*/
	/* This will indicate the maximum number of colors.						*/
	for (i = 255; i >= 0; i--)
	{
		if (histogram[i] != 0)
		{
			max_colors = i + 1;		/* count of colors includes index 0 */
			break;
		}
	}
	free(tiff_row);

	return max_colors;
}

static int depth_for_colors[] = { 1, 2, 4, 8, 16, 32, 64, 128 };


/* PH99
BOOL REXXENTRY DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
    if ( fdwReason == DLL_PROCESS_ATTACH )
    {
        //
    }
    else if ( fdwReason == DLL_PROCESS_DETACH )
    {
        //
    }
    return(TRUE);
}

*/

#ifdef __cplusplus
}
#endif

double my_fit_function( double tx, double ty, int n_par, double* p )
{

    double ret = p[0] + p[1]*tx;

    if(n_par > 2)
          ret += p[2]*ty;
    if(n_par > 3)
    {
        ret += p[3]*tx*tx;
        ret += p[4]*tx*ty;
        ret += p[5]*ty*ty;
    }
    if(n_par > 6)
    {
        ret += p[6]*tx*tx*tx;
        ret += p[7]*tx*tx*ty;
        ret += p[8]*tx*ty*ty;
        ret += p[9]*ty*ty*ty;
    }

    return ret;
}


int Georef_Calculate_Coefficients_Onedir(int n_points, int n_par, double *tx, double *ty, double *y, double *p,
                                        double hintp0, double hintp1, double hintp2)
        /*
        n_points : number of sample points
        n_par :  3, 6, or 10,  6 is probably good
        tx:  sample data independent variable 1
        ty:  sample data independent variable 2
        y:   sample data dependent result
        p:   curve fit result coefficients
        */
{

    int i;
    lm_control_type control;
    lm_data_type data;

    lm_initialize_control( &control );


    for(i=0 ; i<12 ; i++)
        p[i] = 0.;

//    memset(p, 0, 12 * sizeof(double));

    //  Insert hints
    p[0] = hintp0;
    p[1] = hintp1;
    p[2] = hintp2;

    data.user_func = my_fit_function;
    data.user_tx = tx;
    data.user_ty = ty;
    data.user_y = y;
    data.n_par = n_par;
    data.print_flag = 0;

// perform the fit:

            lm_minimize( n_points, n_par, p, lm_evaluate_default, lm_print_default,
                         &data, &control );

// print results:
//            printf( "status: %s after %d evaluations\n",
//                    lm_infmsg[control.info], control.nfev );

            //      Control.info results [1,2,3] are success, other failure
            return control.info;
}


/*
 * This file contains default implementation of the evaluate and printout
 * routines. In most cases, customization of lmfit can be done by modifying
 * these two routines. Either modify them here, or copy and rename them.
 */

void lm_evaluate_default( double* par, int m_dat, double* fvec,
                          void *data, int *info )
/*
        *    par is an input array. At the end of the minimization, it contains
        *        the approximate solution vector.
        *
        *    m_dat is a positive integer input variable set to the number
        *      of functions.
        *
        *    fvec is an output array of length m_dat which contains the function
        *        values the square sum of which ought to be minimized.
        *
        *    data is a read-only pointer to lm_data_type, as specified by lmuse.h.
        *
        *      info is an integer output variable. If set to a negative value, the
        *        minimization procedure will stop.
 */
{
    int i;
    lm_data_type *mydata;
    mydata = (lm_data_type*)data;

    for (i=0; i<m_dat; i++)
    {
        fvec[i] = mydata->user_y[i] - mydata->user_func( mydata->user_tx[i], mydata->user_ty[i], mydata->n_par, par);
    }
    *info = *info; /* to prevent a 'unused variable' warning */
    /* if <parameters drifted away> { *info = -1; } */
}

void lm_print_default( int n_par, double* par, int m_dat, double* fvec,
                       void *data, int iflag, int iter, int nfev )
/*
        *       data  : for soft control of printout behaviour, add control
        *                 variables to the data struct
        *       iflag : 0 (init) 1 (outer loop) 2(inner loop) -1(terminated)
        *       iter  : outer loop counter
        *       nfev  : number of calls to *evaluate
 */
{
    double f, y, tx, ty;
    int i;
    lm_data_type *mydata;
    mydata = (lm_data_type*)data;

    if(mydata->print_flag)
    {
        if (iflag==2) {
            printf ("trying step in gradient direction\n");
        } else if (iflag==1) {
            printf ("determining gradient (iteration %d)\n", iter);
        } else if (iflag==0) {
            printf ("starting minimization\n");
        } else if (iflag==-1) {
            printf ("terminated after %d evaluations\n", nfev);
        }

        printf( "  par: " );
        for( i=0; i<n_par; ++i )
            printf( " %12g", par[i] );
        printf ( " => norm: %12g\n", lm_enorm( m_dat, fvec ) );

        if ( iflag == -1 ) {
            printf( "  fitting data as follows:\n" );
            for( i=0; i<m_dat; ++i ) {
                tx = (mydata->user_tx)[i];
                ty = (mydata->user_ty)[i];
                y = (mydata->user_y)[i];
                f = mydata->user_func( tx, ty, mydata->n_par, par );
                printf( "    tx[%2d]=%8g     ty[%2d]=%8g     y=%12g fit=%12g     residue=%12g\n",
                        i, tx, i, ty, y, f, y-f );
            }
        }
    }       // if print_flag
}





///=================================================================================

/* *********************** high-level interface **************************** */


void lm_initialize_control( lm_control_type *control )
{
    control->maxcall = 100;
    control->epsilon = 1.e-10; //1.e-14;
    control->stepbound = 100; //100.;
    control->ftol = 1.e-14;
    control->xtol = 1.e-14;
    control->gtol = 1.e-14;
}

void lm_minimize( int m_dat, int n_par, double* par,
                  lm_evaluate_ftype *evaluate, lm_print_ftype *printout,
                  void *data, lm_control_type *control )
{
    std::vector<double> fvec(m_dat);
    std::vector<double> diag(n_par);
    std::vector<double> qtf(n_par);
    std::vector<double> fjac(n_par*m_dat);
    std::vector<double> wa1(n_par);
    std::vector<double> wa2(n_par);
    std::vector<double> wa3(n_par);
    std::vector<double> wa4(m_dat);
    std::vector<int> ipvt(n_par);

    // *** perform fit.

          control->info = 0;
          control->nfev = 0;

    // this goes through the modified legacy interface:
          lm_lmdif( m_dat, n_par, par, &fvec[0], control->ftol, control->xtol, control->gtol,
                    control->maxcall*(n_par+1), control->epsilon, &diag[0], 1,
                    control->stepbound, &(control->info),
                    &(control->nfev), &fjac[0], &ipvt[0], &qtf[0], &wa1[0], &wa2[0], &wa3[0], &wa4[0],
                    evaluate, printout, data );

          (*printout)( n_par, par, m_dat, &fvec[0], data, -1, 0, control->nfev );
          control->fnorm = lm_enorm(m_dat, &fvec[0]);
          if (control->info < 0 ) control->info = 10;
}


// ***** the following messages are referenced by the variable info.

const char *lm_infmsg[] = {
    "improper input parameters",
    "the relative error in the sum of squares is at most tol",
    "the relative error between x and the solution is at most tol",
    "both errors are at most tol",
    "fvec is orthogonal to the columns of the jacobian to machine precision",
    "number of calls to fcn has reached or exceeded 200*(n+1)",
    "ftol is too small. no further reduction in the sum of squares is possible",
    "xtol too small. no further improvement in approximate solution x possible",
    "gtol too small. no further improvement in approximate solution x possible",
    "not enough memory",
    "break requested within function evaluation"
};

const char *lm_shortmsg[] = {
    "invalid input",
    "success (f)",
    "success (p)",
    "success (f,p)",
    "degenerate",
    "call limit",
    "failed (f)",
    "failed (p)",
    "failed (o)",
    "no memory",
    "user break"
};


/* *********************** high-level interface **************************** */

/* ************************** implementation ******************************* */
#define BUG 0
#if BUG
#include <stdio.h>
#endif

// the following values seem good for an x86:
//#define LM_MACHEP .555e-16 /* resolution of arithmetic */
//#define LM_DWARF  9.9e-324 /* smallest nonzero number */
// the follwoing values should work on any machine:
 #define LM_MACHEP 1.2e-16
 #define LM_DWARF 1.0e-38

// the squares of the following constants shall not under/overflow:
// these values seem good for an x86:
//#define LM_SQRT_DWARF 1.e-160
//#define LM_SQRT_GIANT 1.e150
// the following values should work on any machine:
 #define LM_SQRT_DWARF 3.834e-20
 #define LM_SQRT_GIANT 1.304e19


void lm_qrfac( int m, int n, double* a, int pivot, int* ipvt,
               double* rdiag, double* acnorm, double* wa);
void lm_qrsolv(int n, double* r, int ldr, int* ipvt, double* diag,
               double* qtb, double* x, double* sdiag, double* wa);
void lm_lmpar( int n, double* r, int ldr, int* ipvt, double* diag, double* qtb,
               double delta, double* par, double* x, double* sdiag,
               double* wa1, double* wa2);

#define MIN(a,b) (((a)<=(b)) ? (a) : (b))
#define MAX(a,b) (((a)>=(b)) ? (a) : (b))
#define SQR(x)   (x)*(x)

// ***** the low-level legacy interface for full control.

void lm_lmdif( int m, int n, double* x, double* fvec, double ftol, double xtol,
               double gtol, int maxfev, double epsfcn, double* diag, int mode,
               double factor, int *info, int *nfev,
               double* fjac, int* ipvt, double* qtf,
               double* wa1, double* wa2, double* wa3, double* wa4,
               lm_evaluate_ftype *evaluate, lm_print_ftype *printout,
               void *data )
{
/*
    *   the purpose of lmdif is to minimize the sum of the squares of
    *   m nonlinear functions in n variables by a modification of
    *   the levenberg-marquardt algorithm. the user must provide a
    *   subroutine evaluate which calculates the functions. the jacobian
    *   is then calculated by a forward-difference approximation.
    *
    *   the multi-parameter interface lm_lmdif is for users who want
    *   full control and flexibility. most users will be better off using
    *   the simpler interface lmfit provided above.
    *
    *   the parameters are the same as in the legacy FORTRAN implementation,
 *   with the following exceptions:
    *      the old parameter ldfjac which gave leading dimension of fjac has
    *        been deleted because this C translation makes no use of two-
    *        dimensional arrays;
    *      the old parameter nprint has been deleted; printout is now controlled
    *        by the user-supplied routine *printout;
    *      the parameter field *data and the function parameters *evaluate and
    *        *printout have been added; they help avoiding global variables.
    *
 *   parameters:
    *
    *    m is a positive integer input variable set to the number
    *      of functions.
    *
    *    n is a positive integer input variable set to the number
    *      of variables. n must not exceed m.
    *
    *    x is an array of length n. on input x must contain
    *      an initial estimate of the solution vector. on output x
    *      contains the final estimate of the solution vector.
    *
    *    fvec is an output array of length m which contains
    *      the functions evaluated at the output x.
    *
    *    ftol is a nonnegative input variable. termination
    *      occurs when both the actual and predicted relative
    *      reductions in the sum of squares are at most ftol.
    *      therefore, ftol measures the relative error desired
    *      in the sum of squares.
    *
    *    xtol is a nonnegative input variable. termination
    *      occurs when the relative error between two consecutive
    *      iterates is at most xtol. therefore, xtol measures the
    *      relative error desired in the approximate solution.
    *
    *    gtol is a nonnegative input variable. termination
    *      occurs when the cosine of the angle between fvec and
    *      any column of the jacobian is at most gtol in absolute
    *      value. therefore, gtol measures the orthogonality
    *      desired between the function vector and the columns
    *      of the jacobian.
    *
    *    maxfev is a positive integer input variable. termination
    *      occurs when the number of calls to lm_fcn is at least
    *      maxfev by the end of an iteration.
    *
    *    epsfcn is an input variable used in determining a suitable
    *      step length for the forward-difference approximation. this
    *      approximation assumes that the relative errors in the
    *      functions are of the order of epsfcn. if epsfcn is less
    *      than the machine precision, it is assumed that the relative
    *      errors in the functions are of the order of the machine
    *      precision.
    *
    *    diag is an array of length n. if mode = 1 (see below), diag is
    *        internally set. if mode = 2, diag must contain positive entries
    *        that serve as multiplicative scale factors for the variables.
    *
    *    mode is an integer input variable. if mode = 1, the
    *      variables will be scaled internally. if mode = 2,
    *      the scaling is specified by the input diag. other
    *      values of mode are equivalent to mode = 1.
    *
    *    factor is a positive input variable used in determining the
    *      initial step bound. this bound is set to the product of
    *      factor and the euclidean norm of diag*x if nonzero, or else
    *      to factor itself. in most cases factor should lie in the
    *      interval (.1,100.). 100. is a generally recommended value.
    *
    *    info is an integer output variable that indicates the termination
 *        status of lm_lmdif as follows:
    *
    *        info < 0  termination requested by user-supplied routine *evaluate;
    *
    *      info = 0  improper input parameters;
    *
    *      info = 1  both actual and predicted relative reductions
    *              in the sum of squares are at most ftol;
    *
    *      info = 2  relative error between two consecutive iterates
    *              is at most xtol;
    *
    *      info = 3  conditions for info = 1 and info = 2 both hold;
    *
    *      info = 4  the cosine of the angle between fvec and any
    *              column of the jacobian is at most gtol in
    *              absolute value;
    *
    *      info = 5  number of calls to lm_fcn has reached or
    *              exceeded maxfev;
    *
    *      info = 6  ftol is too small. no further reduction in
    *              the sum of squares is possible;
    *
    *      info = 7  xtol is too small. no further improvement in
    *              the approximate solution x is possible;
    *
    *      info = 8  gtol is too small. fvec is orthogonal to the
    *              columns of the jacobian to machine precision;
    *
    *    nfev is an output variable set to the number of calls to the
    *        user-supplied routine *evaluate.
    *
    *    fjac is an output m by n array. the upper n by n submatrix
    *      of fjac contains an upper triangular matrix r with
    *      diagonal elements of nonincreasing magnitude such that
    *
    *           t     t       t
    *          p *(jac *jac)*p = r *r,
    *
    *      where p is a permutation matrix and jac is the final
    *      calculated jacobian. column j of p is column ipvt(j)
    *      (see below) of the identity matrix. the lower trapezoidal
    *      part of fjac contains information generated during
    *      the computation of r.
    *
    *    ipvt is an integer output array of length n. ipvt
    *      defines a permutation matrix p such that jac*p = q*r,
    *      where jac is the final calculated jacobian, q is
    *      orthogonal (not stored), and r is upper triangular
    *      with diagonal elements of nonincreasing magnitude.
    *      column j of p is column ipvt(j) of the identity matrix.
    *
    *    qtf is an output array of length n which contains
    *      the first n elements of the vector (q transpose)*fvec.
    *
    *    wa1, wa2, and wa3 are work arrays of length n.
    *
    *    wa4 is a work array of length m.
    *
 *   the following parameters are newly introduced in this C translation:
    *
    *      evaluate is the name of the subroutine which calculates the functions.
    *        a default implementation lm_evaluate_default is provided in lm_eval.c;
    *        alternatively, evaluate can be provided by a user calling program.
 *        it should be written as follows:
    *
    *        void evaluate ( double* par, int m_dat, double* fvec,
    *                       void *data, int *info )
    *        {
    *           // for ( i=0; i<m_dat; ++i )
    *           //     calculate fvec[i] for given parameters par;
    *           // to stop the minimization,
    *           //     set *info to a negative integer.
    *        }
    *
    *      printout is the name of the subroutine which nforms about fit progress.
    *        a default implementation lm_print_default is provided in lm_eval.c;
    *        alternatively, printout can be provided by a user calling program.
 *        it should be written as follows:
    *
    *        void printout ( int n_par, double* par, int m_dat, double* fvec,
    *                       void *data, int iflag, int iter, int nfev )
    *        {
    *           // iflag : 0 (init) 1 (outer loop) 2(inner loop) -1(terminated)
    *           // iter  : outer loop counter
    *           // nfev  : number of calls to *evaluate
    *        }
    *
    *      data is an input pointer to an arbitrary structure that is passed to
    *        evaluate. typically, it contains experimental data to be fitted.
    *
 */
    int i, iter, j;
    double actred, delta, dirder, eps, fnorm, fnorm1, gnorm, par, pnorm,
    prered, ratio, step, sum, temp, temp1, temp2, temp3, xnorm;
    static double p1 = 0.1;
    static double p5 = 0.5;
    static double p25 = 0.25;
    static double p75 = 0.75;
    static double p0001 = 1.0e-4;

    *nfev = 0; // function evaluation counter
    iter = 1;  // outer loop counter
    par = 0;   // levenberg-marquardt parameter
    delta = 0; // just to prevent a warning (initialization within if-clause)
    xnorm = 0; // dito

    temp = MAX(epsfcn,LM_MACHEP);
    eps = sqrt(temp); // used in calculating the Jacobian by forward differences

// *** check the input parameters for errors.

    if ( (n <= 0) || (m < n) || (ftol < 0.)
          || (xtol < 0.) || (gtol < 0.) || (maxfev <= 0) || (factor <= 0.) )
    {
        *info = 0; // invalid parameter
        return;
    }
    if ( mode == 2 )  /* scaling by diag[] */
    {
        for ( j=0; j<n; j++ )  /* check for nonpositive elements */
        {
            if ( diag[j] <= 0.0 )
            {
                *info = 0; // invalid parameter
                return;
            }
        }
    }
#if BUG
    printf( "lmdif\n" );
#endif

// *** evaluate the function at the starting point and calculate its norm.

    *info = 0;
    (*evaluate)( x, m, fvec, data, info );
    (*printout)( n, x, m, fvec, data, 0, 0, ++(*nfev) );
    if ( *info < 0 ) return;
    fnorm = lm_enorm(m,fvec);

// *** the outer loop.

    do {
#if BUG
        printf( "lmdif/ outer loop iter=%d nfev=%d fnorm=%.10e\n",
                iter, *nfev, fnorm );
#endif

// O** calculate the jacobian matrix.

        for ( j=0; j<n; j++ )
{
    temp = x[j];
    step = eps * fabs(temp);
    if (step == 0.) step = eps;
    x[j] = temp + step;
    *info = 0;
    (*evaluate)( x, m, wa4, data, info );
    (*printout)( n, x, m, wa4, data, 1, iter, ++(*nfev) );
    if ( *info < 0 ) return;  // user requested break
    x[j] = temp;
    for ( i=0; i<m; i++ )
        fjac[j*m+i] = (wa4[i] - fvec[i]) / step;
}
#if BUG>1
        // DEBUG: print the entire matrix
        for ( i=0; i<m; i++ )
{
    for ( j=0; j<n; j++ )
        printf( "%.5e ", y[j*m+i] );
    printf( "\n" );
}
#endif

// O** compute the qr factorization of the jacobian.

        lm_qrfac( m, n, fjac, 1, ipvt, wa1, wa2, wa3);

// O** on the first iteration ...

        if (iter == 1)
{
    if (mode != 2)
//      ... scale according to the norms of the columns of the initial jacobian.
    {
        for ( j=0; j<n; j++ )
        {
            diag[j] = wa2[j];
            if ( wa2[j] == 0. )
                diag[j] = 1.;
        }
    }

//      ... calculate the norm of the scaled x and
//          initialize the step bound delta.

    for ( j=0; j<n; j++ )
        wa3[j] = diag[j] * x[j];

    xnorm = lm_enorm( n, wa3 );
    delta = factor*xnorm;
    if (delta == 0.)
        delta = factor;
}

// O** form (q transpose)*fvec and store the first n components in qtf.

        for ( i=0; i<m; i++ )
            wa4[i] = fvec[i];

        for ( j=0; j<n; j++ )
{
    temp3 = fjac[j*m+j];
    if (temp3 != 0.)
    {
        sum = 0;
        for ( i=j; i<m; i++ )
            sum += fjac[j*m+i] * wa4[i];
        temp = -sum / temp3;
        for ( i=j; i<m; i++ )
            wa4[i] += fjac[j*m+i] * temp;
    }
    fjac[j*m+j] = wa1[j];
    qtf[j] = wa4[j];
}

// O** compute the norm of the scaled gradient and test for convergence.

        gnorm = 0;
        if ( fnorm != 0 )
{
    for ( j=0; j<n; j++ )
    {
        if ( wa2[ ipvt[j] ] == 0 ) continue;

        sum = 0.;
        for ( i=0; i<=j; i++ )
            sum += fjac[j*m+i] * qtf[i] / fnorm;
        gnorm = MAX( gnorm, fabs(sum/wa2[ ipvt[j] ]) );
    }
}

        if ( gnorm <= gtol )
{
    *info = 4;
    return;
}

// O** rescale if necessary.

        if ( mode != 2 )
{
    for ( j=0; j<n; j++ )
        diag[j] = MAX(diag[j],wa2[j]);
}

// O** the inner loop.

        do {
#if BUG
            printf( "lmdif/ inner loop iter=%d nfev=%d\n", iter, *nfev );
#endif

// OI* determine the levenberg-marquardt parameter.

            lm_lmpar( n,fjac,m,ipvt,diag,qtf,delta,&par,wa1,wa2,wa3,wa4 );

// OI* store the direction p and x + p. calculate the norm of p.

            for ( j=0; j<n; j++ )
            {
                wa1[j] = -wa1[j];
                wa2[j] = x[j] + wa1[j];
                wa3[j] = diag[j]*wa1[j];
            }
            pnorm = lm_enorm(n,wa3);

// OI* on the first iteration, adjust the initial step bound.

            if ( *nfev<= 1+n ) // bug corrected by J. Wuttke in 2004
                delta = MIN(delta,pnorm);

// OI* evaluate the function at x + p and calculate its norm.

            *info = 0;
            (*evaluate)( wa2, m, wa4, data, info );
            (*printout)( n, x, m, wa4, data, 2, iter, ++(*nfev) );
            if ( *info < 0 ) return;  // user requested break

            fnorm1 = lm_enorm(m,wa4);
#if BUG
            printf( "lmdif/ pnorm %.10e  fnorm1 %.10e  fnorm %.10e"
                    " delta=%.10e par=%.10e\n",
                    pnorm, fnorm1, fnorm, delta, par );
#endif

// OI* compute the scaled actual reduction.

            if ( p1*fnorm1 < fnorm )
                actred = 1 - SQR( fnorm1/fnorm );
            else
                actred = -1;

// OI* compute the scaled predicted reduction and
//     the scaled directional derivative.

            for ( j=0; j<n; j++ )
            {
                wa3[j] = 0;
                for ( i=0; i<=j; i++ )
                    wa3[i] += fjac[j*m+i]*wa1[ ipvt[j] ];
            }
            temp1 = lm_enorm(n,wa3) / fnorm;
            temp2 = sqrt(par) * pnorm / fnorm;
            prered = SQR(temp1) + 2 * SQR(temp2);
            dirder = - ( SQR(temp1) + SQR(temp2) );

// OI* compute the ratio of the actual to the predicted reduction.

            ratio = prered!=0 ? actred/prered : 0;
#if BUG
            printf( "lmdif/ actred=%.10e prered=%.10e ratio=%.10e"
                    " sq(1)=%.10e sq(2)=%.10e dd=%.10e\n",
                    actred, prered, prered!=0 ? ratio : 0.,
                    SQR(temp1), SQR(temp2), dirder );
#endif

// OI* update the step bound.

            if (ratio <= p25)
{
    if (actred >= 0.)
        temp = p5;
    else
        temp = p5*dirder/(dirder + p5*actred);
    if ( p1*fnorm1 >= fnorm || temp < p1 )
        temp = p1;
    delta = temp * MIN(delta,pnorm/p1);
    par /= temp;
}
            else if ( par == 0. || ratio >= p75 )
{
    delta = pnorm/p5;
    par *= p5;
}

// OI* test for successful iteration...

            if (ratio >= p0001)
{

//     ... successful iteration. update x, fvec, and their norms.

    for ( j=0; j<n; j++ )
    {
        x[j] = wa2[j];
        wa2[j] = diag[j]*x[j];
    }
    for ( i=0; i<m; i++ )
        fvec[i] = wa4[i];
    xnorm = lm_enorm(n,wa2);
    fnorm = fnorm1;
    iter++;
}
#if BUG
            else {
    printf( "ATTN: iteration considered unsuccessful\n" );
            }
#endif

// OI* tests for convergence ( otherwise *info = 1, 2, or 3 )

            *info = 0; // do not terminate (unless overwritten by nonzero value)
            if ( fabs(actred) <= ftol && prered <= ftol && p5*ratio <= 1 )
                *info = 1;
            if (delta <= xtol*xnorm)
                *info += 2;
            if ( *info != 0)
                return;

// OI* tests for termination and stringent tolerances.

            if ( *nfev >= maxfev)
                *info = 5;
            if ( fabs(actred) <= LM_MACHEP &&
                 prered <= LM_MACHEP && p5*ratio <= 1 )
                *info = 6;
            if (delta <= LM_MACHEP*xnorm)
                *info = 7;
            if (gnorm <= LM_MACHEP)
                *info = 8;
            if ( *info != 0)
                return;

// OI* end of the inner loop. repeat if iteration unsuccessful.

        } while (ratio < p0001);

// O** end of the outer loop.

    } while (1);

}



void lm_lmpar(int n, double* r, int ldr, int* ipvt, double* diag, double* qtb,
              double delta, double* par, double* x, double* sdiag,
              double* wa1, double* wa2)
{
/*     given an m by n matrix a, an n by n nonsingular diagonal
    *     matrix d, an m-vector b, and a positive number delta,
    *     the problem is to determine a value for the parameter
    *     par such that if x solves the system
    *
    *        a*x = b ,       sqrt(par)*d*x = 0 ,
    *
    *     in the least squares sense, and dxnorm is the euclidean
    *     norm of d*x, then either par is 0. and
    *
    *        (dxnorm-delta) .le. 0.1*delta ,
    *
    *     or par is positive and
    *
    *        abs(dxnorm-delta) .le. 0.1*delta .
    *
    *     this subroutine completes the solution of the problem
    *     if it is provided with the necessary information from the
    *     qr factorization, with column pivoting, of a. that is, if
    *     a*p = q*r, where p is a permutation matrix, q has orthogonal
    *     columns, and r is an upper triangular matrix with diagonal
    *     elements of nonincreasing magnitude, then lmpar expects
    *     the full upper triangle of r, the permutation matrix p,
    *     and the first n components of (q transpose)*b. on output
    *     lmpar also provides an upper triangular matrix s such that
    *
    *         t       t               t
    *        p *(a *a + par*d*d)*p = s *s .
    *
    *     s is employed within lmpar and may be of separate interest.
    *
    *     only a few iterations are generally needed for convergence
    *     of the algorithm. if, however, the limit of 10 iterations
    *     is reached, then the output par will contain the best
    *     value obtained so far.
    *
 *     parameters:
    *
    *    n is a positive integer input variable set to the order of r.
    *
    *    r is an n by n array. on input the full upper triangle
    *      must contain the full upper triangle of the matrix r.
    *      on output the full upper triangle is unaltered, and the
    *      strict lower triangle contains the strict upper triangle
    *      (transposed) of the upper triangular matrix s.
    *
    *    ldr is a positive integer input variable not less than n
    *      which specifies the leading dimension of the array r.
    *
    *    ipvt is an integer input array of length n which defines the
    *      permutation matrix p such that a*p = q*r. column j of p
    *      is column ipvt(j) of the identity matrix.
    *
    *    diag is an input array of length n which must contain the
    *      diagonal elements of the matrix d.
    *
    *    qtb is an input array of length n which must contain the first
    *      n elements of the vector (q transpose)*b.
    *
    *    delta is a positive input variable which specifies an upper
    *      bound on the euclidean norm of d*x.
    *
    *    par is a nonnegative variable. on input par contains an
    *      initial estimate of the levenberg-marquardt parameter.
    *      on output par contains the final estimate.
    *
    *    x is an output array of length n which contains the least
    *      squares solution of the system a*x = b, sqrt(par)*d*x = 0,
    *      for the output par.
    *
    *    sdiag is an output array of length n which contains the
    *      diagonal elements of the upper triangular matrix s.
    *
    *    wa1 and wa2 are work arrays of length n.
    *
 */
    int i, iter, j, nsing;
    double dxnorm, fp, fp_old, gnorm, parc, parl, paru;
    double sum, temp;
    static double p1 = 0.1;
    static double p001 = 0.001;

#if BUG
    printf( "lmpar\n" );
#endif

// *** compute and store in x the gauss-newton direction. if the
//     jacobian is rank-deficient, obtain a least squares solution.

    nsing = n;
    for ( j=0; j<n; j++ )
    {
        wa1[j] = qtb[j];
        if ( r[j*ldr+j] == 0 && nsing == n )
            nsing = j;
        if (nsing < n)
            wa1[j] = 0;
    }
#if BUG
    printf( "nsing %d ", nsing );
#endif
    for ( j=nsing-1; j>=0; j-- )
{
    wa1[j] = wa1[j]/r[j+ldr*j];
    temp = wa1[j];
    for ( i=0; i<j; i++ )
        wa1[i] -= r[j*ldr+i]*temp;
}

    for ( j=0; j<n; j++ )
        x[ ipvt[j] ] = wa1[j];

// *** initialize the iteration counter.
//     evaluate the function at the origin, and test
//     for acceptance of the gauss-newton direction.

    iter = 0;
    for ( j=0; j<n; j++ )
        wa2[j] = diag[j]*x[j];
    dxnorm = lm_enorm(n,wa2);
    fp = dxnorm - delta;
    if (fp <= p1*delta)
{
#if BUG
      printf( "lmpar/ terminate (fp<delta/10\n" );
#endif
        *par = 0;
        return;
}

// *** if the jacobian is not rank deficient, the newton
//     step provides a lower bound, parl, for the 0. of
//     the function. otherwise set this bound to 0..

    parl = 0;
    if (nsing >= n)
{
    for ( j=0; j<n; j++ )
        wa1[j] = diag[ ipvt[j] ] * wa2[ ipvt[j] ] / dxnorm;

    for ( j=0; j<n; j++ )
    {
        sum = 0.;
        for ( i=0; i<j; i++ )
            sum += r[j*ldr+i]*wa1[i];
        wa1[j] = (wa1[j] - sum)/r[j+ldr*j];
    }
    temp = lm_enorm(n,wa1);
    parl = fp/delta/temp/temp;
}

// *** calculate an upper bound, paru, for the 0. of the function.

    for ( j=0; j<n; j++ )
{
    sum = 0;
    for ( i=0; i<=j; i++ )
        sum += r[j*ldr+i]*qtb[i];
    wa1[j] = sum/diag[ ipvt[j] ];
}
    gnorm = lm_enorm(n,wa1);
    paru = gnorm/delta;
    if (paru == 0.)
        paru = LM_DWARF/MIN(delta,p1);

// *** if the input par lies outside of the interval (parl,paru),
//     set par to the closer endpoint.

    *par = MAX( *par,parl);
    *par = MIN( *par,paru);
    if ( *par == 0.)
        *par = gnorm/dxnorm;
#if BUG
    printf( "lmpar/ parl %.4e  par %.4e  paru %.4e\n", parl, *par, paru );
#endif

// *** iterate.

    for ( ; ; iter++ ) {

// *** evaluate the function at the current value of par.

    if ( *par == 0.)
        *par = MAX(LM_DWARF,p001*paru);
    temp = sqrt( *par );
    for ( j=0; j<n; j++ )
        wa1[j] = temp*diag[j];
    lm_qrsolv( n, r, ldr, ipvt, wa1, qtb, x, sdiag, wa2);
    for ( j=0; j<n; j++ )
        wa2[j] = diag[j]*x[j];
    dxnorm = lm_enorm(n,wa2);
    fp_old = fp;
    fp = dxnorm - delta;

// ***       if the function is small enough, accept the current value
//     of par. also test for the exceptional cases where parl
//     is 0. or the number of iterations has reached 10.

    if ( fabs(fp) <= p1*delta
         || (parl == 0. && fp <= fp_old && fp_old < 0.)
         || iter == 10 )
        break; // the only exit from this loop

// *** compute the Newton correction.

    for ( j=0; j<n; j++ )
        wa1[j] = diag[ ipvt[j] ] * wa2[ ipvt[j] ] / dxnorm;

    for ( j=0; j<n; j++ )
    {
        wa1[j] = wa1[j]/sdiag[j];
        for ( i=j+1; i<n; i++ )
            wa1[i] -= r[j*ldr+i]*wa1[j];
    }
    temp = lm_enorm( n, wa1);
    parc = fp/delta/temp/temp;

// *** depending on the sign of the function, update parl or paru.

    if (fp > 0)
        parl = MAX(parl, *par);
    else if (fp < 0)
        paru = MIN(paru, *par);
        // the case fp==0 is precluded by the break condition

// *** compute an improved estimate for par.

    *par = MAX(parl, *par + parc);

    }

}



void lm_qrfac(int m, int n, double* a, int pivot, int* ipvt,
              double* rdiag, double* acnorm, double* wa)
{
/*
    *     this subroutine uses householder transformations with column
    *     pivoting (optional) to compute a qr factorization of the
    *     m by n matrix a. that is, qrfac determines an orthogonal
    *     matrix q, a permutation matrix p, and an upper trapezoidal
    *     matrix r with diagonal elements of nonincreasing magnitude,
    *     such that a*p = q*r. the householder transformation for
    *     column k, k = 1,2,...,min(m,n), is of the form
    *
    *                    t
    *        i - (1/u(k))*u*u
    *
    *     where u has 0.s in the first k-1 positions. the form of
    *     this transformation and the method of pivoting first
    *     appeared in the corresponding linpack subroutine.
    *
 *     parameters:
    *
    *    m is a positive integer input variable set to the number
    *      of rows of a.
    *
    *    n is a positive integer input variable set to the number
    *      of columns of a.
    *
    *    a is an m by n array. on input a contains the matrix for
    *      which the qr factorization is to be computed. on output
    *      the strict upper trapezoidal part of a contains the strict
    *      upper trapezoidal part of r, and the lower trapezoidal
    *      part of a contains a factored form of q (the non-trivial
    *      elements of the u vectors described above).
    *
    *    pivot is a logical input variable. if pivot is set true,
    *      then column pivoting is enforced. if pivot is set false,
    *      then no column pivoting is done.
    *
    *    ipvt is an integer output array of length lipvt. ipvt
    *      defines the permutation matrix p such that a*p = q*r.
    *      column j of p is column ipvt(j) of the identity matrix.
    *      if pivot is false, ipvt is not referenced.
    *
    *    rdiag is an output array of length n which contains the
    *      diagonal elements of r.
    *
    *    acnorm is an output array of length n which contains the
    *      norms of the corresponding columns of the input matrix a.
    *      if this information is not needed, then acnorm can coincide
    *      with rdiag.
    *
    *    wa is a work array of length n. if pivot is false, then wa
    *      can coincide with rdiag.
    *
 */
    int i, j, k, kmax, minmn;
    double ajnorm, sum, temp;
    static double p05 = 0.05;

// *** compute the initial column norms and initialize several arrays.

    for ( j=0; j<n; j++ )
    {
        acnorm[j] = lm_enorm(m, &a[j*m]);
        rdiag[j] = acnorm[j];
        wa[j] = rdiag[j];
        if ( pivot )
            ipvt[j] = j;
    }
#if BUG
    printf( "qrfac\n" );
#endif

// *** reduce a to r with householder transformations.

    minmn = MIN(m,n);
    for ( j=0; j<minmn; j++ )
    {
        if ( !pivot ) goto pivot_ok;

// *** bring the column of largest norm into the pivot position.

        kmax = j;
        for ( k=j+1; k<n; k++ )
            if (rdiag[k] > rdiag[kmax])
                kmax = k;
        if (kmax == j) goto pivot_ok; // bug fixed in rel 2.1

        for ( i=0; i<m; i++ )
        {
            temp        = a[j*m+i];
            a[j*m+i]    = a[kmax*m+i];
            a[kmax*m+i] = temp;
        }
        rdiag[kmax] = rdiag[j];
        wa[kmax] = wa[j];
        k = ipvt[j];
        ipvt[j] = ipvt[kmax];
        ipvt[kmax] = k;

    pivot_ok:

// *** compute the Householder transformation to reduce the
//     j-th column of a to a multiple of the j-th unit vector.

            ajnorm = lm_enorm( m-j, &a[j*m+j] );
    if (ajnorm == 0.)
    {
        rdiag[j] = 0;
        continue;
    }

    if (a[j*m+j] < 0.)
        ajnorm = -ajnorm;
    for ( i=j; i<m; i++ )
        a[j*m+i] /= ajnorm;
    a[j*m+j] += 1;

// *** apply the transformation to the remaining columns
//     and update the norms.

    for ( k=j+1; k<n; k++ )
    {
        sum = 0;

        for ( i=j; i<m; i++ )
            sum += a[j*m+i]*a[k*m+i];

        temp = sum/a[j+m*j];

        for ( i=j; i<m; i++ )
            a[k*m+i] -= temp * a[j*m+i];

        if ( pivot && rdiag[k] != 0. )
        {
            temp = a[m*k+j]/rdiag[k];
            temp = MAX( 0., 1-temp*temp );
            rdiag[k] *= sqrt(temp);
            temp = rdiag[k]/wa[k];
            if ( p05*SQR(temp) <= LM_MACHEP )
            {
                rdiag[k] = lm_enorm( m-j-1, &a[m*k+j+1]);
                wa[k] = rdiag[k];
            }
        }
    }

    rdiag[j] = -ajnorm;
    }
}



void lm_qrsolv(int n, double* r, int ldr, int* ipvt, double* diag,
               double* qtb, double* x, double* sdiag, double* wa)
{
/*
    *     given an m by n matrix a, an n by n diagonal matrix d,
    *     and an m-vector b, the problem is to determine an x which
    *     solves the system
    *
    *        a*x = b ,       d*x = 0 ,
    *
    *     in the least squares sense.
    *
    *     this subroutine completes the solution of the problem
    *     if it is provided with the necessary information from the
    *     qr factorization, with column pivoting, of a. that is, if
    *     a*p = q*r, where p is a permutation matrix, q has orthogonal
    *     columns, and r is an upper triangular matrix with diagonal
    *     elements of nonincreasing magnitude, then qrsolv expects
    *     the full upper triangle of r, the permutation matrix p,
    *     and the first n components of (q transpose)*b. the system
    *     a*x = b, d*x = 0, is then equivalent to
    *
    *             t     t
    *        r*z = q *b ,  p *d*p*z = 0 ,
    *
    *     where x = p*z. if this system does not have full rank,
    *     then a least squares solution is obtained. on output qrsolv
    *     also provides an upper triangular matrix s such that
    *
    *         t       t           t
    *        p *(a *a + d*d)*p = s *s .
    *
    *     s is computed within qrsolv and may be of separate interest.
    *
    *     parameters
    *
    *    n is a positive integer input variable set to the order of r.
    *
    *    r is an n by n array. on input the full upper triangle
    *      must contain the full upper triangle of the matrix r.
    *      on output the full upper triangle is unaltered, and the
    *      strict lower triangle contains the strict upper triangle
    *      (transposed) of the upper triangular matrix s.
    *
    *    ldr is a positive integer input variable not less than n
    *      which specifies the leading dimension of the array r.
    *
    *    ipvt is an integer input array of length n which defines the
    *      permutation matrix p such that a*p = q*r. column j of p
    *      is column ipvt(j) of the identity matrix.
    *
    *    diag is an input array of length n which must contain the
    *      diagonal elements of the matrix d.
    *
    *    qtb is an input array of length n which must contain the first
    *      n elements of the vector (q transpose)*b.
    *
    *    x is an output array of length n which contains the least
    *      squares solution of the system a*x = b, d*x = 0.
    *
    *    sdiag is an output array of length n which contains the
    *      diagonal elements of the upper triangular matrix s.
    *
    *    wa is a work array of length n.
    *
 */
    int i, kk, j, k, nsing;
    double qtbpj, sum, temp;
    double sin, cos, tan, cotan; // these are local variables, not functions
    static double p25 = 0.25;
    static double p5 = 0.5;

// *** copy r and (q transpose)*b to preserve input and initialize s.
//     in particular, save the diagonal elements of r in x.

    for ( j=0; j<n; j++ )
    {
        for ( i=j; i<n; i++ )
            r[j*ldr+i] = r[i*ldr+j];
        x[j] = r[j*ldr+j];
        wa[j] = qtb[j];
    }
#if BUG
    printf( "qrsolv\n" );
#endif

// *** eliminate the diagonal matrix d using a givens rotation.

    for ( j=0; j<n; j++ )
{

// ***       prepare the row of d to be eliminated, locating the
//     diagonal element using p from the qr factorization.

    if (diag[ ipvt[j] ] == 0.)
        goto L90;
    for ( k=j; k<n; k++ )
        sdiag[k] = 0.;
    sdiag[j] = diag[ ipvt[j] ];

// ***       the transformations to eliminate the row of d
//     modify only a single element of (q transpose)*b
//     beyond the first n, which is initially 0..

    qtbpj = 0.;
    for ( k=j; k<n; k++ )
    {

//        determine a givens rotation which eliminates the
//        appropriate element in the current row of d.

        if (sdiag[k] == 0.)
            continue;
        kk = k + ldr * k; // <! keep this shorthand !>
        if ( fabs(r[kk]) < fabs(sdiag[k]) )
        {
            cotan = r[kk]/sdiag[k];
            sin = p5/sqrt(p25+p25*SQR(cotan));
            cos = sin*cotan;
        }
        else
        {
            tan = sdiag[k]/r[kk];
            cos = p5/sqrt(p25+p25*SQR(tan));
            sin = cos*tan;
        }

// ***          compute the modified diagonal element of r and
//        the modified element of ((q transpose)*b,0).

        r[kk] = cos*r[kk] + sin*sdiag[k];
        temp = cos*wa[k] + sin*qtbpj;
        qtbpj = -sin*wa[k] + cos*qtbpj;
        wa[k] = temp;

// *** accumulate the tranformation in the row of s.

        for ( i=k+1; i<n; i++ )
        {
            temp = cos*r[k*ldr+i] + sin*sdiag[i];
            sdiag[i] = -sin*r[k*ldr+i] + cos*sdiag[i];
            r[k*ldr+i] = temp;
        }
    }
    L90:

// *** store the diagonal element of s and restore
//     the corresponding diagonal element of r.

            sdiag[j] = r[j*ldr+j];
    r[j*ldr+j] = x[j];
}

// *** solve the triangular system for z. if the system is
//     singular, then obtain a least squares solution.

    nsing = n;
    for ( j=0; j<n; j++ )
{
    if ( sdiag[j] == 0. && nsing == n )
        nsing = j;
    if (nsing < n)
        wa[j] = 0;
}

    for ( j=nsing-1; j>=0; j-- )
{
    sum = 0;
    for ( i=j+1; i<nsing; i++ )
        sum += r[j*ldr+i]*wa[i];
    wa[j] = (wa[j] - sum)/sdiag[j];
}

// *** permute the components of z back to components of x.

    for ( j=0; j<n; j++ )
        x[ ipvt[j] ] = wa[j];
}

double lm_enorm( int n, double *x )
{
/*     given an n-vector x, this function calculates the
    *     euclidean norm of x.
    *
    *     the euclidean norm is computed by accumulating the sum of
    *     squares in three different sums. the sums of squares for the
    *     small and large components are scaled so that no overflows
    *     occur. non-destructive underflows are permitted. underflows
    *     and overflows do not occur in the computation of the unscaled
    *     sum of squares for the intermediate components.
    *     the definitions of small, intermediate and large components
    *     depend on two constants, LM_SQRT_DWARF and LM_SQRT_GIANT. the main
    *     restrictions on these constants are that LM_SQRT_DWARF**2 not
    *     underflow and LM_SQRT_GIANT**2 not overflow.
    *
    *     parameters
    *
    *    n is a positive integer input variable.
    *
    *    x is an input array of length n.
 */
    int i;
    double agiant, s1, s2, s3, xabs, x1max, x3max, temp;

    s1 = 0;
    s2 = 0;
    s3 = 0;
    x1max = 0;
    x3max = 0;
    agiant = LM_SQRT_GIANT/( (double) n);

    for ( i=0; i<n; i++ )
    {
        xabs = fabs(x[i]);
        if ( xabs > LM_SQRT_DWARF && xabs < agiant )
        {
// **  sum for intermediate components.
            s2 += xabs*xabs;
            continue;
        }

        if ( xabs >  LM_SQRT_DWARF )
        {
// **  sum for large components.
            if (xabs > x1max)
            {
                temp = x1max/xabs;
                s1 = 1 + s1*SQR(temp);
                x1max = xabs;
            }
            else
            {
                temp = xabs/x1max;
                s1 += SQR(temp);
            }
            continue;
        }
// **  sum for small components.
        if (xabs > x3max)
        {
            temp = x3max/xabs;
            s3 = 1 + s3*SQR(temp);
            x3max = xabs;
        }
        else
        {
            if (xabs != 0.)
            {
                temp = xabs/x3max;
                s3 += SQR(temp);
            }
        }
    }

// *** calculation of norm.

    if (s1 != 0)
        return x1max*sqrt(s1 + (s2/x1max)/x1max);
    if (s2 != 0)
    {
        if (s2 >= x3max)
            return sqrt( s2*(1+(x3max/s2)*(x3max*s3)) );
        else
            return sqrt( x3max*((s2/x3max)+(x3max*s3)) );
    }

    return x3max*sqrt(s3);
}



/*-----------------------------------------------------------------------------*
 *
 * Some defines, structs, and helper functions copied from ooDialog
 *
*------------------------------------------------------------------------------*/
#define API_FAILED_MSG            "system API %s() failed; OS error code %d"

/* Struct for the WindowBase object CSelf. */
typedef struct _wbCSelf {
    HWND              hwnd;
    RexxObjectPtr     rexxHwnd;
    RexxObjectPtr     rexxSelf;
    uint32_t          sizeX;
    uint32_t          sizeY;
} CWindowBase;
typedef CWindowBase *pCWindowBase;

inline void systemServiceException(RexxMethodContext *c, char *msg)
{
    c->RaiseException1(Rexx_Error_System_service_user_defined, c->String(msg));
}

void systemServiceExceptionCode(RexxMethodContext *context, const char *msg, const char *arg1, DWORD rc)
{
    TCHAR buffer[512];
    _snprintf(buffer, sizeof(buffer), msg, arg1, rc);
    systemServiceException(context, buffer);
}


inline void pointer2string(char *result, void *pointer)
{
    sprintf(result, "0x%p", pointer);
}

inline void dword2string(char *result, void *dword)
{
    sprintf(result, "%d", (int)dword); // added (int) in V34.0.0.35
}

RexxStringObject pointer2string(RexxMethodContext *c, void *pointer)
{
    char buf[32];
    pointer2string(buf, pointer);
    return c->String(buf);
}

/**
 * Gets a Class object.
 *
 * This is for use for classes visible within the scope of the context, like say
 * .PlainBaseDialog, or .Rect.  Use c->FindClass() to directly get classes from
 * the environment like .Bag or .Directory.
 *
 * @param c     The method context we are operating in.
 * @param name  The name of the class to try and find.
 *
 * @return The class object or null on failure.
 *
 * @remarks  When null is returned an error has been raised: 98.909
 *
 *   98: The language processor detected a specific error during execution. The
 *   associated error gives the reason for the error.
 *
 *   909: Class "class" not found
 */
RexxClassObject rxGetContextClass(RexxMethodContext *c, CSTRING name)
{
    RexxClassObject theClass = c->FindContextClass(name);
    if ( theClass == NULL )
    {
        c->RaiseException1(Rexx_Error_Execution_noclass, c->String(name));
    }
    return theClass;
}

RexxObjectPtr createNewWindow(RexxMethodContext *c, HWND hwnd)
{
    RexxClassObject windowCls = rxGetContextClass(c, "Sat2ChartFUNCS");
    if ( windowCls == NULLOBJECT )
    {
        return c->Nil();
    }

    return c->SendMessage1(windowCls, "NEW", pointer2string(c, hwnd));
}

RexxMethod1(RexxObjectPtr, window_findWindow_cls, CSTRING, title)
{
    HWND hwnd = FindWindow(NULL, title);

    if ( hwnd == NULL )
    {
        return context->Nil();
    }
    else
    {
        return createNewWindow(context, hwnd);
    }
}


RexxMethod2(RexxObjectPtr, window_init, POINTERSTRING, _hwnd, OSELF, self)
{
    HWND hwnd = (HWND)_hwnd;

    if ( !IsWindow(hwnd) )
    {
        context->RaiseException1(Rexx_Error_Incorrect_method_user_defined,
                                 context->String("Argument 1 is not a valid window handle"));
    }
    else
    {
        RexxBufferObject obj = context->NewBuffer(sizeof(CWindowBase));
        if ( obj == NULLOBJECT )
        {
            return NULLOBJECT;
        }

        RECT r = {0};
        if ( GetWindowRect(hwnd, &r) == 0 )
        {
            systemServiceExceptionCode(context, API_FAILED_MSG, "GetWindowRect", GetLastError());
            return NULLOBJECT;
        }

        pCWindowBase pcwb = (pCWindowBase)context->BufferData(obj);
        memset(pcwb, 0, sizeof(CWindowBase));

        pcwb->hwnd     = hwnd;
        pcwb->rexxSelf = self;
        pcwb->rexxHwnd = pointer2string(context, hwnd);

        pcwb->sizeX =  (uint32_t)(r.right - r.left);
        pcwb->sizeY =  (uint32_t)(r.bottom - r.top);

        context->SetObjectVariable("CSELF", obj);
    }
    return NULLOBJECT;
}


RexxMethod3(uint32_t, window_sendKeyPress, uint8_t, key, OPTIONAL_RexxArrayObject, modifiers, CSELF, pCSelf)
{
    LPINPUT keys  = NULL;
    uint32_t ret = 0;
    uint32_t count = 0;

    if ( argumentExists(2) )
    {
        count = (uint32_t)context->ArrayItems(modifiers);
    }
    count++;

    keys = (INPUT *)LocalAlloc(LPTR, sizeof(INPUT) * count * 2);
    if ( keys == NULL )
    {
        context->RaiseException1(Rexx_Error_System_service_user_defined,
                                 context->String("Failed to allocate memory"));
        goto done_out;
    }

    if ( count == 1 )
    {
        keys[0].type = INPUT_KEYBOARD;
        keys[0].ki.dwExtraInfo = GetMessageExtraInfo();
        keys[0].ki.wVk = key;

        keys[1].ki.wVk = key;
        keys[1].type = INPUT_KEYBOARD;
        keys[1].ki.dwFlags = KEYEVENTF_KEYUP;
        keys[1].ki.dwExtraInfo = GetMessageExtraInfo();
    }
    else
    {
        uint32_t      modKey;
        RexxObjectPtr mod;

        // For each key initialize a down key and an up key event. We want the
        // modifier keys pressed down first and released last.
        for ( size_t i = 0; i < count; i++)
        {
            if ( i == count - 1 )
            {
                keys[i].type = INPUT_KEYBOARD;
                keys[i].ki.dwExtraInfo = GetMessageExtraInfo();
                keys[i].ki.wVk = key;

                keys[i + 1].ki.wVk = key;
                keys[i + 1].type = INPUT_KEYBOARD;
                keys[i + 1].ki.dwFlags = KEYEVENTF_KEYUP;
                keys[i + 1].ki.dwExtraInfo = GetMessageExtraInfo();
            }
            else
            {
                mod = context->ArrayAt(modifiers, i + 1);
                if ( mod == NULLOBJECT || ! context->ObjectToUnsignedInt32(mod, &modKey) ||
                     modKey == 0 || modKey > 255 )
                {
                    context->RaiseException1(
                        Rexx_Error_Incorrect_method_user_defined,
                        context->String("Argument 2 does not contain a valid virtual key code"));

                    goto done_out;
                }

                keys[i].type = INPUT_KEYBOARD;
                keys[i].ki.dwExtraInfo = GetMessageExtraInfo();
                keys[i].ki.wVk = modKey;

                keys[i + count + 1].ki.wVk = modKey;
                keys[i + count + 1].type = INPUT_KEYBOARD;
                keys[i + count + 1].ki.dwFlags = KEYEVENTF_KEYUP;
                keys[i + count + 1].ki.dwExtraInfo = GetMessageExtraInfo();
            }
        }
    }

    pCWindowBase pcwb = (pCWindowBase)pCSelf;

    if ( SetForegroundWindow(pcwb->hwnd) )
    {
        ret = SendInput(count * 2, keys, sizeof(INPUT));
    }

done_out:
    if ( keys != NULL )
    {
        LocalFree(keys);
    }
    return ret;
}


/** Window::toForeground()
 *
 *
 *  When this is merged to ooDialog, this function should use:
 *
 *  oodSetForegroundWindow()
 *
 */
RexxMethod1(RexxObjectPtr, window_toForeground, CSELF, pCSelf)
{
    HWND hwnd = ((pCWindowBase)pCSelf)->hwnd;

    SetLastError(0);

    HWND hwndOld =  GetForegroundWindow();
    if (hwndOld != hwnd)
    {
        SetForegroundWindow(hwnd);
    }
    return pointer2string(context, hwndOld);
}

/////////////////////////////////////////////////////////////////////////////
CQuantizer::CQuantizer (UINT nMaxColors, UINT nColorBits)
{
	m_nColorBits = nColorBits < 8 ? nColorBits : 8;
	
	// printf("  nMaxColors=%d nColorBits=%d m_nColorBits=%d\n", nMaxColors,nColorBits,m_nColorBits);  REMOVE  ds.biBitCount

	m_pTree	= NULL;
	m_nLeafCount = 0;
	for	(int i=0; i<=(int) m_nColorBits; i++)
		m_pReducibleNodes[i] = NULL;
	m_nMaxColors = m_nOutputMaxColors = nMaxColors;
	if (m_nMaxColors<16) m_nMaxColors=16;
}
/////////////////////////////////////////////////////////////////////////////
CQuantizer::~CQuantizer	()
{
	if (m_pTree	!= NULL)
		DeleteTree (&m_pTree);
}
/////////////////////////////////////////////////////////////////////////////
BOOL CQuantizer::ProcessImage (HANDLE hImage)
{
	BYTE r,	g, b, a;
	int	i, j;

	BITMAPINFOHEADER ds;
	memcpy(&ds,hImage, sizeof(ds));
	int effwdt = ((((ds.biBitCount * ds.biWidth ) + 31) / 32) * 4);

	int	nPad = effwdt - (((ds.biWidth *	ds.biBitCount) + 7) / 8);
	
	// printf(" BITCOUNT=%d\n",ds.biBitCount); REMOVE 

	BYTE* pbBits = (BYTE*)hImage + *(DWORD*)hImage + ds.biClrUsed * sizeof(RGBQUAD);

	switch (ds.biBitCount) {

	case 1:	// 1-bit DIB
	case 4:	// 4-bit DIB
	case 8:	// 8-bit DIB
		for	(i=0; i<ds.biHeight;	i++) {
			for	(j=0; j<ds.biWidth; j++)	{
				BYTE idx=GetPixelIndex(j,i,ds.biBitCount,effwdt,pbBits);
				BYTE* pal = (BYTE*)(hImage) + sizeof(BITMAPINFOHEADER);
				long ldx = idx*sizeof(RGBQUAD);
				b = pal[ldx];
				g = pal[ldx+1];
				r = pal[ldx+2];
				a = pal[ldx+3];
				AddColor (&m_pTree,	r, g, b, a, m_nColorBits, 0, &m_nLeafCount,
					m_pReducibleNodes);
				while (m_nLeafCount	> m_nMaxColors)
					ReduceTree (m_nColorBits, &m_nLeafCount,
						m_pReducibleNodes);
			}
		}

		break;
	case 24: //	24-bit DIB
		for	(i=0; i<ds.biHeight;	i++) {
			for	(j=0; j<ds.biWidth; j++)	{
				b =	*pbBits++;
				g =	*pbBits++;
				r =	*pbBits++;
				AddColor (&m_pTree,	r, g, b, 0, m_nColorBits, 0, &m_nLeafCount,
					m_pReducibleNodes);
				while (m_nLeafCount	> m_nMaxColors)
					ReduceTree (m_nColorBits, &m_nLeafCount, m_pReducibleNodes);
			}
			pbBits += nPad;
		}
		break;

	default: //	Unrecognized color format
		return FALSE;
	}
	return TRUE;
}
/////////////////////////////////////////////////////////////////////////////
void CQuantizer::AddColor (NODE** ppNode, BYTE r, BYTE g, BYTE b, BYTE a,
	UINT nColorBits, UINT nLevel, UINT*	pLeafCount,	NODE** pReducibleNodes)
{
	static BYTE	mask[8]	= {	0x80, 0x40,	0x20, 0x10,	0x08, 0x04,	0x02, 0x01 };

	// If the node doesn't exist, create it.
	if (*ppNode	== NULL)
		*ppNode	= (NODE*)CreateNode (nLevel, nColorBits, pLeafCount, pReducibleNodes);

	// Update color	information	if it's	a leaf node.
	if ((*ppNode)->bIsLeaf)	{
		(*ppNode)->nPixelCount++;
		(*ppNode)->nRedSum += r;
		(*ppNode)->nGreenSum +=	g;
		(*ppNode)->nBlueSum	+= b;
		(*ppNode)->nAlphaSum += a;
	} else {	// Recurse a level deeper if the node is not a leaf.
		int	shift =	7 -	nLevel;
		int	nIndex =(((r & mask[nLevel]) >> shift) << 2) |
					(((g & mask[nLevel]) >>	shift) << 1) |
					(( b & mask[nLevel]) >> shift);
		AddColor (&((*ppNode)->pChild[nIndex]),	r, g, b, a, nColorBits,
					nLevel + 1,	pLeafCount,	pReducibleNodes);
	}
}
/////////////////////////////////////////////////////////////////////////////
void* CQuantizer::CreateNode (UINT nLevel, UINT	nColorBits,	UINT* pLeafCount,
	NODE** pReducibleNodes)
{
	NODE* pNode = (NODE*)calloc(1,sizeof(NODE));

	if (pNode== NULL) return NULL;

	pNode->bIsLeaf = (nLevel ==	nColorBits)	? TRUE : FALSE;
	if (pNode->bIsLeaf) (*pLeafCount)++;
	else {
		pNode->pNext = pReducibleNodes[nLevel];
		pReducibleNodes[nLevel]	= pNode;
	}
	return pNode;
}
/////////////////////////////////////////////////////////////////////////////
void CQuantizer::ReduceTree	(UINT nColorBits, UINT*	pLeafCount,
	NODE** pReducibleNodes)
{
	int i;
	// Find	the	deepest	level containing at	least one reducible	node.
	for	(i=nColorBits -	1; (i>0) &&	(pReducibleNodes[i]	== NULL); i--);

	// Reduce the node most	recently added to the list at level	i.
	NODE* pNode	= pReducibleNodes[i];
	pReducibleNodes[i] = pNode->pNext;

	UINT nRedSum = 0;
	UINT nGreenSum = 0;
	UINT nBlueSum =	0;
	UINT nAlphaSum = 0;
	UINT nChildren = 0;

	for	(i=0; i<8; i++)	{
		if (pNode->pChild[i] !=	NULL) {
			nRedSum	+= pNode->pChild[i]->nRedSum;
			nGreenSum += pNode->pChild[i]->nGreenSum;
			nBlueSum +=	pNode->pChild[i]->nBlueSum;
			nAlphaSum += pNode->pChild[i]->nAlphaSum;
			pNode->nPixelCount += pNode->pChild[i]->nPixelCount;
			free(pNode->pChild[i]);
			pNode->pChild[i] = NULL;
			nChildren++;
		}
	}

	pNode->bIsLeaf = TRUE;
	pNode->nRedSum = nRedSum;
	pNode->nGreenSum = nGreenSum;
	pNode->nBlueSum	= nBlueSum;
	pNode->nAlphaSum	= nAlphaSum;
	*pLeafCount	-= (nChildren -	1);
}
/////////////////////////////////////////////////////////////////////////////
void CQuantizer::DeleteTree	(NODE**	ppNode)
{
	for	(int i=0; i<8; i++)	{
		if ((*ppNode)->pChild[i] !=	NULL) DeleteTree (&((*ppNode)->pChild[i]));
	}
	free(*ppNode);
	*ppNode	= NULL;
}
/////////////////////////////////////////////////////////////////////////////
void CQuantizer::GetPaletteColors (NODE* pTree,	RGBQUAD* prgb, UINT* pIndex, UINT* pSum)
{
	if (pTree){
		if (pTree->bIsLeaf)	{
			prgb[*pIndex].rgbRed   = (BYTE)min(255,0.5f + ((float)pTree->nRedSum)   / pTree->nPixelCount);
			prgb[*pIndex].rgbGreen = (BYTE)min(255,0.5f + ((float)pTree->nGreenSum) / pTree->nPixelCount);
			prgb[*pIndex].rgbBlue  = (BYTE)min(255,0.5f + ((float)pTree->nBlueSum)  / pTree->nPixelCount);
			prgb[*pIndex].rgbReserved =	(BYTE)min(255,0.5f + ((float)pTree->nAlphaSum) / pTree->nPixelCount);
			if (pSum) pSum[*pIndex] = pTree->nPixelCount;
			(*pIndex)++;
		} else {
			for	(int i=0; i<8; i++)	{
				if (pTree->pChild[i] !=	NULL)
					GetPaletteColors (pTree->pChild[i],	prgb, pIndex, pSum);
			}
		}
	}
}
/////////////////////////////////////////////////////////////////////////////
UINT CQuantizer::GetColorCount ()
{
	return m_nLeafCount;
}
/////////////////////////////////////////////////////////////////////////////
void CQuantizer::SetColorTable (RGBQUAD* prgb)
{
	UINT nIndex	= 0;
	if (m_nOutputMaxColors<16){
		UINT nSum[16];
		RGBQUAD tmppal[16];
		GetPaletteColors (m_pTree, tmppal, &nIndex, nSum);
		if (m_nLeafCount>m_nOutputMaxColors) {
			UINT j,k,nr,ng,nb,na,ns,a,b;
			for (j=0;j<m_nOutputMaxColors;j++){
				a=(j*m_nLeafCount)/m_nOutputMaxColors;
				b=((j+1)*m_nLeafCount)/m_nOutputMaxColors;
				nr=ng=nb=na=ns=0;
				for (k=a;k<b;k++){
					nr+=tmppal[k].rgbRed * nSum[k];
					ng+=tmppal[k].rgbGreen * nSum[k];
					nb+=tmppal[k].rgbBlue * nSum[k];
					na+=tmppal[k].rgbReserved * nSum[k];
					ns+= nSum[k];
				}
				prgb[j].rgbRed   = (BYTE)min(255,0.5f + ((float)nr)/ns);
				prgb[j].rgbGreen = (BYTE)min(255,0.5f + ((float)ng)/ns);
				prgb[j].rgbBlue  = (BYTE)min(255,0.5f + ((float)nb)/ns);
				prgb[j].rgbReserved = (BYTE)min(255,0.5f + ((float)na)/ns);
			}
		} else {
			memcpy(prgb,tmppal,m_nLeafCount * sizeof(RGBQUAD));
		}
	} else {
		GetPaletteColors (m_pTree, prgb, &nIndex, 0);
	}
}
/////////////////////////////////////////////////////////////////////////////
BYTE CQuantizer::GetPixelIndex(long x, long y, int nbit, long effwdt, BYTE *pimage)
{
	if (nbit==8){
		return pimage[y*effwdt + x];
	} else {
		BYTE pos;
		BYTE iDst= pimage[y*effwdt + (x*nbit >> 3)];
		if (nbit==4){
			pos = (BYTE)(4*(1-x%2));
			iDst &= (0x0F<<pos);
			return (iDst >> pos);
		} else if (nbit==1){
			pos = (BYTE)(7-x%8);
			iDst &= (0x01<<pos);
			return (iDst >> pos);
		}
	}
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
TCHAR* FindExtension(const TCHAR * name)
{
	int len = _tcslen(name);
	int i;
	for (i = len-1; i >= 0; i--){
		if (name[i] == '.'){
			return (TCHAR*)(name+i+1);
		}
	}
	return (TCHAR*)(name+len);
}
/**************************** DestroyWin **************************
 * Destroys the window represented by the hwnd
 *
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine2(int, DestroyWin, uintptr_t, hwnd, RexxArrayObject, diagmsgs)
{
   char buffer[512]; 
   int rc,rtnrc;
   BOOL boolrc;
   
	using namespace Gdiplus;
	GdiplusStartupInput gdiplusStartupInput;
	ULONG_PTR gdiplusToken;
	GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);
	{   

      sprintf(buffer,"DIAG DestroyWin Handle=%u ",hwnd);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        
   	// Check if we have a valid window handle
      if ( !IsWindow((HWND)hwnd) )
      {
         sprintf(buffer,"ERROR: DestroyWin Argument 1 is not a valid window handle");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;                         
      }  	
      
   	// Destroy the window    
     boolrc = DestroyWindow((HWND)hwnd); 
     if (boolrc == 0)
     {
         sprintf(buffer,"ERROR: DestroyWindow was NOT successful");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
         rtnrc = 1;                      
     } 
     else
     {
         sprintf(buffer,"DIAG DestroyWindow was successful");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         rtnrc = 0;  
     } 
     
     
   }   
   GdiplusShutdown(gdiplusToken);     
	
   return rtnrc;
}
/**************************** ClipBoard **************************
 * Takes the contents of the clipboard and creates an image file
 *
 * Syntax:	rc = ClipBoard(window_handle, filename, , colors, diffusion, diag, diagmsgs)
 *
 * Params:	window_handle - handle of the window to take a screen shot
 *			   filename - name of the file
 *			   diag    - if gt 0 it will print diagnostic information  
 *			   diagmsgs - array to return diagnostic messages 
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine3(int, ClipBoard, CSTRING, filename, int, diag, RexxArrayObject, diagmsgs) 
{
   char buffer[512]; 
   int rc;
   int rtnrc = 0;
   
	using namespace Gdiplus;
	GdiplusStartupInput gdiplusStartupInput;
	ULONG_PTR gdiplusToken;
	GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);
	{  
      sprintf(buffer,"DIAG ClipBoard Filename=%s",filename);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      if (IsClipboardFormatAvailable(CF_BITMAP) || IsClipboardFormatAvailable(CF_DIB) || IsClipboardFormatAvailable(CF_DIBV5))
      {
         if (OpenClipboard(NULL))
         {
            HANDLE hClipboard = GetClipboardData(CF_DIB);
            
            if (!hClipboard)
            {
                hClipboard = GetClipboardData(CF_DIBV5);
            }
            
            if (hClipboard != NULL && hClipboard != INVALID_HANDLE_VALUE)
            {
               void* dib = GlobalLock(hClipboard);
               
               if (dib)
               {
                  DIB *info = reinterpret_cast<DIB*>(dib);
                  BMP bmp = {0};
                  bmp.header.type = 0x4D42;
                  bmp.header.offset = 54;
                  bmp.header.bfSize = info->biSizeImage + bmp.header.offset;
                  bmp.dib = *info;        
                  
                  // std::ofstream file("C:/Junk/Clipboard_image.bmp", std::ios::out | std::ios::binary);
                  std::ofstream file(filename, std::ios::out | std::ios::binary);
                  if (file)
                  {
                     file.write(reinterpret_cast<char*>(&bmp.header.type), sizeof(bmp.header.type));
                     file.write(reinterpret_cast<char*>(&bmp.header.bfSize), sizeof(bmp.header.bfSize));
                     file.write(reinterpret_cast<char*>(&bmp.header.reserved), sizeof(bmp.header.reserved));
                     file.write(reinterpret_cast<char*>(&bmp.header.offset), sizeof(bmp.header.offset));
                     file.write(reinterpret_cast<char*>(&bmp.dib.biSize), sizeof(bmp.dib.biSize));
                     file.write(reinterpret_cast<char*>(&bmp.dib.biWidth), sizeof(bmp.dib.biWidth));
                     file.write(reinterpret_cast<char*>(&bmp.dib.biHeight), sizeof(bmp.dib.biHeight));
                     file.write(reinterpret_cast<char*>(&bmp.dib.biPlanes), sizeof(bmp.dib.biPlanes));
                     file.write(reinterpret_cast<char*>(&bmp.dib.biBitCount), sizeof(bmp.dib.biBitCount));
                     file.write(reinterpret_cast<char*>(&bmp.dib.biCompression), sizeof(bmp.dib.biCompression));
                     file.write(reinterpret_cast<char*>(&bmp.dib.biSizeImage), sizeof(bmp.dib.biSizeImage));
                     file.write(reinterpret_cast<char*>(&bmp.dib.biXPelsPerMeter), sizeof(bmp.dib.biXPelsPerMeter));
                     file.write(reinterpret_cast<char*>(&bmp.dib.biYPelsPerMeter), sizeof(bmp.dib.biYPelsPerMeter));
                     file.write(reinterpret_cast<char*>(&bmp.dib.biClrUsed), sizeof(bmp.dib.biClrUsed));
                     file.write(reinterpret_cast<char*>(&bmp.dib.biClrImportant), sizeof(bmp.dib.biClrImportant));
                     file.write(reinterpret_cast<char*>(info + 1), bmp.dib.biSizeImage);
                     
                     sprintf(buffer,"DIAG ClipBoard BMP_WIDTH=%d",bmp.dib.biWidth);
                     rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                     sprintf(buffer,"DIAG ClipBoard BMP_HEIGHT=%d",bmp.dib.biHeight);
                     rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                     sprintf(buffer,"DIAG ClipBoard X_PPM=%d",bmp.dib.biXPelsPerMeter);
                     rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                     sprintf(buffer,"DIAG ClipBoard Y_PPM=%d",bmp.dib.biYPelsPerMeter);
                     rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                  }
               }
               GlobalUnlock(dib);
            }
             CloseClipboard();
         }
         else // Unable to open the clipboard
         {
            sprintf(buffer,"ERROR: ClipBoard Unable to Open the Clipboard!");
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            return 1;
         }     
      }
   }
   GdiplusShutdown(gdiplusToken);     
	
   return rtnrc;
} 
/**************************** ScreenShot **************************
 * Takes a screen shot of the window handle that is passed.
 *
 * Syntax:	rc = Screenshot(window_handle, filename, , colors, diffusion, diag, diagmsgs)
 *
 * Params:	window_handle - handle of the window to take a screen shot
 *			   filename - name of the file
 *			   colors - maximum number of colors in output 
 *			   diffusion - error diffusion, either 0=OFF or 1=ON 
 *			   diag    - if gt 0 it will print diagnostic information  
 *			   diagmsgs - array to return diagnostic messages 
 *			   histogram  (optional) an array to return the resulting images histogram changed to histogram in V27.0.0.13
 *			   Quantize  (optional) a flag if = 0 skip Quantizing added in V31.0.0.4
 *
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine8(int, Screenshot, uintptr_t, hwnd, CSTRING, filename, int, colors, int, diffusion, int, diag, RexxArrayObject, diagmsgs, 
               OPTIONAL_RexxArrayObject, histogram, OPTIONAL_int, Quantize)
{
   char buffer[512]; 
   int rc;
   int rtnrc = 0;
   
	using namespace Gdiplus;
	GdiplusStartupInput gdiplusStartupInput;
	ULONG_PTR gdiplusToken;
	GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);
	{   	
	   // int colors = 127;
      // int colors = 64;
      INT numcolors;   

      sprintf(buffer,"DIAG Screenshot Handle=%u Filename=%s Colors=%d Diffusion=%d",hwnd, filename, colors,diffusion);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      
      sprintf(buffer,"DIAG Screenshot Version: 31.0.0.4\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      
   	// See if output is a valid image type
   	TCHAR* extout = FindExtension(filename);
   	_tcslwr(extout);
   	int typeout = CxImage::GetTypeIdFromName(extout);
   	
   	if ( diag>0 )
      {
         sprintf(buffer,"DIAG Screenshot extout=%s typeout=%d",extout,typeout);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }      
       
   	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
      {
         sprintf(buffer,"ERROR: Screenshot Argument 2 is an invalid image type\n");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
   	}  
        
   	// Check if we have a valid window handle
      if ( !IsWindow((HWND)hwnd) )
      {
         sprintf(buffer,"ERROR: Screenshot Argument 1 is not a valid window handle\n");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;                         
      }  	
      
   	// Get the window size  
      RECT r;
      rc = -1;   
      GetWindowRect((HWND)hwnd, &r);   
         
      int before_width =  r.right - r.left ;
      int before_height = r.bottom - r.top ;
      
      // Get the screen resolution 
      DEVMODE dev;
      int Enum_w, Enum_h;
      EnumDisplaySettings(nullptr, ENUM_CURRENT_SETTINGS, &dev);
      Enum_w = dev.dmPelsWidth;
      Enum_h = dev.dmPelsHeight;     
      sprintf(buffer,"DIAG Screenshot Before_width/height=%d %d Screen Resolution:%d %d",before_width,before_height,Enum_w,Enum_h);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      
      double scale = (double)Enum_w / (double)before_width; 
      sprintf(buffer, "%.*g", 3, scale);	
      scale = strtod(buffer, NULL);	
      
      // Adjust the rectangle added in V3.0.4.1
      sprintf(buffer,"DIAG Screenshot Before: scale=%f top=%d bottom=%d left=%d right=%d",scale, r.top,r.bottom,r.left,r.right);
      // printf(buffer); printf("\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
       
      // if the scale is not 1.0 then adjust the top and bottom 
      if (scale > 1.0)
      {
         int incr = (int)ceil(r.top * scale) -  r.top;  // ceil = roundup
         int top_before = r.top;
         r.top += incr;
         r.bottom += incr;
         sprintf(buffer,"DIAG Screenshot adjusted to scale=%f r.top by:%d was=%d now=%d",scale, incr, top_before, r.top);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      }   
       
      int width, height, x, y;
      x = r.left;
      y = r.top;
      
      width =  r.right - r.left ;
      height = r.bottom - r.top ;
      
      width =  (int)(width * scale);
      height = (int)(height * scale); 
      
      sprintf(buffer,"DIAG Screenshot After Width=%d Height=%d",width,height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));             
      
      sprintf(buffer,"DIAG Screenshot After: top=%d bottom=%d left=%d right=%d width=%d height=%d",r.top,r.bottom,r.left,r.right,width,height);
      // printf(buffer); printf("\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
       
      sprintf(buffer,"DIAG Screenshot Screenshot X=%d Y=%d",x,y);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));               
    
      //  Check if the screen dimensions are valid
      if(width <= 0 || height <= 0) 
      {
         sprintf(buffer,"ERROR: Screenshot Invalid screen dimensions width=%d height=%d\n",width, height);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
      }       
      
      // Copy the screen
      HDC scrdc, memdc;
      HBITMAP membit;
      scrdc = ::GetDC(0);
      memdc = CreateCompatibleDC(scrdc);
      membit = CreateCompatibleBitmap(scrdc, width, height);
      HBITMAP hOldBitmap =(HBITMAP) SelectObject(memdc, membit);
      BitBlt(memdc, 0, 0, width, height, scrdc, x, y, SRCCOPY);	
      Gdiplus::Bitmap bitmap(membit, NULL);
     
      // Create a new image from the bitmap 
      CxImage *newImage = new CxImage();     
      
      rc = newImage->CreateFromHBITMAP(membit);
      if (!rc)
      {
         sprintf(buffer,"ERROR: Screenshot Invalid rc from CreateFromHBITMAP RC=%d\n",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         rtnrc = 1;
      }  
      else
      {  
         int resx, resy;
         if(diag > 0)   
         {
            resx = newImage->GetXDPI(); // Get X resolution
            resy = newImage->GetYDPI(); // Get Y resolution
            sprintf(buffer,"DIAG Screenshot BEFORE:Xres=%d Yres=%d",resx, resy);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }  
         
         sprintf(buffer,"DIAG Screenshot Width=%d Height=%d",width,height);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 	             
                                   
         // If the output is NOT a JPG   set resolution to 96ppi, decrease the bpp to 8 and numcolors                 
         if (typeout != CXIMAGE_FORMAT_JPG)
         {
            ULONG outbpp; 
            RGBQUAD * ppal;
         	
            CQuantizer m(16777216,8);
	         m.ProcessImage(newImage->GetDIB());	 
            numcolors = m.GetColorCount();
            // if(diag > 0)   
            // {
               sprintf(buffer,"DIAG Screenshot BEFORE: numcolors=%d",numcolors);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            // } 
         	
            newImage->SetXDPI(96);
            newImage->SetYDPI(96);
            
            bool do_Quantizer = true;
            // If there is a quantize argument
            if ( argumentExists(8) )
            {  
               sprintf(buffer,"DIAG Screenshot Argument 8 exists Quantize=%d",Quantize);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            
               if (Quantize == 0) do_Quantizer = false; // if its 0 don't quantize
               
               sprintf(buffer,"DIAG Screenshot do_Quantizer=%d",do_Quantizer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            }
            
            // Do the quantize if no override
            if (do_Quantizer)
            {
               // CQuantizer q(colors,(colors>16?7:8));   REMOVE
               CQuantizer q(colors,4);     //  was 8
               q.ProcessImage(newImage->GetDIB());           
               ppal=(RGBQUAD*)calloc(256*sizeof(RGBQUAD),1);
               q.SetColorTable(ppal);
            
               bool error_diffusion = (diffusion == 0)	? FALSE : TRUE;       
               rc = newImage->DecreaseBpp(8,error_diffusion,ppal,colors); //  DIFFUSION ADDED IN V9.5.2   
               outbpp = newImage->GetBpp(); 
               numcolors = q.GetColorCount();
            
               // if(diag > 0)   
               // {
                  sprintf(buffer,"DIAG Screenshot DecreaseBpp RC=%d outbpp=%d AFTER: numcolors=%d",rc,outbpp,numcolors);
                  rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));     
               // }   
               if (ppal) free(ppal); // Added in V30.0.6.0    
            }      
         }
         
         // Pass back the type, width, height, X resolution and Y resolution
         int b_resx, b_resy, b_width, b_height;
         b_resx = newImage->GetXDPI(); // Get X resolution
         b_resy = newImage->GetYDPI(); // Get Y resolution
         b_width = newImage->GetWidth();
         b_height = newImage->GetHeight();
         sprintf(buffer,"DIAG Screenshot TYPEOUT=%d XRES=%d YRES=%d ",typeout, b_resx, b_resy);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));       
         sprintf(buffer,"DIAG Screenshot WIDTH=%d HEIGHT=%d NUMCOLORS=%d",b_width,b_height,numcolors);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         
         // If there is a histogram parm pass it back along with the palette histograms
         if ( argumentExists(7) )
         {   
            RGBQUAD color;
            ULONG pixval;
            long x, y;
            uint32_t idx;
            // Get the size of the palette  in bytes
            int palsize = newImage->GetPaletteSize()/4;   
            sprintf(buffer,"DIAG ScreenShot palsize=%d",palsize);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
            
            // Create a array to hold the histogram  of the images, each entry is 6 words long:  
            // color index,  color count, palette colors r,g,b, and pixval
         	ULONG *pArray = (ULONG*) malloc(ENTSIZE*palsize);
         	
         	// Reset the array to zero  to do the image first
         	memset(pArray,0,ENTSIZE*palsize); 
         	   	
         	// Create a histogram of the image 
            int i;
         	// for(i=0; i<palsize; i++)
            for(i=0; i<127; i++)   // pass back 127 palette entries
         	{
               color = newImage->GetPaletteColor(i);
               pixval = Get_Pixval(color);           // added in V17.0.0.19
               pArray[i*ENTLEN] = i; // Save the index for this color
               pArray[i*ENTLEN+2] = color.rgbRed; 
               pArray[i*ENTLEN+3] = color.rgbGreen; 
               pArray[i*ENTLEN+4] = color.rgbBlue; 
               pArray[i*ENTLEN+5] = pixval; // Save the pixval for this color
            }
               
            // Scan all the pixels and get the palette color counts 
         	for(y=0; y<b_height; y++)
            {
         		for(x=0; x<b_width; x++)
               {
                  // Get color index and increment the count for the index
         			idx = newImage->GetPixelIndex(x,y);
                  pArray[idx*ENTLEN+1]++;	// increment the histogram count for this color	
         		}
            }        
            
            // pass back the information in the histogram  changed in V27.0.0.14 to only pass the count and pixval
            // for(int i=0; i<palsize; i++)
            for(int i=0; i<127; i++)
            {
      		   sprintf(buffer,"%010d %d",pArray[i*ENTLEN+1],pArray[i*ENTLEN+5]);  // change to 010d in V30.0.0.12
               rc = context->ArrayAppendString(histogram, buffer, strlen(buffer));
            }  
            
            if (pArray) free(pArray); // Added in V30.0.6.0                        
         }
         
         
         // Save the image
         rc = newImage->Save(filename,typeout);	
         if(!rc)
         {
           sprintf(buffer,"ERROR: Screenshot Invalid rc from Save Image RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           rtnrc = 1;
         }	    
      }	
   		
   	// Delete everything	
   	SelectObject(memdc, hOldBitmap);
   	DeleteObject(memdc);
   	DeleteObject(membit);
   	::ReleaseDC(0,scrdc);
    	  
      // Destroy the image if one exists
      if (newImage->IsValid()) 
      {
         rc = newImage->Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: Screenshot 1. Invalid rc from Destroy Image RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }	 
         
         sprintf(buffer,"DIAG Screenshot destroyed newimage rc=%d",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
      }  
      
   }   
   GdiplusShutdown(gdiplusToken);    
   
   if ( diag>0 ) 
   {
      sprintf(buffer,"DIAG Screenshot Returning RC=%d",rtnrc);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }     
	
   return rtnrc;
}
/**************************** ScreenShot_Mono **************************
 * Takes a screen shot and changes it to black and white
 *
 * Syntax:	rc = Screenshot_Mono(window_handle, filename,diag, diagmsgs)
 *
 * Params:	window_handle - handle of the window to take a screen shot
 *			   filename - name of the file
 *			   diag    - if gt 0 it will print diagnostic information  
 *			   diagmsgs - array to return diagnostic messages 
 *
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine4(int, Screenshot_Mono, uintptr_t, hwnd, CSTRING, filename, int, diag, RexxArrayObject, diagmsgs) 
{
   char buffer[512]; 
   int rc;
   int rtnrc = 0;     
   
	using namespace Gdiplus;
	GdiplusStartupInput gdiplusStartupInput;
	ULONG_PTR gdiplusToken;
	GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);
	{     
      sprintf(buffer,"DIAG Screenshot_Mono Handle=%u Filename=%s",hwnd, filename);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      
      sprintf(buffer,"DIAG Screenshot_Mono Version: 31.0.0.4\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      
   	// See if output is a valid image type
   	TCHAR* extout = FindExtension(filename);
   	_tcslwr(extout);
   	int typeout = CxImage::GetTypeIdFromName(extout);
   	
   	if ( diag>0 )
      {
         sprintf(buffer,"DIAG Screenshot_Mono extout=%s typeout=%d",extout,typeout);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }      
       
   	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
      {
         sprintf(buffer,"ERROR: Screenshot_Mono Argument 2 is an invalid image type\n");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
   	}  
        
   	// Check if we have a valid window handle
      if ( !IsWindow((HWND)hwnd) )
      {
         sprintf(buffer,"ERROR: Screenshot_Mono Argument 1 is not a valid window handle\n");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;                         
      }  	
      
   	// Get the window size  
      RECT r;
      rc = -1;   
      GetWindowRect((HWND)hwnd, &r);
        
      int before_width =  r.right - r.left ;
      int before_height = r.bottom - r.top ;
      
      // Get the screen resolution 
      DEVMODE dev;
      int Enum_w, Enum_h;
      EnumDisplaySettings(nullptr, ENUM_CURRENT_SETTINGS, &dev);
      Enum_w = dev.dmPelsWidth;
      Enum_h = dev.dmPelsHeight;     
      sprintf(buffer,"DIAG Screenshot_Mono Before_width/height=%d %d Screen Resolution:%d %d",before_width,before_height,Enum_w,Enum_h);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      
      double scale = (double)Enum_w / (double)before_width; 
      sprintf(buffer, "%.*g", 3, scale);	
      scale = strtod(buffer, NULL);  	
      
      // Adjust the rectangle  added in V3.0.4.1
      sprintf(buffer,"DIAG Screenshot_Mono Before: scale=%f top=%d bottom=%d left=%d right=%d", scale, r.top,r.bottom,r.left,r.right);
      // printf(buffer); printf("\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
       
      // if the scale is not 1.0 then adjust the top and bottom 
      if (scale > 1.0)
      {
         int incr = (int)ceil(r.top * scale) -  r.top;  // ceil = roundup
         int top_before = r.top;
         r.top += incr;
         r.bottom += incr;
         sprintf(buffer,"DIAG Screenshot_Mono adjusted to scale=%f r.top by:%d was=%d now=%d",scale, incr, top_before, r.top);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      }   
       
      int width, height, x, y;
      x = r.left;
      y = r.top;
      
      width =  r.right - r.left ;
      height = r.bottom - r.top ;
      
      width =  (int)(width * scale);
      height = (int)(height * scale); 
      
      sprintf(buffer,"DIAG Screenshot_Mono After Width=%d Height=%d",width,height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));             
      
      sprintf(buffer,"DIAG Screenshot_Mono After: top=%d bottom=%d left=%d right=%d width=%d height=%d",r.top,r.bottom,r.left,r.right,width,height);
      // printf(buffer); printf("\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      
      if(diag > 0)   
      {
         sprintf(buffer,"DIAG Screenshot_Mono left=%d right=%d top=%d bottom=%d",r.left, r.right, r.top, r.bottom);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         sprintf(buffer,"DIAG Screenshot_Mono Screenshot_Mono X=%d Y=%d",x,y);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));             
      }  
    
      //  Check if the screen dimensions are valid
      if(width <= 0 || height <= 0) 
      {
         sprintf(buffer,"ERROR: Screenshot_Mono Invalid screen dimensions width=%d height=%d\n",width, height);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
      }       
      
      // Copy the screen
      HDC scrdc, memdc;
      HBITMAP membit;
      scrdc = ::GetDC(0);
      memdc = CreateCompatibleDC(scrdc);
      membit = CreateCompatibleBitmap(scrdc, width, height);
      HBITMAP hOldBitmap =(HBITMAP) SelectObject(memdc, membit);
      BitBlt(memdc, 0, 0, width, height, scrdc, x, y, SRCCOPY);	
      Gdiplus::Bitmap bitmap(membit, NULL);
     
      // Create a new image from the bitmap 
      CxImage *newImage = new CxImage();
      rc = newImage->CreateFromHBITMAP(membit);
      if (!rc)
      {
         sprintf(buffer,"ERROR: Screenshot_Mono Invalid rc from CreateFromHBITMAP RC=%d\n",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         rtnrc = 1;
      }  
      else
      {  
         int resx, resy;
         if(diag > 0)   
         {
            resx = newImage->GetXDPI(); // Get X resolution
            resy = newImage->GetYDPI(); // Get Y resolution
            sprintf(buffer,"DIAG Screenshot_Mono BEFORE:Xres=%d Yres=%d",resx, resy);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }  
         
         sprintf(buffer,"DIAG Screenshot_Mono Width=%d Height=%d",width,height);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         
         // set the Bpp to 1 (black and white)     
         rc = newImage->DecreaseBpp(1,0,0,0);               
         
         // Pass back the type, width, height, X resolution and Y resolution
         int b_resx, b_resy, b_width, b_height;
         b_resx = newImage->GetXDPI(); // Get X resolution
         b_resy = newImage->GetYDPI(); // Get Y resolution
         b_width = newImage->GetWidth();
         b_height = newImage->GetHeight();
         sprintf(buffer,"DIAG Screenshot_Mono TYPEOUT=%d XRES=%d YRES=%d ",typeout, b_resx, b_resy);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));       
         sprintf(buffer,"DIAG Screenshot_Mono WIDTH=%d HEIGHT=%d",b_width,b_height);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
         
         // Save the image
         rc = newImage->Save(filename,typeout);	
         if(!rc)
         {
           sprintf(buffer,"ERROR: Screenshot_Mono Invalid rc from Save Image RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           rtnrc = 1;
         }	    
      }	
   		
   	// Delete everything	
   	SelectObject(memdc, hOldBitmap);
   	DeleteObject(memdc);
   	DeleteObject(membit);
   	::ReleaseDC(0,scrdc);
    	  
      // Destroy the image if one exists
      if (newImage->IsValid()) 
      {
         rc = newImage->Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: Screenshot_Mono 1. Invalid rc from Destroy Image RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }	 
         
         sprintf(buffer,"DIAG Screenshot_Mono destroyed newimage rc=%d",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
      }  
      
   }   
   GdiplusShutdown(gdiplusToken);    
   
   if ( diag>0 ) 
   {
      sprintf(buffer,"DIAG Screenshot_Mono Returning RC=%d",rtnrc);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }     
	
   return rtnrc;                          
}

/**************************** Get_GE_Spin_Wheel **************************
 * Gets that the value of the SasPlanet Queue count is zero
 * 
 * Syntax:	rc = Get_GE_Spin_Wheel hwnd, sbimage, comm._diag, diagmsgs, fileout)
 *
 * Params:	hwnd- handle of the window to take a screen shot
 *          data - a directory to get and pass back information                        
 *			   diag    - if gt 0 it will print diagnostic information  
 *			   diagmsgs - array to return diagnostic messages  
 *			   fileout - optional filespec of BMP to store the status bar 
 *			   fileout2 - optional filespec of BMP to store the whole image
 *                           
 * Returns: if the positions array is empty then returns the count of good pixels
 *          else if its not empty returns 1 if all pixels match the positions array pixels, else 0  
 *          else -1  if error
 *********************************************************************/
RexxRoutine6(int, Get_GE_Spin_Wheel, uintptr_t,hwnd,  RexxObjectPtr,data, int,diag, 
                  RexxArrayObject,diagmsgs, OPTIONAL_CSTRING,fileout, OPTIONAL_CSTRING,fileout2)
{
   char buffer[512]; 
   int rc;
   int rtnrc = -1;  // Preset bad RC
   logical_t flag;
   
   // Check if we have a valid window handle
   if ( !IsWindow((HWND)hwnd) )
   {
      sprintf(buffer,"ERROR: Get_GE_Spin_Wheel Argument 1 is not a valid window handle\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return -1;                         
   }  
   
   RexxObjectPtr dataobj;
   RexxDirectoryObject data_dir;
   
   data_dir = (RexxDirectoryObject)data;
   flag = context->IsDirectory(data_dir);
   if ( flag == 0 )
   {
      sprintf(buffer,"ERROR: Get_GE_Spin_Wheel parameter 2 is not a directory!");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return -1;
   }
   
   // Get pointer to the rexx sb_position array
   RexxArrayObject  sb_position_rexx;
   sb_position_rexx = (RexxArrayObject)context->DirectoryAt(data_dir, "SB_POSITION"); 
   // Check its an array
   flag = context->IsArray(sb_position_rexx);
   if ( flag == 0 )
   {
      sprintf(buffer,"ERROR: Get_GE_Spin_Wheel SB_POSITION is not an array!");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return -1;
   } 
   
   // get the initial size of the array
   int initial_sz = context->ArrayItems(sb_position_rexx);
   sprintf(buffer,"DIAG: Get_GE_Spin_Wheel inital sb_position size=%d",initial_sz);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));

   double  *sb_position; // local array    
   if (initial_sz != 0)  // if not zero then allocate that size
   {
      sprintf(buffer,"DIAG: Get_GE_Spin_Wheel transferring rexx %d entries to local array",initial_sz);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      sb_position = (double *)malloc(initial_sz * sizeof(double)); // allocation local array
      
      // Transfer all the rexx data into the local array
      int n;
      double value;
      for(n=0 ; n<initial_sz ; n++)
      { 
         dataobj = context->ArrayAt(sb_position_rexx, n+1);
         flag = context->ObjectToDouble(dataobj, &value);
         sb_position[n] = value;      
      }  
   }
      
	using namespace Gdiplus;
	GdiplusStartupInput gdiplusStartupInput;
	ULONG_PTR gdiplusToken;
	GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);
	{   
      /* KEEP 
      // rc = SetProcessDpiAwarenessContext(DPI_AWARENESS_CONTEXT_SYSTEM_AWARE);  // 
      rc = SetProcessDpiAwarenessContext(DPI_AWARENESS_CONTEXT_SYSTEM_AWARE);
      int blnDpiAware = IsProcessDPIAware(); 
      UINT dpi = GetDpiForWindow((HWND)hwnd);
      sprintf(buffer,"DIAG Get_GE_Spin_Wheel Handle=%u blnDpiAware=%d dpi=%d",hwnd,blnDpiAware,dpi);
      // printf(buffer); printf("\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      */
   	
   	// Get the window size  
      RECT r;
      rc = -1;   
      GetWindowRect((HWND)hwnd, &r);  
      
      int before_width =  r.right - r.left ;
      int before_height = r.bottom - r.top ;
      
      // Get the screen resolution 
      DEVMODE dev;
      int Enum_w, Enum_h;
      EnumDisplaySettings(nullptr, ENUM_CURRENT_SETTINGS, &dev);
      Enum_w = dev.dmPelsWidth;
      Enum_h = dev.dmPelsHeight;     
      sprintf(buffer,"DIAG Get_GE_Spin_Wheel Before_width/height=%d %d Screen Resolution:%d %d",before_width,before_height,Enum_w,Enum_h);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      
      double scale = (double)Enum_w / (double)before_width;    
      sprintf(buffer, "%.*g", 3, scale);	
      scale = strtod(buffer, NULL);
      
      // Adjust the rectangle  added in V3.0.4.1
      sprintf(buffer,"DIAG Get_GE_Spin_Wheel Before: scale=%f top=%d bottom=%d left=%d right=%d", scale, r.top,r.bottom,r.left,r.right);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
       
      // if the scale is not 1.0 then adjust the top and bottom 
      if (scale > 1.0)
      {
         int incr = (int)ceil(r.top * scale) -  r.top;  // ceil = roundup
         int top_before = r.top;
         r.top += incr;
         r.bottom += incr;
         sprintf(buffer,"DIAG Get_GE_Spin_Wheel adjusted to scale=%f r.top by incr=%d was=%d now=%d",scale, incr, top_before, r.top);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      }   
       
      int width, height, x, y;
      x = r.left;
      y = r.top;
      
      width =  r.right - r.left ;
      height = r.bottom - r.top ;
      
      width =  (int)(width * scale);
      height = (int)(height * scale); 
      
      sprintf(buffer,"DIAG Get_GE_Spin_Wheel After Width=%d Height=%d",width,height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));             
      
      sprintf(buffer,"DIAG Get_GE_Spin_Wheel After: top=%d bottom=%d left=%d right=%d width=%d height=%d",r.top,r.bottom,r.left,r.right,width,height);
      // printf(buffer); printf("\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      
      /* KEEP
      // Get width and height to resolve screen resolution problem Works regardless of DPI setting  added in V19.0.1.5S
      WINDOWPLACEMENT wp;
      wp.length = sizeof(WINDOWPLACEMENT);
      if(GetWindowPlacement((HWND)hwnd, &wp))
      {
         sprintf(buffer,"DIAG Get_GE_Spin_Wheel GWP: showcmd:%d left=%d right=%d top=%d bottom=%d",wp.showCmd, wp.rcNormalPosition.left,wp.rcNormalPosition.right,wp.rcNormalPosition.top,wp.rcNormalPosition.bottom);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      }
      */
      
      if(diag > 0)   
      {   
         sprintf(buffer,"DIAG Get_GE_Spin_Wheel X=%d Y=%d",x,y);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));             
      }    
     
      //  Check if the screen dimensions are valid
      if(width <= 0 || height <= 0) 
      {
         sprintf(buffer,"ERROR: Get_GE_Spin_Wheel Invalid screen dimensions width=%d height=%d\n",width, height);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         goto returnit;    
      }       
          
      // Copy the screen
      HDC scrdc, memdc;
      HBITMAP membit;
      scrdc = ::GetDC(0);
      memdc = CreateCompatibleDC(scrdc);
      membit = CreateCompatibleBitmap(scrdc, width, height);      
      // membit = CreateCompatibleBitmap(scrdc, Enum_w, Enum_h);   
      HBITMAP hOldBitmap =(HBITMAP) SelectObject(memdc, membit);
      BitBlt(memdc, 0, 0, width, height, scrdc, x, y, SRCCOPY);	
      Gdiplus::Bitmap bitmap(membit, NULL);
     
      // Create a new image from the bitmap 
      CxImage image;
      // rc = miximage.CreateFromHBITMAP(membit);
      CxImage *imagecopy = new CxImage();
      
      rc = image.CreateFromHBITMAP(membit);
      if (!rc)
      {
         sprintf(buffer,"ERROR: Get_GE_Spin_Wheel Invalid rc from CreateFromHBITMAP RC=%d\n",rc);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         goto returnit;    
      }  
      else
      {  
         int resx, resy;
         if(diag > 0)   
         {
            resx = image.GetXDPI(); // Get X resolution
            resy = image.GetYDPI(); // Get Y resolution
            sprintf(buffer,"DIAG Get_GE_Spin_Wheel BEFORE:Xres=%d Yres=%d",resx, resy);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }   
         
         // If there is an output file2 write the uncropped image to a BMP 
         if ( argumentExists(6) )
         {
         	// See if output is a valid image type
         	TCHAR* extout = FindExtension(fileout2);
         	_tcslwr(extout);
         	int typeout = CxImage::GetTypeIdFromName(extout);
         	
         	if ( diag>0 )
            {
               sprintf(buffer,"DIAG Get_GE_Spin_Wheel fileout2=%s extout=%s typeout=%d",fileout2, extout,typeout);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            }      
             
         	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
            {
               sprintf(buffer,"ERROR: Get_GE_Spin_Wheel Argument 6 is an invalid image type\n");
               printf(buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               goto returnit;   
         	}  
            
            // Save the image   
            rc = image.Save(fileout2,typeout);	
            if(!rc)
            {
              sprintf(buffer,"ERROR: Get_GE_Spin_Wheel Invalid rc from Save fileout2 Image RC=%d",rc);
              printf(buffer);
              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              goto returnit;   
            }	                            
         }
                        	                              
         // Crop the image  into just the status bar
         float angle = 0;
         long newwidth = width;   // changed in V4.0.0.
         long newheight = 60;    // changed in V4.0.0.2
         long left = 0;  // changed in V4.0.0.2
         long top  = height - 60;  // changed in V4.0.0.2
            
         sprintf(buffer,"DIAG Get_GE_Spin_Wheel left=%d top=%d newwidth=%d newheight=%d ",left, top, newwidth, newheight);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));    
             
         rc = image.CropRotatedRectangle(left, top, newwidth, newheight, angle );
         if (!rc)
         {
            sprintf(buffer,"ERROR: Get_GE_Spin_Wheel Invalid rc from crop image RC=%d\n",rc);
            printf(buffer);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            goto returnit;   
         }  

         RGBQUAD color;
         double pixval;  
         RGBQUAD BkgndColor;
         BkgndColor.rgbRed = 0; 
         BkgndColor.rgbGreen = 0; 
         BkgndColor.rgbBlue = 0; 
         
         RGBQUAD WhiteColor;
         WhiteColor.rgbRed = 255; 
         WhiteColor.rgbGreen = 255; 
         WhiteColor.rgbBlue = 255; 
         WhiteColor.rgbReserved = 0;  

         // Create a copy of the cropped image status bar 
         imagecopy->Copy(image);

         bool spinwheel_stopped = true;
         // if the positions array was empty find all the good pixels 
         if (initial_sz == 0)   
         {
            // set the Bpp to 1 (black and white)     
            rc = imagecopy->DecreaseBpp(1,0,0,0); 

            // Make sure the two right hand cols are black
            for(int x=newwidth-2; x<newwidth; x++)
            {
         		for(int y=0; y<newheight; y++)
               {
                  imagecopy->SetPixelColor(x,y,BkgndColor,true); 
               } 
            }
 
            // find the right and left edge of the image
            spinwheel_stopped = false;
            sprintf(buffer,"DIAG Get_GE_Spin_Wheel 1. setting all good pixels ");
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            int first_x = 0;
            int last_x = 0;
            bool col_all_black;
            for(int x=newwidth-1; x>=0; x--) //  scan backwards  skip right haand pixel
            {
               col_all_black = true;  //  preset column all black   
         		for(int y=0; y<newheight; y++)
               {
            		color = imagecopy->GetPixelColor(x,y);
                  pixval = Get_Pixval(color);  
                  // if its white   16777215?      
                  if(pixval == 16777215 && last_x == 0)
                  { 
                     // Check the next pixel to make sure its also white
                     color = imagecopy->GetPixelColor(x-1,y);
                     pixval = Get_Pixval(color);  
                     // if its white   16777215?      
                     if(pixval == 16777215)
                     { 
                        col_all_black = false;
                        // If its the first x of the spinwheel remember it
                        if(first_x == 0) first_x=x;
                        col_all_black = false;   // indicate the column is not all black
                     }
                     else // change the pixel to black
                     {
                        imagecopy->SetPixelColor(x,y,BkgndColor,true); 
                     }
                  }              
         		} 
               // if we have set the first_x and this column is all black set the last_x to stop looking for spinwheel
               if(last_x == 0 && first_x != 0 && col_all_black) last_x = x+1;
            }
            sprintf(buffer,"DIAG Get_GE_Spin_Wheel first_x=%d last_x=%d",first_x,last_x);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));   

            // Now scan the rows and find the top and bottom of the spinwheel
            sprintf(buffer,"DIAG Get_GE_Spin_Wheel 2. setting all good pixels ");
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            int first_y = 0;
            int last_y = 0;
            bool row_all_black;
            for(int y=0; y<newheight; y++)
            {
               row_all_black = true;  //  preset row all black
         		for(int x=last_x; x<newwidth; x++)
               {
            		color = imagecopy->GetPixelColor(x,y);
                  pixval = Get_Pixval(color);  
                  // if not black its in the spinwheel if last_y was not set
                  if(pixval != 0 && last_y == 0)         
                  { 
                     row_all_black = false;
                     // If its the first y of the spinwheel remember it
                     if(first_y == 0) first_y=y;
                     row_all_black = false;   // indicate the row is not all black
                  }             
         		} 
               // if we have set the first_y and this row is all black set the last_y to set 
               if(last_y == 0 && first_y != 0 && row_all_black) last_y = y-1;
            }
            sprintf(buffer,"DIAG Get_GE_Spin_Wheel first_y=%d last_y=%d",first_y,last_y);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 

            // make all pixels to the left of the spinwheel black
            for(int y=0; y<newheight; y++)
            {
         		for(int x=0; x<last_x; x++)
               {
                  imagecopy->SetPixelColor(x,y,BkgndColor,true); 
               } 
            } 

            // make all pixels below the spinwheel black
            for(int y=last_y+1; y<newheight; y++)
            {
         		for(int x=0; x<newwidth; x++)
               {
                  imagecopy->SetPixelColor(x,y,BkgndColor,true); 
               } 
            } 

            // Check that we have a good stopped spinwheel 
            if(first_x > last_x && last_y > first_y)
            {
               // scan all the spinwheel pixels to make sure there no white pixels
               spinwheel_stopped = true;
               bool fnd = false;
               for(int y=first_y; y<=last_y; y++)
               {
                  for(int x=last_x; x<=first_x; x++)
                  {
                     color = imagecopy->GetPixelColor(x,y);
                     pixval = Get_Pixval(color);  
                     // if its not black it a valid spinwheel x,y check the real image 
                     if(pixval != 0)         
                     { 
                        color = image.GetPixelColor(x,y);
                        pixval = Get_Pixval(color);
                        if(pixval > 16776960)         
                        { 
                           sprintf(buffer,"WARNING Get_GE_Spin_Wheel WHITE x=%d y=%d pixval=%f",x,y,pixval);
                           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
                           spinwheel_stopped = false;
                           fnd = true;
                           break;
                        }
                     }
                  }
                  if (fnd) break;
               }
            }

            // If the spinwheel is not stopped set rtnc =0 
            if(!spinwheel_stopped) 
            {
               rtnrc = 0; // return rtnc=0
               sprintf(buffer,"DIAG Get_GE_Spin_Wheel the spinwheel is not stopped rtnrc=%d",rtnrc);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }
            else  // Its stopped get all the pixvals and store in the array
            {

               // find all the good pixels      
               sprintf(buffer,"DIAG Get_GE_Spin_Wheel scanning image ");
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
               int good_count = 0;
               for(int y=0; y<newheight; y++)
               {
            		for(int x=0; x<newwidth; x++)
                  {
                     bool set_black = true;
               		color = imagecopy->GetPixelColor(x,y);
                     pixval = Get_Pixval(color); 
                     // if its not black check the real image pixval
                     if(pixval != 0 ) 
                     {
                        color = image.GetPixelColor(x,y);
                        pixval = Get_Pixval(color);
                        // if(pixval > 6000000 && pixval < 7000000) 
                        if(pixval > 5000000 && pixval < 7000000)  // Changed to 5000000 was 6000000
                        {
                           sprintf(buffer, "%d",x);                                       
                           rc = context->ArrayAppendString(sb_position_rexx, buffer, strlen(buffer));
                           sprintf(buffer, "%d",y);                                       
                           rc = context->ArrayAppendString(sb_position_rexx, buffer, strlen(buffer));  
                           sprintf(buffer, "%f",pixval);                                       
                           rc = context->ArrayAppendString(sb_position_rexx, buffer, strlen(buffer)); 
                           set_black = false;
                           good_count += 1;
                        } 
                     }
            		} 
               }
               rtnrc = good_count; // return the count of good pixels
               sprintf(buffer,"DIAG Get_GE_Spin_Wheel after scanning image rtnrc=%d",rtnrc);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));   
            }
         }

         // else the positions array was not empty use the array to check the spinwheel modified in V3.0.7.3
         else  // check that all the entries in the buffer match the pixvals
         {
            sprintf(buffer,"DIAG Get_GE_Spin_Wheel checking spinwheeel total entries=%d",initial_sz);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            rtnrc = 1;  // assume all match  
            double pixval_ref;
            for(int z=0; z<initial_sz; z+=3)
            {
               // get 3 entries
               x =  (int)sb_position[z];
               y =  (int)sb_position[z+1];
               pixval_ref =  sb_position[z+2];
               color = image.GetPixelColor(x,y);
               pixval = Get_Pixval(color); 
               bool isValid = false;
               // Check if its a close match with initial modified in V3.0.7.2
               if (abs(pixval - pixval_ref) < 1200000 ) // changed was 1000000 
               {
                  rtnrc = 1; 
                  isValid = true;
               }
               
               // If we didn't find a valid match print it and set rtnc=0 and return
               if (isValid == false)
               {
                  sprintf(buffer,"DIAG Get_GE_Spin_Wheel Mismatch at %d %d Initial=%f Now=%f",x,y,pixval_ref,pixval);
                  rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                  rtnrc = 0; 
                  break;
               }                   
            }
            sprintf(buffer,"DIAG Get_GE_Spin_Wheel after checking image rtnrc=%d",rtnrc);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         } 

         // If there is an output file write the cropped status bar to a BMP 
         if ( argumentExists(5) )
         {
         	// See if output is a valid image type
         	TCHAR* extout = FindExtension(fileout);
         	_tcslwr(extout);
         	int typeout = CxImage::GetTypeIdFromName(extout);
         	
         	if ( diag>0 )
            {
               sprintf(buffer,"DIAG Get_GE_Spin_Wheel extout=%s typeout=%d",extout,typeout);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            }      
             
         	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
            {
               sprintf(buffer,"ERROR: Get_GE_Spin_Wheel Argument 4 is an invalid image type\n");
               printf(buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               goto returnit;   
         	}  
            
            // Save the image copy   
            rc = imagecopy->Save(fileout,typeout);	
            if(!rc)
            {
              sprintf(buffer,"ERROR: Get_GE_Spin_Wheel Invalid rc from Save Image RC=%d",rc);
              printf(buffer);
              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              rtnrc = -1;
              goto returnit;   
            }	 
         }               
      } 
   		
   	// Delete everything	
   	SelectObject(memdc, hOldBitmap);
   	DeleteObject(memdc);
   	DeleteObject(membit);
   	::ReleaseDC(0,scrdc); 	
         	  
      // Destroy the image if one exists
      if (image.IsValid()) 
      {
         rc = image.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: Get_GE_Spin_Wheel  1. Invalid rc from Destroy Image RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }	 
         sprintf(buffer,"DIAG Get_GE_Spin_Wheel destroyed newimage rc=%d",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
      }  
         	  
      // Destroy the imagecopy if one exists
      if (imagecopy->IsValid()) 
      {
         rc = imagecopy->Destroy(); // Destroy the imagecopy
         if(!rc)
         {
           sprintf(buffer,"Error: Get_GE_Spin_Wheel  1. Invalid rc from Destroy Image RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }	 
         sprintf(buffer,"DIAG Get_GE_Spin_Wheel destroyed imagecopy rc=%d",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
      }  

   }   
   GdiplusShutdown(gdiplusToken);  
   
   // free the local array
   if (sb_position != NULL && initial_sz > 0)
   {
     free(sb_position);
   } 

   sprintf(buffer,"DIAG Get_GE_Spin_Wheel Returning RC=%d\n",rtnrc);
   // printf(buffer);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        
   returnit:     
       	
   return rtnrc;          
}
/**************************** ScreenShot_tif **************************
 * Takes a screen shot of the window handle that is passed.
 * and creates a .tif
 * Syntax:	rc = Screenshot_tif(window_handle, filename, , colors, diffusion, diag, diagmsgs, crop_amt)
 *
 * Params:	window_handle - handle of the window to take a screen shot
 *			   filename - name of the file
 *			   colors - number of colors in output 
 *			   diffusion - error diffusion, either 0=OFF or 1=ON 
 *			   diag    - if gt 0 it will print diagnostic information  
 *			   diagmsgs - array to return diagnostic messages 
 *			   crop_amt (optional)   the amount to crop the image
 *
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine7(int, Screenshot_tif, uintptr_t, hwnd, CSTRING, filename, int, colors, int, diffusion, int, diag, 
                  RexxArrayObject, diagmsgs,OPTIONAL_int, crop_amt)
{
   char buffer[512]; 
   int rc;
   int rtnrc = 0;
   
	using namespace Gdiplus;
	GdiplusStartupInput gdiplusStartupInput;
	ULONG_PTR gdiplusToken;
	GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);
	{   
	
	   // int colors = 127;
      // int colors = 64;
      int numcolors;   

      sprintf(buffer,"DIAG Screenshot_tif Handle=%u Filename=%s Colors=%d Diffusion=%d",hwnd, filename, colors,diffusion);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      
      sprintf(buffer,"DIAG Screenshot_tif Version: 9.5.9\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      
   	// See if output is a valid image type
   	TCHAR* extout = FindExtension(filename);
   	_tcslwr(extout);
   	int typeout = CxImage::GetTypeIdFromName(extout);
   	
   	if ( diag>0 )
      {
         sprintf(buffer,"DIAG Screenshot_tif extout=%s typeout=%d",extout,typeout);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }      
       
   	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
      {
         sprintf(buffer,"ERROR: Screenshot_tif Argument 2 is an invalid image type\n");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
   	}  
        
   	// Check if we have a valid window handle
      if ( !IsWindow((HWND)hwnd) )
      {
         sprintf(buffer,"ERROR: Screenshot_tif Argument 1 is not a valid window handle\n");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;                         
      }  	
      
   	// Get the window size  
      RECT r;
      rc = -1;   
      GetWindowRect((HWND)hwnd, &r); 
      
      int before_width =  r.right - r.left ;
      int before_height = r.bottom - r.top ;
      
      // Get the screen resolution 
      DEVMODE dev;
      int Enum_w, Enum_h;
      EnumDisplaySettings(nullptr, ENUM_CURRENT_SETTINGS, &dev);
      Enum_w = dev.dmPelsWidth;
      Enum_h = dev.dmPelsHeight;     
      sprintf(buffer,"DIAG Screenshot_tif Before_width/height=%d %d Screen Resolution:%d %d",before_width,before_height,Enum_w,Enum_h);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      
      double scale = (double)Enum_w / (double)before_width; 
      sprintf(buffer, "%.*g", 3, scale);	
      scale = strtod(buffer, NULL);	
      
      // Adjust the rectangle added in V3.0.4.1
      sprintf(buffer,"DIAG Screenshot_tif Before: scale=%f top=%d bottom=%d left=%d right=%d", scale, r.top,r.bottom,r.left,r.right);
      // printf(buffer); printf("\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
       
      // if the scale is not 1.0 then adjust the top and bottom 
      if (scale > 1.0)
      {
         int incr = (int)ceil(r.top * scale) -  r.top;  // ceil = roundup
         int top_before = r.top;
         r.top += incr;
         r.bottom += incr;
         sprintf(buffer,"DIAG Screenshot_tif adjusted to scale=%f r.top by:%d was=%d now=%d",scale, incr, top_before, r.top);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      }   
       
      int width, height, x, y;
      x = r.left;
      y = r.top;
      
      width =  r.right - r.left ;
      height = r.bottom - r.top ;
      
      width =  (int)(width * scale);
      height = (int)(height * scale); 
      
      sprintf(buffer,"DIAG Screenshot_tif After Width=%d Height=%d",width,height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));             
      
      sprintf(buffer,"DIAG Screenshot_tif After: top=%d bottom=%d left=%d right=%d width=%d height=%d",r.top,r.bottom,r.left,r.right,width,height);
      // printf(buffer); printf("\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      
      ULONG inbpp; 
	   ULONG outbpp; 
      RGBQUAD * ppal;
      
      if(diag > 0)   
      {
         sprintf(buffer,"DIAG Screenshot_tif left=%d right=%d top=%d bottom=%d",r.left, r.right, r.top, r.bottom);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         sprintf(buffer,"DIAG Screenshot Screenshot X=%d Y=%d",x,y);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));             
      }  
    
      //  Check if the screen dimensions are valid
      if(width <= 0 || height <= 0) 
      {
         sprintf(buffer,"ERROR: Screenshot_tif Invalid screen dimensions width=%d height=%d\n",width, height);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
      }       
      
      // Copy the screen
      HDC scrdc, memdc;
      HBITMAP membit;
      scrdc = ::GetDC(0);
      memdc = CreateCompatibleDC(scrdc);
      membit = CreateCompatibleBitmap(scrdc, width, height);
      HBITMAP hOldBitmap =(HBITMAP) SelectObject(memdc, membit);
      BitBlt(memdc, 0, 0, width, height, scrdc, x, y, SRCCOPY);	
      Gdiplus::Bitmap bitmap(membit, NULL);
     
      // Create a new image from the bitmap 
      CxImage *newImage = new CxImage();
      rc = newImage->CreateFromHBITMAP(membit);
      if (!rc)
      {
         sprintf(buffer,"ERROR: Screenshot_tif Invalid rc from CreateFromHBITMAP RC=%d\n",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         rtnrc = 1;
      }  
      else
      {  
         int resx, resy;
         if(diag > 0)   
         {
            resx = newImage->GetXDPI(); // Get X resolution
            resy = newImage->GetYDPI(); // Get Y resolution
            sprintf(buffer,"DIAG Screenshot_tif BEFORE:Xres=%d Yres=%d",resx, resy);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }  
         
         sprintf(buffer,"DIAG Screenshot_tif Width=%d Height=%d",width,height);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 	             
         
         // If there is a crop_amt if there is a crop amount then crop the image
         if ( argumentExists(7) )
         {    
            if (crop_amt > 0)
            {
               int crop_width, crop_height;
               // Crop the image  to remove the status bar added in V16.0.0.17
               float angle = 0;
               long newheight = height - crop_amt - 4;  // subtracted the top crop of 4
               sprintf(buffer,"DIAG Screenshot_tif crop_amt=%d newheight=%d ",crop_amt,newheight);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));    
                       
               // rc = newImage->CropRotatedRectangle( 0, 0, width, newheight, angle );
               rc = newImage->CropRotatedRectangle( 0, 4, width, newheight, angle );  // added top crop
               if (!rc)
               {
                  sprintf(buffer,"ERROR: Screenshot_tif Invalid rc from crop image RC=%d\n",rc);
                  rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                  return 1;
               }
               height = newheight;
               crop_width = newImage->GetWidth();
               crop_height = newImage->GetHeight();
               sprintf(buffer,"DIAG Screenshot crop_width=%d crop_height=%d",crop_width,crop_height);	   
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));    
            }      
         }
                                   
         // If the output is NOT a JPG   set resolution to 96ppi, decrease the bpp to 8 and numcolors                 
         if (typeout != CXIMAGE_FORMAT_JPG)
         {
            inbpp = newImage->GetBpp(); 
         	
            CQuantizer m(16777216,8);
	         m.ProcessImage(newImage->GetDIB());	 
            numcolors = m.GetColorCount();
            if(diag > 0)   
            {
               sprintf(buffer,"DIAG Screenshot_tif BEFORE: numcolors=%d",numcolors);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            } 
         	
            newImage->SetXDPI(96);
            newImage->SetYDPI(96);
            // CQuantizer q(colors,(colors>16?7:8));   REMOVE
            CQuantizer q(colors,4);     //  was 8
            q.ProcessImage(newImage->GetDIB());           
            ppal=(RGBQUAD*)calloc(256*sizeof(RGBQUAD),1);
            q.SetColorTable(ppal);
         
            bool error_diffusion = (diffusion == 0)	? FALSE : TRUE;       
            rc = newImage->DecreaseBpp(8,error_diffusion,ppal,colors); //  DIFFUSION ADDED IN V9.5.2   
            outbpp = newImage->GetBpp(); 
            numcolors = q.GetColorCount();
            
            if(diag > 0)   
            {
               sprintf(buffer,"DIAG Screenshot_tif DecreaseBpp RC=%d outbpp=%d AFTER: numcolors=%d",rc,outbpp,numcolors);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));     
            }            
         }
         
         // If the output is a tif then decrease the bpp to 8 and reduce the numcolors to 127
      	if (typeout == 6 ) 
      	{
            int colors = 127;
            int numcolors;
            // image.SetXDPI(96);
            // image.SetYDPI(96);
            // image.Light(10); removed in V9.5.9
            int height = newImage->GetHeight();
      	   int width = newImage->GetWidth();
      	   long x, y;
      	   int idx;  
      
            uint64_t	histogram[256] = {0};
                 
            /* Scan through entire image computing a histogram of colormap indices */
            for(y=0; y<height; y++)
            {
         		for(x=0; x<width; x++)
               {
            	  idx = newImage->GetPixelIndex(x,y);
            		histogram[idx++]++;
               }
            }
            /* Count down from highest colormap index for the first non-zero value.	*/
            /* This will indicate the maximum number of colors.						*/
            for (x = 255; x >= 0; x--)
            {
            	if (histogram[x] != 0)
            	{
            		numcolors = x + 1;		/* count of colors includes index 0 */
            		break;
            	}
            }	  
      	  
      	  /*
      	  CQuantizer q(16777216,8);
      	  q.ProcessImage(image.GetDIB());	  
      	  // if inbpp is 8 and the number of colors is already less or equal t0 127 don't quantize 
      	  numcolors = q.GetColorCount();
      	  */
      	  if ( diag>0 ) 
           {
                sprintf(buffer,"DIAG Screenshot_tif in_numcolors=%d inbpp=%d \n",numcolors,inbpp);
                rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           }
      	  if (numcolors <= colors && inbpp == 8)	 
           {
              if ( diag>0 ) 
              {
                sprintf(buffer,"DIAG Screenshot_tif quantizer bypassed\n");
                rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              }
           } 
           else
           {     
               if (inbpp == 8 || inbpp == 24)   // new in V10.0.0  added inbpp== 24 in V10.0.3
               {
              	  CQuantizer q(colors,(colors>16?7:8));
         	     q.ProcessImage(newImage->GetDIB()); 
         	     ppal=(RGBQUAD*)calloc(256*sizeof(RGBQUAD),1);
         	     q.SetColorTable(ppal);
         	     // rc = image.DecreaseBpp(8,0,ppal,colors);
         	     rc = newImage->DecreaseBpp(8,1,ppal,colors); // V600 Set error diffusion=1
         	     outbpp = newImage->GetBpp(); 
         	     if (!rc)
         	     {
                     sprintf(buffer,"ERROR: Screenshot_tif DecreaseBpp RC=%d outbpp=%d\n",rc,outbpp);
                     rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                     return 1;	
                  } 
            	  numcolors = q.GetColorCount();
            	  if ( diag>0 ) 
                 {
                      sprintf(buffer,"DIAG Screenshot_tif out_numcolors=%d outbpp=%d \n",numcolors,outbpp);
                      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                 } 
               }        
           } 
      	}           
         
         // Pass back the type, width, height, X resolution and Y resolution
         int b_resx, b_resy, b_width, b_height;
         b_resx = newImage->GetXDPI(); // Get X resolution
         b_resy = newImage->GetYDPI(); // Get Y resolution
         b_width = newImage->GetWidth();
         b_height = newImage->GetHeight();
         sprintf(buffer,"DIAG Screenshot TYPEOUT=%d XRES=%d YRES=%d ",typeout, b_resx, b_resy);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));       
         sprintf(buffer,"DIAG Screenshot WIDTH=%d HEIGHT=%d NUMCOLORS=%d",b_width,b_height,numcolors);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         
         // Save the image
         rc = newImage->Save(filename,typeout);	
         if(!rc)
         {
           sprintf(buffer,"ERROR: Screenshot_tif Invalid rc from Save Image RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           rtnrc = 1;
         }	    
      }	
   		
   	// Delete everything	
   	SelectObject(memdc, hOldBitmap);
   	DeleteObject(memdc);
   	DeleteObject(membit);
   	::ReleaseDC(0,scrdc);
    	  
      // Destroy the image if one exists
      if (newImage->IsValid()) 
      {
         rc = newImage->Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: Screenshot_tif 2. Invalid rc from Destroy Image RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }	                         
         sprintf(buffer,"DIAG Screenshot destroyed newimage rc=%d",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
      }  
      
   }   
   GdiplusShutdown(gdiplusToken);    
   
   if ( diag>0 ) 
   {
      sprintf(buffer,"DIAG Screenshot_tif Returning RC=%d",rtnrc);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }      
	
   return rtnrc;
}

/**************************** CountColors **************************
 * Counts the number of colors in the image passed and optionall pass
 * back the value of the SpinWheel
 * Syntax:	rc = CountColors(hwnd, comm._diag, diagmsgs, left_offset, spin_filename)
 *
 * Params:	window_handle - handle of the window to take a screen shot
 *			   diag    - if gt 0 it will print diagnostic information  
 *			   diagmsgs - array to return diagnostic messages 
 *			   Optional left_ offset offset to left side of spin wheel, V7=28, V6=16 
 *			   Optional an output filename to create a bmp for the spin cursor 
 *
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine5(int, CountColors, uintptr_t,hwnd, int,diag, RexxArrayObject,diagmsgs, 
                     OPTIONAL_int,left_offset, OPTIONAL_CSTRING,fileout)
{
   char buffer[512]; 
   int rc;
   int rtnrc = 0;
   
	using namespace Gdiplus;
	GdiplusStartupInput gdiplusStartupInput;
	ULONG_PTR gdiplusToken;
	GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);
	{   
      INT numcolors;   

      sprintf(buffer,"DIAG CountColors Handle=%u\n",hwnd);
      // printf(buffer); printf("\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        
   	// Check if we have a valid window handle
      if ( !IsWindow((HWND)hwnd) )
      {
         sprintf(buffer,"ERROR: CountColors Argument 1 is not a valid window handle\n");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;                         
      }  	
   	
   	// Get the window size  
      RECT r;
      rc = -1;   
      GetWindowRect((HWND)hwnd, &r); 
      
      int before_width =  r.right - r.left ;
      int before_height = r.bottom - r.top ;
      
      // Get the screen resolution 
      DEVMODE dev;
      int Enum_w, Enum_h;
      EnumDisplaySettings(nullptr, ENUM_CURRENT_SETTINGS, &dev);
      Enum_w = dev.dmPelsWidth;
      Enum_h = dev.dmPelsHeight;     
      sprintf(buffer,"DIAG CountColors Before_width/height=%d %d Screen Resolution:%d %d",before_width,before_height,Enum_w,Enum_h);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      
      double scale = (double)Enum_w / (double)before_width;
      sprintf(buffer, "%.*g", 3, scale);	
      scale = strtod(buffer, NULL); 
      
      // Adjust the rectangle added in V3.0.4.1
      sprintf(buffer,"DIAG CountColors Before: scale=%f top=%d bottom=%d left=%d right=%d",scale, r.top,r.bottom,r.left,r.right);
      // printf(buffer); printf("\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
       
      // if the scale is not 1.0 then adjust the top and bottom 
      if (scale > 1.0)
      {
         int incr = (int)ceil(r.top * scale) -  r.top;  // ceil = roundup
         int top_before = r.top;
         r.top += incr;
         r.bottom += incr;
         sprintf(buffer,"DIAG CountColors adjusted to scale=%f r.top by:%d was=%d now=%d",scale, incr, top_before, r.top);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      }   
       
      int width, height, x, y;
      x = r.left;
      y = r.top;
      
      width =  r.right - r.left ;
      height = r.bottom - r.top ;
      
      width =  (int)(width * scale);
      height = (int)(height * scale); 
      
      sprintf(buffer,"DIAG CountColors After Width=%d Height=%d",width,height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));             
      
      sprintf(buffer,"DIAG CountColors After: top=%d bottom=%d left=%d right=%d width=%d height=%d",r.top,r.bottom,r.left,r.right,width,height);
      // printf(buffer); printf("\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              
      if(diag > 0)   
      {
         sprintf(buffer,"DIAG CountColors left=%d right=%d top=%d bottom=%d",r.left, r.right, r.top, r.bottom);
         // printf(buffer); printf("\n");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         sprintf(buffer,"DIAG CountColors X=%d Y=%d",x,y);
         // printf(buffer); printf("\n");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));             
      }  
    
      //  Check if the screen dimensions are valid
      if(width <= 0 || height <= 0) 
      {
         sprintf(buffer,"ERROR: CountColors Invalid screen dimensions width=%d height=%d\n",width, height);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
      }       
      
      // Copy the screen
      HDC scrdc, memdc;
      HBITMAP membit;
      scrdc = ::GetDC(0);
      memdc = CreateCompatibleDC(scrdc);
      membit = CreateCompatibleBitmap(scrdc, width, height);
      HBITMAP hOldBitmap =(HBITMAP) SelectObject(memdc, membit);
      BitBlt(memdc, 0, 0, width, height, scrdc, x, y, SRCCOPY);	
      Gdiplus::Bitmap bitmap(membit, NULL);
     
      // Create a new image from the bitmap 
      CxImage *image = new CxImage();
      
      rc = image->CreateFromHBITMAP(membit);
      if (!rc)
      {
         sprintf(buffer,"ERROR: CountColors Invalid rc from CreateFromHBITMAP RC=%d\n",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         rtnrc = 1;
      }  
      else
      {  
         int resx, resy;
         if(diag > 0)   
         {
            resx = image->GetXDPI(); // Get X resolution
            resy = image->GetYDPI(); // Get Y resolution
            sprintf(buffer,"DIAG CountColors BEFORE:Xres=%d Yres=%d",resx, resy);
            // printf(buffer);printf("\n");
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }   
                         	
         CQuantizer m(16777216,8);
         m.ProcessImage(image->GetDIB());	 
         numcolors = m.GetColorCount();  
         sprintf(buffer,"DIAG CountColors BEFORE: numcolors=%d",numcolors);
         // printf(buffer);printf("\n");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         
         RGBQUAD color;
         // BYTE red,green,blue;                                             
         
         // If there is a left_offset crop the image by the cropamt
         if ( argumentExists(4) )
         {                 
            // Crop the image
            float angle = 0;
            long newwidth = 16;
            long newheight = 16;
            long left = width - left_offset;  
            long top = height - 20;
            
            if(diag > 0)   
            {
               sprintf(buffer,"DIAG CountColors left_offset=%d left=%d top=%d newwidth=%d newheight=%d ",left_offset, left, top, newwidth, newheight);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            }     
                    
            rc = image->CropRotatedRectangle( left, top, newwidth, newheight, angle );
            if (!rc)
            {
               sprintf(buffer,"ERROR: CountColors Invalid rc from crop image RC=%d\n",rc);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               return 1;
            }	 
                                       
         	int countbkg=0;
         	int countfor=0;
         	int height = image->GetHeight();
      	   int width = image->GetWidth();
      	   RGBQUAD BkgndColor;
            BkgndColor.rgbRed = 0; 
            BkgndColor.rgbGreen = 0; 
            BkgndColor.rgbBlue = 0; 
            BkgndColor.rgbReserved = 1; 
            
            // Look at 8 specific pixels and check their color
            ULONG pixval_1;
            ULONG pixval_2;
            ULONG pixval_3;
            ULONG pixval_4;
            ULONG pixval_5;
            ULONG pixval_6;
            ULONG pixval_7;
            ULONG pixval_8;
            ULONG pixval_9;
         
            color = image->GetPixelColor(1,7);
            pixval_1 = Get_Pixval(color);    // added in V17.0.0.19
            color = image->GetPixelColor(7,1);
            pixval_2 = Get_Pixval(color);         // added in V17.0.0.19
            color = image->GetPixelColor(3,3);
            pixval_3 = Get_Pixval(color);      // added in V17.0.0.19 
            color = image->GetPixelColor(8,13);
            pixval_4 = Get_Pixval(color);         // added in V17.0.0.19
            color = image->GetPixelColor(14,7);
            pixval_5 = Get_Pixval(color);          // added in V17.0.0.19
            color = image->GetPixelColor(13,4);
            pixval_6= Get_Pixval(color);         // added in V17.0.0.19                                
            color = image->GetPixelColor(13,11);
            pixval_7 = Get_Pixval(color);         // added in V17.0.0.19
            color = image->GetPixelColor(2,10);
            pixval_8 = Get_Pixval(color);         // added in V17.0.0.19
            color = image->GetPixelColor(8,7); // center
            pixval_9 = Get_Pixval(color);           // added in V17.0.0.19
      		                                                
            sprintf(buffer,"DIAG pixvals: %d %d %d %d %d %d %d %d %d ",pixval_1,pixval_2,pixval_3,pixval_4,pixval_5,pixval_6,pixval_7,pixval_8,pixval_9);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));              
             
            // If there is an output file write spin cursor to a BMP 
            if ( argumentExists(5) )
            {
            	// See if output is a valid image type
            	TCHAR* extout = FindExtension(fileout);
            	_tcslwr(extout);
            	int typeout = CxImage::GetTypeIdFromName(extout);
            	
            	if ( diag>0 )
               {
                  sprintf(buffer,"DIAG CountColors extout=%s typeout=%d",extout,typeout);
                  rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               }      
                
            	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
               {
                  sprintf(buffer,"ERROR: CountColors Argument 4 is an invalid image type\n");
                  rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                  return 1;
            	}  
               
               // Save the image
               rc = image->Save(fileout,typeout);	
               if(!rc)
               {
                 sprintf(buffer,"ERROR: CountColors Invalid rc from Save Image RC=%d\n ",rc);
                 rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                 return 1;
               }	                            
            }  
         } 
      }    	
   		
   	// Delete everything	
   	SelectObject(memdc, hOldBitmap);
   	DeleteObject(memdc);
   	DeleteObject(membit);
   	::ReleaseDC(0,scrdc);
    	  
      // Destroy the image if one exists
      if (image->IsValid()) 
      {
         rc = image->Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: CountColors 1. Invalid rc from Destroy Image RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }	
         sprintf(buffer,"DIAG CountColors destroyed newimage rc=%d",rc);
         // printf(buffer);printf("\n");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
      }  
      
   }   
   GdiplusShutdown(gdiplusToken);    
   
   if ( diag>0 ) 
   {
      sprintf(buffer,"DIAG CountColors Returning RC=%d",rtnrc);
      // printf(buffer);printf("\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }      
	
   return rtnrc;
}
/**************************** CheckQueueCount0 **************************
 * Gets that the value of the SasPlanet Queue count is zero
 * 
 * Syntax:	rc = CheckQueueCount0 hwnd, queueImage0, comm._diag, diagmsgs, fileout)
 *
 * Params:	hwnd- handle of the window to take a screen shot
 *          data - a directory to get and pass back information               
 *			   diag    - if gt 0 it will print diagnostic information  
 *			   diagmsgs - array to return diagnostic messages  
 *			   fileout - optional filespec of BMP to store the screen image
 *                           
 * Returns: if the positions array is empty then returns the count of good pixels
 *          else if its not empty returns 1 if all pixels match the positions array pixels, else 0  
 *          else -1 if error
 *********************************************************************/
RexxRoutine6(int, CheckQueueCount0, uintptr_t,hwnd,  RexxObjectPtr,data, int,diag, RexxArrayObject,diagmsgs, OPTIONAL_CSTRING,fileout, OPTIONAL_CSTRING,fileout2)                  
{
   char buffer[512]; 
   int rc;
   int rtnrc = -1;  // Preset bad RC
   int *local_queueImage0 = NULL;  // local  array  added in V16.0.0.17
   logical_t flag;
   double * flds = NULL;

	using namespace Gdiplus;
	GdiplusStartupInput gdiplusStartupInput;
	ULONG_PTR gdiplusToken;
	GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);
	{    
      sprintf(buffer,"DIAG CheckQueueCount0 Handle=%u\n",hwnd);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        
   	// Check if we have a valid window handle
      if ( !IsWindow((HWND)hwnd) )
      {
         sprintf(buffer,"ERROR: CheckQueueCount0 Argument 1 is not a valid window handle\n");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return -1;                    
      }  

      RexxDirectoryObject data_dir;  
      data_dir = (RexxDirectoryObject)data;
      flag = context->IsDirectory(data_dir);
      if ( flag == 0 )
      {
         sprintf(buffer,"ERROR: CheckQueueCount0 parameter 2 is not a directory!");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return -1;
      }

   	// Get the window size  
      RECT r;
      rc = -1;   
      GetWindowRect((HWND)hwnd, &r); 
      
      int before_width =  r.right - r.left ;
      int before_height = r.bottom - r.top ;
      
      // Get the screen resolution 
      DEVMODE dev;
      int Enum_w, Enum_h;
      EnumDisplaySettings(nullptr, ENUM_CURRENT_SETTINGS, &dev);
      Enum_w = dev.dmPelsWidth;
      Enum_h = dev.dmPelsHeight;     
      sprintf(buffer,"DIAG CheckQueueCount0 Before_width/height=%d %d Screen Resolution:%d %d",before_width,before_height,Enum_w,Enum_h);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      
      double scale = (double)Enum_w / (double)before_width;
      sprintf(buffer, "%.*g", 3, scale);	
      scale = strtod(buffer, NULL); 	
      
      // Adjust the rectangle added in V3.0.4.1
      sprintf(buffer,"DIAG CheckQueueCount0 Before: scale=%f top=%d bottom=%d left=%d right=%d",scale, r.top,r.bottom,r.left,r.right);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
       
      // if the scale is not 1.0 then adjust the top and bottom 
      if (scale > 1.0)
      {
         int incr = (int)ceil(r.top * scale) -  r.top;  // ceil = roundup
         int top_before = r.top;
         r.top += incr;
         r.bottom += incr;
         sprintf(buffer,"DIAG CheckQueueCount0 adjusted to scale=%f r.top by:%d was=%d now=%d",scale, incr, top_before, r.top);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      }   
       
      int width, height, x, y;
      x = r.left;
      y = r.top;
      
      width =  r.right - r.left ;
      height = r.bottom - r.top ;
      
      width =  (int)(width * scale);
      height = (int)(height * scale); 
      
      sprintf(buffer,"DIAG CheckQueueCount0 After Width=%d Height=%d",width,height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      
      // printf("DEBUG CheckQueueCount0 After Width=%d Height=%d\n",width,height);            
      
      sprintf(buffer,"DIAG CheckQueueCount0 After: top=%d bottom=%d left=%d right=%d width=%d height=%d",r.top,r.bottom,r.left,r.right,width,height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      
      //  Check if the screen dimensions are valid
      if(width <= 0 || height <= 0) 
      {
         sprintf(buffer,"ERROR: CheckQueueCount0 Invalid screen dimensions width=%d height=%d\n",width, height);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         goto returnit;    
      }       
          
      // Copy the screen
      HDC scrdc, memdc;
      HBITMAP membit;
      scrdc = ::GetDC(0);
      memdc = CreateCompatibleDC(scrdc);
      membit = CreateCompatibleBitmap(scrdc, width, height);
      HBITMAP hOldBitmap =(HBITMAP) SelectObject(memdc, membit);
      BitBlt(memdc, 0, 0, width, height, scrdc, x, y, SRCCOPY);	
      Gdiplus::Bitmap bitmap(membit, NULL);
     
      // Create a new image from the bitmap 
      CxImage *image = new CxImage();
      
      rc = image->CreateFromHBITMAP(membit);
      if (!rc)
      {
         sprintf(buffer,"ERROR: CheckQueueCount0 Invalid rc from CreateFromHBITMAP RC=%d\n",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         goto returnit;    
      }  
      else
      {                                       
         RGBQUAD color;
         
         double pixval,pixval2; 
          
         // White
         RGBQUAD WhitekColor;
         WhitekColor.rgbRed = 255; 
         WhitekColor.rgbGreen = 255; 
         WhitekColor.rgbBlue = 255;
        
         // Black
         RGBQUAD BlackColor;
         BlackColor.rgbRed = 0; 
         BlackColor.rgbGreen = 0; 
         BlackColor.rgbBlue = 0; 
         
         if ( argumentExists(5) )
         {
         	// See if output is a valid image type
         	TCHAR* extout = FindExtension(fileout);
         	_tcslwr(extout);
         	int typeout = CxImage::GetTypeIdFromName(extout);
         	
            sprintf(buffer,"DIAG CheckQueueCount0 fileout=%s extout=%s typeout=%d",fileout, extout,typeout);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                   
         	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
            {
               sprintf(buffer,"ERROR: CheckQueueCount0 Argument 5 is an invalid image type\n");
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               goto returnit;   
         	}  
            
            // Save the image   
            rc = image->Save(fileout,typeout);	
            if(!rc)
            {
              sprintf(buffer,"ERROR: CheckQueueCount0 Invalid rc from Save fileout Image RC=%d",rc);
              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              goto returnit;   
            }	                            
         } 	
         
         image->GrayScale();
           
         // get the maximum status bar height   add in V9.0.3.1
         int Max_statusbar_height = (int)(16 * scale); 
          
         // Crop the image status bar up to Max_statusbar_height
         float angle = 0;
         long newwidth = width;   
         long newheight = Max_statusbar_height;    
         long left = 0;  
         long top  = height - Max_statusbar_height; 
         sprintf(buffer,"DIAG CheckQueueCount0 cropping image Max_statusbar_height=%d left=%d top=%d newwidth=%d newheight=%d ",Max_statusbar_height,left, top, newwidth, newheight);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));      
         rc = image->CropRotatedRectangle(left, top, newwidth, newheight, angle );
         if (!rc)
         {
            sprintf(buffer,"ERROR: CheckQueueCount0 Invalid rc from CropRotatedRectangle RC=%d\n",rc);
            printf(buffer);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            goto returnit;   
         } 
         
         // Scan the image and change any pixels that are 1,1,1 (65793) to black
      	for(int y=0; y<Max_statusbar_height; y++)
         {
      		for(x=0; x<width; x++)
            {
               color = image->GetPixelColor(x,y);
               pixval = Get_Pixval(color); 
               if(pixval == 65793) image->SetPixelColor(x,y,BlackColor);
      		}
         }        
         
         // sets the maximum number of colors to use for SetPixelColor & HueColor
         image->SetClrImportant(0);	 
      	// Get the size of the palette  in bytes add in V9.0.3.1
         int palsize = image->GetPaletteSize()/4;   
   
         // Get the .biClrUsed;
         BITMAPINFOHEADER ds;
         HANDLE hImage = image->GetDIB();
         memcpy(&ds,hImage, sizeof(ds));
         int clrused = ds.biClrUsed; 
         
         sprintf(buffer,"DIAG CheckQueueCount0 palsize=%d clrused=%d",palsize,clrused); 
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
           
         // Pseudo Colorize the image add in V9.0.3.1
         image->HuePalette(); 
         
         // Create a array to hold the histogram  of the image, each entry is 7 words long: palette colors r,g,b, color index, 
         // color count,new index  and pixval
         palsize = image->GetPaletteSize()/4;   
      	ULONG *pArray = (ULONG*) malloc(ENTSIZE*palsize);
      	
      	// Reset the array to zero  to do the image first
      	memset(pArray,0,ENTSIZE*palsize); 
      
         // Fill the array with the information from the palette
      	for(int i=0; i<palsize; i++)
      	{
            color = image->GetPaletteColor(i);
            pixval = Get_Pixval(color);    // added in V17.0.0.19
            pArray[i*ENTLEN] = color.rgbRed; 
            pArray[i*ENTLEN+1] = color.rgbGreen; 
            pArray[i*ENTLEN+2] = color.rgbBlue; 
            pArray[i*ENTLEN+3] = i; // Save the index for this color
            pArray[i*ENTLEN+6] = (ULONG)pixval; // Save the pixval for this color
         }
            
         // Scan all the pixels and increment the palette color counts 
         int idxx;
      	for(y=0; y<height; y++)
         {
      		for(x=0; x<width; x++)
            {
               // Get color index and increment the count for the index
      			idxx = image->GetPixelIndex(x,y);
               pArray[idxx*ENTLEN+4]++;	// increment the histogram count for this color	
      		}
         }    
                          
         // Sort the array by the count value in descending order   
         sortENT( pArray, 4, palsize, 1);  
         
         // Get the highest pixval and color
         double highest_pixval = pArray[0*ENTLEN+6];
         int highest_count = pArray[0*ENTLEN+4];
         int rgbRed = pArray[0*ENTLEN]; 
         int rgbGreen  = pArray[0*ENTLEN+1]; 
         int rgbBlue  = pArray[0*ENTLEN+2]; 
         
         sprintf(buffer,"DIAG CheckQueueCount0 highest_pixval=%f highest_count=%d",highest_pixval,highest_count);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         
         sprintf(buffer,"DIAG CheckQueueCount0 highest_pixval rgbRed=%d rgbGreen%d rgbBlue=%d",rgbRed,rgbGreen,rgbBlue);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         
         // Set the palette color of first to black 
         image->SetPaletteColor(0,rgbRed,rgbGreen,rgbBlue);
         
         // Set the palette color of 2nd to While
         image->SetPaletteColor(1,WhitekColor);
         
         // Free the pArray  // added in V9.0.3.1
         if(pArray) free(pArray);
         
         // turn all the pixels in the status bar that are not near the highest_pixval to white added in V9.0.3.1
         double val_diff;
         for(int y=0; y<Max_statusbar_height; y++)
         {
            for(int x=0; x<width; x++)  
            {
         		color = image->GetPixelColor(x,y);
               pixval = Get_Pixval(color); 
               val_diff = fabs(pixval - highest_pixval);
               if (pixval != highest_pixval) image->SetPixelIndex(x,y,1); // Set to white                
      		} 
         } 
         
         // make the bottom  of the status bar the highest_pixval index
         y = 0;
         for(int x=0; x<width; x++)  
         {
      		image->SetPixelIndex(x,y,0); // zero highest_pixval index                  
   		} 
       
         // make the 1st line  of the status bar the highest_pixval index
         y = Max_statusbar_height - 1;
         for(int x=0; x<width; x++)  
         {
      		image->SetPixelIndex(x,y,0); // zero highest_pixval index                  
   		} 
         
         // make the 2nd line  of the status bar the highest_pixval index
         y = Max_statusbar_height - 2;
         for(int x=0; x<width; x++)  
         {
      		image->SetPixelIndex(x,y,0); // zero highest_pixval index                  
   		} 
         
         //Make the ist index color black
         image->SetPaletteColor(0,BlackColor);
         
         // If there is an output file2 write the uncropped image to a BMP  -- added in V9.0.0.6
         if ( argumentExists(6) )
         {
         	// See if output is a valid image type
         	TCHAR* extout = FindExtension(fileout2);
         	_tcslwr(extout);
         	int typeout = CxImage::GetTypeIdFromName(extout);
         	
            sprintf(buffer,"DIAG CheckQueueCount0 fileout2=%s extout=%s typeout=%d",fileout2, extout,typeout);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                   
         	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
            {
               sprintf(buffer,"ERROR: CheckQueueCount0 Argument 6 is an invalid image type\n");
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               goto returnit;   
         	}  
            
            // Save the image   
            rc = image->Save(fileout2,typeout);	
            if(!rc)
            {
              sprintf(buffer,"ERROR: CheckQueueCount0 Invalid rc from Save fileout2 Image RC=%d",rc);
              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              goto returnit;   
            }	                            
         }
         
         // Find the first separator scanning from the bottom of the status bar
         bool sep_found = false;
         int sep_start_y = 0;
         int sep_start_x = 0;
         int sep_end_x = 0;
         int first_sep_width = 0;
         for(int y=Max_statusbar_height - 1; y>0; y--)
         {
            for(int x=0; x<150; x++)     // scan horizontally only to 150
            {
         		color = image->GetPixelColor(x,y);
               pixval = Get_Pixval(color); 
               if (pixval == 0); // if black do nothing
               else  // else its white 
               {  //must be a separator scan for the end
                  sep_found = true;
                  // assume the bottom of the separator is bad
                  y = y-1; // get up one 
                  sep_start_y = y;  // set the start y 
                  sep_start_x = x;  // Set the start x
                  // find the end x
                  for(sep_end_x=x; sep_end_x<150; sep_end_x++)
                  {
                     color = image->GetPixelColor(sep_end_x,y);
                     pixval = Get_Pixval(color); 
                     if (pixval == 0) // if black must be the end
                     {
                        first_sep_width = sep_end_x - x;  // set the first separator width
                        break;
                     }
                  }
               } 
               if(sep_found) break;      
      		} 
            if(sep_found) break;
         } 
         
         sprintf(buffer,"DIAG CheckQueueCount0 sep_found=%d sep_start_x/y=%d %d sep_end_x=%d first_sep_width=%d",sep_found,sep_start_x,sep_start_y,sep_end_x,first_sep_width);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));      

         // Allocate 20 fields with each entry containg: fld_start, fld_end, percent of width, num_white_pixels in field, 
         //  number of chars in field and the the ist characters width
         int max_num_flds = 20; // maximum number fields
         int num_ents = 6; 
         int entsize = sizeof(double) * num_ents;  // size of one entry
         flds = (double*) malloc(entsize*max_num_flds);	
         // Reset the array to zero
         memset(flds,0,entsize*max_num_flds);
         
         // mark the start end of the first field that we found above
         flds[0*num_ents] = 0; // mark first start
         flds[0*num_ents+1] = sep_start_x; // mark 1st end
         flds[1*num_ents] = sep_end_x; // mark 2nd start
           
         // Find the start and end position of all the rest of the fields added in V3.0.6.9
         y=sep_start_y; 
         bool fnd;
         int fldnum = 0;
         // set the first field's start to 0 because sometimes there is something in 0 
         flds[fldnum*num_ents] = 0;
         int x_start = sep_end_x; // start the x after first separator
         int num_flds = 1;   // already have first one
         int sep_width = 0;
         int end_field = 0;
         int start_field = 0;
         for (fldnum=1; fldnum<20; fldnum++)
         {
            // printf("DEBUG fldnum=%d x_start=%d num_flds=%d\n",fldnum,x_start,num_flds);
            // scan starting past the end of the last separator looking for start of next separator
            fnd = false;
            sep_width = 0; // set the separator width to zero
            for (x=x_start; x<width; x++)
            {
               color = image->GetPixelColor(x,y);
               pixval = Get_Pixval(color); 
               if(pixval == 0); // If its black do nothing added in V8.0.9.9)
               else // if not black its start of a separator
               {
                  end_field = x-1; //remember the end-1 of the separator start in case it is a valid separator
                  //find the end of the separator but it must be same width as first separator to be valid
                  for (int xs=x+1; xs<width; xs++)
                  {
                     sep_width += 1;
                     color = image->GetPixelColor(xs,y);
                     pixval = Get_Pixval(color);  
                     // if its black its the end of the separator
                     if(pixval == 0) //  added  in V8.0.9.9
                     {   
                        // start_field = xs; // remember start in case its a valid separator
                        // if the separator is the same length as the first separator update the fields
                        if(sep_width == first_sep_width)
                        {    
                           // printf("DEBUG fldnum=%d pixval=%f sep_width=%d first_sep_width=%d xs=%d\n",fldnum,pixval,sep_width,first_sep_width,xs);
                           flds[fldnum*num_ents+1] = end_field; // save x-1 as the end of the field
                           flds[(fldnum+1)*num_ents] = xs; // save xs as start of next field
                           x_start = xs;
                           fnd = true;
                           num_flds += 1;
                           break;  
                        }
                     } 
                     if(fnd) break; // if we found the end of a separator break  
                  }
               }
               if(fnd) break; // if we found the end of a separator break
            }
            if (x >= width -1)  break;
         }
         num_flds += 1;   // include last field
         flds[fldnum*num_ents] = x_start;     // set last fields start
         flds[fldnum*num_ents+1] = x - 1; // set last fields end which should be at the width -1
         
         // list the start and end of all the fields
         int fld_str=0;
         int fld_end=0;
         for (int fldnum=0; fldnum<num_flds; fldnum++)
         {
            fld_str = (int)flds[fldnum*num_ents];   // get the start of the field
            fld_end = (int)flds[fldnum*num_ents+1]; // get the end of the field
            sprintf(buffer,"DIAG: num_flds=%d fldnum=%d fld_str=%d fld_end=%d",num_flds,fldnum,fld_str,fld_end);
            // printf("DEBUG: CheckQueueCount0 num_flds=%d fldnum=%d fld_str=%d fld_end=%d\n",num_flds,fldnum,fld_str,fld_end);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }
         
         // Check if there are more than 10 fields, if so die added in V8.0.11.0
         if(num_flds > 10)   
         {
           sprintf(buffer,"ERROR: CheckQueueCount0 There are more than 10 num_flds=%d",num_flds);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           goto returnit;   
         }	   
         sprintf(buffer,"DIAG CheckQueueCount0 num_flds=%d",num_flds);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
 
         fld_str=0;
         fld_end=0;
         
         int statusbar_start = 0;
         int statusbar_height = Max_statusbar_height - 1;
         int statusbar_bottom = Max_statusbar_height - 1;
         
         // determine the real width of the last field
         // get the start and end of the last field
         int lastfld = num_flds-1;
         fld_str = (int)flds[lastfld*num_ents];   // get the start of the last field
         fld_end = (int)flds[lastfld*num_ents+1]; // get the end of the last field
         fnd = false;
         y = 0;
         // scan backwards from the end
         for (y=statusbar_start; y<statusbar_height; y++)
         {
            for (x=width-1; x>=fld_str; x--)
            {
               color = image->GetPixelColor(x,y);
               pixval = Get_Pixval(color); 
               // if its not black we found the real end set it
               // If its black do nothing in V8.0.9.9)
               if(pixval == 0);  
               else  // added in V8.0.9.9
               {
                  flds[lastfld*num_ents+1] = x; // Update the end of last field
                  fnd = true;
                  break;
               }
            }
            if (fnd) break; 
         }
         
         // list the start and end of all the fields
         for (fldnum=0; fldnum<num_flds; fldnum++)
         {
            fld_str = (int)flds[fldnum*num_ents];   // get the start of the field
            fld_end = (int)flds[fldnum*num_ents+1]; // get the end of the field
            sprintf(buffer,"DIAG: num_flds=%d fldnum=%d fld_str=%d fld_end=%d",num_flds,fldnum,fld_str,fld_end);
            // printf("DEBUG: num_flds=%d fldnum=%d fld_str=%d fld_end=%d\n",num_flds,fldnum,fld_str,fld_end);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }

         // Scan the flds array and determine the percentage each field width is of the total width
         int diff;
         double perc = 0.0;
         for (fldnum=0; fldnum<num_flds; fldnum++)
         {
            fld_str = (int)flds[fldnum*num_ents];   // get the start of the field
            fld_end = (int)flds[fldnum*num_ents+1]; // get the end of the field
            diff = fld_end - fld_str;
            perc = (double)diff / (double)width;  
            sprintf(buffer, "%.*g",2, perc);	
            perc = strtod(buffer, NULL); 	
            flds[fldnum*num_ents+2] = perc; // save as third entry 
         }

         // Determine the number of white pixels in each field
         int total_pixels; 
         for (fldnum=0; fldnum<num_flds; fldnum++)
         {
            total_pixels = 0;
            fld_str = (int)flds[fldnum*num_ents];   // get the start of the field
            fld_end = (int)flds[fldnum*num_ents+1]; // get the end of the field
            for (y=statusbar_start; y<statusbar_height - statusbar_start; y++)
            {
               for (x=fld_str; x<=fld_end; x++)
               {
                  color = image->GetPixelColor(x,y);
                  pixval = Get_Pixval(color); 
                  // If its black do nothing added in V8.0.9.9)
                  if(pixval == 0); 
                  // if its not black count it 
                  else // added in V8.0.9.9
                  {
                     total_pixels += 1;
                  }
               } 
            }
            flds[fldnum*num_ents+3] = total_pixels; // save total_pixels as fourth double 
         }
         
         // Determine the number of characters in each field
         int total_chars; 
         int first_char_width = 0;
         int first_char_str = 0;
         for (fldnum=0; fldnum<num_flds; fldnum++)
         {
            total_chars = 0;
            
            fld_str = (int)flds[fldnum*num_ents];   // get the start of the field
            fld_end = (int)flds[fldnum*num_ents+1]; // get the end of the field
            pixval = 0;
            // scan until the pixval is zero which means we found the bottom of first char
            for (y=statusbar_start; y<statusbar_height - statusbar_start; y++)
            {
               for (x=fld_str; x<=fld_end; x++)
               {
                  color = image->GetPixelColor(x,y);
                  pixval = Get_Pixval(color); 
                  // If its black do nothing added in V8.0.9.9)
                  if(pixval == 0); 
                  // if its not black its the start of the first character
                  else // added in V8.0.9.9 
                  {
                     total_chars += 1;
                     first_char_str = x;
                     break;
                  }
               } 
               // If its black do nothing added  in V8.0.9.9)
               if(pixval == 0);
               else break;  // added in V8.0.9.9  // we have found the bottom left of 1st character in this field
            }

            // if y equals 0 then make the total chars one to eliminate fields with extenders
            if (y == 0) flds[fldnum*num_ents+4] = 1; 
            else
            {
               // now find the start and end of each character by scanning the bottoms
               fnd = false;
               int char_num = 0;
               for (int xs=x; xs<=fld_end; xs++)
               {  // if we found the end and now its not black its a new character, count it
                  color = image->GetPixelColor(xs,y);
                  pixval = Get_Pixval(color); 
                  if (fnd && pixval != 0)
                  {
                     total_chars += 1;
                     char_num += 1;
                     fnd = false;
                  }
                  else if(fnd == false)
                  {  // check that there are 2 pixels in a row that are black
                     color = image->GetPixelColor(xs+1,y); // get the value of the next char
                     pixval2 = Get_Pixval(color); 
                     // if there are two blacks chars in a row its the end of the character
                     if(pixval == 0 && pixval2 == 0) 
                     {
                        fnd = true;
                        // if this is the first char then calculate its size
                        if (char_num == 0) first_char_width = xs - first_char_str;
                     } 
                  }
               }
               flds[fldnum*num_ents+4] = total_chars; // save total characters as fifth entry
               flds[fldnum*num_ents+5] = first_char_width; // save first_char_width as 6th entry
            }
            total_pixels = (int)flds[fldnum*num_ents+3]; 
            perc = flds[fldnum*num_ents+2];   // get the precentage
            // print it 
            sprintf(buffer,"DIAG field %d x/y=%d %d pixval=%f starts at %d ends at %d has %d pixels, width percentage=%f and %d characters 1st width=%d",
                           fldnum+1,x,y,pixval,fld_str,fld_end,total_pixels,perc,total_chars,first_char_width);                  
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         }

         // determine if the last field is Tile path Info
         int last_fld = num_flds -1;
         perc = flds[last_fld*num_ents+2];   // get the percentage
         bool tile_path_info = false;
         // if its over 20 % of the total width then its on
         if (perc > 0.20) tile_path_info = true;

         int queue0_fld = 0;
         switch (num_flds) 
         {
            case 1:
               queue0_fld = 1;
               break;
            case 2:
               if(tile_path_info) queue0_fld = 1;
               else queue0_fld = 2;
               break;
         	case 3:
               if(tile_path_info) queue0_fld = 2;
               else queue0_fld = 3;
               break;
            case 4:
               if(tile_path_info) queue0_fld = 3;
               else queue0_fld = 4;
               break;
            case 5:
               if(tile_path_info) queue0_fld = 4;
               else queue0_fld = 5;
               break;
            case 6:
               if(tile_path_info) queue0_fld = 5;
               else queue0_fld = 6;
               break;
            case 7:
               if(tile_path_info) queue0_fld = 6;
               else queue0_fld = 7;
               break; 
            case 8:
               queue0_fld = 7;
               break;  	     	
         }
         queue0_fld -= 1;
         
         fld_str = (int)flds[queue0_fld*num_ents];   // get the start of the field
         fld_end = (int)flds[queue0_fld*num_ents+1]; // get the end of the field
         sprintf(buffer,"DIAG field queue0_fld=%d starts at %d ends at %d ",queue0_fld,fld_str,fld_end);                                     
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
                      
         sprintf(buffer,"DIAG CheckQueueCount0 fld_str=%d fld_end=%d",fld_str,fld_end);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         
         // printf("DEBUG fldnum=%d fld_str=%d fld_end=%d\n",fldnum,fld_str,fld_end);
         
         // Find the the right bottom side of the "0"
         int x_bot_right = 0; 
         int y_bot_right = 0;
         // start x at 3rd separator - 1 
         fnd = false;
         for (y=statusbar_bottom; y>=statusbar_start; y--) 
         {       
            for (x=fld_end; x >= fld_str; x--) // scan x backwards 
            {
               color = image->GetPixelColor(x,y);
               pixval = Get_Pixval(color);  
               // If its black do nothing added  in V8.0.9.9)
               if(pixval == 0);
               // if not black its the right side of the 0, save the x and y
               else  
               { 
                  x_bot_right = x+1;   
                  y_bot_right = y; 
                  fnd = true; 
                  break;
               }
            } 
            if(fnd) break;
         }
         sprintf(buffer,"DIAG CheckQueueCount0 x_bot_right=%d y_bot_right=%d",x_bot_right,y_bot_right);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
        
         // Now find the the left bottom side of the "0"
         int x_bot_left = 0; 
         int y_bot_left = 0; 
         fnd = false;
         // start at the right side of the zero - 1 
         for (x=x_bot_right -1; x >= fld_str; x--) // scan x backwards 
         {
            color = image->GetPixelColor(x,y);
            pixval = Get_Pixval(color);  
            // if black its one past the left bottom side of the 0
            if(pixval == 0) 
            { 
               x_bot_left = x; 
               y_bot_left = y;  
               fnd = true; 
               break;
            }
         } 
         sprintf(buffer,"DIAG CheckQueueCount0 x_bot_left=%d y_bot_left=%d",x_bot_left,y_bot_left);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         
         // scan up ont the right until we hit a non white
         int x_top_right = 0;
         int y_top_right = 0;
         for (y=y_bot_right -1; y >=statusbar_start; y--) // scan y up
         {
            color = image->GetPixelColor(x,y);
            pixval = Get_Pixval(color); 
            // If its black do nothing added in V8.0.9.9)
            if(pixval == 0); 
            // if its not a black break 
            else break;
  
         }    ; 
         // contine scanning up on the right until we hit a 0
         for (y=y-1; y >=statusbar_start; y--) // scan y up
         {
            color = image->GetPixelColor(x,y);
            pixval = Get_Pixval(color);  
            // if its black 
            if(pixval == 0) 
            { 
               x_top_right = x; 
               y_top_right = y;  
               fnd = true; 
               break;
            }   
         } 
         sprintf(buffer,"DIAG CheckQueueCount0 x_top_right=%d y_top_right=%d",x_top_right,y_top_right);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
         // scan up  the left until we hit a non 0
         int x_top_left = 0;
         int y_top_left = 0;
         for (y=y_bot_left -1; y >=statusbar_start; y--) // scan y up
         {
            color = image->GetPixelColor(x,y);
            pixval = Get_Pixval(color);  
            // If its black do nothing added in V8.0.9.9)
            if(pixval == 0); 
            // if its black 
            else break;
         }  
         // Contine scanning up until we hit a 0  ;
         for (y=y-1; y >=statusbar_start; y--) // scan y up
         {
            color = image->GetPixelColor(x,y);
            pixval = Get_Pixval(color);  
            // if its black 
            if(pixval == 0) if(pixval == 0) 
            { 
               x_top_left = x; 
               y_top_left = y;  
               fnd = true; 
               break;
            }   
         } 
          
         // Determine if its a complete zero
         if ((y_bot_left - y_top_left) > 5 && y_top_right == y_top_left && x_bot_left == x_top_left ) rtnrc= 1;
         else rtnrc = 0;
         
         sprintf(buffer,"DIAG CheckQueueCount0 rtnrc=%d x_top_left=%d y_top_left_side=%d",rtnrc,x_top_left,y_top_left);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         
         // Reduce the number of colors and return the numcolors  added in V3.0.6.5
         int numcolors = 0;
         CQuantizer m(16777216,8);
         m.ProcessImage(image->GetDIB()); 	 
         numcolors = m.GetColorCount();  
         sprintf(buffer,"DIAG CheckQueueCount0 BEFORE: numcolors=%d",numcolors);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));        
      }    	
   		
   	// Delete everything	
   	SelectObject(memdc, hOldBitmap);
   	DeleteObject(memdc);
   	DeleteObject(membit);
   	::ReleaseDC(0,scrdc);
         	  
      // Destroy the image if one exists
      if (image->IsValid()) 
      {
         rc = image->Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: CheckQueueCount0  1. Invalid rc from Destroy Image RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }	 
         sprintf(buffer,"DIAG CheckQueueCount0 destroyed newimage rc=%d",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
      }  
      
   }   
   GdiplusShutdown(gdiplusToken);  
        
   sprintf(buffer,"DIAG CheckQueueCount0 freeing arrays\n");
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));

   // free the local array
   if (flds)  free(flds);
    
   sprintf(buffer,"DIAG CheckQueueCount0 Returning RC=%d",rtnrc);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
     
   returnit:  
	
   return rtnrc;
}
/**************************** Screen_Info **************************
 * gets the screenbinformation: resoultion, scale, pixel width/height, bpp, xres, yres
 * 
 * Syntax:	rc = Screen_Info(hwnd,outout_dir,comm._diag, diagmsgs)
 *
 * Params:	hwnd- handle of the window like SasPlanet 
 *          data - A directory to pass back the scale, width, height, bpp, xres and yres added in V8.0.9.4
 *			   diag    - if gt 0 it will print diagnostic information  
 *			   diagmsgs - array to return diagnostic messages  
 *			   
 *                           
 * Returns: 0 if OK else 1
 *********************************************************************/
RexxRoutine4(double, Screen_Info, uintptr_t,hwnd, RexxObjectPtr,data, int,diag, RexxArrayObject,diagmsgs)                  
{
   char buffer[512]; 
   int rc;
   logical_t flag;
   RexxObjectPtr dataobj;

	using namespace Gdiplus;
	GdiplusStartupInput gdiplusStartupInput;
	ULONG_PTR gdiplusToken;
	GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);
	{    
      sprintf(buffer,"DIAG Screen_Info Handle=%u",hwnd);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        
   	// Check if we have a valid window handle
      if ( !IsWindow((HWND)hwnd) )
      {
         sprintf(buffer,"ERROR: Screen_Info Argument 1 is not a valid window handle\n");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;                    
      }  
      
      // Check its a valid directory 
      RexxDirectoryObject data_dir;  
      data_dir = (RexxDirectoryObject)data;
      flag = context->IsDirectory(data_dir);
      if ( flag == 0 )
      {
         sprintf(buffer,"ERROR: Screen_Info parameter 2 is not a directory!");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
      }

   	// Get the window size  
      RECT r;   
      GetWindowRect((HWND)hwnd, &r); 
      sprintf(buffer,"DIAG Screen_Info left=%d right=%d top=%d bottom=%d",r.left,r.right,r.top,r.bottom);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      
      int before_width =  r.right - r.left ;
      int before_height = r.bottom - r.top ;
      sprintf(buffer,"DIAG Screen_Info before_width=%d before_height=%d",before_width,before_height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      
      // Get the screen resolution 
      DEVMODE dev;
      int32_t Enum_w, Enum_h;
      EnumDisplaySettings(nullptr, ENUM_CURRENT_SETTINGS, &dev);
      Enum_w = dev.dmPelsWidth;
      Enum_h = dev.dmPelsHeight;     
      sprintf(buffer,"DIAG Screen_Info Before_width/height=%d %d Screen Resolution:%d %d",before_width,before_height,Enum_w,Enum_h);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
            
      // Pass back the screen resolution
      dataobj = context->Int32ToObject(Enum_w); 
      context->DirectoryPut(data_dir, dataobj, "ENUM_W"); 
      dataobj = context->Int32ToObject(Enum_h); 
      context->DirectoryPut(data_dir, dataobj, "ENUM_H"); 		
    
      double scale = (double)Enum_w / (double)before_width;
      sprintf(buffer, "%.*g", 3, scale);	
      scale = strtod(buffer, NULL); 
      
      // Pass back the scale
      dataobj = context->DoubleToObject(scale); 
      context->DirectoryPut(data_dir, dataobj, "SCALE"); 	
         
      // Adjust the rectangle added in V3.0.4.1
      sprintf(buffer,"DIAG Screen_Info Before: scale=%f top=%d bottom=%d left=%d right=%d",scale, r.top,r.bottom,r.left,r.right);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));      
         
      // if the scale is not 1.0 then adjust the top and bottom 
      if (scale > 1.0)
      {
         int incr = (int)ceil(r.top * scale) -  r.top;  // ceil = roundup
         int top_before = r.top;
         r.top += incr;
         r.bottom += incr;
         sprintf(buffer,"DIAG Screen_Info adjusted to scale=%f r.top by:%d was=%d now=%d",scale, incr, top_before, r.top);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      }   
       
      int32_t width, height, x, y;
      x = r.left;
      y = r.top;
      
      width =  r.right - r.left ;
      height = r.bottom - r.top ;
      
      width =  (int32_t)(width * scale);
      height = (int32_t)(height * scale); 
      
      sprintf(buffer,"DIAG Screen_Info After Width=%d Height=%d",width,height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      
      //  Check if the screen dimensions are valid
      if(width <= 0 || height <= 0) 
      {
         sprintf(buffer,"ERROR: CheckQueueCount0 Invalid screen dimensions width=%d height=%d\n",width, height);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;    
      }       
          
      // Copy the screen
      HDC scrdc, memdc;
      HBITMAP membit;
      scrdc = ::GetDC(0);
      memdc = CreateCompatibleDC(scrdc);
      membit = CreateCompatibleBitmap(scrdc, width, height);
      HBITMAP hOldBitmap =(HBITMAP) SelectObject(memdc, membit);
      BitBlt(memdc, 0, 0, width, height, scrdc, x, y, SRCCOPY);	
      Gdiplus::Bitmap bitmap(membit, NULL);
     
      // Create a new image from the bitmap 
      CxImage *image = new CxImage();
      
      rc = image->CreateFromHBITMAP(membit);
      if (!rc)
      {
         sprintf(buffer,"ERROR: CheckQueueCount0 Invalid rc from CreateFromHBITMAP RC=%d\n",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;    
      }             
      
      sprintf(buffer,"DIAG Screen_Info After: top=%d bottom=%d left=%d right=%d width=%d height=%d",r.top,r.bottom,r.left,r.right,width,height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      
      // Pass back the screen pixel width and height
      dataobj = context->Int32ToObject(width); 
      context->DirectoryPut(data_dir, dataobj, "PIXEL_WIDTH"); 
      dataobj = context->Int32ToObject(height); 
      context->DirectoryPut(data_dir, dataobj, "PIXEL_HEIGHT");
      
      //  pass back the BPP XRES and YRES
      int32_t resx, resy;
      resx = image->GetXDPI(); // Get X resolution
      resy = image->GetYDPI(); // Get Y resolution   
      
      // Get the bpp
      int32_t bpp;
      bpp  =  image->GetBpp(); // Get the bpp 
      
      // get the number of monitors connected
      MonitorRects monitors;
      dataobj = context->Int32ToObject(monitors.rcMonitors.size()); 
      context->DirectoryPut(data_dir, dataobj, "NUM_MONITORS"); 
      // cout << "You have " << monitors.rcMonitors.size() << " monitors connected.";
      
      dataobj = context->Int32ToObject(bpp); 
      context->DirectoryPut(data_dir, dataobj, "BPP"); 
      dataobj = context->Int32ToObject(resx); 
      context->DirectoryPut(data_dir, dataobj, "XRES"); 
      dataobj = context->Int32ToObject(resy); 
      context->DirectoryPut(data_dir, dataobj, "YRES");
      sprintf(buffer,"DIAG Screen_Info Image pixel Width/Height=%d %d Bpp=%d Xres=%d Yres=%d",width,height,bpp,resx, resy);  
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      
      // Destroy the image if one exists
      if (image->IsValid()) 
      {
         rc = image->Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: Screen_Info Invalid rc from Destroy Image RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }	 
         sprintf(buffer,"DIAG Screen_Info destroyed image rc=%d",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
      }  

      return 0;
   }
}
/**************************** Taken_Blocks **************************
 * Allocates an array the size of an image 
 * and marks blocks around a x/y offset as taken 
 * or it checks the x/y to see if its taken
 * 
 * Syntax:	rc = Taken_blocks(hwnd,outout_dir,comm._diag, diagmsgs)
 *
 * Params:	 
 *          data - A directory to pass parms back and forth it contains:
 *               - function to perform:
 *                  - 0 = Allocate storage the size of the pixel_width * pixel_height
 *                  - 1 = check if the xp/yp is taken
 *                  - 2 = block the xp/yp  
 *                  - 9 = deallocate the storage
 *               - image width
 *               - image height
 *               - xp offset into the image
 *               - yp offset into the image
 *               - block width
 *               - block height
 *               - address of the block
 *               - a return code 0 if ok  else 1
 *			   diag    - if gt 0 it will print diagnostic information  
 *			   diagmsgs - array to return diagnostic messages  
 *			   
 *                           
 * Returns: 0 if OK else 1
 *********************************************************************/
RexxRoutine3(double, Taken_blocks, RexxObjectPtr,data, int,diag, RexxArrayObject,diagmsgs)                  
{
   char buffer[512]; 
   int rc;
   logical_t flag;
   RexxObjectPtr dataobj;
   HANDLE hwnd = 0;

	using namespace Gdiplus;
	GdiplusStartupInput gdiplusStartupInput;
	ULONG_PTR gdiplusToken;
	GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);
	{        
      // Check its a valid directory 
      RexxDirectoryObject dir_data;  
      dir_data = (RexxDirectoryObject)data;
      flag = context->IsDirectory(dir_data);
      if ( flag == 0 )
      {
         sprintf(buffer,"ERROR: Taken_blocks parameter 2 is not a directory!");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
      }
      
      // printf("DEBUG IN Taken_blocks\n");

      //Get all necessary values that are not arrays
      int func,pixel_width,pixel_height,xp,yp,block_width,block_height,return_code;
      dataobj = (RexxObjectPtr)context->DirectoryAt(dir_data, "func"); 
      flag = context->ObjectToInt32(dataobj, &func);          
      dataobj = (RexxObjectPtr)context->DirectoryAt(dir_data, "pixel_width"); 
      flag = context->ObjectToInt32(dataobj, &pixel_width); 
      dataobj = (RexxObjectPtr)context->DirectoryAt(dir_data, "pixel_height");
      flag = context->ObjectToInt32(dataobj, &pixel_height);  
      dataobj = (RexxObjectPtr)context->DirectoryAt(dir_data, "xp");
      flag = context->ObjectToInt32(dataobj, &xp);  
      dataobj = (RexxObjectPtr)context->DirectoryAt(dir_data, "yp");
      flag = context->ObjectToInt32(dataobj, &yp); 
      dataobj = (RexxObjectPtr)context->DirectoryAt(dir_data, "block_width");
      flag = context->ObjectToInt32(dataobj, &block_width);  
      dataobj = (RexxObjectPtr)context->DirectoryAt(dir_data, "block_height");
      flag = context->ObjectToInt32(dataobj, &block_height);  
      
      // printf("DEBUG IN Taken_blocks  Got all the values\n");
      
      // if func is zero allocate the array 
      // intptr_t * ar_ptr = 0;
      char * ar_ptr;
      SIZE_T size = pixel_width * pixel_height;
      if(func == 0)   // Allocate the buffer and pass back a pointer
      {
         return_code = 1;  //Set bad dc
         // printf("DEBUG IN Taken_block creating the heap size=%d\n",size);
         
         // Allocate a private heap the size of pixel_width * pixel_height
         hwnd =  GetProcessHeap();
         // hwnd = HeapCreate(0, 0, size );
         if (hwnd == NULL )
      	{	
            sprintf(buffer,"ERROR: Taken_blocks unable to create  heap pixel_width/height=%d %d",pixel_width, pixel_height);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            return 1;    
      	}
         
         // printf("DEBUG IN Taken_block allocating the array size=%d\n",size);
         
         if (!(ar_ptr = (char *)HeapAlloc(hwnd,HEAP_ZERO_MEMORY,size)))
      	{	
            sprintf(buffer,"ERROR: Taken_blocks unable to allocate array pixel_width/height=%d %d",pixel_width, pixel_height);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            return 1;    
      	}
         // check we got all we asked for
         SIZE_T actual_size = HeapSize( hwnd,0,(char *)ar_ptr);
         if (actual_size != size)
      	{	
            sprintf(buffer,"ERROR: Taken_blocks unable to allocate all the array pixel_width/height=%d %d",pixel_width, pixel_height);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            return 1;    
      	}
         
         // printf("DEBUG IN Taken_block passing back apointer to the array \n");
         
         // Pass back the the ptr to the array
         RexxPointerObject p = context->NewPointer(ar_ptr);
         context->DirectoryPut(dir_data, p, "ar_ptr"); 
         
         // Set good return code
         return_code = 0;
         dataobj = context->Int32ToObject(return_code); 
         context->DirectoryPut(dir_data, dataobj, "return_code"); 
      }  
      else if(func == 1)  // check if certain xp/yp is taken
      {  
         // printf("DEBUG IN Taken_block Function 1 xp/yp=%d %d \n",xp,yp);
         
         // get the pointer to the array
         RexxPointerObject ptr = (RexxPointerObject)context->DirectoryAt(dir_data, "ar_ptr");
         ar_ptr = (char *)context->PointerValue(ptr);
         
         // printf("DEBUG IN Taken_block Function 1 got the ar_ptr \n"); 
         
         // pass back the value at the xp yp
         if (ar_ptr[xp + yp]) return_code =  1;
         else return_code =  0;
         
         // DEBUG return_code = ar_ptr[xp + yp];
         
         // printf("DEBUG IN Taken_block return_code=%d \n",return_code); 
         
         dataobj = context->Int32ToObject(return_code); 
         context->DirectoryPut(dir_data, dataobj, "return_code"); 
      }
      else if(func == 2)  // block the array around the xp/yp
      {  
         // printf("DEBUG IN Taken_block Function 2 xp/yp=%d %d block_width/height=%d %d \n",xp,yp,block_width,block_height);
         
         // get the pointer to the array
         RexxPointerObject ptr = (RexxPointerObject)context->DirectoryAt(dir_data, "ar_ptr");
         ar_ptr = (char *)context->PointerValue(ptr);
         
         // printf("DEBUG IN Taken_block Function 2 got the ar_ptr \n"); 
         
         // fill the block  with the ul_xp ul_yp values
         int ul_xp,ul_yp, org_ul_xp,bx_xp,bx_yp;
         ul_xp = xp - (block_width / 2);
         ul_yp = yp - (block_height / 2);
         org_ul_xp = ul_xp;
         // printf("DEBUG IN Taken_block xp/yp=%d %d block_width/height=%d %d ul_xp/yp=%d %d \n",xp,yp,block_width,block_height,ul_xp,ul_yp); 
         
         if(diag > 0)
         {
            sprintf(buffer,"DIAG: Taken_blocks Taken_block xp/yp=%d %d block_width/height=%d %d ul_xp/yp=%d %d \n",xp,yp,block_width,block_height,ul_xp,ul_yp);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }
         for(bx_yp=0;bx_yp<block_height; bx_yp++)
         {
            for(bx_xp=0;bx_xp<block_width; bx_xp++)
            {
               ar_ptr[ul_xp + ul_yp] = true; // mark this pixel taken
               if(diag > 1)
               {
                  sprintf(buffer,"DIAG: DEBUG bx_xp/yp=%d %d ul_xp/yp=%d %d \n",bx_xp,bx_yp,ul_xp,ul_yp);
                  rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               }   
               
               ul_xp += 1;
            }
            ul_xp = org_ul_xp; 
            ul_yp += 1;
         }
         
         return_code = 0;
         
         // printf("DEBUG IN Taken_block Function 2 return_code=%d \n",return_code); 
         
         dataobj = context->Int32ToObject(return_code); 
         context->DirectoryPut(dir_data, dataobj, "return_code"); 
      }
      else   // Deallocate the storage
         // printf("DEBUG IN Taken_block Function 9 freeing the storage"); 
      {  // get the pointer to the array
         RexxPointerObject ptr = (RexxPointerObject)context->DirectoryAt(dir_data, "ar_ptr");
         ar_ptr = (char *)context->PointerValue(ptr);
         
         BOOL therc = HeapFree(hwnd,0,ar_ptr);
         if(therc == 0)
         {
            sprintf(buffer,"ERROR: Taken_blocks unable to free the array");
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            return 1;    
         }
         
         /* DEBUG
         // Free the heap
         therc = HeapDestroy(hwnd);
         if(therc == 0)
         {
            sprintf(buffer,"ERROR: Taken_blocks unable to destroy the heap");
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            return 1;    
         }
         */
         
         // Set good return code
         return_code = 0;
         dataobj = context->Int32ToObject(return_code); 
         context->DirectoryPut(dir_data, dataobj, "return_code"); 
      }   	

      return 0;
   }
}
          
/**************************** GetZoomLevel **************************
 * gets the zoom level from the StatusBar
 * 
 * Syntax:	rc = GetZoomLevel(hwnd, queueImage0, comm._diag, diagmsgs, fileout)
 *
 * Params:	hwnd- handle of the window to take a screen shot
 *          zoom_level - if >0 initialize all the zoom levels
 *                       or if 0 then return the current zoom level
 *          data - a directory to get and pass back information               
 *			   diag    - if gt 0 it will print diagnostic information  
 *			   diagmsgs - array to return diagnostic messages  
 *			   fileout - optional filespec of BMP to store the screen image
 *			   fileout2 - optional filespec of BMP to store the screen image
 *                           
 * Returns: if the positions array is empty then returns an image of the zoom level in the array
 *          else if its not empty returns 1 if all pixels match the positions array pixels, else 0  
 *          else -1 if error
 *********************************************************************/
RexxRoutine7(int, GetZoomLevel, uintptr_t,hwnd, int,zoom_level, RexxObjectPtr,data, int,diag, RexxArrayObject,diagmsgs, OPTIONAL_CSTRING,fileout,OPTIONAL_CSTRING,fileout2) 
                  
{
   char buffer[512]; 
   int rc;
   int rtnrc = -1;  // Preset bad RC
   int *local_queueImage0 = NULL;  // local  array  added in V16.0.0.17
   logical_t flag;
   double * flds = NULL;

   sprintf(buffer,"DIAG GetZoomLevel Handle=%u zoom_level=%d",hwnd,zoom_level);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));

	using namespace Gdiplus;
	GdiplusStartupInput gdiplusStartupInput;
	ULONG_PTR gdiplusToken;
	GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);
	{    
        
   	// Check if we have a valid window handle
      if ( !IsWindow((HWND)hwnd) )
      {
         sprintf(buffer,"ERROR: GetZoomLevel Argument 1 is not a valid window handle\n");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return -1;                    
      }  

      RexxDirectoryObject data_dir;  
      data_dir = (RexxDirectoryObject)data;
      flag = context->IsDirectory(data_dir);
      if ( flag == 0 )
      {
         sprintf(buffer,"ERROR: GetZoomLevel parameter 2 is not a directory!");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return -1;
      }

      //Get the scale passed
      double scale;
      RexxObjectPtr dataobj;
      dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "SCALE");
      flag = context->ObjectToDouble(dataobj, &scale);

   	// Get the window size  
      RECT r;
      rc = -1;   
      GetWindowRect((HWND)hwnd, &r); 
      sprintf(buffer,"DIAG GetZoomLevel left=%d right=%d top=%d bottom=%d",r.left,r.right,r.top,r.bottom);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      
      int before_width =  r.right - r.left ;
      int before_height = r.bottom - r.top ;
      sprintf(buffer,"DIAG GetZoomLevel before_width=%d before_height=%d",before_width,before_height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      
      // Get the screen resolution 
      DEVMODE dev;
      int Enum_w, Enum_h;
      EnumDisplaySettings(nullptr, ENUM_CURRENT_SETTINGS, &dev);
      Enum_w = dev.dmPelsWidth;
      Enum_h = dev.dmPelsHeight;     
      sprintf(buffer,"DIAG GetZoomLevel Before_width/height=%d %d Screen Resolution:%d %d",before_width,before_height,Enum_w,Enum_h);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  

      int width, height, x, y;
      
      // Adjust the rectangle added in V3.0.4.1
      sprintf(buffer,"DIAG GetZoomLevel Before: scale=%f top=%d bottom=%d left=%d right=%d",scale, r.top,r.bottom,r.left,r.right);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
       
      // if the scale is not 1.0 then adjust the top and bottom 
      int incr;
      if (scale > 1.0)
      {
         incr = (int)ceil(r.top * scale) -  r.top;  // ceil = roundup
         int top_before = r.top;
         r.top += incr;
         r.bottom += incr;
         sprintf(buffer,"DIAG GetZoomLevel adjusted to scale=%f r.top by:%d was=%d now=%d",scale, incr, top_before, r.top);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
         // if this is for initialization adjust the left and right
         if (zoom_level != 0) 
         {      
            incr = (int)ceil(r.left * scale) -  r.left;  // ceil = roundup
            int left_before = r.left;
            r.left += incr;
            r.right += incr;
         }
      }

      x = r.left;
      y = r.top;

      width =  r.right - r.left ;
      height = r.bottom - r.top ;
      
      width =  (int)(width * scale);
      height = (int)(height * scale); 
      
      sprintf(buffer,"DIAG GetZoomLevel After Width=%d Height=%d",width,height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));             
      
      sprintf(buffer,"DIAG GetZoomLevel After: top=%d bottom=%d left=%d right=%d width=%d height=%d",r.top,r.bottom,r.left,r.right,width,height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   
      //  Check if the screen dimensions are valid
      if(width <= 0 || height <= 0) 
      {
         sprintf(buffer,"ERROR: GetZoomLevel Invalid screen dimensions width=%d height=%d\n",width, height);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         goto returnit;    
      }       
          
      // Copy the screen
      HDC scrdc, memdc;
      HBITMAP membit;
      scrdc = ::GetDC(0);
      memdc = CreateCompatibleDC(scrdc);
      membit = CreateCompatibleBitmap(scrdc, width, height);
      HBITMAP hOldBitmap =(HBITMAP) SelectObject(memdc, membit);
      BitBlt(memdc, 0, 0, width, height, scrdc, x, y, SRCCOPY);	
      Gdiplus::Bitmap bitmap(membit, NULL);
     
      // Create a new image from the bitmap 
      CxImage image;
      CxImage *imagecopy = new CxImage();
      
      rc = image.CreateFromHBITMAP(membit);
      if (!rc)
      {
         sprintf(buffer,"ERROR: GetZoomLevel Invalid rc from CreateFromHBITMAP RC=%d\n",rc);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         goto returnit;    
      }  
      else
      {  
         int resx, resy;
         if(diag > 0)   
         {
            resx = image.GetXDPI(); // Get X resolution
            resy = image.GetYDPI(); // Get Y resolution
            sprintf(buffer,"DIAG GetZoomLevel BEFORE:Xres=%d Yres=%d",resx, resy);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }   
        
      
         // If there is an output file2 write the whole image to a BMP 
         if ( argumentExists(7) )
         {
         	// See if output is a valid image type
         	TCHAR* extout = FindExtension(fileout2);
         	_tcslwr(extout);
         	int typeout = CxImage::GetTypeIdFromName(extout);   	
         	
            sprintf(buffer,"DIAG GetZoomLevel fileout2=%s extout=%s typeout=%d",fileout2, extout,typeout);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                  
             
         	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
            {
               sprintf(buffer,"ERROR: GetZoomLevel Argument 7 is an invalid image type\n");
               printf(buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               goto returnit;   
         	}  
            
            // Save the image   
            rc = image.Save(fileout2,typeout);	
            if(!rc)
            {
              sprintf(buffer,"ERROR: GetZoomLevel Invalid rc from Save fileout2 Image RC=%d",rc);
              printf(buffer);
              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              goto returnit;   
            }	                            
         }            
       
         RGBQUAD color;
         
         double pixval; 
          
         // White
         RGBQUAD WhitekColor;
         WhitekColor.rgbRed = 255; 
         WhitekColor.rgbGreen = 255; 
         WhitekColor.rgbBlue = 255;
        
         // Black
         RGBQUAD BkgndColor;
         BkgndColor.rgbRed = 0; 
         BkgndColor.rgbGreen = 0; 
         BkgndColor.rgbBlue = 0; 
      
         // Create a copy of the image
         imagecopy->Copy(image);

         bool fnd = false;

         // If this an initialization call remove the border
         if (zoom_level != 0) 
         {   // NOTE: the image is upside down.  

            // set the copy to all black and white     
            // rc = imagecopy->DecreaseBpp(1,0,0,0); 

            int bdr_top_y = 0;
            int bdr_bot_y = height -1; 
            int bdr_left_x = 0;
            int bdr_right_x = width -1;

            // Determine how thick the border at the top is
            fnd = false;
            int x = 4; 
            for(int y=4; y<height; y++)
            {
         		// for(int x=1; x<width; x++)
               //{
            		color = imagecopy->GetPixelColor(x,y);
                  pixval = Get_Pixval(color); 
                  // if its black we have found the border top y
                  if(pixval == 0)
                  {
                     bdr_top_y = y -1; 
                     // fnd = true;
                     break; 
                  }                
         		//} 
               //if (fnd) break;
            }  
            // Determine how thick the border at the bottom is
            fnd = false;
            for(int y=height-4; y>=0; y--)
            {
         		// for(int x=1; x<width; x++)
               // {
            		color = imagecopy->GetPixelColor(x,y);
                  pixval = Get_Pixval(color); 
                  // if its black we have found the row the border bottom ends 
                  if(pixval == 0 )
                  {
                     bdr_bot_y = y +1; 
                     // fnd = true;
                     break; 
                  }                      
         		//} 
               // if (fnd) break;
            }  
            // Determine how thick the border at the left
            fnd = false;
            int y = bdr_top_y+1;
            for(int x=1; x<width; x++)
            {
         		// for(int y=1; y<height; y++)
               // {
            		color = imagecopy->GetPixelColor(x,y);
                  pixval = Get_Pixval(color); 
                  // if its black we have found the col the border left 
                  if(pixval == 0 )
                  {
                     bdr_left_x = x -1; 
                     // fnd = true;
                     break; 
                  // }                  
         		} 
               if (fnd) break;
            }  
            // Determine how thick the border at the right is
            fnd = false;
            for(int x=width-2; x>=0; x--)
            {
         		for(int y=1; y<height; y++)
               {
            		color = imagecopy->GetPixelColor(x,y);
                  pixval = Get_Pixval(color); 
                  // if its black we have found the col the border left 
                  if(pixval == 0)
                  {
                     bdr_right_x = x +1; 
                     fnd = true;
                     break; 
                  }                    
         		} 
               if (fnd) break;
            }  
         }

         // Determine the start of the status bar added in V3.0.6.3
         x = 0;
         int statusbar_start = 9999;
         for (y=0; y<height; y++)
         {
            color = imagecopy->GetPixelColor(x,y);
            pixval = Get_Pixval(color); 
            // if its black assume the bottom of the status bar
            if(pixval == 0)  
            {
               statusbar_start = y;  // set the start (actually the bottom)
               break;
            }
            
         }
         sprintf(buffer,"DIAG GetZoomLevel statusbar_start=%d",statusbar_start);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));

         int statusbar_height = 0;
         int statusbar_bottom = 0;
         // If this is not an initialization call Determine the height of the status bar
         if (zoom_level == 0) 
         {
         
            x = 1;  // start at 1 because sometime status bar has a someting in 0
            for (y=statusbar_start; y<height; y++)
            {
               color = imagecopy->GetPixelColor(x,y);
               pixval = Get_Pixval(color);
               // of its not black must be bottom of status bar +1 
               if(pixval != 0) 
               {
                  statusbar_bottom = y -1;
                  break;   
               }
               statusbar_height += 1; // increment the status bar height
            }
         }
         else  
         {
            statusbar_height = height -1; // for the init call use the height -1 
            statusbar_bottom  = 0; // same for the bottom
         }
         sprintf(buffer,"DIAG GetZoomLevel statusbar_bottom=%d statusbar_height=%d",statusbar_bottom, statusbar_height);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 

         long newwidth,newheight;
         // Crop the image  into just the status bar
         // If this is not an initialization call Determine the height of the status bar
         if (zoom_level == 0) 
         {
            float angle = 0;
            newwidth = width;   // changed in V4.0.0.
            newheight = statusbar_height;    // changed in V4.0.0.2
            long left = 0;  // changed in V4.0.0.2
            long top  = height - statusbar_height;  // changed in V4.0.0.2
               
            sprintf(buffer,"DIAG GetZoomLevel left=%d top=%d newwidth=%d newheight=%d ",left, top, newwidth, newheight);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
                
            rc = imagecopy->CropRotatedRectangle(left, top, newwidth, newheight, angle );
            if (!rc)
            {
               sprintf(buffer,"ERROR: GetZoomLevel Invalid rc from crop image RC=%d\n",rc);
               printf(buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               goto returnit;   
            }  

            // set the copy to all black and white     
            // rc = imagecopy->DecreaseBpp(1,0,0,0); 
            height = newheight;
            width = newwidth; 
         }
  
         // If there is an output file write the imagcopy to it  
         if ( argumentExists(6) )
         {
         	// See if output is a valid image type
         	TCHAR* extout = FindExtension(fileout);
         	_tcslwr(extout);
         	int typeout = CxImage::GetTypeIdFromName(extout);
         	
            sprintf(buffer,"DIAG GetZoomLevel fileout=%s extout=%s typeout=%d",fileout, extout,typeout);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                   
         	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
            {
               sprintf(buffer,"ERROR: GetZoomLevel Argument 6 is an invalid image type\n");
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               goto returnit;   
         	}  
            
            // Save the image   
            rc = imagecopy->Save(fileout,typeout);	
            if(!rc)
            {
              sprintf(buffer,"ERROR: GetZoomLevel Invalid rc from Save fileout2 Image RC=%d",rc);
              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              goto returnit;   
            }	                            
         }
         
         // NOTE: the iage is flipped upside down 
              
         // Check the status bar if less than 5 or gt 99 its probably off, return an error
         if(statusbar_height < 5 || statusbar_height > 99)   
         {
           sprintf(buffer,"ERROR: GetZoomLevel statusbar_height=%d",statusbar_height);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           goto returnit;   
         }	   
         sprintf(buffer,"DIAG GetZoomLevel statusbar_height=%d",statusbar_height);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 

         // Create an array to hold information about each field
         int max_num_flds = 20; // maximum number fields
         // number of doubles per entries : fld_start, fld_end, percent of width, num_white_pixels in field, 
         //  number of chars in field and the the ist chaaracters width
         int num_ents = 6; 
         int entsize = sizeof(double) * num_ents;  // size of one entry
         flds = (double*) malloc(entsize*max_num_flds);	
         // Reset the array to zero
         memset(flds,0,entsize*max_num_flds);

         // find the row that the separators start at the bottom 
         fnd = false;
         int sep_row = 0;
         int y_start = statusbar_bottom;
         for(int y=y_start; y<height; y++)
         {
      		for(int x=0; x<width -3; x++)  // start to x=1 because sometimes there is something in 0 
            {
         		color = imagecopy->GetPixelColor(x,y);
               pixval = Get_Pixval(color); 
               // if its white its the row the sep starts 
               if(pixval == 16777215 ) 
               {
                  sep_row = y;
                  fnd = true;
                  break;
               }                 
      		} 
            if (fnd) break;
         } 

         sprintf(buffer,"DIAG GetZoomLevel y_start=%d x=%d sep_row=%d",y_start,x,sep_row);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
           
         // Find the start and end position of all the fields 
         y=sep_row; 
         fnd;
         int fldnum = 0;
         // set the first field's start to 1 because sometimes there is something in 0 
         flds[fldnum*num_ents] = 1;
         int x_start = 1;
         int num_flds = 0;
         for (fldnum=0; fldnum<20; fldnum++)
         {
            // scan starting from the bottom of the status bar up
            fnd = false;
            for (x=x_start; x<width; x++)
            {
               color = imagecopy->GetPixelColor(x,y);
               pixval = Get_Pixval(color);  
               // if not black its start of a separator find so set flds end
               if(pixval != 0)
               {
                  flds[fldnum*num_ents+1] = x-1; // save x-1 as the end of the field
      
                  num_flds += 1;
                  //find the end of the separator
                  for (int xs=x+1; xs<width; xs++)
                  {
                     color = imagecopy->GetPixelColor(xs,y);
                     pixval = Get_Pixval(color);  
                     // if its black its the end of the separator
                     if(pixval == 0) 
                     {     
                        flds[(fldnum+1)*num_ents] = xs; // save xs as start of next field
                        x_start = xs;
                        fnd = true;
                        break;
                     }   
                  }
               }
               if(fnd) break; // if we found the end of a separator break
            }
            if (x >= width -1)  break;
         }
         flds[fldnum*num_ents+1] = x - 1; // save last fields end which should be at the width -1
         int fld_str=0;
         int fld_end=0;

         // determine the real width of the last field
         // get the start and end of the last field
         fld_str = (int)flds[num_flds*num_ents];   // get the start of the field
         fld_end = (int)flds[num_flds*num_ents]; // get the end of the field
         fnd = false;
         // scan backwards from the end
         for (x=fld_end; x>=fld_str; x--)
         {
            for (y=statusbar_start; y<statusbar_height - statusbar_start; y++)
            {
               color = imagecopy->GetPixelColor(x,y);
               pixval = Get_Pixval(color); 
               // if its not black we found the real end set it
               if(pixval != 0)  
               {
                  flds[num_flds*num_ents+1] = x;
                  fnd = true;
                  break;
               }
            }
            if (fnd) break; 
         }

         num_flds +=1; // add one more to account for the last field

         // list the start and end of all the fields
         for (fldnum=0; fldnum<num_flds; fldnum++)
         {
            fld_str = (int)flds[fldnum*num_ents];   // get the start of the field
            fld_end = (int)flds[fldnum*num_ents+1]; // get the end of the field
            sprintf(buffer,"DIAG: num_flds=%d fldnum=%d fld_str=%d fld_end=%d",num_flds,fldnum,fld_str,fld_end);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }
         
         // Determine the number of white pixels in each field
         int total_pixels; 
         for (fldnum=0; fldnum<num_flds; fldnum++)
         {
            total_pixels = 0;
            fld_str = (int)flds[fldnum*num_ents];   // get the start of the field
            fld_end = (int)flds[fldnum*num_ents+1]; // get the end of the field
            for (y=statusbar_start; y<statusbar_height - statusbar_start; y++)
            {
               for (x=fld_str+2; x<=fld_end-1; x++)
               {
                  color = imagecopy->GetPixelColor(x,y);
                  pixval = Get_Pixval(color); 
                  // if its not black count it
                  if(pixval != 0)  
                  {
                     total_pixels += 1;
                  }
               } 
            }
            flds[fldnum*num_ents+3] = total_pixels; // save total_pixels as fourth double 
         }
         
         // create the arrays for all the fields
         int array_num=0;
         for (fldnum=0; fldnum<num_flds; fldnum++)
         {
            total_pixels = (int)flds[fldnum*num_ents+3]; 
            if (zoom_level == 0) array_num = 0;
            else array_num = (fldnum +1) + ((zoom_level-1) * 8);
            // create an array for this field
            RexxArrayObject obj;      
            obj = context->NewArray(total_pixels);
            sprintf(buffer,"%d",array_num);
            dataobj = context->String(buffer);
            context->DirectoryPut(data_dir, obj, buffer);
            if (zoom_level == 0) break; // only do the first field for a non-init call
         }

         // get the x,ys for each field
         char ar_num[10];
         RexxArrayObject zm_array;
         int top_y, bot_y, left_x,right_x,top_x,bot_x;
         int total_xys;
         for (fldnum=0; fldnum<num_flds; fldnum++)
         {    
            fld_str = (int)flds[fldnum*num_ents];   // get the start of the field
            fld_end = (int)flds[fldnum*num_ents+1]; // get the end of the field
           
            if (zoom_level == 0) array_num = 0;
            else array_num = (fldnum +1) + ((zoom_level-1) * 8);
            sprintf(ar_num,"%d",array_num); // Convert the ararray_num to a string 
            zm_array = (RexxArrayObject)context->DirectoryAt(data_dir, ar_num);

            sprintf(buffer,"DIAG GetZoomLevel scanning for xyz in Z%s fld_str=%d fld_end=%d",ar_num,fld_str,fld_end);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));

            // find the top of the Z data 
            double pixval2;
            fnd = false;
            for (y=0; y<height; y++)
            {
               for (x=fld_str+2; x<fld_end-1; x++)
               {
                  color = imagecopy->GetPixelColor(x,y);
                  pixval = Get_Pixval(color); 
                  color = imagecopy->GetPixelColor(x+1,y);
                  pixval2 = Get_Pixval(color); 
                  // if its not black use it
                  if(pixval != 0 && pixval2 != 0)  
                  {
                     top_y = y;
                     top_x = x;
                     fnd = true;
                     break;
                  }
               } 
               if (fnd) break;
            }

            // find the bottom of the data 
            fnd = false;
            for (y=height-1; y>=0; y--)
            {
               for (x=fld_str+2; x<fld_end-1; x++)
               {
                  color = imagecopy->GetPixelColor(x,y);
                  pixval = Get_Pixval(color);
                  // if its not black use it
                  if(pixval != 0) 
                  {
                     bot_y = y;
                     bot_x = x;
                     fnd = true;
                     break;
                  }
               } 
               if (fnd) break;
            }
            
            // find the left side 
            fnd = false;
            for (x=fld_str+2; x<fld_end-1; x++)
            {
               for (y=0; y<height;y++)
               {
                  color = imagecopy->GetPixelColor(x,y);
                  pixval = Get_Pixval(color); 
                  // if its not black use it
                  if(pixval != 0)  
                  {
                     left_x = x;
                     fnd = true;
                     break;
                  }
               } 
               if (fnd) break;
            }
            
            // find the right side 
            fnd = false;
            for (x=fld_end-1; x>=fld_str+2; x--)
            {
               for (y=0; y<height;y++)
               {
                  color = imagecopy->GetPixelColor(x,y);
                  pixval = Get_Pixval(color); 
                  // if its not black use it
                  if(pixval != 0)  
                  {
                     right_x = x;
                     fnd = true;
                     break;
                  }
               } 
               if (fnd) break;
            }

            // scan the data rectangle and store the x,y s
            int offset_x = 0;
            int offset_y = 0;
            total_xys = 0;
            sprintf(buffer,"DIAG GetZoomLevel scan rect with top_y=%d bot_y=%d left_x=%d right_x=%d",top_y,bot_y,left_x,right_x);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));

            // scan rect with top_y=6 bot_y=14 left_x=12 right_x=25
            for (y=top_y; y<=bot_y; y++)
            {  
               offset_x = 0; // reset x
               for (x=left_x; x<=right_x; x++)
               {
                  color = imagecopy->GetPixelColor(x,y);
                  pixval = Get_Pixval(color); 
                  // if its not black use it
                  if(pixval != 0)  
                  {     
                     // save in Z array   
                     sprintf(buffer," %d %d",offset_x,offset_y); 
                     rc = context->ArrayAppendString(zm_array, buffer, strlen(buffer)); 
                     total_xys += 1;
                  }
                  offset_x += 1;
               } 
               offset_y += 1;
            }
         
            flds[fldnum*num_ents+4] = (double)total_xys; // save total_xys as fifth double 

            sprintf(buffer,"DIAG GetZoomLevel Z%s top_x=%d top_y=%d bot_x=%d bot_y=%d left_x=%d right_x=%d total_xys=%d",ar_num,top_x,top_y,bot_x,bot_y,left_x,right_x,total_xys);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));

            // only do the first field for a non-init call   
            if (zoom_level == 0) break; 
            
         }

         // List all the fields
         int total_chars; 
         int first_char_width = 0;
         int first_char_str = 0;
         for (fldnum=0; fldnum<num_flds; fldnum++)
         {
            total_chars = 0;
            
            fld_str = (int)flds[fldnum*num_ents];   // get the start of the field
            fld_end = (int)flds[fldnum*num_ents+1]; // get the end of the field
            
            total_pixels = (int)flds[fldnum*num_ents+3]; 
            total_xys = (int)flds[fldnum*num_ents+4]; 
            
            // print it 
            sprintf(buffer,"DIAG GetZoomLevel field %d starts at %d ends at %d has %d pixels %d xyz pixels",
                           fldnum+1,fld_str,fld_end,total_pixels,total_xys);                  
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         }

         // If there is an output file write the imagcopy to it
         if ( argumentExists(6) )
         {
         	// See if output is a valid image type
         	TCHAR* extout = FindExtension(fileout);
         	_tcslwr(extout);
         	int typeout = CxImage::GetTypeIdFromName(extout);
         	
            sprintf(buffer,"DIAG GetZoomLevel fileout=%s extout=%s typeout=%d",fileout, extout,typeout);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                   
         	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
            {
               sprintf(buffer,"ERROR: GetZoomLevel Argument 6 is an invalid image type\n");
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               goto returnit;   
         	}  
            
            // Save the image   
            rc = imagecopy->Save(fileout,typeout);	
            if(!rc)
            {
              sprintf(buffer,"ERROR: GetZoomLevel Invalid rc from Save fileout2 Image RC=%d",rc);
              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              goto returnit;   
            }	                            
         }

         // If this zoom_level=0 see which of the other arrays has the closest match
         if (zoom_level == 0)
         {
            CString str1,str2; 
            // get a pointer to the Z0 array
            int sz,sz_z0;
            array_num = 0;
            // RexxStringObject str1,str2;
            sprintf(ar_num,"%d",array_num); // Convert the ararray_num to a string 
            RexxArrayObject z0 = (RexxArrayObject)context->DirectoryAt(data_dir, ar_num);
            sz_z0 = context->ArrayItems(z0); 
            int best_cmp = 0;
            int best_zm = 0;
            // scan  all the zoom arrays and compare
            for(int array_num=1; array_num<25; array_num++)
            {  
               sprintf(ar_num,"%d",array_num); // Convert the array_num to a string 
               zm_array = (RexxArrayObject)context->DirectoryAt(data_dir, ar_num);
               if (zm_array == NULLOBJECT) continue; // forget if no array why??
               sz = context->ArrayItems(zm_array); 
          
               // compare all the entries to z0
               int good_cmp = 0;
               for(int n = 1; n < sz_z0; n++)
               {
                  dataobj = context->ArrayAt(z0, n);
                  str1 = context->CString(dataobj);
                  // scan all of the zm_array looking for a match
                  for(int m = 1; m < sz; m++)
                  {
                     dataobj = context->ArrayAt(zm_array, m);
                     str2 = context->CString(dataobj);
                     if (!strcmp(str1,str2))
                     {
                        good_cmp += 1;
                        break;
                     }
                  }
               } 

               // see if this one is the best
               if (good_cmp > best_cmp)
               {
                  best_cmp = good_cmp;
                  best_zm = array_num; 
               } 
            }
            rtnrc = best_zm; // return the best match
         } 
         else rtnrc = 0;
         
         // Reduce the number of colors and return the numcolors  added in V3.0.6.5
         int numcolors = 0;
         CQuantizer m(16777216,8);
         m.ProcessImage(image.GetDIB());	 
         numcolors = m.GetColorCount();  
         sprintf(buffer,"DIAG GetZoomLevel BEFORE: numcolors=%d",numcolors);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }    	
   		
   	// Delete everything	
   	SelectObject(memdc, hOldBitmap);
   	DeleteObject(memdc);
   	DeleteObject(membit);
   	::ReleaseDC(0,scrdc);


      // Destroy the image if one exists
      if (image.IsValid()) 
      {
         rc = image.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: GetZoomLevel  1. Invalid rc from Destroy Image RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }	 
         sprintf(buffer,"DIAG GetZoomLevel destroyed image rc=%d",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
      }  
         	  
      // Destroy the imagecopy if one exists
      if (imagecopy->IsValid()) 
      {
         rc = imagecopy->Destroy(); // Destroy the imagecopy
         if(!rc)
         {
           sprintf(buffer,"Error: GetZoomLevel  1. Invalid rc from Destroy Image RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }	 
         sprintf(buffer,"DIAG GetZoomLevel destroyed imagecopy rc=%d",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
      }  
      
   }   
   GdiplusShutdown(gdiplusToken);  
        
   sprintf(buffer,"DIAG GetZoomLevel freeing arrays\n");
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));

   // free the local array
   if (flds)  free(flds);
    
   sprintf(buffer,"DIAG GetZoomLevel Returning RC=%d",rtnrc);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
     
   returnit:    

   return rtnrc;
}
/**************************** Check_View **************************
 * Checks if the GE View pulldown
 *
 * Syntax:	rc = Check_View(window_handle, filename, diag, diagmsgs)
 *
 * Params:	window_handle - handle of the window to take a screen shot
 *			   diag    - if gt 0 it will print diagnostic information  
 *			   diagmsgs - array to return diagnostic messages 
 *			   filename - optional name of the file to save the window as a BMP
 *
 * Return: 0 if OK else -1
 *********************************************************************/
RexxRoutine4(int, Check_View, uintptr_t,hwnd, int,diag, RexxArrayObject,diagmsgs, OPTIONAL_CSTRING,fileout)
{
   char buffer[2048]; 
   int rc;
   int rtnrc = -1;  // Preset bad RC
   
	using namespace Gdiplus;
	GdiplusStartupInput gdiplusStartupInput;
	ULONG_PTR gdiplusToken;
	GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);
	{   
      if ( diag>0 ) 
      {
         sprintf(buffer,"DIAG Check_View Handle=%u fileout=%s",hwnd, fileout);
         // printf(buffer); printf("\n");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }       
        
   	// Check if we have a valid window handle
      if ( !IsWindow((HWND)hwnd) )
      {
         sprintf(buffer,"ERROR: Check_View Argument 1 is not a valid window handle\n");
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         goto returnit;                            
      }      
      
   	// Get the window size  
      RECT r;
      rc = -1;   
      GetWindowRect((HWND)hwnd, &r); 
      
      int before_width =  r.right - r.left ;
      int before_height = r.bottom - r.top ;
      
      // Get the screen resolution 
      DEVMODE dev;
      int Enum_w, Enum_h;
      EnumDisplaySettings(nullptr, ENUM_CURRENT_SETTINGS, &dev);
      Enum_w = dev.dmPelsWidth;
      Enum_h = dev.dmPelsHeight;     
      sprintf(buffer,"DIAG Check_View Before_width/height=%d %d Screen Resolution:%d %d",before_width,before_height,Enum_w,Enum_h);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      
      double scale = (double)Enum_w / (double)before_width;  
      sprintf(buffer, "%.*g", 3, scale);	
      scale = strtod(buffer, NULL); 	
      
      // Adjust the rectangle  added in V3.0.4.1
      sprintf(buffer,"DIAG Check_View Before: scale=%f top=%d bottom=%d left=%d right=%d",scale, r.top,r.bottom,r.left,r.right);
      // printf(buffer); printf("\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
       
      // if the scale is not 1.0 then adjust the top and bottom 
      if (scale > 1.0)
      {
         int incr = (int)ceil(r.top * scale) -  r.top;  // ceil = roundup
         int top_before = r.top;
         r.top += incr;
         r.bottom += incr;
         sprintf(buffer,"DIAG Check_View adjusted to scale=%f r.top by:%d was=%d now=%d",scale, incr, top_before, r.top);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      }   
       
      int width, height, x, y;
      x = r.left;
      y = r.top;
      
      width =  r.right - r.left ;
      height = r.bottom - r.top ;
      
      width =  (int)(width * scale);
      height = (int)(height * scale); 
      
      sprintf(buffer,"DIAG Check_View After Width=%d Height=%d",width,height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));             
      
      sprintf(buffer,"DIAG Check_View After: top=%d bottom=%d left=%d right=%d width=%d height=%d",r.top,r.bottom,r.left,r.right,width,height);
      // printf(buffer); printf("\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));                 
      //  Check if the screen dimensions are valid
      if(width <= 0 || height <= 0) 
      {
         sprintf(buffer,"ERROR: Check_View Invalid screen dimensions width=%d height=%d\n",width, height);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         goto returnit;     
      }       
      
      // Copy the screen
      HDC scrdc, memdc;
      HBITMAP membit;
      scrdc = ::GetDC(0);
      memdc = CreateCompatibleDC(scrdc);
      membit = CreateCompatibleBitmap(scrdc, width, height);
      HBITMAP hOldBitmap =(HBITMAP) SelectObject(memdc, membit);
      BitBlt(memdc, 0, 0, width, height, scrdc, x, y, SRCCOPY);	
      Gdiplus::Bitmap bitmap(membit, NULL);
     
      // Create a new image from the bitmap 
      CxImage *newImage = new CxImage();
      rc = newImage->CreateFromHBITMAP(membit);
      if (!rc)
      {
         sprintf(buffer,"ERROR: Check_View Invalid rc from CreateFromHBITMAP RC=%d\n",rc);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         goto returnit;     
      }  
      else
      {  
        
         //  Change to B & W image
         ULONG outbpp; 
         if (!newImage->IsGrayScale()) newImage->GrayScale();
			// REMOVED in V17.0.0.15 newImage->HuePalette();       
         
         rc = newImage->AdaptiveThreshold();
         if (!rc)
         {
            sprintf(buffer,"ERROR: Check_View Invalid rc from AdaptiveThreshold RC=%d\n",rc);
            printf(buffer);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            goto returnit2;     
         }
         
         // If there is an output file write the screen to a BMP 
         if ( argumentExists(4) )
         {         
         	// See if output is a valid image type
         	TCHAR* extout = FindExtension(fileout);
         	_tcslwr(extout);
         	int typeout = CxImage::GetTypeIdFromName(extout);
         	
            
            sprintf(buffer,"DIAG Check_View extout=%s typeout=%d",extout,typeout);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                          
         	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
            {
               sprintf(buffer,"ERROR: Check_View Argument 4 is an invalid image type\n");
               printf(buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               goto returnit2;   
         	}  
            
            // Save the image   
            rc = newImage->Save(fileout,typeout);	
            if(!rc)
            {
              sprintf(buffer,"ERROR: Check_View Invalid rc from Save Image RC=%d\n ",rc);
              printf(buffer);
              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              goto returnit2;   
            }	                            
         }
         
         // Flip it over
         rc = newImage->Flip();
         if (!rc)
         {
            sprintf(buffer,"ERROR: Check_View Invalid rc from Flip RC=%d\n",rc);
            printf(buffer);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            goto returnit2;     
         }         
                    
         outbpp = newImage->GetBpp(); 
         // Pass back the type, width, height, X resolution and Y resolution
         int b_resx, b_resy, b_width, b_height;
         b_resx = newImage->GetXDPI(); // Get X resolution
         b_resy = newImage->GetYDPI(); // Get Y resolution
         b_width = newImage->GetWidth();
         b_height = newImage->GetHeight();
         sprintf(buffer,"DIAG Check_View XRES=%d YRES=%d ",b_resx, b_resy);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));       
         sprintf(buffer,"DIAG Check_View WIDTH=%d HEIGHT=%d ",b_width,b_height);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            
         RGBQUAD color;    
         ULONG pixval;                  
            
         // Find the start and end x of the pulldown
         int good_cnt=0;
         int left_x=0;
         int right_x=0; 
         int y = 23;  // the image is NOT upside down changed to 23 from 0  in V26.0.0.8 for GE Pro
         for(x=0; x<b_width; x++)
         {         // white = 16777215
            color = newImage->GetPixelColor(x,y);
            pixval = Get_Pixval(color);
            if(pixval==0)  // if black
            {
               if (good_cnt < 200)  // if we don't have a lot then reset
               {
                  good_cnt = 0;
                  left_x = 0;
               }
               else
               {
                  if (right_x == 0)
                     right_x = x;
               }
            }
            else  // must be white
            { 
               if (right_x == 0)   // if no end has been set start counting
               {
                  good_cnt += 1; // increment good
                  if (left_x == 0)
                     left_x = x;
               }
            }        
         }
         sprintf(buffer,"DIAG Check_View good_cnt=%d left_x=%d right_x=%d \n",good_cnt, left_x, right_x);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
         
         // If we found the pull down then see how tall it is
         int bottom_y = 0; 
         if (good_cnt > 200) 
         {     
            for(x=left_x,y=23; y<b_height; y++)  // Changed y to 23 was 0 in V26.0.0.8 for GE Pro
            {        
               color = newImage->GetPixelColor(x,y);
               pixval = Get_Pixval(color);
               if(pixval == 0)  // if black its the bottom
               {
                  bottom_y = y;
                  sprintf(buffer,"DIAG Check_View bottom_y=%d \n",y);
                  rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
                  break;
               }
            }
         }    
         
         // If we found the View pulldown write it out 
         if ( diag>1 &&  bottom_y > 0)
         {
            sprintf(buffer,"DIAG Check_View Screen Image:");
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            int offset;
            for (y=23; y < bottom_y+1; y++)   // Changed y to 23 was 0 in V26.0.0.8 for GE Pro
            {    
         		for(x=left_x,offset=0; x < right_x+1; x++,offset++)
               {
            		color = newImage->GetPixelColor(x,y);
                  pixval = Get_Pixval(color);
                  if(pixval == 0)
                     buffer[offset] = 32;   // space = black
                  else  
                     buffer[offset] = 88;  // X = white         
               }
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            }  
         }                       
         
         // If we found the pull down bottom then scan down to see if any check marks
         if (bottom_y > 0) 
         {        
            int look_offset = 12;   // was 9 changed to 12 in V26.0.0.8
            int lookfor=0; // start looking for black
            int start_mark = 0;
            int pix_below = 0;
            for(x=left_x+look_offset,y=23; y < bottom_y+1; y++) // Changed y to 23 was 0 in V26.0.0.8 for GE Pro
            {        
               color = newImage->GetPixelColor(x,y);
               pixval = Get_Pixval(color);
               if(pixval == lookfor)  // if lookfor found
               {
                  color = newImage->GetPixelColor(x,y+1);   // check the color below it
                  pix_below = Get_Pixval(color); 
                  if (pix_below == pixval) // if not same color forget it probably a separator line
                  {
                     if (pixval == 0) start_mark = y;   // if its the start remember it
                     else  // if its an end then log it
                     {
                        sprintf(buffer,"DIAG Check_View look_offset=%d start_mark=%d end_mark=%d \n",look_offset, start_mark, y);
                        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
                     }
                     color = newImage->GetPixelColor(x,y-1);   // now use the color above it
                     lookfor = Get_Pixval(color); 
                  }
               }
            }
            rtnrc = 0;
         }                  
      }	
      
      returnit2:	
   	// Delete everything	
   	SelectObject(memdc, hOldBitmap);
   	DeleteObject(memdc);
   	DeleteObject(membit);
   	::ReleaseDC(0,scrdc);
    	  
      // Destroy the image if one exists
      if (newImage->IsValid()) 
      {
         rc = newImage->Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: Check_View  1. Invalid rc from Destroy Image RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }	 
         sprintf(buffer,"DIAG Check_View destroyed newimage rc=%d",rc);
         // printf(buffer);printf("\n");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
      }  
      
   }   
   GdiplusShutdown(gdiplusToken);    
    
   sprintf(buffer,"DIAG Check_View Returning RC=%d",rtnrc);
   // printf(buffer);printf("\n");
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
   
   returnit: 
    	
   return rtnrc;       
}
/**************************** ConvertImage **************************
  Converts image files from one type to another
 * Syntax:	rc = ConvertImage(filein, fileout, mixfilein, diag, diagmsgs)
 *
 * Params:		
 *			   filein - filename of the input file
 *			   fileout  - filename of the outputfile file
 *			   mixfilein  - (optional) filename of the file to mix with input image 
 *			   diag    - if specified it will print diag information 
 *			   diagmsgs    - an array to hold diagnostic and error messages  
 *			   histogram  (optional) an array to return the resulting images pallette
 *			   maxcolors     (optional) the maximum number of colors added in V6.0.2.1 
 *
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine7(int, ConvertImage, CSTRING, filein, CSTRING, fileout, OPTIONAL_CSTRING, mixfilein, int, diag, RexxArrayObject, diagmsgs, OPTIONAL_RexxArrayObject, histogram, OPTIONAL_int, maxcolors)
{
   char buffer[512]; 
   int rc;
                                                 
   if ( diag>0 )
   {
      sprintf(buffer,"DIAG ConvertImage input=%s output=%s",filein, fileout);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   } 
   	
	ULONG inbpp; 
	ULONG outbpp; 
	RGBQUAD * ppal;
   int colors;

	// See if arg 1 is a valid image type
	TCHAR* extin = FindExtension(filein);
	TCHAR* extout = FindExtension(fileout);
	_tcslwr(extin);
	int typein = CxImage::GetTypeIdFromName(extin);	
	_tcslwr(extout);
	int typeout = CxImage::GetTypeIdFromName(extout);
	
	if ( diag>0 )
   {
      sprintf(buffer,"DIAG ConvertImage extin=%s extout=%s typein=%d typeout=%d ENTSIZE=%d",extin,extout,typein,typeout,ENTSIZE);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   } 
	
	if (typein == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: ConvertImage Argument 1 is an invalid image type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}
   // See if arg 2 is a valid image type
	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: ConvertImage Argument 2 in an invalid type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;    
	}	

   // If there is an optional maxcolors  argument  set the colors added in V6.0.2.1
   if ( argumentExists(7) ) colors = maxcolors;  // set the colors to the maxcolors  
   else  colors = 127;
   sprintf(buffer,"DIAG ConvertImage maximum colors %d",colors);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
	
	CxImage image;
	if(!image.Load(filein,typein))
	{
      sprintf(buffer,"ERROR: ConvertImage Load image file:%s failed\n",filein);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      // image.Destroy();
      if (image.IsValid()) 
      {
         rc = image.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: ConvertImage 1. Invalid rc from Destroy Image RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }
      }	
      return 1;
	}

   inbpp = image.GetBpp();  // Added in V6.0.1.2
   sprintf(buffer,"DIAG ConvertImage inbpp=%d",inbpp);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));

   // RGBQUAD * ppal;
   if (inbpp == 8 || inbpp == 24)   // new in V6.0.1.2
   {
      int numcolors;
      int reduced_colors = colors;
      // if there is a mixin file then reduce the input image to half_the colors 
      if ( argumentExists(3) )  reduced_colors = colors / 2; 
      
      // If the reduced colors = 16777216 then don't quantize or decreease the bpp added in V8.0.1.7  
      if (reduced_colors == 16777216 );
      else
      {   
        sprintf(buffer,"DIAG ConvertImage reducing colors to reduced_colors=%d",reduced_colors);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        CQuantizer q(reduced_colors,(reduced_colors>16?7:8));
        q.ProcessImage(image.GetDIB()); 
        ppal=(RGBQUAD*)calloc(256*sizeof(RGBQUAD),1);
        q.SetColorTable(ppal);
        rc = image.DecreaseBpp(8,1,ppal,reduced_colors); // V600 Set error diffusion=1
        ULONG outbpp = image.GetBpp(); 
        if (!rc)
        {
            sprintf(buffer,"ERROR: ConvertImage DecreaseBpp RC=%d outbpp=%d\n",rc,outbpp);
            printf(buffer);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            return 1;	
         } 
         numcolors = q.GetColorCount(); 
         sprintf(buffer,"DIAG ConvertImage 1. out_numcolors=%d outbpp=%d",numcolors,outbpp);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      
      } 
   }    

	// If there is a mixin file load it and mix with the input image
	if ( argumentExists(3) ) 
   {
      
      // print the mix filename
      sprintf(buffer,"DIAG ConvertImage MixFilein=%s",mixfilein);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));

      // See if arg 3 is a valid image type
   	TCHAR* mixinext = FindExtension(mixfilein);
   	_tcslwr(mixinext);
   	int mixtypein = CxImage::GetTypeIdFromName(mixinext);
   	if (mixtypein == CXIMAGE_FORMAT_UNKNOWN) 
      {
         sprintf(buffer,"ERROR: Argumnet 3 in an invalid type\n");
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         // image.Destroy();
         if (image.IsValid()) 
         {
            rc = image.Destroy(); // Destroy the image
            if(!rc)
            {
              sprintf(buffer,"Error: ConvertImage 2. Invalid rc from Destroy Image RC=%d\n ",rc);
              printf(buffer);
              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              return 1;
            }
         }	
         return 1;      
   	}
   	// Create a mix image
   	CxImage miximage;
   	// Load the mix file image
   	if(!miximage.Load(mixfilein, mixtypein))
   	{
         sprintf(buffer,"ERROR: ConvertImage Error loading mixfile:%s\n",mixfilein);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         // image.Destroy();
         if (image.IsValid()) 
         {
            rc = image.Destroy(); // Destroy the image
            if(!rc)
            {
              sprintf(buffer,"Error: ConvertImage 3. Invalid rc from Destroy Image RC=%d\n ",rc);
              printf(buffer);
              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              return 1;
            }
         }	
         // miximage.Destroy();
         if (miximage.IsValid()) 
         {
            rc = miximage.Destroy(); // Destroy the image
            if(!rc)
            {
              sprintf(buffer,"Error: ConvertImage 4. Invalid rc from Destroy Image RC=%d\n ",rc);
              printf(buffer);
              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              return 1;
            }
         }
         return 1;   	
   	}
   	
   
   	// ======================================================================================================	
   	//  Mix the foreground image with the background image  both must have 1/2 the maxcolors either 63 or 128
   	//  It does this by taking the first 63 or 128 colors of the GE background and adding in the 
   	//  63 or 128 colors of the foreground which is the chart.
      // ====================================================================================================== 
   	RGBQUAD color;
      int half_colors = colors / 2;
   	BYTE red,green,blue;
   	int countbkg=0;
   	int countfor=0;
   	int height = image.GetHeight();
	   int width = image.GetWidth();
	   long x, y;
	   RGBQUAD BkgndColor;
      BkgndColor.rgbRed = 0; 
      BkgndColor.rgbGreen = 0; 
      BkgndColor.rgbBlue = 0; 
      BkgndColor.rgbReserved = 1; 
      int count, numcolors;
      ULONG pixval;
      uint32_t fgidx,idx,newidx;
	   		
		// Get the size of the palette  in bytes
      int palsize = image.GetPaletteSize()/4;   

      sprintf(buffer,"DIAG ConvertImage palsize=%d half_colors=%d",palsize,half_colors);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
         	            
   	// Create a array to hold the histogram  of the images, each entry is 7 words long: palette colors r,g,b, color index, 
      // color count,new index  and pixval
   	ULONG *pArray = (ULONG*) malloc(ENTSIZE*palsize);
   	
   	// Reset the array to zero  to do the image first
   	memset(pArray,0,ENTSIZE*palsize); 
   	   	
   	// Create a histogram of the image 

   	for(int i=0; i<palsize; i++)
   	{
         color = image.GetPaletteColor(i);
         pixval = Get_Pixval(color);           // added in V17.0.0.19
         pArray[i*ENTLEN] = color.rgbRed; 
         pArray[i*ENTLEN+1] = color.rgbGreen; 
         pArray[i*ENTLEN+2] = color.rgbBlue; 
         pArray[i*ENTLEN+3] = i; // Save the index for this color
         pArray[i*ENTLEN+6] = pixval; // Save the pixval (colorref) for this color
      }
         
      // Scan all the pixels and get the palette color counts 
   	for(y=0; y<height; y++)
      {
   		for(x=0; x<width; x++)
         {
            // Get color index and increment the count for the index
   			idx = image.GetPixelIndex(x,y);
            pArray[idx*ENTLEN+4]++;	// increment the histogram count for this color	
   		}
      }    
       
   	if ( diag>1 ) // if diag > 1  print the histogram
   	{       
         sprintf(buffer,"DIAG ConvertImage histogram image before sort:\n");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));    
      	// list the values in the histogram 
      	numcolors = listENT(pArray, palsize,diagmsgs,context);
      	sprintf(buffer,"DIAG ConvertImage image Numcolors=%d \n",numcolors);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));        	
      }	
      
      if ( diag>1 )  // If DIAG  IS GT 1 then print all the image palette colors  
      {
         for(int i=0; i<palsize; i++)
         {
            color = image.GetPaletteColor(i);
   		   sprintf(buffer,"DIAG ConvertImage DIAG image palette  %d: rgb(%d,%d,%d) \n",i,color.rgbRed,color.rgbGreen,color.rgbBlue);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }
      }              
      
      // ***********************************************************************************************
      //  NOW  ADD THE MIXIMAGE (Chart) 
      // ***********************************************************************************************
   	// Reset the array to zero  to do the mix image
   	memset(pArray,0,ENTSIZE*palsize); 
   	
	   uint32_t transidx = miximage.GetTransIndex();  // Get the transparent color index
   	color = miximage.GetTransColor();              // and the transparent color
      pixval = Get_Pixval(color);    // added in V17.0.0.19
  		
		// Get the size of the palette  in bytes
      palsize = miximage.GetPaletteSize()/4;   

      sprintf(buffer,"DIAG ConvertImage palsize=%d  transidx:%d colors:rgb(%d,%d,%d)(%d)\n",palsize,transidx,color.rgbRed,color.rgbGreen,color.rgbBlue,pixval);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));        	
   	
   	// Create a histogram of the miximage 

   	for(int i=0; i<palsize; i++)
   	{
         color = miximage.GetPaletteColor(i);
         pixval = Get_Pixval(color);    // added in V17.0.0.19
         pArray[i*ENTLEN] = color.rgbRed; 
         pArray[i*ENTLEN+1] = color.rgbGreen; 
         pArray[i*ENTLEN+2] = color.rgbBlue; 
         pArray[i*ENTLEN+3] = i; // Save the index for this color
         pArray[i*ENTLEN+6] = pixval; // Save the pixval for this color
      }
         
      // Scan all the pixels and increment the palette color counts 
   	for(y=0; y<height; y++)
      {
   		for(x=0; x<width; x++)
         {
            // Get color index and increment the count for the index
   			idx = miximage.GetPixelIndex(x,y);
            pArray[idx*ENTLEN+4]++;	// increment the histogram count for this color	
   		}
      }  
      
      if ( diag>1 ) // if diag > 1 turned on print the histogram
   	{       
         sprintf(buffer,"DIAG ConvertImage histogram miximage:\n");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));    
      	// list the values in the histogram 
      	numcolors = listENT(pArray, palsize,diagmsgs,context);
      	sprintf(buffer,"DIAG ConvertImage miximage Numcolors=%d \n",numcolors);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));         	
      }                                        
   
   	// Add the miximage to the background palette by adding half_colors to the index
   	// Also reset the transidx to new value
   	BOOL transidx_reset = false;
   	numcolors = 0;
   	for(int i=0; i<half_colors; i++)   // half_colors added in V6.0.5.1 was 63
   	{
         red = (BYTE)pArray[i*ENTLEN]; 
         green = (BYTE)pArray[i*ENTLEN+1]; 
         blue = (BYTE)pArray[i*ENTLEN+2]; 
         pixval = red * 65536;
   		pixval += green * 256;
   		pixval += blue;
         idx =   pArray[i*ENTLEN+3]; 
         count = pArray[i*ENTLEN+4];
         newidx = idx + half_colors; // Add half_colors to the index added in V6.0.5.1 was 63
         pArray[i*ENTLEN+5] = newidx; // save the new index
            	         
         // if this is the transidx, reset to new value
         if (idx == transidx && transidx_reset == false)
         {
            sprintf(buffer,"DIAG ConvertImage transidx reset from:%d to %d \n",transidx,newidx);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));                      
            transidx = newidx;
            transidx_reset = true;            
         }   
         
         // Add the foreground color to the background palette
         image.SetPaletteColor(newidx,red,green,blue);    
      }
  
      if ( diag>1 ) // if diag > 1 turned on print the histogram
   	{       
         sprintf(buffer,"DIAG ConvertImage histogram miximage after adding %d to idx:",half_colors);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));    
      	// list the values in the histogram 
      	numcolors = listENT(pArray, palsize,diagmsgs,context);
      	sprintf(buffer,"DIAG ConvertImage miximage Numcolors=%d transidx=%d\n",numcolors,transidx);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      }                       
                                 
      // Scan all the foreground pixels and reset the pixel index to the new value       
   	for(int newindex,y=0; y<height; y++)
      {
   		for(x=0; x<width; x++)
         {
            // Get color index and increment the count for the index
   			idx = miximage.GetPixelIndex(x,y);
   			newindex =  pArray[idx*ENTLEN+5];   // get new index value
            miximage.SetPixelIndex(x,y,newindex); // Update the pixel with new index
   		}
      } 
      
      if ( diag>1 )  // If DIAG  IS GT 1 then print all the image palette colors  
      {
         for(int i=0; i<palsize; i++)
         {
            color = image.GetPaletteColor(i);
            red = color.rgbRed;
   		   green = color.rgbGreen;
   		   blue = color.rgbBlue;
   		   sprintf(buffer,"DIAG ConvertImage DIAG image palette before doing transparancy %d: rgb(%d,%d,%d) \n",i,red,green,blue);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }
      }            

      // Scan the mix (foreground) image and if a pixel is transparent, add half_colors
      // and put back the the original background's color in the palette
      countbkg = 0;
      countfor = 0;
   	for(y=0; y<height; y++)
      {
   		for(x=0; x<width; x++)
         {
            // Get color index
            // yflip = height - y - 1;
   			fgidx = miximage.GetPixelIndex(x,y);  // Get foregrounds index for this pixel
   			// if the pixel is not transparent reset the pixel to the foregrounds index
   			if (fgidx != transidx) 
   			{
   			   image.SetPixelIndex(x,y,fgidx);  // Change the background's pixel index to the foreground's
            }   			
   		}
      }
      
      // if (miximage.IsValid()) {miximage.Destroy();}   
      if (miximage.IsValid()) 
         {
            rc = miximage.Destroy(); // Destroy the image
            if(!rc)
            {
              sprintf(buffer,"Error: ConvertImage 5. Invalid rc from Destroy Image RC=%d\n ",rc);
              printf(buffer);
              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              return 1;
            }
         }
      free(pArray); 
   } 
   
	inbpp = image.GetBpp(); 
   sprintf(buffer,"DIAG ConvertImage typein=%d typeout=%d inbpp=%d",typein,typeout,inbpp);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  

   // If the output is a tif then decrease the bpp to 8 and reduce the numcolors to maxcolors
	if (typeout == 6 ) 
	{
      int numcolors;
      // image.SetXDPI(96);
      // image.SetYDPI(96);
      // image.Light(10); removed in V9.5.9
      int height = image.GetHeight();
	   int width = image.GetWidth();
	   long x, y;
	   int idx;

      uint64_t	histogram[256] = {0};
      
      // if there is an alpha chaNNEL strip off  added in V24.0.0.19
      if(image.AlphaIsValid())
      {
         sprintf(buffer,"DIAG ConvertImage stripping off alpha channel");
          rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         // strip off the alpha layer added in V24.0.0.19
         RGBQUAD c={255,255,255,0};
   	   image.SetTransColor(c);
   	   image.AlphaStrip();
      } 
           
      /* Scan through entire image computing a histogram of colormap indices	*/
      for(y=0; y<height; y++)
      {
   		for(x=0; x<width; x++)
         {
      	  idx = image.GetPixelIndex(x,y);
      		histogram[idx++]++;
         }
      }
      
      /* Count down from highest colormap index for the first non-zero value.	*/
      /* This will indicate the maximum number of colors.						*/
      for (x = 255; x >= 0; x--)
      {
      	if (histogram[x] != 0)
      	{
      		numcolors = x + 1;		/* count of colors includes index 0 */
      		break;
      	}
      }	  

     sprintf(buffer,"DIAG ConvertImage in_numcolors=%d inbpp=%d",numcolors,inbpp);
     rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
    
	  if (numcolors <= colors && inbpp == 8)	 
     {
        
       sprintf(buffer,"DIAG ConvertImage quantizer bypassed\n");
       rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        
     } 
     else
     {     
         if (inbpp == 8 || inbpp == 24)   // new in V10.0.0  added inbpp== 24 in V10.0.3
         {
        	  CQuantizer q(colors,(colors>16?7:8));
   	     q.ProcessImage(image.GetDIB()); 
   	     ppal=(RGBQUAD*)calloc(256*sizeof(RGBQUAD),1);
   	     q.SetColorTable(ppal);
   	     // rc = image.DecreaseBpp(8,0,ppal,colors);
   	     rc = image.DecreaseBpp(8,1,ppal,colors); // V600 Set error diffusion=1
   	     outbpp = image.GetBpp(); 
   	     if (!rc)
   	     {
               sprintf(buffer,"ERROR: ConvertImage DecreaseBpp RC=%d outbpp=%d\n",rc,outbpp);
               printf(buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               return 1;	
            } 
      	   numcolors = q.GetColorCount(); 
            sprintf(buffer,"DIAG ConvertImage 2. out_numcolors=%d outbpp=%d",numcolors,outbpp);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         }        
     } 
	}
	
   // If the output is a jpg set the type
	if (typeout == 3)  
	{	
      rc = image.SetType(typeout);
	  	if ( diag>0 ) 
      {
         sprintf(buffer,"DIAG ConvertImage RC from SetType=%d",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }		     
	}
   
   // Pass back the type, width, height, X resolution and Y resolution
   int b_resx, b_resy, b_width, b_height;
   b_resx = image.GetXDPI(); // Get X resolution
   b_resy = image.GetYDPI(); // Get Y resolution
   b_width = image.GetWidth();
   b_height = image.GetHeight();
   sprintf(buffer,"DIAG ConvertImage TYPEOUT=%d XRES=%d YRES=%d ",typeout, b_resx, b_resy);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));       
   sprintf(buffer,"DIAG ConvertImage WIDTH=%d HEIGHT=%d ",b_width,b_height);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
   
   // If there is a histogram parm pass it back along with the counts
   if ( argumentExists(6) )
   {   
      RGBQUAD color;
      ULONG pixval;
      long x, y;
      uint32_t idx;
      // Get the size of the palette  in bytes
      int palsize = image.GetPaletteSize()/4;   
      sprintf(buffer,"DIAG ConvertImage palsize=%d",palsize);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      
      // Create a array to hold the histogram  of the images, each entry is 6 words long:  
      // color index,  color count, palette colors r,g,b, and pixval
   	ULONG *pArray = (ULONG*) malloc(ENTSIZE*palsize);
   	
   	// Reset the array to zero  to do the image first
   	memset(pArray,0,ENTSIZE*palsize); 
   	   	
   	// Create a histogram of the image 
      int i;
      for(i=0; i<colors; i++)   // pass back the palette entries  
   	{
         color = image.GetPaletteColor(i);
         pixval = Get_Pixval(color);           // added in V17.0.0.19
         pArray[i*ENTLEN] = i; // Save the index for this color
         pArray[i*ENTLEN+2] = color.rgbRed; 
         pArray[i*ENTLEN+3] = color.rgbGreen; 
         pArray[i*ENTLEN+4] = color.rgbBlue; 
         pArray[i*ENTLEN+5] = pixval; // Save the pixval for this color
      }
         
      // Scan all the pixels and get the palette color counts 
   	for(y=0; y<b_height; y++)
      {
   		for(x=0; x<b_width; x++)
         {
            // Get color index and increment the count for the index
   			idx = image.GetPixelIndex(x,y);
            pArray[idx*ENTLEN+1]++;	// increment the histogram count for this color	
   		}
      }        
      
      // pass back the information in the histogram  changed in V27.0.0.14 to only pass the count and pixval
      for(int i=0; i<colors; i++) // changed to colors from 127 in V6.0.2.1
      {
		   sprintf(buffer,"%010d %d",pArray[i*ENTLEN+1],pArray[i*ENTLEN+5]);    // change to 010d in V30.0.0.12
         rc = context->ArrayAppendString(histogram, buffer, strlen(buffer));
      }  

      // Free the pArray   // added in V3.0.6.9	
      if(pArray) free(pArray);                       
   }  	

   rc = image.Save(fileout,typeout);	
   if (!rc)
   {
      sprintf(buffer,"ERROR: ConvertImage Invalid rc from save image RC=%d\n",rc);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }	
	
	// if (image.IsValid()) {image.Destroy();}
   if (image.IsValid()) 
   {
      rc = image.Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: ConvertImage 2. Invalid rc from Destroy Image RC=%d\n ",rc);
        printf(buffer);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }
   }	

   return 0;
}   
/**************************** FindGrid **************************
  Takes two GE images one with the Grid on the other with it off
  and removes everything but the grid
 * Syntax:	rc = FindGrid(filein, filein2, fileout, grid_array, diag, diagmsgs, maxcolors)
 *
 * Params:		
 *			   filein1 - filename of the input file with the grid
 *			   filein2 - filename of the input file without the grid
 *			   fileout  - filename of the outputfile file
 			   grid_array - an array to send back the grid data
 *			   diag    - if specified it will print diag information 
 *			   diagmsgs    - an array to hold diagnostic and error messages  
 *			   maxcolors   -  (optional) the maximum number of colors added in V6.0.2.1 
 *
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine7(int, FindGrid, CSTRING, filein1, CSTRING, filein2, CSTRING, fileout, RexxArrayObject, grid_array,   int, diag, RexxArrayObject, diagmsgs, int, maxcolors)
{
   char buffer[512]; 
   int rc;
                                                 
   if ( diag>0 )
   {
      sprintf(buffer,"DIAG FindGrid filein1=%s filein2=%s output=%s",filein1, filein2, fileout);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   } 
   	
	ULONG inbpp1,inbpp2; 
	RGBQUAD * ppal1;
   int colors;
   
   /* DEBUR
   RexxObjectPtr dataobj;
   logical_t flag;
   RexxDirectoryObject grid_dir;
   grid_dir = (RexxDirectoryObject)data;
   flag = context->IsDirectory(grid_dir);
   if ( flag == 0 )   
   {
      sprintf(buffer,"ERROR: FindGrid parameter 4 is not a directory!");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }
   */

	// See if arg 1 is a valid image type
	TCHAR* extin1 = FindExtension(filein1);
   TCHAR* extin2 = FindExtension(filein2);
	TCHAR* extout = FindExtension(fileout);
	_tcslwr(extin1);
	int typein1 = CxImage::GetTypeIdFromName(extin1);
   _tcslwr(extin2);
   int typein2 = CxImage::GetTypeIdFromName(extin2);		
	_tcslwr(extout);
	int typeout = CxImage::GetTypeIdFromName(extout);
	
	if ( diag>0 )
   {
      sprintf(buffer,"DIAG FindGrid extin1=%s extin2=%s extout=%s typein1=%d typein2=%d typeout=%d",extin1,extin2,extout,typein1,typein2,typeout);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   } 
	
	if (typein1 == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: FindGrid Argument 1 is an invalid image type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}
   if (typein2 == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: FindGrid Argument 3 is an invalid image type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}
   // See if arg 2 is a valid image type
	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: FindGrid Argument 2 in an invalid type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;    
	}	

   // If there is an optional maxcolors  argument  set the colors added in V6.0.2.1
   colors =  maxcolors;
   sprintf(buffer,"DIAG FindGrid maximum colors %d",colors);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
	
   // Load the image with the Grid
	CxImage image1;
	if(!image1.Load(filein1,typein1))
	{
      sprintf(buffer,"ERROR: FindGrid Load image file:%s failed\n",filein1);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      if (image1.IsValid()) 
      {
         rc = image1.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: FindGrid 1. Invalid rc from Destroy Image1 RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }
      }	
      return 1;
	}
   
   image1.GrayScale();
   // rc = image1.DecreaseBpp(1,1,NULL,2); 
   
   // Load the image without the Grid
	CxImage image2;
	if(!image2.Load(filein2,typein2))
	{
      sprintf(buffer,"ERROR: FindGrid Load image file:%s failed\n",filein2);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      if (image2.IsValid()) 
      {
         rc = image2.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: FindGrid 1. Invalid rc from Destroy Image2 RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }
      }	
      return 1;
	}
   
   // Pass back the type, width, height, X resolution and Y resolution
   int b_resx, b_resy, b_width, b_height;
   b_resx = image2.GetXDPI(); // Get X resolution
   b_resy = image2.GetYDPI(); // Get Y resolution
   b_width = image2.GetWidth();
   b_height = image2.GetHeight();
   sprintf(buffer,"DIAG FindGrid TYPEOUT=%d XRES=%d YRES=%d ",typeout, b_resx, b_resy);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));       
   sprintf(buffer,"DIAG FindGrid WIDTH=%d HEIGHT=%d NUMCOLORS=%d",b_width,b_height,colors);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   image2.GrayScale();
   // rc = image2.DecreaseBpp(1,1,NULL,2); 

   inbpp1 = image1.GetBpp(); 
   inbpp2 = image2.GetBpp(); 
   int height = image1.GetHeight();
	int width = image1.GetWidth();
   sprintf(buffer,"DIAG FindGrid inbpp1=%d inbpp2=%d width=%d height=%d",inbpp1,inbpp2,width,height);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));

   // RGBQUAD * ppal;
   // if this is 8 bit or 24 bit
   if (inbpp1 == 8 || inbpp1 == 24)   
   {
      int numcolors;
      int reduced_colors = colors;
      
      // If the reduced colors = 16777216 then don't quantize or decreease the bpp added in V8.0.1.7  
      if (reduced_colors == 16777216 );
      else
      {   
        sprintf(buffer,"DIAG FindGrid reducing colors to reduced_colors=%d",reduced_colors);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        CQuantizer q1(reduced_colors,(reduced_colors>16?7:8));
        q1.ProcessImage(image1.GetDIB()); 
        ppal1=(RGBQUAD*)calloc(256*sizeof(RGBQUAD),1);
        q1.SetColorTable(ppal1);
        rc = image1.DecreaseBpp(8,1,ppal1,reduced_colors); 
        ULONG outbpp1 = image1.GetBpp(); 
        if (!rc)
        {
            sprintf(buffer,"ERROR: FindGrid DecreaseBpp RC=%d outbpp1=%d\n",rc,outbpp1);
            printf(buffer);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            return 1;	
         } 
         numcolors = q1.GetColorCount(); 
         sprintf(buffer,"DIAG FindGrid 1. image1 out_numcolors=%d outbpp1=%d",numcolors,outbpp1);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         
        CQuantizer q2(reduced_colors,(reduced_colors>16?7:8));
        q2.ProcessImage(image2.GetDIB()); 
        q2.SetColorTable(ppal1);
        rc = image2.DecreaseBpp(8,1,ppal1,reduced_colors); 
        // ULONG outbpp2 = image2.GetBpp(); 
        if (!rc)
        {
            sprintf(buffer,"ERROR: FindGrid DecreaseBpp RC=%d outbpp1=%d\n",rc,outbpp1);
            printf(buffer);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            return 1;	
         } 
         numcolors = q2.GetColorCount(); 
         sprintf(buffer,"DIAG FindGrid 1. image2 out_numcolors=%d outbpp1=%d",numcolors,outbpp1);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      } 
   }   
   
   RGBQUAD WhiteColor;
   WhiteColor.rgbRed = 255; 
   WhiteColor.rgbGreen = 255; 
   WhiteColor.rgbBlue = 255; 
   WhiteColor.rgbReserved = 0;    
   
   RGBQUAD BlackColor;
   BlackColor.rgbRed = 0; 
   BlackColor.rgbGreen = 0; 
   BlackColor.rgbBlue = 0; 
   BlackColor.rgbReserved = 0;   
   
   RGBQUAD color;   

   // Scan the images and leave only the grid
   // ULONG pixval1,pixval2;
   double pixval1,pixval2;
   int x,y,yflip;
	for(y=0; y<height; y++)
   {
      yflip = height - y - 1;
		for(x=0; x<width; x++)
      {
			color = image1.GetPixelColor(x,yflip);  // Get the color at this x,y
         pixval1 = (double)Get_Pixval(color);  
         color = image2.GetPixelColor(x,yflip);  // Get the color at this x,y
         pixval2 = (double)Get_Pixval(color); 
         
         double fabs_result = fabs(pixval1 - pixval2);
                     
			if (fabs_result < 1000000.0)  
			{  
            // make the pixel black        
			   image1.SetPixelColor(x,yflip,BlackColor,false);
         }
         else
         {
            // make the pixel white        
			   image1.SetPixelColor(x,yflip,WhiteColor,false);
         }
        		
		}
   }
   
   ULONG *V_Array = (ULONG*) malloc(ENTSIZE*1000); // Vertical array
   ULONG *H_Array = (ULONG*) malloc(ENTSIZE*1000); // Horizontal array
   
   // Determine which vertical lines are grid lines and make them so
   double y_white = 0;
   double percent = 0.0;
   int prev_x = 0;
   int v_ent_num = 0;
   for(x=0; x<width; x++)          
   {
      y_white = 0;
      // Scan down and see how many are white
		for(y=0; y<height; y++)
      {
         yflip = height - y - 1;
         color = image1.GetPixelColor(x,yflip);  // Get the color at this x,y
         pixval1 = (double)Get_Pixval(color);  
         // if not black increment the count of white
         if(pixval1 != 0) y_white ++;  
      }
      
      percent = (y_white/height) * 100; 
      
      // If 70% of the pixels are white assume its a vertical grid line.
      if(percent > 70.0) 
      {
         
         sprintf(buffer,"Vertical: x/y= %d %d y_white=%f height=%d percent=%f\n",x,y,y_white,height,percent);
         rc = context->ArrayAppendString(grid_array, buffer, strlen(buffer));
             
         V_Array[v_ent_num*ENTLEN] = x; 
         V_Array[v_ent_num*ENTLEN+1] = y; 
         V_Array[v_ent_num*ENTLEN+2] = (ULONG)percent; 
         V_Array[v_ent_num*ENTLEN+3] = (ULONG)y_white;  
         v_ent_num += 1;
         
         // Make the whole line white
         for(y=0; y<height; y++)
         {
            yflip = height - y - 1;
            image1.SetPixelColor(x,yflip,WhiteColor,false);
         } 
         prev_x = x;     
      }    
   }
   
   // Determine which horizontal lines are grid lines and make them so
   double x_white = 0;
   int prev_y = 0;
   int h_ent_num = 0;
   for(y=0; y<height; y++)          
   {
      yflip = height - y - 1;
      x_white = 0;
      // Scan across and see how many are white
		for(x=0; x<width; x++)
      {
         
         color = image1.GetPixelColor(x,yflip);  // Get the color at this x,y
         pixval1 = (double)Get_Pixval(color);  
         // if not black increment the count of white
         if(pixval1 != 0) x_white ++;  
      }
      
      percent = (x_white/width) * 100; 
      // If 70% of the pixels are white assume its a horizontal grid line.
      if(percent > 70.0)
      {
         // if(prev_y != y-1)
         // {
            sprintf(buffer,"Horizontal: x/y= %d %d x_white=%f width=%d percent=%f\n",x,y,x_white,width,percent);
            rc = context->ArrayAppendString(grid_array, buffer, strlen(buffer));
            
            H_Array[h_ent_num*ENTLEN] = x; 
            H_Array[h_ent_num*ENTLEN+1] = y; 
            H_Array[h_ent_num*ENTLEN+2] = (ULONG)percent; 
            H_Array[h_ent_num*ENTLEN+3] = (ULONG)x_white;
            h_ent_num += 1;
            
            // Make the whole line white
            for(x=0; x<width; x++)
            {
               // yflip = height - y - 1;
               image1.SetPixelColor(x,yflip,WhiteColor,false);
            }  
            prev_y = y; 
      }    
   }  
        
   rc = image1.Save(fileout,typeout);	
   if (!rc)
   {
      sprintf(buffer,"ERROR: FindGrid Invalid rc from save image RC=%d\n",rc);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }	
	
	// if (image.IsValid()) {image1.Destroy();}
   if (image1.IsValid()) 
   {
      rc = image1.Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: FindGrid 2. Invalid rc from Destroy Image RC=%d\n ",rc);
        printf(buffer);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }
   }	
   
   // if (image2.IsValid()) {image2.Destroy();}
   if (image2.IsValid()) 
   {
      rc = image2.Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: FindGrid 2. Invalid rc from Destroy Image RC=%d\n ",rc);
        printf(buffer);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }
   }	

   free(V_Array);
   free(H_Array);

   return 0;
}   
/**************************** Drawtext **************************
  Draws some text on a File
 * Syntax:	rc = Drawtext(filein, text, font, font_size, font_weight, xp,yp, colorref diag, diagmsgs, maxcolors)
 *
 * Params:	
 *       RexxObjectPtr,data -- directory block with:   
 *			   filein - filename of an image to draw on
 *			   depths   - an array of depths
 *			   xps     - an array of xp for above depth
 *			   yps     - an array of yp for above depth
 *			   font   - the font name to use 
 *			   size   - the size of the font 
 *			   weight - font weight
 *          underline -- flag to make it underlined 
 *          color  - the color for the text  s/b a COLORREF     
 *          italic -  flag to make it italic
 *	   diag    - if specified it will print diag information 
 *	   diagmsgs    - an array to save diagnostic and error messages   
 *
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine3(int, Drawtext, RexxObjectPtr,data, int, diag, RexxArrayObject, diagmsgs)      
{
   char buffer[512]; 
   int rc;
   
   logical_t flag;
   
   // Get the data directory  DEBUGRHERE
   RexxObjectPtr dataobj;
   RexxDirectoryObject data_dir;
   data_dir = (RexxDirectoryObject)data;
   flag = context->IsDirectory(data_dir);
   if ( flag == 0 )
   {
      sprintf(buffer,"ERROR: Drawtext parameter 1 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 0;
   }
   
   //Get all necessary values that are not arrays
   CSTRING filein;
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "filein"); 
   filein = context->ObjectToStringValue(dataobj); 
   
   CSTRING font;
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "font"); 
   font = context->ObjectToStringValue(dataobj); 
    
   uint32_t color;
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "color");
   flag = context->ObjectToUnsignedInt32(dataobj, &color);
   
   uint32_t iso_div;
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "iso_div");
   flag = context->ObjectToUnsignedInt32(dataobj, &iso_div);
   
   int font_size;
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "font_size");    
   flag = context->ObjectToInt32(dataobj, &font_size);  
   
   int font_weight;
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "font_weight");    
   flag = context->ObjectToInt32(dataobj, &font_weight); 
   
   logical_t italic;
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "italic");    
   flag = context->Logical(dataobj, &italic);
   
   logical_t underline;
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "underline");    
   flag = context->Logical(dataobj, &underline);
                                                                 
   RexxArrayObject xps;
   xps = (RexxArrayObject)context->DirectoryAt(data_dir, "xps");
   
   RexxArrayObject yps;
   yps = (RexxArrayObject)context->DirectoryAt(data_dir, "yps");
   
   RexxArrayObject depths;
   depths = (RexxArrayObject)context->DirectoryAt(data_dir, "depths");
   
   RexxArrayObject iso_xps;
   iso_xps = (RexxArrayObject)context->DirectoryAt(data_dir, "iso_xps");
   
   RexxArrayObject iso_yps;
   iso_yps = (RexxArrayObject)context->DirectoryAt(data_dir, "iso_yps");
   
   RexxArrayObject iso_depths;
   iso_depths = (RexxArrayObject)context->DirectoryAt(data_dir, "iso_depths");
   
   logical_t isolines;  // flag to indicate the data is isolones
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "isolines");
   flag = context->ObjectToLogical(dataobj, &isolines);  
   
   // printf("DEBUGR before after 2 values isolines=%d\n",isolines);
   
   // Red
   RGBQUAD RedColor;
   RedColor.rgbRed = 255; 
   RedColor.rgbGreen = 32; 
   RedColor.rgbBlue = 39;
   
   // Orange
   RGBQUAD OrangeColor;
   OrangeColor.rgbRed = 255; 
   OrangeColor.rgbGreen = 211; 
   OrangeColor.rgbBlue = 33;
   
   RGBQUAD BlackColor;
   BlackColor.rgbRed = 0; 
   BlackColor.rgbGreen = 0; 
   BlackColor.rgbBlue = 0; 
   BlackColor.rgbReserved = 0; 
   
   RGBQUAD WhiteColor;    // added in V15.0.5.15
   WhiteColor.rgbRed = 255; 
   WhiteColor.rgbGreen = 255; 
   WhiteColor.rgbBlue = 255; 
   WhiteColor.rgbReserved = 0; 

	// See if arg 1 is a valid image type
	TCHAR* extin = FindExtension(filein);
	_tcslwr(extin);
	int typein = CxImage::GetTypeIdFromName(extin);
	
	if (typein == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: DrawText Argument 1 is an invalid image type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}
	
   // Load the image with the Grid
	CxImage image;
   
   int height = image.GetHeight(); 
   
	if(!image.Load(filein,typein))
	{
      sprintf(buffer,"ERROR: DrawText Load image file:%s failed\n",filein);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      if (image.IsValid()) 
      {
         rc = image.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: DrawText 1. Invalid rc from Destroy Image1 RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }
      }	
      return 1;
	}
   
   RGBQUAD color_passed = image.RGBtoRGBQUAD(color); // get the color passed as a RGBQUAD
   
   // get the bpp  added in V12.0.0.4
   ULONG bpp = image.GetBpp(); 
   // if the BPP is 8 add our 3 colors
   if(bpp == 8)
   {
      // Get the size of the palette  in bytes
      int palsize = image.GetPaletteSize()/4; 
      int b_width = image.GetWidth();
      int b_height = image.GetHeight(); 
      BYTE red, green, blue;
      double pixval;
       
      // Get the number of colors used 
      BITMAPINFOHEADER ds;
      HANDLE hImage = image.GetDIB();
      
      memcpy(&ds,hImage, sizeof(ds));
      int clrused = ds.biClrUsed;
      
      sprintf(buffer,"DIAG GetDepths palsize=%d clrused=%d\n",palsize,clrused);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));     		
   
   	// Create a array to hold the histogram  of the images, each entry is 7 words long: palette colors r,g,b, color index, 
      // color count,new index  and pixval
   	ULONG *pArray = (ULONG*) malloc(ENTSIZE*palsize);
   	
   	// Reset the array to zero  to do the image first
   	memset(pArray,0,ENTSIZE*palsize); 
   
      // Fill the array with the information from the palette
   	for(int i=0; i<palsize; i++)
   	{
         RGBQUAD pal_color = image.GetPaletteColor(i);
         pixval = Get_Pixval(pal_color);    // added in V17.0.0.19
         pArray[i*ENTLEN] = pal_color.rgbRed; 
         pArray[i*ENTLEN+1] = pal_color.rgbGreen; 
         pArray[i*ENTLEN+2] = pal_color.rgbBlue; 
         pArray[i*ENTLEN+3] = i; // Save the index for this color
         pArray[i*ENTLEN+6] = (ULONG)pixval; // Save the pixval for this color
      }
      
      // Scan all the pixels and increment the palette color counts 
      int idxx, x,y;
   	for(y=0; y<b_height; y++)
      {
   		for(x=0; x<b_width; x++)
         {
            // Get color index and increment the count for the index
   			idxx = image.GetPixelIndex(x,y);
            pArray[idxx*ENTLEN+4]++;	// increment the histogram count for this color	
   		}
      }    
                       
      // Sort the array by the count value in ascending order   
      sortENT( pArray, 4, palsize, 0); 
      
      // set our colors as the lowest  and one black
      image.SetPaletteColor(0,RedColor); 
      image.SetPaletteColor(1,OrangeColor); 
      image.SetPaletteColor(2,color_passed); 
      image.SetPaletteColor(3,WhiteColor); 
      image.SetPaletteColor(4,BlackColor); 
      
      // write back the palette skipping first 5
      for(int i=5; i<palsize; i++)
      {
         pixval = pArray[i*ENTLEN+6];
         if (pixval == 0) continue;
         red = (BYTE)pArray[i*ENTLEN];
         green = (BYTE)pArray[i*ENTLEN+1];
         blue = (BYTE)pArray[i*ENTLEN+2];
         image.SetPaletteColor(i,red, green, blue);
      }
      
      // Scan all the image and change pixels  with an index 0-2 to 5 (black)
   	for(y=0; y<b_height; y++)
      {
   		for(x=0; x<b_width; x++)
         {
            // Get color index and increment the count for the index
   			idxx = image.GetPixelIndex(x,y);
            if(idxx == 0 || idxx == 1 || idxx == 2) image.SetPixelIndex(x,y,5);
   		}
      }    
      
      // Free the pArray  
      if(pArray) free(pArray);
   }
      
   // Get the nuber of depths    
   int num_depths = context->ArrayItems(depths);  
   
   RGBQUAD c;	
   int n,m,xp,yp,xp2,yp2,start_x,end_x,start_y,end_y,start_y_flip,end_y_flip;
   CSTRING depth;
   
   // Get all the depths and write on the image
   for(n=0 ; n<num_depths ; n++)  
   { 
      // get a depth
      dataobj = context->ArrayAt(depths, n+1);
      depth = context->ObjectToStringValue(dataobj);
      // get the xp for the depth
      dataobj = context->ArrayAt(xps, n+1);
      flag = context->ObjectToInt32(dataobj, &xp); 
      // get the yp for the depth
      dataobj = context->ArrayAt(yps, n+1);
      flag = context->ObjectToInt32(dataobj, &yp); 
      
      // if the depth is less than 1 meter make it red 
      double depth_dbl = strtod(depth, NULL);	
      if(depth_dbl < 1.0) c = RedColor; 
      else if(depth_dbl < 2.0) c = OrangeColor; 
      else c = color_passed; // else use the color passed
      
      // Draw the depth on the image
      image.DrawString(0,xp,yp,
               depth, // text
               c,
   				font,  // font
   				font_size,
   				font_weight,
   				(BYTE)italic,
   				(BYTE)underline,
   				true);                     
   }  
   
   // Get the nuber of iso_depths    
   int num_iso_depths = context->ArrayItems(iso_depths);  
   
   // if there are iso_depths
   {
      image.Flip();
      int iso_count = 0,tot_lines;  // count the the number of iso lines 
      // Get all the isolines
      // printf("DEBUG processing num_depths=%d\n",num_depths);
      for(n=0 ; n<num_iso_depths ; n+=2)  
      { 
         if(n >= num_iso_depths) break;
         // get a depth
         dataobj = context->ArrayAt(iso_depths, n+1);
         depth = context->ObjectToStringValue(dataobj);
         
         c = WhiteColor; // make the isolines white added in V15.0.5.15
         
         // get the xp for the depth
         dataobj = context->ArrayAt(iso_xps, n+1);
         flag = context->ObjectToInt32(dataobj, &xp); 
         // get the yp for the depth
         dataobj = context->ArrayAt(iso_yps, n+1);
         flag = context->ObjectToInt32(dataobj, &yp); 
         
         // Get the next entry for the end coordinates
         m = n +2;
         if(m >= num_iso_depths) break;
         // get a depth
         dataobj = context->ArrayAt(iso_depths, m);
         depth = context->ObjectToStringValue(dataobj);
         // get the xp for the depth
         dataobj = context->ArrayAt(iso_xps, m);
         flag = context->ObjectToInt32(dataobj, &xp2); 
         // get the yp for the depth
         dataobj = context->ArrayAt(iso_yps, m);
         flag = context->ObjectToInt32(dataobj, &yp2); 
         
         // If this is the start of a new iso line get the total lines
         if (xp2 == 999999)
         { 
            tot_lines = yp2;
            iso_count = 0;
            if (tot_lines > 5)
               iso_div = tot_lines / 5;
            // printf("DEBUG n=%d depth=%s xp/yp=%d %d tot_lines=%d iso_div=%d\n",n, depth, xp, yp, tot_lines,iso_div);
         }
         
         // printf("DEBUGR0 n=%d depth=%s xp/yp=%d %d xp2/yp2=%d %d\n",n, depth, xp, yp, xp2,yp2);
         
         // if this is the start of an isoline or every iso_div isos print the depth
         // if ((xp2 == 999999 && yp2 == 999999) || (iso_count % iso_div == 0))
         if ((xp2 == 999999) || (iso_count % iso_div == 0))  // removed yp == 999999 in V15.0.5.15
         {
         
            // printf("DEBUGR1 n=%d depth=%s xp/yp=%d %d xp2/yp2=%d %d\n",n, depth, xp, yp, xp2,yp2);
            xp -= 4;yp += 6;  // adjust the position to put it on the line
            image.Flip();
            // Draw the depth on the image
            image.DrawString(0,xp,yp,
                     depth, // text
                     c,
         				font,  // font
         				font_size,
         				font_weight,
         				(BYTE)italic,
         				(BYTE)underline,
         				true); 
            image.Flip();
         }  
         else  // else draw a line
         {
            start_x = xp;
            end_x = xp2;
            start_y = yp;
            end_y = yp2;
            start_y_flip = height - start_y - 1;   
            end_y_flip = height - end_y - 1;
            // image.Flip();
            // printf("DEBUGR2 n=%d depth=%s xp/yp=%d %d xp2/yp2=%d %d\n",n, depth, xp, yp, xp2,yp2);
            // image.DrawLine	(start_x,end_x,start_y_flip,end_y_flip,c,false);
            image.DrawLine	(start_x,end_x,start_y,end_y,c,false);
            // image.Flip();
         }  
         iso_count += 1;      
      }
      image.Flip();
   }
   
   
   // Pass back the type, width, height, X resolution and Y resolution
   int b_resx, b_resy, b_width, b_height;
   b_resx = image.GetXDPI(); // Get X resolution
   b_resy = image.GetYDPI(); // Get Y resolution
   b_width = image.GetWidth();
   b_height = image.GetHeight();
   sprintf(buffer,"DIAG DrawText XRES=%d YRES=%d ",b_resx, b_resy);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));       
   sprintf(buffer,"DIAG DrawText WIDTH=%d HEIGHT=%d",b_width,b_height);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
        
   rc = image.Save(filein,typein);	
   if (!rc)
   {
      sprintf(buffer,"ERROR: DrawText Invalid rc from save image RC=%d\n",rc);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }	
	
	// if (image.IsValid()) {image.Destroy();}
   if (image.IsValid()) 
   {
      rc = image.Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: DrawText 2. Invalid rc from Destroy Image RC=%d\n ",rc);
        printf(buffer);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }
   }	

   return 0;
}                                                   
/**************************** Colorize_depths **************************
  Colozizes each depth on the image
 * Syntax:	rc = Colorize_depths(filein, text, font, font_size, font_weight, xp,yp, colorref diag, diagmsgs, maxcolors)
 *
 * Params:		
 *			   filein - filename of an image to draw on
 *			   depths   - an array of depths
 *			   xps     - an array of xp for above depth
 *			   yps     - an array of yp for above depth
 *			   font   - the font name to use 
 *			   size   - the size of the font 
 *			   weight - font weight
            color  - the color for the text  s/b a COLORREF
 *			   diag    - if specified it will print diag information 
 *			   diagmsgs    - an array to hold diagnostic and error messages   
 *
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine10(int, Colorize_depths, CSTRING, filein, RexxArrayObject, depths, RexxArrayObject, xps, RexxArrayObject, yps, CSTRING, font, int, size, int, weight, int, color, int, diag, RexxArrayObject, diagmsgs)
             
{
   char buffer[512]; 
   int rc;
                                                 
   if ( diag>0 )
   {
      sprintf(buffer,"DIAG Colorize_depths filein=%s font=%s",filein, font);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   } 
   
   // Red
   RGBQUAD RedColor;
   RedColor.rgbRed = 255; 
   RedColor.rgbGreen = 32; 
   RedColor.rgbBlue = 39;
   
   // Orange
   RGBQUAD OrangeColor;
   OrangeColor.rgbRed = 255; 
   OrangeColor.rgbGreen = 211; 
   OrangeColor.rgbBlue = 33;

	// See if arg 1 is a valid image type
	TCHAR* extin = FindExtension(filein);
	_tcslwr(extin);
	int typein = CxImage::GetTypeIdFromName(extin);
	
	if (typein == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: Colorize_depths Argument 1 is an invalid image type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}
	
   // Load the image with the Grid
	CxImage image;
   
	if(!image.Load(filein,typein))
	{
      sprintf(buffer,"ERROR: Colorize_depths Load image file:%s failed\n",filein);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      if (image.IsValid()) 
      {
         rc = image.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: Colorize_depths 1. Invalid rc from Destroy Image RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }
      }	
      return 1;
	}
   
   // Flip it
   image.Flip();
   
   // Get the number of depths    
   int num_depths = context->ArrayItems(depths);  
   
   // Get all the depths and write on the image
   int n,xp,yp;
   CSTRING depth;
   RexxObjectPtr dataobj;
   logical_t flag;
   // CXTEXTINFO m_text;
   // CxImage::tagCxTextInfo m_text;
   for(n=0 ; n<num_depths ; n++)  
   { 
      // get a depth
      dataobj = context->ArrayAt(depths, n+1);
      depth = context->ObjectToStringValue(dataobj);
      // get the xp for the depth
      dataobj = context->ArrayAt(xps, n+1);
      flag = context->ObjectToInt32(dataobj, &xp); 
      // get the yp for the depth
      dataobj = context->ArrayAt(yps, n+1);
      flag = context->ObjectToInt32(dataobj, &yp); 
      // if the depth is less than 1 meter make it red 
      RGBQUAD c;	
      double depth_dbl = strtod(depth, NULL);	
      if(depth_dbl < 1.0) c = RedColor; 
      else if(depth_dbl < 2.0) c = OrangeColor; 
      else c = image.RGBtoRGBQUAD(color); // else use the color passed
      ULONG font_weight = 400;
      // set the pixel color
      image.SetPixelColor(xp,yp,c,false);            
   }  
   
   // Pass back the type, width, height, X resolution and Y resolution
   int b_resx, b_resy, b_width, b_height;
   b_resx = image.GetXDPI(); // Get X resolution
   b_resy = image.GetYDPI(); // Get Y resolution
   b_width = image.GetWidth();
   b_height = image.GetHeight();
   sprintf(buffer,"DIAG Colorize_depths XRES=%d YRES=%d ",b_resx, b_resy);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));       
   sprintf(buffer,"DIAG Colorize_depths WIDTH=%d HEIGHT=%d",b_width,b_height);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
        
   rc = image.Save(filein,typein);	
   if (!rc)
   {
      sprintf(buffer,"ERROR: Colorize_depths Invalid rc from save image RC=%d\n",rc);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }	
   
   // Flip it  back
   image.Flip();
	
	// if (image.IsValid()) {image1.Destroy();}
   if (image.IsValid()) 
   {
      rc = image.Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: Colorize_depths 2. Invalid rc from Destroy Image RC=%d\n ",rc);
        printf(buffer);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }
   }	

   return 0;
}   
/**************************** GetDepths **************************
  Takes an imge of a SasPlanet or C-Map chart and removes all but the depths 
  and saves it to an output file
 * Syntax:	rc = GetDepths(filein, fileout)
 *
 * Params:	data - directory of stuff like: threshold and isolines	
 *			   fileinn  - filename of the colored bmp screenshot of sasplanet or C-map or geo-reffrenced image
 *			   fileout  - filename of the outputfile file
 *			   diag     - comm._diag
 *			   diagmsgs - out messages array
 *
 * Return: 0 if OK else 1
 *********************************************************************/
// DEBUGR RexxRoutine5(int, GetDepths, CSTRING, filein,CSTRING, fileout, uint32_t, threshold, int, diag, RexxArrayObject, diagmsgs)
RexxRoutine5(int, GetDepths, RexxObjectPtr,data, CSTRING, filein,CSTRING, fileout, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512]; 
   int rc;
   logical_t flag;
   
   RexxObjectPtr dataobj;
   RexxDirectoryObject data_dir;
   
   data_dir = (RexxDirectoryObject)data;
   flag = context->IsDirectory(data_dir);
   if ( flag == 0 )
   {
      sprintf(buffer,"ERROR: GetDepths parameter 2 is not a directory!");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return -1;
   }
   
   int32_t threshold;
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "THRESHOLD");
   flag = context->ObjectToInt32(dataobj, &threshold); 
   
   logical_t isolines;
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "ISOLINES");
   flag = context->ObjectToLogical(dataobj, &isolines);  
                                              
   sprintf(buffer,"DIAG GetDepths threshold=%d isolines=%d filein=%s fileout=%s",threshold, isolines, filein, fileout);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   RGBQUAD WhiteColor;
   WhiteColor.rgbRed = 255; 
   WhiteColor.rgbGreen = 255; 
   WhiteColor.rgbBlue = 255; 
   WhiteColor.rgbReserved = 0;    
   
   RGBQUAD BlackColor;
   BlackColor.rgbRed = 0; 
   BlackColor.rgbGreen = 0; 
   BlackColor.rgbBlue = 0; 
   BlackColor.rgbReserved = 0;   
   
	// See if arg 1 is a valid image type
	TCHAR* extin1 = FindExtension(filein);
	TCHAR* extout = FindExtension(fileout);
	_tcslwr(extin1);
	int typein1 = CxImage::GetTypeIdFromName(extin1);
	_tcslwr(extout);
	int typeout = CxImage::GetTypeIdFromName(extout);
	
   sprintf(buffer,"DIAG GetDepths extin1=%s extout=%s typein1=%d typeout=%d",extin1,extout,typein1,typeout);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
	
	if (typein1 == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: GetDepths Argument 1 is an invalid image type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}
   
   // See if arg 2 is a valid image type
	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: GetDepths Argument 2 in an invalid type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;    
	}	
	
   // Load the image with the input image
	CxImage image1;
	if(!image1.Load(filein,typein1))
	{
      sprintf(buffer,"ERROR: GetDepths Load image file:%s failed\n",filein);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      if (image1.IsValid()) 
      {
         rc = image1.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: GetDepths 1. Invalid rc from Destroy Image1 RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }
      }	
      return 1;
	}
   
   // Pass back the type, width, height, X resolution and Y resolution
   int b_resx, b_resy, width, height;
   b_resx = image1.GetXDPI(); // Get X resolution
   b_resy = image1.GetYDPI(); // Get Y resolution
   width = image1.GetWidth();
   height = image1.GetHeight();
   sprintf(buffer,"DIAG GetDepths TYPEOUT=%d XRES=%d YRES=%d ",typeout, b_resx, b_resy);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));       
   sprintf(buffer,"DIAG GetDepths WIDTH=%d HEIGHT=%d ",width,height);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));   
   
   // Reduce the colors to 256
   int colors = 256; 
   RGBQUAD * ppal;
   int numcolors;
   CQuantizer q(colors,(colors>16?7:8));
   q.ProcessImage(image1.GetDIB()); 
   ppal=(RGBQUAD*)calloc(256*sizeof(RGBQUAD),1);
   q.SetColorTable(ppal);
   rc = image1.DecreaseBpp(8,1,ppal,colors); // V600 Set error diffusion=1
   ULONG outbpp = image1.GetBpp(); 
   if (!rc)
   {
      sprintf(buffer,"ERROR: GetDepths DecreaseBpp RC=%d outbpp=%d\n",rc,outbpp);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;	
   } 

   // Count the number of colors
   numcolors = q.GetColorCount(); 
   sprintf(buffer,"DIAG GetDepths out_numcolors=%d outbpp=%d width=%d height=%d",numcolors,outbpp, width, height);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   
   RGBQUAD color;
	ULONG pixval;
	int red,green,blue;
	int count=0;
   long x, y;
	uint8_t idx;
	
	// Get the size of the palette  in bytes
   int palsize = image1.GetPaletteSize()/4;   
   // Get the number of colors used 
   BITMAPINFOHEADER ds;
   HANDLE hImage = image1.GetDIB();
   memcpy(&ds,hImage, sizeof(ds));
   int clrused = ds.biClrUsed;
   sprintf(buffer,"DIAG GetDepths palsize=%d clrused=%d\n",palsize,clrused);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));     		

	// Create a array to hold the histogram  of the images, each entry is 7 words long: palette colors r,g,b, color index, 
   // color count,new index  and pixval
	ULONG *pArray = (ULONG*) malloc(ENTSIZE*palsize);
	
	// Reset the array to zero  to do the image first
	memset(pArray,0,ENTSIZE*palsize); 

   // Fill the array with the information from the palette
	for(int i=0; i<palsize; i++)
	{
      color = image1.GetPaletteColor(i);
      pixval = Get_Pixval(color);    // added in V17.0.0.19
      pArray[i*ENTLEN] = color.rgbRed; 
      pArray[i*ENTLEN+1] = color.rgbGreen; 
      pArray[i*ENTLEN+2] = color.rgbBlue; 
      pArray[i*ENTLEN+3] = i; // Save the index for this color
      pArray[i*ENTLEN+6] = pixval; // Save the pixval for this color
   }
      
   // Scan all the pixels and increment the palette color counts 
	for(y=0; y<height; y++)
   {
		for(x=0; x<width; x++)
      {
         // Get color index and increment the count for the index
			idx = image1.GetPixelIndex(x,y);
         pArray[idx*ENTLEN+4]++;	// increment the histogram count for this color	
		}
   }    
                          
   // Sort the array by the count value in descending order   
   sortENT( pArray, 4, palsize, 1);  	
   
   // Scan the array histogram and see if there is a normal land color
	uint32_t lowrange = 0;
	uint32_t hirange = 0;
	ULONG landcolor = -1;
	ULONG landidx =0;                              
	for(int i=0; (i<palsize && landcolor == -1); i++)
	{
      pixval = pArray[i*ENTLEN+6];
      if (pixval == 0) continue;
      red = pArray[i*ENTLEN];
      green = pArray[i*ENTLEN+1];
      blue = pArray[i*ENTLEN+2];
      idx = (uint8_t)pArray[i*ENTLEN+3];
      count = pArray[i*ENTLEN+4];

      //  Changed the following to accommodate ENC charts            11899449  11842484            added test for green in V9.5.0
      // if (pixval < 11800000 || pixval > 17097568 || blue > 202 || green < 140) continue; 
      if (pixval >= 11800000 && pixval <= 17097568 && blue <= 202 && green >= 140)
      { // forget entry if it does not meet land criteria
         // save the landcolor and  its index
         landcolor = pixval; landidx=pArray[i*ENTLEN+3];
         sprintf(buffer,"DIAG GetDepths Criteria MET idx=%d rgb(%d,%d,%d)(%d) count=%d\n",idx,red,green,blue, pixval,count);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         // break;
      } 
   }   
   
   // If no landcolor was found, scan again looking for a land color like the one in the NGA charts eg: 24400.kap
   if (landcolor == -1) 
   {
      for(int i=0; (i<palsize && landcolor == -1); i++)
      {
         pixval = pArray[i*ENTLEN+6];
         if (pixval == 0) continue;
         red = pArray[i*ENTLEN];
         green = pArray[i*ENTLEN+1];
         blue = pArray[i*ENTLEN+2];
         idx = (uint8_t)pArray[i*ENTLEN+3];
         count = pArray[i*ENTLEN+4];
         // Added in V9.5.2   to accomodate those strange charts    14867688   14867688   14413809
         if (pixval >= 14413800 && pixval <= 15129400)
         {
            // save the landcolor and  its index
            landcolor = pixval; landidx=pArray[i*ENTLEN+3];
            sprintf(buffer,"DIAG GetDepths Criteria MET Strange idx=%d rgb(%d,%d,%d)(%d) count=%d\n",idx,red,green,blue, pixval,count);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            // break;    
         }
      }      
   }      
   
   // If no landcolor was found, pass back a message to the caller 
   if (landcolor == -1)   
   {
      sprintf(buffer,"GetDepths no landcolor was found\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }
   else //  else landcolor was found set the range
   {
      // If the fuzz is valid, then set the range with it
      int range = 0;
      if (threshold>0)
      {     
         range = landcolor / 100 * threshold;
         lowrange = landcolor - range;
         hirange = landcolor + range;
      }
      else {lowrange=landcolor; hirange=landcolor;}
   
      sprintf(buffer,"DIAG GetDepths landcolor=%d landindex=%d range=%d lowrange=%d hirange=%d\n",landcolor, landidx, range, lowrange, hirange);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
       
   	// Set the background color to the landcolor
   	red = landcolor/65536;
      ULONG pixval2 = landcolor - red*65536;
      green = pixval2/256;
      pixval2 -= green*256;
      blue = pixval2;
      RGBQUAD BkgndColor;
      // RGBQUAD * ppal;
      BkgndColor.rgbRed = red; 
      BkgndColor.rgbGreen = green; 
      BkgndColor.rgbBlue = blue; 	
       
      sprintf(buffer,"DIAG GetDepths landcolor=%d rgb(%d,%d,%d)\n",landcolor,red,green,blue);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }        
   
   // Scan the image and make all land black
   int yflip;
	for(y=0; y<height; y++)
   {
      yflip = height - y - 1;
		for(x=0; x<width; x++)
      {
			color = image1.GetPixelColor(x,yflip);  // Get the color at this x,y  4276545 
         pixval = Get_Pixval(color); 
         // if isolines is checked added in V15.0.5.8   
         if (isolines == 0) 
         {   
            // Remove any lines and the Navionics water mark
            if(pixval == 609908 || pixval == 4276545 || pixval == 80219 || pixval == 5461587 || 
               pixval == 4210752 || pixval == 14426 || pixval == 5395794 || pixval == 14803425) 
                   image1.SetPixelColor(x,yflip,WhiteColor,false);
         }
   
         if(color.rgbGreen > 183 &&  (color.rgbBlue == 0 || color.rgbBlue == 1))  // if its the green Sasplanet reef color
            image1.SetPixelColor(x,yflip,WhiteColor,false);  // make it white later will be black
         else // else check if its a land color
         {   
            pixval = Get_Pixval(color); 
            // set any land to white  
            if (pixval >= lowrange && pixval <= hirange && color.rgbBlue <= 202 && color.rgbGreen >= 140)
               image1.SetPixelColor(x,yflip,WhiteColor,false);  // make it white later will be black
         }
		}
   }
   
   // set the Bpp to 1 (black and white)     
   rc = image1.DecreaseBpp(1,0,0,0); 
   
   // Scan the image and reverse white and black
	for(y=0; y<height; y++)
   {
      yflip = height - y - 1;
		for(x=0; x<width; x++)
      {
			color = image1.GetPixelColor(x,yflip);  // Get the color at this x,y
         pixval = Get_Pixval(color); 
         // if its black make it white   
         if(pixval == 0) image1.SetPixelColor(x,yflip,WhiteColor,false);
         else image1.SetPixelColor(x,yflip,BlackColor,false);  // else make it black
		}
   }
         
   rc = image1.Save(fileout,typeout);	
   if (!rc)
   {
      sprintf(buffer,"ERROR: GetDepths Invalid rc from save image RC=%d\n",rc);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }	
	
	// if (image.IsValid()) {image1.Destroy();}
   if (image1.IsValid()) 
   {
      rc = image1.Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: GetDepths 2. Invalid rc from Destroy Image RC=%d\n ",rc);
        printf(buffer);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }
   }	
   
   free(pArray);	
   
   return 0;
}   
/**************************** MergeFGBG **************************
  Mixes two images together one being a foreground image that has a land color
  and the other the being the GE backgound.  Both are 24 bit RGBA images.
  It fill first read the foreground image and determine the land color then destoys it.  
  Then read in the foreground and background image and replaces any pixels in the foreground
  that landcolor with the background pixels.
  
 * Syntax:	rc = MergeFGBG(filein, fileout, threshold, diag, diagmsgs, maxcolors)
 *
 * Params:	
 *          data     -  a directory of other data added in V11.0.5.3
 *			   filein   - filename of the foreground input file
 *			   filein2  - filename of the backgound (Google earth) file
 *			   fileout  - filename of the final output file
 *			   threshold - the Threshold percentage for the land colors 
 *			   diag    - if specified it will print diag information 
 *			   diagmsgs    - an array to hold diagnostic and error messages  
 *			   maxcolors   - the maximum number of colors 
 *			   make_water_transparent - make the water transparent
 *			   depths - optional a file with just the depths
 *
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine10(int, MergeFGBG, RexxObjectPtr, data, CSTRING, filein, CSTRING, filein2,  CSTRING, fileout, uint32_t, threshold, int, diag, RexxArrayObject, diagmsgs, int, maxcolors, int, make_water_transparent,OPTIONAL_CSTRING,depths)
{
   char buffer[512]; 
   int rc;
   logical_t flag;
                                                  
   sprintf(buffer,"DIAG MergeFGBG make_water_transparent=%d threshold=%d filein=%s filein2=%s fileout=%s",make_water_transparent, threshold, filein, filein2, fileout);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   RexxObjectPtr dataobj;
   RexxDirectoryObject data_dir;
 
   data_dir = (RexxDirectoryObject)data;
   flag = context->IsDirectory(data_dir);
   if ( flag == 0 )
   {
      sprintf(buffer,"ERROR: MergeFGBG parameter 1 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 0;
   }

	// See if filein is a valid image type
	TCHAR* extin = FindExtension(filein);
	TCHAR* extin2 = FindExtension(filein2);
   TCHAR* extout = FindExtension(fileout);
	_tcslwr(extin);
	int typein = CxImage::GetTypeIdFromName(extin);	
	_tcslwr(extin2);
	int typein2 = CxImage::GetTypeIdFromName(extin2);
	_tcslwr(extout);
	int typeout = CxImage::GetTypeIdFromName(extout);
	
   sprintf(buffer,"DIAG MergeFGBG extin=%s extout=%s typein=%d typein2=%d typeout=%d ENTSIZE=%d",extin,extin2,typein,typein2,typeout, ENTSIZE);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
	
	if (typein == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: MergeFGBG Argument 1 is an invalid image type\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}

   // See if arg 2 is a valid image type
	if (typein2 == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: MergeFGBG Argument 2 in an invalid type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;    
	}	

   // See if arg 3 is a valid image type
	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: MergeFGBG Argument 2 in an invalid type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;    
	}	
   
   double latdtm;
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "LATDTM");
   flag = context->ObjectToDouble(dataobj, &latdtm);
   
   double londtm;
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "LONDTM");
   flag = context->ObjectToDouble(dataobj, &londtm);
   
   double ppd_lat;
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "PPD_LAT");
   flag = context->ObjectToDouble(dataobj, &ppd_lat);
   
   double ppd_lon;
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "PPD_LON");
   flag = context->ObjectToDouble(dataobj, &ppd_lon);
   
   int make_transparent;  // Get make land transparent
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "MAKE_TRANSPARENT");
   flag = context->ObjectToInt32(dataobj, &make_transparent);
   
   int lat_off = int(latdtm / 3600 * ppd_lat);
   int lon_off = int(londtm / 3600 * ppd_lon);
   
   sprintf(buffer,"DIAG MergeFGBG make_transparent=%d lat/lon dtm=%f %f ppd_lat/lon=%f %f lat/lon_off=%d %d",make_transparent, latdtm, londtm, ppd_lat, ppd_lon,lat_off,lon_off);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
   
   sprintf(buffer,"DIAG MergeFGBG lat/lon dtm=%f %f ppd_lat/lon=%f %f",latdtm, londtm, ppd_lat, ppd_lon);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
   
   RGBQUAD color;
   ULONG pixval,pixval2;
   int count=0;
   long x, y;
   CxImage image;
   CxImage image2;
   int height;
	int width;

   // if the make water transparent is off
   if(make_water_transparent == 0)
   {
      //  load the foreground image and determine the land color
      
   	if(!image.Load(filein,typein))
   	{	
         sprintf(buffer,"ERROR: MergeFGBG Load image file:%s failed\n",filein);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         // image.Destroy();
         if (image.IsValid()) 
         {
            rc = image.Destroy(); // Destroy the image
            if(!rc)
            {
              sprintf(buffer,"Error: MergeFGBG 1. Invalid rc from Destroy Image RC=%d\n ",rc);
              printf(buffer);
              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              return 1;
            }
         }
         return 1;
   	}	
   
      ULONG inbpp = image.GetBpp();  // Added in V6.0.1.2
      sprintf(buffer,"DIAG MergeFGBG inbpp=%d",inbpp);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   
      // if the inbpp is not 24 its an error
      if(inbpp != 24)
   	{	
         sprintf(buffer,"ERROR: MergeFGBG the inbpp is not 24 it is%d",inbpp);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         // image.Destroy();
         if (image.IsValid()) 
         {
            rc = image.Destroy(); // Destroy the image
            if(!rc)
            {
              sprintf(buffer,"Error: MergeFGBG 1. Invalid rc from Destroy Image RC=%d\n ",rc);
              printf(buffer);
              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              return 1;
            }
         }
         return 1;
   	}	
   
      RGBQUAD * ppal;
      int colors = maxcolors; 
      int numcolors;
      
      RGBQUAD WhiteColor;
      WhiteColor.rgbRed = 255; 
      WhiteColor.rgbGreen = 255; 
      WhiteColor.rgbBlue = 255; 
      WhiteColor.rgbReserved = 0; 
      
      // Reduce the colors to 63 or 256
      CQuantizer q(colors,(colors>16?7:8));
      q.ProcessImage(image.GetDIB()); 
      ppal=(RGBQUAD*)calloc(256*sizeof(RGBQUAD),1);
      q.SetColorTable(ppal);
      rc = image.DecreaseBpp(8,1,ppal,colors); // V600 Set error diffusion=1
      ULONG outbpp = image.GetBpp(); 
      if (!rc)
      {
         sprintf(buffer,"ERROR: MergeFGBG DecreaseBpp RC=%d outbpp=%d\n",rc,outbpp);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;	
      } 
   
      // Count the number of colors
      numcolors = q.GetColorCount(); 
      height = image.GetHeight();
	   width = image.GetWidth();
      sprintf(buffer,"DIAG MergeFGBG out_numcolors=%d outbpp=%d width=%d height=%d",numcolors,outbpp, width, height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));       
   	 
   	
   	int red,green,blue;
   	uint8_t idx;
   	
   	// Get the size of the palette  in bytes
      int palsize = image.GetPaletteSize()/4;   
      // Get the number of colors used 
      BITMAPINFOHEADER ds;
      HANDLE hImage = image.GetDIB();
      memcpy(&ds,hImage, sizeof(ds));
      int clrused = ds.biClrUsed;
      sprintf(buffer,"DIAG MergeFGBG palsize=%d clrused=%d\n",palsize,clrused);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));     		
   
   	// Create a array to hold the histogram  of the images, each entry is 7 words long: palette colors r,g,b, color index, 
      // color count,new index  and pixval
   	ULONG *pArray = (ULONG*) malloc(ENTSIZE*palsize);
   	
   	// Reset the array to zero  to do the image first
   	memset(pArray,0,ENTSIZE*palsize); 
   
      // Fill the array with the information from the palette
   	for(int i=0; i<palsize; i++)
   	{
         color = image.GetPaletteColor(i);
         pixval = Get_Pixval(color);    // added in V17.0.0.19
         pArray[i*ENTLEN] = color.rgbRed; 
         pArray[i*ENTLEN+1] = color.rgbGreen; 
         pArray[i*ENTLEN+2] = color.rgbBlue; 
         pArray[i*ENTLEN+3] = i; // Save the index for this color
         pArray[i*ENTLEN+6] = pixval; // Save the pixval for this color
      }
         
      // Scan all the pixels and increment the palette color counts 
   	for(y=0; y<height; y++)
      {
   		for(x=0; x<width; x++)
         {
            // Get color index and increment the count for the index
   			idx = image.GetPixelIndex(x,y);
            pArray[idx*ENTLEN+4]++;	// increment the histogram count for this color	
   		}
      }    
                             
      // Sort the array by the count value in descending order   
      sortENT( pArray, 4, palsize, 1);  	
      
      // Scan the array histogram and see if there is a normal land color
   	uint32_t lowrange = 0;
   	uint32_t hirange = 0;
   	ULONG landcolor = -1;
   	ULONG landidx =0;                              
   	for(int i=0; (i<palsize && landcolor == -1); i++)
   	{
         pixval = pArray[i*ENTLEN+6];
         if (pixval == 0) continue;
         red = pArray[i*ENTLEN];
         green = pArray[i*ENTLEN+1];
         blue = pArray[i*ENTLEN+2];
         idx = (uint8_t)pArray[i*ENTLEN+3];
         count = pArray[i*ENTLEN+4];
   
         //  Changed the following to accommodate ENC charts            11899449  11842484            added test for green in V9.5.0
         // if (pixval < 11800000 || pixval > 17097568 || blue > 202 || green < 140) continue; 
         if (pixval >= 11800000 && pixval <= 17097568 && blue <= 202 && green >= 140)
         { // forget entry if it does not meet land criteria
            // save the landcolor and  its index
            landcolor = pixval; landidx=pArray[i*ENTLEN+3];
            sprintf(buffer,"DIAG MergeFGBG Criteria MET idx=%d rgb(%d,%d,%d)(%d) count=%d\n",idx,red,green,blue, pixval,count);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            // break;
         } 
      }   
      
      // If no landcolor was found, scan again looking for a land color like the one in the NGA charts eg: 24400.kap
      if (landcolor == -1) 
      {
         for(int i=0; (i<palsize && landcolor == -1); i++)
         {
            pixval = pArray[i*ENTLEN+6];
            if (pixval == 0) continue;
            red = pArray[i*ENTLEN];
            green = pArray[i*ENTLEN+1];
            blue = pArray[i*ENTLEN+2];
            idx = (uint8_t)pArray[i*ENTLEN+3];
            count = pArray[i*ENTLEN+4];
            // Added in V9.5.2   to accomodate those strange charts    14867688   14867688   14413809
            if (pixval >= 14413800 && pixval <= 15129400)
            {
               // save the landcolor and  its index
               landcolor = pixval; landidx=pArray[i*ENTLEN+3];
               sprintf(buffer,"DIAG MergeFGBG Criteria MET Strange idx=%d rgb(%d,%d,%d)(%d) count=%d\n",idx,red,green,blue, pixval,count);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
               // break;    
            }
         }      
      }       
      
      // If no landcolor was found, pass back a message to the caller 
      if (landcolor == -1)   
      {
         sprintf(buffer,"MergeFGBG no landcolor was found\n");
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }
      else //  else landcolor was found set the range
      {
         // If the fuzz is valid, then set the range with it
         int range = 0;
         if (threshold>0)
         {     
            range = landcolor / 100 * threshold;
            lowrange = landcolor - range;
            hirange = landcolor + range;
         }
         else {lowrange=landcolor; hirange=landcolor;}
      
         sprintf(buffer,"DIAG MergeFGBG landcolor=%d landindex=%d range=%d lowrange=%d hirange=%d\n",landcolor, landidx, range, lowrange, hirange);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
          
      	// Set the background color to the landcolor
      	red = landcolor/65536;
         pixval2 = landcolor - red*65536;
         green = pixval2/256;
         pixval2 -= green*256;
         blue = pixval2;
         RGBQUAD BkgndColor;
         // RGBQUAD * ppal;
         BkgndColor.rgbRed = red; 
         BkgndColor.rgbGreen = green; 
         BkgndColor.rgbBlue = blue; 	
          
         sprintf(buffer,"DIAG MergeFGBG landcolor=%d rgb(%d,%d,%d)\n",landcolor,red,green,blue);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }
   
      // Destroy the image and free the histogram array
      // if (image.IsValid()) {image.Destroy();}
      if (image.IsValid()) 
      {
         rc = image.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: MergeFGBG 3. Invalid rc from Destroy Image RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }
      }
      
      free(pArray);	

      // Reload the foreground image
   	if(!image.Load(filein,typein))
   	{	
         sprintf(buffer,"ERROR: MergeFGBG Load image file:%s failed\n",filein);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         // image.Destroy();
         if (image.IsValid()) 
         {
            rc = image.Destroy(); // Destroy the image
            if(!rc)
            {
              sprintf(buffer,"Error: MergeFGBG 1. Invalid rc from Destroy Image RC=%d\n ",rc);
              printf(buffer);
              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              return 1;
            }
         }
         return 1;
   	}	 
   
      BITMAPINFOHEADER ds2;
      hImage = image.GetDIB();
      memcpy(&ds2,hImage, sizeof(ds2));
      clrused = ds2.biClrUsed;
      sprintf(buffer,"DIAG MergeFGBG foreground clrused=%d",clrused);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
   
      // If there is a valid land color load the background and then replace all the pixels in the foreground with any that are land color 
      if (landcolor != -1)  
      {
         // Load the background image
      	if(!image2.Load(filein2,typein2))
      	{	
            sprintf(buffer,"ERROR: MergeFGBG Load image file:%s failed\n",filein2);
            printf(buffer);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            if (image.IsValid()) 
            {
               rc = image2.Destroy(); // Destroy the image
               if(!rc)
               {
                 sprintf(buffer,"Error: MergeFGBG 1. Invalid rc from Destroy Image RC=%d\n ",rc);
                 printf(buffer);
                 rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                 return 1;
               }
            }
            return 1;
      	}	
         
         BITMAPINFOHEADER ds3;
         hImage = image.GetDIB();
         memcpy(&ds3,hImage, sizeof(ds3));
         clrused = ds3.biClrUsed;
         sprintf(buffer,"DIAG MergeFGBG background clrused=%d",clrused);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      
         ULONG inbpp2 = image2.GetBpp();  // Added in V6.0.1.2
         sprintf(buffer,"DIAG MergeFGBG inbpp2=%d",inbpp2);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      
         // if the inbpp is not 24 its an error
         if(inbpp != 24)
      	{	
            sprintf(buffer,"ERROR: MergeFGBG the inbpp2 is not 24 it is%d",inbpp);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            // image.Destroy();
            if (image.IsValid()) 
            {
               rc = image2.Destroy(); // Destroy the image
               if(!rc)
               {
                 sprintf(buffer,"Error: MergeFGBG 1. Invalid rc from Destroy Image RC=%d\n ",rc);
                 printf(buffer);
                 rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                 return 1;
               }
            }
            return 1;
      	}	
       
         // if the make land transparent is checked
         if(make_transparent == 1) 
         {
            sprintf(buffer,"DIAG MergeFGBG making land tranparent make_transparent=%d",make_transparent);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
         
            // Scan both images and replace any pixels in the foreground that have the landcolor with the same pixel from the background
            count = 0;
         	for(y=0; y<height; y++)
            {
               int yflip = height - y - 1;
               // LONG idxcnt;
               // uint32_t idx;
         		for(x=0; x<width; x++)
               {
         			color = image.GetPixelColor(x,yflip);      
                  pixval = Get_Pixval(color);    // added in V17.0.0.19
         			if (pixval >= lowrange && pixval <= hirange && color.rgbBlue <= 202 && color.rgbGreen >= 140)  // added test for green in V9.5.0
         			// if ((pixval >= lowrange && pixval <= hirange && color.rgbBlue <= 202 && color.rgbGreen >= 140) || (pixval >= 2000000 && pixval <= 11400000) )  
                  //if ((pixval >= lowrange && pixval <= hirange && color.rgbBlue <= 202 && color.rgbGreen >= 140)|| (color.rgbBlue >= 199 && color.rgbBlue <= 255) ) 
                  {
                     if ( diag>2 )
                     {
                        sprintf(buffer,"DIAG MergeFGBG X=%d Y=%d red=%d green=%d blue=%d pixval=%d\n",x,y,color.rgbRed,color.rgbGreen,color.rgbBlue,pixval);
                        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                     }  
                     // replace with the GE background color
                     color = image2.GetPixelColor(x,yflip);          
         			   image.SetPixelColor(x,yflip,color,false);
         			   count += 1;
                  }	
                  else  // not a land color See if we should make the water transparent added in V11.0.5.1
                  {
                     // if there is a depths file do nothing
                     if (argumentExists(10));
                     else  // It must sasplanet
                     {  
                        if(make_water_transparent) // Should we make the land transparent?  
                        {      
                           bool make_white = false;    
                           if (color.rgbGreen < 150 && color.rgbBlue < 199) // must be a depth unit from sasplanet depth
                              make_white = true; 
                           if(make_white)   
                           {
                              image.SetPixelColor(x,yflip,WhiteColor,false);  // Make it white
                           }
                           else  // make it the GE pixval
                           {
                              // replace with the background color
                              color = image2.GetPixelColor(x,yflip);          
                  			   image.SetPixelColor(x,yflip,color,false);
                  			   count += 1;
                           }
                        }
                     }
                  }		
         		}
            }
         }
      }
      
      // if (image2.IsValid()) {image.Destroy();}
      if (image2.IsValid()) 
      {
         rc = image2.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: MergeFGBG 3. Invalid rc from Destroy Image RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }
      }	
   }
   else // The make water is checked load the background as the image
   { 
   	if(!image.Load(filein2,typein2))
   	{	
         sprintf(buffer,"ERROR: MergeFGBG Load GE BACKGROUND image file:%s failed\n",filein);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         // image.Destroy();
         if (image.IsValid()) 
         {
            rc = image.Destroy(); // Destroy the image
            if(!rc)
            {
              sprintf(buffer,"Error: MergeFGBG 1. Invalid rc from Destroy Image RC=%d\n ",rc);
              printf(buffer);
              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              return 1;
            }
         }
         return 1;
   	}	
      // Flip it
      image.Flip();
      
      height = image.GetHeight();
	   width = image.GetWidth();
      sprintf(buffer,"DIAG MergeFGBG GE Background Image width=%d height=%d",width, height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));   
	
      // If depths file exists
      if(!argumentExists(10)) // if no is there a depth file
      {
         sprintf(buffer,"ERROR: MergeFGBG depth file does not exist!");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }
      else
      {       
         sprintf(buffer,"DIAG MergeFGBG making lat/lon_off=%d %d water make_water_transparent=%d",lat_off,lon_off,make_water_transparent);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));     
      
         CxImage depths_image;  
         // if there is a depths file load it
         if ( argumentExists(10) )
         {  
            // Load the depths added in V11.0.5.2
         	if(!depths_image.Load(depths,typein))
         	{	
               sprintf(buffer,"ERROR: MergeFGBG Load depths_image file:%s failed\n",depths);
               printf(buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               if (depths_image.IsValid()) 
               {
                  rc = depths_image.Destroy(); // Destroy the image
                  if(!rc)
                  {
                    sprintf(buffer,"Error: MergeFGBG 1. Invalid rc from Destroy depths_image RC=%d\n ",rc);
                    printf(buffer);
                    rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                    return 1;
                  }
               }
               return 1;
            }
      	}	 
         
         // Flip it
         depths_image.Flip();
         
         // Scan the depth units image and the image and put in the depth units
         count = 0;
         int xoffs,yoffs;
      	for(y=0; y<height; y++)
         {
            yoffs = y+lat_off;
            if(yoffs < height)  // if the we are within the height
            {
         		for(x=0; x<width; x++)
               {
                  xoffs = y+lat_off;
                  if(xoffs < width)  // if the we are within the width
                  {
            			color = depths_image.GetPixelColor(x+lon_off,y+lat_off);      
                     pixval = Get_Pixval(color);    // added in V17.0.0.19
                     if(pixval > 0)  // if its greater than 0 add it
                     {
                        if ( diag>2 )
                        {             
                           sprintf(buffer,"DIAG MergeFGBG X=%d Y=%d red=%d green=%d blue=%d pixval=%d\n",x,y,color.rgbRed,color.rgbGreen,color.rgbBlue,pixval);
                           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                        }  
                        // replace with the depths color        
            			   image.SetPixelColor(x+lon_off,y+lat_off,color,false);
                        count += 1;
                     }
                  }
               }
            }
         }	
         
         sprintf(buffer,"DIAG MergeFGBG replaced count=%d",count);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));     
         
         if (depths_image.IsValid())  // Destroy the image
         {
            rc = depths_image.Destroy(); // Destroy the image
            if(!rc)
            {
              sprintf(buffer,"Error: MergeFGBG 1. Invalid rc from Destroy depths_image RC=%d\n ",rc);
              printf(buffer);
              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              return 1;
            }
         }
         
         // Flip it  again
         image.Flip();
      }   	
   }

   // Save the modified foreground
   rc = image.Save(fileout,typeout);	
   if (!rc)
   {
      sprintf(buffer,"ERROR: MergeFGBG Invalid rc from save image RC=%d\n",rc);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }	
	
	// if (image.IsValid()) {image.Destroy();}
   if (image.IsValid()) 
   {
      rc = image.Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: MergeFGBG 2. Invalid rc from Destroy Image RC=%d\n ",rc);
        printf(buffer);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }
   }		
	
   return 0;
}   
/**************************** SetAlpha **************************
  Sets the alpha layer of a file from a gray scale mask
 * Syntax:	rc = SetAlpha(filein, fileout, maskfile, diag, diagmsgs)
 *
 * Params:		
 *			   filein - filename of the input file
 *			   maskfile  - filename of the mask file created by GDAL: 
 *			   gdal_translate -b 4 Havana_MH_1.tif Havana_MH_1_Mask.tif
 *			   diag    - if specified it will print diag information 
 *			   diagmsgs    - an array to hold diagnostic and error messages  

 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine4(int, SetAlpha, CSTRING, filein, CSTRING, maskfile, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512]; 
   int rc;
   
   sprintf(buffer,"DIAG SetAlpha input=%s maskfile=%s",filein,maskfile);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   	
	// See if arg 1 is a valid image type
	TCHAR* extin = FindExtension(filein);
	_tcslwr(extin);
	int typein = CxImage::GetTypeIdFromName(extin);	
	
	if (typein == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: SetAlpha Argument 1 is an invalid image type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}
   
   // See if arg 2 is a valid image type
	TCHAR* maskinext = FindExtension(maskfile);
	_tcslwr(maskinext);
	int masktypein = CxImage::GetTypeIdFromName(maskinext);
	if (masktypein == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: SetAlpha Argument maskfile in an invalid type");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
      return 1;      
	}
   	
	CxImage image;
	if(!image.Load(filein,typein))
	{
      sprintf(buffer,"ERROR: SetAlpha Load image file:%s failed\n",filein);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      // image.Destroy();
      if (image.IsValid()) 
      {
         rc = image.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: SetAlpha 1. Invalid rc from Destroy Image RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }
      }	
      return 1;
	}
	
	// load the maskfile image
	CxImage maskimage;
	// Load the maskfile image
	if(!maskimage.Load(maskfile, masktypein))
	{
      sprintf(buffer,"ERROR: SetAlpha Error loading maskfile:%s",maskfile);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      // image.Destroy();
      if (image.IsValid()) 
      {
         rc = image.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: SetAlpha 3. Invalid rc from Destroy Image RC=%d",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }
      }	
      // maskimage.Destroy();
      if (maskimage.IsValid()) 
      {
         rc = maskimage.Destroy(); // Destroy the maskimage
         if(!rc)
         {
           sprintf(buffer,"Error: SetAlpha 4. Invalid rc from Destroy Image RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }
      }
      return 1;   	
	}
   
   // create the alpha layer from the maskfile
   image.AlphaSet(maskimage);
 		
      
   // if (maskimage.IsValid()) {maskimage.Destroy();}   
   if (maskimage.IsValid()) 
   {
      rc = maskimage.Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: SetAlpha 5. Invalid rc from Destroy maskimage RC=%d\n ",rc);
        printf(buffer);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }
   }

   // Save the modified Filein with a new alpha
   rc = image.Save(filein,typein);	
   if (!rc)
   {
      sprintf(buffer,"ERROR: SetAlpha Invalid rc from save image RC=%d\n",rc);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }	
	
	// if (image.IsValid()) {image.Destroy();}
   if (image.IsValid()) 
   {
      rc = image.Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: SetAlpha 2. Invalid rc from Destroy Image RC=%d\n ",rc);
        printf(buffer);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }
   }	
   
   return 0;
}   
/**************************** CropImage **************************
  Crops an image 
 * Syntax:	rc = CropImage(filein, fileout, left ,top, right, bottom, comm._diag, diagmsgs) 
 *
 * Params:		
 *			   filein - filename of the input file
 *			   fileout  - filename of the outputfile file
 *			   left     - left amount to crop
 *			   top      - top amount to crop
 *			   right    - right amount to crop
 *			   bottom   - bottom amount to crop     
 *			   diag    - if specified it will print diag information 
 *			   diagmsgs    - an array to hold diagnostic and error messages 
 *			   colors  (optional) if present is the number of colors to quantize to Quantize added in V31.0.0.4
 *
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine9(int, CropImage, CSTRING, filein, CSTRING, fileout, int, left, int, top, int, right, int, bottom, int, diag, RexxArrayObject, diagmsgs , OPTIONAL_int, colors)
{
   char buffer[512]; 
   int rtnrc = 0;
   int rc;
   if ( diag>0 )
   {sprintf(buffer,"DIAG CropImage input=%s output=%s",filein, fileout);rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));} 
   	
	ULONG inbpp; 

	// See if arg 1 is a valid image type
	TCHAR* extin = FindExtension(filein);
	_tcslwr(extin);
	int typein = CxImage::GetTypeIdFromName(extin);
	if (typein == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: CropImage Argument 1 is an invalid image type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}
   // See if arg 2 is a valid image type
	TCHAR* extout = FindExtension(fileout);
	_tcslwr(extout);
	int typeout = CxImage::GetTypeIdFromName(extout);
	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: CropImage Argument 2 in an invalid type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;    
	}	
	
	CxImage image;
	if(!image.Load(filein,typein))
	{
      sprintf(buffer,"ERROR: CropImage Load image file:%s failed\n",filein);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      // image.Destroy();
      if (image.IsValid()) 
      {
         rc = image.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: CropImage 1. Invalid rc from Destroy Image RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }
      }
      return 1;
	}

   INT numcolors; 
   CQuantizer m(16777216,8);
   m.ProcessImage(image.GetDIB());	 
   numcolors = m.GetColorCount();	
   
	inbpp = image.GetBpp(); 

   sprintf(buffer,"DIAG CropImage typein= %d typeout=%d inbpp=%d numcolors=%d\n",typein,typeout,inbpp,numcolors);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   sprintf(buffer,"DIAG CropImage left=%d top=%d right=%d bottom=%d\n",left,top,right,bottom);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));      
   
   bool do_Quantizer = false;
   // If there is a quantize argument
   if ( argumentExists(9) )
   {  
      sprintf(buffer,"DIAG CropImage Argument 9 exists colors=%d",colors);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      if (colors > 0 ) do_Quantizer = true; // if its 0 don't quantize
      sprintf(buffer,"DIAG CropImage do_Quantizer=%d",do_Quantizer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }
   
   // Do the quantize if no override
   if (do_Quantizer)
   {
      ULONG outbpp; 
      RGBQUAD * ppal;
      INT numcolors; 
      bool error_diffusion = true; 
      CQuantizer q(colors,4);     //  was 8
      q.ProcessImage(image.GetDIB());           
      ppal=(RGBQUAD*)calloc(256*sizeof(RGBQUAD),1);
      q.SetColorTable(ppal);
    
      rc = image.DecreaseBpp(8,error_diffusion,ppal,colors);    
      outbpp = image.GetBpp(); 
      numcolors = q.GetColorCount();
   
      sprintf(buffer,"DIAG CropImage DecreaseBpp RC=%d outbpp=%d AFTER: numcolors=%d",rc,outbpp,numcolors);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));       
      if (ppal) free(ppal); // Added in V30.0.6.0    
   }        
    
   // Crop the image
   float angle = 0;
   int newwidth, newheight;
   newwidth = right - left;
   newheight = bottom - top;
   rc = image.CropRotatedRectangle( left, top, newwidth, newheight, angle );
   if (!rc)
   {
      sprintf(buffer,"ERROR: CropImage Invalid rc from crop image RC=%d\n",rc);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }	

   // Pass back the type, width, height, X resolution and Y resolution
   int b_resx, b_resy, b_width, b_height;
   b_resx = image.GetXDPI(); // Get X resolution
   b_resy = image.GetYDPI(); // Get Y resolution  nn
   b_width = image.GetWidth();
   b_height = image.GetHeight();
   sprintf(buffer,"DIAG CropImage TYPEOUT=%d XRES=%d YRES=%d ",typeout, b_resx, b_resy);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));       
   sprintf(buffer,"DIAG CropImage WIDTH=%d HEIGHT=%d ",b_width,b_height);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));    	

   // Save the image in the output file
   rc = image.Save(fileout,typeout);	
   if (!rc)
   {
      sprintf(buffer,"ERROR: CropImage Invalid rc from save image RC=%d\n",rc);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }	
	
	// if (image.IsValid()) {image.Destroy();}
   if (image.IsValid()) 
   {
      rc = image.Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: CropImage 2. Invalid rc from Destroy Image RC=%d\n ",rc);
        printf(buffer);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }
   }
   
   return rtnrc;
}   
/**************************** ReSizeImage**************************
  Resizes an image 
 * Syntax:	rc = ReSizeImage(filein, fileout, newx, newy, diag, diagmsgs)
 *
 * Params:		
 *			   filein - filename of the input file
 *			   fileout  - filename of the outputfile file
 *			   newx     - new width
 *			   newy     - new height
 *			   diag    - if specified it will print diag information 
 *			   diagmsgs    - an array to hold diagnostic and error messages    
 *
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine6(int, ReSizeImage, CSTRING, filein, CSTRING, fileout, int, newx, int, newy, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512]; 
   int rtnrc = 0;
   int rc;
   if ( diag>0 )
   {sprintf(buffer,"DIAG ReSizeImage input=%s output=%s",filein, fileout);rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));} 
   	
	ULONG inbpp; 

	// See if arg 1 is a valid image type
	TCHAR* extin = FindExtension(filein);
	_tcslwr(extin);
	int typein = CxImage::GetTypeIdFromName(extin);
	if (typein == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: ReSizeImage Argument 1 is an invalid image type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}
   // See if arg 2 is a valid image type
	TCHAR* extout = FindExtension(fileout);
	_tcslwr(extout);
	int typeout = CxImage::GetTypeIdFromName(extout);
	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: ReSizeImage Argument 2 in an invalid type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;    
	}	
	
	CxImage image;
	if(!image.Load(filein,typein))
	{
      sprintf(buffer,"ERROR: ReSizeImage Load image file:%s failed\n",filein);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      // image.Destroy();
      if (image.IsValid()) 
      {
         rc = image.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: ResizeImage 1. Invalid rc from Destroy Image RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }
      }
      return 1;
	}

   INT numcolors; 
   CQuantizer m(16777216,8);
   m.ProcessImage(image.GetDIB());	 
   numcolors = m.GetColorCount();	
   
	inbpp = image.GetBpp(); 

	if ( diag>0 ) 
   {
      sprintf(buffer,"DIAG ReSizeImage typein= %d typeout=%d inbpp=%d numcolors=%d\n",typein,typeout,inbpp,numcolors);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      sprintf(buffer,"DIAG Resize newx=%d newy=%d\n",newx,newy);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));      
   }   
    
   // Resize the image
   rc = image.Resample2( (long)newx, (long)newy, (CxImage::InterpolationMethod)7, (CxImage::OverflowMethod)1,NULL,true);
   // rc = image.Resample2( (long)newx, (long)newy);
   if (!rc)
   {
      sprintf(buffer,"ERROR: ReSizeImage Invalid rc from crop image RC=%d\n",rc);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }	

   // Pass back the type, width, height, X resolution and Y resolution
   int b_resx, b_resy, b_width, b_height;
   b_resx = image.GetXDPI(); // Get X resolution
   b_resy = image.GetYDPI(); // Get Y resolution
   b_width = image.GetWidth();
   b_height = image.GetHeight();
   sprintf(buffer,"DIAG ReSizeImage TYPEOUT=%d XRES=%d YRES=%d ",typeout, b_resx, b_resy);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));       
   sprintf(buffer,"DIAG ReSizeImage WIDTH=%d HEIGHT=%d ",b_width,b_height);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));    	

   // Save the image in the output file
   rc = image.Save(fileout,typeout);	
   if (!rc)
   {
      sprintf(buffer,"ERROR: ResIze Invalid rc from save image RC=%d\n",rc);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }	
	
	// if (image.IsValid()) {image.Destroy();}
   if (image.IsValid()) 
      {
         rc = image.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: ResizeImage 1. Invalid rc from Destroy Image RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }
      }
   
   return rtnrc;
}   
/**************************** ExpandImage**************************
  Expands an image 
 * Syntax:	rc = ExpandImage(filein, fileout, newx, newy, diag, diagmsgs)
 *
 * Params:		
 *			   filein - filename of the input file
 *			   fileout  - filename of the outputfile file
 *			   newx     - new width
 *			   newy     - new height
 *			   diag    - if specified it will print diag information 
 *			   diagmsgs    - an array to hold diagnostic and error messages    
 *
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine6(int, ExpandImage, CSTRING, filein, CSTRING, fileout, int, newx, int, newy, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512]; 
   int rtnrc = 0;
   int rc;
   if ( diag>0 )
   {sprintf(buffer,"DIAG ExpandImage input=%s output=%s",filein, fileout);rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));} 
   	
	ULONG inbpp; 

	// See if arg 1 is a valid image type
	TCHAR* extin = FindExtension(filein);
	_tcslwr(extin);
	int typein = CxImage::GetTypeIdFromName(extin);
	if (typein == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: ExpandImage Argument 1 is an invalid image type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}
   // See if arg 2 is a valid image type
	TCHAR* extout = FindExtension(fileout);
	_tcslwr(extout);
	int typeout = CxImage::GetTypeIdFromName(extout);
	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: ExpandImage Argument 2 in an invalid type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;    
	}	
	
	CxImage image;
	if(!image.Load(filein,typein))
	{
      sprintf(buffer,"ERROR: ExpandImage Load image file:%s failed\n",filein);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      // image.Destroy();
      if (image.IsValid()) 
      {
         rc = image.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: ExpandImage 1. Invalid rc from Destroy Image RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }
      }
      return 1;
	}

   INT numcolors; 
   CQuantizer m(16777216,8);
   m.ProcessImage(image.GetDIB());	 
   numcolors = m.GetColorCount();	
   
	inbpp = image.GetBpp(); 

	if ( diag>0 ) 
   {
      sprintf(buffer,"DIAG ExpandImage typein= %d typeout=%d inbpp=%d numcolors=%d\n",typein,typeout,inbpp,numcolors);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      sprintf(buffer,"DIAG ExpandImage newx=%d newy=%d\n",newx,newy);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));      
   }   
    
   // Expand the image
   RGBQUAD Canvascolor;
   Canvascolor.rgbRed = 0; 
   Canvascolor.rgbGreen = 0; 
   Canvascolor.rgbBlue = 0; 
   Canvascolor.rgbReserved = 1;       
   rc = image.Expand( (long)newx, (long)newy, Canvascolor);
   if (!rc)
   {
      sprintf(buffer,"ERROR: ExpandImage Invalid rc from crop image RC=%d\n",rc);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }	

   // Pass back the type, width, height, X resolution and Y resolution
   int b_resx, b_resy, b_width, b_height;
   b_resx = image.GetXDPI(); // Get X resolution
   b_resy = image.GetYDPI(); // Get Y resolution
   b_width = image.GetWidth();
   b_height = image.GetHeight();
   sprintf(buffer,"DIAG ExpandImage TYPEOUT=%d XRES=%d YRES=%d ",typeout, b_resx, b_resy);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));       
   sprintf(buffer,"DIAG ExpandImage WIDTH=%d HEIGHT=%d ",b_width,b_height);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));    	

   // Save the image in the output file
   rc = image.Save(fileout,typeout);	
   if (!rc)
   {
      sprintf(buffer,"ERROR: ResIze Invalid rc from save image RC=%d\n",rc);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }	
	
	// if (image.IsValid()) {image.Destroy();}
   if (image.IsValid()) 
      {
         rc = image.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: ExpandImage 2. Invalid rc from Destroy Image RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }
      }
	
	return rtnrc;
}	
   
/**************************** Rotate Image **************************
  rotates an iamge around its center point keeping the same size 
 * Syntax:	rc = RotateImage(filein, fileout, angle, diag, diagmsgs)
 *
 * Params:		
 *			   filein - filename of the input file
 *			   fileout  - filename of the outputfile file
 *			   angle    - number of degrees to rotate (positive values rotate clockwise)
 *			   diag    - if specified it will print diag information 
 *			   diagmsgs    - an array to hold diagnostic and error messages    
 *
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine5(int, RotateImage, CSTRING, filein, CSTRING, fileout, float, angle, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512]; 
   int rtnrc = 0;
   int rc;
   if ( diag>0 )
   {
      sprintf(buffer,"DIAG RotateImage input=%s output=%s, angle=%f",filein, fileout, angle);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   } 
   	
	ULONG inbpp; 

	// See if arg 1 is a valid image type
	TCHAR* extin = FindExtension(filein);
	_tcslwr(extin);
	int typein = CxImage::GetTypeIdFromName(extin);
	if (typein == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: RotateImage Argument 1 is an invalid image type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}
   // See if arg 2 is a valid image type
	TCHAR* extout = FindExtension(fileout);
	_tcslwr(extout);
	int typeout = CxImage::GetTypeIdFromName(extout);
	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: RotateImage Argument 2 in an invalid type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;    
	}	
	
	CxImage image;
	if(!image.Load(filein,typein))
	{
      sprintf(buffer,"ERROR: RotateImage Load image file:%s failed\n",filein);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      image.Destroy();
      return 1;
	}
   
	inbpp = image.GetBpp(); 

	if ( diag>0 ) 
   {
      sprintf(buffer,"DIAG RotateImage typein= %d typeout=%d inbpp=%d\n",typein,typeout,inbpp);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));    
   }   
    
   // Rotata the image
   CxImage::InterpolationMethod intm;
	CxImage::OverflowMethod overm;
	intm=CxImage::IM_BILINEAR;
	overm=CxImage::OM_BACKGROUND;
   rc = image.Rotate2( angle, NULL, intm, overm, 0, 1, 1 );
   if (!rc)
   {
      sprintf(buffer,"ERROR: RotateImage Invalid rc from crop image RC=%d\n",rc);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }	

   // Pass back the type, width, height, X resolution and Y resolution
   int b_resx, b_resy, b_width, b_height;
   b_resx = image.GetXDPI(); // Get X resolution
   b_resy = image.GetYDPI(); // Get Y resolution
   b_width = image.GetWidth();
   b_height = image.GetHeight();
   sprintf(buffer,"DIAG RotateImage TYPEOUT=%d XRES=%d YRES=%d ",typeout, b_resx, b_resy);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));       
   sprintf(buffer,"DIAG RotateImage WIDTH=%d HEIGHT=%d ",b_width,b_height);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));    	

   // Save the image in the output file
   rc = image.Save(fileout,typeout);	
   if (!rc)
   {
      sprintf(buffer,"ERROR: RotateImage Invalid rc from save image RC=%d\n",rc);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }	
	
	// if (image.IsValid()) {image.Destroy();}
   if (image.IsValid()) 
      {
         rc = image.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: ExpandImage 1. Invalid rc from Destroy Image RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }
      }
   
   return rtnrc;
}  
/**************************** AlphaSet **************************
  Sets any pixels in a GIF file that are with a certain range to transparent
 * Syntax:	rc = AlphaSet(filein, fileout, fuzz, diag(optional))
 * Params:		
 *			   filein - filename of the input file
 *			   fileout  - filename of the outputfile file
 *			   fuzz - percent fuzz
 *			   maxcolors - the maxcolors added in V6.0.2.1
 *			   diag    - if specified it will print diag information 
 *			   diagmsgs    - an array to hold diagnostic and error messages     
 *
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine6(int, AlphaSet, CSTRING, filein, CSTRING, fileout, uint32_t, fuzz, int, maxcolors, int, diag, RexxArrayObject, diagmsgs)
{
   int rc;
   int rtnrc = 0;
   char buffer[512]; 
  
   sprintf(buffer,"DIAG AlphaSet input=%s output=%s fuzz=%d \n",filein, fileout, fuzz);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
   
	// See if arg 1 is a valid image type
	TCHAR* extin = FindExtension(filein);
	_tcslwr(extin);
	int typein = CxImage::GetTypeIdFromName(extin);
	if (typein == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: AlphaSet Argument 1 is an invalid image type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}
   // See if arg 2 is a valid image type
	TCHAR* extout = FindExtension(fileout);
	_tcslwr(extout);
	int typeout = CxImage::GetTypeIdFromName(extout);
	if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: AlphaSet Argumnent 2 in an invalid type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;    
	}	   
	
	CxImage image;
	if(!image.Load(filein,typein))
	{	
      sprintf(buffer,"ERROR: AlphaSet Load image file:%s failed\n",filein);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      // image.Destroy();
      if (image.IsValid()) 
      {
         rc = image.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: AlphaSet 1. Invalid rc from Destroy Image RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }
      }
      return 1;
	}	

   ULONG inbpp = image.GetBpp();  // Added in V6.0.1.2
   sprintf(buffer,"DIAG AlphaSet inbpp=%d",inbpp);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));

   RGBQUAD * ppal;
   int colors = maxcolors; // added in V6.0.2.1
   int numcolors;
   if (inbpp == 8 || inbpp == 24)   // new in V6.0.1.2
   {
  	  CQuantizer q(colors,(colors>16?7:8));
     q.ProcessImage(image.GetDIB()); 
     ppal=(RGBQUAD*)calloc(256*sizeof(RGBQUAD),1);
     q.SetColorTable(ppal);
     rc = image.DecreaseBpp(8,1,ppal,colors); // V600 Set error diffusion=1
     ULONG outbpp = image.GetBpp(); 
     if (!rc)
     {
         sprintf(buffer,"ERROR: AlphaSet DecreaseBpp RC=%d outbpp=%d\n",rc,outbpp);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;	
      } 
	   numcolors = q.GetColorCount(); 
      sprintf(buffer,"DIAG AlphaSet out_numcolors=%d outbpp=%d",numcolors,outbpp);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   }      
	 
	RGBQUAD color;
	ULONG pixval,pixval2;
	int height = image.GetHeight();
	int width = image.GetWidth();
	int red,green,blue;
	int count=0;
   long x, y;
	uint8_t idx;	
	
	// Get the size of the palette  in bytes
   int palsize = image.GetPaletteSize()/4;   
   // Get the number of colors used 
   BITMAPINFOHEADER ds;
   HANDLE hImage = image.GetDIB();
   memcpy(&ds,hImage, sizeof(ds));
   int clrused = ds.biClrUsed;
   sprintf(buffer,"DIAG AlphaSet palsize=%d clrused=%d\n",palsize,clrused);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));     		

   //  make sure the input only has 63 or 128 colors or less  ADDED IN v9.5.1
   colors = numcolors / 2; // added in V6.0.4.1
   // RGBQUAD * ppal;
  	// CQuantizer q(colors,(colors>16?7:8));
  	CQuantizer q(colors,4);     //  was 8  changed in V9.5.3
   q.ProcessImage(image.GetDIB()); 
   ppal=(RGBQUAD*)calloc(256*sizeof(RGBQUAD),1);
   q.SetColorTable(ppal);
   rc = image.DecreaseBpp(8,0,ppal,colors);

	// Create a array to hold the histogram  of the images, each entry is 7 words long: palette colors r,g,b, color index, 
   // color count,new index  and pixval
	ULONG *pArray = (ULONG*) malloc(ENTSIZE*palsize);
	
	// Reset the array to zero  to do the image first
	memset(pArray,0,ENTSIZE*palsize); 

   // Fill the array with the information from the palette
	for(int i=0; i<palsize; i++)
	{
      color = image.GetPaletteColor(i);
      pixval = Get_Pixval(color);    // added in V17.0.0.19
      pArray[i*ENTLEN] = color.rgbRed; 
      pArray[i*ENTLEN+1] = color.rgbGreen; 
      pArray[i*ENTLEN+2] = color.rgbBlue; 
      pArray[i*ENTLEN+3] = i; // Save the index for this color
      pArray[i*ENTLEN+6] = pixval; // Save the pixval for this color
   }
      
   // Scan all the pixels and increment the palette color counts 
	for(y=0; y<height; y++)
   {
		for(x=0; x<width; x++)
      {
         // Get color index and increment the count for the index
			idx = image.GetPixelIndex(x,y);
         pArray[idx*ENTLEN+4]++;	// increment the histogram count for this color	
		}
   }    
    
	if ( diag>1 ) // if diag turned on print the histogram
	{       
      sprintf(buffer,"DIAG AlphaSet histogram image before sort:\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));    
   	// list the values in the histogram 
   	numcolors = listENT(pArray, palsize,diagmsgs,context);
      sprintf(buffer,"DIAG Alphaset image Numcolors=%d \n",numcolors);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));      	
   }	
                          
   // Sort the array by the count value in descending order   
   sortENT( pArray, 4, palsize, 1);  
   
	if ( diag>1 ) // if diag turned on print the histogram
	{       
      sprintf(buffer,"DIAG AlphaSet histogram image after sort by count:\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));    
   	// list the values in the histogram 
   	numcolors = listENT(pArray, palsize,diagmsgs,context);
   	sprintf(buffer,"DIAG Alphaset image Numcolors=%d \n",numcolors);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));        	
   }	           	
   
   // Scan the array histogram and see if there is a normal land color
	uint32_t lowrange = 0;
	uint32_t hirange = 0;
	ULONG landcolor = -1;
	ULONG landidx =0;                              
	for(int i=0; (i<palsize && landcolor == -1); i++)
	{
      pixval = pArray[i*ENTLEN+6];
      if (pixval == 0) continue;
      red = pArray[i*ENTLEN];
      green = pArray[i*ENTLEN+1];
      blue = pArray[i*ENTLEN+2];
      idx = (uint8_t)pArray[i*ENTLEN+3];
      count = pArray[i*ENTLEN+4];

      //  Changed the following to accommodate ENC charts            11899449  11842484            added test for green in V9.5.0
      // if (pixval < 11800000 || pixval > 17097568 || blue > 202 || green < 140) continue; 
      if (pixval >= 11800000 && pixval <= 17097568 && blue <= 202 && green >= 140)
      { // forget entry if it does not meet land criteria
         // save the landcolor and  its index
         landcolor = pixval; landidx=pArray[i*ENTLEN+3];
         sprintf(buffer,"DIAG AlphaSet Criteria MET idx=%d rgb(%d,%d,%d)(%d) count=%d\n",idx,red,green,blue, pixval,count);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         // break;
      } 
   }   
   
   // If no landcolor was found, scan again looking for a land color like the one in the NGA charts eg: 24400.kap
   if (landcolor == -1) 
   {
      for(int i=0; (i<palsize && landcolor == -1); i++)
      {
         pixval = pArray[i*ENTLEN+6];
         if (pixval == 0) continue;
         red = pArray[i*ENTLEN];
         green = pArray[i*ENTLEN+1];
         blue = pArray[i*ENTLEN+2];
         idx = (uint8_t)pArray[i*ENTLEN+3];
         count = pArray[i*ENTLEN+4];
         // Added in V9.5.2   to accomodate those strange charts    14867688   14867688   14413809
         if (pixval >= 14413800 && pixval <= 15129400)
         {
            // save the landcolor and  its index
            landcolor = pixval; landidx=pArray[i*ENTLEN+3];
            sprintf(buffer,"DIAG AlphaSet Criteria MET Strange idx=%d rgb(%d,%d,%d)(%d) count=%d\n",idx,red,green,blue, pixval,count);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            // break;    
         }
      }      
   }       
   
   // If no landcolor was found, pass back a message to the caller 
   if (landcolor == -1)   
   {
      sprintf(buffer,"Alphaset no landcolor was found\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      
   	if(!image.Save(fileout,typeout))
   	{	  
         context->RaiseException1(
               Rexx_Error_Incorrect_method_user_defined,
               context->String("Invalid rc from image save"));	
   		return 1;
   	}	
      
      // if (image.IsValid()) {image.Destroy();}
      if (image.IsValid()) 
      {
         rc = image.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: AlphaSet 2. Invalid rc from Destroy Image RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }
      }
      
      free(pArray);	   	
   	
      return rtnrc;      
   }
   
   // If the fuzz is valid, then set the range with it
   int range = 0;
   if (fuzz>0)
   {     
      range = landcolor / 100 * fuzz;
      lowrange = landcolor - range;
      hirange = landcolor + range;
   }
   else {lowrange=landcolor; hirange=landcolor;}
   

   sprintf(buffer,"DIAG AlphaSet landcolor=%d landindex=%d range=%d lowrange=%d hirange=%d\n",landcolor, landidx, range, lowrange, hirange);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
    
	// Set the background color to the landcolor
	red = landcolor/65536;
   pixval2 = landcolor - red*65536;
   green = pixval2/256;
   pixval2 -= green*256;
   blue = pixval2;
   RGBQUAD BkgndColor;
   // RGBQUAD * ppal;
   BkgndColor.rgbRed = red; 
   BkgndColor.rgbGreen = green; 
   BkgndColor.rgbBlue = blue; 	
    
   
   sprintf(buffer,"DIAG AlphaSet landcolor=%d rgb(%d,%d,%d)\n",landcolor,red,green,blue);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        
	image.SetTransIndex(landidx);  // Set the background color's index
	image.SetTransColor(BkgndColor); // Set the transparent color
	
	rc = image.GetTransIndex();	
	
   sprintf(buffer,"DIAG AlphaSet GetTransindex RC=%d\n",rc);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));      
	
	// Scan all the pixels and set any within low and high range and color criteria to the transparent color
	count = 0;
	for(y=0; y<height; y++)
   {
      int yflip = height - y - 1;
      // LONG idxcnt;
      // uint32_t idx;
		for(x=0; x<width; x++)
      {
			color = image.GetPixelColor(x,yflip);
         pixval = Get_Pixval(color);    // added in V17.0.0.19
			// if (pixval >= lowrange && pixval <= hirange && blue <= 200 && idxcnt>5000)  REMOVE
			if (pixval >= lowrange && pixval <= hirange && color.rgbBlue <= 202 && color.rgbGreen >= 140)  // added test for green in V9.5.0
			{
            if ( diag>2 )
            {
               sprintf(buffer,"DIAG AlphaSet X=%d Y=%d red=%d green=%d blue=%d pixval=%d\n",x,y,color.rgbRed,color.rgbGreen,color.rgbBlue,pixval);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            }            
			   image.SetPixelColor(x,yflip,BkgndColor,true);
			   image.SetPixelIndex(x,yflip,(BYTE)landidx);
			   count += 1;
         }			
		}
   }
   
   sprintf(buffer,"DIAG AlphaSet Count=%d \n",count);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));   
		
	if(!image.Save(fileout,typeout))
	{	  
      context->RaiseException1(
            Rexx_Error_Incorrect_method_user_defined,
            context->String("Invalid rc from image save"));	
		return 1;
	}	
   
   // if (image.IsValid()) {image.Destroy();}
   if (image.IsValid()) 
      {
         rc = image.Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: AlphaSet 3. Invalid rc from Destroy Image RC=%d\n ",rc);
           printf(buffer);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }
      }
   
   free(pArray);	   	
	
   return rtnrc;
} 

/**************************** tif2BSB **************************
  Converts a tif file to a BSB (Kap) chart
* Takes a .kap template and a tif file 
* Syntax:	rc = Tif2BSB(filein, fileout, hdrfile, numcolors(optional), diag(optional))
 *
 * Params:		
 *			   filein - filename of the tif input file
 *			   fileout  - filename of the kap outputfile file
 *			   hdrfile  - filename of the hdr file 
 *			   numcolors  - number of colors in the output (optional)
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages   
 *
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine6(int, Tif2BSB, CSTRING, filein, CSTRING, fileout, CSTRING, hdrfile, OPTIONAL_int, numcolors, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512]; 
   int rtnrc = 0;
   int rc;
  
   if ( diag>0 )
   {
      sprintf(buffer,"DIAG Tif2BSB Filein=%s Fileout=%s Hdrfile=%s\n",filein, fileout,hdrfile);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   } 

	BSBImage	image;
	int			i, j, depth = 0, num_colors = -1, *index;
	FILE		*tmpl_file, *out;
	TIFF*		tif;
	uint16_t	*red, *green, *blue;
	uint16_t	bits_per_sample, photometric, planar_config;
	uint8_t		*tiff_row, *scratch;
	char		line[1024];
	
	// If the numcolors is provided then use it
	if ( argumentExists(4) ) num_colors = numcolors;
	if ( diag>0 ) 
   {
      sprintf(buffer,"DIAG Tif2BSB numcolors=%d\n",num_colors);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }    	
	
	// See if arg 1 is a valid image type
	TCHAR* extin = FindExtension(filein);
	_tcslwr(extin);
	int typein = CxImage::GetTypeIdFromName(extin);
	if (typein == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"ERROR: Tif2BSB Argument 1 is an invalid image type\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}	
 
	tif = TIFFOpen(filein, "rb");
	if (! tif)
	{
      sprintf(buffer,"ERROR: Tif2BSB Could not TIFFOpen \"%s\"\n", filein);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;		
	}

	TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &image.width);
	TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &image.height);
	TIFFGetFieldDefaulted(tif, TIFFTAG_BITSPERSAMPLE, &bits_per_sample);
	TIFFGetFieldDefaulted(tif, TIFFTAG_PHOTOMETRIC, &photometric);
	TIFFGetFieldDefaulted(tif, TIFFTAG_PLANARCONFIG, &planar_config);

	/* Check for supported TIFF files */
	if (image.width == 0 || image.height == 0)
	{
      sprintf(buffer,"ERROR: Tif2BSB Invalid TIFF file (width=%d,height=%d)\n", image.width, image.height);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;		
	}	
   
   if ( diag>0 )
   {
      sprintf(buffer,"DIAG Tif2BSB width=%d height=%d\n",image.width, image.height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }    
	
	if (photometric != PHOTOMETRIC_PALETTE)
	{
      sprintf(buffer,"ERROR: Tif2BSB No support for TIFF files with PHOTOMETRIC=%d (only images containing color maps supported)\n", photometric);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;		
	}	
	
	if (bits_per_sample != 8 && bits_per_sample != 4)
	{
      sprintf(buffer,"ERROR: Tif2BSB No support for TIFF files with BITSPERSAMPLE=%d (only 4 or 8 bits per sample supported)\n", bits_per_sample);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;		
	}	
	
	if (planar_config != PLANARCONFIG_CONTIG)
	{
      sprintf(buffer,"ERROR: Tif2BSB No support for TIFF files with PLANARCONFIG=%d (only single plane images supported)\n", planar_config);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;	
   }	

	/* TIFFGetField allocates memory for color fields - TIFFClose frees it. */
	TIFFGetField(tif, TIFFTAG_COLORMAP, &red, &green, &blue);

	tiff_row = (uint8_t *)malloc(image.width);
	if (! tiff_row)
	{
      sprintf(buffer,"ERROR: Tif2BSB Cannot allocate %d bytes for image row\n", image.width);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;	
   }		
	

	if (num_colors == -1)
	{
		num_colors = count_tiff_colors(tif, &image, bits_per_sample);
		sprintf(buffer,"DIAG Tif2BSB num_colors in image=%d\n",num_colors);
		printf(buffer); printf("\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
	}	

	if (num_colors <= 0)
	{
      sprintf(buffer,"ERROR: Tif2BSB no colors found in image\n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;	
   }	

	/* The BSB format cannot cope with more than 127 colors */
	if (num_colors >= 128)
	{
      sprintf(buffer,"ERROR: Tif2BSB Too many colors for BSB format (%d > 127 max.)\n", num_colors);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;	
   }

	/* Given num_colors in input image, compute required "depth"	*/
	/* E.g. num_colors=35, depth=6  (2^6 = 64)						*/
	for (i = 0; i < (int)sizeof(depth_for_colors); i++)
	{
		if (num_colors < depth_for_colors[i])
			break;
		depth = i + 1;
	}
	image.depth = depth;

	out = fopen(fileout, "wb");
	if (! out)
	{
      sprintf(buffer,"ERROR: Tif2BSB Could not Open \"%s\"\n", fileout);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;	
   }	

	/* Copy in all text lines from template excluding RGB and IFM tags */
	tmpl_file = fopen(hdrfile, "rb");
	while (fgets(line, sizeof(line), tmpl_file) != NULL)
	{
		if (strncmp("RGB/", line, 4) == 0)
			continue;
		if (strncmp("IFM/", line, 4) == 0)
			continue;
		if (line[0] == 0x1a)
			break;
		fputs(line, out);
	}
	fclose(tmpl_file);
   

	/* Write IFM tag */
	fprintf(out, "IFM/%d\r\n", depth);
	/* Write RGB tags for colormap */
	for (i = 0; i < num_colors; i++)
	{                                     
		fprintf(out, "RGB/%d,%d,%d,%d\r\n",
					i+1,
					red[i] >> 8,
					green[i] >> 8,
					blue[i] >> 8
					);
	}
	fputc(0x1a, out);
	fputc('\0', out);
	fputc(depth, out);

	index = (int *)malloc((image.height + 1) * sizeof(int));

	scratch = (uint8_t *)malloc(image.width + 8);	/* max space encoded line can take */

	/* Read rows from tif, write to bsb */
	for (j = 0; j < image.height; j++)
	{
		int		len;

		/* Record start-of-row file position for index table */
		index[j] = ftell(out);

		read_8bpp_tiff_line(tif, bits_per_sample, image.width, tiff_row, j);

		/* Compress raster and write to BSB file */
		len = bsb_compress_row(&image, j, tiff_row, scratch);
		fwrite(scratch, len, 1, out);
	}
	free(scratch);
	free(tiff_row);
	TIFFClose(tif);

	/* record start-of-index-table file position in the index table */
	index[image.height] = ftell(out);

	bsb_write_index(out, image.height, index);
	free(index);

	fclose(out);
	
	return rtnrc;
}	



/**************************** BSB2Tif **************************
  Converts a BSB (Kap) chart to a TIF file
* Syntax:	rc = BSB2Tif(filein, fileout, diag, diagmsgs)
 *
 * Params:		
 *			   filein - filename of the kap chart
 *			   fileout  - filename of the tif output file
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages   
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine4(int, BSB2Tif, CSTRING, filein, CSTRING, fileout, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512]; 
   int rtnrc = 0;
   int rc;
   
   sprintf(buffer,"DIAG BSB2Tif Filein=%s Fileout=%s",filein, fileout);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 

	BSBImage	image;
	int			i;
	uint16_t	red[256], green[256], blue[256];
	TIFF*		tif;
	uint8_t		*buf;

	rc = bsb_open_header((char *)filein, &image);
   if(rc != 0)
   {
      sprintf(buffer,"ERROR BSB2Tif: Open Failed rc=%d",rc);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   }
   
   sprintf(buffer,"DIAG BSB2Tif image.height=%d",image.height);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 

	/* Initialise colormap entries */
	memset(red, 0, sizeof(red));
	memset(green, 0, sizeof(green));
	memset(blue, 0, sizeof(blue));

	/* convert BSB colormap to libtiff style colormap */
	for (i = 0; i < image.num_colors; i++)
	{
		red[i] = image.red[i] * 65536 / 256;
		green[i] = image.green[i] * 65536 / 256;
		blue[i] = image.blue[i] * 65536 / 256;
	}

	buf = (uint8_t *)malloc(image.width);
	if (! buf)
   {
      sprintf(buffer,"ERROR BSB2Tif: Buf Failed rc=%d",rc);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
		return 1;
   }

	/* Open tif file for output (truncates any existing file) */
	tif = TIFFOpen(fileout, "w");
	if (! tif)
	{
      sprintf(buffer,"ERROR: BSB2TIF Could not Open \"%s\"\n", fileout);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;	
   }	

	/* From TIFF spec, these are required fields for palette-color images */
	TIFFSetField(tif,TIFFTAG_IMAGEWIDTH, image.width);
	TIFFSetField(tif,TIFFTAG_IMAGELENGTH, image.height);
	/* Can't rely on tiff library having LZW support so use PACKBITS */
	TIFFSetField(tif,TIFFTAG_COMPRESSION, COMPRESSION_PACKBITS);
	/* BSB is strictly a colormap only format */
	TIFFSetField(tif,TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_PALETTE);
	TIFFSetField(tif,TIFFTAG_ROWSPERSTRIP, 1);
	TIFFSetField(tif,TIFFTAG_BITSPERSAMPLE, 8);
	/* CONTIG means R G B grouped together in each STRIP */
	TIFFSetField(tif,TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
	TIFFSetField(tif,TIFFTAG_COLORMAP, red, green, blue);

	/* Read rows from bsb file, write to tif file */
	for (i = 0; i < image.height; i++)
	{
		bsb_seek_to_row(&image, i);
		bsb_read_row(&image, buf);
		TIFFWriteScanline(tif, buf, i, 0);
	}

	TIFFClose(tif);
	free(buf);

	bsb_close(&image);

	return rtnrc;
} 

BOOL CALLBACK EnumWindowsProc(HWND hwnd,LPARAM lParam)
{   
    DWORD           i;
	 // RECT			r;
    // TCHAR           buf[MAX_PATH];
    TCHAR           buf[1000]; // changed in in V15.0.1.3
    PTASK_LIST_ENUM te = (PTASK_LIST_ENUM)lParam;
    PTASK_LIST      tlist = te->tlist;
    DWORD           numTasks = te->numtasks;

	i=1+tlist[0].dwProcessId;
	// max 700 windows (I think it's enough)
	// DEBUGR if (i<500)
   if (i<700) // change to 700 in V15.0.1.3
   {
		// discard strange windows
		// GetClientRect(hwnd,&r);
		// if (r.right>4 && r.bottom>4  && IsWindowEnabled(hwnd))
		// if (IsWindow(hwnd)) // is the handle valid?
		// {
		// if (IsWindowEnabled(hwnd)) commented in V12.0.0.22
      //{
      /* DEBUGR
      GetClientRect(hwnd,&r);
      if (r.right>4 && r.bottom>4) // added V15.0.1.3
      {  
      */
         if (IsWindow(hwnd)) // is the handle valid? added in V15.0.1.3
         {
      		// get the window caption
            int title_size = GetWindowText(hwnd, buf, 1000 );
      		if (title_size < 1 || title_size > 256);
            else 
            {
               if (buf[0] != '{')  // forget windows that start with a {  
               { 
      			   tlist[i].hwnd = hwnd;				 // store hwnd
      			   _tcscpy( tlist[i].WindowTitle, buf ); // store window title
            		tlist[0].dwProcessId++;
                  //Get the status of the window
                  //  strcpy(tlist[i].status, "Enabled ");  commented in V12.0.0.22
                  if (IsWindowEnabled(hwnd)) strcpy(tlist[i].status, "Enabled "); // uncommented in V12.0.0.22
                  else strcpy(tlist[i].status, "Disabled ");            // uncommented in V12.0.0.22
                  if (IsWindowVisible(hwnd)) strcat(tlist[i].status, "Visible ");
                  else strcat(tlist[i].status, "Invisible ");
                  if (IsZoomed(hwnd)) strcat(tlist[i].status, "Zoomed ");
                  else if (IsIconic(hwnd)) strcat(tlist[i].status, "Minimized ");
                  if (GetForegroundWindow() == hwnd) strcat(tlist[i].status, "Foreground ");  
                  tlist[0].dwProcessId++;  			   
      			}
      		}
         }	
      // }	
		//}
	}
	return TRUE;
}

/**************************** ListWindows **************************
  Returns list of window data
* Syntax:	rc = ListWindows(directory, diag, diagmsgs)
 *
 * Params:		
 *			   directory - directory of data which has mutiple arrays (window_handle, window_title, Window_class, Window_state, id )
 *			             - also has the window title which may be shorten for Google Eearth, OpenCPN and OpenCPN_GooglePlugin 
 *			   in_pid  - if 0 all windows will be returned, if 1 then only specific windows will be returned, otherwise its a pid of a window to return
 *			             along with the specific windows           
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages 
 *			   handles_taken -  an optional directory of handles taken so they don't return if in the list
 *			   layer_name - an optional layer_name
 * Return: 0 if OK else 1  and fills arrays with data
 *********************************************************************/
RexxRoutine6(int, ListWindows, RexxObjectPtr, data, int, in_pid, int, diag, RexxArrayObject, diagmsgs, OPTIONAL_RexxObjectPtr, handles_taken, OPTIONAL_CSTRING, layer_name)
{
   int rc;
   int id;
   HWND  hwnd, firstchild, childnext;
   RexxObjectPtr dataobj;
   RexxDirectoryObject data_dir;
   logical_t flag;
   DWORD	pid;
   bool return_win;
   char retstr[5000];        // moved to start in V32.0.0.2
   char child_retstr[5000];
   char title[1000];
   char idstr[100]; 
   char process_id[100];
   char class_name[256]; 
   char child_class_name[256];  
   char child_title[256]; 
   int class_name_length;
   char child_status[256]; 
   char wnd_title[1000];
   char child_idstr[100]; 
   char handle[100];
   char child_handle[100];
   char handle_buffer[100];
   char buffer[1000];
   char buffer2[100];
   bool downloads_array_exists = false;
   
   flag = context->IsArray(diagmsgs);
   if ( flag == 0 )   
   {
      printf("ERROR: ListWindows parameter 4 diagmsgs is not an array!");
      // rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }
 
   data_dir = (RexxDirectoryObject)data;
   flag = context->IsDirectory(data_dir);
   if ( flag == 0 )   
   {
      sprintf(buffer,"ERROR: ListWindows parameter 1 is not a directory!");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }

   // Get pointer to the downloads array  added in V4.0.2.17
   RexxArrayObject downloads_array;
   downloads_array = (RexxArrayObject)context->DirectoryAt(data_dir, "downloads"); 
   if (downloads_array != NULLOBJECT)
      {// Check its an array
      flag = context->IsArray(downloads_array);
      if ( flag == 0 )
      {
         sprintf(buffer,"ERROR: ListWindows downloads is not an array!");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return -1;
      } 
      downloads_array_exists = true;
      sprintf(buffer,"DIAG LISTWINDOWS optional downloads array found and is valid");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   }
   
   // if there is the optional handles_taken check its valid directory
   RexxDirectoryObject handles_taken_dir;
   if( argumentExists(5) )
   { 
      handles_taken_dir = (RexxDirectoryObject)handles_taken;
      flag = context->IsDirectory(handles_taken_dir);
      if ( flag == 0 )
      {
         sprintf(buffer,"ERROR: ListWindows parameter 5 is not a directory!");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
      }
      else
      {
         if ( diag>0 )
         { 
            sprintf(buffer,"DIAG LISTWINDOWS optional handles_taken_dir found and is valid");
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         }   
      }
   }
   
   // if there is the optional layer_name get it
   if( argumentExists(6) )
   { 
      if(StrCmpN(layer_name,"No",2))
      {
         sprintf(buffer,"DIAG LISTWINDOWS layer_name=%s found",layer_name);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      }  
   } 
	//structures to contain the task info
   TASK_LIST_ENUM    te;
	// DEBUGR TASK_LIST   tlist[500];
   TASK_LIST   tlist[700];  // increased to 700 was 500 in V15.0.1.3
	//reserved slot for task enum
    tlist[0].dwProcessId = 0;
	// task list
    te.tlist = tlist;
	// get the windows
   EnumWindows(EnumWindowsProc,(LPARAM)(&te));

   // printf("LISTWINDOWS HERE 1\n");
   
   // If pid = 0 and diag GT 0 added in V12.0.3.2
   if ( in_pid == 0 && diag>0 )
   {         
      sprintf(buffer,"DIAG LISTWINDOWS IN_PID is zero returning all the windows");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   }  

	int numtasks=tlist[0].dwProcessId;
   
   sprintf(buffer,"DIAG LISTWINDOWS the numtasks=%d",numtasks); // added n V15.0.1.3
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 

	// process the list
	for(int32_t i=1;i<numtasks;i++)
   {
      return_win = false;
      
      hwnd = (HWND)tlist[i].hwnd ; // get the window handle
      // Check if we have a valid window handle if not forget it   // added in V32.0.0.3
      if (!IsWindow((HWND)hwnd) )  continue;  
      
      // add title to the retstr output
      strcpy(title, tlist[i].WindowTitle); // get the window title 
      int title_len = strlen(title); 
      if(title_len < 1 || title_len > 256) continue;
      sprintf(retstr,"title=%s",title);  

      // Add the handle 
      pointer2string(handle, hwnd); // convert the handle to string
      sprintf(buffer2,"%s",handle); 
      sprintf(buffer,"handle=%s",handle); 
      strcpy(handle_buffer, buffer2); // save the string handle for diagnostics 
      strcat(retstr,buffer);

      // add the class name and add 
      class_name_length = GetClassName(hwnd, class_name, 255);
      if(class_name_length < 1 || class_name_length > 256) continue;
      sprintf(buffer,"class=%s",class_name);  
      strcat(retstr,buffer);
      
       // add the id
      id = GetDlgCtrlID(hwnd);
      sprintf(idstr,"%d",id);
      sprintf(buffer,"id=%s",idstr);  
      strcat(retstr,buffer);
      
      // Add the pid  added in V12.0.0.23
      GetWindowThreadProcessId(hwnd, &pid);     
      dword2string(process_id, (void *)pid); // convert the pid to string
      sprintf(buffer,"process=%s",process_id);  
      strcat(retstr,buffer);
 
      // add the status
      sprintf(buffer,"status=%s",tlist[i].status);  
      strcat(retstr,buffer);
        
      // return all the windows if pid is 0 moved here in V12.0.3.2
      if (in_pid == 0) return_win = true;
      else
      {       
         if (in_pid == pid &&  strstr(tlist[i].status,"Enabled"))
         {
            return_win = true;  
            strcpy(handle,"PID_WIN");   
            if ( diag>0 )
            {         
               sprintf(buffer,"DIAG LISTWINDOWS FOUND PID_WIN i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }          
         }             
               
         // if this is the Google Earth set the handle as "Google Earth"
         if(!strcmp(title,"Google Earth") || !strcmp(title,"Google Earth Pro"))
         {
            if (!strcmp(class_name,"QWidget") || !strcmp(class_name,"Qt5QWindowIcon"))  // added  Qt5QWindowIcon in V25.0.0.12
            {
               // Check if has a firstchild
               (HWND)firstchild = GetWindow(hwnd, GW_CHILD);
               if (firstchild != NULL)
               {   
                  GetWindowText(firstchild,wnd_title,sizeof(wnd_title)); // added in V12.0.0.22
                  // check its first child is called menubar added in V12.0.0.22
                  if (!strcmp(wnd_title,"menubar") || !strcmp(wnd_title,"menubarWindow")) // added  menubarWindown in V25.0.0.12 
                  {     
                     return_win = true;
                     strcpy(handle,"Google Earth");
                     if ( diag>0 )
                     {
                        sprintf(buffer,"DIAG LISTWINDOWS FOUND GE i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
                        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
                     }
                  }
               }
            }
         }
            
         // if this is the OpenCPN set "OpenCPN" as the handle
         // else if(!memcmp(title,"OpenCPN ",8) && !strcmp(class_name,"wxWindowClassNR"))
         else if(!memcmp(title,"OpenCPN ",8) && (!strcmp(class_name,"wxWindowClassNR") || !strcmp(class_name,"wxWindowNR") ))  // changed in V17.0.1.7
         {
            strcpy(handle,"OpenCPN");
            return_win = true; 
            if ( diag>0 )
            {         
               sprintf(buffer,"DIAG LISTWINDOWS FOUND OPENCPN i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            } 
            
            // Check if the firstchild is the openCPN GE Plugin
            firstchild = GetWindow(hwnd, GW_CHILD);
            if (firstchild != NULL)
            {
               GetWindowText(firstchild, child_title, 255);  // get the title
               class_name_length = GetClassName(firstchild, child_class_name, 255);  // get the classname
               // if (!strcmp(child_title,"GoogleEarth") && IsWindowVisible(firstchild) && !strcmp(child_class_name,"wxWindowClassNR")) 
               if (!strcmp(child_title,"GoogleEarth") && IsWindowVisible(firstchild) && (!strcmp(child_class_name,"wxWindowClassNR") || !strcmp(child_class_name,"wxWindowNR"))) 
               {
                  // Create another string for the plugin 
                  sprintf(child_retstr,"title=%s",child_title);  
            
                  // Add the handle 
                  pointer2string(child_handle, firstchild); // convert the handle to string
                  sprintf(buffer,"handle=%s",child_handle);  
                  strcat(child_retstr,buffer);
            
                  // add the class name and add 
                  sprintf(buffer,"class=%s",child_class_name);  
                  strcat(child_retstr,buffer);
                  
                   // add the id
                  id = GetDlgCtrlID(firstchild);
                  sprintf(child_idstr,"%d",id);
                  sprintf(buffer,"id=%s",child_idstr);  
                  strcat(child_retstr,buffer);
             
                  // add the status
                  strcpy(child_status, "Visible ");
                  if (IsWindowEnabled(firstchild)) strcat(child_status, "Enabled ");
                  else strcat(child_status, "Disabled ");
                  if (IsZoomed(firstchild)) strcat(child_status, "Zoomed ");
                  else if (IsIconic(firstchild)) strcat(child_status, "Minimized ");
                  if (GetForegroundWindow() == firstchild) strcat(child_status, "Foreground ");  	               
                  
                  sprintf(buffer,"status=%s",child_status);  
                  strcat(child_retstr,buffer);               
                  
                  // add the string to the directory with the OPENCPN_GE_PLUGIN as the index
                  dataobj = context->String(child_retstr);
                  context->DirectoryPut(data_dir, dataobj, "OPENCPN_GE_PLUGIN");
                  
                  if ( diag>0 )
                  {               
                     sprintf(buffer,"DIAG LISTWINDOWS FOUND OPENCPN_GE_PLUGIN DOCKED title=%s",child_title);
                     rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));         
                  }
               }
            }          
         } 
         
         // if this is the OpenCPN GE PLUGIN and its visible set "OPENCPN_GE_PLUGIN" as the handle
         if(!strcmp(title,"GoogleEarth") &&  (!strcmp(class_name,"wxWindowClassNR") ||!strcmp(class_name,"wxWindowNR")) && strstr(tlist[i].status,"Visible"))
         {
            return_win = true;
            strcpy(handle,"OPENCPN_GE_PLUGIN"); 
            
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND OPENCPN_GE_PLUGIN NOT DOCKED i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));   
            }         
         }       
         
         // if this is the Sat2Chart Plugin set the the handle as GE2KAP_PLUGIN
         if(!memcmp(title,"GE2KAP_PLUGIN_",14) &&  !strcmp(class_name,"#32770"))
         {
            return_win = true;
            strcpy(handle,"GE2KAP_PLUGIN"); 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND GE2KAP_PLUGIN i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));   
            }         
         }    
         
         // if this is the Google Earth Pro New Placemark window 
         if (!strcmp(title,"Google Earth - New Placemark") && strstr(tlist[i].status,"Visible") )                
         {
            // printf("DEBUG HERE\n");
            return_win = true;
            strcpy(handle,"GEPRO_PLACEMARK"); 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND GEPRO_PLACEMARK i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));   
            }         
         }    
         
         // if this is the OpenCPN Active Logbook  set the the handle as OPENCPN_ACTIVE_LOGBOOK
         if(StrStrI(title,"Active Logbook") &&  !strcmp(class_name,"#32770"))
         {
            return_win = true;
            strcpy(handle,"OPENCPN_ACTIVE_LOGBOOK"); 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND OPENCPN_ACTIVE_LOGBOOK i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }           
         }  
             
         // if this is a ocpnToolbarDialog     added in V12.0.0.23
         if(!strcmp(title,"ocpnToolbarDialog") &&  !strcmp(class_name,"#32770"))
         {
            return_win = true;
            strcpy(handle,"ocpnToolbarDialog"); 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND ocpnToolbarDialog i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }           
         } 
         
         // if this is the Typing Test Window    added in V19.0.2.1
         if(!strcmp(title,"Typing Test Please wait") &&  !strcmp(class_name,"#32770"))
         {
            return_win = true;
            strcpy(handle,"TYPING TEST");
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND TYPING TEST i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }           
         }       
         
         // if this is the Google Earth Help About Window  added in V14.0.0.20
         if (!strcmp(class_name,"QWidget") && strstr(tlist[i].status,"Visible") )
         {
            // Check if has a firstchild named qt_msgbox_label
            (HWND)firstchild = GetWindow(hwnd, GW_CHILD);
            if (firstchild != NULL)
            {   // Check if the second child is the copyright_label_
               (HWND)childnext = GetWindow(firstchild, GW_HWNDNEXT);
               if (childnext != NULL)
               { 
                  GetWindowText(childnext,wnd_title,sizeof(wnd_title)); 
                  if (!StrCmpN(wnd_title,"copyright_label_",16) &&  strstr(tlist[i].status,"Enabled"))  // check its first child is called menubar changed to use StrCmpN
                  {                                                                                // in V12.0.0.39 and that is Enabled
                     return_win = true;                                                                  
                     strcpy(handle,"GE HELP_ABOUT");
                     if ( diag>0 )
                     {
                        sprintf(buffer,"DIAG LISTWINDOWS FOUND GE HELP_ABOUTs i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
                        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
                     }
                  }      
               }
            }
         }                           
         
         // if this is the Google Earth message window"
         if(!strcmp(title,"Google Earth") || !strcmp(title,"Google Earth Pro"))
         {
            if (!strcmp(class_name,"QWidget"))
            {
               // Check if has a firstchild
               (HWND)firstchild = GetWindow(hwnd, GW_CHILD);
               if (firstchild != NULL)
               {   
                  GetWindowText(firstchild,wnd_title,sizeof(wnd_title)); // added in V12.0.0.22
   
                  if (strcmp(wnd_title,"menubar") &&  strstr(tlist[i].status,"Enabled"))  // check its first child is NOT called menubar added in V12.0.0.22
                  {                                                                       // and that is Enabled
                     return_win = true;                                                                
                     strcpy(handle,"GE MSG");
                     if ( diag>0 )
                     {
                        sprintf(buffer,"DIAG LISTWINDOWS FOUND GE MSG i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
                        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
                     }
                  }
               }
            }
         }      
         
         // if this is the Google Earth warning message
         if (!strcmp(class_name,"QWidget"))
         {
            // Check if has a firstchild named qt_msgbox_label
            (HWND)firstchild = GetWindow(hwnd, GW_CHILD);
            if (firstchild != NULL)
            {   
               GetWindowText(firstchild,wnd_title,sizeof(wnd_title)); // added in V12.0.0.22
   
               // if (strstr(wnd_title,"qt_msgbox_") &&  strstr(tlist[i].status,"Enabled"))  // check its first child is called menubar added in V12.0.0.22
               if (!StrCmpN(wnd_title,"qt_msgbox_",10) &&  strstr(tlist[i].status,"Enabled"))  // check its first child is called menubar changed to use StrCmpN
               {                                                                                // in V12.0.0.39 and that is Enabled
                  return_win = true;                                                                  
                  strcpy(handle,"GE WARNING MSG");
                  if ( diag>0 )
                  {
                     sprintf(buffer,"DIAG LISTWINDOWS FOUND GE WARNING MSG i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
                     rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
                  }
               }
            }
         }
         
        // Check if Google Translate window  added in V13.0.0.27
         if( !StrCmpN(title,"Google Translate - ",19) && strstr(tlist[i].status,"Visible") )
         {                 
            return_win = true;
            strcpy(handle,"GOOGLE TRANSLATE"); 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND GOOGLE TRANSLATE i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }           
         }  
         
         // Check if SasPlanet added in V15.0.1.3
         //DEBUGR if(!StrCmpN(title,"SAS.",4)  && !strcmp(class_name,"TfrmMain") && strstr(tlist[i].status,"Visible") )
         if(StrStrI(title,"SAS.")  && !strcmp(class_name,"TfrmMain") && strstr(tlist[i].status,"Visible") )
         {                 
            return_win = true;
            strcpy(handle,"SASPLANET"); 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND SASPLANET i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }    
             
            
            // If there is both a windows taken and a downloads array added in V15.0.1.3 
            if(argumentExists(5) && downloads_array_exists) 
            { 
               // if check if its in the handles_taken_dir
               RexxObjectPtr obj;
               obj = context->DirectoryAt(handles_taken_dir, handle_buffer);
               // if  handle has not been used add it to the downloads array
               if(obj ==  NULLOBJECT)
               { 
                  rc = context->ArrayAppendString(downloads_array, handle_buffer, strlen(handle_buffer));
      
                  sprintf(buffer,"DIAG LISTWINDOWS added handle:%s",handle_buffer);
                  rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
                  
                  if ( diag>0 )                                                                                                          
                  {                                                                                                              
                     sprintf(buffer,"DIAG LISTWINDOWS FOUND APPENDED to SasPlanet list i=%d handle=%s pid=%d title=%s status=%s ",i,handle_buffer,pid, title,tlist[i].status);
                     rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
                  } 
               } 
            }           
         }
         
         // Check if Sat2Chart added in V29.0.1.9  added in_pid > 0 in V10.0.6.1
         if(in_pid > 0 && !StrCmpN(title,"Sat2Chart",9) && !strcmp(class_name,"#32770") )
         {                 
            return_win = true;
            strcpy(handle,"Sat2Chart"); 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND Sat2Chart i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }           
         }
                           
         // Check if SasPlanet error like Line Write Failure added in V18.0.1.9
         if( (!StrCmpN(title,"Attention",9) || !StrCmpN(title,"Aten",5)) && !strcmp(class_name,"#32770") && strstr(tlist[i].status,"Visible") )
         {                 
            return_win = true;
            strcpy(handle,"SASPLANET_ATTENTION"); 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND SASPLANET_ATTENTION i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }           
         }    
   
         // Check if SasPlanet Export window added in V4.0.2.6
         if( !StrCmpN(title,"Export",6) && !strcmp(class_name,"TfrmProgressSimple") && strstr(tlist[i].status,"Visible") )
         {                 
            return_win = true;
            strcpy(handle,"SASPLANET_EXPORT"); 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND SASPLANET_EXPORT i=%d,title=%s class=%s pid=%d status=%s handle=%s",i,title,class_name,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }           
         } 
            
         // Check if SasPlanet Error  added in V18.0.1.11
         if( !StrCmpN(title,"SAS.",4) && !strcmp(class_name,"#32770") && strstr(tlist[i].status,"Visible") )
         {                 
            return_win = true;
            strcpy(handle,"SASPLANET_ERROR"); 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND SASPLANET_ERROR i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }           
         }      
         
         // Check if Error  added in V31
         if( !StrCmpN(title,"Error",5) && !strcmp(class_name,"#32770") && strstr(tlist[i].status,"Visible") )
         {                 
            return_win = true;
            strcpy(handle,"ERROR"); 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND ERROR i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }           
         }        
         
         // Check if SasPlanet Open added in V28.0.1.3 added support for German, French and Spanish in V30.0.7.0
         // added support for swedish in v1.0.0.1
         if( (!StrCmpN(title,"Open",4) || !StrCmpN(title,"ffnen",6) || !StrCmpN(title,"Ouvrir",6) || 
               !StrCmpN(title,"Abierto",7) || !StrCmpN(title,"ppna",5) || !StrCmpN(title,"Abrir",5) || !StrCmpN(title,"Apri",4)) 
              && !strcmp(class_name,"#32770") && strstr(tlist[i].status,"Visible") )
         {                 
            return_win = true;
            strcpy(handle,"SASPLANET_OPEN"); 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND SASPLANET_OPEN i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }           
         }     
              
         // Check if SasPlanet Download running window   added in V28.0.1.11
         if( !strcmp(class_name,"TfrmProgressDownload") && strstr(tlist[i].status,"Visible") )
         {   
            // If there is both a windows taken and a downloads array added in V4.0.2.17 
            if(argumentExists(5) && downloads_array_exists) 
            { 
               // if  handle has not been used add it to the downloads array
               RexxObjectPtr obj;
               obj = context->DirectoryAt(handles_taken_dir, handle_buffer);
               if(obj ==  NULLOBJECT)
               { 
                  rc = context->ArrayAppendString(downloads_array, handle_buffer, strlen(handle_buffer));
      
                  sprintf(buffer,"DIAG LISTWINDOWS added handle:%s",handle_buffer);
                  rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
               } 
            } 
      
            return_win = true;
            // Check if its a layer download window   
            if(argumentExists(6) )    // if there is an optional layer   
            {
   
               if(strstr(title,layer_name) && StrCmpN(layer_name,"No",2))  // if this windows is that layer and its not "No"
               {
                  strcpy(handle,"DOWNLOAD_LAYER"); 
                  if ( diag>0 )
                  {
                     sprintf(buffer,"DIAG LISTWINDOWS FOUND DOWNLOAD_LAYER i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
                     rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
                  } 
               }
               else   //  added in V31.0.2.3
               {
                  strcpy(handle,"DOWNLOADRUNNING"); 
                  if ( diag>0 )
                  {
                     sprintf(buffer,"DIAG LISTWINDOWS FOUND DOWNLOADRUNNING i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
                     rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
                  } 
               }   
            }     
            else 
            {
               strcpy(handle,"DOWNLOADRUNNING"); 
               if ( diag>0 )
               {
                  sprintf(buffer,"DIAG LISTWINDOWS FOUND DOWNLOADRUNNING i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
                  rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
               } 
            }          
         }                 
                            
         // Check if SasPlanet  Selection Manager added in V15.0.1.2
         if(!strcmp(class_name,"TfrmRegionProcess") && strstr(tlist[i].status,"Visible") ) // removed title in V19.0.0.0
         {                 
            return_win = true;
            strcpy(handle,"SELECTION MANAGER"); 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND SELECTION MANAGER i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }           
         }    
         
         // Check if SasPlanet  Selection by coordinates added in V15.0.1.3
         if(!strcmp(class_name,"TfrmLonLatRectEdit") && strstr(tlist[i].status,"Visible") )    // removed title in V19.0.0.0
         {                 
            return_win = true;
            strcpy(handle,"SELECTION BY COORDINATES"); 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND SELECTION BY COORDINATES i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }           
         }        
         
         // Check if SasPlanet  Go to... by coordinates added in V15.0.1.3
         if(!strcmp(class_name,"TfrmGoTo") && strstr(tlist[i].status,"Visible") )    // removed title in V19.0.0.0
         {                 
            return_win = true;
            strcpy(handle,"GO TO..."); 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND GO TO... i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }           
         }   
         
         // Check if Worker Window added in V30.0.9.18
         if(!strcmp(class_name,"Worker Window")  )  
         {                 
            return_win = true;
            strcpy(handle,"WORKER"); 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND WORKER i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }           
         }   
         
         // Check if SasPlanet  Interface Options added in V18.0.1.9
         if(!strcmp(class_name,"TfrmMapLayersOptions") && strstr(tlist[i].status,"Visible") )    // removed title in V19.0.0.0
         {                 
            return_win = true;
            strcpy(handle,"INTERFACE_OPTIONS"); 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND INTERFACE_OPTIONS i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }           
         }   
                                                                                                                                                                  
         // Check if SasPlanet  Tfrm.... and not any of the the above and its Visible window   added in V18.0.1.9  
         // Added TfrmAdvancedSystemCare_Monitor to the NOT list in V11.0.0.1 
         if( !StrCmpN(class_name,"Tfrm",4)  && !(!strcmp(class_name,"TfrmLonLatRectEdit") || !strcmp(class_name,"TfrmProgressSimple") ||
                       !strcmp(class_name,"TfrmGoTo") || !strcmp(class_name,"TfrmMain") || !strcmp(class_name,"TfrmRegionProcess") || 
                       !strcmp(class_name,"TfrmProgressDownload") || !strcmp(class_name,"TfrmMapLayersOptions") 
                       || !strcmp(class_name,"TfrmMarksMultiEdit") || !strcmp(class_name,"TfrmAdvancedSystemCare_Monitor")) 
                       && strstr(tlist[i].status,"Visible") )
         {                 
            return_win = true;
            strcpy(handle,"TFRM..."); 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND TFRM... i=%d,title=%s class=%s pid=%d status=%s handle=%s",i,title,class_name,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }           
         }                                                        
         
         // Check if Sat2Chart DOWNLOADS WINDOW  added in V13.0.0.27
         if( !StrCmpN(title,"Index of /Downloads - ",20) && strstr(tlist[i].status,"Visible") ) // changed to 20 from 22 for Microsoft Edge
         {                 
            return_win = true;
            strcpy(handle,"Sat2Chart DOWNLOADS"); 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND Sat2Chart DOWNLOADS i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }           
         }        
         
         // NOTE THIS MUST BE LAST WINDOW TEST
         // Change certain locations in title if not lowercase -- change in V12.0.0.31 
         if (title[6] == '') title[6] = ' '; // If char is uppercase blank change to lowercase blank NOTE compare is for an uppercase blank
         if (title[12] == '') title[12] = ' '; // If char is uppercase blank change to lowercase blank NOTE compare is for an uppercase blank
         if (title[13] == '') title[13] = '-'; // if 13th char is a &ndash make in a normAL DASH   NOTE compare is for an uppercase dash
         if (title[14] == '') title[14] = ' '; // If char is uppercase blank change to lowercase blank NOTE compare is for an uppercase blank
         // if this is the New Image Overlay window  or Edit Image window  
         // added check for Qt5QWindowToolSaveBits for GE Pro V7.3.0
         if( !StrCmpN(title,"Google Earth - ",15) &&  strcmp(title,"Google Earth - New Placemark") && (!strcmp(class_name,"QTool")|| !strcmp(class_name,"Qt5QWindowToolSaveBits")) && strstr(tlist[i].status,"Visible") )// changed to !StrCmpN in V12.0.0.39
         {            //      
            return_win = true;
            strcpy(handle,"IMAGE OVERLAY"); 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS FOUND GE IMAGE OVERLAY i=%d,title=%s pid=%d status=%s handle=%s",i,title,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }           
         }     
      }                                             
      
      /* DEBUG
      // return all the windows if pid is 0
      if (in_pid == 0)                 
         return_win = true;
      */   
      
      // if there is a handles_taken directory and this window is selected check it hasn't been used already  
      if( argumentExists(5) && return_win )
      {
         RexxObjectPtr obj;
         obj = context->DirectoryAt(handles_taken_dir, handle_buffer);
         if(obj ==  NULLOBJECT)
         {
            if ( diag>0 )
            {        
               sprintf(buffer,"DIAG LISTWINDOWS  i=%d,title=%s class=%s, pid=%d status=%s handle=%s has not been used and will be returned",i,title,class_name,pid,tlist[i].status,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }
         } 
         else 
         {
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG LISTWINDOWS title%s handle=%s is used already and will not be returned",title,handle_buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            }
         }
         // IF NOT NULL THEN DON'T USE IT
         if (obj != NULLOBJECT)
         {
            return_win = false; 
         }
      }
      
      if (return_win)  // if we are to return this window do so
      {
         /* DEBUG
         // Check if Sat2Chart added in V29.0.1.9
         if(!StrCmpN(title,"Sat2Chart",9) && !strcmp(class_name,"#32770") )
         { 
            printf("LISTWINDOWS FOUND %s\n",retstr);
         }
         */
      
         // add the string to the directory with the handle as the index
         dataobj = context->String(retstr);
         context->DirectoryPut(data_dir, dataobj, handle); 
      }        
	} 
   
   return 0;  
}   
/**************************** Georef_Calculate_Coefficients_Proj **************************
  Calculates the georef coefficients
* Syntax:	rc = Georef_Calculate_Coefficients_Proj(Cpoints, diag, diagmsgs)
 *
 * Params:		
 *			   data - directory of data needed
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages   
 * Return: 0 if OK else 1  and fills in the pwx, pwy, wpx and wpy array with geo references
 *********************************************************************/
RexxRoutine3(int, Calculate_GeoRefs, RexxObjectPtr, data, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512]; 
   int rc;
   int  r1, r2, r3, r4;
   int mp;   
   int nRefpoint,sz,n;
   int txmin,txmax,tymin,tymax;
   double lonmin,lonmax,latmin,latmax;
   double value;
   logical_t flag;
   double *tx, *ty, *lon, *lat, *pwx, *pwy, *wpx, *wpy;  // local tx array
   RexxArrayObject tx_rexx,ty_rexx,lon_rexx,lat_rexx,pwx_rexx,pwy_rexx,wpx_rexx,wpy_rexx;
   
   RexxObjectPtr dataobj;
   RexxDirectoryObject data_dir;
 
   data_dir = (RexxDirectoryObject)data;
   flag = context->IsDirectory(data_dir);
   if ( flag == 0 )
   {
      sprintf(buffer,"ERROR: GEOREF parameter 1 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 0;
   }
   
   //Get all necessary values that are not arrays
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "lonmin");
   flag = context->ObjectToDouble(dataobj, &lonmin);
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "lonmax");
   flag = context->ObjectToDouble(dataobj, &lonmax);   
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "latmax");
   flag = context->ObjectToDouble(dataobj, &latmax);  
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "latmin");
   flag = context->ObjectToDouble(dataobj, &latmin); 
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "tymin");
   flag = context->ObjectToInt32(dataobj, &tymin);       
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "tymax");
   flag = context->ObjectToInt32(dataobj, &tymax);     
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "txmax");
   flag = context->ObjectToInt32(dataobj, &txmax);  
   dataobj = (RexxObjectPtr)context->DirectoryAt(data_dir, "txmin");
   flag = context->ObjectToInt32(dataobj, &txmin);            

   if ( diag>0 )
   {
      sprintf(buffer,"DIAG GEOREF latmin=%f,latmax=%f,lonmin=%f,lonmax=%f\n",latmin,latmax,lonmin,lonmax);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      sprintf(buffer,"DIAG GEOREF txmin=%d,txmax=%d,tymin=%d,tymax=%d\n",txmin,txmax,tymin,tymax);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }   
 
   // Get pointers to all the rexx arrays
   tx_rexx = (RexxArrayObject)context->DirectoryAt(data_dir, "tx");
   ty_rexx = (RexxArrayObject)context->DirectoryAt(data_dir, "ty");
   lat_rexx = (RexxArrayObject)context->DirectoryAt(data_dir, "lat");
   lon_rexx = (RexxArrayObject)context->DirectoryAt(data_dir, "lon");
   pwx_rexx = (RexxArrayObject)context->DirectoryAt(data_dir, "pwx");
   pwy_rexx = (RexxArrayObject)context->DirectoryAt(data_dir, "pwy");
   wpx_rexx = (RexxArrayObject)context->DirectoryAt(data_dir, "wpx"); 
   wpy_rexx = (RexxArrayObject)context->DirectoryAt(data_dir, "wpy"); 
    
   sz = context->ArrayItems(tx_rexx);  // the number of items
   
   nRefpoint = sz;  // this is the number of reference points too
   
  
   if ( diag>0 )
   {
      sprintf(buffer,"DIAG GEOREF number of items in  tx=%d\n",sz);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }
     
   // Allocate all the local arrays
   tx  = (double *)malloc(nRefpoint * sizeof(double));
   ty  = (double *)malloc(nRefpoint * sizeof(double));
   lon = (double *)malloc(nRefpoint * sizeof(double));
   lat = (double *)malloc(nRefpoint * sizeof(double));

   pwx = (double *)malloc(12 * sizeof(double));
   wpx = (double *)malloc(12 * sizeof(double));
   pwy = (double *)malloc(12 * sizeof(double));
   wpy = (double *)malloc(12 * sizeof(double));
   
    //      Zero out the points
   pwx[6] = pwy[6] = wpx[6] = wpy[6] = 0.;
   pwx[7] = pwy[7] = wpx[7] = wpy[7] = 0.;
   pwx[8] = pwy[8] = wpx[8] = wpy[8] = 0.;
   pwx[9] = pwy[9] = wpx[9] = wpy[9] = 0.;
   pwx[3] = pwy[3] = wpx[3] = wpy[3] = 0.;
   pwx[4] = pwy[4] = wpx[4] = wpy[4] = 0.;
   pwx[5] = pwy[5] = wpx[5] = wpy[5] = 0.;
   pwx[0] = pwy[0] = wpx[0] = wpy[0] = 0.;
   pwx[1] = pwy[1] = wpx[1] = wpy[1] = 0.;
   pwx[2] = pwy[2] = wpx[2] = wpy[2] = 0.;
   
   // Transfer all the rexx arrays into local arrays
   for(n=0 ; n<nRefpoint ; n++)
   { 
      dataobj = context->ArrayAt(tx_rexx, n+1);
      flag = context->ObjectToDouble(dataobj, &value);
      tx[n] = value; 
      dataobj = context->ArrayAt(ty_rexx, n+1);
      flag = context->ObjectToDouble(dataobj, &value);
      ty[n] = value; 
      dataobj = context->ArrayAt(lat_rexx, n+1);
      flag = context->ObjectToDouble(dataobj, &value);
      lat[n] = value; 
      dataobj = context->ArrayAt(lon_rexx, n+1);
      flag = context->ObjectToDouble(dataobj, &value);
      lon[n] = value;                   
   }   
   
   mp =6;  // changed to 6 was 3 in V17.0.1.13   DEBUG
   // mp = 3;  
   
   sprintf(buffer,"DIAG GEOREF lonmin=%f lonmax=%f txmax=%d txmin=%d",lonmin,lonmax,txmin,txmax);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
   double hint1 = lonmin - (txmin * (lonmax - lonmin) /(txmax - txmin));
   double hint2 = (lonmax - lonmin) /(txmax - txmin);
   sprintf(buffer,"DIAG GEOREF calculating pwx hint1=%f hint2=%f",hint1, hint2);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));           
   r1 = Georef_Calculate_Coefficients_Onedir(nRefpoint, mp, tx, ty, lon, pwx,
               lonmin - (txmin * (lonmax - lonmin) /(txmax - txmin)),
                             (lonmax - lonmin) /(txmax - txmin),
                              0.);
   hint1 = latmin - (tymin * (latmax - latmin) /(tymax - tymin));
   double hint3 = (latmax - latmin) /(tymax - tymin);
   sprintf(buffer,"DIAG GEOREF calculating pwy hint1=%f hint3=%f",hint1, hint3);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));        
   r2 = Georef_Calculate_Coefficients_Onedir(nRefpoint, mp, tx, ty, lat, pwy,
               latmin - (tymin * (latmax - latmin) /(tymax - tymin)),
                             0.,
                             (latmax - latmin) /(tymax - tymin));

 //      (northing/easting) to pixel(tx,ty)
 //      Calculate and use a linear equation for p[0..2] to hint the solver
 //      Make a dummy double array
   // std::vector<double> pnull(nRefpoint, 1.0);
   hint1 = txmin - ((txmax - txmin) * lonmin / (lonmax - lonmin));
   hint2 = (txmax - txmin) / (lonmax - lonmin);
   sprintf(buffer,"DIAG GEOREF calculating wpx hint1=%f hint2=%f",hint1, hint2);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   r3 = Georef_Calculate_Coefficients_Onedir(nRefpoint, mp, lon, lat, tx, wpx, 
               txmin - ((txmax - txmin) * lonmin / (lonmax - lonmin)),
                             (txmax - txmin) / (lonmax - lonmin),
                              0.0);                                               

   hint1 = tymin - ((tymax - tymin) * latmin / (latmax - latmin));
   hint3 = (tymax - tymin) / (latmax - latmin);
   sprintf(buffer,"DIAG GEOREF calculating wpy hint1=%f hint3=%f",hint1, hint3);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));        
   r4 = Georef_Calculate_Coefficients_Onedir(nRefpoint, mp, lon, lat, ty, wpy, 
               tymin - ((tymax - tymin) * latmin / (latmax - latmin)),
                             0.0,
                             (tymax - tymin) / (latmax - latmin));
   // Added in V17.0.1.13 
   int therc =1;                         
   if((r1) && (r1 < 4) && (r2) && (r2 < 4) && (r3) && (r3 < 4) && (r4) && (r4 < 4))
      therc = 0;   
      
   sprintf(buffer,"DIAG GEOREF r1=%d r2=%d r3=%d r4=%d therc=%d\n",r1,r2,r3,r4,therc);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));                             
                        
   // Transfer pwx,pwy,wpx and wpy local arrays to the rexx arrays
   for(n=0 ; n<10 ; n++)
   { 
      sprintf(buffer,"%.10e\0",pwx[n]);   // %12.9g
      rc = context->ArrayAppendString(pwx_rexx, buffer, strlen(buffer));
      sprintf(buffer,"DIAG GEOREF pwx[%d]=%f",n,pwx[n]);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      sprintf(buffer,"%.10e\0",pwy[n]);
      rc = context->ArrayAppendString(pwy_rexx, buffer, strlen(buffer)); 
      sprintf(buffer,"DIAG GEOREF pwy[%d]=%f",n,pwy[n]);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));        
      sprintf(buffer,"%.10e\0",wpx[n]);
      rc = context->ArrayAppendString(wpx_rexx, buffer, strlen(buffer));  
      sprintf(buffer,"DIAG GEOREF wpx[%d]=%f",n,wpx[n]);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));             
      sprintf(buffer,"%.10e\0",wpy[n]);
      rc = context->ArrayAppendString(wpy_rexx, buffer, strlen(buffer));  
      sprintf(buffer,"DIAG GEOREF wpy[%d]=%f",n,wpy[n]);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));                             
   } 
    
   // free all the arrays we allocated
   free(tx );
   free(ty );
   free(lon );
   free(lat );
   
   free(pwx );
   free(wpx );
   free(pwy );
   free(wpy ); 
                                        
   if((r1) && (r1 < 4) && (r2) && (r2 < 4) && (r3) && (r3 < 4) && (r4) && (r4 < 4))
      return 0;
   else
      return 1;     
}  

/**************************** GetDatum **************************
  Gets the datum information
* Syntax:	rc = GetDatum(datum, Cpoints, diag, diagmsgs)
 *
 * Params:	datum - the datum to find parms for	
 *			   data - directory of data needed
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages   
 * Return: the index for the datum else -1 if not found
 *********************************************************************/
RexxRoutine4(int, GetDatum, CSTRING, datum,  RexxObjectPtr, data, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512]; 
   int rc;   
   // CSTRING datum;
   double flattening, semimajor_axis;
   logical_t flag;
   
   RexxObjectPtr dataobj;
   RexxDirectoryObject data_dir;
   
   extern struct DATUM const gDatum[];
   extern struct ELLIPSOID const gEllipsoid[];


   // short nDatums = sizeof(gDatum)/sizeof(struct DATUM);
   
   data_dir = (RexxDirectoryObject)data;
   flag = context->IsDirectory(data_dir);
   if ( flag == 0 )
   {
      sprintf(buffer,"ERROR: Getdatum parameter 1 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }
   
   if ( diag>0 )
   {
      sprintf(buffer,"DIAG GetDatum datum=%s",datum);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }  
   
   int i = 0;
   while (i < 102)
   {
         if(!strcmp(datum, gDatum[i].name))
         {  
            flattening = gEllipsoid[gDatum[i].ellipsoid].invf;        // flattening
            semimajor_axis = gEllipsoid[gDatum[i].ellipsoid].a;      // semimajor axis
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG GetDatum datum=%s found at index:%d Flattening=%f Semimajor_axis=%f ",datum,i,flattening,semimajor_axis);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            }  
             
            // Transfer the datum parameters back to rexx caller in data_dir
            dataobj = context->DoubleToObject(flattening);
            context->DirectoryPut(data_dir, dataobj, "flattening");
            dataobj = context->DoubleToObject(semimajor_axis);
            context->DirectoryPut(data_dir, dataobj, "semimajor_axis");
            dataobj = context->DoubleToObject(gDatum[i].dx);
            context->DirectoryPut(data_dir, dataobj, "datum_dx");
            dataobj = context->DoubleToObject(gDatum[i].dy);
            context->DirectoryPut(data_dir, dataobj, "datum_dy");
            dataobj = context->DoubleToObject(gDatum[i].dz);
            context->DirectoryPut(data_dir, dataobj, "datum_dz");                        
            return i;
         }
         
         i++;
   }	
   return -1;
}  

/**************************** CalcArcTan2 **************************
 * Calculates the ArcTan2 of two numbers  and returns it
 *********************************************************************/
RexxRoutine4(RexxObjectPtr, CalcArcTan2, double, x, double, y, OPTIONAL_uint32_t, precision, OPTIONAL_CSTRING, units)
{
    TrigFormatter formatter(context, argumentExists(3), precision, units);
    // calculate and return
    return formatter.evaluateArc(x, y, ARCTAN2);
} 

/**************************** lat2tiley **************************
 * Calculates the tiley for a given latitude
 *********************************************************************/
RexxRoutine2(int,lat2tiley, double, lat, int, z)
{
    return (int)(floor((1.0 - log( tan(lat * SPI/180.0) + 1.0 / cos(lat * SPI/180.0)) / SPI) / 2.0 * pow(2.0, z))); 
} 

/**************************** lon2tilex **************************
 * Calculates the tiley for a given latitude
 *********************************************************************/
RexxRoutine2(int,lon2tilex, double, lon, int, z)
{
    return (int)(floor((lon + 180.0) / 360.0 * pow(2.0, z))); 
} 

/**************************** merc_y **************************
 * Converts Mercator latitude to UTM northing meters
 * See: https://wiki.openstreetmap.org/wiki/Mercator#Elliptical_.28true.29_Mercator_Projection
 *********************************************************************/
RexxRoutine1(double,merc_y, double, lat)
{  
        lat = fmin (89.5, fmax (lat, -89.5));
        double phi = deg_rad(lat);
        double sinphi = sin(phi);
        double con = ECCENT * sinphi;
        con = pow((1.0 - con) / (1.0 + con), COM);
        double ts = tan(0.5 * (M_PI * 0.5 - phi)) / con;
        return 0 - R_MAJOR * log(ts);   
}

/**************************** merc_lat **************************
 * Converts Mercator northing meters to latitude
 * See: https://wiki.openstreetmap.org/wiki/Mercator#Elliptical_.28true.29_Mercator_Projection
 *********************************************************************/
RexxRoutine1(double,merc_lat, double, y)
{
        return (((atan(exp(y/6378137))-SPI/4)*360)/SPI); 
}

/**************************** merc_x **************************
 * Converts Mercator to longitude to UTM easting meters
 * See: https://wiki.openstreetmap.org/wiki/Mercator#Elliptical_.28true.29_Mercator_Projection
 *********************************************************************/
RexxRoutine1(double,merc_x, double, lon)
{
      return R_MAJOR * deg_rad(lon);
}

/**************************** merc_lon **************************
 * Converts Mercator easting meters to longitude
 * See: https://wiki.openstreetmap.org/wiki/Mercator#Elliptical_.28true.29_Mercator_Projection
 *********************************************************************/
RexxRoutine1(double,merc_lon, double, x)
{
      return (x/6378137*180/SPI);
}

/**************************** GetTileX **************************
 * Converts longitude to X
 * See: https://wiki.openstreetmap.org/wiki/Mercator#Elliptical_.28true.29_Mercator_Projection
 *********************************************************************/
RexxRoutine1(double,GetTileX, double, longitude)
{
   return 0.5 + longitude / 360.0;
}
/**************************** Varint_Encode **************************
 * Converts a unsigned 64 bit number to a Google buffer varint
 * returns the number of bytes in the returned variant
 * See: https://stackoverflow.com/questions/19758270/read-varint-from-linux-sockets
**************************************************************/

RexxRoutine2(int,Varint_Encode, RexxStringObject, varint, uint64_t, value)
// int encode_unsigned_varint(uint8_t *const buffer, uint64_t value)
{
    int encoded = 0;
    // uint32_t therc = 0;
    // RexxStringObject therc;
    char buffer[25]; 
    
    // printf("DEBUG Varint_Encode value=%I64d\n",value);
    // process all the bytes in value
    do
    {
        uint8_t next_byte = value & 0x7F;
        value >>= 7;

        if (value)
            next_byte |= 0x80;

        buffer[encoded++] = next_byte;
        // buffer[encoded] = '\0';

    } while (value);
    
    // Copy the buffer into the varint CSTRING varint
    const byte *dataptr = reinterpret_cast<const byte*>( context->StringData(varint) );
    buffer[encoded] = '\0';
    // printf("DEBUG Varint_Encode buffer length=%d\n",strlen(buffer));
    // Pass back the buffer into varint CSTRING 
    strcpy((char *)dataptr,buffer);
   
    return encoded;
}
/* KEEP  USING an array to pass back the encoded value
RexxRoutine2(int,Varint_Encode, RexxArrayObject, data, uint64_t, value)
// int encode_unsigned_varint(uint8_t *const buffer, uint64_t value)
{
   int encoded = 0;
   int therc = 0;
   int rc;
   char buffer[25]; 
    
   
   // printf("DEBUG Varint_Encode value=%I64d\n",value);
   
   do
   {
     uint8_t next_byte = value & 0x7F;
     value >>= 7;
   
     if (value)
         next_byte |= 0x80;
   
     sprintf(buffer,"%u\0",next_byte); 
     rc = context->ArrayAppendString(data, buffer, strlen(buffer));
     encoded += 1;
     // buffer[encoded++] = next_byte;
   
   } while (value);
   
   return encoded;
}
*/
/**************************** Varint_Decode **************************
 * Converts a google varint to a a 64 bit Number
 * returns the value of the varint
 * also sets the decoded bytes to the number of bytes decoded
 * See: https://stackoverflow.com/questions/19758270/read-varint-from-linux-sockets
**************************************************************/
RexxRoutine2(uint64_t,Varint_Decode, CSTRING, varint, RexxStringObject, decoded_bytes)
// RexxRoutine1(uint64_t,Varint_Decode, CSTRING, varint)
// uint64_t decode_unsigned_varint( const uint8_t *const data, int &decoded_bytes )
{
    int i = 0;
    uint64_t decoded_value = 0;
    int shift_amount = 0;
    int num_decoded = 0;
    char buffer [20];

    do 
    {
        // printf("DEBUG varint=%u\n",data[i]);
        decoded_value |= (uint64_t)(varint[i] & 0x7F) << shift_amount;     
        shift_amount += 7;
    } while ( (varint[i++] & 0x80) != 0 );
    
    // set how many were decoded
    num_decoded = i;  
    sprintf(buffer,"%d",num_decoded);
    const byte *dataptr = reinterpret_cast<const byte*>( context->StringData(decoded_bytes) );
    // Pass back the number decoded 
    strcpy((char *)dataptr,buffer);

    return decoded_value;
}
/**************************** GetTileY **************************
 * Converts the Latitude into Y for an Ellipsoid
 * See: https://akalinovsky.wordpress.com/2011/11/28/en-custom-map-control-for-windows-phone-7-from-scratch-part-1/
 *********************************************************************/
RexxRoutine1(double,GetTileY, double, latitude)
{ 
      double radiusa = 6378137; 
      double radiusb =  6356752; 
      
      // double e = sqrt (sqrt(radiusa) - sqrt(radiusb)) / radiusa; 
      // double e = 0.081819790992;
      /*
      
      double minus_R = RadiusA - RadiusB; 
      double plus_R = RadiusA + RadiusB;
      double temp = fabs(minus_R * plus_R);
      double e =  sqrt(temp * (RadiusA + RadiusB)) / max(RadiusA, RadiusB);
      // double e =  sqrt(temp * (RadiusA + RadiusB)) / max(RadiusA, RadiusB);
      */
      double e = sqrt(abs(radiusa - radiusb) * (radiusa + radiusb)) / max(radiusa, radiusb);
      return 0.5 * (1 - ((log((1 + sin(latitude * SPI / 180.0)) / (1 - sin(latitude * SPI / 180.0)))) / 2.0 - e * (log((1 + e * sin(latitude * SPI / 180.0)) / (1 - e * sin(latitude * SPI / 180.0)))) / 2.0) / PI);
}

/**************************** GetTileY_Spheroid **************************
 * Converts the Latitude into Y in spheroid
 * See: https://akalinovsky.wordpress.com/2011/11/28/en-custom-map-control-for-windows-phone-7-from-scratch-part-1/
 *********************************************************************/
RexxRoutine1(double,GetTileY_Spheroid, double, latitude)
{   // return 0.5 * (1 - ((Math.Log((1 + Math.Sin(latitude * Math.PI / 180.0)) / (1 - Math.Sin(latitude * Math.PI / 180.0)))) / 2.0) / Math.PI);
   return 0.5 * (1 - ((log((1 + sin(latitude * SPI / 180.0)) / (1 - sin(latitude * SPI / 180.0)))) / 2.0) / SPI);
}
/**************************** cYToLat **************************
 * * Transforms pixel coordinate ellipsoid Y to latitude
 * * @see mobac.mapsources.mapspace.MercatorPower2MapSpace#cYToLat(int, int)
 *********************************************************************/
RexxRoutine2(double,cYToLat, int, y, int, zoom)
{ 
	int y2 = y + falseNorthing(zoom);
	double latitude = exp(-y2 / radius(zoom));
	latitude = cphi2(latitude);
	return -1 * rad_deg(latitude);
}
/**************************** cYToLatSpheroid **************************
 * * Transforms pixel coordinate Y to latitude    Spheroid version
 * * @see https://www.javatips.net/api/mobac-ext-master/src/main/java/mobac/mapsources/mapspace/MercatorPower2MapSpace.java
 *********************************************************************/
RexxRoutine2(double,cYToLatSpheroid, int, y, int, zoom)
{
		y += falseNorthing(zoom);
		double latitude = (SPI / 2) - (2 * atan(exp(-1.0 * y / radius(zoom))));
		return -1 * rad_deg(latitude);
}
/**************************** cXToLon **************************
 * * Transforms pixel coordinate Y to longitude   Ellipsoid version
 * * @see https://www.javatips.net/api/mobac-ext-master/src/main/java/mobac/mapsources/mapspace/MercatorPower2MapSpace.java
 *********************************************************************/
RexxRoutine2(double,cXToLon, int, y, int, zoom)
{
	return ((360.0 * y) / getMaxPixels(zoom)) - 180.0;
}   

/**************************** degreesToPixelsY **************************
 * * converts degrees into pixels Y
 * * @see Rhomobile
 *********************************************************************/
RexxRoutine2(uint64_t,degreesToPixelsY, double, n, int, zoom)
{
    // Merkator projection
    double MAX_SIN = 0.99627207622;
    double sin_phi = sin(n*SPI/180);
    // MAX_SIN - maximum value of sine allowed by Merkator projection
    // (~85.0 degrees of north latitude)
    if (sin_phi < -MAX_SIN) sin_phi = -MAX_SIN;
    if (sin_phi > MAX_SIN) sin_phi = MAX_SIN;
    double ath = rho_math_atanh(sin_phi);
    double val = 256 * rho_math_pow2(zoom) * (1 - ath/SPI)/2;
    return (uint64_t)val;
}

/**************************** degreeEllipsoidToPixelsY **************************
 * * converts degrees into pixels Y   for an  Ellipsoid
 * * @see Rhomobile
 *********************************************************************/
RexxRoutine2(uint64_t,degreeEllipsoidToPixelsY, double, lat, int, zoom)
{
    // Merkator projection
    // double MAX_SIN = 0.99627207622;
    // double sin_phi = sin(n*SPI/180);
    // MAX_SIN - maximum value of sine allowed by Merkator projection
    // (~85.0 degrees of north latitude)
    
    double lat_rads = deg_rad(lat);
    double tmp = tan(M_PI_4 + lat_rads / 2.0); 
    double pow_tmp = pow(tan(M_PI_4 + (asin(YANDEX_E * sin(deg_rad(lat)))) / 2.0), YANDEX_E);
    return (uint64_t)((YANDEX_A - (YANDEX_Rn * log (tmp / pow_tmp))) * YANDEX_F)>>(23-zoom);
    
}
/**************************** getTileNumberX **************************
 * * converts Spheroid Longitude to tilex
 * * @see osmand.util
 *********************************************************************/
RexxRoutine2(double,getTileNumberX, double, zoom, double, longitude )
{
   /*
   public static double getTileNumberX(float zoom, double longitude){
   if(longitude == 180d) {
			return getPowZoom(zoom) - 1;
		}
		longitude = checkLongitude(longitude);
		return (longitude + 180d)/360d * getPowZoom(zoom);
   */
   if(longitude == 180.0) 
   {
   	return getPowZoom(zoom) - 1;
   }
   longitude = checkLongitude(longitude);
   return (longitude + 180.0)/360.0 * getPowZoom(zoom);
}
/**************************** getTileNumberY **************************
 *  converts Spheroid Latitude to tiley
 * * @see osmand.util
 *********************************************************************/
RexxRoutine2(double,getTileNumberY, double, zoom, double, latitude )
{
   
   double eval = log(tan(toRadians(latitude)) + 1/cos(toRadians(latitude)) );
   return (1 - eval / SPI) / 2 * getPowZoom(zoom);
}
/**************************** getTileEllipsoidNumberY **************************
 *  converts Latitude to Ellipsoid tiley
 * * @see osmand.util
 *********************************************************************/
RexxRoutine2(double,getTileEllipsoidNumberY, double, zoom, double, latitude )
{

   double E2 = (double) latitude * SPI / 180;
   double sradiusa = 6378137;
   double sradiusb = 6356752;
   double J2 = (double) sqrt((sradiusa * sradiusa) - (sradiusb * sradiusb))	/ sradiusa;
   double M2 = (double) log((1 + sin(E2))
   		/ (1 - sin(E2)))/ 2- J2	* log((1 + J2 * sin(E2))/ (1 - J2 * sin(E2))) / 2;
   double B2 = getPowZoom(zoom);
   return B2 / 2 - M2 * B2 / 2 / SPI;
}

/**************************** getLatitudeFromEllipsoidTileY **************************
 *  get Latitude from  Ellipsoid tiley
 * * @see osmand.util       DEBUG SOME TIMES GIVES nan use
 *                         pixely = ul_tiley * 256
                           lat = cYToLat(pixely, zm) 
 *********************************************************************/
RexxRoutine2(double,getLatitudeFromEllipsoidTileY, double, zoom, double, tileNumberY )
{

   double MerkElipsK = 0.0000001;
   double sradiusa = 6378137;
   double sradiusb = 6356752;
   double FExct = (double) sqrt((sradiusa * sradiusa)  - (sradiusb * sradiusb)) / sradiusa;
   double TilesAtZoom = getPowZoom(zoom);
   double result = (tileNumberY - TilesAtZoom / 2) / -(TilesAtZoom / (2 * SPI));
   result = (2 * atan(exp(result)) - SPI / 2) * 180 / SPI;
   double Zu = result / (180 / SPI);
   double yy = (tileNumberY - TilesAtZoom / 2);
   double Zum1 = Zu;
   Zu = asin(1 - ((1 + sin(Zum1)) * pow(1 - FExct * sin(Zum1), FExct))
   		/ (exp((2 * yy) / -(TilesAtZoom / (2 * SPI))) * pow(1 + FExct * sin(Zum1), FExct)));
   while (abs(Zum1 - Zu) >= MerkElipsK) 
   {
   	Zum1 = Zu;
   	Zu = asin(1 - ((1 + sin(Zum1)) * pow(1 - FExct * sin(Zum1), FExct))
   			/ (exp((2 * yy) / -(TilesAtZoom / (2 * SPI))) * pow(1 + FExct * sin(Zum1), FExct)));
   }
   
   return Zu * 180 / SPI;
}
/**************************** getLongitudeFromTile **************************
 *  get Longitude from tilex
 * * @see osmand.util
 *********************************************************************/
RexxRoutine2(double,getLongitudeFromTile, double, zoom, double, x )
{  /*  
   i	public static double getLongitudeFromTile(float zoom, double x) {
		return x / getPowZoom(zoom) * 360.0 - 180.0;
   */
	return x / getPowZoom(zoom) * 360.0 - 180.0;
}

/**************************** getLatitudeFromTile **************************
 *  get latitude from tiley
 * * @see osmand.util
 *********************************************************************/
RexxRoutine2(double,getLatitudeFromTile, double, zoom, double, y )
{
	int sign = y < 0 ? -1 : 1;
	return atan(sign*sinh(SPI * (1 - 2 * y / getPowZoom(zoom)))) * 180.0 / SPI;
}
/**************************** GetLatitude **************************
 * * Transforms pixel coordinate Y to latitude
 * * @see https://akalinovsky.wordpress.com/2011/11/28/en-custom-map-control-for-windows-phone-7-from-scratch-part-1/
 *********************************************************************/
RexxRoutine1(double,GetLatitude, double, tileY)
{ 

   bool S = false;
   double math_e = 2.71828182845905;
   double e = 0.081819790992;   //excentricity
   double precision = 0.0000001;
   if (tileY > 0.5)
   {
       S = true;
       tileY = 1 - tileY;
   }
   double y1 = (1.0 - tileY * 2) * SPI;
   double teta0 = 2.0 * atan(pow(math_e, y1)) - SPI / 2.0;

   // double e = sqrt(abs(RadiusA - RadiusB) * (RadiusA + RadiusB)) / max(RadiusA, RadiusB);
   double teta = CalcTeta(teta0, y1, e);
   while (abs(teta - teta0) > precision)
   {
       teta0 = teta;
       teta = CalcTeta(teta, y1, e);
   }
   return (S ? -1 : 1) * teta * 180.0 / SPI;
}

/**************************** GetLongitude **************************
 * * Transforms pixel coordinate Y to latitude
 * * @see https://akalinovsky.wordpress.com/2011/11/28/en-custom-map-control-for-windows-phone-7-from-scratch-part-1/
 *********************************************************************/
RexxRoutine1(double,GetLongitude, double, tileX)
{ 
   return (tileX - 0.5) * 360.0;
}

/**************************** GetLongitude_Spheroid **************************
 * * Transforms pixel coordinate Y to latitude
 * * @see https://akalinovsky.wordpress.com/2011/11/28/en-custom-map-control-for-windows-phone-7-from-scratch-part-1/
 * *********************************************************************/
RexxRoutine1(double,GetLongitude_Spheroid, double, tileX)
{ 
   return (tileX - 0.5) * 360.0;
}

/**************************** MercatorOnEllipsoid_LAt2Y **************************
 * * Transforms latitude to pixel Y which is between -1 to + 1
 * * @seehttps://searchcode.com/file/102993188/Src/CoordConvert/u_CoordConverterMercatorOnEllipsoid.pas
 *********************************************************************/
RexxRoutine1(double,MercatorOnEllipsoid_LAt2Y, double, lat)
{  
  double VRadiusA = 6378137;
  double VRadiusB = 6356752;
  double sqRadA = VRadiusA * VRadiusA;
  double sqRadB = VRadiusB * VRadiusB;
  double FExct = sqrt(sqRadA - sqRadB) / VRadiusA;
  double z = sin(lat * SPI / 180);
  double c = (1 / (2 * SPI));
  return (0.5 - c * (rho_math_atanh(z) - FExct * rho_math_atanh(FExct * z)));
}   
/**************************** MercatorOnEllipsoid_Y2Lat **************************
 * * Transforms pixel Y which is between -1 to + 1 to Latitude
 * * @seehttps://searchcode.com/file/102993188/Src/CoordConvert/u_CoordConverterMercatorOnEllipsoid.pas
 *********************************************************************/
RexxRoutine1(double,MercatorOnEllipsoid_Y2Lat, double, y)
{   
   double Y,yy,Zu,e_y,Zum1,VSin;
   double MerkElipsK = 0.000000001;
   double VRadiusA = 6378137;
   double VRadiusB = 6356752;
   double sqRadA = VRadiusA * VRadiusA;
   double sqRadB = VRadiusB * VRadiusB;
   double FExct = sqrt(sqRadA - sqRadB) / VRadiusA;
   if (y > 0.5)  yy = (y - 0.5);
   else yy = (0.5 - y); 
   yy = yy * (2 * SPI);
   Zu = 2 * atan(exp(yy)) - SPI / 2;
   e_y = exp(2 * yy);
   Y = Zu * (180 / SPI);
   // repeat
   do
   {
      Zum1 = Zu;
      VSin = sin(Zum1);
      Zu = asin(1 - (1 + VSin) * pow((1 - FExct * VSin) / (1 + FExct * VSin), FExct) / e_y);
   } while ((abs(Zum1 - Zu) >= MerkElipsK) ); 
   
   if (y > 0.5) Y = -Zu * 180 / SPI;
   else  Y = Zu * 180 / SPI; 
   return Y;
} 
/**************************** MercatorOnShere_Lat2Y **************************
 * * Transforms latitude to pixel Y which is between -1 to + 1
 * * @seehttps://searchcode.com/file/102993188/Src/CoordConvert/u_CoordConverterMercatorOnShere.pas
 *********************************************************************/
RexxRoutine1(double,MercatorOnShere_Lat2Y, double, lat)
{ 
   double z = sin(lat * SPI / 180);
   double c = 1 / (2 * SPI);
   return 0.5 - 0.5 * log((1 + z) / (1 - z)) * c;
}    
/**************************** MercatorOnShere_Y2Lat **************************
 * *  Transforms pixel Y which is between -1 to + 1 to Latitude
 * * @seehttps://searchcode.com/file/102993188/Src/CoordConvert/u_CoordConverterMercatorOnShere.pas
 *********************************************************************/
RexxRoutine1(double,MercatorOnShere_Y2Lat, double, lat)
{   
   double Y = -(lat - 0.5) * (2 * SPI);
   Y = (2 * atan(exp(Y)) - SPI / 2) * 180 / SPI;
   return Y;
}                            
/****************************  mercator_offset **************************
 * Converts local pixel_x/y to mercator x/y used by Yandex
 * See: https://searchcode.com/file/65796262/src/map_management.c
 * 
 * Params:	
 *			   info - directory containing zoom, pixel_x, pixel_y
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages   
 * Return: the 0 if OK else 1  and if 0 then the mercator_x and mercator_y
 *********************************************************************/
RexxRoutine3(int, mercator_offset, RexxObjectPtr, info, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512]; 
   int rc;
   logical_t flag;
   
   RexxObjectPtr dataobj;
   RexxDirectoryObject info_dir;
   
   info_dir = (RexxDirectoryObject)info;
   flag = context->IsDirectory(info_dir);
   if ( flag == 0 )
   {
      sprintf(buffer,"ERROR: WidePath parameter 1 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }
   
   // get the zoom
   int zoom;
   dataobj = (RexxObjectPtr)context->DirectoryAt(info_dir, "ZOOM");
   flag = context->ObjectToInt32(dataobj, &zoom);  
   
   // get the pixel_x
   double pixel_x;
   dataobj = (RexxObjectPtr)context->DirectoryAt(info_dir, "PIXEL_X");
   flag = context->ObjectToDouble(dataobj, &pixel_x);  
   
   // get the pixel_y
   double pixel_y;
   dataobj = (RexxObjectPtr)context->DirectoryAt(info_dir, "PIXEL_Y");
   flag = context->ObjectToDouble(dataobj, &pixel_y);  

   sprintf(buffer,"DIAG mercator_offset zoom=%d pixel_x=%f pixel_y=%f ",zoom,pixel_x,pixel_y);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
 
//--------------Yandex repository offset ------------------------
   
	 double lat,lon,tmp,pow_tmp,mercator_x,mercator_y,lat_rads;  
    lat=pixel2lat(zoom,pixel_y);
    lon=pixel2lon(zoom,pixel_x);
    lat_rads = deg_rad(lat);
    
    sprintf(buffer,"DIAG mercator_offset lat=%f lon=%f lat_rads=%f ",lat,lon,lat_rads);
    rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
    
    tmp = tan(M_PI_4 + lat_rads / 2.0);   
    // pow_tmp = pow( (tan(M_PI_4 + (asin(YANDEX_E * sin(deg_rad(lat)))) / 2.0)), YANDEX_E);
    pow_tmp = pow(tan(M_PI_4 + (asin(YANDEX_E * sin(deg_rad(lat)))) / 2.0), YANDEX_E);
    mercator_x = (int)((YANDEX_Rn * deg_rad (lon) + YANDEX_A) * YANDEX_F)>>(23-zoom);    
    mercator_y = (int)((YANDEX_A - (YANDEX_Rn * log (tmp / pow_tmp))) * YANDEX_F)>>(23-zoom); 
//--------------Yandex repository offset ------------------------
    // pass back the mercator_x
    dataobj = context->DoubleToObject(mercator_x); 
    context->DirectoryPut(info_dir, dataobj, "MERCATOR_PIXELX"); 
    // pass back the mercator_y
    dataobj = context->DoubleToObject(mercator_y); 
    context->DirectoryPut(info_dir, dataobj, "MERCATOR_PIXELY"); 
    return 0;
}

/**************************** MemUsage **************************
 * Returns the amount of memory being used added in V31
 *********************************************************************/
RexxRoutine0(int, MemUsage)
{
   //get the handle to this process
   auto myHandle = GetCurrentProcess();
   //to fill in the process' memory usage details
   PROCESS_MEMORY_COUNTERS pmc;
   //return the usage (bytes), if I may
   if (GetProcessMemoryInfo(myHandle, &pmc, sizeof(pmc)))
       return(pmc.WorkingSetSize);
   else
       return 0;
} 
/**************************** GetBuf*******************************
  Allocates memory used to trigger a Garbage Collection added in V31
 *********************************************************************/
RexxRoutine1(int, GetBuf,  int, Size)
{   
   RexxBufferObject obj;
   obj = context->NewBuffer(Size);
   if (obj == NULL) return 1;
   return 0; 
} 
/**************************** ReleaseArray *******************************
  Release an Array added in V31
 *********************************************************************/
RexxRoutine1(int, ReleaseArray,  RexxArrayObject, obj)
{   
   // RexxObjectPtr obj;
   // obj = context->GetCallerContext();  
   context->ReleaseGlobalReference(obj);    
   // context.activity->getInstance()->ReleaseGlobalReference(obj);    
   // context->ReleaseLocalReference(theArray);  
   return 0;
} 
/**************************** RunPGM *******************************
  Uses RexxCreateInterpreter to run a pgm  added in V31
 *********************************************************************/
RexxRoutine1(int, RunPGM ,  RexxArrayObject, common_Array)
{
    // RXSYSEXIT exit_list[9];              /* Exit list array                   */
    short    rexxrc = 0;                 /* return code from rexx             */
    int   rc;                            /* actually running program RC       */
    // CONSTRXSTRING argument;              /* rexxstart argument                */
    RXSTRING rxretbuf;                   // program return buffer
   rc = 0;                              /* set default return  
   
   /* By setting the strlength of the output RXSTRING to zero, we   */
   /* force the interpreter to allocate memory and return it to us. */
   /* We could provide a buffer for the interpreter to use instead. */
   rxretbuf.strlength = 0L;          /* initialize return to empty*/ 
   
   // register IO exit

    /* Here we call the interpreter.                                  */
    rc=REXXSTART(0,              /* number of arguments   */
                 NULL,      /* array of arguments    */
                 "Test_Thread2.rex",        /* name of REXX file     */
                 NULL,           /* No INSTORE used       */
                 "CMD",          /* Command env. name     */
                 RXCOMMAND,      /* Code for how invoked  */
                 NULL,           /* exits for this call   */
                 &rexxrc,        /* Rexx program output   */
                 &rxretbuf );    /* Rexx program output   */

    /* free memory allocated for the return result */
    if (rc==0)
    {
        RexxFreeMemory(rxretbuf.strptr);
    }
    RexxDeregisterExit("MY_IOC",NULL);     // remove the exit in exe exit list
    // return interpeter or rexx program return code
    return rc ? rc : rexxrc;   
}


/**************************** UNZIPIT **************************
Unzips a compressed filein to the fileout
* Syntax:	rc = UnzipIt(filein, fileout,  diag, diagmsgs)
 *
 * Params:		
 *			   filein - filename of the KMZ file
 *			   fileout  - filename of the KML file
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages   
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine4(int, UnzipIt, CSTRING, filein, CSTRING, fileout, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512]; 
   int rtnrc = 0;
   int rc;
  
   if ( diag>0 )
   {
      sprintf(buffer,"DIAG UnzipIt Filein=%s Fileout=%s\n",filein, fileout);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }  
   
   HUNZIP		huz;
	DWORD		result;
   
   // Create a the fileout name added in V18.0.1.13
   char filenameoutkml[MAXFILENAME+16];
   strncpy(filenameoutkml, _T(fileout),MAXFILENAME-1);
   /* strncpy doesnt append the trailing NULL, of the string is too long. */
   filenameoutkml[ MAXFILENAME ] = '\0';   
	
	// Open th ZIP archive pointed to by filein
	// if (!(result = UnzipOpenFileA(&huz, _T(filein), 0)))    
   if (!(result = UnzipOpenFile(&huz, _T(filein), 0)))    
	{
		ZIPENTRY	ze;
		DWORD		numitems;
    
		// Find out how many items are in the archive.
		ze.Index = (DWORD)-1;
		// if ((result = UnzipGetItemW(huz, &ze))) 
      if ((result = UnzipGetItem(huz, &ze))) 
      {
         sprintf(buffer,"DIAG Unzipit UnzipGetItemW failedresult=%d",result);   // added in V18.0.1.13
         // printf(buffer);printf("\n");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));      
         UnzipClose(huz);
			return 1;
      } 
      
		numitems = ze.Index;
      if ( diag>0 )  // added in V32.0.0.6
      {
         sprintf(buffer,"DIAG UnzipIt numitems=%d",numitems);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }

		// Unzip each item, using the name stored (in the zip) for that item.
		for (ze.Index = 0; ze.Index < numitems; ze.Index++)
		{
         // if (!(result = UnzipGetItemA(huz, &ze)))    
         if (!(result = UnzipGetItem(huz, &ze))) 
         {
            // if((result = UnzipItemToFileA(huz, filenameoutkml, &ze)))  
            // if((result = UnzipItemToFileA(huz, _T(fileout), &ze)))   
            // if((result = UnzipItemToFile(huz, _T(fileout), &ze)))
            if((result = UnzipItemToFileA(huz, ze.Name, &ze)))
            {
               sprintf(buffer,"DIAG Unzipit UnzipItemToFileA failed result=%d",result);   // added in V18.0.1.13
               // printf(buffer);printf("\n");
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));   
            
             UnzipClose(huz);
             return 1;
            }
            if ( diag>0 )  // added in V32.0.0.6
            {
               sprintf(buffer,"DIAG UnzipIt unzipped index%d Name=%s",ze.Index,ze.Name);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            } 
         }
         
         else
         {  
             sprintf(buffer,"DIAG Unzipit UnzipGetItemA failed result=%d",result);   // added in V18.0.1.13
             // printf(buffer);printf("\n");
             rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));      
             UnzipClose(huz); 
             return 1;
         }          
		}
		
		// Done unzipping files, so close the ZIP archive.
		UnzipClose(huz);
	}	
	else
	{ 
      sprintf(buffer,"DIAG Unzipit UnzipOpenFileA failedresult=%d",result);   // added in V18.0.1.13
      // printf(buffer);printf("\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));      
		return 1;
   }	

	return rtnrc;
} 

/**************************** ZIPIT **************************
Zips a uncompressed filein(KML)) to the fileout KMZ
* Syntax:	rc = ZipIt(filein, fileout,  diag, diagmsgs)
 *
 * Params:		
 *			   filein - filename of the KML file
 *			   imagein - filename of the JPG image file
 *			   fileout  - filename of the KMZ file
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages   
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine5(int, ZipIt, CSTRING, filein, CSTRING, imagein, CSTRING, fileout, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512]; 
   int rtnrc = 0;
   int rc;
   zipFile zf;
   int errclose;
   zlib_filefunc_def ffunc;
   int opt_overwrite=0;
   int err=0;
   int size_buf=0;
   void* buf=NULL;
   const char* password=NULL;
   int opt_exclude_path=1;
   int opt_compress_level=Z_DEFAULT_COMPRESSION;
  
   if ( diag>0 )
   {
      sprintf(buffer,"DIAG ZipIt Filein=%s Fileout=%s\n",filein, fileout);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }  
   
    size_buf = WRITEBUFFERSIZE;
    buf = (void*)malloc(size_buf);
    if (buf==NULL)
    {
        printf("Error allocating memory\n");
        return ZIP_INTERNALERROR;
    }   
	
   if ( diag>0 )
   {
      sprintf(buffer,"DIAG ZipIt Issuing  zipOpen2\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }
	// Create a ZIP archive with the fileout name
	
	fill_win32_filefunc(&ffunc);
   zf = zipOpen2(_T(fileout),(opt_overwrite==2) ? 2 : 0,NULL,&ffunc);
   
   if (zf == NULL)
   {
         printf("error opening %s\n",_T(fileout));
         err= ZIP_ERRNO;
   }
   else
   {
      if ( diag>0 )
      {
         sprintf(buffer,"DIAG ZipIt Opening zip fileout=%s\n",fileout);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }  
       
      // Add the input file to the archive 
         
       FILE * fin;
       int size_read;
       // char* filenameinzip;
       char filenameinzip[MAXFILENAME+16];
       const char *savefilenameinzip;
       zip_fileinfo zi;
       unsigned long crcFile=0;
       
       strncpy(filenameinzip, _T(filein),MAXFILENAME-1);
       /* strncpy doesnt append the trailing NULL, of the string is too long. */
       filenameinzip[ MAXFILENAME ] = '\0';

       zi.tmz_date.tm_sec = zi.tmz_date.tm_min = zi.tmz_date.tm_hour =
       zi.tmz_date.tm_mday = zi.tmz_date.tm_mon = zi.tmz_date.tm_year = 0;
       zi.dosDate = 0;
       zi.internal_fa = 0;
       zi.external_fa = 0;
       filetime(filenameinzip,&zi.tmz_date,&zi.dosDate);   

       /*the path name saved, should not include a leading slash. */
       /*if it did, windows/xp and dynazip couldn't read the zip file. */
       savefilenameinzip = filenameinzip;
       while( savefilenameinzip[0] == '\\' || savefilenameinzip[0] == '/' )
       {
           savefilenameinzip++;
       }

       /*should the zip file contain any path at all?*/
       if( opt_exclude_path )
       {
           const char *tmpptr;
           const char *lastslash = 0;
           for( tmpptr = savefilenameinzip; *tmpptr; tmpptr++)
           {
               if( *tmpptr == '\\' || *tmpptr == '/')
               {
                   lastslash = tmpptr;
               }
           }
           if( lastslash != NULL )
           {
               savefilenameinzip = lastslash+1; // base filename follows last slash.
           }
       }

       /**/
       err = zipOpenNewFileInZip3(zf,savefilenameinzip,&zi,
                        NULL,0,NULL,0,NULL /* comment*/,
                        (opt_compress_level != 0) ? Z_DEFLATED : 0,
                        opt_compress_level,0,
                        /* -MAX_WBITS, DEF_MEM_LEVEL, Z_DEFAULT_STRATEGY, */
                        -MAX_WBITS, DEF_MEM_LEVEL, Z_DEFAULT_STRATEGY,
                        password,crcFile);

       if (err != ZIP_OK)
           printf("error in opening %s in zipfile\n",filenameinzip);
       else
       {
           fin = fopen(filenameinzip,"rb");
           if (fin==NULL)
           {
               err=ZIP_ERRNO;
               printf("error in opening %s for reading\n",filenameinzip);
           }
       }

       if (err == ZIP_OK)
           do
           {
               err = ZIP_OK;
               size_read = (int)fread(buf,1,size_buf,fin);
               if (size_read < size_buf)
                   if (feof(fin)==0)
               {
                   printf("error in reading %s\n",filenameinzip);
                   err = ZIP_ERRNO;
               }

               if (size_read>0)
               {
                   err = zipWriteInFileInZip (zf,buf,size_read);
                   if (err<0)
                   {
                       printf("error in writing %s in the zipfile\n",
                                        filenameinzip);
                   }

               }
           } while ((err == ZIP_OK) && (size_read>0));

       if (fin)
           fclose(fin);

       if (err<0)
           err=ZIP_ERRNO;
       else
       {
           err = zipCloseFileInZip(zf);
           if (err!=ZIP_OK)
               printf("error in closing %s in the zipfile\n",
                           filenameinzip);
       }
       
       // Add the image jpg file
       
       strncpy(filenameinzip, _T(imagein),MAXFILENAME-1);
       /* strncpy doesnt append the trailing NULL, of the string is too long. */
       filenameinzip[ MAXFILENAME ] = '\0';

       zi.tmz_date.tm_sec = zi.tmz_date.tm_min = zi.tmz_date.tm_hour =
       zi.tmz_date.tm_mday = zi.tmz_date.tm_mon = zi.tmz_date.tm_year = 0;
       zi.dosDate = 0;
       zi.internal_fa = 0;
       zi.external_fa = 0;
       filetime(filenameinzip,&zi.tmz_date,&zi.dosDate);   

       /*the path name saved, should not include a leading slash. */
       /*if it did, windows/xp and dynazip couldn't read the zip file. */
       savefilenameinzip = filenameinzip;
       while( savefilenameinzip[0] == '\\' || savefilenameinzip[0] == '/' )
       {
           savefilenameinzip++;
       }

       /*should the zip file contain any path at all?*/
       if( opt_exclude_path )
       {
           const char *tmpptr;
           const char *lastslash = 0;
           for( tmpptr = savefilenameinzip; *tmpptr; tmpptr++)
           {
               if( *tmpptr == '\\' || *tmpptr == '/')
               {
                   lastslash = tmpptr;
               }
           }
           if( lastslash != NULL )
           {
               savefilenameinzip = lastslash+1; // base filename follows last slash.
           }
       }

       /**/
       err = zipOpenNewFileInZip3(zf,savefilenameinzip,&zi,
                        NULL,0,NULL,0,NULL /* comment*/,
                        (opt_compress_level != 0) ? Z_DEFLATED : 0,
                        opt_compress_level,0,
                        /* -MAX_WBITS, DEF_MEM_LEVEL, Z_DEFAULT_STRATEGY, */
                        -MAX_WBITS, DEF_MEM_LEVEL, Z_DEFAULT_STRATEGY,
                        password,crcFile);

       if (err != ZIP_OK)
           printf("error in opening %s in zipfile\n",filenameinzip);
       else
       {
           fin = fopen(filenameinzip,"rb");
           if (fin==NULL)
           {
               err=ZIP_ERRNO;
               printf("error in opening %s for reading\n",filenameinzip);
           }
       }

       if (err == ZIP_OK)
           do
           {
               err = ZIP_OK;
               size_read = (int)fread(buf,1,size_buf,fin);
               if (size_read < size_buf)
                   if (feof(fin)==0)
               {
                   printf("error in reading %s\n",filenameinzip);
                   err = ZIP_ERRNO;
               }

               if (size_read>0)
               {
                   err = zipWriteInFileInZip (zf,buf,size_read);
                   if (err<0)
                   {
                       printf("error in writing %s in the zipfile\n",
                                        filenameinzip);
                   }

               }
           } while ((err == ZIP_OK) && (size_read>0));

       if (fin)
           fclose(fin);

       if (err<0)
           err=ZIP_ERRNO;
       else
       {
           err = zipCloseFileInZip(zf);
           if (err!=ZIP_OK)
               printf("error in closing %s in the zipfile\n",
                           filenameinzip);
       }
       
   }       
   
   errclose = zipClose(zf,NULL);
        if (errclose != ZIP_OK)
            printf("error in closing %s\n",_T(fileout));   
   
   return err;     
} 

/**************************** WidePath **************************
  Determines which charts are close to to the path of check points
  Returns the lat/lons of the chart centre points
 * Syntax:	rc = WidePath(info, diag, diagmsgs)
 *
 * Params:	
 *			   info - directory of chart information (array of latlon check_points, array of chart lat/lon centre points)
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages   
 * Return: the 0 if OK else 1  and if OK then the chartbits in the ChartInfo
 *********************************************************************/
RexxRoutine3(int, WidePath, RexxObjectPtr, info, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512]; 
   int rc,sz, sz2;
   logical_t flag;
   double *ck_points;  // local  array  of lat/lon check points
   double *chart_cps;  // local  array  of lat/lon chart centre points
   RexxArrayObject ck_points_rexx;
   RexxArrayObject chart_cps_rexx;  
   RexxObjectPtr dataobj;
   RexxDirectoryObject info_dir;
   
   info_dir = (RexxDirectoryObject)info;
   flag = context->IsDirectory(info_dir);
   if ( flag == 0 )
   {
      sprintf(buffer,"ERROR: WidePath parameter 1 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }
   
   // get the max_range
   int max_range;
   dataobj = (RexxObjectPtr)context->DirectoryAt(info_dir, "MAX_RANGE");
   flag = context->ObjectToInt32(dataobj, &max_range);  
   
   // array to return the chart lat/lons
   RexxArrayObject lat_lons;
   lat_lons = (RexxArrayObject)context->DirectoryAt(info_dir, "LAT_LONS");  
   
   // Get pointer to the rexx check points array
   ck_points_rexx = (RexxArrayObject)context->DirectoryAt(info_dir, "CK_POINTS");  
   sz = context->ArrayItems(ck_points_rexx);  
   ck_points = (double *)malloc(sz * sizeof(double)); 
   sprintf(buffer,"DIAG: WidePath max_range=%d sz=%d",max_range,sz);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   
   // Transfer all the rexx check points into local array
   int n,m;
   double value;
   for(n=0 ; n<sz ; n++)
   { 
      dataobj = context->ArrayAt(ck_points_rexx, n+1);
      flag = context->ObjectToDouble(dataobj, &value);
      ck_points[n] = value;      
   }  
   
   // Get pointer to the rexx chart centre points array
   chart_cps_rexx = (RexxArrayObject)context->DirectoryAt(info_dir, "CHART_CPS");  
   sz2 = context->ArrayItems(chart_cps_rexx);  
   chart_cps = (double *)malloc(sz2 * sizeof(double));
   
   sprintf(buffer,"DIAG: WidePath sz2=%d",sz2);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   
   // Transfer all the rexx chart centre points into local array
   for(n=0 ; n<sz2 ; n++)
   { 
      dataobj = context->ArrayAt(chart_cps_rexx, n+1);
      flag = context->ObjectToDouble(dataobj, &value);
      chart_cps[n] = value;      
   } 
   
   // for each check point scan all the charts to see if with max_range
   double ck_lat, ck_lon, chart_lat, chart_lon, range;
   for(n=0 ; n<sz ; n += 2)
   {
      ck_lat = ck_points[n];  // Get check point lat/lon 
      ck_lon = ck_points[n+1]; 
         
      for(m=0 ; m<sz2 ; m += 2)
      {
         chart_lat = chart_cps[m];  // Get chart lat/lon 
         chart_lon = chart_cps[m+1]; 
         // if we have not selected this chart already
         if (chart_lat != 99.0)
         {
            // get the range to the chart centre
            range = DistGreatCircle( ck_lat, ck_lon, chart_lat, chart_lon );
            // if the range is lessthan or equal to the max_range save the lat/lon
            if (range <= max_range)
            {
               sprintf(buffer,"%f %f",chart_lat,chart_lon);
               rc = context->ArrayAppendString(lat_lons, buffer, strlen(buffer));
               // mark the chart lat/lon as selected
               chart_cps[m] = 99.0;
            }
         }
      }      
   }
   
   // free all the arrays we allocated
   free(ck_points);   
   free(chart_cps);
   
   return 0; 
}

/**************************** CalcCharts **************************
  Calculates the charts needed for a polygon
* Syntax:	rc = CalcCharts(chartinfo, diag, diagmsgs)
 *
 * Params:	
 *			   chartinfo - directory of chart information (boundaries, chart_width, chart_height)
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages   
 * Return: the 0 if OK else 1  and if OK then the chartbits in the ChartInfo
 *********************************************************************/
RexxRoutine3(int, CalcCharts, RexxObjectPtr, chartinfo, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512]; 
   int rc,sz_xys,sz_cps,n,value_int;   
   int32_t chart_width_sec,chart_height_sec;
   int32_t total_charts,bmp_width_sec,bmp_height_sec,type, scale;
   logical_t flag;
   
   int *xy_coordinates;  // local  array  added in V16.0.0.17
   int *chart_cp_xys;  // local  array  added in 27.0.0.11
   
   RexxArrayObject chart_cp_xys_rexx;  // added in V27.0.0.11
   RexxObjectPtr dataobj;
   RexxDirectoryObject chart_dir;
   
   chart_dir = (RexxDirectoryObject)chartinfo;
   flag = context->IsDirectory(chart_dir);
   if ( flag == 0 )
   {
      sprintf(buffer,"ERROR: CalcCharts parameter 1 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }
   
   //Get all necessary values that are not arrays
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "chart_height_sec");
   flag = context->ObjectToInt32(dataobj, &chart_height_sec);  
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "chart_width_sec");
   flag = context->ObjectToInt32(dataobj, &chart_width_sec);            
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "bmp_height_sec");
   flag = context->ObjectToInt32(dataobj, &bmp_height_sec);  
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "bmp_width_sec");
   flag = context->ObjectToInt32(dataobj, &bmp_width_sec);  
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "total_charts");
   flag = context->ObjectToInt32(dataobj, &total_charts);    
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "type");  // type 0 = path 1= polygon
   flag = context->ObjectToInt32(dataobj, &type);  
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "scale");         
   flag = context->ObjectToInt32(dataobj, &scale);  
   
   sprintf(buffer,"CalcCharts Total_charts=%d Type=%d BMP WIDTH/HEIGHT: %d %d ",total_charts,type,bmp_width_sec,bmp_height_sec);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));   
                                               
   // array to return the selected chart cps
   RexxArrayObject selected_cps;
   selected_cps = (RexxArrayObject)context->DirectoryAt(chart_dir, "selected_cps");        
   
   // Get pointer to the chart_cp_xys and allocate an array for it   added in V27.0.0.11
   chart_cp_xys_rexx = (RexxArrayObject)context->DirectoryAt(chart_dir, "chart_cp_xys");  
   sz_cps = context->ArrayItems(chart_cp_xys_rexx);  // the number of chart_cp_xys
   chart_cp_xys  = (int *)malloc(sz_cps * sizeof(int));
   
   // Transfer all the rexx chart_cp_xys into local array
   for(n=0 ; n<sz_cps ; n++)
   { 
      dataobj = context->ArrayAt(chart_cp_xys_rexx, n+1);
      flag = context->ObjectToInt32(dataobj, &value_int);
      chart_cp_xys[n] = value_int;      
   }  
   
   // Get pointer to the xy rexx xy_coordinates   and allocate an array for it
   RexxArrayObject xy_coordinates_rexx;  // added in V16.0.0.17
   xy_coordinates_rexx = (RexxArrayObject)context->DirectoryAt(chart_dir, "xy_coordinates");  
   sz_xys = context->ArrayItems(xy_coordinates_rexx);  // the number of coordinates
   xy_coordinates  = (int *)malloc(sz_xys * sizeof(int));
   
   // Transfer all the rexx xy_coordinates into local array
   for(n=0 ; n<sz_xys ; n++)
   { 
      dataobj = context->ArrayAt(xy_coordinates_rexx, n+1);
      flag = context->ObjectToInt32(dataobj, &value_int);
      xy_coordinates[n] = value_int;      
   }     
   
   // Check all the xy coordinates to see if they are valid   added in V28.0.1.15
   sprintf(buffer,"CalcCharts Total_xy_coordinates=%d",sz_xys/2);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   bool bad_coor = false;
   int yp,xp;
   for(n=0 ; n<sz_xys ; n += 2)
   {
      xp = xy_coordinates[n];  
      yp = xy_coordinates[n+1]; 
      if (xp < 0 || xp >= bmp_width_sec)  bad_coor = true;
      if (yp < 0 || yp >= bmp_height_sec) bad_coor = true;
      if (bad_coor) break; 
   }  
   if(bad_coor)
   {
      sprintf(buffer,"Error: the polygon coordinates are invalid");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;  
   }
      
   // Create a new image 
   CxImage *newImage = new CxImage();
   DWORD wBpp = 1; 
   newImage->Create(bmp_width_sec,bmp_height_sec,wBpp,1); 
   
   if (newImage == NULL)
   {
      sprintf(buffer,"Error: 1. the polygon is too large");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;  
   } 
   
   int numcolors;
   numcolors = newImage->GetNumColors();
   // printf("NUMCOLORS %d\n",numcolors); 
   
   int palette_size;
   palette_size = newImage->GetPaletteSize(); 
   // printf("palette SIZE %d\n",palette_size); 
   
   int idx;
   idx = newImage->GetPixelIndex(0,0);
   // printf("PIXEL 0,0  INDEX=%d\n",idx);
   
   RGBQUAD WhiteColor;
   WhiteColor.rgbRed = 255; 
   WhiteColor.rgbGreen = 255; 
   WhiteColor.rgbBlue = 255; 
   WhiteColor.rgbReserved = 0;    
   
   RGBQUAD BlackColor;
   BlackColor.rgbRed = 0; 
   BlackColor.rgbGreen = 0; 
   BlackColor.rgbBlue = 0; 
   BlackColor.rgbReserved = 0;     
   
   newImage->SetPaletteColor(0,WhiteColor);    // Set entire bmp to white
   newImage->SetPaletteColor(1,BlackColor);
   newImage->Clear(0);
   
   // RGBQUAD WhatColor;
   int x,y;    
   RGBQUAD Color;
   Color.rgbRed = 0; 
   Color.rgbGreen = 0; 
   Color.rgbBlue = 0; 
   Color.rgbReserved = 1; 
   x = 0; y=0;    
   
   //  draw the polygon
   int sel_rc;
   if (type == 1) // if polygon then do this
   {
      POINT* p = (POINT*)malloc(sz_xys/2*sizeof(POINT));  
      long i = 0;
      sprintf(buffer,"CalcCharts  Drawing Polygon Totals points=%d",sz_xys/2);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      
      for(n=0 ; n<sz_xys ; n += 2, i++)
      {
         p[i].x = xy_coordinates[n];  // added in V1.0.0.17
         p[i].y = xy_coordinates[n+1];  // added in V1.0.0.17         
         
         if ( diag>0 )
         {
            sprintf(buffer,"CalcCharts POINT %d xpos/ypos:%d %d",i,p[i].x,p[i].y);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         }
		}
      
      // Use all the points to  draw the polygon
      sel_rc = newImage->SelectionCreate();
      if (sel_rc == NULL)
      {
         sprintf(buffer,"Error: 2. the polygon is too large");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         return 1;  
      } 
      sel_rc = newImage->SelectionClear();
      if (sel_rc == NULL)
      {
         sprintf(buffer,"Error: 3. the polygon is too large");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         return 1;  
      } 
		sel_rc = newImage->SelectionAddPolygon(p,sz_xys/2); 
      if (sel_rc == NULL)
      {
         sprintf(buffer,"Error: 4. the polygon is too large");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         return 1;  
      } 
		
		free(p);
   }  
           
   // Check which charts should be used
   int sx,sy;
   x = 0; y = 0;
   sprintf(buffer,"CalcCharts chart_width_sec:%d chart_height_sec:%d\n",chart_width_sec, chart_height_sec);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));    
   
   // Get each chart cp and see if it is within the scan range added in V27.0.0.11
   int maxy,maxx,cps_num;
   for(cps_num=0;cps_num<sz_cps;cps_num+=2)   
   {
      bool select_chart = false;   // moved after the for in V27.0.0.12
      // Get the chart centre point  x y
      xp = chart_cp_xys[cps_num];  
      yp = chart_cp_xys[cps_num+1]; 
      
      if ( diag>0 )
      {
         sprintf(buffer,"DIAG CalcCharts checking chart CP XP=%d YP=%d",xp,yp);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }  
      
      // if its selected already forget it
      if(xp >= 0) // if not selected already 
      {
         // first see if the cp is within the polygon added in V28.0.1.2
         if (newImage->SelectionIsInside(xp,yp))
         { 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG CalcCharts CP at XP=%d YP=%d is inside polygon",xp,yp);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            }  
            select_chart = true;   // select this chart
         }
         else // else check if its within range
         {
            sy = yp-chart_height_sec;
            maxy = yp+chart_height_sec;
            if (sy<0) 
            {
               sy = 0;
            }
            
            // scan the expanded chart area to see if there is a bit on
            for(y=sy; y<maxy; y++) // changed in V16.0.0.17  
            { 
               if(y >= bmp_height_sec) break;
               sx = xp - chart_width_sec;
               maxx = xp+chart_width_sec;
               if(sx<0) 
               {
                  sx = 0;
               }
               
               for(x=sx; x<maxx; x++) // changed in V16.0.0.17 
               {
                  if( x >= bmp_width_sec) break;
                  
                  if (newImage->SelectionIsInside(x,y))
                  { 
                     if ( diag>0 )
                     {
                        sprintf(buffer,"DIAG CalcCharts FOUND inside at XP=%d YP=%d SY=%d SX=%d  X=%d Y=%d",xp,yp,sy,sx,y,x);
                        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                     }  
                     select_chart = true;   // select this chart
                     break;
                  }
               
               }  
               if (select_chart) break; 
            }         
         }
      
         // if the chart was selected pass its cps back
         if (select_chart)
         {
            sprintf(buffer,"%d %d",xp,yp);
            rc = context->ArrayAppendString(selected_cps, buffer, strlen(buffer));
            // mark the chart as selected
            chart_cp_xys[cps_num]  = -1; 
         } 
      } 
   }
       
    
   // Destroy the image if one exists
   if (newImage->IsValid()) 
   {
      rc = newImage->Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: CalcCharts 1. Invalid rc from Destroy Image RC=%d\n ",rc);
        printf(buffer);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }
   
      sprintf(buffer,"DIAG CalcCharts destroyed newimage rc=%d\n",rc);
      // printf(buffer);printf("\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
   }  
   
   // Destroy the selection if one exists
   if (newImage->SelectionIsValid()) 
   {
      rc = newImage->SelectionDelete();; // Destroy the selection
      if(!rc)
      {
        sprintf(buffer,"Error: CalcCharts 1. Invalid rc from SelectionDestroy Image RC=%d\n ",rc);
        printf(buffer);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }
      sprintf(buffer,"DIAG CalcCharts destroyed selection rc=%d\n",rc);
      // printf(buffer);printf("\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
   }     
   
   // free all the arrays we allocated
   free(xy_coordinates);   
   free(chart_cp_xys); 
   
   return 0;
}  
/**************************** Calc_HLG_Tiles **************************
  Calculates the tiles needed for an HLG
* Syntax:	rc = Calc_HLG_Tiles(chartinfo, diag, diagmsgs)
 *
 * Params:	
 *			   Dir - directory of data
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages   
 * Return: the 0 if OK else 1  and if OK then the chartbits in the Dir
 *********************************************************************/
RexxRoutine3(int, Calc_HLG_Tiles, RexxObjectPtr, Dir, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512]; 
   int rc;
   logical_t flag;
   
   RexxObjectPtr dataobj;
   
   RexxDirectoryObject dir_data = (RexxDirectoryObject)Dir;
   flag = context->IsDirectory(dir_data);
   if ( flag == 0 )
   {
      sprintf(buffer,"ERROR: Calc_HLG_Tiles parameter 1 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }
   
   //Get all necessary values that are not arrays
   int bmp_width,bmp_height;         
   dataobj = (RexxObjectPtr)context->DirectoryAt(dir_data, "bmp_width"); 
   flag = context->ObjectToInt32(dataobj, &bmp_width); 
   dataobj = (RexxObjectPtr)context->DirectoryAt(dir_data, "bmp_height");
   flag = context->ObjectToInt32(dataobj, &bmp_height); 
   
   // printf("DEBUG bmp_width=%d bmp_height=%d\n",bmp_width,bmp_height); 
   
   sprintf(buffer,"Calc_HLG_Tiles bmp width/height=%d %d",bmp_width,bmp_height);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));   
                                               
   // Rexx array to return the selected tiles
   RexxArrayObject selected_tiles = (RexxArrayObject)context->DirectoryAt(dir_data, "selected_tiles");        
   
   // Get pointer to the hlg points and allocate an array for it   
   RexxArrayObject hlg_xys_rexx = (RexxArrayObject)context->DirectoryAt(dir_data, "hlg_xys");  
   int num_hlgs = context->ArrayItems(hlg_xys_rexx);  // the number of hlgs
   int * hlg_xys  = (int *)malloc(num_hlgs * sizeof(int));
   
   // printf("DEBUG Transferringrexx hlg_xys into local array\n");
   
   // Transfer all the rexx hlg_xys into local array
   int n,value_int;
   for(n=0 ; n<num_hlgs ; n++)
   { 
      dataobj = context->ArrayAt(hlg_xys_rexx, n+1);
      flag = context->ObjectToInt32(dataobj, &value_int);
      hlg_xys[n] = value_int;      
   }  
   
   // printf("DEBUG creating image bmp_width=%d bmp_height=%d\n",bmp_width,bmp_height);
      
   // Create a new image the size of the gif
   CxImage *newImage = new CxImage();
   DWORD wBpp = 1; 
   newImage->Create(bmp_width,bmp_height,wBpp,1); 
   
   if (newImage == NULL)
   {
      sprintf(buffer,"Error: 5. the polygon is too large");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;  
   } 
   
    // printf("DEBUG creating array of all the HLG points\n");
   
   // create an array of all the HLG points
   POINT* p = (POINT*)malloc(num_hlgs/2*sizeof(POINT));  
   long i = 0;
   sprintf(buffer,"Calc_HLG_Tiles  Drawing Polygon Totals points=%d",num_hlgs/2);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   // Get all the hlg offsets
   for(n=0 ; n<num_hlgs ; n += 2, i++)
   {
      // Create the point from the hlg x/y
      p[i].x = hlg_xys[n];  // added in V1.0.0.17
      p[i].y = hlg_xys[n+1];  // added in V1.0.0.17         
      
      if ( diag>0 )
      {
         sprintf(buffer,"Calc_HLG_Tiles POINT %d xpos/ypos:%d %d",i,p[i].x,p[i].y);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      }
      // printf("DEBUG POINT %d xpos/ypos:%d %d\n",i,p[i].x,p[i].y);
	}
   
   // printf("DEBUG creating Create a new Selection\n");
   
   // Create a new Selection
   bool sel_rc = newImage->SelectionCreate();
   if (!sel_rc)
   {
      sprintf(buffer,"Error: 6. the polygon is too large");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;  
   } 
   // printf("DEBUG creatiClear the selection\n");
   // Clear the selection 
   sel_rc = newImage->SelectionClear();
   if (!sel_rc)
   {
      sprintf(buffer,"Error: 7. the polygon is too large");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;  
   } 
   // printf("DEBUG Make the polygon from all the HLG points num_hlgs/2=%d\n",num_hlgs/2);
   // Make the polygon from all the HLG points
	sel_rc = newImage->SelectionAddPolygon(p,num_hlgs/2); 
   if (!sel_rc)
   {
      sprintf(buffer,"Error: 8. the polygon is too large");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;  
   } 
   
   // printf("DEBUG 	free(\n");
	
	free(p);  
   
   // printf("DEBUG checking each bit within the bmp to see if its within the polygon\n");
   
   // check each bit within the bmp to see if its within the polygon
   int xp,yp;
   for(yp=0; yp<bmp_height; yp++) 
   {       
      for(xp=0; xp<bmp_width; xp++) 
      {
         if (newImage->SelectionIsInside(xp,yp))
         { 
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG Calc_HLG_Tiles FOUND inside at XP=%d YP=%d ",xp,yp);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            }  
            
            // put in the selected_tiles array
            sprintf(buffer,"%d %d",xp,yp);
            rc = context->ArrayAppendString(selected_tiles, buffer, strlen(buffer));
         }
      }
   }     
   
   // printf("DEBUG Destroy the image\n");
     
           
   // Destroy the image if one exists
   if (newImage->IsValid()) 
   {
      rc = newImage->Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: Calc_HLG_Tiles 1. Invalid rc from Destroy Image RC=%d\n ",rc);
        printf(buffer);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }
   
      sprintf(buffer,"DIAG Calc_HLG_Tiles destroyed newimage rc=%d\n",rc);
      // printf(buffer);printf("\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
   }  
   
   // Destroy the selection if one exists
   if (newImage->SelectionIsValid()) 
   {
      rc = newImage->SelectionDelete();; // Destroy the selection
      if(!rc)
      {
        sprintf(buffer,"Error: Calc_HLG_Tiles 1. Invalid rc from SelectionDestroy Image RC=%d\n ",rc);
        printf(buffer);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }
      sprintf(buffer,"DIAG Calc_HLG_Tiles destroyed selection rc=%d\n",rc);
      // printf(buffer);printf("\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
   }     
   
   // free all the arrays we allocated  
   free(hlg_xys); 
   
   return 0;
}  
/**************************** Add_HLG_Points **************************
  Adds points to an HLG so there enough to pickup all tiles
* Syntax:	rc = Add_HLG_Points(chartinfo, diag, diagmsgs)
 *
 * Params:	
 *			   Dir - directory of data
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages   
 * Return: the 0 if OK else 1  and if OK then the chartbits in the Dir
 *********************************************************************/
RexxRoutine3(int, Add_HLG_Points, RexxObjectPtr, Dir, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512]; 
   char taken_string[512]; 
   int rc;
   logical_t flag;
   
   // printf("DEBUG  Add_HLG_Points ENTERED \n");    
   
   RexxObjectPtr dataobj;
   
   RexxDirectoryObject dir_data = (RexxDirectoryObject)Dir;
   flag = context->IsDirectory(dir_data);
   if ( flag == 0 )
   {
      sprintf(buffer,"ERROR: Add_HLG_Points parameter 1 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }
   
   //Get all necessary values that are not arrays
   int zm,projection;        
   dataobj = (RexxObjectPtr)context->DirectoryAt(dir_data, "zm"); 
   flag = context->ObjectToInt32(dataobj, &zm); 
   dataobj = (RexxObjectPtr)context->DirectoryAt(dir_data, "projection");
   flag = context->ObjectToInt32(dataobj, &projection); 
   
   double wpdist; //  = .027 -- nm = 50 meters 
   dataobj = (RexxObjectPtr)context->DirectoryAt(dir_data, "wpdist");
   flag = context->ObjectToDouble(dataobj, &wpdist);  
   
   sprintf(buffer,"DIAG: Add_HLG_Points zm=%d projection=%d wpdist=%f",zm,projection,wpdist);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   // printf("DEBUG zm=%d projection=%d wpdist=%f\n",zm,projection,wpdist);                                            
                                               
   // Rexx array to return the points
   RexxArrayObject hlg_tile_x_ys = (RexxArrayObject)context->DirectoryAt(dir_data, "hlg_tile_x_ys"); 
   
   // The wp_lats
   RexxArrayObject wp_lats = (RexxArrayObject)context->DirectoryAt(dir_data, "wp_lats");  
   // The wp_lons
   RexxArrayObject wp_lons = (RexxArrayObject)context->DirectoryAt(dir_data, "wp_lons"); 
   
   int total_points = context->ArraySize(wp_lats);
   
   RexxDirectoryObject taken = (RexxDirectoryObject)context->DirectoryAt(dir_data, "taken"); 
    
   double wp_lat,wp_lon;
   int n, m, tilex,tiley;
   double tile_lat, tile_lon, next_lat, next_lon;
   double bearing, rangetonext,back_bearing;
   
   // printf("DEBUG total_points=%d\n",total_points);
        
   // add extra points
   for (int p=1; p<=total_points; p++)
   {
      // Get the lat/lon of the wp
      dataobj = context->ArrayAt(wp_lats, p);     
      flag = context->ObjectToDouble(dataobj, &wp_lat); 
      dataobj = context->ArrayAt(wp_lons, p);     
      flag = context->ObjectToDouble(dataobj, &wp_lon); 
      
      // get the tilex/y and tilex/y lat/lon of this point
      L_Get_Tile_XY(wp_lat,wp_lon,zm,projection,tilex,tiley,tile_lat,tile_lon);
      
      // printf("DEBUG 1. wp#%d lat/lon=%f %f tile/x/y=%d %d tile_lat/lon=%f %f\n",p,wp_lat,wp_lon,tilex,tiley,tile_lat,tile_lon);
      
      // if  tilex/y has not been used 
      RexxObjectPtr obj;
      // if  tilex/y has not been used add it to the downloads array
      sprintf(taken_string,"%d %d",tilex,tiley);
      // printf("DEGUG wp#%d lat/lon=%f %f tile/x/y=%d %d tile_lat/lon=%f %f taken_string=%s\n",p,wp_lat,wp_lon,tilex,tiley,tile_lat,tile_lon,taken_string);
      obj = context->DirectoryAt(taken, taken_string);
      if(obj ==  NULLOBJECT)
      { 
         sprintf(buffer,"%d %d %f %f",tilex,tiley,tile_lat,tile_lon);
         rc = context->ArrayAppendString(hlg_tile_x_ys, buffer, strlen(buffer));
         // add the string to the directory with the taken_string as the index
         dataobj = context->Int32ToObject(1);
         context->DirectoryPut(taken, dataobj, taken_string);
         
         // printf("DEBUG wp#%d PUT WP tile/x/y=%d %d tile_lat/lon=%f %f\n",p,tilex,tiley,tile_lat,tile_lon); 
         
         // Mark it as taken
         dataobj = context->Int32ToObject(1);
         context->DirectoryPut(taken,dataobj,taken_string);     
      } 
      
      //  test these values
      int test_tilex = tilex;
      int test_tiley = tiley;
      
      //  make this wp the ck lat/lon to find next point
      double ck_lat = wp_lat;
      double ck_lon = wp_lon; 
      
      // Get the next point
      n = p + 1;
      // if its past the end break out
      if(n>total_points) break;
      
      // Get the lat/lon of the next
      dataobj = context->ArrayAt(wp_lats, n);     
      flag = context->ObjectToDouble(dataobj, &next_lat); 
      dataobj = context->ArrayAt(wp_lons, n);     
      flag = context->ObjectToDouble(dataobj, &next_lon); 
      
      // printf("DEBUG 2. next wp=#%d lat/lon=%f %f\n",n,next_lat,next_lon);
      
      // scan to the next wp and add any new tilex/y s found
      for (m=1; p<10000; m++)
      {
         // get the bearing from the current ck to the next    
         DistanceBearingMercator(next_lat,next_lon,ck_lat,ck_lon, &bearing, &rangetonext);
         // rangetonext = rangetonext/1852; // change to nautical miles
         
         // printf("DEBUG bearing to next=%f rangetonext=%f wpdist=%f\n",bearing,rangetonext,wpdist);
         
         // If the range to the next ck point is less than the wpdist break out
         if (rangetonext < wpdist) break;
      
         // calculate where the next ck should be
         ll_gc_ll( ck_lat,ck_lon,bearing,wpdist,&ck_lat, &ck_lon );
         
         // printf("DEBUG 3. lp=%d bearing=%f rangetonext=%f next ck_lat/lon=%f %f\n",m,bearing,rangetonext,ck_lat,ck_lon);
         
         // get the tilex/y and tilex/y lat/lon of this knew ck point
         L_Get_Tile_XY(ck_lat,ck_lon,zm,projection,tilex,tiley,tile_lat,tile_lon);
         
         // if we are in another tile add it 
         if (test_tilex != tilex || test_tiley != tiley)
         {
            // if  handle has not been used add it to the downloads array
            sprintf(taken_string,"%d %d",tilex,tiley);
            obj = context->DirectoryAt(taken, taken_string);
            if(obj ==  NULLOBJECT)
            { 
               // Add it to the output
               sprintf(buffer,"%d %d %f %f",tilex,tiley,tile_lat,tile_lon);
               rc = context->ArrayAppendString(hlg_tile_x_ys, buffer, strlen(buffer));
               
               // printf("DEBUG lp%d PUT EXTRA CK tile/x/y=%d %d tile_lat/lon=%f %f taken_string=%s\n",m,tilex,tiley,tile_lat,tile_lon,taken_string); 
               
               // update the test tilex/y
               test_tilex = tilex;
               test_tiley = tiley;
              
               // Mark it as taken
               dataobj = context->Int32ToObject(1);
               context->DirectoryPut(taken,dataobj, taken_string); 
            }         
         }
      }
      //  calculate the back_bearing from the last ckpt back to the last waypoint
      DistanceBearingMercator(ck_lat,ck_lon,wp_lat,wp_lon, &back_bearing, &rangetonext);
   }    
   
   return 0;
}  
/**************************** Calc_Bdr ***********************************
 Calculates the border points of a polygon
* Syntax:	rc = Calc_Bdr(chartinfo, diag, diagmsgs)
 *
 * Params:	
 *			   Dir - directory of data
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages   
 * Return: the 0 if OK else 1  and if OK then the chartbits in the Dir
 *********************************************************************/
RexxRoutine3(int, Calc_Bdr, RexxObjectPtr, Dir, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512]; 
   int rc;
   logical_t flag;
   
   RexxObjectPtr dataobj;
   
   RexxDirectoryObject dir_data = (RexxDirectoryObject)Dir;
   flag = context->IsDirectory(dir_data);
   if ( flag == 0 )
   {
      sprintf(buffer,"ERROR: Calc_Bdr parameter 1 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }
   
   //Get all necessary values that are not arrays
   int bmp_width,bmp_height,bmp_ul_tilex,bmp_ul_tiley;         
   dataobj = (RexxObjectPtr)context->DirectoryAt(dir_data, "bmp_width"); 
   flag = context->ObjectToInt32(dataobj, &bmp_width); 
   dataobj = (RexxObjectPtr)context->DirectoryAt(dir_data, "bmp_height");
   flag = context->ObjectToInt32(dataobj, &bmp_height);  
   dataobj = (RexxObjectPtr)context->DirectoryAt(dir_data, "bmp_ul_tilex");
   flag = context->ObjectToInt32(dataobj, &bmp_ul_tilex);  
   dataobj = (RexxObjectPtr)context->DirectoryAt(dir_data, "bmp_ul_tiley");
   flag = context->ObjectToInt32(dataobj, &bmp_ul_tiley);   
                                               
   // Rexx array to return the border points
   RexxArrayObject bdrs = (RexxArrayObject)context->DirectoryAt(dir_data, "bdrs");  
   
   // Create an array the size of the bmp_width * bmp_height 
   int array_size = bmp_width * bmp_height; 
   char * tiles_all = (char *)malloc(array_size);
   memset(tiles_all,0,array_size); 
   
   /* KEEP
   char **tiles_all;
   tiles_all=(char **) malloc(bmp_height*sizeof(char *));
   for(int i=0;i<10;i++)
   {
       tiles_all[i]=(char *) malloc(bmp_width);
       memset(tiles_all[i],0,bmp_width); 
   }
   printf("DEBUG 1 array_size=%d tiles_all[0][49]=%d\n",array_size,tiles_all[0][49]);
   */
   
   // get all the tile_xys~items  and mark the tiles array entries with their tilex/y
   RexxArrayObject tile_xys_rexx = (RexxArrayObject)context->DirectoryAt(dir_data, "tile_xys");  
   int sz = context->ArraySize(tile_xys_rexx);
   
   sprintf(buffer,"DIAG Calc_Bdr bmp width/height=%d %d UL=%d %d array_size=%d sz=%d",bmp_width,bmp_height,bmp_ul_tilex,bmp_ul_tiley,array_size,sz);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   // printf("DEBUG bmp width/height=%d %d UL=%d %d array_size=%d sz=%d\n",bmp_width,bmp_height,bmp_ul_tilex,bmp_ul_tiley,array_size,sz);
   
   int n,tilex,tiley,x_ofs,y_ofs;
   for (n=1; n<=sz; n+=2)
   {
      dataobj = context->ArrayAt(tile_xys_rexx, n);     // get the tilex
      flag = context->ObjectToInt32(dataobj, &tilex); 
      dataobj = context->ArrayAt(tile_xys_rexx, n+1);   // get the tiley
      flag = context->ObjectToInt32(dataobj, &tiley); 
      // calulate the offsets 
      x_ofs = (tilex - bmp_ul_tilex);
      y_ofs = (tiley - bmp_ul_tiley); 
      // printf("DEBUG n=%d tilex/y=%d %d x/y_ofs=%d %d \n",n,tilex,tiley,x_ofs,y_ofs);
      tiles_all[x_ofs + (y_ofs * bmp_width)] = true;  // mark the tiles_all array entry as used
   }  
    
   // Find the first entry at top scanning from top to bottom  left to right 
   int Bdr_pos = 1;
   int xp,yp,first_x,first_y;
   bool fnd = false;
   for(yp=0; yp<bmp_height; yp++)
   {
      // printf("DEBUG yp=%d\n",yp);
      for(xp=0; xp<bmp_width;xp++)
      {
         // printf("DEBUG yp=%d xp=%d tiles_all[yp][xp]=%d\n",yp,xp,tiles_all[yp][xp]);
         // if it has data use it
         if (tiles_all[xp + (yp * bmp_width)])
         {
            // put in the brds array  a ply border point
            sprintf(buffer,"%d %d %d",xp,yp,Bdr_pos);
            rc = context->ArrayAppendString(bdrs, buffer, strlen(buffer));
            fnd = true;
            // Save the first xy
            first_x = xp;
            first_y = yp;
            sprintf(buffer,"DIAG Calc_Bdr first_x=%d first_y=%d",first_x,first_y);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
            // printf("DEBUG first_x=%d first_y=%d",first_x,first_y); 
            break;
         }
      }
      if(fnd) break;
   }  
   
   // set to go down
   int x_incr = 0;
   int y_incr = 1;
   int Rule = 1;  // look for change in the tri_1 or curr
   
   // Set the prev_trival                    
   byte prev_trival = Trival(xp,yp,x_incr,y_incr,bmp_width,bmp_height, tiles_all); // get the intial trival
   
   sprintf(buffer,"DIAG Calc_Bdr prev_trival=%x",prev_trival);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   yp += 1;  //  go down 1
   
   // Use the bloodhound algorithm to find all the border points
   int lp = 0;                                           
   bool success = false; // flag to indicate success
   int therc = 1; // Bad rc
   int tri_1 = 0;
   int tri_2 = 0;
   int curr = 0;
   int prev_tri_1 = 0;
   int prev_tri_2 = 0;
   int prev_curr = 0;
   byte trival = 0;
   byte bit80 = 0x80;  
   byte bit40 = 0x40;  
   byte bit20 = 0x20;   
   bool create_bdr = false;
   bool error = false; // indictes and error ocurred
   for(lp = 1; lp < 100000; lp++)
   {
      // Get the trival of the next tile at xp,yp
      trival = Trival(xp,yp,x_incr,y_incr,bmp_width,bmp_height, tiles_all); // get the intial trival
      // Set the tri_1 curr and tri_2
      if ((trival & bit80) == bit80) tri_1 = 1;else tri_1 = 0;
      if ((trival & bit40) == bit40) curr = 1;else curr = 0;
      if ((trival & bit20) == bit20) tri_2 = 1;else tri_2 = 0; 
      sprintf(buffer,"DIAG Calc_Bdr lp=%d xp/yp=%d %d Rule=%d trival=%x tri_1=%d curr=%d tri_2=%d",lp,xp,yp,Rule,trival,tri_1,curr,tri_2);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      // printf("DEBUG lp=%d xp/yp=%d %d Rule=%d trival=%x tri_1=%d curr=%d tri_2=%d\n",lp,xp,yp,Rule,trival,tri_1,curr,tri_2); 
      
      // Set the prev_tri_1 prev_curr and prev_tri_2
      if ((prev_trival & bit80) == bit80) prev_tri_1 = 1;else prev_tri_1 = 0;
      if ((prev_trival & bit40) == bit40) prev_curr = 1;else prev_curr = 0;
      if ((prev_trival & bit20) == bit20) prev_tri_2 = 1;else prev_tri_2 = 0; 
      sprintf(buffer,"DIAG Calc_Bdr lp=%d xp/yp=%d %d Rule=%d prev_trival=%x prev_tri_1=%d prev_curr=%d prev_tri_2=%d",lp,xp,yp,Rule, prev_trival,prev_tri_1,prev_curr,prev_tri_2);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      
      // Set the direction 
      int direction = 0;
      if(x_incr == 1) direction = 1; // Going right
      else if(x_incr == -1) direction = 2; // Going left
      if(y_incr == 1) direction = 3;  // Going down
      else if(y_incr == -1) direction = 4; // Going up
      sprintf(buffer,"DIAG Calc_Bdr lp=%d xp/yp=%d %d Rule=%d x_incr=%d y_incr=%d direction=%d",lp,xp,yp,Rule,x_incr,y_incr,direction);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      
      create_bdr = false;
   
      switch(Rule)  // TriGs
      {
         case 1 :  // Rule 1 Moving Down and Right
            switch(direction)
            {
               case 3 :  // -- Rule 1 Going Down
                  if(prev_tri_1 != tri_1)  create_bdr = true;// tri_1 mismatch
                  else if(xp==0 && !tiles_all[xp+1+(yp*bmp_width)]) create_bdr = true;// at the left wall and nothing to the right
                  else if(xp==bmp_width-1 && !tiles_all[xp-1+(yp*bmp_width)]) create_bdr = true;// at the right wall and noting to the left
                  else if(!tiles_all[xp-1+(yp*bmp_width)] && !tiles_all[xp+1+(yp*bmp_width)]) create_bdr = true;// if nothing to the left or right
                  else if(yp==bmp_height-1) // if  at the bottom nothing below
                  {
                     Rule = 2;  // Change to rule 2
                     Bdr_pos = 2;
                     create_bdr = true;
                  }
                  else if(yp < bmp_height-1 && !tiles_all[xp+((yp+1)*bmp_width)] ) // if  nothing below
                  {
                     Rule = 2;  // Change to rule 2
                     Bdr_pos = 2;
                     create_bdr = true;
                  }
             
                  break; // breal going down
                  
               case 2 :  // -- Rule 1 Going Left 
               
                  if(tri_1 == 1) //if there is a tile above 
                  {
                     Rule = 4;   //  Change to Rule 4
                     create_bdr = true;
                  } 
                  else if(xp==0) create_bdr = true; // if  at left wall nothing to the left
                  else if(!tiles_all[xp-1+(yp*bmp_width)]) create_bdr = true; // if  nothing to the left
                  
                  break;  // break going left
                     
               default:
                  
                  break;
            }
            
            break; // Break Rule 1  
            
         case 2 :  // Rule 2 Down or Right
           
            switch(direction)
            {
               case 3 :  // -- Rule 2 Going Down
               
                  if(tri_1 == 1) create_bdr = true; // if there is a tile to the right 
                  else if(xp==0 && !tiles_all[xp+1+(yp*bmp_width)]) create_bdr = true; // if at left wall and nothing to the right
                  else if(xp==bmp_width-1 && !tiles_all[xp-1+(yp*bmp_width)]) create_bdr = true; // if at right wall and nothing to the left
                  else if(!tiles_all[xp-1+(yp*bmp_width)] && !tiles_all[xp+1+(yp*bmp_width)]) create_bdr = true; // if nothing to the left or right
                  else if(yp == bmp_height-1) create_bdr = true; // if  at the bottom nothing below
                  else if(!tiles_all[xp+((yp+1)*bmp_width)]) create_bdr = true; // if  nothing below
                  break; // break going down
                  
              case 1 :  // -- Rule 2 Going Right   
              
                  if(prev_tri_2 != tri_2)  create_bdr = true;// tri_2 mismatch
                  else if( xp==bmp_width-1 ) // if  at the right wall nothing to the right
                  {
                     Rule = 3;  // Change to rule 3
                     Bdr_pos = 3;
                     create_bdr = true;
                  } 
                  else if(!tiles_all[xp+1+(yp*bmp_width)]) // if  nothing to the right
                  {
                     Rule = 3;  // Change to rule 3
                     Bdr_pos = 3;
                     create_bdr = true;
                  } 
                  break;  // break going right
               
               default:
                  
                  break;
            }
            
            break; // Break Rule 2
            
         case 3 :  // Rule 3 Going Up or Right
           
            switch(direction)
            {
               case 4 :  // Rule 3 Going Up
                  if(prev_tri_2 != tri_2) create_bdr = true; // tri_2 mismatch
                  else if(yp==0) create_bdr = true; // at the top nothing above
                  else if(!tiles_all[xp+((yp-1)*bmp_width)] ) create_bdr = true; // if nothing above
                  break; // break going up
                  
               case 1 :  // Rule 3 Going Right 
                  if(prev_tri_1 != tri_1) create_bdr = true; // tri_1 mismatch
                  else if(tri_2 == 1) //  if something below
                  {
                     Rule = 2;
                     Bdr_pos = 2;
                     create_bdr = true;
                  }
                  else if(xp==bmp_width-1) create_bdr = true; // at right wall nothing to the right 
                  else if(!tiles_all[xp+1+(yp*bmp_width)]) create_bdr = true; // if  nothing to the right 
                  
                  break; // break going right
                  
               default:
                  
                  break;
            }
            
         break;  // Break Rule 3
         
         case 4 :  // Rule 4 Going Up or Left
           
            switch(direction)
            {
               case 4 : // Going Up 
                  if(tri_2 == 1)  // if tile to the right
                  {
                     Rule = 3;  //  change to Rule 3
                     Bdr_pos = 3;
                     y_incr = -1;  // Go up
                     x_incr = 0;
                     create_bdr = true;
                  }
                  else if(yp==0) create_bdr = true; //if at the top nothing above
                  else if(!tiles_all[xp+((yp-1)*bmp_width)]) create_bdr = true; //if  nothing above
                  
                  break;  // break going up
                  
               case 2 : // Going Left
                  if(prev_tri_1 != tri_1) create_bdr = true; // if  tri_1 mismatch
                  else if(xp==0)   // if at the left wall nothing to the left 
                  {
                     Rule = 1;  //  change to Rule 
                     y_incr = 0; 
                     x_incr = -1; // Go Left
                     create_bdr = true;
                  }
                  else if(!tiles_all[xp-1+(yp*bmp_width)])   // if nothing to the left 
                  {
                     Rule = 1;  //  change to Rule 
                     y_incr = 0; 
                     x_incr = -1; // Go Left
                     create_bdr = true;
                  }
                  
                  break; // break going left   
               
               default: 
               
                  break;
            }
            
         break;  // Break Rule 4   
      }
      
      // ReSet the direction 
      direction = 0;
      if(x_incr == 1) direction = 1; // Going right
      else if(x_incr == -1) direction = 2; // Going left
      if(y_incr == 1) direction = 3;  // Going down
      else if(y_incr == -1) direction = 4; // Going up
      sprintf(buffer,"DIAG Calc_Bdr lp=%d create_bdr=%d xp/yp=%d %d Rule=%d x_incr=%d y_incr=%d direction=%d",lp,create_bdr,xp,yp,Rule,x_incr,y_incr,direction);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      
      // if create_bdr go create a new bdrs and determine where to go
      if (create_bdr)
      {
         // Check that the xp,yp are valid 
         if(!tiles_all[xp+(yp*bmp_width)])
         {
            sprintf(buffer,"Error: Calc_Bdr lp=%d xp/yp=%d %d is not valid!",lp,xp,yp);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            error = true;
            break;
         }
         
         // Check if this a special case
         bool special = false;
         // there is no where to go  
         if (tri_1 == 0 && tri_2 == 0)
         { 
            special = true;
            // depending on which direction we are going
            switch(direction)
            {
            
               case 3 : // Going Down
                  // put in the brds array  a ply border point
                  sprintf(buffer,"%d %d %d",xp,yp,2);  // LL
                  rc = context->ArrayAppendString(bdrs, buffer, strlen(buffer));
                  sprintf(buffer,"%d %d %d",xp,yp,3);  // LR
                  rc = context->ArrayAppendString(bdrs, buffer, strlen(buffer));
                  break;
                  
               case 4 : // Going Up
                  // put in the brds array  a ply border point
                  sprintf(buffer,"%d %d %d",xp,yp,4);  // UR
                  rc = context->ArrayAppendString(bdrs, buffer, strlen(buffer));
                  sprintf(buffer,"%d %d %d",xp,yp,1);  // UL
                  rc = context->ArrayAppendString(bdrs, buffer, strlen(buffer));
                  break;   
                  
               case 1 : // Going Right
                  // put in the brds array  a ply border point
                  sprintf(buffer,"%d %d %d",xp,yp,3);  // LR
                  rc = context->ArrayAppendString(bdrs, buffer, strlen(buffer));
                  sprintf(buffer,"%d %d %d",xp,yp,4);  // UR
                  rc = context->ArrayAppendString(bdrs, buffer, strlen(buffer));
                  break; 
                  
               case 2 : // Going Left
                  // put in the brds array  a ply border point
                  sprintf(buffer,"%d %d %d",xp,yp,1);  // UL
                  rc = context->ArrayAppendString(bdrs, buffer, strlen(buffer));
                  sprintf(buffer,"%d %d %d",xp,yp,2);  // LL
                  rc = context->ArrayAppendString(bdrs, buffer, strlen(buffer));
                  break; 
                      
               default: 
                  break;   
            } 
         } 
           
         // if not special create a border
         if(!special)
         {
            // put in the brds array  a ply border point
            sprintf(buffer,"%d %d %d",xp,yp,Bdr_pos); 
            rc = context->ArrayAppendString(bdrs, buffer, strlen(buffer));
         }
         
         Bdr_pos = 0;  // Reset the border position
         
         //  Calculate what direction to move next depending on Rule and the direction we are moving now
         
         switch(Rule)   // ACtions
         {
            case 1 :  // Rule 1 Going Down or Left
               switch(direction)
               {
                  case 3 :  // -- Rule 1 Going Down  
                     
                     // printf("DEBUG Rule=%d direction=%d\n",Rule,direction);
                  
                     if(xp==0 && !tiles_all[xp+1+(yp*bmp_width)]) // if at the left wall and nothing to the right
                     {
                        Rule = 3;      // Change to rule 3
                        y_incr = -1;   // Go up 
                        x_incr = 0; 
                     }
                     else if(xp==bmp_width-1 && !tiles_all[xp-11+(yp*bmp_width)]) // if at the right wall and nothing to the left
                     {
                        Rule = 3;      // Change to rule 3
                        y_incr = -1;   // Go up 
                        x_incr = 0; 
                     }
                     else if(!tiles_all[xp-1+(yp*bmp_width)] && !tiles_all[xp+1+(yp*bmp_width)]) // if nothing to the left or right
                     {
                        Rule = 3;      // Change to rule 3
                        y_incr = -1;   // Go up 
                        x_incr = 0; 
                     }
                     else // Default
                     {
                        y_incr = 0;   
                        x_incr = -1; // Go left  
                     }
                     break;  // Break going down
                     
                  case 2 :  // -- Rule 1 Going Left 
                  
                     if(yp==bmp_height-1) // if at the bottom nothing below
                     {
                        Rule = 2;   // change to rule 2
                        y_incr = 0;   
                        x_incr = 1; // Go right  
                     }
                     else if(!tiles_all[xp+((yp+1)*bmp_width)]) // if nothing below
                     {
                        Rule = 2;   // change to rule 2
                        y_incr = 0;   
                        x_incr = 1; // Go right  
                     }
                     else // Default
                     {
                        y_incr = 1; // Go down  
                        x_incr = 0; 
                     }
                     break; // Break Going Left
                        
                  default: 
                     break;
               }
               
               break; // Break Rule 1 
               
            case 2 :  // Rule 2 Going Down or Right
        
               switch(direction)
               {
                  case 3 :  // -- Rule 2 Going Down
                     if(yp==bmp_height-1 && tiles_all[xp+1+(yp*bmp_width)]) // if at the bottom and tile to the right
                     {
                        y_incr = 0;   
                        x_incr = 1; // Go right  
                     }
                     else if(!tiles_all[xp+((yp+1)*bmp_width)] &&  tiles_all[xp+1+(yp*bmp_width)]) // if nothing below and tile to the right
                     {
                        y_incr = 0;   
                        x_incr = 1; // Go right  
                     }
                     else if(yp==bmp_height-1) // if  at the bottom nothing below
                     {
                        Rule = 3;  // Change to rule 3
                        y_incr = -1; // Go up  
                        x_incr = 0; 
                     } 
                     else if(!tiles_all[xp+((yp+1)*bmp_width)] ) // if  nothing below
                     {
                        Rule = 3;  // Change to rule 3
                        y_incr = -1; // Go up  
                        x_incr = 0; 
                     } 
                     else if(tiles_all[xp-1+(yp*bmp_width)]) // if tile to the left
                     {
                        Rule = 1;  // Change to rule 1
                        Bdr_pos = 2;  // But put the border pos on the LL
                        y_incr = 0; 
                        x_incr = -1; // Go left
                     }
                     else if(!tiles_all[xp-1+(yp*bmp_width)] && tiles_all[xp+1+(yp*bmp_width)]) // if Nothing to right or left 
                     {
                        Rule = 3;     // Change to rule 3
                        y_incr = -1;  // Go up 
                        x_incr = 0; 
                     }
                     else  // Default
                     {
                        y_incr = 0;    
                        x_incr = 1;  // go right 
                     }
                      
                     break; // Break going down
                     
                  case 1 :  // Rule 2 Going right   
                     y_incr = 1; // Go down    
                     x_incr = 0; 
                     
                     break; // Break going right
                  
                  default: 
                     break;
               }
            
                break;  // Break Rule 2 
                
            case 3 :  // Rule 3 Going Up or Right
        
               switch(direction)
               {
                  case 4 :  // -- Rule 3 Going Up
                     if(yp==0 && xp == bmp_width-1) // hit upper right corner
                     {
                        Rule = 1;  // Change to Rule 1
                        y_incr = 0; 
                        x_incr = -1; // go left 
                     }
                     else if(tri_2 == 0 && (yp > 1 && !tiles_all[xp+((yp-1)*bmp_width)])) // if nothing to right or above
                     {
                        Rule = 4;  // Change to Rule 4
                        y_incr = 0; 
                        x_incr = -1; // go right 
                     }
                     else if(yp > 0 && !tiles_all[xp+((yp-1)*bmp_width)] && xp==bmp_width-1)  // if  at the right side and nothing above
                     {
                        Rule = 4; // Change to Rule 4
                        x_incr = -1;  // go left
                        y_incr = 0; 
                     }
                     else   // else Default
                     { 
                        y_incr = 0;    
                        x_incr = 1;  // go right  
                     }
                     
                     break; // Break going up
                     
                  case 1 :  // -- Rule 3 Going right
                     if(yp==bmp_height-1 && xp == bmp_width-1) // Going right on bottom and hitting lower right corner
                     {
                        Rule = 4;    // Change to Rule 4
                        y_incr = -1; // go up  
                        x_incr = 0; 
                     }
                     else if(yp==0 && !tiles_all[xp+((yp+1)*bmp_width)] ) // at the top and nothing below
                     {
                        Rule = 1;   // change to rule 1
                        y_incr = 0;    
                        x_incr = -1; //  Go left    
                     }
                     else if(!tiles_all[xp+((yp-1)*bmp_width)] && !tiles_all[xp+((yp+1)*bmp_width)] ) // if  Nothing above or below
                     {
                        Rule = 4;     // change to 4
                        y_incr = 0;  
                        x_incr = -1; // go right
                     }
                     else if(!tiles_all[xp+((yp-1)*bmp_width)]) // if  Nothing above
                     {
                        Rule = 4;    // Change to Rule 4
                        y_incr = 0; 
                        x_incr = -1; // go left 
                     }
                     else // Default
                     {
                        y_incr = -1;  // go up
                        x_incr = 0;
                     }
                     
                     break; // Break going right
               }
               
               break; // Break Rule 3
               
            case 4 :  // Rule 4 Going Up or Left
        
               switch(direction)
               {
                  case 4 :  // -- Rule 4 Going Up
                     if(yp==0 && xp == bmp_width-1) // hit upper right corner
                     {
                        Rule = 2;     // Change to Rule 2
                        y_incr = 0;    
                        x_incr = -1;  // Go left
                     }
                     else if(tri_2 == 1) // if there is tile to the right
                     {
                        Rule = 3;     //  Change to Rule 3
                        Bdr_pos = 4;  // Put the border in the UR
                        y_incr = 0;   
                        x_incr = 1;   // Go right
                     }
                     else if(tri_1 == 0 && yp > 1 && !tiles_all[xp+((yp-1)*bmp_width)] ) // if nothing to left or above
                     {
                        Rule = 1;    // Change to Rule 
                        y_incr = 1;  // go down 
                        x_incr = 0;
                     }
                     else // Default
                     {
                        y_incr = 0;    
                        x_incr = -1;    // go left 
                     }
                     
                     break; // Rule 4 Going up
                     
                  case 2 :  // -- Rule 4 Going left    
                     y_incr = -1;  // go up
                     x_incr = 0;
                     
                     break; // Rule 4 Going left
               }
               
               break; // Break Rule 4
               
         }  // End switch Rule 
         
         if(Bdr_pos == 0) Bdr_pos = Rule; // If we did not set the border posion make it the Rule
         
         prev_trival = Trival(xp,yp,x_incr,y_incr,bmp_width,bmp_height, tiles_all); // Update the prev_trival
      
      }
      else  prev_trival = trival; // else not creating a border save the previous trival 
      
      if(error) break; // break if an error occurred quit
      
      // increment the xp/yp 
      xp += x_incr;
      yp += y_incr;
     
      // check that the xp,yp do not go outside the tile_all array
      if(xp < 0) xp = 0;
      if(xp > bmp_width-1) xp = bmp_width-1;
      if(yp < 0) yp = 0;
      if(yp > bmp_height-1) yp = bmp_height-1;
      
      // see if we reached the start
      if (xp == first_x && yp == first_y)
      {
         success = true;
         sprintf(buffer,"DIAG Calc_Bdr lp=%d Success xp/yp=%d %d",lp,xp,yp);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         therc = 0;
         break;
      }
      
      // if(lp >= 550) break;  // DEBUG   
      
   } // end of for lp
   
   
   // free the array we allocated 
   /* KEEP
   for(int i=0;i<10;i++)
       free(tiles_all[i]);
   */    
   free(tiles_all); 
   
   return therc;
}  
/**************************** CalcWidePath **************************
  Calculates the charts needed for a wide path  added in V29.0.3.10
* Syntax:	rc = CalcWidePath(chartinfo, diag, diagmsgs)
 *
 * Params:	
 *			   chartinfo - directory of chart information (boundaries, chart_width, chart_height)
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages   
 * Return: the 0 if OK else 1  and if OK then the chartbits in the ChartInfo
 *********************************************************************/
RexxRoutine3(int, CalcWidePath, RexxObjectPtr, chartinfo, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[1000]; 
   int rc,n;
   logical_t flag;
   
   // int *xy_coordinates;  // local  array  
   
   RexxObjectPtr dataobj;
   RexxDirectoryObject chart_dir;
   
   chart_dir = (RexxDirectoryObject)chartinfo;
   flag = context->IsDirectory(chart_dir);
   if ( flag == 0 )
   {
      sprintf(buffer,"ERROR: CalcWidePath parameter 1 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }
                                
   //Get all necessary values that are not arrays
   double max_ck_dist,max_range,chart_width,chart_height,nl_sec,wl_sec,fract_width_sec,fract_height_sec;
   int num_charts_in_width,num_charts_in_height;
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "max_ck_dist");
   flag = context->ObjectToDouble(dataobj, &max_ck_dist);  
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "max_range");
   flag = context->ObjectToDouble(dataobj, &max_range); 
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "chart_width");
   flag = context->ObjectToDouble(dataobj, &chart_width); 
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "chart_height");
   flag = context->ObjectToDouble(dataobj, &chart_height); 
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "nl_sec");
   flag = context->ObjectToDouble(dataobj, &nl_sec); 
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "wl_sec");
   flag = context->ObjectToDouble(dataobj, &wl_sec); 
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "fract_width_sec");
   flag = context->ObjectToDouble(dataobj, &fract_width_sec); 
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "fract_height_sec");
   flag = context->ObjectToDouble(dataobj, &fract_height_sec); 
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "num_charts_in_width");
   flag = context->ObjectToInt32(dataobj, &num_charts_in_width); 
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "num_charts_in_height");
   flag = context->ObjectToInt32(dataobj, &num_charts_in_height); 
   
   double half_width,half_height;
   half_width = (chart_width / 2) + chart_width * .05;     // Added .1 extra in V29.0.3.11
   half_height = (chart_height / 2)  + chart_height * .05;
   
   // Convert the latitude to 0 - 180 for comparison  added in V32.0.0.5
   nl_sec += 324000; // increase by 90 * 3600  
   
   sprintf(buffer,"CalcWidePath max_ck_dist=%f max_range=%f chart_width=%f chart_height=%f",max_ck_dist,max_range,chart_width,chart_height);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
   sprintf(buffer,"CalcWidePath num_charts_in_width=%d num_charts_in_height=%d",num_charts_in_width,num_charts_in_height);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
   sprintf(buffer,"CalcWidePath nl/wl_sec=%f %f fract_width/height_sec=%f %f",nl_sec,wl_sec,fract_width_sec,fract_height_sec);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));    
   
   // if the wl is negative make it positive by adding 360 degrees so we test from 0 - 360 added in V32.0.0.5
   if (wl_sec < 0)
   {
      wl_sec += 1296000; // increase by 360 * 3600 
      sprintf(buffer,"CalcWidePath wl_sec was negative made it positive, now=%f",wl_sec);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));    
   }  
                                               
   // array to return the selected chart cps
   RexxArrayObject selected_cps;
   selected_cps = (RexxArrayObject)context->DirectoryAt(chart_dir, "selected_cps");        
   
   // Get pointer to the cps_latlon and allocate an array for it  
   RexxArrayObject cps_latlon_rexx;  
   double value_dbl;
   double *cps_latlon;  // local  array  
   int sz_cps;
   cps_latlon_rexx = (RexxArrayObject)context->DirectoryAt(chart_dir, "cps_latlon");  
   sz_cps = context->ArrayItems(cps_latlon_rexx);  // the number of cps_latlon
   cps_latlon = (double *)malloc(sz_cps * sizeof(double) ); 
   // Transfer all the rexx cps_latlon into local array actually has lat,lon
   for(n=0 ; n<sz_cps ; n++)
   { 
      dataobj = context->ArrayAt(cps_latlon_rexx, n+1);
      flag = context->ObjectToDouble(dataobj, &value_dbl);
      cps_latlon[n] = value_dbl;      
   }  
   
   // Get pointer to the cps_xpyp and allocate an array for it  
   RexxArrayObject cps_xpyp_rexx;  
   int value_int;
   int *cps_xpyp;  // local  array  
   int sz_cps_xpyp;
   cps_xpyp_rexx = (RexxArrayObject)context->DirectoryAt(chart_dir, "cps_xpyp");  
   sz_cps_xpyp = context->ArrayItems(cps_xpyp_rexx);  // the number of cps_xpxp
   cps_xpyp = (int *)malloc(sz_cps_xpyp * sizeof(int) );
   // Transfer all the rexx cps_xpyp into local array actually has xp,yp
   for(n=0 ; n<sz_cps_xpyp ; n++)
   { 
      dataobj = context->ArrayAt(cps_xpyp_rexx, n+1);
      flag = context->ObjectToInt32(dataobj, &value_int);
      cps_xpyp[n] = value_int;      
   }  
   
   // Get pointer to the ckpts_latlon and allocate an array for it  
   RexxArrayObject ckpts_latlon_rexx;  
   double *ckpts_latlon;  // local  array  
   int sz_ckpts;
   ckpts_latlon_rexx = (RexxArrayObject)context->DirectoryAt(chart_dir, "ckpts_latlon");  
   sz_ckpts = context->ArrayItems(ckpts_latlon_rexx);  // the number of ckpts 
   ckpts_latlon = (double *)malloc(sz_ckpts * sizeof(double));
   
   sprintf(buffer,"CalcWidePath sz_cps=%d sz_ckpts=%d",sz_cps, sz_ckpts);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
   
   // Transfer all the rexx ckps_latlon into local array actually has lat,lon of ckpt
   for(n=0 ; n<sz_ckpts ; n++)
   { 
      dataobj = context->ArrayAt(ckpts_latlon_rexx, n+1);
      flag = context->ObjectToDouble(dataobj, &value_dbl);
      ckpts_latlon[n] = value_dbl;      
   }  
   
   if ( diag>1 )
   {
      // list all cps actually has lat,lon,xp,yp
      double latval,lonval;
      int xval,yval;
      int offset=0;
      for(n=0 ; n<sz_cps ; n+=2)
      { 
        
         latval = cps_latlon[n]; 
         lonval = cps_latlon[n+1];
         xval = cps_xpyp[n]; 
         yval = cps_xpyp[n+1];   
         sprintf(buffer,"DIAG CalcWidePath offset=%d %f %f %d %d",offset,latval,lonval,xval,yval);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
         offset+=4;                
      }  
   }
   
   double chart_width_sec,chart_height_sec;
   chart_width_sec = int(chart_width * 3600);
   chart_height_sec = int(chart_height * 3600);
   
   sprintf(buffer,"DIAG CalcWidePath Processing checkpoints Chart_width/height_sec=%f %f",chart_width_sec,chart_height_sec);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));

   
   // Process each chart ckpts
   int ck_off,ck_num,nxt_ck_off;
   double ck_lat,ck_lon,nxt_ck_lat,nxt_ck_lon;
   int fnd_prim=0;int fnd_sec=0; int fnd_fscan=0; int total=0;
   int not_fnd_box = 0; 
   int fnd_box = 0;
   bool getout = false;
   double xx_range;
   double xx_bearing,scan_bearing,xx_bearing_minus_90;
   int lp,maxlp;
   for(ck_off=0,ck_num=1;ck_off<sz_ckpts;ck_off+=2,ck_num+=1)   
   {
      // Get the ckpt lat/lon
      ck_lat = ckpts_latlon[ck_off];  
      ck_lon = ckpts_latlon[ck_off+1]; 
      
      if ( diag>0 )
      {
         sprintf(buffer,"DIAG CalcWidePath checking ckpt#%d ck_off=%d sz_ckpts=%d ck_lat/lon=%f %f",ck_num,ck_off,sz_ckpts,ck_lat,ck_lon);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }
      
      // get the next ck num
      nxt_ck_off = ck_off + 2;
      if (nxt_ck_off < sz_ckpts) 
      {
         // Get the next ckpt lat/lon
         nxt_ck_lat = ckpts_latlon[nxt_ck_off];  
         nxt_ck_lon = ckpts_latlon[nxt_ck_off+1]; 
      
         // get the bearing from the next ck back to the current    
         DistanceBearingMercator( ck_lat,ck_lon,nxt_ck_lat,nxt_ck_lon, &xx_bearing, &xx_range );
         maxlp = 35;    // changed to 19 in V30.0.3.4 was 15
         xx_bearing_minus_90 = xx_bearing - 90;
         if (xx_bearing_minus_90 >= 360) xx_bearing_minus_90 = xx_bearing_minus_90 - 360; // make sure not over 360
         if (xx_bearing_minus_90 < 0) xx_bearing_minus_90 = xx_bearing_minus_90 + 360; // make not negative
      }
      else  // on last checkpoint
      {
         getout = true; 
         maxlp = 35;     // scan in complete circle at end   changed to 19 in V30.0.3.4 was 15
      }  
        
      // using the bearing find any charts within range
      double scan_dist;
      scan_dist = max_range;  // range to scan in 
      scan_bearing = xx_bearing_minus_90;  // start at -90
      // scan in a circle or ckpt scan in a complete circle around the checkpoint at 22.5 degree increments added in V30.0.1.3
      for (lp=1;lp<=maxlp;lp++)
      { 
         
         if (scan_bearing >= 360) scan_bearing = scan_bearing - 360; // make sure not over 360
         if (scan_bearing < 0) scan_bearing = scan_bearing + 360; // make not negative
         if ( diag>0 )
         {
            sprintf(buffer,"DIAG CalcWidePath starting lp=%d xx_bearing=%f scan_bearing=%f xx_bearing_minus_90=%f",lp,xx_bearing,scan_bearing,xx_bearing_minus_90);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         } 
         
         if ( diag>0 )
         {
            sprintf(buffer,"DIAG CalcWidePath checking ckpt%d lp=%d ck_lat/lon=%f %f nxt_lat/lon=%f %f xx_bearing=%f scan_bearing=%f ",ck_num,lp,ck_lat,ck_lon,nxt_ck_lat,nxt_ck_lon,xx_bearing,scan_bearing);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         } 
            
         // do from the max_range by the  max_ck_dist added in V29.0.3.6
         // double total_dist;
         // for(total_dist=0;total_dist<=max_range;total_dist+=max_ck_dist) 
         for(scan_dist=max_range;scan_dist>0;scan_dist-=max_ck_dist) // changed in V30.0.1.3 to use scan_dist
         {    
            double tst_lat,tst_lon;
            ll_gc_ll( ck_lat,ck_lon,scan_bearing,scan_dist, &tst_lat, &tst_lon );
            // rc = self~RhumbLineApparentLLRngBrgToLL(ck_lat,ck_lon,scan_bearing,scan_dist)
            // parse var rc tst_lat tst_lon .
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG CalcWidePath checking ckpt%d lp=%d tst_lat/lon=%f %f scan_bearing=%f scan_dist=%f",ck_num,lp,tst_lat,tst_lon,scan_bearing,scan_dist);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            }
            
            total++;
            
            double cklat,cklon;
            cklat=tst_lat;
            cklon=tst_lon;
            
            // deterine the offset from nl/wl
            double nl,sl,wl,el;
            
            double lat_sec,lon_sec;
            int off_width,off_height,uncom_xp,uncom_yp,xp,yp,uplf_xp,uplf_yp,max_xp,max_yp,prim_xp,prim_yp;
            // convert the lat/lon to seconds
            lat_sec = tst_lat * 3600; 
            lon_sec = tst_lon * 3600;   
            
            // if the lon_sec is negative make it positive by adding 360 degrees so we test from 0 - 360 added in V32.0.0.5
            if (lon_sec < 0)
            {
               lon_sec += 1296000; // increase by 360 * 3600 
               sprintf(buffer,"CalcWidePath lon_sec was negative made it positive, now=%f",lon_sec);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));    
            }  
            
            // Convert the lat_sec to 0 - 180 for comparison added in V32.0.0.5
            lat_sec += 324000; // increase by 90 * 3600  
   
            off_width = int(fabs(lon_sec - wl_sec));
            off_height = int(fabs(lat_sec - nl_sec));
              
            // determine the current chart offset
            uncom_xp = int(off_width / chart_width_sec);
            uncom_yp = int(off_height / chart_height_sec);
            
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG CalcWidePath uncompensated lat/lon_sec=%f %f off_width/height=%d off_height=%d uncom_xp=%d uncom_yp=%d ",lat_sec,lon_sec,off_width,off_height,uncom_xp,uncom_yp);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            } 
            
            // compensate the lat/ lon above  fract_width_sec/fract_height_sec 
            lat_sec = int((lat_sec + (uncom_yp * fract_height_sec)));
            lon_sec = int((lon_sec + (uncom_xp * fract_width_sec)));
                  
            // Rexx code
            // xp = abs(lon_sec - wl_sec) % chart_width_sec + 1
            // yp = abs(lat_sec - nl_sec) % chart_height_sec + 1
            
            xp = int(fabs(lon_sec - wl_sec) / chart_width_sec) + 1;
            yp = int(fabs(lat_sec - nl_sec) / chart_height_sec) + 1;
            
            prim_xp = xp;   // save for secondary 
            prim_yp = yp;
            
            if ( diag>0 )
            {
               sprintf(buffer,"DIAG CalcWidePath compensated lat/lon_sec=%f %f xp/yp=%d %d",lat_sec,lon_sec,xp,yp);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            }     
            
            bool found,rtcd;
            found = false;      // indicates whether we found the entry
            double cp_lat,cp_lon;
            cp_lat = NULL;
            int offs;   
            // make sure xp and yp are OK
            if (xp <= num_charts_in_width && xp >= 1 && yp <= num_charts_in_height && yp >= 1)
            {
               offs = ( (xp-1)*2) + ( (yp-1) * 2 * num_charts_in_width) ;  // get the offset into the cps_latlon
               
               // get the cps_latlon entry for this xp/yp
               if ( diag>0 )                                                                                                 
               {
                  sprintf(buffer,"DIAG CalcWidePath offs=%d xp/yp=%d %d",offs,xp,yp);
                  rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               } 
               
               // Get the cp lat/lon and xp yp
               cp_lat = cps_latlon[offs];  
               cp_lon = cps_latlon[offs+1]; 
               xp = cps_xpyp[offs];
               yp = cps_xpyp[offs+1];  
               
               if ( diag>0 )
               {
                  sprintf(buffer,"DIAG CalcWidePath cp_lat/lon=%f %f xp/yp=%d %d",cp_lat,cp_lon,xp,yp);
                  rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               } 
               
               // if the cp_lat is not NULL then check if it is valid 
               if (cp_lat)
               { 
                  // verify that the this is the proper chart cp  
                  nl = cp_lat + half_height;
                  sl = cp_lat - half_height;
                  wl = cp_lon - half_width;
                  wl = adjustAngleToPI(wl);   
                  el = cp_lon + half_width;
                  el = adjustAngleToPI(el);   
                  // check if the input lat lon is in this chart
                  if ( diag>0 )
                  {
                     sprintf(buffer,"DIAG CalcWidePath primary ckinbounds nl/wl=%f %f sl/el=%f %f cklat/lon=%f %f",nl,wl,sl,el,cklat,cklon);
                     rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                  }
                  
                  rtcd = ckinbounds(nl,wl,sl,el,cklat, cklon);
                  // if its within then 
                  if(rtcd)
                  { 
                     if ( diag>0 )
                     {
                        sprintf(buffer,"DIAG CalcWidePath found in primary cp_lat/lon=%f %f xp/yp=%d %d",cp_lat,cp_lon,xp,yp);
                        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                     }
                     found = true; 
                     fnd_prim++;
                  }
                  else 
                  {
                     if ( diag>0 )
                     {
                        sprintf(buffer,"DIAG CalcWidePath NOT found in primary cp_lat/lon=%f %f xp/yp=%d %d",cp_lat,cp_lon,xp,yp);
                        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                     }
                  }
               
               }
               else  // the cp_lat is NULL
               {
                  if ( diag>0 )
                  {
                     sprintf(buffer,"DIAG CalcWidePath NOT found in primary cp_lat is NULL xp/yp=%d %d",xp,yp);
                     rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                  }
               }
            }
             
            // If it wasn't found then try secondary with one xp less
            if (found == false)
            {
               xp = prim_xp - 1;
               yp = prim_yp;
               // make sure xp and yp are OK
               if (xp <= num_charts_in_width && xp >= 1 && yp <= num_charts_in_height && yp >= 1)
               {   
                  offs = ( (xp-1)*2) + ( (yp-1) * 2 * num_charts_in_width) ;  // get the offset into the cps_latlon
                  if ( diag>0 )
                  {
                     sprintf(buffer,"DIAG CalcWidePath secondary offs=%d xp/yp=%d %d",offs,xp,yp);
                     rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                  }
                  // Get the cp lat/lon and xp yp
                  cp_lat = cps_latlon[offs];  
                  cp_lon = cps_latlon[offs+1]; 
                  xp = cps_xpyp[offs];
                  yp = cps_xpyp[offs+1];  
                  if (cp_lat) // if the entry is valid
                  { 
                     nl = cp_lat + half_height;
                     sl = cp_lat - half_height;
                     wl = cp_lon - half_width;
                     wl = adjustAngleToPI(wl);   
                     el = cp_lon + half_width;
                     el = adjustAngleToPI(el);   
                     // check if the input lat lon is in this chart
                     // check if the input lat lon is in this chart
                     if ( diag>0 )
                     {
                        sprintf(buffer,"DIAG CalcWidePath secondary ckinbounds nl/wl=%f %f sl/el=%f %f cklat/lon=%f %f",nl,wl,sl,el,cklat,cklon);
                        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                     } 
                  
                     rtcd = ckinbounds(nl,wl,sl,el,cklat, cklon);
                     // if its within then 
                     if(rtcd)
                     {  
                        found = true;
                        fnd_sec++;
                        if ( diag>0 )
                        {
                           sprintf(buffer,"DIAG CalcWidePath found in secondary cp_lat/lon=%f %f xp/yp=%d %d",cp_lat,cp_lon,xp,yp);
                           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                        }
                     }
                     else
                     {  
                        if ( diag>0 )
                        {
                           sprintf(buffer,"DIAG CalcWidePath NOT found in secondary cp_lat/lon=%f %f xp/yp=%d %d",cp_lat,cp_lon,xp,yp);
                           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                        }
                     }
                  }
                  else
                  {
                     if ( diag>0 )
                     {
                        sprintf(buffer,"DIAG CalcWidePath found in secondary  but its NULL ");
                        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                     }
                  }
               } 
            } 
                         
            // make chart size a little bigger
            half_width = (chart_width / 2) + chart_width * .1;
            half_height = (chart_height / 2)  + chart_height * .1;
               
            // if not found try a box scan with the centre being the uncompensated  xp yp
            if (found == false)   
            {  
               uplf_xp = prim_xp- 2;
               uplf_yp = prim_yp - 2;
               uplf_xp = uncom_xp - 2;
               uplf_yp = uncom_yp - 2;
               if (uplf_xp > num_charts_in_width -4 ) uplf_xp = num_charts_in_width-4;
               if (uplf_xp <= 0) uplf_xp = 1;
               if (uplf_yp > num_charts_in_height -4) uplf_yp = num_charts_in_height-4;
               if (uplf_yp <= 0) uplf_yp = 1;
               max_xp = xp + 4;
               if (max_xp > num_charts_in_width) max_xp = num_charts_in_width;
               max_yp = yp + 4;
               if (max_yp > num_charts_in_height) max_yp = num_charts_in_height;
               if ( diag>0 )
               {
                  sprintf(buffer,"DIAG CalcWidePath doing boxscan uplf_xp/yp=%d %d max_xp/yp= %d %d num_charts_w/h %d %d",uplf_xp,uplf_yp,max_xp,max_yp,num_charts_in_width,num_charts_in_height);
                  rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               }
                  
               for(yp=uplf_yp;yp<=max_yp;yp++)
               {
                  for(xp=uplf_xp;xp<=max_xp;xp++)
                  {
                     offs = ( (xp-1)*2) + ( (yp-1) * 2 * num_charts_in_width) ;  // get the offset into the cps_latlon
                     if ( diag>1 )
                     {
                        sprintf(buffer,"DIAG CalcWidePath boxscan offs=%d xp/yp=%d %d",offs,xp,yp);
                        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                     }
                     if(offs < sz_cps && offs > 0) // if valid offset 
                     {
                        // Get the cp lat/lon and xp yp
                        cp_lat = cps_latlon[offs];  
                        cp_lon = cps_latlon[offs+1]; 
                        xp = cps_xpyp[offs];
                        yp = cps_xpyp[offs+1];  
                        if (cp_lat) // if the entry is valid
                        { 
                           nl = cp_lat + half_height;
                           sl = cp_lat - half_height;
                           wl = cp_lon - half_width;
                           wl = adjustAngleToPI(wl);   
                           el = cp_lon + half_width;
                           el = adjustAngleToPI(el);   
                           // check if the input lat lon is in this chart
                           // check if the input lat lon is in this chart
                           if ( diag>1 )
                           {
                              sprintf(buffer,"DIAG CalcWidePath boxscan ckinbounds nl/wl=%f %f sl/el=%f %f cklat/lon=%f %f",nl,wl,sl,el,cklat,cklon);
                              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                           } 
                           rtcd = ckinbounds(nl,wl,sl,el,cklat, cklon);
                           // if its within then 
                           if(rtcd)
                           {  
                              found = true;
                              fnd_box++;
                              if ( diag>0 )
                              {
                                 sprintf(buffer,"DIAG CalcWidePath found in boxscan cp_lat/lon=%f %f xp/yp=%d %d",cp_lat,cp_lon,xp,yp);
                                 rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                              }
                              break;
                           }
                        }
                     }
                  }
                  if (found) break;
               }
               
               if (found == false)   
               {  
                  if ( diag>0 )
                  {
                     sprintf(buffer,"DIAG CalcWidePath NOT found in boxscan cp_lat/lon=%f %f xp/yp=%d %d",cp_lat,cp_lon,xp,yp);
                     rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                  }
                  
                  not_fnd_box++; // increment the not found in boxscan
               }
            
            }   
            
            if (found)  // we found a chart 
            {
               if (cp_lat)
               {
                  // add the chart cp to the selected list
                  sprintf(buffer,"%f %f %d %d",cp_lat,cp_lon,xp,yp);   // changed xp,yp from %f to %d in V30.0.0.4
                  rc = context->ArrayAppendString(selected_cps, buffer, strlen(buffer));
                  
                  if ( diag>0 )
                  {
                     sprintf(buffer,"DIAG CalcWidePath adding chart at cp_lat/lon=%f %f xp/yp=%d %d",cp_lat,cp_lon,xp,yp);
                     rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                  }
               }
               else
               {
                  // self~logmsg('Get_chart_cp_lat_lon already in chart_latlons xp/yp='xp yp' cp lat/lon:'latlon)
               }
            } 
         } // end scan_dist changing
         scan_bearing+=10;   // increase the scan bearing  changed to 10 in V30.0.3.8 
      } // end of FOR semi circle loop
      // finshed loops in one direction.  Reverse it if wwe found one more charts
      if (getout) break;   // if last ckpt get out
   }   // end of ckpts
      
   // print the totals
   int total_fnd;
   total_fnd = fnd_prim+fnd_sec+fnd_box;
   sprintf(buffer,"DIAG CalcWidePath fnd_prim=%d fnd_sec=%d fnd_box=%d total not found in box scan scan=%d total found=%d ",fnd_prim,fnd_sec,fnd_box,not_fnd_box,total);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
     

   // free all the arrays we allocated
   free(ckpts_latlon);  
   free(cps_latlon);   
   free(cps_xpyp);     
   
   return 0; 
}  

/**************************** CalcSplit **************************
  Calculates how many times a chart list should be split so that it fits in the comm._max_quilt_size 
* Syntax:	rc = CalcSplit(chartinfo, diag, diagmsgs)
 *
 * Params:	
 *			   chartinfo - directory of chart information (boundaries, chart_width, chart_height)
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages   
 * Return: split amount 0 - n or returns -1 if there is an error
 *********************************************************************/
RexxRoutine3(int, CalcSplit, RexxObjectPtr, chartinfo, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[1000]; 
   int rc,n,m,p,q;
   logical_t flag;
   
   // int *xy_coordinates;  // local  array  
   
   RexxObjectPtr dataobj;
   RexxDirectoryObject chart_dir;      // RexxDirectoryObject)
   
   chart_dir = (RexxDirectoryObject)chartinfo;
   flag = context->IsDirectory(chart_dir);
   if ( flag == 0 )
   {
      sprintf(buffer,"ERROR: CalcSplit parameter 1 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return -1;
   }
    
   //Get all necessary values that are not arrays   ObjectToStringValue
  
   int range,use_SP_tiles;
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "range");  // the zoom level
   flag = context->ObjectToInt32(dataobj, &range); 
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "use_SP_tiles");   // flag to indicate we are using sp tiles is on = 1 or off=0
   flag = context->ObjectToInt32(dataobj, &use_SP_tiles); 
   
   // doubles:
   double dpp_lat,dpp_lon,max_quilt_size;
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "dpp_lat");
   flag = context->ObjectToDouble(dataobj, &dpp_lat);  
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "dpp_lon");
   flag = context->ObjectToDouble(dataobj, &dpp_lon);   
   dataobj = (RexxObjectPtr)context->DirectoryAt(chart_dir, "max_quilt_size");
   flag = context->ObjectToDouble(dataobj, &max_quilt_size); 
   
   // Get pointer to the splits_array rexx array    added in V33.0.0.7
   RexxArrayObject splits_array;
   splits_array = (RexxArrayObject)context->DirectoryAt(chart_dir, "splits_array");  
    
   // Get pointer to the  diaginfo rexx array    added in V30.0.3.1
   RexxArrayObject diaginfo;
   diaginfo = (RexxArrayObject)context->DirectoryAt(chart_dir, "DIAGINFO");  
   
   // Get pointer to the chart_list and allocate an array for it  
   RexxArrayObject chart_list_rexx;  
   double value_dbl;
   double *chart_list;  // local  array  
   int sz_cl;
   chart_list_rexx = (RexxArrayObject)context->DirectoryAt(chart_dir, "chart_list");  
   sz_cl = context->ArrayItems(chart_list_rexx);  // the number of entries in the chart list
   chart_list = (double *)malloc(sz_cl * sizeof(double) * 4 );  //local array actually has ul_lat ul_lon lr_lat lr_lon
   
   sprintf(buffer,"CalcSplit max_quilt_size=%f dpp_lat=%f dpp_lon=%f range=%d chart_list entries=%d",max_quilt_size,dpp_lat,dpp_lon,range,sz_cl);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
   
   // Transfer all the rexx chart_list entries  into local array actually has ul_lat ul_lon lr_lat lr_lon
   RexxDirectoryObject cd; // chart data directory
   q = 0; //offset into the chart_list array
   for(n=0 ; n<sz_cl ; n++)
   { 
      cd = (RexxDirectoryObject)context->ArrayAt(chart_list_rexx, n+1);
      if ( cd != NULLOBJECT )
      {      
         CSTRING bounds;                             
         dataobj = (RexxObjectPtr)context->DirectoryAt(cd, "BOUNDS"); 
         if ( dataobj != NULLOBJECT)
         {           
            bounds = context->ObjectToStringValue(dataobj); 
            
            if ( diag>1 )
            {
               sprintf(buffer,"DIAG CalcSplit bounds=%s",bounds);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            }
            // convert the bounds string to the lat lons  23.144581 -82.433568 23.138031 -82.418977
            p=0;  // offset into buffer 
            int len = lstrlen(bounds);
            // parse out the bounds to get all the ul_lat ul_lon and the lr_lat lr_lon
            for(m=0;m <= len ;m++)
            {  
               if (bounds[m] ==  0x20 || m == len) // if we found a blank  or end of string
               {
                  buffer[p] =  '\0'; // put in the string delimiter
                  sscanf(buffer, "%lf", &value_dbl);   // convert from string to double
                  if ( diag>1 )
                  {
                     sprintf(buffer,"DIAG CalcSplit m=%d p=%d len=%d value_dbl=%f",m,p,len,value_dbl);
                     rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                  }
                  chart_list[q] = value_dbl;  // put it in the chart_list array
                  q++; 
                  p = 0;
               }
               buffer[p] = bounds[m] ;
               p++;   
            }
         }
      }
   }  
   
   sprintf(buffer,"DIAG CalcSplit Total chart_list entries=%d",q);        
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));

   // If we are using Sasplanet tiles Convert all the lat/lons to the tile lat/lons
   if (use_SP_tiles == 1)
   {  
      double lat,lon,xp,yp,tile_lat,tile_lon,sinLatitude,mapSize;
      int zm,zoom_level,pixelx,pixely,tilex,tiley;
      zoom_level = range;
      zm = zoom_level - 1; // get the zoom -1
      m = int(pow(2.0,zm));  //  multiplier for mapsize
      mapSize = 256 * m; 
      /* RExx:
       zoom_level = cd["RANGE"]
      zm = zoom_level - 1 -- get the zoom -1
      m = RxCalcPower(2.0,zm)  -- multiplier for mapsize
      mapsize = 256 * m   
      */
      sprintf(buffer,"DIAG CalcSplit zoom_level=%d zm=%d m=%d mapsize=%f",zoom_level,zm,m,mapSize);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      int chart_num=1;
      for(n=0 ; n<q ; n+=4) // by 4
      {  
         // DO the UL 
         lat = chart_list[n]; // get the ul_lat
         lon = chart_list[n+1]; // get the ul_lon     
         xp = (lon + 180) / 360;
         sinLatitude = sin(lat * DEG_TO_RAD);
         yp = 0.5 - log((1.0 + sinLatitude) / (1.0 - sinLatitude)) / (4.0 * ONEPI);
         pixelx = int((xp * mapSize + 0.5));
         pixely = int((yp * mapSize + 0.5));
         tilex = (pixelx /256);
         tiley = (pixely / 256);
         if ( diag>1 )
         {
            sprintf(buffer,"DIAG CalcSplit chart#%d UL xp/yp=%f %f pixelx/y=%d %d tilex/y=%d %d",chart_num,xp,yp,pixelx,pixely,tilex,tiley);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }
          
         // determine the lat/lon of the UL pixel of the UL tile 
         pixelx = tilex * 256;
         pixely = tiley * 256;
         xp = (pixelx / mapSize) - 0.5;
         yp = 0.5 - (pixely / mapSize);
         tile_lat = 90 - 360 * atan(exp(-yp * 2.0 * ONEPI)) / ONEPI;
         tile_lon = 360 * xp;
         if ( diag>1 )
         {
            sprintf(buffer,"DIAG CalcSplit chart#%d UL xp/yp=%f %f pixelx/y=%d %d tilex/y=%d %d tile_la/lon=%f %f",chart_num,xp,yp,pixelx,pixely,tilex,tiley,tile_lat,tile_lon);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }
         
         if ( diag>0 )
         {
            sprintf(buffer,"DIAG CalcSplit chart#%d UL was=%f %f now=%f %f",chart_num,chart_list[n],chart_list[n+1],tile_lat,tile_lon);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }
         chart_list[n] = tile_lat;  // replace the UL lat and lon
         chart_list[n+1] = tile_lon;
         
         // do the LR
         lat = chart_list[n+2]; // get the lr_lat
         lon = chart_list[n+3]; // get the lr_lon   
         xp = (lon + 180) / 360;
         sinLatitude = sin(lat * DEG_TO_RAD);
         yp = 0.5 - log((1.0 + sinLatitude) / (1.0 - sinLatitude)) / (4.0 * ONEPI);
         pixelx = int((xp * mapSize + 0.5));
         pixely = int((yp * mapSize + 0.5));
         tilex = (pixelx /256);
         tiley = (pixely / 256);
         // determine the lat/lon of the UL pixel of the UL tile 
         pixelx = (tilex * 256) + 255  ;
         pixely = (tiley * 256) + 255;
         xp = (pixelx / mapSize) - 0.5;
         yp = 0.5 - (pixely / mapSize);
         tile_lat = 90 - 360 * atan(exp(-yp * 2.0 * ONEPI)) / ONEPI;
         tile_lon = 360 * xp;
         if ( diag>0 )
         {
            sprintf(buffer,"DIAG CalcSplit chart#%d LR was=%f %f now=%f %f",chart_num,chart_list[n+2],chart_list[n+3],tile_lat,tile_lon);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }
         chart_list[n+2] = tile_lat;  // replace the  LR lat and lon
         chart_list[n+3] = tile_lon;   
         chart_num++;              
      }  
   } 
   
   // Split the list into groups less than the maximum size
   int list_num = 1;
   int splits,start;

   start = 0;
   double ul_lat, ul_lon,lr_lat,lr_lon,latdiff,londiff,width_pixels,height_pixels,proj_size;
   for(splits=1;splits<10000;splits++)
   {
      
      sprintf(buffer,"DIAG CalcSplit q=%d splits=%d start=%d list_num=%d",q,splits,start,list_num);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      
      // add the list_num to the splits_array
      sprintf(buffer,"%d",list_num);  
      rc = context->ArrayAppendString(splits_array, buffer, strlen(buffer));
      
      // scan the chart list until the size of the chart is gt the max_quilt_size 
      // chart boundaries  
      double UL_LAT = -999.0; 
      double UL_LON = 999.0;     
      double LR_LAT = 999.0;
      double LR_LON = -999.0;
      bool ul_lon0,lr_lon0;
      
      for(n=start ; n<q ; n+=4) // by 4
      {  
         
         ul_lat = chart_list[n];   // get the ul_lat
         ul_lon = chart_list[n+1]; // get the ul_lon 
         lr_lat = chart_list[n+2]; // get the lr_lat
         lr_lon = chart_list[n+3]; // get the lr_lon  
         
         // see if chart extends the boundaries of the quilted image  quilt_data UL_LAT UL_LON LR_LAT LR_LON
         if (ul_lat > UL_LAT) UL_LAT = ul_lat; 
         if (lr_lat < LR_LAT) LR_LAT = lr_lat;
         
         // make the lons from 0 360 to do compare  -- added in V23.0.0.7
         // if (ul_lon < 0) ul_lon += 360;  // make bewteen 0 and 360
         // if (lr_lon < 0) lr_lon += 360;
         // if (UL_LON < 0) UL_LON += 360;  // make bewteen 0 and 360
         // if (LR_LON < 0) LR_LON += 360;
         
         // check if UL crossing lon0   added in V23.0.0.7
         if (fabs(UL_LON) != 999 && fabs(ul_lon - UL_LON) > 200) 
         {
            ul_lon0 = true;
            sprintf(buffer,"DIAG CalcSplit UL_LON0 CROSSED UL_LON=%f ul_lon=%f",UL_LON,ul_lon);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }
         else  ul_lon0 = false;
         // if we crossed longitude 0  
         if (ul_lon0) // if we crossed lon0
         {
            if (ul_lon > UL_LON) UL_LON = ul_lon;
         }
         else // we didn't cross lon0
         {
            if (ul_lon < UL_LON) UL_LON = ul_lon;
         }
         
         // check if LR crossing lon0   added in V23.0.0.7
         if (fabs(LR_LON) != 999 && fabs(lr_lon - LR_LON) > 200) 
         {
            lr_lon0 = true;
            sprintf(buffer,"DIAG LR_LON0 CROSSED");
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }
         else  lr_lon0 = false;
         // if we crossed longitude 0  
         if (lr_lon0) // if we crossed lon0
         {
            if (lr_lon < LR_LON) LR_LON = lr_lon;
         }
         else // we didn't cross lon0
         {
            if (lr_lon > LR_LON) LR_LON = lr_lon;
         }
       
         // adjust the lons 
         UL_LON =  adjustAngleToPI(UL_LON);
         LR_LON =  adjustAngleToPI(LR_LON);
         
         // check if the resulting chart is greater than the maximum
         latdiff = fabs(DiffLatLon(false,UL_LAT,LR_LAT));
         londiff = fabs(DiffLatLon(true,LR_LON,UL_LON));
         if ( diag>0 )
         {
            sprintf(buffer,"DIAG CalcSplit lon0_ul_lr='%d %d UL_LAT/LON=%f %f LR_LAT/LON %f %f lat/londiff=%f %f",ul_lon0,lr_lon0,UL_LAT,UL_LON,LR_LAT,LR_LON,latdiff,londiff);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }
         
         width_pixels = londiff / dpp_lon +.5; 
         height_pixels = latdiff / dpp_lat +.5;
         
         // if the proj_size is gt the maximum quit this set 
         proj_size = width_pixels * height_pixels; 
         if (proj_size > max_quilt_size) 
         {
            
            sprintf(buffer,"DIAG CalcSplit n=%d width/height_pixels=%f %f proj_size=%f list_num=%d",n,width_pixels,height_pixels,proj_size,list_num);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            
            start = n; // start at this chart  
            break; 
         }
         list_num ++; // increment the list number
   
      }  // end of n=start
      if(n >= q) break;
   }  // end of splits
   
   sprintf(buffer,"DIAG CalcSplit end of splits n=%d q=%d width/height_pixels=%f %f proj_size=%f list_num=%d",n,q,width_pixels,height_pixels,proj_size,list_num);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));

   
   sprintf(buffer,"DIAG  splits=%d",splits);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   
   // free all the arrays we allocated
   free(chart_list); 
   
   return splits; 
}  
/**************************** Quilt **************************
  Creates one large BMP from multiple input images
 * Syntax:	rc = Quilt(Quilt_Data, diag, diagmsgs)
 *
 * Params:	
 *			   Quilt_Data - directory of quilt information:
 *			   (bmp_width/height, chart width/height, degrees/pixel x/y 
 *			       array of images, array of the images start x/y offsets)
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages 
 *			   The following were parameters and now they are in the Quilt_data as flags
 *			   check_size   if = 1 then check size and return
 *			   bypass_red  - if = 1 then bypass setting last palette entry to red   added in V27.0.0.6
 * Return: the 0 if OK else 1  and if OK then the chartbits in the Quilt_Data
 *********************************************************************/
RexxRoutine3(int, Quilt, RexxObjectPtr, Quilt_Data,  int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512];
   
   int rc,num_images,n,num_entries,num_ovl_tiles;      
   int32_t gif_width_pixels, gif_height_pixels, chart_width, chart_height, x_offset, y_offset;
   logical_t flag,outside_border,check_size,bypass_red;
   int32_t alpha_value, x,y;
   RexxArrayObject gif_image_names, x_start_offsets, y_start_offsets, palette;
   RexxArrayObject ovl_tiles, ovl_x_start_offsets, ovl_y_start_offsets;
  
   RexxObjectPtr dataobj;
   RexxDirectoryObject quilt_dir;
   
   quilt_dir = (RexxDirectoryObject)Quilt_Data;
   flag = context->IsDirectory(quilt_dir);
   if ( flag == 0 )
   {
      sprintf(buffer,"Error: Quilt parameter 2 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }  
   
   CSTRING filename;
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "FILENAME"); 
   filename = context->ObjectToStringValue(dataobj);
   
   // See if filename is a valid image type
	TCHAR* extout = FindExtension(filename);
	_tcslwr(extout);
	int typeout = CxImage::GetTypeIdFromName(extout);
   if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"Error: Quilt filename is an invalid image type.");
      // printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}  
   sprintf(buffer,"DIAG Quilt filename:%s type=%d",filename,typeout);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));   
        
   //Get all necessary values that are not arrays
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "GIF_WIDTH_PIXELS");
   flag = context->ObjectToInt32(dataobj, &gif_width_pixels);
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "GIF_HEIGHT_PIXELS");
   flag = context->ObjectToInt32(dataobj, &gif_height_pixels); 
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "CHART_WIDTH");
   flag = context->ObjectToInt32(dataobj, &chart_width);  
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "CHART_HEIGHT");
   flag = context->ObjectToInt32(dataobj, &chart_height);   
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "OUTSIDE_BORDER");
   flag = context->ObjectToLogical(dataobj, &outside_border); 
   // following added in V28.0.1.3 
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "CHECK_SIZE");
   flag = context->ObjectToLogical(dataobj, &check_size);  
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "BYPASS_RED");
   flag = context->ObjectToLogical(dataobj, &bypass_red);  
   int rectangle; // added in V30.0.0.4 
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "RECTANGLE");
   flag = context->ObjectToInt32(dataobj, &rectangle); 
   // following added in V6.0.2.1
   int maxcolors; 
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "MAXCOLORS");
   flag = context->ObjectToInt32(dataobj, &maxcolors);   
                                                 
   // added in V27.0.0.8
   sprintf(buffer,"DIAG Quilt maxcolors=%d gif_width_pixels=%d gif_height_pixels=%d rectangle=%d",maxcolors,gif_width_pixels, gif_height_pixels,rectangle);  
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   sprintf(buffer,"DIAG Quilt chart_width=%dpixels chart_height=%dpixels",chart_width, chart_height);  
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));   
   // Print the flags added in V28.0.1.3  
   sprintf(buffer,"DIAG Quilt outside_border=%d check_size=%d bypass_red=%d",outside_border,check_size,bypass_red);  
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
     
   // Create a new output gif image 
   CxImage *newImage = new CxImage();
   DWORD wBpp = 8;  // bits per pixel
   newImage->Create(gif_width_pixels,gif_height_pixels,wBpp,2);  // Change to 2 in V33.0.0.7= CXIMAGE_FORMAT_GIF         
   if (newImage == NULL)
   {
      sprintf(buffer,"Error: Quilt Unable to create output file:%s",filename);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;  
   }
   
   // Get the the images width and height
   int image_width, image_height;
   image_width = newImage->GetWidth();
   image_height = newImage->GetHeight();
   
   // Get the bpp
   int bpp;
   bpp  =  newImage->GetBpp(); // Get the bpp added in V27.0.0.6
   
   //  pass back the XRES and YRES
   int32_t resx, resy;
   resx = newImage->GetXDPI(); // Get X resolution
   resy = newImage->GetYDPI(); // Get Y resolution
      
   dataobj = context->Int32ToObject(resx); 
   context->DirectoryPut(quilt_dir, dataobj, "XRES"); 
   dataobj = context->Int32ToObject(resy); 
   context->DirectoryPut(quilt_dir, dataobj, "YRES");
   sprintf(buffer,"DIAG Quilt Image Width/Height=%d %d Bpp=%d Xres=%d Yres=%d",image_width,image_height,bpp,resx, resy);  // bpp added in V27.0.0.6
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
    
   // If we did not create an image of the proper size then quit
   if ((image_width != gif_width_pixels) | (image_height != gif_height_pixels))
   {
      sprintf(buffer,"Error: Quilt Unable to create output GIF of proper size should be:%d %d is:%d %d",gif_width_pixels,gif_height_pixels,image_width,image_height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1; 
   } 
   
   // If there is check_size is true then return    
   if (check_size == 1 ) // added in V28.0.1.3
   {    
      sprintf(buffer,"DIAG Quilt check_size is true, returning");  // added in V28.0.1.3
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      
      // try to Save  the mage up to 3 times added in V33.0.0.7
      for(int lp=0; lp<3; lp++)
      {
         rc = newImage->Save(filename,typeout);	
         if (!rc)
         {
            if (lp == 2)
            {
               sprintf(buffer,"Error: Quilt Check Size Invalid rc from Save Image RC=%d\n ",rc);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               return 1;
            }
            sprintf(buffer,"Warning: Quilt Check Size Invalid rc from Save Image RC=%d\n ",rc);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }
         else break; 
      }
       
      /* Save the image
      rc = newImage->Save(filename,typeout);	
      if(!rc)
      {
        sprintf(buffer,"Error: Quilt Invalid rc from Save Image RC=%d\n ",rc);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }	
      */
        
      // Destroy the image if one exists
      if (newImage->IsValid()) 
      {
         rc = newImage->Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: Quilt 1. Invalid rc from Destroy Image RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }	 
         sprintf(buffer,"DIAG Quilt destroyed newimage rc=%d",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
      }  
      return 0;  
   } 
   
   // Get pointer to the palette
   palette = (RexxArrayObject)context->DirectoryAt(quilt_dir, "PALETTE"); 
   num_entries = context->ArrayItems(palette);  // the number of palette entries  
   if (num_entries != 256) 
   {
      sprintf(buffer,"Error: Quilt the number of palette entries is not 256.");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}   
   
   // set the whole image to fully transparent
   newImage->AlphaCreate();
   //newImage->AlphaClear();
   newImage->AlphaSet(1);
   alpha_value = newImage->AlphaGet(0,0); // get the alpha value
   sprintf(buffer,"DIAG Quilt Before alpha_value=%d",alpha_value);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
     
   // Get pointer to the  GIF_IMAGE_NAMES rexx array
   gif_image_names = (RexxArrayObject)context->DirectoryAt(quilt_dir, "GIF_IMAGE_NAMES");  
   num_images = context->ArrayItems(gif_image_names);  // the number of images
   
   // Get pointer to the  overlay tile filespecs rexx array  added in V27.0.0.6
   ovl_tiles = (RexxArrayObject)context->DirectoryAt(quilt_dir, "OVL_TILES");  
   num_ovl_tiles = context->ArrayItems(ovl_tiles);  // the number of  tiles  
   sprintf(buffer,"DIAG Quilt num_images=%d num_ovl_tiles=%d ",num_images,num_ovl_tiles);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));    
   
   // array for the border points added i V23.0.0.4
   RexxArrayObject ply_points;
   ply_points = (RexxArrayObject)context->DirectoryAt(quilt_dir, "PLY_POINTS");  
    
   // Get pointer to the  xy_start_offsets array
   x_start_offsets = (RexxArrayObject)context->DirectoryAt(quilt_dir, "X_START_OFFSETS");  
   y_start_offsets = (RexxArrayObject)context->DirectoryAt(quilt_dir, "Y_START_OFFSETS");
   ovl_x_start_offsets = (RexxArrayObject)context->DirectoryAt(quilt_dir, "OVL_X_START_OFFSETS");  
   ovl_y_start_offsets = (RexxArrayObject)context->DirectoryAt(quilt_dir, "OVL_Y_START_OFFSETS");
        
   sprintf(buffer,"DIAG Quilt num_images=%d  GIF WIDTH/HEIGHT: %d %d ",num_images,gif_width_pixels,gif_height_pixels);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   // Flip it over so x=0,y=0 are at upperright
   newImage->Flip();
   
   int numcolors;
   numcolors = newImage->GetNumColors();
   // printf("NUMCOLORS %d\n",numcolors); 
   
   int palette_size;
   palette_size = newImage->GetPaletteSize(); 
   // printf("palette SIZE %d\n",palette_size); 
   
   int idx;
   idx = newImage->GetPixelIndex(0,0);
   // ULONG pixval;   
   RGBQUAD color;    
   
   RGBQUAD  WhiteColor;  
   WhiteColor.rgbRed = 255; 
   WhiteColor.rgbGreen = 255; 
   WhiteColor.rgbBlue = 255; 
   WhiteColor.rgbReserved = 0;    
   
   RGBQUAD BlackColor;
   BlackColor.rgbRed = 0; 
   BlackColor.rgbGreen = 0; 
   BlackColor.rgbBlue = 0; 
   BlackColor.rgbReserved = 0;  
     
    // initialize the palette 
   uint32_t colorref;
   BYTE *p = (BYTE *)(uint32_t *)&colorref;
   BYTE red,green,blue;
   for(n=0 ; n<256 ; n++)
   {
      dataobj = context->ArrayAt(palette, n+1);
      flag = context->ObjectToUnsignedInt32(dataobj, &colorref);
      
      // get the palette colors Note they are reversed format is rrggbb00
      red = p[0]; 
      green = p[1];  
      blue = p[2];      
      newImage->SetPaletteColor(n,red, green, blue);
   }
   
   // sets the maximum number of colors to use for  SetPixelColor	
   newImage->SetClrImportant(maxcolors); // added maxcolors in V6.0.2.1
      
   // Read each of the gif images    
   // keep track of the max x and y which will be used   by the find gap code added in V27.0.0.6                              
   int max_x = 0;  
   int max_y = 0;   
   CSTRING  name;
   // Create a new image from the gif file     
   CxImage *loadImage = new CxImage(); // Create the image only once
   for(n=0 ; n<num_images ; n++)
   { 
      //  get the name of tile
      dataobj = context->ArrayAt(gif_image_names, n+1);
      name = context->ObjectToStringValue(dataobj);
      
      // printf("DEBUG name=%s\n",name);

      // try to load the tile up to 3 times added in V32.0.0.7
      for(int lp=0; lp<3; lp++)
      {
         rc = loadImage->Load(name,3);  // Changed to 3 which is JPG type in V31     
         if (!rc && lp == 3)
         {
            sprintf(buffer,"Error: Quilt Invalid rc from Load map tile RC=%d file:%s",rc,name );
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            return 1;
         }
         else break; 
      }
      
      // Flip it over
      loadImage->Flip();
      
      dataobj = context->ArrayAt(x_start_offsets, n+1);
      flag = context->ObjectToInt32(dataobj, &x_offset);
      dataobj = context->ArrayAt(y_start_offsets, n+1);
      flag = context->ObjectToInt32(dataobj, &y_offset);
      
      // get the realwidth and height of the image added in V33.0.0.6
      chart_width = loadImage->GetWidth();
      chart_height = loadImage->GetHeight();
      
      if(diag > 0) // Added in V33.0.0.7  
      {
         sprintf(buffer,"DIAG Quilt quilting map tile at x/y chart_width/height=%d %d offset=%d %d filename=%s",chart_width,chart_height,x_offset,y_offset,name);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }
       
      // Transfer all the pixels from the input gif to output gif
      long x,y,ox,oy;
   	for(y=0,oy=y_offset; y<chart_height; y++,oy++)
      {
         if (oy>=gif_height_pixels) break;  // if the oy are greater than or equal than the gif height don't transfer the pixel  break added in V27.0.0.6
   		for(x=0,ox=x_offset; x<chart_width; x++,ox++)
         {
            if (ox>=gif_width_pixels) break; // if the ox are greater than or equal than the gif width don't transfer the pixel  break added in V27.0.0.6
            // Get pixel color
   			color = loadImage->GetPixelColor(x,y);
            newImage->SetPixelColor(ox,oy,color,false);
            newImage->AlphaSet(ox,oy,0); // added in V23.0.0.19
   		}
         if (ox>max_x) max_x = ox; // added in V27.0.0.6
      }  
      if (oy>max_y) max_y = oy; // added in V27.0.0.6 
             
      // Destroy the input image if one exists
      if (loadImage->IsValid()) 
      {
         rc = loadImage->Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: Quilt 2. Invalid rc from Destroy Image RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }	 
      }  
         
   }
   
   // if there any overlay images then add them too   added in V27.0.0.6
   if(num_ovl_tiles!=0)
   {
      // Read each of the overlay mages
      CSTRING  name;
      CxImage *loadImage = new CxImage();  // create a new image only once  moved here in V31
      for(n=0 ; n<num_ovl_tiles ; n++)
      { 
         
         dataobj = context->ArrayAt(ovl_tiles, n+1);
         name = context->ObjectToStringValue(dataobj);
         
         // try to load the tile up to 3 times added in V32.0.0.7
         for(int lp=0; lp<3; lp++)
         {
            rc = loadImage->Load(name,1);  
            if (!rc && lp == 3)
            {
               sprintf(buffer,"Error: Quilt Invalid rc from Load layer tile RC=%d file:%s",rc,name );
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               return 1;
            }
            else break; 
         }
         
         // Flip it over
         // loadImage->Flip();
         
         dataobj = context->ArrayAt(ovl_x_start_offsets, n+1);
         flag = context->ObjectToInt32(dataobj, &x_offset);
         dataobj = context->ArrayAt(ovl_y_start_offsets, n+1);
         flag = context->ObjectToInt32(dataobj, &y_offset);
         
         if(diag > 0)  // added in V33.0.0.7 
         {
            sprintf(buffer,"DIAG Quilt quilting layer tile at x/y chart_width/height=%d %d offset=%d %d filename=%s",chart_width,chart_height,x_offset,y_offset,name);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }
          
         // Transfer all the pixels that have the alpha value not equal 0 from the input overlay tile to output gif
         long x,y,ox,oy;
      	for(y=0,oy=y_offset; y<chart_height; y++,oy++)
         {
            if (oy>=gif_height_pixels) break;     // added in V27.0.0.6
      		for(x=0,ox=x_offset; x<chart_width; x++,ox++)
            {          
               // the image is flipped so get the flipped y 
               long yflip = loadImage->GetHeight() - y - 1;
               // get the color  and the alpha value 
		         RGBQUAD rgb=loadImage->GetPixelColor(x,yflip);
		         if (loadImage->AlphaIsValid()) rgb.rgbReserved=loadImage->AlphaGet(x,yflip);
		         else rgb.rgbReserved=loadImage->GetPaletteColor(loadImage->GetPixelIndex(x,yflip)).rgbReserved;
               // if alphha is not zero 
               if(rgb.rgbReserved != 0)     
               {    
                  // if the ox >= if_width_pixels transfer the pixel  break added in V27.0.0.6
                  if (ox>=gif_width_pixels) break;    // added in V27.0.0.6
                  // set the pixel in the GIF the same as the tile color
                  newImage->SetPixelColor(ox,oy,rgb,true);
                  newImage->AlphaSet(ox,oy,0);
               }
      		}
            if (ox>max_x) max_x = ox; // added in V27.0.0.6
         }  
         if (oy>max_y) max_y = oy; // added in V27.0.0.6  
         
         
         // Destroy the input image if one exists
         if (loadImage->IsValid()) 
         {
            rc = loadImage->Destroy(); // Destroy the image
             if(!rc)
            {
              sprintf(buffer,"Error: Quilt 3. Invalid rc from Destroy Image RC=%d\n ",rc);
              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              return 1;
            }	 
         }  
      }
   }   

   sprintf(buffer,"DIAG Quilt after pixel transfer max_x=%d max_y=%d gif_width=%d gif_height=%d",max_x, max_y, gif_width_pixels,gif_height_pixels);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
    
   // If outside_border is not set, check that the chart data spans the whole gif, if not turn on outside_border added in V27.0.0.7 
   if (outside_border == false)   // if not set already
   {  
      // determine the differenc added in V27.0.0.8
      int diff_x, diff_y;
      diff_x = abs(gif_width_pixels - max_x);
      diff_y = abs(gif_height_pixels - max_y);
      sprintf(buffer,"DIAG Quilt max_x=%d gif_width=%d diff_x=%d  max_y=%d gif_height=%d diff_y=%d",max_x,gif_width_pixels,diff_x,max_y, gif_height_pixels,diff_y);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      if ((diff_x > 3) || (diff_y > 4))   // change diff_x to > 3 in V34.0.0.3
      {
         outside_border = true;
         sprintf(buffer,"WARNING: Quilt chart data does not span entire gif outside_border was set for filename=%s",filename);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         // pass it back to the rexx routine
         dataobj = context->LogicalToObject(outside_border); 
         context->DirectoryPut(quilt_dir, dataobj, "OUTSIDE_BORDER"); 
      }
   }
   
   // Check again
   alpha_value = newImage->AlphaGet(0,0); // get the alpha value
   sprintf(buffer,"DIAG Quilt After alpha_value=%d outside_border=%d",alpha_value, outside_border);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   
   // If outside_border is not set, check there are no gaps in the chart data, if so turn on outside_border  added in V27.0.0.6
   if (outside_border == false) // if not set alreeady 
   {    
      logical_t all_empty;
      // make sure we don't scan off the image
      if(max_x>gif_width_pixels)  max_x=gif_width_pixels;
      if(max_y>gif_height_pixels) max_y=gif_height_pixels;
      
      // first find the first row with chart data   added in V27.0.0.6
      int first_x = -1;
      int first_y = 0;
      for(y=0; y<gif_height_pixels; y++) 
      { 
         for(x=0; x<gif_width_pixels; x++) 
         {
            alpha_value = newImage->AlphaGet(x,y); // get the alpha value
            if (alpha_value != 1)  // if the alpha value is not set then its chart data
            { 
               // Save the first xy
               first_x = x;
               first_y = y;
               break;
            }
         }
         if (first_x != -1) break;
      }
      
      // first scan all the rows
      sprintf(buffer,"DIAG Quilt First_y=%d max_x=%d max_y=%d gif_width=%d gif_height=%d scanning rows for gaps",first_y, max_x, max_y, gif_width_pixels,gif_height_pixels);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   	for(y=first_y; y<max_y; y++) // change to use max_y in V27.0.0.6
      {
         all_empty = true;
   		for(x=0; x<max_x; x++)    // change to use max_x in V27.0.0.6
         {            
            alpha_value = newImage->AlphaGet(x,y); // get the alpha value
            if (alpha_value != 1)  
            {
               all_empty = false;  // if its chart data then this col is OK 
               break;
            }  
         } 
         //   if there is no chart_data in this col set the outside_border      
         if (all_empty)
         {
            outside_border = true;
            sprintf(buffer,"WARNING: Quilt row=%d is empty, outside_border was set for filename=%s",y,filename);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            // pass it back to the rexx routine
            dataobj = context->LogicalToObject(outside_border); 
            context->DirectoryPut(quilt_dir, dataobj, "OUTSIDE_BORDER"); 
            break;
         }  
      }
       
      // Then scan all the cols
      for(x=1; x<max_x; x++)    // start col 1 changed in V27.0.0.9     change to use max_x in V27.0.0.6
      {
         all_empty = true;
   		for(y=0; y<max_y; y++)    // change to use max_y in V27.0.0.6
         {            
            alpha_value = newImage->AlphaGet(x,y); // get the alpha value
            if (alpha_value != 1)  
            {
               all_empty = false;  // if its chart data then this column is OK 
               break;
            }  
         }  
         //   if there is no chart_data in this column set the outside_border      
         if (all_empty)
         {
            outside_border = true;
            sprintf(buffer,"WARNING: Quilt col=%d is empty, outside_border was set for filename=%s",x,filename);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            // pass it back to the rexx routine
            dataobj = context->LogicalToObject(outside_border); 
            context->DirectoryPut(quilt_dir, dataobj, "OUTSIDE_BORDER"); 
            break;
         } 
      }
   }    
       
   // if we are not using the outside_border use the bloodhound algorithm
   if (outside_border == false && num_ovl_tiles == 0)  // added num_ovl_tiles == 0 in V31.0.2.7
   {
      // scan horizontally until we find the first vertical line
      int xl, xr, yt, yb, av_l, av_r, av_t, av_b, s_av_l, s_av_r, s_av_t, s_av_b, ply_x, ply_y;
      int first_x, first_y;
      bool going_down = true;
      bool going_right = true;
      bool av1 = true;
      bool av2 = true;
      
      int bps = 1; // number of border points added in V27.0.0.7
      first_x = -1;
      for(y=0; y<gif_height_pixels; y++) 
      { 
         for(x=0; x<gif_width_pixels; x++) 
         {
            alpha_value = newImage->AlphaGet(x,y); // get the alpha value
            if (alpha_value != 1)  // if the alpha value is not 
            { 
               // Create the first border bointS
               sprintf(buffer,"%d %d VS 1",x,y);
               rc = context->ArrayAppendString(ply_points, buffer, strlen(buffer)); 
               // Save the first xy
               first_x = x;
               first_y = y;
               ply_x = x;
               ply_y = y;
               break;
            }
         }
         if (first_x != -1) break;
      }
      sprintf(buffer,"DIAG Quilt #0 BP=%d first_x=%d first_y=%d ply_x=%d ply_y=%d av1=%d av2=%d going_down=%d going_right=%d",bps,first_x,first_y,ply_x,ply_y,av1,av2,going_down,going_right);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      
      // do until we get back to the first xy
      int lp = 0;
      for (lp=1; lp<1000; lp++)
      {
         //follow the vertical line until we hit a horizontal line or the top or bottom  
         x = ply_x;
         xl = x-1;
         xr = x+1;
         
         if(going_down) y =  ply_y +1;
         else y = ply_y -1;
             
         // get the start alpha values on either side
         if (xl>-1 && xl<gif_width_pixels) 
         { 
            s_av_l = newImage->AlphaGet(xl,y);
            av1 = true;
         }
         else 
         {
            s_av_l = 2;  // left of the edge
            av_l = 2;
            av1 = false;
         }
         if (xr>-1 && xr<gif_width_pixels) 
         { 
            s_av_r = newImage->AlphaGet(xr,y);
            av2 = true;
         }
         else 
         {
            s_av_r = 2;  // right of the edge
            av_r = 2;
            av2 = false;
         }
         sprintf(buffer,"DIAG Quilt #1 LP=%d x=%d y=%d ply_x=%d ply_y=%d xl=%d xr=%d av1=%d av2=%d s_av_l=%d av_l=%d s_av_r=%d av_r=%d ",lp, x, y,ply_x,ply_y,xl,xr,av1,av2,s_av_l,av_l,s_av_r,av_r);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         
         if(going_down)
         {
            sprintf(buffer,"DIAG Quilt following Vertical line GOING DOWN" );
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            for(; y<gif_height_pixels; y++) 
            {
               if (av1) av_l = newImage->AlphaGet(xl,y); // get the alpha value  to the left
               if (av2) av_r = newImage->AlphaGet(xr,y); // get the alpha value  to the right
               // if we hit a change or the bottom then its a new point
               if (av_l != s_av_l || av_r != s_av_r || y == gif_height_pixels-1)
               {
                  if (y == gif_height_pixels-1) ply_y = y;  // the bottom
                  else if (av_l != s_av_l && s_av_l == 0) ply_y=y-1;// Inside change
                  else if (av_r != s_av_r && s_av_r == 0) ply_y=y-1;// Inside change
                  else ply_y=y; // must be an outside change
                  ply_x=x;
                  break;  
               }
            }
            // determine which way the next horizontal line is going
            if (s_av_r != av_r) going_right = true;  // its going right
            else going_right = false;  // its going left
            if (y >= gif_height_pixels-1) going_right = true; // Force going to right if we have hit the bottom
            sprintf(buffer,"%d %d V %d",ply_x,ply_y,going_right); 
            rc = context->ArrayAppendString(ply_points, buffer, strlen(buffer));
            bps += 1;
            sprintf(buffer,"DIAG Quilt #2 BP=%d going_right=%d x=%d y=%d ply_x=%d ply_y=%d xl=%d xr=%d av1=%d av2=%d s_av_l=%d av_l=%d s_av_r=%d av_r=%d ",bps,going_right,x, y,ply_x,ply_y,xl,xr,av1,av2,s_av_l,av_l,s_av_r,av_r);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            if(going_right)
            {
               sprintf(buffer,"DIAG Quilt following Horizontal line GOING RIGHT");
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            }  
            else   // its Going LEFT
            {
               sprintf(buffer,"DIAG Quilt following Horizontal line GOING LEFT");
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            }  
         }
         else // vertical line is going up
         {
            sprintf(buffer,"DIAG Quilt following Vertical line GOING UP" );
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            // scan the rows 
            for(; y>-1; y--) 
            {
               if (av1) av_l = newImage->AlphaGet(xl,y); // get the alpha value  to the left
               if (av2) av_r = newImage->AlphaGet(xr,y); // get the alpha value  to the right
               // if we hit a change then its a new point
               if (av_l != s_av_l || av_r != s_av_r || y == 0)
               {
                  if (y == 0) ply_y=y; // if at the top 
                  else if (av_l != s_av_l && s_av_l == 0) ply_y=y+1;// Inside change
                  else if (av_r != s_av_r && s_av_r == 0) ply_y=y+1;// Inside change
                  else ply_y=y; // must be an outside change
                  ply_x=x;
                  break;  
               }
            }
            
            // determine which way the next horizontal line is going
            if (s_av_r != av_r) going_right = true;  // its going right
            else going_right = false;  // its going left
            if (y <= 0) going_right = false;; // Force going to left if we have hit the top
            sprintf(buffer,"%d %d V %d",ply_x,ply_y,going_right);
            rc = context->ArrayAppendString(ply_points, buffer, strlen(buffer));
            bps += 1;  
             
            sprintf(buffer,"DIAG Quilt #3 BP=%d going_right=%d x=%d y=%d ply_x=%d ply_y=%d xl=%d xr=%d av1=%d av2=%d s_av_l=%d av_l=%d s_av_r=%d av_r=%d ", bps,going_right,x, y,ply_x,ply_y,xl,xr,av1,av2,s_av_l,av_l,s_av_r,av_r);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }
         
         //follow the horizontal line until we hit a vertical line or the left or right side 
         y = ply_y;
         yt = y-1;
         yb = y+1;
         
         if(going_right) x = ply_x +1;
         else x = ply_x -1;
           
         // get the start alpha values above and below
         if (yt>-1 && yt<gif_height_pixels) 
         {
            s_av_t = newImage->AlphaGet(x,yt);
            av1 = true;
         }
         else 
         {
            s_av_t = 2;  // above the top
            av_t = 2; 
            av1 = false;
         }
         if (yb>-1 && yb<gif_height_pixels) 
         {
            s_av_b = newImage->AlphaGet(x,yb);
            av2 = true;
         }
         else 
         {
            s_av_b = 2;  // below the bottom
            av_b = 2;
            av2 = false;
         }
      
         sprintf(buffer,"DIAG Quilt #4 x=%d y=%d ply_x=%d ply_y=%d yt=%d yb=%d av1=%d av2=%d s_av_l=%d av_l=%d s_av_r=%d av_r=%d ",x, y,ply_x,ply_y,yt,yb,av1,av2,s_av_l,av_l,s_av_r,av_r);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         
         if(going_right)
         {
            for(; x<gif_width_pixels; x++) 
            {
               if (av1) av_t = newImage->AlphaGet(x,yt); // get the alpha value above
               if (av2) av_b = newImage->AlphaGet(x,yb); // get the alpha value below
               // if we hit a change or the bottom then its a new point
               if (av_t != s_av_t || av_b != s_av_b || x == gif_width_pixels-1)
               {
                  if (x == gif_width_pixels-1) ply_x=x; // at the right side
                  else if (av_t != s_av_t && s_av_t == 0) ply_x = x-1;   // Inside change
                  else if (av_b != s_av_b && s_av_b == 0) ply_x = x-1; // Inside change
                  else ply_x = x;  // outside change 
                  ply_y=y;
                  break;  
               }
            }
            // determine which way the next vertical line is going
            if (s_av_t != av_t) going_down = false;  // its going up
            else going_down = true;  // its going down
            if (x >= gif_width_pixels-1) going_down = false; // Force going up if we have hit the right side
            sprintf(buffer,"%d %d H %d",ply_x,ply_y,going_down);
            rc = context->ArrayAppendString(ply_points, buffer, strlen(buffer)); 
            bps += 1;
            
            sprintf(buffer,"DIAG Quilt #5 BP=%d going_down=%d x=%d y=%d ply_x=%d ply_y=%d yt=%d yb=%d av1=%d av2=%d s_av_t=%d av_t=%d s_av_b=%d av_b=%d going_down=%d",bps,going_down,x, y,ply_x,ply_y,yt,yb,av1,av2,s_av_t,av_t,s_av_b,av_b,going_down);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         }
         else // follow the horizontal going left
         {
            for(; x>-1; x--) 
            {
               if (av1) av_t = newImage->AlphaGet(x,yt); // get the alpha value above
               if (av2) av_b = newImage->AlphaGet(x,yb); // get the alpha value below
               // if we hit a change or the bottom then its a new point
               if (av_t != s_av_t || av_b != s_av_b || x == 0)
               {
                  if (x == 0) ply_x=x; // at the top
                  else if (av_t != s_av_t && s_av_t == 0) ply_x = x+1;   // Inside change
                  else if (av_b != s_av_b && s_av_b == 0) ply_x = x+1; // Inside change
                  else ply_x = x;  // outside change 
                  ply_y=y;
                  break;
               }
            }
            // determine which way the next vertical line is going
            if (s_av_t != av_t) going_down = false;  // its going up
            else going_down = true;  // its going down
            if (x <=0 ) going_down = true; // Force going down if we have hit the left side
            sprintf(buffer,"%d %d H %d",ply_x,ply_y,going_down);
            rc = context->ArrayAppendString(ply_points, buffer, strlen(buffer)); 
            bps += 1;
         
            sprintf(buffer,"DIAG Quilt #6 BP=%d going_down=%d x=%d y=%d ply_x=%d ply_y=%d yt=%d yb=%d av1=%d av2=%d s_av_t=%d av_t=%d s_av_b=%d av_b=%d",bps,going_down,x, y,ply_x,ply_y,yt,yb,av1,av2,s_av_t,av_t,s_av_b,av_b);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         } 
         
         lp += 1;
         if (lp > 1000) 
         {
            sprintf(buffer,"Error: Quilt too many ply_points LP=%d",lp );
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            return 1; 
         }
      
         sprintf(buffer,"DIAG Quilt end of loop=%d ply_x=%d ply_y=%d  going_right=%d going_down=%d",lp, ply_x,ply_y,going_right,going_down);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         
         // if we are back at the first point we are finished
         if (ply_x == first_x && ply_y == first_y)
         { 
            sprintf(buffer,"DIAG Quilt LP=%d FINISHED ply_x == first_x && ply_y == first_y",lp);  // added ,lp in V31
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            break;
         }  
      
      } // while (ply_x != first_x && ply_y != first_y);
   } 
   if (outside_border)  // Added in V32.0.0.1 to prevent red colors
   {  
      if ( bypass_red != 1 ) // if there is no bypass_red then set the last palette entry to red added in V28.0.1.3
      {
         // set the last palette entry to red 
         newImage->SetPaletteColor(126,250, 6, 6);
         sprintf(buffer,"DIAG Quilt using outside_border set last palette entry to red");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         
         // Make all the non data areas red that have the alpha set
         long x,y;
      	for(y=0; y<gif_height_pixels; y++)
         {
      		for(x=0; x<gif_width_pixels; x++)
            {            
               alpha_value = newImage->AlphaGet(x,y); // get the alpha value
               if (alpha_value == 1)  // if its set make this pixel red by changing its palette index to 126
               { 
                  newImage->SetPixelIndex(x,y,126); // Update the pixel with new index
               }
      		}
         }  
      }         
   }  
      
   // Flip it over again so it looks right
   newImage->Flip();  
   
   // try to Save  the image up to 10 times added in V33.0.0.7
   for(int lp=0; lp<10; lp++)
   {
      rc = newImage->Save(filename,typeout);	
      if (!rc)
      {
         if (lp == 9)
         {
            sprintf(buffer,"Error: Quilt Invalid rc from Save Image RC=%d\n ",rc);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            return 1;
         }
         sprintf(buffer,"Warning: Quilt Invalid rc from Save Image RC=%d\n ",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }
      else break; 
   }             

   /* Save the image
   rc = newImage->Save(filename,typeout);	
   if(!rc)
   {
     sprintf(buffer,"Error: Quilt Invalid rc from Save Image RC=%d\n ",rc);
     rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
     return 1;
   }
   */	    
   
   // Destroy the image if one exists
   if (newImage->IsValid()) 
   {
      newImage->AlphaDelete();	// added in V31
      rc = newImage->Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: Quilt 4. Invalid rc from Destroy Image RC=%d\n ",rc);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }	 
      sprintf(buffer,"DIAG Quilt destroyed newimage rc=%d\n",rc);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
   }   
     
   return 0;
}  

/**************************** Quilt_Blobs **************************
  Creates one large BMP from multiple mbtiles blob images 
 * Syntax:	rc = Quilt_Blobs(Quilt_Data, diag, diagmsgs)
 *
 * Params:	
 *			   Quilt_Data - directory of quilt information:
 *			   (bmp_width/height, chart width/height, degrees/pixel x/y 
 *			       array of images, array of the images start x/y offsets)
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages 
 *			   The following were parameters and now they are in the Quilt_data as flags
 *			   check_size   if = 1 then check size and return
 *			   bypass_red  - if = 1 then bypass setting last palette entry to red   added in V27.0.0.6
 * Return: the 0 if OK else 1  and if OK then the chartbits in the Quilt_Data
 *********************************************************************/
RexxRoutine3(int, Quilt_Blobs, RexxObjectPtr, Quilt_Data,  int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512];
   
   int rc,n,num_entries;      
   
   int32_t alpha_value;
   RexxArrayObject x_start_offsets, y_start_offsets, palette;
  
   RexxObjectPtr dataobj;
   RexxDirectoryObject quilt_dir;

   logical_t flag;
   
   quilt_dir = (RexxDirectoryObject)Quilt_Data;
   flag = context->IsDirectory(quilt_dir);
   if ( flag == 0 )
   {
      sprintf(buffer,"Error: Quilt_Blobs parameter 2 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }  
   
   CSTRING filename;
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "FILENAME"); 
   filename = context->ObjectToStringValue(dataobj);
   
   // printf("DEBUG filename=%s",filename);
   
   // See if filename is a valid image type
	TCHAR* extout = FindExtension(filename);
	_tcslwr(extout);
	int typeout = CxImage::GetTypeIdFromName(extout);
   if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"Error: Quilt_Blobs filename is an invalid image type.");
      // printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}  
   sprintf(buffer,"DIAG Quilt_Blobs filename:%s type=%d",filename,typeout);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));   
    
   //Get all necessary values that are not arrays
   int32_t gif_width_pixels, gif_height_pixels, x_offset, y_offset;
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "GIF_WIDTH_PIXELS");
   flag = context->ObjectToInt32(dataobj, &gif_width_pixels);
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "GIF_HEIGHT_PIXELS");
   flag = context->ObjectToInt32(dataobj, &gif_height_pixels);  
   int maxcolors; 
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "MAXCOLORS");
   flag = context->ObjectToInt32(dataobj, &maxcolors); 
   int transindex; 
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "TRANSINDEX");
   flag = context->ObjectToInt32(dataobj, &transindex); 
   
   // printf("DEBUG gif_width_pixels=%d height_pixels=%d",gif_width_pixels,gif_height_pixels);
                                                                  
   // added in V27.0.0.8
   sprintf(buffer,"DIAG Quilt_Blobs transindex=%d maxcolors=%d gif_width_pixels=%d gif_height_pixels=%d ",transindex,maxcolors,gif_width_pixels, gif_height_pixels);  
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
    
   // Create a new output gif image 
   CxImage *newImage = new CxImage();
   DWORD wBpp = 8;  // bits per pixel
   newImage->Create(gif_width_pixels,gif_height_pixels,wBpp,typeout);  // Change to 2 in V8.2.0.5      
   if (newImage == NULL)
   {
      sprintf(buffer,"Error: Quilt_Blobs Unable to create output file:%s",filename);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;  
   }      
   
   // Get the the images width and height
   int image_width, image_height;
   image_width = newImage->GetWidth();
   image_height = newImage->GetHeight();
   
   // Get the bpp
   int bpp;
   bpp  =  newImage->GetBpp(); // Get the bpp added in V27.0.0.6
   
   //  pass back the XRES and YRES
   int32_t resx, resy;
   resx = newImage->GetXDPI(); // Get X resolution
   resy = newImage->GetYDPI(); // Get Y resolution
      
   dataobj = context->Int32ToObject(resx); 
   context->DirectoryPut(quilt_dir, dataobj, "XRES"); 
   dataobj = context->Int32ToObject(resy); 
   context->DirectoryPut(quilt_dir, dataobj, "YRES");

   sprintf(buffer,"DIAG Quilt_Blobs Image Width/Height=%d %d Bpp=%d Xres=%d Yres=%d",image_width,image_height,bpp,resx, resy);  // bpp added in V27.0.0.6
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));

   int gif_size = gif_width_pixels * gif_height_pixels;
   // printf("image_width=%d gif_width_pixels=%d image_height=%d gif_height_pixels=%d gif_size=%d\n",image_width,gif_width_pixels,image_height,gif_height_pixels,gif_size);
    
   // If we did not create an image of the proper size then quit
   if ((image_width != gif_width_pixels) || (image_height != gif_height_pixels))
   {
      sprintf(buffer,"Error: Quilt_Blobs Unable to create output GIF of proper size should be:%d %d is:%d %d",gif_width_pixels,gif_height_pixels,image_width,image_height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1; 
   } 
   
   // Get pointer to the palette
   palette = (RexxArrayObject)context->DirectoryAt(quilt_dir, "PALETTE"); 
   num_entries = context->ArrayItems(palette);  // the number of palette entries  
   if (num_entries != 256) 
   {
      sprintf(buffer,"Error: Quilt_Blobs the number of palette entries is not 256.");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}  
   
   // set the whole image to fully transparent
   newImage->AlphaCreate();
   //newImage->AlphaClear();
   newImage->AlphaSet(1);
   newImage->SetTransIndex(transindex);  // Set the transindex
   alpha_value = newImage->AlphaGet(0,0); // get the alpha value
   sprintf(buffer,"DIAG Quilt_Blobs Before alpha_value=%d",alpha_value);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
    
   // Get pointer to the  xy_start_offsets array
   x_start_offsets = (RexxArrayObject)context->DirectoryAt(quilt_dir, "X_START_OFFSETS");  
   y_start_offsets = (RexxArrayObject)context->DirectoryAt(quilt_dir, "Y_START_OFFSETS");

   // Flip it over so x=0,y=0 are at upper right
   newImage->Flip();
   
   int numcolors;
   numcolors = newImage->GetNumColors();
   
   int palette_size;
   palette_size = newImage->GetPaletteSize(); 
   
   int idx = newImage->GetPixelIndex(0,0);
   // ULONG pixval;   
   RGBQUAD color; 
   color.rgbRed = 255; 
   color.rgbGreen = 255; 
   color.rgbBlue = 255; 
   color.rgbReserved;    
   
   RGBQUAD  WhiteColor;  
   WhiteColor.rgbRed = 255; 
   WhiteColor.rgbGreen = 255; 
   WhiteColor.rgbBlue = 255; 
   WhiteColor.rgbReserved = 0;    
   
   RGBQUAD BlackColor;
   BlackColor.rgbRed = 0; 
   BlackColor.rgbGreen = 0; 
   BlackColor.rgbBlue = 0; 
   BlackColor.rgbReserved = 0;  

   // Set all the palette to black
     
   // initialize the palette 
   uint32_t colorref;
   BYTE *p = (BYTE *)(uint32_t *)&colorref;
   BYTE red,green,blue;
   for(n=0 ; n<256 ; n++)
   {
      dataobj = context->ArrayAt(palette, n+1);
      flag = context->ObjectToUnsignedInt32(dataobj, &colorref);  
      // get the palette colors Note they are reversed format is rrggbb00
      red = p[0]; 
      green = p[1];  
      blue = p[2];      
      newImage->SetPaletteColor(n,red, green, blue);
   }
   
   // sets the maximum number of colors to use for  SetPixelColor	
   newImage->SetClrImportant(maxcolors); // added maxcolors in V6.0.2.1
   newImage->SetTransIndex(transindex);  // Set the transindex

   // Get the blobs array  
   RexxArrayObject blobs_array_rexx = (RexxArrayObject)context->DirectoryAt(quilt_dir, "BLOBS_ARRAY");  
   int num_blobs = context->ArrayItems(blobs_array_rexx); 

   // Get a pointer the blob_filespecs array
   RexxArrayObject blob_filespecs = (RexxArrayObject)context->DirectoryAt(quilt_dir, "BLOB_FILESPECS"); 
   
   // printf("DEBUG reading the blobs\n");  
      
   // Read each of the blob images                         
   int max_x = 0;  
   int max_y = 0;   
   
   // RexxBufferObject blob;
   // RexxBufferStringObject blob;
   // size_t sz;
   // byte * blob_array;
   // POINTER dataptr; 
   for(n=0 ; n<num_blobs ; n++)
   {   
      RexxStringObject blob = (RexxStringObject)context->ArrayAt(blobs_array_rexx, n+1);
      const char *dataptr = reinterpret_cast<const char*>( context->StringData(blob) );
      // dataptr = (POINTER)context->StringData(blob);    

      int dataLength = context->StringLength(blob);                                                                               
      
      /* KEEP write out the blob
      CSTRING  filespec;
      dataobj = context->ArrayAt(blob_filespecs, n+1);
      filespec = context->ObjectToStringValue(dataobj);  
   	FILE * out = fopen(filespec, "wb");
   	if (! out)
   	{
         sprintf(buffer,"ERROR: Quilt_Blobs Could not Open \"%s\"\n", filespec);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;	
      }	
      fwrite(dataptr, dataLength, 1, out); 
      fclose(out);
      // Load the image back
      CxImage *blobimage = new CxImage();
      rc = blobimage->Load(filespec,1);
      if (!rc)
      {
         sprintf(buffer,"Error: Quilt_Blobs Invalid rc from Load RC=%d file:%s",rc,filespec );
         // printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
      } 
      // delete the file
      if (remove(filespec) != 0) 
         printf("Unable to delete the file\n"); 
       */

      // Determine the image type from the blob's first two characters
      // printf("%d %d",dataptr[0],dataptr[1]);
      int type;
      if ((dataptr[0] == -1) & (dataptr[1] == -40)) type = 3;  //JPG
      else type = 4;  //PNG
      
      // printf("DEBUG blobnum=%d type=%d dataLength=%d \n",n,type,dataLength);
      
      // Build the image from the blob
      CxImage *blobimage = new CxImage();
      BOOL OK = blobimage->Decode((byte *)dataptr,dataLength,type);
      if(!OK)
      {
         sprintf(buffer,"ERROR: Quilt_Blobs Invalid rc from Decode\n");
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;  
      }  
      
      // Check if the image has an alpha layer
      // if(blobimage->AlphaIsValid()) printf("The image has a Alpha Channel\n");
        
      /* KEEP
      byte * image_bits = blobimage->GetBits(); 
      long pDib_size = blobimage->GetSize(); 
      printf("pDib_size=%d dataLength=%d\n",pDib_size,dataLength);
      // memcpy((byte *)pDib,dataptr,dataLength);	
      memcpy(image_bits,dataptr,dataLength);
      */
      
      dataobj = context->ArrayAt(x_start_offsets, n+1);
      flag = context->ObjectToInt32(dataobj, &x_offset);
      dataobj = context->ArrayAt(y_start_offsets, n+1);
      flag = context->ObjectToInt32(dataobj, &y_offset);

      // printf("DEBUG Quilt_Blobs image:%d dataLength=%d x/y offsets %d %d\n",n, dataLength, x_offset, y_offset); 
      
      // get the realwidth and height of the image added in V33.0.0.6
      int chart_width = blobimage->GetWidth();
      int chart_height = blobimage->GetHeight();
      
      if(diag > 0) // Added in V33.0.0.7  
      {
         sprintf(buffer,"DIAG Quilt_Blobs quilting map tile at x/y chart_width/height=%d %d offset=%d %d",chart_width,chart_height,x_offset,y_offset);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }      
      
      /* DEBUG
      alpha_value = newImage->AlphaGet(0,0); // get the alpha value
      printf("DEBUG ALPHA VALUE of 0,0 for %d is %d\n",n, alpha_value);
      */
      
      // blobimage->Flip(); 
       
      // printf("DEBUG Quilt_Blobs transferring pixels\n");
       
      // Transfer all the pixels from the blobimage to the newimage
      long x,y,ox,oy;
   	for(y=0,oy=y_offset; y<chart_height; y++,oy++)
      {
         if (oy>=gif_height_pixels) break;  // if the oy are greater than or equal than the gif height don't transfer the pixel  break added in V27.0.0.6
   		for(x=0,ox=x_offset; x<chart_width; x++,ox++)
         {
            if (ox>=gif_width_pixels) break; // if the ox are greater than or equal than the gif width don't transfer the pixel  break added in V27.0.0.6
            // Get pixel color
   			color = blobimage->GetPixelColor(x,y);
            int alpha_value = blobimage->AlphaGet(x,y);
            // printf("DEBUG ALPHA VALUE of %d at x/y=%d %d is %d\n",n,x,y,alpha_value);
            newImage->SetPixelColor(ox,oy,color,false); 
            newImage->AlphaSet(ox,oy,255); // make the alpha value 255 
   		}
         if (ox>max_x) max_x = ox; // added in V27.0.0.6
      }  
      if (oy>max_y) max_y = oy; // added in V27.0.0.6 
             
      // Destroy the input image if one exists
      if (blobimage->IsValid()) 
      {
         rc = blobimage->Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: Quilt_Blobs 2. Invalid rc from Destroy Image RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }	 
      }        
   }

   // Scan all the pixels and find all that are tranparent and change them to white and set their transindex
   long x,y;
	for(y=0; y<gif_height_pixels; y++)
   {
		for(x=0; x<gif_width_pixels; x++)
      {            
         alpha_value = newImage->AlphaGet(x,y); // get the alpha value
         if (alpha_value == 1)  // if its transparent then set the index to the transindex and make it white
         { 
            newImage->SetPixelColor(x,y,WhiteColor,false);  // And make it white
            newImage->SetPixelIndex(x,y,transindex); // Update the pixel with new index
         }
		}
   }
   
   newImage->Flip(); 
   
   // try to Save  the image up to 10 times added in V33.0.0.7
   for(int lp=0; lp<10; lp++)
   {
      rc = newImage->Save(filename,typeout);	
      if (!rc)
      {
         if (lp == 9)
         {
            sprintf(buffer,"Error: Quilt_Blobs Invalid rc from Save Image RC=%d\n ",rc);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            return 1;
         }
         sprintf(buffer,"Warning: Quilt_Blobs Invalid rc from Save Image RC=%d\n ",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }
      else break; 
   }             
   
   // Destroy the image if one exists
   if (newImage->IsValid()) 
   {
      newImage->AlphaDelete();	// added in V31
      rc = newImage->Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: Quilt_Blobs 4. Invalid rc from Destroy Image RC=%d\n ",rc);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }	 
      sprintf(buffer,"DIAG Quilt_Blobs destroyed newimage rc=%d\n",rc);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
   }   
     
   return 0;
}  
/**************************** Quilt_Blobs_2 **************************
  Creates one large BMP from multiple mbtiles blob images 
 * Syntax:	rc = Quilt_Blobs_2(Quilt_Data, diag, diagmsgs)
 * added in V10.0.9.2 for sqlite3 cache
 * Params:	
 *			   Quilt_Data - directory of quilt information:
 *			   (bmp_width/height, chart width/height, degrees/pixel x/y 
 *			       array of images, array of the images start x/y offsets)
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages 
 *			   The following were parameters and now they are in the Quilt_data as flags
 *			   check_size   if = 1 then check size and return
 *			   bypass_red  - if = 1 then bypass setting last palette entry to red   added in V27.0.0.6
 * Return: the 0 if OK else 1  and if OK then the chartbits in the Quilt_Data
 *********************************************************************/
RexxRoutine3(int, Quilt_Blobs_2, RexxObjectPtr, Quilt_Data,  int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512];
   
   int rc,n,num_entries;      
   
   int32_t alpha_value;
   RexxArrayObject x_start_offsets, y_start_offsets, palette;
  
   RexxObjectPtr dataobj;
   RexxDirectoryObject quilt_dir;

   logical_t flag;
   
   quilt_dir = (RexxDirectoryObject)Quilt_Data;
   flag = context->IsDirectory(quilt_dir);
   if ( flag == 0 )
   {
      sprintf(buffer,"Error: Quilt_Blobs_2 parameter 2 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }  
   
   CSTRING filename;
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "FILENAME"); 
   filename = context->ObjectToStringValue(dataobj);
   
   // See if filename is a valid image type
	TCHAR* extout = FindExtension(filename);
	_tcslwr(extout);
	int typeout = CxImage::GetTypeIdFromName(extout);
   if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"Error: Quilt_Blobs_2 filename is an invalid image type.");
      // printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}  
   sprintf(buffer,"DIAG Quilt_Blobs_2 filename:%s type=%d",filename,typeout);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));   
    
   //Get all necessary values that are not arrays
   int32_t gif_width_pixels, gif_height_pixels, x_offset, y_offset;
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "GIF_WIDTH_PIXELS");
   flag = context->ObjectToInt32(dataobj, &gif_width_pixels);
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "GIF_HEIGHT_PIXELS");
   flag = context->ObjectToInt32(dataobj, &gif_height_pixels);  
   int maxcolors; 
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "MAXCOLORS");
   flag = context->ObjectToInt32(dataobj, &maxcolors); 
   int transindex; 
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "TRANSINDEX");
   flag = context->ObjectToInt32(dataobj, &transindex); 
                                                                  
   // added in V27.0.0.8
   sprintf(buffer,"DIAG Quilt_Blobs_2 transindex=%d maxcolors=%d gif_width_pixels=%d gif_height_pixels=%d ",transindex,maxcolors,gif_width_pixels, gif_height_pixels);  
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
    
   // Create a new output gif image 
   CxImage *newImage = new CxImage();
   DWORD wBpp = 8;  // bits per pixel
   newImage->Create(gif_width_pixels,gif_height_pixels,wBpp,typeout);  // Change to 2 in V8.2.0.5      
   if (newImage == NULL)
   {
      sprintf(buffer,"Error: Quilt_Blobs_2 Unable to create output file:%s",filename);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;  
   }      
   
   // Get the the images width and height
   int image_width, image_height;
   image_width = newImage->GetWidth();
   image_height = newImage->GetHeight();
   
   // Get the bpp
   int bpp;
   bpp  =  newImage->GetBpp(); // Get the bpp added in V27.0.0.6
   
   //  pass back the XRES and YRES
   int32_t resx, resy;
   resx = newImage->GetXDPI(); // Get X resolution
   resy = newImage->GetYDPI(); // Get Y resolution
      
   dataobj = context->Int32ToObject(resx); 
   context->DirectoryPut(quilt_dir, dataobj, "XRES"); 
   dataobj = context->Int32ToObject(resy); 
   context->DirectoryPut(quilt_dir, dataobj, "YRES");

   sprintf(buffer,"DIAG Quilt_Blobs_2 Image Width/Height=%d %d Bpp=%d Xres=%d Yres=%d",image_width,image_height,bpp,resx, resy);  // bpp added in V27.0.0.6
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));

   int gif_size = gif_width_pixels * gif_height_pixels;
   // printf("image_width=%d gif_width_pixels=%d image_height=%d gif_height_pixels=%d gif_size=%d\n",image_width,gif_width_pixels,image_height,gif_height_pixels,gif_size);
    
   // If we did not create an image of the proper size then quit
   if ((image_width != gif_width_pixels) || (image_height != gif_height_pixels))
   {
      sprintf(buffer,"Error: Quilt_Blobs_2 Unable to create output GIF of proper size should be:%d %d is:%d %d",gif_width_pixels,gif_height_pixels,image_width,image_height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1; 
   } 
   
   // Get pointer to the palette
   palette = (RexxArrayObject)context->DirectoryAt(quilt_dir, "PALETTE"); 
   num_entries = context->ArrayItems(palette);  // the number of palette entries  
   if (num_entries != 256) 
   {
      sprintf(buffer,"Error: Quilt_Blobs_2 the number of palette entries is not 256.");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}  
   
   // set the whole image to fully transparent
   newImage->AlphaCreate();
   //newImage->AlphaClear();
   newImage->AlphaSet(1);
   newImage->SetTransIndex(transindex);  // Set the transindex
   alpha_value = newImage->AlphaGet(0,0); // get the alpha value
   sprintf(buffer,"DIAG Quilt_Blobs_2 Before alpha_value=%d",alpha_value);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
    
   // Get pointer to the  xy_start_offsets array
   x_start_offsets = (RexxArrayObject)context->DirectoryAt(quilt_dir, "X_START_OFFSETS");  
   y_start_offsets = (RexxArrayObject)context->DirectoryAt(quilt_dir, "Y_START_OFFSETS");

   // Flip it over so x=0,y=0 are at upper right
   newImage->Flip();
   
   int numcolors;
   numcolors = newImage->GetNumColors();
   
   int palette_size;
   palette_size = newImage->GetPaletteSize(); 
   
   int idx = newImage->GetPixelIndex(0,0);
   // ULONG pixval;   
   RGBQUAD color; 
   color.rgbRed = 255; 
   color.rgbGreen = 255; 
   color.rgbBlue = 255; 
   color.rgbReserved;    
   
   RGBQUAD  WhiteColor;  
   WhiteColor.rgbRed = 255; 
   WhiteColor.rgbGreen = 255; 
   WhiteColor.rgbBlue = 255; 
   WhiteColor.rgbReserved = 0;    
   
   RGBQUAD BlackColor;
   BlackColor.rgbRed = 0; 
   BlackColor.rgbGreen = 0; 
   BlackColor.rgbBlue = 0; 
   BlackColor.rgbReserved = 0;  

   // Set all the palette to black
     
   // initialize the palette 
   uint32_t colorref;
   BYTE *p = (BYTE *)(uint32_t *)&colorref;
   BYTE red,green,blue;
   for(n=0 ; n<256 ; n++)
   {
      dataobj = context->ArrayAt(palette, n+1);
      flag = context->ObjectToUnsignedInt32(dataobj, &colorref);  
      // get the palette colors Note they are reversed format is rrggbb00
      red = p[0]; 
      green = p[1];  
      blue = p[2];      
      newImage->SetPaletteColor(n,red, green, blue);
   }
   
   // sets the maximum number of colors to use for  SetPixelColor	
   newImage->SetClrImportant(maxcolors); // added maxcolors in V6.0.2.1
   newImage->SetTransIndex(transindex);  // Set the transindex

   // Get the blobs array  
   RexxArrayObject blobs_array_rexx = (RexxArrayObject)context->DirectoryAt(quilt_dir, "BLOBS_ARRAY");  
   int num_blobs = context->ArrayItems(blobs_array_rexx); 

   // Get a pointer the blob_filespecs array
   RexxArrayObject blob_filespecs = (RexxArrayObject)context->DirectoryAt(quilt_dir, "BLOB_FILESPECS");   
      
   // Read each of the blob images                         
   int max_x = 0;  
   int max_y = 0;   
   
   // RexxBufferObject blob;
   // RexxBufferStringObject blob;
   // size_t sz;
   // byte * blob_array;
   // POINTER dataptr; 
   for(n=0 ; n<num_blobs ; n++)
   {   
      RexxStringObject blob = (RexxStringObject)context->ArrayAt(blobs_array_rexx, n+1);
      const char *dataptr = reinterpret_cast<const char*>( context->StringData(blob) );
      // dataptr = (POINTER)context->StringData(blob);    

      int dataLength = context->StringLength(blob);                                                                               
      
      /* KEEP write out the blob
      CSTRING  filespec;
      dataobj = context->ArrayAt(blob_filespecs, n+1);
      filespec = context->ObjectToStringValue(dataobj);  
   	FILE * out = fopen(filespec, "wb");
   	if (! out)
   	{
         sprintf(buffer,"ERROR: Quilt_Blobs_2 Could not Open \"%s\"\n", filespec);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;	
      }	
      fwrite(dataptr, dataLength, 1, out); 
      fclose(out);
      // Load the image back
      CxImage *blobimage = new CxImage();
      rc = blobimage->Load(filespec,1);
      if (!rc)
      {
         sprintf(buffer,"Error: Quilt_Blobs_2 Invalid rc from Load RC=%d file:%s",rc,filespec );
         // printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
      } 
      // delete the file
      if (remove(filespec) != 0) 
         printf("Unable to delete the file\n"); 
       */

      // Determine the image type from the blob's first two characters
      // printf("%d %d",dataptr[0],dataptr[1]);
      int type;
      if ((dataptr[0] == -1) & (dataptr[1] == -40)) type = 3;  //JPG
      else type = 4;  //PNG
      
      // Build the image from the blob
      CxImage *blobimage = new CxImage();
      BOOL OK = blobimage->Decode((byte *)dataptr,dataLength,type);
      if(!OK)
      {
         sprintf(buffer,"ERROR: Quilt_Blobs_2 Invalid rc from Decode\n");
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;  
      }  
      
      // Check if the image has an alpha layer
      // if(blobimage->AlphaIsValid()) printf("The image has a Alpha Channel\n");
        
      /* KEEP
      byte * image_bits = blobimage->GetBits(); 
      long pDib_size = blobimage->GetSize(); 
      printf("pDib_size=%d dataLength=%d\n",pDib_size,dataLength);
      // memcpy((byte *)pDib,dataptr,dataLength);	
      memcpy(image_bits,dataptr,dataLength);
      */
      
      dataobj = context->ArrayAt(x_start_offsets, n+1);
      flag = context->ObjectToInt32(dataobj, &x_offset);
      dataobj = context->ArrayAt(y_start_offsets, n+1);
      flag = context->ObjectToInt32(dataobj, &y_offset);

      // printf("DEBUG Quilt_Blobs_2 image:%d dataLength=%d x/y offsets %d %d\n",n, dataLength, x_offset, y_offset); 
      
      // get the realwidth and height of the image added in V33.0.0.6
      int chart_width = blobimage->GetWidth();
      int chart_height = blobimage->GetHeight();
      
      if(diag > 0) // Added in V33.0.0.7  
      {
         sprintf(buffer,"DIAG Quilt_Blobs_2 quilting map tile at x/y chart_width/height=%d %d offset=%d %d",chart_width,chart_height,x_offset,y_offset);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }      
      
      /* DEBUG
      alpha_value = newImage->AlphaGet(0,0); // get the alpha value
      printf("DEBUG ALPHA VALUE of 0,0 for %d is %d\n",n, alpha_value);
      */
      
      blobimage->Flip(); 
       
      // Transfer all the pixels from the blobimage to the newimage
      long x,y,ox,oy;
   	for(y=0,oy=y_offset; y<chart_height; y++,oy++)
      {
         if (oy>=gif_height_pixels) break;  // if the oy are greater than or equal than the gif height don't transfer the pixel  break added in V27.0.0.6
   		for(x=0,ox=x_offset; x<chart_width; x++,ox++)
         {
            if (ox>=gif_width_pixels) break; // if the ox are greater than or equal than the gif width don't transfer the pixel  break added in V27.0.0.6
            // Get pixel color
   			color = blobimage->GetPixelColor(x,y);
            int alpha_value = blobimage->AlphaGet(x,y);
            // printf("DEBUG ALPHA VALUE of %d at x/y=%d %d is %d\n",n,x,y,alpha_value);
            newImage->SetPixelColor(ox,oy,color,false); 
            newImage->AlphaSet(ox,oy,255); // make the alpha value 255 
   		}
         if (ox>max_x) max_x = ox; // added in V27.0.0.6
      }  
      if (oy>max_y) max_y = oy; // added in V27.0.0.6 
             
      // Destroy the input image if one exists
      if (blobimage->IsValid()) 
      {
         rc = blobimage->Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: Quilt_Blobs_2 2. Invalid rc from Destroy Image RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }	 
      }        
   }

   // Scan all the pixels and find all that are tranparent and change them to white and set their transindex
   long x,y;
	for(y=0; y<gif_height_pixels; y++)
   {
		for(x=0; x<gif_width_pixels; x++)
      {            
         alpha_value = newImage->AlphaGet(x,y); // get the alpha value
         if (alpha_value == 1)  // if its transparent then set the index to the transindex and make it white
         { 
            newImage->SetPixelColor(x,y,WhiteColor,false);  // And make it white
            newImage->SetPixelIndex(x,y,transindex); // Update the pixel with new index
         }
		}
   }
   
   newImage->Flip(); 
   
   // try to Save  the image up to 10 times added in V33.0.0.7
   for(int lp=0; lp<10; lp++)
   {
      rc = newImage->Save(filename,typeout);	
      if (!rc)
      {
         if (lp == 9)
         {
            sprintf(buffer,"Error: Quilt_Blobs_2 Invalid rc from Save Image RC=%d\n ",rc);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            return 1;
         }
         sprintf(buffer,"Warning: Quilt_Blobs_2 Invalid rc from Save Image RC=%d\n ",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }
      else break; 
   }             
   
   // Destroy the image if one exists
   if (newImage->IsValid()) 
   {
      newImage->AlphaDelete();	// added in V31
      rc = newImage->Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: Quilt_Blobs_2 4. Invalid rc from Destroy Image RC=%d\n ",rc);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }	 
      sprintf(buffer,"DIAG Quilt_Blobs_2 destroyed newimage rc=%d\n",rc);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
   }   
     
   return 0;
} 
  /**************************** Quilt_NV_Blobs **************************
  Creates one large BMP from multiple NV Blobs
 * Syntax:	rc = Quilt_NV_Blobs(Quilt_Data, diag, diagmsgs)
 *
 * Params:	
 *			   Quilt_Data - directory of quilt information:
 *			   (bmp_width/height, chart width/height, degrees/pixel x/y 
 *			       array of images, array of the images start x/y offsets)
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages 
 *			   The following were parameters and now they are in the Quilt_data as flags
 * Return: the 0 if OK else 1  and if OK then the chartbits in the Quilt_Data
 *********************************************************************/
RexxRoutine3(int, Quilt_NV_Blobs, RexxObjectPtr, Quilt_Data,  int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512];
   
   int rc,n,num_entries;      
   
   int32_t alpha_value;
   RexxArrayObject x_start_offsets, y_start_offsets, palette;
  
   RexxObjectPtr dataobj;
   RexxDirectoryObject quilt_dir;

   logical_t flag;
   
   quilt_dir = (RexxDirectoryObject)Quilt_Data;
   flag = context->IsDirectory(quilt_dir);
   if ( flag == 0 )
   {
      sprintf(buffer,"Error: Quilt_NV_Blobs parameter 2 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }  
   
   CSTRING filename;
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "FILENAME"); 
   filename = context->ObjectToStringValue(dataobj);
   
   // See if filename is a valid image type
	TCHAR* extout = FindExtension(filename);
	_tcslwr(extout);
	int typeout = CxImage::GetTypeIdFromName(extout);
   if (typeout == CXIMAGE_FORMAT_UNKNOWN) 
   {
      sprintf(buffer,"Error: Quilt_NV_Blobs filename is an invalid image type.");
      // printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}  
   sprintf(buffer,"DIAG Quilt_NV_Blobs filename:%s type=%d",filename,typeout);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));   
    
   //Get all necessary values that are not arrays
   int32_t gif_width_pixels, gif_height_pixels, x_offset, y_offset;
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "GIF_WIDTH_PIXELS");
   flag = context->ObjectToInt32(dataobj, &gif_width_pixels);
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "GIF_HEIGHT_PIXELS");
   flag = context->ObjectToInt32(dataobj, &gif_height_pixels);  
   int maxcolors; 
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "MAXCOLORS");
   flag = context->ObjectToInt32(dataobj, &maxcolors); 
   int transindex; 
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "TRANSINDEX");
   flag = context->ObjectToInt32(dataobj, &transindex); 
   logical_t rotate180; 
   dataobj = (RexxObjectPtr)context->DirectoryAt(quilt_dir, "ROTATE180");
   flag = context->ObjectToLogical(dataobj, &rotate180);
   
                                                                  
   // added in V27.0.0.8
   sprintf(buffer,"DIAG Quilt_NV_Blobs transindex=%d maxcolors=%d gif_width_pixels=%d gif_height_pixels=%d ",transindex,maxcolors,gif_width_pixels, gif_height_pixels);  
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
    
   // Create a new output gif image 
   CxImage *newImage = new CxImage();
   // DEBUG DWORD wBpp = 8;  // bits per pixel
   DWORD wBpp = 24;  // bits per pixel
   newImage->Create(gif_width_pixels,gif_height_pixels,wBpp,typeout);  // Change to 2 in V8.2.0.5      
   if (newImage == NULL)
   {
      sprintf(buffer,"Error: Quilt_NV_Blobs Unable to create output file:%s",filename);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;  
   }      
   
   // Get the the images width and height
   int image_width, image_height;
   image_width = newImage->GetWidth();
   image_height = newImage->GetHeight();
   
   // Get the bpp
   int bpp;
   bpp  =  newImage->GetBpp(); // Get the bpp added in V27.0.0.6
   
   //  pass back the XRES and YRES
   int32_t resx, resy;
   resx = newImage->GetXDPI(); // Get X resolution
   resy = newImage->GetYDPI(); // Get Y resolution
      
   dataobj = context->Int32ToObject(resx); 
   context->DirectoryPut(quilt_dir, dataobj, "XRES"); 
   dataobj = context->Int32ToObject(resy); 
   context->DirectoryPut(quilt_dir, dataobj, "YRES");

   sprintf(buffer,"DIAG Quilt_NV_Blobs Image Width/Height=%d %d Bpp=%d Xres=%d Yres=%d",image_width,image_height,bpp,resx, resy);  // bpp added in V27.0.0.6
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));

   int gif_size = gif_width_pixels * gif_height_pixels;
   // printf("image_width=%d gif_width_pixels=%d image_height=%d gif_height_pixels=%d gif_size=%d\n",image_width,gif_width_pixels,image_height,gif_height_pixels,gif_size);
    
   // If we did not create an image of the proper size then quit
   if ((image_width != gif_width_pixels) || (image_height != gif_height_pixels))
   {
      sprintf(buffer,"Error: Quilt_NV_Blobs Unable to create output GIF of proper size should be:%d %d is:%d %d",gif_width_pixels,gif_height_pixels,image_width,image_height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1; 
   } 
   
   // Get pointer to the palette
   palette = (RexxArrayObject)context->DirectoryAt(quilt_dir, "PALETTE"); 
   num_entries = context->ArrayItems(palette);  // the number of palette entries  
   if (num_entries != 256) 
   {
      sprintf(buffer,"Error: Quilt_NV_Blobs the number of palette entries is not 256.");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
	}  
   
   // set the whole image to fully transparent
   newImage->AlphaCreate();
   //newImage->AlphaClear();
   newImage->AlphaSet(0);
   newImage->SetTransIndex(transindex);  // Set the transindex
   alpha_value = newImage->AlphaGet(0,0); // get the alpha value
   sprintf(buffer,"DIAG Quilt_NV_Blobs Before alpha_value=%d",alpha_value);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
    
   // Get pointer to the  xy_start_offsets array
   x_start_offsets = (RexxArrayObject)context->DirectoryAt(quilt_dir, "X_START_OFFSETS");  
   y_start_offsets = (RexxArrayObject)context->DirectoryAt(quilt_dir, "Y_START_OFFSETS");

   // Flip it over so x=0,y=0 are at upper right
   // if(flip == 1) 
   newImage->Flip();
   
   int numcolors;
   numcolors = newImage->GetNumColors();
   
   int palette_size;
   palette_size = newImage->GetPaletteSize(); 
   
   int idx = newImage->GetPixelIndex(0,0);
   // ULONG pixval;   
   RGBQUAD color; 
   color.rgbRed = 255; 
   color.rgbGreen = 255; 
   color.rgbBlue = 255; 
   color.rgbReserved;    
   
   RGBQUAD  WhiteColor;  
   WhiteColor.rgbRed = 255; 
   WhiteColor.rgbGreen = 255; 
   WhiteColor.rgbBlue = 255; 
   WhiteColor.rgbReserved = 0;    
   
   RGBQUAD BlackColor;
   BlackColor.rgbRed = 0; 
   BlackColor.rgbGreen = 0; 
   BlackColor.rgbBlue = 0; 
   BlackColor.rgbReserved = 0;  

   // Set all the palette to black

   // Get the blobs array  
   RexxArrayObject blobs_array_rexx = (RexxArrayObject)context->DirectoryAt(quilt_dir, "BLOBS_ARRAY");  
   int num_blobs = context->ArrayItems(blobs_array_rexx); 

   // Get a pointer the blob_filespecs array
   RexxArrayObject blob_filespecs = (RexxArrayObject)context->DirectoryAt(quilt_dir, "BLOB_FILESPECS");   
      
   // Read each of the blob images                         
   int max_x = 0;  
   int max_y = 0;   
   
   // RexxBufferObject blob;
   // RexxBufferStringObject blob;
   // size_t sz;
   // byte * blob_array;
   // POINTER dataptr; 
   for(n=0 ; n<num_blobs ; n++)
   {   
      RexxStringObject blob = (RexxStringObject)context->ArrayAt(blobs_array_rexx, n+1);
      const char *dataptr = reinterpret_cast<const char*>( context->StringData(blob) );
      // dataptr = (POINTER)context->StringData(blob);    

      int dataLength = context->StringLength(blob);                                                                               
      
      /* KEEP write out the blob
      CSTRING  filespec;
      dataobj = context->ArrayAt(blob_filespecs, n+1);
      filespec = context->ObjectToStringValue(dataobj);  
   	FILE * out = fopen(filespec, "wb");
   	if (! out)
   	{
         sprintf(buffer,"ERROR: Quilt_NV_Blobs Could not Open \"%s\"\n", filespec);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;	
      }	
      fwrite(dataptr, dataLength, 1, out); 
      fclose(out);
      // Load the image back
      CxImage *blobimage = new CxImage();
      rc = blobimage->Load(filespec,1);
      if (!rc)
      {
         sprintf(buffer,"Error: Quilt_NV_Blobs Invalid rc from Load RC=%d file:%s",rc,filespec );
         // printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
      } 
      // delete the file
      if (remove(filespec) != 0) 
         printf("Unable to delete the file\n"); 
       */

      // Determine the image type from the blob's first two characters
      // printf("%d %d",dataptr[0],dataptr[1]);
      int type;
      if ((dataptr[0] == -1) & (dataptr[1] == -40)) type = 3;  //JPG
      else type = 4;  //PNG
      
      // Build the image from the blob
      CxImage *blobimage = new CxImage();
      BOOL OK = blobimage->Decode((byte *)dataptr,dataLength,type);
      if(!OK)
      {
         sprintf(buffer,"ERROR: Quilt_NV_Blobs Invalid rc from Decode\n");
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;  
      }  
      
      // Check if the image has an alpha layer
      // if(blobimage->AlphaIsValid()) printf("The image has a Alpha Channel\n");
        
      /* KEEP
      byte * image_bits = blobimage->GetBits(); 
      long pDib_size = blobimage->GetSize(); 
      printf("pDib_size=%d dataLength=%d\n",pDib_size,dataLength);
      // memcpy((byte *)pDib,dataptr,dataLength);	
      memcpy(image_bits,dataptr,dataLength);
      */
      
      dataobj = context->ArrayAt(x_start_offsets, n+1);
      flag = context->ObjectToInt32(dataobj, &x_offset);
      dataobj = context->ArrayAt(y_start_offsets, n+1);
      flag = context->ObjectToInt32(dataobj, &y_offset);

      // printf("DEBUG Quilt_NV_Blobs image:%d dataLength=%d x/y offsets %d %d\n",n, dataLength, x_offset, y_offset); 
      
      // get the realwidth and height of the image added in V33.0.0.6
      int chart_width = blobimage->GetWidth();
      int chart_height = blobimage->GetHeight();
      
      if(diag > 0) // Added in V33.0.0.7  
      {
         sprintf(buffer,"DIAG Quilt_NV_Blobs quilting map tile at x/y chart_width/height=%d %d offset=%d %d",chart_width,chart_height,x_offset,y_offset);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }      
       
      blobimage->Flip();
        
      // printf("DEBUG blob_num=%d chart_width/height=%d %d offset=%d %d\n",n, chart_width,chart_height,x_offset,y_offset);
       
      // Transfer all the pixels from the blobimage to the newimage
      long x,y,ox,oy;
   	for(y=0,oy=y_offset; y<chart_height; y++,oy++)
      {
         if (oy>=gif_height_pixels) break;  // if the oy are greater than or equal than the gif height don't transfer the pixel  break added in V27.0.0.6
   		for(x=0,ox=x_offset; x<chart_width; x++,ox++)
         {
            if (ox>=gif_width_pixels) break; // if the ox are greater than or equal than the gif width don't transfer the pixel  break added in V27.0.0.6
            
            // Get pixel color
   			color = blobimage->GetPixelColor(x,y);
            int alpha_value = blobimage->AlphaGet(x,y);
            if(alpha_value == 0); // if its transparent do nothing
            else // else transfer it and make it 255
            {
               // printf("DEBUG ALPHA VALUE of %d at x/y=%d %d is %d\n",n,x,y,alpha_value);
               newImage->SetPixelColor(ox,oy,color,false); 
               newImage->AlphaSet(ox,oy,255); // make the alpha value 255 
               // newImage->AlphaSet(ox,oy,0); // make the alpha value 255 
            }
   		}
         if (ox>max_x) max_x = ox; // added in V27.0.0.6
      }  
      if (oy>max_y) max_y = oy; // added in V27.0.0.6 
             
      // Destroy the input image if one exists
      if (blobimage->IsValid()) 
      {
         rc = blobimage->Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: Quilt_NV_Blobs 2. Invalid rc from Destroy Image RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }	 
      }        
   }
   
   // printf("DEBUG setting the transindex\n");
   
   // determine the extents of the inner polygon
   int x,y;
   int ul_x = 99999999;
   int ul_y = 99999999;
   int lr_x = 0;
   int lr_y = 0;
   
   // scan for the top to bottom left to right
	for(y=0; y<gif_height_pixels; y++)
   {
		for(x=0; x<gif_width_pixels; x++)
      {            
         alpha_value = newImage->AlphaGet(x,y); // get the alpha value
         /* DEBUG
         // if its NOT transparent 
         if(y > 31 && y < 33)
         {
            printf("x=%d y=%d alpha_value=%d\n",x,y,alpha_value);
         }
         */
         if (alpha_value == 255)  // if the alpha is set
         {   
            if(x < ul_x) ul_x= x;
            if(x > lr_x) lr_x= x;
         }
		}
   }
   
   // scan for the  left to right top to bottom
   for(x=0; x<gif_width_pixels; x++)
   {
		for(y=0; y<gif_height_pixels; y++)
      {            
         alpha_value = newImage->AlphaGet(x,y); // get the alpha value
         // if its NOT transparent 
         if (alpha_value ==  255)  // if the alpha is set
         {   
            if(y < ul_y) ul_y = y;
            if(y > lr_y) lr_y = y;
         }
		}
   }  
   
   // flip it back 
   newImage->Flip(); 
    
      
   // Crop the image
   float angle = 0;
   int newwidth, newheight;
   newwidth = lr_x - ul_x;
   newheight = lr_y - ul_y;
   
   sprintf(buffer,"DIAG: Quilt_NV_Blobs 2. UL=%d %d LR=%d %d newwidth=%d newheight=%d",ul_x,ul_y,lr_x,lr_y,newwidth,newheight);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   

   rc = newImage->CropRotatedRectangle( ul_x, ul_y, newwidth, newheight, angle);
   if (!rc)
   {
      sprintf(buffer,"ERROR: CropImage Invalid rc from crop image RC=%d\n",rc);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }	
   
   // if the rotae180 is true then do it
   if(rotate180)
   {
      // Rotate the image
      CxImage::InterpolationMethod intm;
   	CxImage::OverflowMethod overm;
   	intm=CxImage::IM_BILINEAR;
   	overm=CxImage::OM_BACKGROUND;
      float rt_angle = 180;
      rc = newImage->Rotate2( rt_angle, NULL, intm, overm, 0, 1, 1 );
      if (!rc)
      {
         sprintf(buffer,"ERROR: RotateImage failed RC=%d\n",rc);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
      }
   }
   
   // Get the the images width and height
   image_width = newImage->GetWidth();
   image_height = newImage->GetHeight();
   sprintf(buffer,"DIAG Quilt_NV_Blobs Image output Width/Height=%d %d",image_width,image_height);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   
   // pass back the xres yres image_width image_height
   dataobj = context->Int32ToObject(resx); 
   context->DirectoryPut(quilt_dir, dataobj, "XRES"); 
   dataobj = context->Int32ToObject(resy); 
   context->DirectoryPut(quilt_dir, dataobj, "YRES");
   dataobj = context->Int32ToObject(image_width); 
   context->DirectoryPut(quilt_dir, dataobj, "IMAGE_WIDTH");
   dataobj = context->Int32ToObject(image_height); 
   context->DirectoryPut(quilt_dir, dataobj, "IMAGE_HEIGHT");
   
   
   // try to Save  the image up to 10 times added in V33.0.0.7
   for(int lp=0; lp<10; lp++)
   {
      rc = newImage->Save(filename,typeout);	
      if (!rc)
      {
         if (lp == 9)
         {
            sprintf(buffer,"Error: Quilt_NV_Blobs Invalid rc from Save Image RC=%d\n ",rc);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            return 1;
         }
         sprintf(buffer,"Warning: Quilt_NV_Blobs Invalid rc from Save Image RC=%d\n ",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }
      else break; 
   }           
   
   // Destroy the image if one exists
   if (newImage->IsValid()) 
   {
      newImage->AlphaDelete();	// added in V31
      rc = newImage->Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: Quilt_NV_Blobs 4. Invalid rc from Destroy Image RC=%d\n ",rc);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }	 
      sprintf(buffer,"DIAG Quilt_NV_Blobs destroyed newimage rc=%d\n",rc);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));	
   }   
     
   return 0;
}   
/**************************** Tile_Histogram **************************
  Creates a Histogram from SasPlanet tiles
 * Syntax:	rc = Tile_Histogram(Tile_Data, diag, diagmsgs)
 *
 * Params:	
 *			   Tile_Data - directory of Tile_Histogram information:
 *			   an array containing all tile files 
 *			  an array to pass back the tile histogram.  Each entry in the count and the color value
 *			  A flag to indicate to use diffusion or not
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages 
 * Return: the 0 if OK else 1  and if OK then the histogram will be in the array
 *********************************************************************/
RexxRoutine3(int, Tile_Histogram, RexxObjectPtr, Tile_data,  int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512];
   
   int rc,num_tiles,n;                               
   logical_t flag, diffusion;

   RexxArrayObject tiles, histogram;
                        
   RexxObjectPtr dataobj;
   RexxDirectoryObject Tile_Info;
   
   // get a pointer to the tile data
   Tile_Info = (RexxDirectoryObject)Tile_data;
   flag = context->IsDirectory(Tile_Info);
   if ( flag == 0 )
   {
      sprintf(buffer,"Error: Tile_Histogram parameter 2 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }  

   // following added in V6.0.8.1
   int maxcolors; 
   dataobj = (RexxObjectPtr)context->DirectoryAt(Tile_Info, "MAXCOLORS");
   flag = context->ObjectToInt32(dataobj, &maxcolors); 
   if (maxcolors == NULL) maxcolors = 127; 
   
   // Get pointer to the  histogram rexx array
   histogram = (RexxArrayObject)context->DirectoryAt(Tile_Info, "HISTOGRAM");  
   
   // Get pointer to the  tile filespecs rexx array
   tiles = (RexxArrayObject)context->DirectoryAt(Tile_Info, "TILES");  
   num_tiles = context->ArrayItems(tiles);  // the number of  tiles 
   
   // Get pointer to the  overlay tile filespecs rexx array 
   int num_ovl_tiles;  
   RexxArrayObject ovl_tiles;
   ovl_tiles = (RexxArrayObject)context->DirectoryAt(Tile_Info, "OVL_TILES");  
   num_ovl_tiles = context->ArrayItems(ovl_tiles);  // the number of  tiles 
    
   sprintf(buffer,"DIAG Tile_Histogram num_tiles=%d num_ovl_tiles=%d ",num_tiles,num_ovl_tiles);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
  
   // Get a pointer to a flag to indicate if we should use diffusion
   dataobj = (RexxObjectPtr)context->DirectoryAt(Tile_Info, "DIFFUSION");
   diffusion = context->ObjectToLogical(dataobj, &diffusion);    
   
   // Create a array to hold the histogram  of all  images, each entry is 7 words long:       
   // color index,  color count, palette colors r,g,b, and pixval
   int palsize = maxcolors;   // changed to 256 in V6.0.8.1
	ULONG *pArray = (ULONG*) malloc(HISTSIZE*palsize*(num_tiles+num_ovl_tiles));
   
   sprintf(buffer,"DIAG Tile_histogram maxcolors=%d diffusion=%d num_tiles=%d num_ovl_tiles=%d",maxcolors, diffusion,num_tiles,num_ovl_tiles);
   // printf(buffer); printf("\n");
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   
   // Reset the array to zero  to do the image first
   memset(pArray,0,HISTSIZE*palsize*num_tiles);   
 
   // Read each of the tile images
   CSTRING  name;
   for(n=0 ; n<num_tiles ; n++)
   { 
      dataobj = context->ArrayAt(tiles, n+1);
      name = context->ObjectToStringValue(dataobj); 
      
      CxImage *loadImage = new CxImage();
               
      rc = loadImage->Load(name,1);
      if (!rc)
      {
         sprintf(buffer,"Error: Tile_Histogram Invalid rc from Load RC=%d file:%s",rc,name );
         // printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
      } 
      
      // reduce the palette to maxcolors
      int colors=maxcolors;   // changed to 256 in V6.0.8.1 was 127
      RGBQUAD * ppal;
      CQuantizer q(colors,4);     //  was 8
      q.ProcessImage(loadImage->GetDIB());           
      ppal=(RGBQUAD*)calloc(256*sizeof(RGBQUAD),1);
      q.SetColorTable(ppal);
      bool error_diffusion = (diffusion == 0)	? FALSE : TRUE;                
      rc = loadImage->DecreaseBpp(8,error_diffusion,ppal,colors); 
      
      // Create a histogram of the image 
      int i;
      RGBQUAD color;
      ULONG pixval;
   	// for(i=0; i<palsize; i++)
      for(i=0; i<maxcolors; i++)   // pass back maxcolors palette entries changed in V6.0.8.1 was 127
   	{
         color = loadImage->GetPaletteColor(i);
         pixval = Get_Pixval(color);           // added in V17.0.0.19
         pArray[i*HISTLEN] = i; // Save the index for this color
         pArray[i*HISTLEN+2] = pixval; // Save the pixval for this color
      }
      
      // Get the the images width and height
      int b_width, b_height;
      b_width = loadImage->GetWidth();
      b_height = loadImage->GetHeight();
         
      // Scan all the pixels and get the palette color counts 
      long x, y;
      uint32_t idx;
   	for(y=0; y<b_height; y++)
      {
   		for(x=0; x<b_width; x++)
         {
            // Get color index and increment the count for the index
   			idx = loadImage->GetPixelIndex(x,y);
            pArray[idx*HISTLEN+1]++;	// increment the histogram count for this color	
   		}
      }        
      
      // pass back the count and pixval in the histogram  changed in V27.0.0.14 to only pass the count and pixval
      // for(int i=0; i<palsize; i++)
      for(int i=0; i<maxcolors; i++)
      {
		   sprintf(buffer,"%010d %d",pArray[i*HISTLEN+1],pArray[i*HISTLEN+2]);    // change count to %010d in V30.0.0.10
         rc = context->ArrayAppendString(histogram, buffer, strlen(buffer));
      }                         
     
      // Destroy the input image if one exists
      if (loadImage->IsValid()) 
      {
         rc = loadImage->Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: Tile_Histogram 1. Invalid rc from Destroy Image RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }	 
      }   
   } 
   
   // if there any overlay images then add them too     
   if(num_ovl_tiles!=0)
   {
      // Read each of the overlay mages
      CSTRING  name;
      for(n=0 ; n<num_ovl_tiles ; n++)
      { 
         dataobj = context->ArrayAt(ovl_tiles, n+1);
         name = context->ObjectToStringValue(dataobj);
         
         // load the overlay tile 
         CxImage *loadImage = new CxImage();
         rc = loadImage->Load(name,1);
         if (!rc)
         {
            sprintf(buffer,"Error: Tile_Histogram Invalid rc from Load  overlay RC=%d file:%s",rc,name );
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            return 1;
         } 
         
         // reduce the palette to maxcolrs
         int colors=maxcolors;   // changed in V6.0.8.1 was 127
         RGBQUAD * ppal;
         CQuantizer q(colors,4);     //  was 8
         q.ProcessImage(loadImage->GetDIB());           
         ppal=(RGBQUAD*)calloc(256*sizeof(RGBQUAD),1);
         q.SetColorTable(ppal);
         bool error_diffusion = (diffusion == 0)	? FALSE : TRUE;                
         rc = loadImage->DecreaseBpp(8,error_diffusion,ppal,colors); 
         
         // Create a histogram of the image 
         int i;
         RGBQUAD color;
         ULONG pixval;
      	// for(i=0; i<palsize; i++)
         for(i=0; i<maxcolors; i++)   // pass back maxcolors palette entries  changed in V6.0.8.1 was 127
      	{
            color = loadImage->GetPaletteColor(i);
            pixval = Get_Pixval(color);           // added in V17.0.0.19
            pArray[i*HISTLEN] = i; // Save the index for this color
            pArray[i*HISTLEN+2] = pixval; // Save the pixval for this color
         }
         
         // Get the the images width and height
         int b_width, b_height;
         b_width = loadImage->GetWidth();
         b_height = loadImage->GetHeight();
            
         // Scan all the pixels and get the palette color counts 
         long x, y;
         uint32_t idx;
      	for(y=0; y<b_height; y++)
         {
      		for(x=0; x<b_width; x++)
            {
               // Get color index and increment the count for the index
      			idx = loadImage->GetPixelIndex(x,y);
               pArray[idx*HISTLEN+1]++;	// increment the histogram count for this color	
      		}
         }        
         
         // pass back the count and pixval in the histogram  changed in V27.0.0.14 to only pass the count and pixval
         // for(int i=0; i<palsize; i++)
         for(int i=0; i<maxcolors; i++)  // changed in V6.0.8.1
         {
   		   sprintf(buffer,"%010d %d",pArray[i*HISTLEN+1],pArray[i*HISTLEN+2]);  // changed to %010d in V30.0.0.11
            rc = context->ArrayAppendString(histogram, buffer, strlen(buffer));
         }                         
        
         // Destroy the input image if one exists
         if (loadImage->IsValid()) 
         {
            rc = loadImage->Destroy(); // Destroy the image
            if(!rc)
            {
              sprintf(buffer,"Error: Tile_Histogram 2. Invalid rc from Destroy Image RC=%d\n ",rc);
              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              return 1;
            }	 
         }             
      }
   }
   
   free(pArray);  // free the array printf()
   return 0;
}  
/**************************** Tile_Histogram_Blobs **************************
  Creates a Histogram from SasPlanet tile blobs
 * Syntax:	rc = Tile_Histogram_Blobs(Tile_Data, diag, diagmsgs)
 *
 * Params:	
 *			   Tile_Data - directory of Tile_Histogram_Blobs information:
 *			   an array containing all tile files 
 *			  an array to pass back the tile histogram.  Each entry in the count and the color value
 *			  A flag to indicate to use diffusion or not
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages 
 * Return: the 0 if OK else 1  and if OK then the histogram will be in the array
 *********************************************************************/
RexxRoutine3(int, Tile_Histogram_Blobs, RexxObjectPtr, Tile_data,  int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512];
   
   int rc,n;                               
   logical_t flag, diffusion;
                        
   RexxObjectPtr dataobj;
   RexxDirectoryObject Tile_Info;
   
   // get a pointer to the tile data
   Tile_Info = (RexxDirectoryObject)Tile_data;
   flag = context->IsDirectory(Tile_Info);
   if ( flag == 0 )
   {
      sprintf(buffer,"Error: Tile_Histogram_Blobs parameter 2 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }  

   // following added in V6.0.8.1
   int maxcolors; 
   dataobj = (RexxObjectPtr)context->DirectoryAt(Tile_Info, "MAXCOLORS");
   flag = context->ObjectToInt32(dataobj, &maxcolors); 
   if (maxcolors == NULL) maxcolors = 127; 
   
   // Get pointer to the  histogram rexx array
   RexxArrayObject histogram = (RexxArrayObject)context->DirectoryAt(Tile_Info, "HISTOGRAM");  
  
   // Get a pointer to a flag to indicate if we should use diffusion
   dataobj = (RexxObjectPtr)context->DirectoryAt(Tile_Info, "DIFFUSION");
   diffusion = context->ObjectToLogical(dataobj, &diffusion);    
   
   // Get the histogram blobs array  
   RexxArrayObject blobs_array_rexx = (RexxArrayObject)context->DirectoryAt(Tile_Info, "HISTOGRAM_BLOBS");  
   int num_blobs = context->ArrayItems(blobs_array_rexx);   

   /* Get the blobs lengths array  
   RexxArrayObject blobs_lengths_rexx = (RexxArrayObject)context->DirectoryAt(Tile_Info, "BLOB_LENGTHS"); 
   */ 

   // Get a pointer the Tile filespecs array
   RexxArrayObject tile_filespecs = (RexxArrayObject)context->DirectoryAt(Tile_Info, "TILE_FILESPECS");   

   // Create a array to hold the histogram  of all  images, each entry is 7 words long:       
   // color index,  color count, palette colors r,g,b, and pixval
   int palsize = maxcolors;   // changed to 256 in V6.0.8.1
	ULONG *pArray = (ULONG*) malloc(HISTSIZE*palsize*(num_blobs));
   
   sprintf(buffer,"DIAG Tile_Histogram_Blobs maxcolors=%d diffusion=%d num_blobs=%d",maxcolors, diffusion,num_blobs);
   // printf(buffer); printf("\n");
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   
   // Reset the array to zero  to do the image first
   memset(pArray,0,HISTSIZE*palsize*num_blobs);    
   // POINTER dataptr; 
   // Read each of the histogram blobs
   for(n=0 ; n<num_blobs ; n++)
   {   
      RexxStringObject blob = (RexxStringObject)context->ArrayAt(blobs_array_rexx, n+1);
      const char *dataptr = reinterpret_cast<const char*>( context->StringData(blob) );
      // dataptr = (POINTER)context->StringData(blob);

      //sz = context->BufferStringLength(blob); 
      int dataLength = context->StringLength(blob);

      // std::copy( pBuffer, pBuffer + blob_array.size(), &blob_array[0]);
      // blob_array  = (byte *)malloc(dataLength * sizeof(BYTE));
      // memcpy(blob_array,dataptr,dataLength);   
      
      // print some data
      // printf("%x %x %x %x %x %x %x %x \n", dataptr[0], dataptr[1], dataptr[2], dataptr[3], dataptr[4], dataptr[5],  dataptr[6], dataptr[7]); 
      // printf("%x %x %x %x %x %x %x %x \n", dataptr[24], dataptr[25], dataptr[26], dataptr[27], dataptr[28], dataptr[29],  dataptr[30], dataptr[31]); 

      /* KEEP  write out the blob
      CSTRING  filespec;
      dataobj = context->ArrayAt(tile_filespecs, n+1);
      filespec = context->ObjectToStringValue(dataobj);  
   	FILE * out = fopen(filespec, "wb");
   	if (! out)
   	{
         sprintf(buffer,"ERROR: Tile_Histogram_Blobs Could not Open \"%s\"\n", filespec);
         printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;	
      }	
      fwrite(dataptr, dataLength, 1, out); 
      fclose(out);

      // Load the image back
      CxImage *blobimage = new CxImage();
      rc = blobimage->Load(filespec,1);
      if (!rc)
      {
         sprintf(buffer,"Error: Tile_Histogram Invalid rc from Load RC=%d file:%s",rc,filespec );
         // printf(buffer);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
      } 
   
      // delete the file
      if (remove(filespec) != 0) 
         printf("Unable to delete the file\n"); 
      */

      // Determine the image type from the blob's first two characters
      int type;
      // printf("%d %d",dataptr[0],dataptr[1]);
      if ((dataptr[0] == -1) & (dataptr[1] == -40)) type = 3;  //JPG
      else type = 4;  //PNG
      
      // Build the image from the blob
      CxImage *blobimage = new CxImage();
      BOOL OK = blobimage->Decode((byte *)dataptr,dataLength,type);  
      
      byte * image_bits = blobimage->GetBits(); 
      // printf("Blob24-31 %x %x %x %x %x %x %x %x \n", image_bits[24], image_bits[25], image_bits[26], image_bits[27], image_bits[28], image_bits[29],  image_bits[30], image_bits[31]); 
     
      /* KEEP get the image information
      int chart_width = blobimage->GetWidth();
      int chart_height = blobimage->GetHeight();
      int bpp = blobimage->GetBpp(); 
      int type = blobimage->GetType();
      long xdpi = blobimage->GetXDPI();
      long ydpi = blobimage->GetYDPI();  
      int colortype = blobimage->GetColorType();   
      int isvalid = blobimage->IsValid();
      // Get the effdt
      HANDLE hImage = blobimage->GetDIB();
      BITMAPINFOHEADER ds;
   	memcpy(&ds,hImage, sizeof(ds));
   	int effwdt = ((((ds.biBitCount * ds.biWidth ) + 31) / 32) * 4);
      printf("blob#%d type=%d bpp=%d chart_width=%d chart_height=%d effwdt=%d xdpi=%d ydpi=%d colortype=%d isvalid=%d\n",n,type,bpp,chart_width,chart_height,effwdt,xdpi,ydpi,colortype,isvalid);
      */
     
      /* Save the image for diagnostic purpose  TODO COMMENT
      CSTRING  filespec;
      dataobj = context->ArrayAt(tile_filespecs, n+1);
      filespec = context->ObjectToStringValue(dataobj);  
      printf("IMAGE2 %x %x %x %x %x %x %x %x \n", image_bits[24], image_bits[25], image_bits[26], image_bits[27], image_bits[28], image_bits[29],  image_bits[30], image_bits[31]);
      rc = blobimage->Save(filespec,4);  // PNG	
      if(!rc)
      {
        sprintf(buffer,"ERROR: Tile_Histogam Invalid rc from Save Image RC=%d\n ",rc);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }	
      */    
              
      // reduce the palette to maxcolors
      int colors=maxcolors;   // changed to 256 in V6.0.8.1 was 127
      RGBQUAD * ppal;
      CQuantizer q(colors,4);     //  was 8
      q.ProcessImage(blobimage->GetDIB());           
      ppal=(RGBQUAD*)calloc(256*sizeof(RGBQUAD),1);
      q.SetColorTable(ppal);
      bool error_diffusion = (diffusion == 0)	? FALSE : TRUE;                
      rc = blobimage->DecreaseBpp(8,error_diffusion,ppal,colors); 
      
      // Create a histogram of the image 
      int i;
      RGBQUAD color;
      ULONG pixval;
   	// for(i=0; i<palsize; i++)
      for(i=0; i<maxcolors; i++)   // pass back maxcolors palette entries changed in V6.0.8.1 was 127
   	{
         color = blobimage->GetPaletteColor(i);
         pixval = Get_Pixval(color);           // added in V17.0.0.19
         pArray[i*HISTLEN] = i; // Save the index for this color
         pArray[i*HISTLEN+2] = pixval; // Save the pixval for this color
      }
      
      // Get the the images width and height
      int b_width, b_height;
      b_width = blobimage->GetWidth();
      b_height = blobimage->GetHeight();
         
      // Scan all the pixels and get the palette color counts 
      long x, y;
      uint32_t idx;
   	for(y=0; y<b_height; y++)
      {
   		for(x=0; x<b_width; x++)
         {
            // Get color index and increment the count for the index
   			idx = blobimage->GetPixelIndex(x,y);
            pArray[idx*HISTLEN+1]++;	// increment the histogram count for this color	
   		}
      }        
      
      // pass back the count and pixval in the histogram  changed in V27.0.0.14 to only pass the count and pixval
      // for(int i=0; i<palsize; i++)
      for(int i=0; i<maxcolors; i++)
      {
		   sprintf(buffer,"%010d %d",pArray[i*HISTLEN+1],pArray[i*HISTLEN+2]);    // change count to %010d in V30.0.0.10
         rc = context->ArrayAppendString(histogram, buffer, strlen(buffer));
      }                         
     
      // Destroy the input image if one exists
      if (blobimage->IsValid()) 
      {
         rc = blobimage->Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: Tile_Histogram_Blobs 1. Invalid rc from Destroy Image RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }	 
      }     
   }
   
   free(pArray);  // free the array printf()
   return 0;
}  
/**************************** Create_PNG_Blobs **************************
  Creates the PNG blob files used to create an Mbtiles data set
 * Syntax:	rc = Create_PNG_Blobs(Png_Data, diag, diagmsgs)
 *
 * Params:	
 *			   Png_Data - directory of Create_PNG_Blobs information:
 *			    -- an array containing all image files 
 *			    -- a directory to pass back the png blobs
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages 
 * Return: the 0 if OK else 1  and if OK then the histogram will be in the array
 *********************************************************************/
RexxRoutine3(int, Create_PNG_Blobs, RexxObjectPtr, Png_Data,  int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512];

   char png_name[100];  // buffer to create the png name
   
   int rc,n;                               
   logical_t flag;
                        
   RexxObjectPtr dataobj;
   RexxDirectoryObject Png_Info;
   
   // printf("DEBUG Create_PNG_Blobs\n");
   
   // get a pointer to the tile data
   Png_Info = (RexxDirectoryObject)Png_Data;
   flag = context->IsDirectory(Png_Info);
   if ( flag == 0 )
   {
      sprintf(buffer,"Error: Create_PNG_Blobs parameter 2 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }  
   
   // Get the png directory to store the blobs 
   RexxDirectoryObject png_blobs_rexx = (RexxDirectoryObject)context->DirectoryAt(Png_Info, "PNG_BLOBS"); 
   // Check its a directory 
   flag = context->IsDirectory(png_blobs_rexx);
   if ( flag == 0 )
   {
      sprintf(buffer,"ERROR: Create_PNG_Blobs PNG_BLOBS is not a directory !");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }  
   sprintf(buffer,"DIAG Create_PNG_Blobs found PNG_BLOBS directory");
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));   

   // Get the temporary directory, includes a backslash
   dataobj = (RexxObjectPtr)context->DirectoryAt(Png_Info, "TEMP_DIR"); 
   CSTRING temp_dir = context->ObjectToStringValue(dataobj);
   
   // Get the Unique tick time 
   dataobj = (RexxObjectPtr)context->DirectoryAt(Png_Info, "UNIQUE"); 
   CSTRING unique = context->ObjectToStringValue(dataobj);

   // Get pointers to the rexx arrays
   RexxArrayObject x_start_offsets = (RexxArrayObject)context->DirectoryAt(Png_Info, "X_START_OFFSETS");  
   RexxArrayObject y_start_offsets = (RexxArrayObject)context->DirectoryAt(Png_Info, "Y_START_OFFSETS");
   RexxArrayObject start_tilexs = (RexxArrayObject)context->DirectoryAt(Png_Info, "START_TILEX");  
   RexxArrayObject start_tileys = (RexxArrayObject)context->DirectoryAt(Png_Info, "START_TILEY");  
   RexxArrayObject tile_image_widths = (RexxArrayObject)context->DirectoryAt(Png_Info, "TILE_IMAGE_WIDTH");
   RexxArrayObject tile_image_heights = (RexxArrayObject)context->DirectoryAt(Png_Info, "TILE_IMAGE_HEIGHT");  
   RexxArrayObject new_widths = (RexxArrayObject)context->DirectoryAt(Png_Info, "NEW_WIDTHS");  
   RexxArrayObject new_heights = (RexxArrayObject)context->DirectoryAt(Png_Info, "NEW_HEIGHTS");  
       
   // Get pointer to image files array 
   RexxArrayObject image_files_array_rexx = (RexxArrayObject)context->DirectoryAt(Png_Info, "IMAGE_FILES"); 
   int num_images = context->ArrayItems(image_files_array_rexx); 
   
   sprintf(buffer,"DIAG: Create_PNG_Blobs num_images=%d",num_images);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   
   // printf("DEBUG Create_PNG_Blobs reading all image files num_images=%d \n", num_images);

   // Read each of the image bmp files
   int x_offset,y_offset,start_tilex,start_tiley,tile_image_width,tile_image_height,new_width,new_height;
   CSTRING filespec;
   CxImage *loadImage = new CxImage(); // Create the image only once
   for(n=0 ; n<num_images ; n++)
   {   
      //  get the name of image bmp file
      dataobj = context->ArrayAt(image_files_array_rexx, n+1);
      filespec = context->ObjectToStringValue(dataobj);
      
      // printf("DEBUG Create_PNG_Blobs filespec=%s\n",filespec); 
   
      // Load the bmp image file 
      rc = loadImage->Load(filespec,CXIMAGE_FORMAT_BMP);   
      if (!rc)
      {
         sprintf(buffer,"Error: Create_PNG_Blobs Invalid rc from Load bmp image RC=%d filespec:%s",rc,filespec );
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
      } 

      // Get all the x/y offsets and start_tile_x,start_tile_y for this file and the tile iamge width and height
      dataobj = context->ArrayAt(x_start_offsets, n+1);
      flag = context->ObjectToInt32(dataobj, &x_offset);
      dataobj = context->ArrayAt(y_start_offsets, n+1);
      flag = context->ObjectToInt32(dataobj, &y_offset); 
      dataobj = context->ArrayAt(start_tilexs, n+1);
      flag = context->ObjectToInt32(dataobj, &start_tilex);  
      dataobj = context->ArrayAt(start_tileys, n+1);
      flag = context->ObjectToInt32(dataobj, &start_tiley);
      dataobj = context->ArrayAt(tile_image_widths, n+1);
      flag = context->ObjectToInt32(dataobj, &tile_image_width); 
      dataobj = context->ArrayAt(tile_image_heights, n+1);
      flag = context->ObjectToInt32(dataobj, &tile_image_height); 
      dataobj = context->ArrayAt(new_widths, n+1);
      flag = context->ObjectToInt32(dataobj, &new_width); 
      dataobj = context->ArrayAt(new_heights, n+1);
      flag = context->ObjectToInt32(dataobj, &new_height); 
      
      sprintf(buffer,"DIAG Create_PNG_Blobs x/y_offsets=%d %d start_tilex/y=%d %d tile_image_width/height=%d %d new wdth/height=%d %d",x_offset,y_offset,start_tilex,start_tiley,tile_image_width,tile_image_height,new_width,new_height);
      // sprintf(buffer,"DIAG Create_PNG_Blobs x/y_offsets=%d %d start_tilex/y=%d %d tile_image_width/height=%d %d",x_offset,y_offset,start_tilex,start_tiley,tile_image_width,tile_image_height);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      
      // printf("DEBUG Create_PNG_Blobs x/y_offsets=%d %d start_tilex/y=%d %d tile_image_width/height=%d %d",x_offset,y_offset,start_tilex,start_tiley,tile_image_width,tile_image_height);    

      // if the new_width is zero bypass the resample, it was done by the caller added in V8.0.2.3
      if(new_width == 0 )
      {
         sprintf(buffer,"DIAG: Create_PNG_Blobs new_width is zero  ReSample bypassed");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }
      else
      { 
         // Expand the bmp image 
         // give a diag -- added in V11.0.1.1
         sprintf(buffer,"DIAG: Create_PNG_Blobs expanding image new_width=%d new_height=%d",new_width,new_height);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         
         rc = loadImage->Resample((long)new_width,(long)new_height);
         if (!rc)
         {
            sprintf(buffer,"ERROR: Create_PNG_Blobs Invalid rc from ReSample image RC=%d\n",rc);
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            return 1;
         }	 
      }  

      // Get the the new images width and height
      int b_width, b_height;
      b_width = loadImage->GetWidth();
      b_height = loadImage->GetHeight();
      // Get BPP
      int bpp = loadImage->GetBpp(); 
      sprintf(buffer,"DIAG Create_PNG_Blobs bpp=%d width/height=%d %d filespec=%s",bpp,b_width,b_height,filespec);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
      
      // printf("DEBUG Create_PNG_Blobs bpp=%d width/height=%d %d filespec=%s\n",bpp,b_width,b_height,filespec);

      // Flip it over
      loadImage->Flip();                

      // Create a new tile image
      CxImage *TileImage = new CxImage();
      DWORD wBpp = 24;
      TileImage->Create(tile_image_width,tile_image_height,wBpp,CXIMAGE_FORMAT_BMP);      
      if (TileImage == NULL)
      {
         sprintf(buffer,"Error: Create_PNG_Blobs unable to create tile image");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
         return 1;  
      }

      // Clear the image to white
      // TileImage->Clear(255);
      
      //Create an Alpha layer all pixels transparent
      TileImage->AlphaCreate();
      TileImage->AlphaSet(0);

      // Flip it over
      TileImage->Flip();
             
      /* KEEP
      byte * load_image_bits = loadImage->GetBits(); 
      byte * tile_image_bits = TileImage->GetBits(); 
      int total_bytes = b_width * b_height;
      int image_start_byte = x_offset + y_offset;
   	for(int z=0; z<total_bytes; z++)
      {	
         tile_image_bits[image_start_byte] = load_image_bits[z];
         image_start_byte += 1;  
      } 
      */ 

      // transfer all the pixel colors from the bmp to the tile image
      long x, y;
      RGBQUAD pixel;
      int start_x = x_offset; 
      for(y=0; y<b_height; y++)
      {
         x_offset = start_x;
   		for(x=0; x<b_width; x++)
         {
            pixel = loadImage->GetPixelColor(x,y,false);
            TileImage->SetPixelColor(x_offset, y_offset,pixel);
            TileImage->AlphaSet(x_offset,y_offset,255); // make the alpha value 255 opaque
            x_offset += 1;
            if(x_offset >= tile_image_width) break;
   		}
         y_offset += 1;
         if(y_offset >= tile_image_height) break;
      }     

      // Flip it over
      TileImage->Flip();

      // Create  PNGs from all the tiles and store them in the output directory
      int total_bytes = tile_image_width * tile_image_height;
      int num_tiles_x = tile_image_width / 256; 
      int num_tiles_y = tile_image_height / 256; 
      int topx = 0;
      int topy = 0;
      
      sprintf(buffer,"DIAG: Create_PNG_Blobs total_bytes=%d num_tiles_x=%d num_tiles_y=%d",total_bytes,num_tiles_x,num_tiles_y);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
    
      for(y=start_tiley; y<start_tiley + num_tiles_y; y++)
      {
   		for(x=start_tilex; x<start_tilex + num_tiles_x; x++)
         {
            // Build a PNG image for each tile
            CxImage *pngimage = new CxImage();
   
            // Split the tile image into individual tiles
            if(diag > 0)   
            {
               sprintf(buffer,"DIAG Create_PNG_Blobs cropping tile image topx/y=%d %d",topx,topy); 
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
            } 
            rc = TileImage->CropRotatedRectangle(topx, topy, 256, 256, 0 , pngimage);
            if (!rc)
            {
               sprintf(buffer,"ERROR: Create_PNG_Blobs Invalid rc from crop image RC=%d\n",rc);
               printf(buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               return 1;  
            }  

            /* KEEP   Encode2RGBA
            byte *pImage_RGBA = NULL;
            long size = 262144; // 256 x 256 X 4 (RGBA)
            rc = pngimage->Encode2RGBA(pImage_RGBA,size,false);
            if(!rc)
            {
               sprintf(buffer,"ERROR: Create_PNG_Blobs Invalid rc from Encode2RGBA RC=%d",rc);
               printf(buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               return 1;  
            }  
            if(pImage_RGBA)
            {
               pngimage->FreeMemory(pImage_RGBA);
            }
            */

            sprintf(png_name,"PNG_%d_%d.png",x,y); // Create a name using the tiles x/y

            // Check if the png already exists in the png_blobs_rexx
            RexxObjectPtr obj = context->DirectoryAt(png_blobs_rexx, png_name);
            if(obj !=  NULLOBJECT)
            { 
               sprintf(buffer,"DIAG Create_PNG_Blobs png %s already exists",png_name);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 

               // Retrieve the old png image
               RexxStringObject blob = (RexxStringObject)context->DirectoryAt(png_blobs_rexx, png_name);
               const char *dataptr = reinterpret_cast<const char*>( context->StringData(blob) );

               int dataLength = context->StringLength(blob); 

               // Build a png image from the blob
               CxImage *old_blobimage = new CxImage();
               BOOL OK = old_blobimage->Decode((byte *)dataptr,dataLength,CXIMAGE_FORMAT_PNG);
               if(!OK)
               {
                  sprintf(buffer,"ERROR: Create_PNG_Blobs Invalid rc from Decode\n");
                  printf(buffer);
                  rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                  return 1;  
               }  

               // transfer all pixels that are not transparent from the old blob to new
               // transfer all the pixel colors from the bmp to the tile image
               long x, y;
               RGBQUAD pixel;
               int alpha_value;
               for(y=0; y<b_height; y++)
               {
            		for(x=0; x<b_width; x++)
                  {
                     alpha_value = old_blobimage->AlphaGet(x,y);
                     // If its transparent forget it
                     if (alpha_value == 0) continue;  
                     pixel = old_blobimage->GetPixelColor(x,y,true);
                     pngimage->SetPixelColor(x, y,pixel,true);
            		}
               }    
               // Destroy the old_blobimage if one exists
               if (old_blobimage->IsValid()) 
               {
                  rc = old_blobimage->Destroy(); // Destroy the image
                  if(!rc)
                  {
                    sprintf(buffer,"Error: Create_PNG_Blobs 1. Invalid rc from Destroy old_blobimage RC=%d\n ",rc);
                    rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                    return 1;
                  }	 
               }   
            }  

            // Encode the pngimage into a buffer
            byte *pImage_RGB = NULL;
            // long size = 196608; // 256 x 256 X 3 (RGB)
            long size = 262144; // 256 x 256 X 4 (RGBA)
            rc = pngimage->Encode(pImage_RGB,size,CXIMAGE_FORMAT_PNG);
            if(!rc)
            {
               sprintf(buffer,"ERROR: Create_PNG_Blobs Invalid rc from Encode RC=%d",rc);
               printf(buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               return 1;  
            } 
           
            // Put the png buffer into the png_blobs_rexx
            RexxStringObject obj2 = context->NewString((CSTRING)pImage_RGB, size);
            context->DirectoryPut(png_blobs_rexx, obj2, png_name);  

            /* KEEP COMMENT KEEP write out the blob  for diagnostic purposes 
            // Create a name using the tiles x/y  added unique
            sprintf(buffer,"%sPNG_%d_%d_%s_.png",temp_dir,x,y,unique); 
         	FILE * out = fopen(buffer, "wb");
         	if (! out)
         	{
               sprintf(buffer,"ERROR: Create_PNG_Blobs Could not open file \n");
               printf(buffer);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               return 1;	
            }	
            fwrite(pImage_RGB, size, 1, out); 
            fclose(out);
             KEEP */

            // free the pImage_RGB
            if(pImage_RGB)
            {
               pngimage->FreeMemory(pImage_RGB);
            }

            /* save it to a file for diagnostic purposes only comment 
            sprintf(buffer,"%sPNG_%d_%d.png",temp_dir,x,y); // Create a name using the tiles x/y 
            rc = pngimage->Save(buffer,CXIMAGE_FORMAT_PNG);	
            if (!rc)
            {   
               sprintf(buffer,"Error: Create_PNG_Blobs Invalid rc from Save TileImage RC=%d\n ",rc);
               rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
               return 1;
            } 
            */

            // Destroy the png image if one exists
            if (pngimage->IsValid()) 
            {
               rc = pngimage->Destroy(); // Destroy the image
               if(!rc)
               {
                 sprintf(buffer,"Error: Create_PNG_Blobs 1. Invalid rc from Destroy pngimage RC=%d\n ",rc);
                 rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
                 return 1;
               }	 
            } 
            topx += 256;    
         }
         topy += 256;
         topx = 0; 
      }    
  

      /* KEEP For debugging comment after
      sprintf(buffer,"%sTILE_IMAGE_%d.bmp",temp_dir,n);
      rc = TileImage->Save(buffer,CXIMAGE_FORMAT_BMP);	
      if (!rc)
      {   
         sprintf(buffer,"Error: Create_PNG_Blobs Invalid rc from Save TileImage RC=%d\n ",rc);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         return 1;
      } 
      */   

      // Flip it over
      TileImage->Flip();          
     
      // Destroy the bmp input image if one exists
      if (loadImage->IsValid()) 
      {
         rc = loadImage->Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: Create_PNG_Blobs 1. Invalid rc from Destroy Image RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }	 
      }     

      // Destroy the tile image if one exists
      if (TileImage->IsValid()) 
      {
         rc = TileImage->Destroy(); // Destroy the image
         if(!rc)
         {
           sprintf(buffer,"Error: Create_PNG_Blobs 1. Invalid rc from Destroy TileImage RC=%d\n ",rc);
           rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
           return 1;
         }	 
      }     

   }
   return 0;
}  
/**************************** Get_Blob **************************
  Gets a Blob    added in V9.0.4.9
 * Syntax:	rc = Get_Blob(Dir, diag, diagmsgs)
 *
 * Params:	
 *			   Dir - directory of information:
 *			    -- the filesepc of the blob in the SasPlanet cache
 *			    -- the blob name in the form zm'_'row'_'column
 *			    return the blob in the dir entry "BLOB"
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages 
 * Return: the 0 if OK else 1  and if OK then the histogram will be in the array
 *********************************************************************/
RexxRoutine3(int, Get_Blob, RexxObjectPtr, Dir,  int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512];
   
   int rc;                               
   logical_t flag;
                        
   RexxObjectPtr dataobj;
   
   // get a pointer to the directory
   RexxDirectoryObject Dir_Info = (RexxDirectoryObject)Dir;
   flag = context->IsDirectory(Dir_Info);
   if ( flag == 0 )
   {
      sprintf(buffer,"Error: Get_Blob parameter 1 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }  
   
   // Get the SasPlanet cache filespec
   dataobj = (RexxObjectPtr)context->DirectoryAt(Dir_Info, "FILESPEC"); 
   CSTRING filespec = context->ObjectToStringValue(dataobj);
   
   // Create a new image from the filespec     
   CxImage *loadImage = new CxImage();
   
   //  Determine the image type
	TCHAR* extin = FindExtension(filespec);	
	_tcslwr(extin);
	int typein = CxImage::GetTypeIdFromName(extin);
    
   rc = loadImage->Load(filespec,3);  // 3 is a JPG but this will also load PNGs     
   if (!rc)
   {
      sprintf(buffer,"Error: Get_Blob Invalid rc from Load map tile RC=%d file:%s",rc,filespec);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }
   
   /* KEEP
   // Get the JPEQuality added in V9.0.4.11
   float jpegq = 0.0;
   if(typein == CXIMAGE_FORMAT_JPG)
   {
      jpegq = loadImage->GetJpegQualityF();     
      printf("DEBUG: JPeq Quality=%f",jpegq);
     
   }  AlphaIsValid
   */
   
   // Check if the image has an alpha layer
   // if(loadImage->AlphaIsValid()) printf("The image has a Alpha Channel\n");
                      
   // Encode the loadImage into a buffer
   byte *pImage_RGB = NULL;
   // long size = 196608; // 256 x 256 X 3 (RGB)
   long size = 1000000; // 256 x 256 X 4 (RGBA)  
   rc = loadImage->Encode(pImage_RGB,size,typein);  
   if(!rc)
   {
      sprintf(buffer,"ERROR: Get_Blob Invalid rc from Encode RC=%d",rc);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;  
   } 
   
   // Put the blob into the Dir as "BLOB" 
   RexxStringObject obj2 = context->NewString((CSTRING)pImage_RGB, size);
   context->DirectoryPut(Dir_Info, obj2, "BLOB");  
   context->ReleaseLocalReference(obj2); // added in V12.0.4.0

   // free the pImage_RGB
   if(pImage_RGB)
   {
      loadImage->FreeMemory(pImage_RGB);
   }
   
   // Destroy the loadImage if one exists
   if (loadImage->IsValid()) 
   {
      rc = loadImage->Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: Get_Blob 1. Invalid rc from Destroy loadImage RC=%d\n ",rc);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }	 
   } 
   
   sprintf(buffer,"DIAG Get_Blob returning");
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));

   return 0;
}                                        
/**************************** Merge_Tiles **************************
  Merges a Map and Layer filespec    added in V9.0.4.9
 * Syntax:	rc = Merge_Tiles(Dir, diag, diagmsgs)
 *
 * Params:	
 *			   Dir - directory of information:
 *			    -- the filesepc of the map tile in the SasPlanet cache
 *			    -- the filesepc of the layer tile in the SasPlanet cache
 *			    
 *			    return the merged blob in the dir entry "BLOB"
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages 
 * Return: the 0 if OK else 1  and if OK then the histogram will be in the array
 *********************************************************************/
RexxRoutine3(int, Merge_Tiles, RexxObjectPtr, Dir,  int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512];
   
   int rc;                               
   logical_t flag;
                        
   RexxObjectPtr dataobj;
   
   // get a pointer to the directory
   RexxDirectoryObject Dir_Info = (RexxDirectoryObject)Dir;
   flag = context->IsDirectory(Dir_Info);
   if ( flag == 0 )
   {
      sprintf(buffer,"Error: Merge_Tiles parameter 1 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }  
   
   // Get the SasPlanet cache Map filespec
   dataobj = (RexxObjectPtr)context->DirectoryAt(Dir_Info, "FILESPEC"); 
   CSTRING filespec = context->ObjectToStringValue(dataobj);
   
   // Get the SasPlanet cache layer filespec
   dataobj = (RexxObjectPtr)context->DirectoryAt(Dir_Info, "LAYER_FILESPEC"); 
   CSTRING layer_filespec = context->ObjectToStringValue(dataobj);
   
   // Create a new image from the map filespec     
   CxImage *map_Image = new CxImage();
   
   //  Determine the image type
	TCHAR* extin = FindExtension(filespec);	
	_tcslwr(extin);
	int map_type = CxImage::GetTypeIdFromName(extin);
    
   rc = map_Image->Load(filespec,map_type);     
   if (!rc)
   {
      sprintf(buffer,"Error: Merge_Tiles Invalid rc from Load map tile RC=%d file:%s",rc,filespec);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }
   
   // Check if the map_Image has an alpha channel
   bool map_has_alpha = map_Image->AlphaIsValid();
   
   /* KEEP
   // Create a transparent alpha layer for the map
   bool ac_rc = map_Image->AlphaCreate();
   if(ac_rc == false)
   {
      sprintf(buffer,"ERROR: Merge_Tiles Invalid rc from AlphaCreate RC=%d",ac_rc);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;  
   }   
   map_Image->AlphaSet(0);
   */
   
   // Create a new image from the layer filespec     
   CxImage *layer_Image = new CxImage();
   
   //  Determine the image type
	extin = FindExtension(layer_filespec);	
	_tcslwr(extin);
	int layer_type = CxImage::GetTypeIdFromName(extin);
    
   rc = layer_Image->Load(layer_filespec,layer_type);      
   if (!rc)
   {
      sprintf(buffer,"Error: Merge_Tiles Invalid rc from Load map tile RC=%d file:%s",rc,layer_filespec);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }
   
   // layer_Image->AlphaCreate();
   // layer_Image->AlphaSet(0)
   
   // Get the width and height of the layer
   int b_width = layer_Image->GetWidth();
   int b_height = layer_Image->GetHeight();
   // printf("DEBUG width/height=%d %d ",b_width,b_height);
   
   // check if the layer is completely white, if so forget merging added in V10.0.7.4
   long x, y;
   RGBQUAD pixel;
   double pixval;   
   bool all_white = true;
   for(y=0; y<b_height; y++)
   {
		for(x=0; x<b_width; x++)
      {      
         pixel = layer_Image->GetPixelColor(x,y,true);
         pixval = Get_Pixval(pixel);
         if(pixval == 16777215) continue; // if while forget it
         all_white = false;  // found a non-white pixel
         break;
		}
      if (all_white == false) break;
   }  
   
   // if not all white transfer the pixels
   if (all_white == false)
   {
      // Flip both images
      map_Image->Flip(); 
      layer_Image->Flip(); 
      
      // long x, y;
      
      // transfer all pixels that are not transparent from the layer_Image to the map_image
      for(y=0; y<b_height; y++)
      {
   		for(x=0; x<b_width; x++)
         {      
            pixel = layer_Image->GetPixelColor(x,y,true);
            pixval = Get_Pixval(pixel);
            if(pixval == 0) continue;  // if its black skip it 
            map_Image->SetPixelColor(x,y,pixel,true);   
   		}
      }  
      
      map_Image->Flip(); // flip the map back
   }
      
   // map_Image->AlphaStrip();  // Remove the alpha layer

   // Encode the map_image  into a buffer
   byte *pImage_RGB = NULL;
   // long size = 262144; // 256 x 256 X 4 (RGBA)
   long size = 1000000;
   bool enc_rc = map_Image->Encode(pImage_RGB,size,map_type); 
   if(enc_rc == false)
   {
      sprintf(buffer,"ERROR: Merge_Tiles Invalid rc from Encode RC=%d",enc_rc);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;  
   } 
   // else printf("DEBUG image is all white layer_filespec=%s",layer_filespec);
   
   // Put the blob into the Dir as "BLOB"
   RexxStringObject obj2 = context->NewString((CSTRING)pImage_RGB, size);
   context->DirectoryPut(Dir_Info, obj2, "BLOB");  
   
   // Free the  pImage_RGB buffer
   if(pImage_RGB)
   {
      map_Image->FreeMemory(pImage_RGB);
   }   

  
   // Destroy the map_Image if one exists
   if (map_Image->IsValid()) 
   {
      rc = map_Image->Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: Merge_Tiles 1. Invalid rc from Destroy map_Image RC=%d\n ",rc);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }	 
   } 
   
   // Destroy the layer_Image if one exists
   if (layer_Image->IsValid()) 
   {
      rc = layer_Image->Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: Merge_Tiles 1. Invalid rc from Destroy layer_Image RC=%d\n ",rc);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }	 
   } 
   sprintf(buffer,"DIAG Merge_Tiles returning 0");
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   return 0;
}                                         
/**************************** Merge_2_Blobs **************************
  Merges two blobs into one
 * Syntax:	rc = Merge_2_Blobs(Png_Data, diag, diagmsgs)
 *
 * Params:	
 *			   Png_Data - directory of Merge_2_Blobs information:
 *			    -- an array containing all image files 
 *			    -- an directory to pass back the png blobs
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages 
 * Return: the 0 if OK else 1  
 *********************************************************************/
RexxRoutine3(int, Merge_2_Blobs, RexxObjectPtr, Png_Data,  int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512];
   
   int rc;                               
   logical_t flag;
                        
   RexxObjectPtr dataobj;
   RexxDirectoryObject Png_Info;
   
   // get a pointer to the png data
   Png_Info = (RexxDirectoryObject)Png_Data;
   flag = context->IsDirectory(Png_Info);
   if ( flag == 0 )
   {
      sprintf(buffer,"Error: Merge_2_Blobs parameter 2 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }  
   
   // Get the temporary directory, includes a backslash
   dataobj = (RexxObjectPtr)context->DirectoryAt(Png_Info, "TEMP_DIR"); 
   CSTRING temp_dir = context->ObjectToStringValue(dataobj);
   
   // Get the destination filename
   dataobj = (RexxObjectPtr)context->DirectoryAt(Png_Info, "DEST_FILENAME"); 
   CSTRING dest_filename = context->ObjectToStringValue(dataobj);
   
   // Get the file type
   TCHAR* extin = FindExtension(dest_filename);
	_tcslwr(extin);
	int typein = CxImage::GetTypeIdFromName(extin);	
   
   sprintf(buffer,"DIAG: Merge_2_Blobs typein=%d",typein);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   
   // Get the Dest blob
   RexxObjectPtr obj = context->DirectoryAt(Png_Info, "DEST_BLOB");
   if(obj ==  NULLOBJECT)
   { 
      sprintf(buffer,"ERROR:  Merge_2_Blobs the DEST_BLOB does not exist !");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }                        
      
   // Retrieve the DEST_BLOB
   RexxStringObject dest_blob = (RexxStringObject)context->DirectoryAt(Png_Info, "DEST_BLOB");
   const char *dest_dataptr = reinterpret_cast<const char*>( context->StringData(dest_blob) );
   int dest_dataLength = context->StringLength(dest_blob);
   
   // Determine the dest_blob image type from the blob's first two characters
   // printf("%d %d",dataptr[0],dataptr[1]);
   int dest_type;
   if ((dest_dataptr[0] == -1) & (dest_dataptr[1] == -40)) dest_type = 3;  //JPG
   else dest_type = 4;  //PNG
   
   sprintf(buffer,"DIAG: Merge_2_Blobs type=%d",dest_type);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   // Build a png image from the dest blob
   CxImage *dest_blobimage = new CxImage();
   BOOL OK = dest_blobimage->Decode((byte *)dest_dataptr,dest_dataLength,dest_type);
   if(!OK)
   {
      sprintf(buffer,"ERROR: Merge_2_Blobs Invalid rc from Decode for dest blob \n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;  
   } 
   
   // if(dest_blobimage->AlphaIsValid()) printf("The dest_blobimage image has a Alpha Channel\n"); 
   
   dest_blobimage->Flip();  
   
   // Get the source blob
   RexxObjectPtr obj2 = context->DirectoryAt(Png_Info, "SOURCE_BLOB");
   if(obj ==  NULLOBJECT)
   { 
      sprintf(buffer,"ERROR: the SOURCE_BLOB does not exist !");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }  
     
   // Retrieve the SOURCE_BLOB
   RexxStringObject source_blob = (RexxStringObject)context->DirectoryAt(Png_Info, "SOURCE_BLOB");
   const char *dataptr_source = reinterpret_cast<const char*>( context->StringData(source_blob) );
   int source_dataLength = context->StringLength(source_blob); 
   
   // Determine the source_blob image type from the blob's first two characters
   // printf("%d %d",dataptr_source[0],dataptr_source[1]);
   int source_type;
   if ((dataptr_source[0] == -1) & (dataptr_source[1] == -40)) source_type = 3;  //JPG
   else source_type = 4;  //PNG
   
   // Build a png image from the source blob
   CxImage *source_blobimage = new CxImage();
   OK = source_blobimage->Decode((byte *)dataptr_source,source_dataLength,source_type);
   if(!OK)
   {
      sprintf(buffer,"ERROR: Merge_2_Blobs Invalid rc from Decode for source blob \n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;  
   }  
   
   // if(source_blobimage->AlphaIsValid()) printf("The source_blobimage image has a Alpha Channel\n");
   
   source_blobimage->Flip(); 
   
   // Get the width and height
   int b_width = source_blobimage->GetWidth();
   int b_height = source_blobimage->GetHeight();
   
   /* DEBUG
   // check if the source has alpha=255
   long x, y;
   RGBQUAD pixel;
   int alpha_value;
   BOOL has_opaque = false;
   for(y=0; y<b_height; y++)
   {
		for(x=0; x<b_width; x++)
      {
         alpha_value = source_blobimage->AlphaGet(x,y);
         // If its opaque
         if (alpha_value == 255) 
         {
            has_opaque = true; 
            break;
         }
		}
      if (has_opaque) break; 
   }  
   
   if(diag > 0)
   {
      sprintf(buffer,"DIAG: Merge_2_Blobs has_opaque=%d",has_opaque);     
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }
   
   // transfer all pixels that are not transparent from the source blob to the dest blob
   for(y=0; y<b_height; y++)
   {
		for(x=0; x<b_width; x++)
      {
         if (has_opaque == false) // if no opaques copy the pixel from source to dest
         {  
            pixel = source_blobimage->GetPixelColor(x,y,true);
            dest_blobimage->SetPixelColor(x, y,pixel,false);
            dest_blobimage->AlphaSet(x,y,255); 
         } 
         else  // the image has a mixture so test the alpha value
         {  
            alpha_value = source_blobimage->AlphaGet(x,y);
            // If its transparent forget it
            if (alpha_value == 0) continue;  
            pixel = source_blobimage->GetPixelColor(x,y,true);
            dest_blobimage->SetPixelColor(x, y,pixel,true);
         }      
		}
   }  
   */
   

   /* DEBUG
   // new in V10.0.2.14
   long x, y;
   RGBQUAD pixel;
   double pixval;  
   int alpha_value; 
   // transfer all pixels that are not transparent from the source_blobimage to the dest_blobimage
   for(y=0; y<b_height; y++)
   {
		for(x=0; x<b_width; x++)
      {      
         // alpha_value = source_blobimage->AlphaGet(x,y);
         // If its transparent forget it
         // if (alpha_value == 0) continue;  
         pixel = source_blobimage->GetPixelColor(x,y,true);
         pixval = Get_Pixval(pixel);
         if(pixval == 0) continue;  // if its white skip it 
         dest_blobimage->SetPixelColor(x, y,pixel,true);
         // dest_blobimage->AlphaSet(x,y,255); 
         // dest_blobimage->SetPixelColor(x,y,pixel,true);   
		}
   } 
   */
   
   // NEW  in V10.0.9.2
   // check if the layer is completely white, if so forget merging added in V10.0.7.4
   long x, y;
   RGBQUAD pixel;
   double pixval;   
   bool all_white = true;
   for(y=0; y<b_height; y++)
   {
		for(x=0; x<b_width; x++)
      {      
         pixel = source_blobimage->GetPixelColor(x,y,true);
         pixval = Get_Pixval(pixel);
         if(pixval == 16777215) continue; // if while forget it
         all_white = false;  // found a non-white pixel
         break;
		}
      if (all_white == false) break;
   }  
   
   // if not all white transfer the pixels
   if (all_white == false)
   {
      // Flip both images
      dest_blobimage->Flip(); 
      source_blobimage->Flip(); 
      
      // long x, y;
      
      // transfer all pixels that are not transparent from the layer_Image to the map_image
      for(y=0; y<b_height; y++)
      {
   		for(x=0; x<b_width; x++)
         {      
            pixel = source_blobimage->GetPixelColor(x,y,true);
            pixval = Get_Pixval(pixel);
            if(pixval == 0) continue;  // if its black skip it 
            dest_blobimage->SetPixelColor(x,y,pixel,true);  
            dest_blobimage->AlphaSet(x,y,255); // added in 11.0.1.1
   		}
      }  
      
      dest_blobimage->Flip(); // flip the dest_blobimage back
   }
   
   dest_blobimage->Flip(); 
   
   // dest_blobimage->AlphaStrip();  // Remove the alpha layer
   
   /* Save the dest_blobimage for diagnostics 
   sprintf(buffer,"%s_AFTER_UPDATING_%s",temp_dir,dest_filename); 
   rc = dest_blobimage->Save(buffer,typein);	
   if(!rc)
   {
     sprintf(buffer,"ERROR: Merge_2_Blobs Invalid rc from Save Image RC=%d\n ",rc);
     rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
     return 1;
   }	  
   */    

   // Encode the dest_image  into a buffer
   byte *pImage_RGB = NULL;
   // long size = 262144; // 256 x 256 X 4 (RGBA)
   long size = 1000000;    
   rc = dest_blobimage->Encode(pImage_RGB,size,dest_type); 
   if(!rc)
   {
      sprintf(buffer,"ERROR: Merge_2_Blobs Invalid rc from Encode RC=%d",rc);
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;  
   } 
   
   // Put the png buffer into the directory
   RexxStringObject obj3 = context->NewString((CSTRING)pImage_RGB, size);
   context->DirectoryPut(Png_Info, obj3, "DEST_BLOB_UPDATED"); 
   
   context->ReleaseLocalReference(obj3); // added in V12.0.3.3
   
   // Free the  pImage_RGB buffer
   if(pImage_RGB)
   {
      dest_blobimage->FreeMemory(pImage_RGB);
   } 
   
   /* KEEP write out the blob  for diagnostic purposes 
   sprintf(buffer,"%s_AFTER_%s",temp_dir,dest_filename); 
	FILE * out = fopen(buffer, "wb");
	if (! out)
	{
      sprintf(buffer,"ERROR: Merge_2_Blobs Could not open file \n");
      printf(buffer);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;	
   }	
   fwrite(pImage_RGB, size, 1, out); 
   fclose(out);
   */
     
   // Destroy the dest_blobimage if one exists
   if (dest_blobimage->IsValid()) 
   {
      rc = dest_blobimage->Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: Merge_2_Blobs 1. Invalid rc from Destroy dest_blobimage RC=%d\n ",rc);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }	 
   }       
   
   // Destroy the source_blobimage if one exists
   if (source_blobimage->IsValid()) 
   {
      rc = source_blobimage->Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: Merge_2_Blobs 1. Invalid rc from Destroy source_blobimage RC=%d\n ",rc);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }	 
   }  
   
   return 0;    
}  
/**************************** Create_From_Array **************************
  Creates a BMP from a matrix passed
 * Syntax:	rc = Create_From_Array(Png_Data, diag, diagmsgs)
 *
 * Params:	
 *			   Png_Data - directory of information:
 *			    -- the width, height,bpp,bytesperline,temporary directory, dest_filename
 *			    -- a matrix of GBR0 
 *			   diag    - if specified it will print diag information  
 *			   diagmsgs    - an array to hold diagnostic and error messages 
 * Return: the 0 if OK else 1  
 *********************************************************************/
RexxRoutine3(int, Create_From_Array, RexxObjectPtr, Png_Data,  int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512];
   
   int rc;                               
   logical_t flag;
                        
   RexxObjectPtr dataobj;
   RexxDirectoryObject Png_Info;
   
   // printf("DEBUG Create_From_Array Entered\n");
   
   // get a pointer to the png data
   Png_Info = (RexxDirectoryObject)Png_Data;
   flag = context->IsDirectory(Png_Info);
   if ( flag == 0 )
   {
      sprintf(buffer,"Error: Create_From_Array parameter 2 is not a directory!\n");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
   }  
   
   // Get the width, height and bytes per line
   int32_t width, height, Bytesperline,Bpp; 
   dataobj = (RexxObjectPtr)context->DirectoryAt(Png_Info, "WIDTH");
   flag = context->ObjectToInt32(dataobj, &width);
   dataobj = (RexxObjectPtr)context->DirectoryAt(Png_Info, "HEIGHT");
   flag = context->ObjectToInt32(dataobj, &height);
   dataobj = (RexxObjectPtr)context->DirectoryAt(Png_Info, "BYTESPERLINE");
   flag = context->ObjectToInt32(dataobj, &Bytesperline);
   dataobj = (RexxObjectPtr)context->DirectoryAt(Png_Info, "BPP");
   flag = context->ObjectToInt32(dataobj, &Bpp);
   
   int matrix_size = width*height*3;
   
   sprintf(buffer,"DIAG: Create_From_Array width=%d height=%d bytesperline=%d bpp=%d matrix_size=%d",width,height,Bytesperline,Bpp,matrix_size);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   // printf("DEBUG width=%d height=%d bytesperline=%d bpp=%d matrix_size=%d\n",width,height,Bytesperline,Bpp,matrix_size);
   
   // Get the temporary directory, includes a backslash
   dataobj = (RexxObjectPtr)context->DirectoryAt(Png_Info, "TEMP_DIR"); 
   CSTRING temp_dir = context->ObjectToStringValue(dataobj);
   
   // Get the destination filename
   dataobj = (RexxObjectPtr)context->DirectoryAt(Png_Info, "DEST_FILENAME"); 
   CSTRING dest_filename = context->ObjectToStringValue(dataobj);  
   
   // Get pointer to the matrix_array
   RexxArrayObject matrix_array = (RexxArrayObject)context->DirectoryAt(Png_Info, "MATRIX_ARRAY"); 
   int num_entries = context->ArrayItems(matrix_array);  // the number of matrix_array entries 
     
   sprintf(buffer,"DIAG: Create_From_Array the number of matrix_array=%d",num_entries);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   // printf("DEBUG matrix_array num_entries=%d\n",num_entries);   
   
   // if the number of entries in the array * 3 is greater than the matrix size make the matrix size = to the arra* 3
   if(num_entries * 3 > matrix_size)
   {
      matrix_size = num_entries * 3;
      
      sprintf(buffer,"DIAG: Create_From_Array increased the matrix_size=%d to array size * 3\n",matrix_size);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      // printf("DEBUG increased the matrix_size=%d to array size * 3\n",matrix_size); 
   }
   
   // Allocate a buffer for the matrix
   BYTE * pixelMatrix = (byte *)malloc(matrix_size); 
   
   // Transfer all the RGBS  into pixelMatrix
   uint32_t colorref;
   BYTE *p = (BYTE *)(uint32_t *)&colorref;
   uint32_t red,green,blue;
   int offset = 0;
   for(int n=0 ; n<num_entries ; n++)
   { 
      dataobj = context->ArrayAt(matrix_array, n+1);
      flag = context->ObjectToUnsignedInt32(dataobj, &colorref);
      blue = p[0];  
      green = p[1]; 
      red = p[2]; 
         
      pixelMatrix[offset] = blue;  
      pixelMatrix[offset+1] = green;  
      pixelMatrix[offset+2] = red; 
      
      /* DEBUG
      if(n<100 )  
      {
         printf("DEBUG  n==%d colorref=%u blue=%u green=%u red=%u \n",n,colorref, blue,green,red); 
         printf("DEBUG  n==%d pixelMatrix =%hhX %hhX %hhX\n",n, pixelMatrix[offset],pixelMatrix[offset+1],pixelMatrix[offset+2]); 
      } 
      */
       
      offset +=3;   
   }  
   
   // printf("DEBUG issuing the CreateFromMatrix\n");
    
   // Build a bmp image from the matrix
   CxImage *matrix_image = new CxImage();
   // BOOL OK = matrix_image->CreateFromMatrix(pixelMatrix,width,height,Bpp,Bytesperline,false); 
   BOOL OK = matrix_image->CreateFromArray(pixelMatrix,width,height,Bpp,Bytesperline,false);     
   if(!OK)
   {
      sprintf(buffer,"Error: Create_From_Array Unable to create matrix_image");
      return 1;  
   }      
 
   // matrix_image->Flip();  
  
   // printf("DEBUG saving to dest_filename=%s\n",dest_filename);
   rc = matrix_image->Save(dest_filename,CXIMAGE_FORMAT_BMP);	
   if (!rc)
   {  
      sprintf(buffer,"Error: Create_From_Array Invalid rc from Save Image RC=%d\n ",rc);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      return 1;
      
   }   
   
   free(pixelMatrix);
       
   // Destroy the matrix_image if one exists
   if (matrix_image->IsValid()) 
   {
      rc = matrix_image->Destroy(); // Destroy the image
      if(!rc)
      {
        sprintf(buffer,"Error: Create_From_Array 1. Invalid rc from Destroy matrix_image RC=%d\n ",rc);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }	 
   }       
   
   return 0;    
}  
/**************************** TileXYToQuadKey **************************
  Creates a quadkey from the  tile xy, levelofdetail
 * Syntax:	rc = TileXYToQuadKey(Data, diag, diagmsgs)
 *
 *    Data is a directory that contains the tilex, tiley, levelof detail (zoom -1) and 
 *       returns the quadkey in it.
 *       rc = 0 if OK, else 1
 *********************************************************************/
RexxRoutine3(int, TileXYToQuadKey, RexxObjectPtr, Data, int, diag, RexxArrayObject, diagmsgs)
{
   
   int rc;   
   logical_t flag;
   char buffer[512];
   
   RexxObjectPtr dataobj;
   RexxDirectoryObject data;
   data = (RexxDirectoryObject)Data;
   flag = context->IsDirectory(data);
   if ( flag == 0 )
   {
      return -1;
   }  
   
   int tileX, tileY, levelOfDetail; 
   
   dataobj = (RexxObjectPtr)context->DirectoryAt(data, "TILEX");
   flag = context->ObjectToInt32(dataobj, &tileX);
   dataobj = (RexxObjectPtr)context->DirectoryAt(data, "TILEY");
   flag = context->ObjectToInt32(dataobj, &tileY);
   dataobj = (RexxObjectPtr)context->DirectoryAt(data, "LEVELOFDETAIL");
   flag = context->ObjectToInt32(dataobj, &levelOfDetail);
   
   if ( diag>1 )
   {
      sprintf(buffer," DIAG TileXYToQuadKey tileX=%d tileY=%d levelOfDetail=%d",tileX, tileY,levelOfDetail);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
   }
   
   int number_of_quad_chars = 256;
   CHAR * quadchars;
   quadchars  = (char *)malloc(number_of_quad_chars * sizeof(CHAR));    
   int ofs = 0; 
   for (int i = levelOfDetail; i > 0; i--)
   {
       char digit = '0';
       int mask = 1 << (i - 1);
       if ((tileX & mask) != 0)
       {
           digit++;
       }
       if ((tileY & mask) != 0)
       {
           digit++;
           digit++;
       }
       quadchars[ofs] = digit;
       ofs += 1;
   }
   // pass back the quadkey
   quadchars[ofs] = '\0';  // delimit the string   
   
   if ( diag>1 )
   {
      sprintf(buffer," DIAG TileXYToQuadKey quadchars=%s",quadchars);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   } 
   dataobj = context->String(quadchars);
   context->DirectoryPut(data, dataobj, "QUADKEY");  
   
   free(quadchars);
   
   return 0;
}   
/**************************** QuadKeyToTileXY **************************
  Converts a quad key to its tile X/Y and  levelOfDetail
 * Syntax:	rc = QuadKeyToTileXY(Data, diag, diagmsgs)
 *
 *       The Data is a directory that contains the quadkey
 *       returns the tilex, tiley, levelof detail (zoom -1)
 *       rc = 0 if OK, else 1
 *********************************************************************/
RexxRoutine3(int, QuadKeyToTileXY, RexxObjectPtr, Data, int, diag, RexxArrayObject, diagmsgs)
{
   CSTRING quadKey;
   char buffer[512];
   
   int rc;  
   logical_t flag; 
  
   RexxObjectPtr dataobj;
   RexxDirectoryObject data;
   
   data = (RexxDirectoryObject)Data;
   flag = context->IsDirectory(data);
   if ( flag == 0 )
   {
      return -1;
   }  
   
   //double quaddbl;
   
   dataobj = (RexxObjectPtr)context->DirectoryAt(data, "QUADKEY");
   quadKey = context->ObjectToStringValue(dataobj);
   
   if ( diag>1 )
   {
      sprintf(buffer," DIAG QuadKeyToTileXY quadKey=%s",quadKey);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   }
   
   rc = 0;
   int levelOfDetail,tileX, tileY;
   tileX = tileY = 0;
   levelOfDetail = strlen(quadKey);
   for (int i = levelOfDetail; i > 0; i--)
   {
       int mask = 1 << (i - 1);
       switch (quadKey[levelOfDetail - i])
       {
           case '0':
               break;

           case '1':
               tileX |= mask;
               break;

           case '2':
               tileY |= mask;
               break;

           case '3':
               tileX |= mask;
               tileY |= mask;
               break;

           default:
              rc = -1;
              break;
       }
   }
   // pass baqck the data
   dataobj = context->Int32ToObject(tileX); 
   context->DirectoryPut(data, dataobj, "TILEX"); 
   dataobj = context->Int32ToObject(tileY); 
   context->DirectoryPut(data, dataobj, "TILEY"); 
   dataobj = context->Int32ToObject(levelOfDetail); 
   context->DirectoryPut(data, dataobj, "LEVELOFDETAIL"); 
   
   return rc;      
}
/**************************** GE_DeCrypt **************************
 GE_DeCrypts a GE JPG or DXT  
 * Syntax:	rc = GE_DeCrypt(Data, diag, diagmsgs)
 *
 *       The Data is a directory that contains an array with the JPG/DXT data to GE_DeCrypt
 *       from which a BMP is created.
 *       rc = 0 if OK, else 1
 *********************************************************************/
RexxRoutine3(int, GE_DeCrypt, RexxObjectPtr, Data, int, diag, RexxArrayObject, diagmsgs)
{
   using namespace std;
   int rc;   
   logical_t flag;
   char buffer[512];
   /*
   static const BYTE BMP_HEAD[54] = {
   // 0    1     2     3     4     5     6     7     8     9     A     B     C     D     E     F
	0x42, 0x4D, 0x36, 0x00, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x36, 0x00, 0x00, 0x00, 0x28, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x01, 0x00, 0x18, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
   */
   // Modified for RGBA 32bpp   size = 262198 = x'00040036'
   static const BYTE BMP_HEAD[54] = {
   // 0    1     2     3     4     5     6     7     8     9     A     B     C     D     E     F
	0x42, 0x4D, 0x36, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x36, 0x00, 0x00, 0x00, 0x28, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x01, 0x00, 0x20, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
   
   static const BYTE key[1024] = {
   0x94, 0x64, 0x87, 0x4E, 0x66, 0x00, 0x00, 0x42, 0x45, 0xF4, 0xBD, 0x0B, 0x79, 0xE2, 0x6A, 0x45,
	0x22, 0x05, 0x92, 0x2C, 0x17, 0xCD, 0x06, 0x71, 0xF8, 0x49, 0x10, 0x46, 0x67, 0x51, 0x00, 0x42,
	0x25, 0xC6, 0xE8, 0x61, 0x2C, 0x66, 0x29, 0x08, 0xC6, 0x34, 0xDC, 0x6A, 0x62, 0x25, 0x79, 0x0A,
	0x77, 0x1D, 0x6D, 0x69, 0xD6, 0xF0, 0x9C, 0x6B, 0x93, 0xA1, 0xBD, 0x4E, 0x75, 0xE0, 0x41, 0x04,
	0x5B, 0xDF, 0x40, 0x56, 0x0C, 0xD9, 0xBB, 0x72, 0x9B, 0x81, 0x7C, 0x10, 0x33, 0x53, 0xEE, 0x4F,
	0x6C, 0xD4, 0x71, 0x05, 0xB0, 0x7B, 0xC0, 0x7F, 0x45, 0x03, 0x56, 0x5A, 0xAD, 0x77, 0x55, 0x65,
	0x0B, 0x33, 0x92, 0x2A, 0xAC, 0x19, 0x6C, 0x35, 0x14, 0xC5, 0x1D, 0x30, 0x73, 0xF8, 0x33, 0x3E,
	0x6D, 0x46, 0x38, 0x4A, 0xB4, 0xDD, 0xF0, 0x2E, 0xDD, 0x17, 0x75, 0x16, 0xDA, 0x8C, 0x44, 0x74,
	0x22, 0x06, 0xFA, 0x61, 0x22, 0x0C, 0x33, 0x22, 0x53, 0x6F, 0xAF, 0x39, 0x44, 0x0B, 0x8C, 0x0E,
	0x39, 0xD9, 0x39, 0x13, 0x4C, 0xB9, 0xBF, 0x7F, 0xAB, 0x5C, 0x8C, 0x50, 0x5F, 0x9F, 0x22, 0x75,
	0x78, 0x1F, 0xE9, 0x07, 0x71, 0x91, 0x68, 0x3B, 0xC1, 0xC4, 0x9B, 0x7F, 0xF0, 0x3C, 0x56, 0x71,
	0x48, 0x82, 0x05, 0x27, 0x55, 0x66, 0x59, 0x4E, 0x65, 0x1D, 0x98, 0x75, 0xA3, 0x61, 0x46, 0x7D,
	0x61, 0x3F, 0x15, 0x41, 0x00, 0x9F, 0x14, 0x06, 0xD7, 0xB4, 0x34, 0x4D, 0xCE, 0x13, 0x87, 0x46,
	0xB0, 0x1A, 0xD5, 0x05, 0x1C, 0xB8, 0x8A, 0x27, 0x7B, 0x8B, 0xDC, 0x2B, 0xBB, 0x4D, 0x67, 0x30,
	0xC8, 0xD1, 0xF6, 0x5C, 0x8F, 0x50, 0xFA, 0x5B, 0x2F, 0x46, 0x9B, 0x6E, 0x35, 0x18, 0x2F, 0x27,
	0x43, 0x2E, 0xEB, 0x0A, 0x0C, 0x5E, 0x10, 0x05, 0x10, 0xA5, 0x73, 0x1B, 0x65, 0x34, 0xE5, 0x6C,
	0x2E, 0x6A, 0x43, 0x27, 0x63, 0x14, 0x23, 0x55, 0xA9, 0x3F, 0x71, 0x7B, 0x67, 0x43, 0x7D, 0x3A,
	0xAF, 0xCD, 0xE2, 0x54, 0x55, 0x9C, 0xFD, 0x4B, 0xC6, 0xE2, 0x9F, 0x2F, 0x28, 0xED, 0xCB, 0x5C,
	0xC6, 0x2D, 0x66, 0x07, 0x88, 0xA7, 0x3B, 0x2F, 0x18, 0x2A, 0x22, 0x4E, 0x0E, 0xB0, 0x6B, 0x2E,
	0xDD, 0x0D, 0x95, 0x7D, 0x7D, 0x47, 0xBA, 0x43, 0xB2, 0x11, 0xB2, 0x2B, 0x3E, 0x4D, 0xAA, 0x3E,
	0x7D, 0xE6, 0xCE, 0x49, 0x89, 0xC6, 0xE6, 0x78, 0x0C, 0x61, 0x31, 0x05, 0x2D, 0x01, 0xA4, 0x4F,
	0xA5, 0x7E, 0x71, 0x20, 0x88, 0xEC, 0x0D, 0x31, 0xE8, 0x4E, 0x0B, 0x00, 0x6E, 0x50, 0x68, 0x7D,
	0x17, 0x3D, 0x08, 0x0D, 0x17, 0x95, 0xA6, 0x6E, 0xA3, 0x68, 0x97, 0x24, 0x5B, 0x6B, 0xF3, 0x17,
	0x23, 0xF3, 0xB6, 0x73, 0xB3, 0x0D, 0x0B, 0x40, 0xC0, 0x9F, 0xD8, 0x04, 0x51, 0x5D, 0xFA, 0x1A,
	0x17, 0x22, 0x2E, 0x15, 0x6A, 0xDF, 0x49, 0x00, 0xB9, 0xA0, 0x77, 0x55, 0xC6, 0xEF, 0x10, 0x6A,
	0xBF, 0x7B, 0x47, 0x4C, 0x7F, 0x83, 0x17, 0x05, 0xEE, 0xDC, 0xDC, 0x46, 0x85, 0xA9, 0xAD, 0x53,
	0x07, 0x2B, 0x53, 0x34, 0x06, 0x07, 0xFF, 0x14, 0x94, 0x59, 0x19, 0x02, 0xE4, 0x38, 0xE8, 0x31, 
	0x83, 0x4E, 0xB9, 0x58, 0x46, 0x6B, 0xCB, 0x2D, 0x23, 0x86, 0x92, 0x70, 0x00, 0x35, 0x88, 0x22,
	0xCF, 0x31, 0xB2, 0x26, 0x2F, 0xE7, 0xC3, 0x75, 0x2D, 0x36, 0x2C, 0x72, 0x74, 0xB0, 0x23, 0x47,
	0xB7, 0xD3, 0xD1, 0x26, 0x16, 0x85, 0x37, 0x72, 0xE2, 0x00, 0x8C, 0x44, 0xCF, 0x10, 0xDA, 0x33,
	0x2D, 0x1A, 0xDE, 0x60, 0x86, 0x69, 0x23, 0x69, 0x2A, 0x7C, 0xCD, 0x4B, 0x51, 0x0D, 0x95, 0x54,
	0x39, 0x77, 0x2E, 0x29, 0xEA, 0x1B, 0xA6, 0x50, 0xA2, 0x6A, 0x8F, 0x6F, 0x50, 0x99, 0x5C, 0x3E, 
	0x54, 0xFB, 0xEF, 0x50, 0x5B, 0x0B, 0x07, 0x45, 0x17, 0x89, 0x6D, 0x28, 0x13, 0x77, 0x37, 0x1D,
	0xDB, 0x8E, 0x1E, 0x4A, 0x05, 0x66, 0x4A, 0x6F, 0x99, 0x20, 0xE5, 0x70, 0xE2, 0xB9, 0x71, 0x7E, 
	0x0C, 0x6D, 0x49, 0x04, 0x2D, 0x7A, 0xFE, 0x72, 0xC7, 0xF2, 0x59, 0x30, 0x8F, 0xBB, 0x02, 0x5D, 
	0x73, 0xE5, 0xC9, 0x20, 0xEA, 0x78, 0xEC, 0x20, 0x90, 0xF0, 0x8A, 0x7F, 0x42, 0x17, 0x7C, 0x47,
	0x19, 0x60, 0xB0, 0x16, 0xBD, 0x26, 0xB7, 0x71, 0xB6, 0xC7, 0x9F, 0x0E, 0xD1, 0x33, 0x82, 0x3D,
	0xD3, 0xAB, 0xEE, 0x63, 0x99, 0xC8, 0x2B, 0x53, 0xA0, 0x44, 0x5C, 0x71, 0x01, 0xC6, 0xCC, 0x44, 
	0x1F, 0x32, 0x4F, 0x3C, 0xCA, 0xC0, 0x29, 0x3D, 0x52, 0xD3, 0x61, 0x19, 0x58, 0xA9, 0x7D, 0x65, 
	0xB4, 0xDC, 0xCF, 0x0D, 0xF4, 0x3D, 0xF1, 0x08, 0xA9, 0x42, 0xDA, 0x23, 0x09, 0xD8, 0xBF, 0x5E, 
	0x50, 0x49, 0xF8, 0x4D, 0xC0, 0xCB, 0x47, 0x4C, 0x1C, 0x4F, 0xF7, 0x7B, 0x2B, 0xD8, 0x16, 0x18,
	0xC5, 0x31, 0x92, 0x3B, 0xB5, 0x6F, 0xDC, 0x6C, 0x0D, 0x92, 0x88, 0x16, 0xD1, 0x9E, 0xDB, 0x3F, 
	0xE2, 0xE9, 0xDA, 0x5F, 0xD4, 0x84, 0xE2, 0x46, 0x61, 0x5A, 0xDE, 0x1C, 0x55, 0xCF, 0xA4, 0x00, 
	0xBE, 0xFD, 0xCE, 0x67, 0xF1, 0x4A, 0x69, 0x1C, 0x97, 0xE6, 0x20, 0x48, 0xD8, 0x5D, 0x7F, 0x7E, 
	0xAE, 0x71, 0x20, 0x0E, 0x4E, 0xAE, 0xC0, 0x56, 0xA9, 0x91, 0x01, 0x3C, 0x82, 0x1D, 0x0F, 0x72,
	0xE7, 0x76, 0xEC, 0x29, 0x49, 0xD6, 0x5D, 0x2D, 0x83, 0xE3, 0xDB, 0x36, 0x06, 0xA9, 0x3B, 0x66,
	0x13, 0x97, 0x87, 0x6A, 0xD5, 0xB6, 0x3D, 0x50, 0x5E, 0x52, 0xB9, 0x4B, 0xC7, 0x73, 0x57, 0x78, 
	0xC9, 0xF4, 0x2E, 0x59, 0x07, 0x95, 0x93, 0x6F, 0xD0, 0x4B, 0x17, 0x57, 0x19, 0x3E, 0x27, 0x27, 
	0xC7, 0x60, 0xDB, 0x3B, 0xED, 0x9A, 0x0E, 0x53, 0x44, 0x16, 0x3E, 0x3F, 0x8D, 0x92, 0x6D, 0x77,
	0xA2, 0x0A, 0xEB, 0x3F, 0x52, 0xA8, 0xC6, 0x55, 0x5E, 0x31, 0x49, 0x37, 0x85, 0xF4, 0xC5, 0x1F, 
	0x26, 0x2D, 0xA9, 0x1C, 0xBF, 0x8B, 0x27, 0x54, 0xDA, 0xC3, 0x6A, 0x20, 0xE5, 0x2A, 0x78, 0x04,
	0xB0, 0xD6, 0x90, 0x70, 0x72, 0xAA, 0x8B, 0x68, 0xBD, 0x88, 0xF7, 0x02, 0x5F, 0x48, 0xB1, 0x7E, 
	0xC0, 0x58, 0x4C, 0x3F, 0x66, 0x1A, 0xF9, 0x3E, 0xE1, 0x65, 0xC0, 0x70, 0xA7, 0xCF, 0x38, 0x69,
	0xAF, 0xF0, 0x56, 0x6C, 0x64, 0x49, 0x9C, 0x27, 0xAD, 0x78, 0x74, 0x4F, 0xC2, 0x87, 0xDE, 0x56, 
	0x39, 0x00, 0xDA, 0x77, 0x0B, 0xCB, 0x2D, 0x1B, 0x89, 0xFB, 0x35, 0x4F, 0x02, 0xF5, 0x08, 0x51,
	0x13, 0x60, 0xC1, 0x0A, 0x5A, 0x47, 0x4D, 0x26, 0x1C, 0x33, 0x30, 0x78, 0xDA, 0xC0, 0x9C, 0x46,
	0x47, 0xE2, 0x5B, 0x79, 0x60, 0x49, 0x6E, 0x37, 0x67, 0x53, 0x0A, 0x3E, 0xE9, 0xEC, 0x46, 0x39,
	0xB2, 0xF1, 0x34, 0x0D, 0xC6, 0x84, 0x53, 0x75, 0x6E, 0xE1, 0x0C, 0x59, 0xD9, 0x1E, 0xDE, 0x29,
	0x85, 0x10, 0x7B, 0x49, 0x49, 0xA5, 0x77, 0x79, 0xBE, 0x49, 0x56, 0x2E, 0x36, 0xE7, 0x0B, 0x3A,
	0xBB, 0x4F, 0x03, 0x62, 0x7B, 0xD2, 0x4D, 0x31, 0x95, 0x2F, 0xBD, 0x38, 0x7B, 0xA8, 0x4F, 0x21,
	0xE1, 0xEC, 0x46, 0x70, 0x76, 0x95, 0x7D, 0x29, 0x22, 0x78, 0x88, 0x0A, 0x90, 0xDD, 0x9D, 0x5C,
	0xDA, 0xDE, 0x19, 0x51, 0xCF, 0xF0, 0xFC, 0x59, 0x52, 0x65, 0x7C, 0x33, 0x13, 0xDF, 0xF3, 0x48,
	0xDA, 0xBB, 0x2A, 0x75, 0xDB, 0x60, 0xB2, 0x02, 0x15, 0xD4, 0xFC, 0x19, 0xED, 0x1B, 0xEC, 0x7F,
	0x35, 0xA8, 0xFF, 0x28, 0x31, 0x07, 0x2D, 0x12, 0xC8, 0xDC, 0x88, 0x46, 0x7C, 0x8A, 0x5B, 0x22};
   
   RexxObjectPtr dataobj;
   RexxDirectoryObject data;
   data = (RexxDirectoryObject)Data;
   flag = context->IsDirectory(data);
   if ( flag == 0 )
   {
      sprintf(buffer,"Error: GE_DeCrypt parameter 1 is not a directory!");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;
   }  
   
   // Get the input filename
   CSTRING input_filename;
   dataobj = (RexxObjectPtr)context->DirectoryAt(data, "INPUT_FILENAME"); 
   input_filename = context->ObjectToStringValue(dataobj);
   
   // Get the offset  input the input file
   uint32_t offset;
   dataobj = (RexxObjectPtr)context->DirectoryAt(data, "OFFSET"); 
   flag = context->ObjectToUnsignedInt32(dataobj, &offset);
   
   // Get the size
   uint32_t size;   
   dataobj = (RexxObjectPtr)context->DirectoryAt(data, "SIZE"); 
   flag = context->ObjectToUnsignedInt32(dataobj, &size);
   
   sprintf(buffer,"DIAG GE_DeCrypt input_filename=%s size=%d offset=%d",input_filename, size, offset);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      
   BYTE* memblock;
   
   ifstream file (input_filename, ios::in|ios::binary|ios::ate);
   if (file.is_open())
   {
      // if the size passed is zero then get the entire file size
      if(size == 0) size = (long)file.tellg();
      
      sprintf(buffer,"DIAG GE_DeCrypt size=%d",size);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
    
      memblock = (BYTE*)malloc(size);
      file.seekg (offset, ios::beg);
      file.read ((char *)memblock, size);
      file.close();
   }
   else // Unable to open the input file
   {
      sprintf(buffer,"Error: GE_DeCrypt Unable to open Input file!");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;
   }
   
   // if this a valid GE image
   /*
   const CRYPTED_JPEG   = $A6EF9107; //$0791EFA6;
   const CRYPTED_DXT1   = $77B3CBB7; //$B7CBB377;
   const DECRYPTED_JPEG = $E0FFD8FF; //$FFD8FFE0;
   const DECRYPTED_DXT1 = $31A3824F; //$4F82A331;
   */
   DWORD	 imgtype;
   memcpy(&imgtype, memblock,4);
   CxImage* newImage = new CxImage();
   if (imgtype == 0xA6EF9107 || imgtype == 0x77B3CBB7 || imgtype == 0xE0FFD8FF || imgtype == 0x31A3824F)
   {
      // if encrypted JPEG or DXT then Decrypt it
      if (imgtype == 0xA6EF9107 || imgtype == 0x77B3CBB7)
      {
         // do the the GE DeCryption
         sprintf(buffer,"DIAG GE_DeCrypt Decrypting image");
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         int j=16;
         for (uint32_t i=0; i<(unsigned int)size; i++)
         {
            memblock[i] = memblock[i] ^ key[j+8];
            j++;
            if ((j % 8) == 0) j = j+16;
            if (j >= 1016) j = (j+8) % 24;
         }
         // if its a JPEG create Create an image from it 
         if (imgtype == 0xA6EF9107)
         {
            sprintf(buffer,"DIAG GE_DeCrypt Decoding JPEG image");
            rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
            
            // Create an image from the decrypted data
            rc = newImage->Decode(memblock,size,CXIMAGE_FORMAT_JPG);
            if(!rc)
            {
              sprintf(buffer,"ERROR: GE_DeCrypt invalid rc from Decode RC=%d\n ",rc);
              rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
              return 1;
            }
         }
      }
      // if its a DXT then Decode it
      BYTE *bmp_output, *dxt_decode;
      if (imgtype == 0x77B3CBB7 || imgtype == 0x31A3824F)
      {
         sprintf(buffer,"DIAG GE_DeCrypt Decoding DXT image size_int32=%d",sizeof (uint32_t));
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
         bmp_output = (BYTE*)malloc(262198);   // allocate the output bmp block of data
         memcpy(bmp_output,BMP_HEAD,54);   // put in the BMP header
         dxt_decode = bmp_output+54;  // put the decoded data past the header
         uint32_t x=0;
         uint32_t y=0;
         int i,j;
         memblock+=24; // move past header to first block to decompress
         for(i=0; i<64; i+=8)
         {
		       for(j=0; j<64; j+=8)
            {
               DecompressBlockBC1(x,y,8,memblock,(unsigned char*)dxt_decode);
               memblock+=8;  // move to next block to decompress
               x+=8;
            }
            y+=8;
         }
      }
      // Create an image from the decoded DXT data
      rc = newImage->Decode(bmp_output,262198,CXIMAGE_FORMAT_BMP);
      if(!rc)
      {
        sprintf(buffer,"ERROR: GE_DeCrypt invalid rc from DXT Decode RC=%d\n ",rc);
        rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
        return 1;
      }
   }
   else
   {
      sprintf(buffer,"Error: Unkown file type!");
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
      return 1;
   }
    
   
   sprintf(buffer,"DIAG GE_DeCrypt Reducing number of colors to 127");
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));  
   
	// Reduce the number of colors to 127
   ULONG outbpp; 
   RGBQUAD * ppal;
   INT numcolors, colors;
   colors = 127;   
	
   CQuantizer m(16777216,8);
   m.ProcessImage(newImage->GetDIB());	 
   numcolors = m.GetColorCount();
   if(diag > 0)   
   {
      sprintf(buffer,"DIAG GE_DeCrypt BEFORE: numcolors=%d",numcolors);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   } 
	
   newImage->SetXDPI(96);
   newImage->SetYDPI(96);
   // CQuantizer q(colors,(colors>16?7:8));   REMOVE
   CQuantizer q(colors,4);     //  was 8
   q.ProcessImage(newImage->GetDIB());           
   ppal=(RGBQUAD*)calloc(256*sizeof(RGBQUAD),1);
   q.SetColorTable(ppal);

   bool error_diffusion = true; //  (diffusion == 0)	? FALSE : TRUE;       
   rc = newImage->DecreaseBpp(8,error_diffusion,ppal,colors); //  DIFFUSION ADDED IN V9.5.2   
   outbpp = newImage->GetBpp(); 
   numcolors = q.GetColorCount();
   
   if(diag > 0)   
   {
      sprintf(buffer,"DIAG GE_DeCrypt DecreaseBpp RC=%d outbpp=%d AFTER: numcolors=%d",rc,outbpp,numcolors);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));     
   }     
     
   // Get pointer to the palette
   RexxArrayObject palette;
   palette = (RexxArrayObject)context->DirectoryAt(data, "PALETTE");  
      
   RGBQUAD color;
   ULONG pixval;
   long x, y;
   uint32_t idx;

   // Get the size of the palette  in bytes
   int palsize = newImage->GetPaletteSize()/4;   
   sprintf(buffer,"DIAG GE_DeCrypt palsize=%d",palsize);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   // Create a array to hold the histogram  of the images, each entry is 6 words long:  
   // color index,  color count, palette colors r,g,b, and pixval
	ULONG *pArray = (ULONG*) malloc(ENTSIZE*palsize);
	
	// Reset the array to zero  to do the image first
	memset(pArray,0,ENTSIZE*palsize); 
	   	
	// Create a histogram of the image 
   int i;
	// for(i=0; i<palsize; i++)
   for(i=0; i<127; i++)   // pass back 127 palette entries
	{
      color = newImage->GetPaletteColor(i);
      pixval = Get_Pixval(color);           // added in V17.0.0.19
      pArray[i*ENTLEN] = i; // Save the index for this color
      pArray[i*ENTLEN+2] = color.rgbRed; 
      pArray[i*ENTLEN+3] = color.rgbGreen; 
      pArray[i*ENTLEN+4] = color.rgbBlue; 
      pArray[i*ENTLEN+5] = pixval; // Save the pixval for this color
   }
      
   // Scan all the pixels and get the palette color counts 
   int b_width, b_height;
   b_width = newImage->GetWidth();
   b_height = newImage->GetHeight();
	for(y=0; y<b_height; y++)
   {
		for(x=0; x<b_width; x++)
      {
         // Get color index and increment the count for the index
			idx = newImage->GetPixelIndex(x,y);
         pArray[idx*ENTLEN+1]++;	// increment the histogram count for this color	
		}
   }        
   
   // pass back the information in the histogram
   // for(int i=0; i<palsize; i++)
   for(int i=0; i<127; i++)
   {
	   sprintf(buffer,"%d %d %d %d %d %d",pArray[i*ENTLEN],pArray[i*ENTLEN+1],pArray[i*ENTLEN+2],pArray[i*ENTLEN+3],pArray[i*ENTLEN+4],pArray[i*ENTLEN+5]);
      rc = context->ArrayAppendString(palette, buffer, strlen(buffer));
   }                                
   
   // Get the output filename
   CSTRING filename;
   dataobj = (RexxObjectPtr)context->DirectoryAt(data, "FILENAME"); 
   filename = context->ObjectToStringValue(dataobj);
   
   // See if output is a valid image type
	TCHAR* extout = FindExtension(filename);
	_tcslwr(extout);
	int typeout = CxImage::GetTypeIdFromName(extout);
	
	if ( diag>0 )
   {
      sprintf(buffer,"DIAG GE_DeCrypt filename=%s extout=%s typeout=%d",filename, extout,typeout);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }      
   
   // Save the image
   rc = newImage->Save(filename,typeout);	
   if(!rc)
   {
     sprintf(buffer,"ERROR: GE_DeCrypt Invalid rc from Save Image RC=%d\n ",rc);
     rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
     return 1;
   }	      
   
   // Pass back the type, width, height, X resolution and Y resolution  as diagostics
   int b_resx, b_resy;
   b_resx = newImage->GetXDPI(); // Get X resolution
   b_resy = newImage->GetYDPI(); // Get Y resolution
   sprintf(buffer,"DIAG GE_DeCrypt TYPEOUT=%d XRES=%d YRES=%d ",typeout, b_resx, b_resy);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));       
   sprintf(buffer,"DIAG GE_DeCrypt WIDTH=%d HEIGHT=%d NUMCOLORS=%d",b_width,b_height,numcolors);
   rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer)); 
   
   //  pass back the XRES and YRES in the data directory
   dataobj = context->Int32ToObject(b_resx); 
   context->DirectoryPut(data, dataobj, "XRES"); 
   dataobj = context->Int32ToObject(b_resy); 
   context->DirectoryPut(data, dataobj, "YRES");
   
   /* test div  function 
   div_t divresult;
   divresult = div (38,5);
   printf ("38 div 5 => %d, remainder %d.\n", divresult.quot, divresult.rem);
   */

   free(memblock);
   // free(bmp_output);
   free(pArray);
   
   return 0;  
}         
/**************************** Load_library **************************
 * Issues a LoadLibrary for a specific library
 *
 * Syntax:	rc = Load_library(library_name, diag, diagmsgs)
 *
 * Params:	library_name - the library to load
 *			   diag    - if gt 0 it will print diagnostic information  
 *			   diagmsgs - array to return diagnostic messages 
 *
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine3(int, Load_library, CSTRING, library_name, int, diag, RexxArrayObject, diagmsgs)
{
   logical_t success;
   success = context->LoadLibrary(library_name);
   return success;
}   
/**************************** WMCopydata **************************
 * Issues a WMCopydata to a window
 *
 * Syntax:	rc = WMCopydata (my_window_handle, window_name, textlen, text, diag, diagmsgs)
 *
 * Params:	my_window_handle - my window handle
 *			   window_name - name of window to send to
 *			   textlen -  length of text string to send 
 *			   text -  text string to send
 *			   diag    - if gt 0 it will print diagnostic information  
 *			   diagmsgs - array to return diagnostic messages 
 *
 * Return: 0 if OK else 1
 *********************************************************************/
RexxRoutine6(int, WMCopydata, uintptr_t, hwnd, CSTRING, window_name, int, textlen, CSTRING, text, int, diag, RexxArrayObject, diagmsgs)
{
   char buffer[512]; 
   int rc;
   int rtnrc = 0;
   HWND hwDispatch;
   if ( diag>0 ) 
   {
      sprintf(buffer,"DIAG WMCopydata Handle=%u Window_name=%s Textlen='%d Text=%s",hwnd, window_name, textlen, text);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }       
       
   char windowname[MAXFILENAME+16];       
   strncpy(windowname, _T(window_name),MAXFILENAME-1);
   /* strncpy doesnt append the trailing NULL, of the string is too long. */
   windowname[ MAXFILENAME ] = '\0'; 
   
   hwDispatch = FindWindow( NULL, windowname );  
   
   if( hwDispatch != NULL )
   {
      
      char txt[MAXFILENAME]; 
      strncpy(txt, _T(text),textlen);     
      txt[ textlen ] = '\0';      
      
      COPYDATASTRUCT cpd;
      // cpd.dwData = 0;
      cpd.dwData = Sat2Chart;          // function identifier
      cpd.cbData = textlen;  // size of txt
      cpd.lpData = &txt;           // txt   
      
      if ( diag>0 ) 
      {
         sprintf(buffer,"DIAG WMCopydata cbData=%d lpData=%s",cpd.cbData, (char *)cpd.lpData);
         rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
      }        
      
      SendMessage( hwDispatch,
                WM_COPYDATA,
                (WPARAM)(HWND) hwnd,
                (LPARAM) (LPVOID) &cpd );
   
   }   
   else 
      rtnrc = 1;
   
   if ( diag>0 ) 
   {
      sprintf(buffer,"DIAG WMCopydata Returning RC=%d",rtnrc);
      rc = context->ArrayAppendString(diagmsgs, buffer, strlen(buffer));
   }      
	
   return rtnrc;
}

RexxMethodEntry window_methods[] =
{
    REXX_METHOD(window_findWindow_cls,  window_findWindow_cls),
    REXX_METHOD(window_init,            window_init),
    REXX_METHOD(window_sendKeyPress,    window_sendKeyPress),
    REXX_METHOD(window_toForeground,    window_toForeground),
    REXX_LAST_METHOD()
};

// build the actual entry list
RexxRoutineEntry Sat2Chart_routines[] = {       
    REXX_TYPED_ROUTINE(Load_library, Load_library), 
    REXX_TYPED_ROUTINE(WMCopydata, WMCopydata), 
    REXX_TYPED_ROUTINE(ZipIt, ZipIt), 
    REXX_TYPED_ROUTINE(WidePath, WidePath),
    REXX_TYPED_ROUTINE(CalcCharts, CalcCharts),
    REXX_TYPED_ROUTINE(Calc_HLG_Tiles, Calc_HLG_Tiles),
    REXX_TYPED_ROUTINE(Add_HLG_Points, Add_HLG_Points),
    REXX_TYPED_ROUTINE(Calc_Bdr, Calc_Bdr),
    REXX_TYPED_ROUTINE(CalcSplit, CalcSplit),
    REXX_TYPED_ROUTINE(CalcWidePath, CalcWidePath),
    REXX_TYPED_ROUTINE(Quilt, Quilt),
    REXX_TYPED_ROUTINE(Quilt_Blobs, Quilt_Blobs),
    REXX_TYPED_ROUTINE(Quilt_Blobs_2, Quilt_Blobs_2),
    REXX_TYPED_ROUTINE(Quilt_NV_Blobs, Quilt_NV_Blobs),
    REXX_TYPED_ROUTINE(Tile_Histogram, Tile_Histogram),
    REXX_TYPED_ROUTINE(Tile_Histogram_Blobs, Tile_Histogram_Blobs),
    REXX_TYPED_ROUTINE(Create_PNG_Blobs, Create_PNG_Blobs),
    REXX_TYPED_ROUTINE(Get_Blob, Get_Blob),   
    REXX_TYPED_ROUTINE(Merge_Tiles, Merge_Tiles),   
    REXX_TYPED_ROUTINE(Merge_2_Blobs, Merge_2_Blobs), 
    REXX_TYPED_ROUTINE(Create_From_Array, Create_From_Array),
    REXX_TYPED_ROUTINE(TileXYToQuadKey, TileXYToQuadKey),  
    REXX_TYPED_ROUTINE(QuadKeyToTileXY, QuadKeyToTileXY),
    REXX_TYPED_ROUTINE(GE_DeCrypt, GE_DeCrypt),
    REXX_TYPED_ROUTINE(UnzipIt, UnzipIt), 
    REXX_TYPED_ROUTINE(CalcArcTan2, CalcArcTan2), 
    REXX_TYPED_ROUTINE(lat2tiley, lat2tiley),
    REXX_TYPED_ROUTINE(lon2tilex, lon2tilex),
    REXX_TYPED_ROUTINE(merc_y, merc_y),   
    REXX_TYPED_ROUTINE(merc_lat, merc_lat), 
    REXX_TYPED_ROUTINE(merc_x, merc_x),    
    REXX_TYPED_ROUTINE(merc_lon, merc_lon),
    REXX_TYPED_ROUTINE(GetTileX, GetTileX),    
    REXX_TYPED_ROUTINE(Varint_Encode, Varint_Encode),   
    REXX_TYPED_ROUTINE(Varint_Decode, Varint_Decode),       
    REXX_TYPED_ROUTINE(GetTileY, GetTileY),
    REXX_TYPED_ROUTINE(GetTileY_Spheroid, GetTileY_Spheroid),   
    REXX_TYPED_ROUTINE(cYToLat, cYToLat), 
    REXX_TYPED_ROUTINE(cYToLatSpheroid, cYToLatSpheroid),
    REXX_TYPED_ROUTINE(cXToLon, cXToLon),     
    REXX_TYPED_ROUTINE(degreesToPixelsY, degreesToPixelsY),    
    REXX_TYPED_ROUTINE(degreeEllipsoidToPixelsY, degreeEllipsoidToPixelsY),  
    REXX_TYPED_ROUTINE(getTileNumberX, getTileNumberX),  
    REXX_TYPED_ROUTINE(getTileNumberY, getTileNumberY),    
    REXX_TYPED_ROUTINE(getTileEllipsoidNumberY, getTileEllipsoidNumberY),    
    REXX_TYPED_ROUTINE(getLatitudeFromEllipsoidTileY, getLatitudeFromEllipsoidTileY), 
    REXX_TYPED_ROUTINE(getLongitudeFromTile, getLongitudeFromTile),   
    REXX_TYPED_ROUTINE(getLatitudeFromTile, getLatitudeFromTile),
    REXX_TYPED_ROUTINE(GetLatitude, GetLatitude),
    REXX_TYPED_ROUTINE(GetLongitude, GetLongitude),   
    REXX_TYPED_ROUTINE(GetLongitude_Spheroid, GetLongitude_Spheroid), 
    REXX_TYPED_ROUTINE(MercatorOnEllipsoid_LAt2Y, MercatorOnEllipsoid_LAt2Y),   
    REXX_TYPED_ROUTINE(MercatorOnEllipsoid_Y2Lat, MercatorOnEllipsoid_Y2Lat),
    REXX_TYPED_ROUTINE(MercatorOnShere_Lat2Y, MercatorOnShere_Lat2Y),  
    REXX_TYPED_ROUTINE(MercatorOnShere_Y2Lat, MercatorOnShere_Y2Lat),
    REXX_TYPED_ROUTINE(mercator_offset, mercator_offset),
    REXX_TYPED_ROUTINE(MemUsage,MemUsage),   
    REXX_TYPED_ROUTINE(GetBuf,GetBuf), 
    REXX_TYPED_ROUTINE(ReleaseArray,  ReleaseArray), 
    REXX_TYPED_ROUTINE(RunPGM,  RunPGM), 
    REXX_TYPED_ROUTINE(GetDatum, GetDatum),
    REXX_TYPED_ROUTINE(RotateImage, RotateImage),
    REXX_TYPED_ROUTINE(ListWindows, ListWindows),
    REXX_TYPED_ROUTINE(Calculate_GeoRefs, Calculate_GeoRefs),
    REXX_TYPED_ROUTINE(CropImage, CropImage),
    REXX_TYPED_ROUTINE(ReSizeImage, ReSizeImage),
    REXX_TYPED_ROUTINE(ExpandImage, ExpandImage),
    REXX_TYPED_ROUTINE(ConvertImage, ConvertImage),  
    REXX_TYPED_ROUTINE(FindGrid, FindGrid), 
    REXX_TYPED_ROUTINE(Drawtext, Drawtext), 
    REXX_TYPED_ROUTINE(Colorize_depths, Colorize_depths), 
    REXX_TYPED_ROUTINE(GetDepths, GetDepths),   
    REXX_TYPED_ROUTINE(MergeFGBG, MergeFGBG),    
    REXX_TYPED_ROUTINE(SetAlpha, SetAlpha),
    REXX_TYPED_ROUTINE(AlphaSet, AlphaSet),
    REXX_TYPED_ROUTINE(Tif2BSB, Tif2BSB),
    REXX_TYPED_ROUTINE(BSB2Tif, BSB2Tif),   
    REXX_TYPED_ROUTINE(DestroyWin, DestroyWin),
    REXX_TYPED_ROUTINE(ClipBoard, ClipBoard),
    REXX_TYPED_ROUTINE(Screenshot, Screenshot),
    REXX_TYPED_ROUTINE(Screenshot_Mono, Screenshot_Mono), 
    REXX_TYPED_ROUTINE(Get_GE_Spin_Wheel, Get_GE_Spin_Wheel),
    REXX_TYPED_ROUTINE(Screenshot_tif, Screenshot_tif),
    REXX_TYPED_ROUTINE(Check_View, Check_View),
    REXX_TYPED_ROUTINE(CountColors, CountColors),   
    REXX_TYPED_ROUTINE(CheckQueueCount0, CheckQueueCount0),      
    REXX_TYPED_ROUTINE(GetZoomLevel, GetZoomLevel), 
    REXX_TYPED_ROUTINE(Screen_Info, Screen_Info), 
    REXX_TYPED_ROUTINE(Taken_blocks, Taken_blocks),    
    REXX_LAST_ROUTINE()
};  


RexxPackageEntry window_package_entry =
{
    STANDARD_PACKAGE_HEADER
    REXX_INTERPRETER_4_0_0,              // needs at least the 4.0.0 interpreter
    "Sat2ChartFUNCS",                    // name of the package       
    "2.0.0",                             // package information
    NULL,                                // no load/unload functions
    NULL,
    Sat2Chart_routines,                  // the exported functions
    window_methods                       // the exported methods
};

// package loading stub.  
OOREXX_GET_PACKAGE(window);
